# SPDX-License-Identifier: MIT
"""DOI organization data."""

from collections.abc import Sequence

from fnattr.vlju.types.doi import DOI

ORGANIZATION = {
    10: {
        1001: "American Medical Association",
        1002: "Wiley (John Wiley & Sons)",
        1006: "Elsevier - Academic Press",
        1007: "Springer-Verlag",
        1013: "Springer Science and Business Media LLC",
        1016: "Elsevier",
        1017: "Cambridge University Press",
        1021: "American Chemical Society",
        1023: "Springer (Kluwer Academic Publishers)",
        1024: "Hogrefe Publishing Group",
        1026: "Hogrefe Publishing Group",
        1027: "Hogrefe Publishing Group",
        1029: "American Geophysical Union",
        1033: "EBSCO Publishing",
        1034: "Wiley (Munksgaard)",
        1036: "McGraw-Hill Professional",
        1037: "American Psychological Association",
        1038: "Springer Science and Business Media LLC",
        1039: "The Royal Society of Chemistry",
        1042: "Portland Press",
        1043: "Allen Press",
        1044: "American Speech-Language-Hearing Association",
        1045: "Corporation for National Research Initiatives",
        1046: "Wiley (Blackwell Publishing)",
        1049: "Institution of Electrical Engineers",
        1051: "EDP Sciences",
        1053: "Elsevier - WB Saunders",
        1054: "Elsevier- Churchill Livingstone",
        1055: "Georg Thieme Verlag KG",
        1056: "New England Journal of Medicine",
        1057: "Springer Science and Business Media LLC",
        1061: "American Society of Civil Engineers",
        1063: "American Institute of Physics",
        1065: "Springer - Ecomed Publishers",
        1066: "Blackhorse Publishing",
        1067: "Elsevier - Mosby",
        1068: "Pion Ltd.",
        1070: "Turpion",
        1071: "CSIRO Publishing",
        1073: "Proceedings of the National Academy of Sciences",
        1074: "American Society for Biochemistry and Molecular Biology",
        1075: "John Benjamins Publishing Company",
        1076: "Informa UK (Swets & Zeitlinger Publishers)",
        1078: "Elsevier - Urban & Fischer Verlag",
        1079: "CABI Publishing",
        1080: "Informa UK (Taylor & Francis)",
        1081: "Informa UK (Marcel Dekker)",
        1083: "The Rockefeller University Press",
        1084: "The Rockefeller University Press",
        1085: "The Rockefeller University Press",
        1086: "The University of Chicago Press",
        1087: "ALPSP -Association of Learned and Professional"
              "Society Publishers",
        1088: "IOP Publishing",
        1089: "Mary Ann Liebert",
        1090: "American Mathematical Society",
        1091: "American Society for Cell Biology",
        1092: "Metapress",
        1093: "Oxford University Press",
        1094: "Scientific Societies",
        1095: "Society for the Study of Reproduction",
        1096: "Federation of American Society for Experimental Biology",
        1097: "Ovid Technologies (Wolters Kluwer)- Lippincott"
              "Williams & Wilkins",
        1098: "The Royal Society",
        1099: "Microbiology Society",
        1100: "Hindawi (The Scientific World)",
        1101: "Cold Spring Harbor Laboratory",
        1102: "E-MED LTD.",
        1103: "American Physical Society",
        1104: "American Society of Plant Biologists",
        1105: "American Society of Plant Biologists",
        1106: "Sage Publications  - Technomic Publishing Company",
        1107: "International Union of Crystallography",
        1108: "Emerald (MCB UP)",
        1109: "Institute of Electrical and Electronics Engineers",
        1110: "Cold Spring Harbor Laboratory",
        1111: "Wiley (Blackwell Publishing)",
        1112: "London Mathematical Society",
        1113: "Wiley (The Physiological Society)",
        1114: "Springer (Kluwer Academic Publishers- Biomedical"
              "Engineering Society (BMES))",
        1115: "ASME International",
        1116: "American Vacuum Society",
        1117: "SPIE - International Society for Optical Engineering",
        1118: "American Association of Physicists in Medicine",
        1119: "American Association of Physics Teachers",
        1120: "American College of Medical Physics",
        1121: "Acoustical Society of America",
        1122: "Society of Rheology",
        1123: "Human Kinetics",
        1124: "American Society for Pharmacology & Experimental Therapeutics",
        1126: "American Association for the Advancement of Science (AAAS)",
        1127: "Schweizerbart",
        1128: "American Society for Microbiology",
        1130: "Geological Society of America",
        1134: "Pleiades Publishing",
        1135: "Institute of Organic Chemistry & Biochemistry,Academy"
              "of Sciences of the Czech Republic",
        1136: "BMJ",
        1137: "Society for Industrial and Applied Mathematics",
        1138: "International Bone and Mineral Society",
        1139: "Canadian Science Publishing",
        1140: "Springer-Verlag",
        1142: "World Scientific",
        1143: "Japan Society of Applied Physics",
        1144: "Geological Society of London",
        1145: "Association for Computing Machinery",
        1146: "Annual Reviews",
        1147: "IBM",
        1148: "Radiological Society of North America",
        1149: "The Electrochemical Society",
        1151: "NIDA/RTI International",
        1152: "American Physiological Society",
        1154: "International Centre for Diffraction Data",
        1155: "Hindawi Limited",
        1157: "Elsevier - Ediciones Doyma",
        1158: "American Association for Cancer Research",
        1159: "S. Karger AG",
        1160: "Thieme",
        1161: "Ovid Technologies Wolters Kluwer -American Heart Association",
        1162: "MIT Press",
        1163: "Brill",
        1164: "American Thoracic Society",
        1165: "American Thoracic Society",
        1166: "American Scientific Publishers",
        1167: "Association for Research in Vision and Ophthalmology (ARVO)",
        1172: "American Society for Clinical Investigation",
        1174: "Fundacion Infancia y Aprendizaje",
        1175: "American Meteorological Society",
        1176: "American Psychiatric Association Publishing",
        1177: "SAGE Publications",
        1179: "Maney Publishing",
        1180: "The Mineralogical Society",
        1182: "American Society of Hematology",
        1183: "European Respiratory Society",
        1185: "Informa UK (Librapharm)",
        1186: "Springer (Biomed Central Ltd.)",
        1187: "American Society for Cell Biology",
        1188: "Oncology Nursing Society",
        1189: "Society for Leukocyte Biology",
        1190: "Society of Exploration Geophysicists",
        1191: "Arnold Publishers",
        1192: "Royal College of Psychiatrists",
        1193: "Earthquake Engineering Research Institute",
        1194: "American Society for Biochemistry and Molecular"
              "Biology,Journal of Lipid Research",
        1196: "Wiley (New York Academy of Sciences)",
        1197: "Elsevier- Hanley and Belfus Inc.",
        1198: "Informa UK (American Statistical Association)",
        1199: "BioOne- American Society of Plant Biologists (Arabidopsis)",
        1200: "American Society of Clinical Oncology",
        1201: "CRC Press",
        1202: "American Industrial Hygiene Association",
        1203: "Ovid Technologies (Wolters Kluwer)- International"
              "Pediatric Research Foundation",
        1205: "Elsevier - Institution of Chemical Engineers",
        1206: "BioOne (American Museum of Natural History)",
        1207: "Informa UK (Lawrence Erlbaum Associates, Inc.)",
        1208: "American Association of Pharmaceutical Scientists",
        1209: "IOP Publishing - Europhysics Letters",
        1210: "The Endocrine Society",
        1211: "Royal Pharmaceutical Society",
        1212: "Ovid Technologies (Wolters Kluwer) - American"
              "Academy of Neurology",
        1213: "Ovid Technologies (Wolters Kluwer)  - Anesthesia & Analgesia",
        1214: "Institute of Mathematical Statistics",
        1215: "Duke University Press",
        1216: "Rocky Mountain Mathematics Consortium",
        1227: "Ovid Technologies (Wolters Kluwer)  - Neurosurgery",
        1238: "Royal Swedish Academy of Sciences",
        1239: "Applied Probability Trust",
        1240: "Elsevier- Spektrum Akademischer Verlag",
        1241: "Japan Science and Technology Agency",
        1242: "The Company of Biologists",
        1243: "Research Publishing Services - Professional"
              "Engineering Publishing",
        1245: "Springer - Society of Surgical Oncology",
        1246: "The Chemical Society of Japan",
        1247: "Japan Society for Cell Biology",
        1248: "Pharmaceutical Society of Japan",
        1249: "Ovid Technologies (Wolters-Kluwer)- American"
              "College of Sports Medicine",
        1250: "Acoustical Society of Japan",
        1251: "Springer (Biological Procedures Online)",
        1252: "The Society of Chemical Engineers Japan",
        1253: "The Japanese Circulation Society",
        1254: "The Japanese Pharmacological Society",
        1255: "IM Publications Open LLP",
        1256: "Wiley (Royal Meteorological Society)",
        1257: "American Economic Association",
        1258: "The Royal Society of Medicine",
        1259: "British Institute of Radiology",
        1260: "Multiscience",
        1261: "Cold Spring Harbor Laboratory",
        1262: "Japanese Society of Animal Reproduction",
        1263: "The Society for Biotechnology, Japan",
        1264: "Japanese Society of Microbial Ecology",
        1265: "Japanese Society for Hygiene",
        1266: "The Genetics Society of Japan",
        1267: "Japan Society of Histochemistry and Cytochemistry",
        1269: "The Japan Radiation Research Society",
        1270: "Japanese Society of Breeding",
        1271: "Japan Society for Bioscience, Biotechnology,"
              "and Agrochemistry",
        1272: "The Medical Association of Nippon Medical School",
        1273: "Chem-Bio Informatics Society",
        1274: "BioOne (Japanese Society of Mammalian Ova Research)",
        1287: "Institute for Operations Research and the Management Sciences",
        1288: "Ovid Technologies (Wolters Kluwer)  - Triological Society",
        1289: "Environmental Health Perspectives",
        1290: "BioOne (Society for In Vitro Biology)",
        1291: "Japanese Society of Hypertension",
        1292: "Japanese Society of Veterinary Science",
        1293: "The Japanese Society of Toxicologic Pathology",
        1294: "Japan Society of Equine Science",
        1295: "Society of Polymer Science, Japan",
        1297: "The Japanese Society for Pediatric Endocrinology",
        1298: "Japanese Physical Therapy Association",
        1299: "The Japan Society of Mechanical Engineers",
        1300: "Informa UK (Haworth Press, Inc.,)",
        1301: "Wiley (International Life Sciences Institute)",
        1302: "British Editorial Society of Bone and Joint Surgery",
        1303: "Japanese Society of Applied Entomology & Zoology",
        1304: "Encyclopedia of Life",
        1305: "Notre Dame Journal of Formal Logic",
        1306: "AAPG/Datapages",
        1307: "Michigan Mathematical Journal",
        1308: "The Royal College of Surgeons of England",
        1309: "American Society for Clinical Pathology",
        1310: "Thomas Land Publishers, Inc.",
        1331: "American Pharmacists Association",
        1332: "Bristol University Press",
        1344: "Edicions de la Universitat de Barcelona",
        1345: "Harvey Whitney Books Co.",
        1346: "Clay Minerals Society",
        1347: "Uitgeverij Boom",
        1348: "Wiley (British Psychological Society)",
        1349: "Dartmouth College Library Press",
        1350: "Vathek Publishing",
        1351: "International Union of Pure and Applied Chemistry",
        1352: "American Association on Intellectualand Developmental"
              "Disabilities",
        1353: "Project MUSE",
        1354: "American College of Veterinary Pathologists",
        1355: "ISEAS - Yusof Ishak Institute",
        1357: "Journal of Marine Research/Yale",
        1358: "Thomson Reuters (Prous Science)",
        1359: "Wiley(American Society Bone & Mineral Research)",
        1360: "Science in China Press",
        1361: "Springer - ASM International",
        1362: "Westburn Publishers",
        1363: "Guttmacher Institute",
        1364: "The Optical Society",
        1365: "Springer Fachmedien Wiesbaden GmbH",
        1366: "Society for Applied Spectroscopy",
        1367: "Elsevier - Ambulatory Pediatric Associates",
        1368: "BioOne (Society of Protozoologists)",
        1369: "Journal of Histochemistry & Cytochemistry",
        1370: "Annals of Family Medicine",
        1371: "Public Library of Science",
        1372: "Muse - College Literature",
        1373: "American Association for Clinical Chemistry",
        1374: "Muse - Gallaudet University Press",
        1375: "Cambridge University Press (Australian Academic Press)",
        1376: "Muse - Penn State Press",
        1377: "Project Hope",
        1378: "American College of Chest Physicians",
        1379: "Springer - Cell Stress Society International",
        1380: "The Surface Science Society of Japan",
        1381: "Springer - FD Communications",
        1382: "Fibre Diffraction Review",
        1383: "Elsevier - Medicine Publishing Company",
        1384: "The Surface Analysis Society of Japan",
        1385: "Springer - Humana Press",
        1386: "Intellect",
        1387: "UPV/EHU Press",
        1390: "mEDRA",
        1399: "Verlag Eugen Ulmer",
        1472: "mEDRA",
        1484: "Brepols Publishers, NV",
        1489: "The International Leprosy Association",
        1490: "Muse - Indiana University Press",
        1491: "Muse - George Washington University Institute"
              "for Ethnographic Research",
        1492: "Muse - University of St.Thomas - Logos",
        1493: "Muse - University of North Carolina Press",
        1494: "Muse - University of Minnesota Press",
        1495: "Muse - Journal of Social History",
        1496: "Muse - The Americas",
        1497: "Muse - Music Library Association",
        1498: "Muse - University of Hawaii Press",
        1499: "Muse - Kent State University Press",
        1500: "Muse - University of Nebraska Press",
        1501: "Ankara University",
        1502: "Muse- Wayne State University Press",
        1503: "Joule Inc.",
        1504: "Inderscience Enterprises Ltd.",
        1505: "Commonwealth Forestry Association",
        1506: "Wiley (Canadian Academic Accounting Association)",
        1507: "The Japan Endocrine Society",
        1508: "International Society of Cytology",
        1509: "American Marketing Association",
        1510: "European Association of Cardiothoracic Surgery",
        1511: "Sigma Xi",
        1512: "Indiana University Mathematics Journal",
        1513: "American Thoracic Society",
        1515: "Walter de Gruyter GmbH",
        1516: "Wiley (International Journal of Psychoanalysis)",
        1517: "Informa UK (Ashley Publications)",
        1518: "Human Factors & Ergonomics",
        1519: "Ovid Technologies (Wolters Kluwer) - National"
              "Strength & Conditioning",
        1520: "ASTM International",
        1521: "Guilford Publications",
        1522: "Universite du Quebec a Chicoutimi",
        1523: "Society for Neuroscience",
        1524: "Oldenbourg Wissenschaftsverlag",
        1525: "University of California Press",
        1526: "Wiley (Rural Sociological Society)",
        1527: "The Japanese Society for Artificial Intelligence",
        1528: "Sage Publications- Academy of Traumatology",
        1529: "Elsevier - Biophysical Society",
        1530: "Bioscientifica",
        1531: "Informa UK (Taylor & Francis Books)",
        1532: "Carden Jennings Publishing Co.",
        1533: "Woodhead Publishing",
        1534: "The Genetics Society of America",
        1535: "Intel Corp",
        1536: "International Heart Journal Association",
        1537: "Anthropological Society of Nippon",
        1538: "Japanese Association for Laboratory Animal Science",
        1539: "Japan Society for Occupational Health",
        1540: "Japan Society of Smooth Muscle Research",
        1541: "Institute Electrical Engineers Japan",
        1542: "American Academy of Pediatrics",
        1553: "Verlag der Osterreichischen Akademie der Wissenschaften",
        1554: "BioOne (Society for the Study of Evolution)",
        1555: "Mars Informatics",
        1556: "Akademiai Kiado Zrt.",
        1557: "Materials Research Society",
        1558: "Equinox Publishing",
        1559: "Cartography & Geographic Information Society",
        1560: "Laser Pages Publishing Ltd",
        1561: "Now Publishers",
        1562: "BioOne (American Society for Photobiology)",
        1563: "American Academy of Implant Dentistry",
        1564: "Research Information Ltd",
        1571: "Patricia Seybold Group",
        1575: "MBLWHOI Library",
        1576: "Wiley Blackwell"
              " (Royal College of Obstetricians & Gynecologist)",
        1577: "American Fisheries Society",
        1578: "Aquatic Mammals Journal",
        1579: "BioOne (Royal Swedish Academy of Sciences)",
        1580: "Elsevier - Wilderness Medical Society",
        1581: "Wiley (Robotic Publications)",
        1582: "Illuminating Engineering Society of North America",
        1583: "International Society of Endovascular Specialists",
        1584: "Pesticide Science Society of Japan",
        1585: "Japan Society of Plasma Science and Nuclear Fusion Research",
        1586: "Informa Healthcare (Expert Reviews, LTD)",
        1587: "Institute of Electronics, Information and Communications"
              "Engineers",
        1588: "Pearblossom Private School",
        1589: "The Society of Physical Therapy Science",
        1590: "SciELO",
        1591: "Competence Network Pediatric Oncology & Hematology",
        1592: "Pharmacotherapy",
        1593: "Neoplasia Press",
        1595: "Johnson Matthey",
        1596: "The World Bank",
        1597: "American Cleft Palate Association",
        1598: "International Reading Association",
        1599: "Milbank Memorial Fund",
        1600: "American Society of Plant Taxonomists",
        1601: "NamesForLife, LLC",
        1602: "Elsevier - American Society for Experimental"
              "Neurotherapeutics",
        1603: "BioOne (Entomological Society of America)",
        1606: "Alliance for Children and Families",
        1608: "BioOne (Freshwater Biological Association)",
        1609: "Association for the Advancement of Artificial"
              "Intelligence (AAAI)",
        1613: "AI Access Foundation",
        1614: "Weed Science Society of America",
        1615: "Begell House Inc.",
        1616: "IFIS Publishing",
        1617: "Springer - RILEM Publishing",
        1618: "Japanese Society of Human-Environment",
        1619: "Mathematical Society of Japan",
        1620: "Tohoku University Medical Press",
        1621: "Nuclear Receptor Signaling Atlas",
        1622: "SagePublications - National Association of School Nurses",
        1623: "Informa UK (IAHS Ltd.)",
        1624: "Lamaze International",
        1625: "Japanese Society for Complementary & Alternative Medicine",
        1626: "Crop Science Society of Japan",
        1627: "Japan Petroleum Institute",
        1628: "Mohr Siebeck",
        1629: "UKSG",
        1630: "Sage Publications - International Institute"
              "for Environment and Development",
        1631: "Zhejiang University Press",
        1632: "Modern Language Association",
        1633: "KISTI",
        1634: "Alphamed Press",
        1635: "BioOne (Academy of Natural Sciences of Philadelphia)",
        1636: "BioOne (American Arachnological Society)",
        1637: "BioOne (American Association of Avian Pathologists)",
        1638: "BioOne (American Association of Zoo Veterinarians)",
        1639: "BioOne (American Bryological and Lichenological Society)",
        1640: "BioOne (American Fern Society)",
        1641: "BioOne (American Institute of Biological Sciences)",
        1642: "BioOne (American Ornithologists Union)",
        1643: "BioOne (American Society of Icthyologists and Herpetologists)",
        1644: "BioOne (American Society of Mammalogists)",
        1645: "BioOne (American Society of Parasitologists)",
        1646: "BioOne (Association for Tropical Biology)",
        1647: "BioOne (Association of Avian Veterinarians)",
        1648: "BioOne (Association of Field Ornithologists)",
        1649: "BioOne (Coleopterists Society)",
        1650: "BioOne (Cooper Ornithological Society)",
        1651: "BioOne (Crustacean Society)",
        1652: "BioOne (EUCC - The Coastal Union)",
        1653: "BioOne (Florida Entomological Society)",
        1654: "BioOne (Helminthological Society of Washington)",
        1655: "BioOne (Herpetologists League)",
        1656: "BioOne (Humboldt Field Research Institute)",
        1657: "BioOne (Institute of Arctic and Alpine Research)",
        1658: "BioOne (International Association of Vegetation Science)",
        1659: "BioOne (International Mountain Society and"
              "United Nations University)",
        1660: "BioOne (Kansas Academy of Science)",
        1661: "BioOne (Micropaleontology Press)",
        1662: "BioOne (National Association of Biology Teachers)",
        1663: "BioOne (New York Botanical Garden)",
        1664: "BioOne (New York Entomological Society)",
        1665: "BioOne (Orthopterist's Society)",
        1666: "BioOne (Paleontological Society)",
        1667: "BioOne (Radiation Research Society)",
        1668: "BioOne (Society for Integrative and Comparative Biology)",
        1669: "BioOne (Society for Sedimentary Geology)",
        1670: "BioOne (Society for the Study of Amphibians & Reptiles)",
        1671: "BioOne (Society of Vertebrate Paleontology)",
        1672: "BioOne (Society of Wetland Scientists)",
        1673: "BioOne (University of Arizona Library)",
        1674: "BioOne (University of Notre Dame)",
        1675: "BioOne (Waterbird Society)",
        1676: "BioOne (Wilson Ornithological Society)",
        1677: "BioScientifica (Society for Endocrinology)",
        1678: "Society of Rheology Japan",
        1679: "International Society Histology & Cytology",
        1680: "Thomas Telford Ltd.",
        1681: "American Society of Nephrology",
        1682: "Journal of Rehabilitation Research & Development",
        1684: "John Libbey Eurotext",
        1685: "SIMAI",
        1783: "Faculty of Family Planning",
        1784: "British Institute of Non-Destructive Testing",
        1785: "Seismological Society",
        1786: "Organisation for Economic Co-Operation and Development",
        1787: "Organisation for Economic Co-Operation and Development",
        1887: "Informa UK(IOP Books)",
        1888: "Informa UK (Encyclopedia of Astronomy and Astrophysics)",
        1889: "Society for Information Display",
        1890: "Ecological Society of America",
        1891: "Springer Publishing Company",
        1892: "Wiley (American College of Veterinary Internal Medicine)",
        1893: "BioOne (Beta Beta Biological Society)",
        1894: "BioOne (Southwestern Association of Naturalists)",
        1895: "WormBook",
        1896: "BioOne (Conservation International)",
        1897: "Wiley (Society of Environmental Toxicology and Chemistry)",
        1898: "BioOne (Northwestern Naturalist)",
        1899: "BioOne (Society for Freshwater Science)",
        1900: "JCFCorp SG PTE LTD",
        1901: "Society for the Experimental Analysis of Behavior",
        1902: "American Academy of Periodontology",
        1920: "Institute for Fiscal Studies",
        1921: "Whiting & Birch Ltd",
        1923: "Collegium Basilea",
        1956: "Journal of Global Economy",
        2104: "Monash University",
        2105: "American Public Health Association",
        2106: "Journal of Bone and Joint Surgery",
        2107: "Confederacion Academica Nipona, Espanola y Latinoamericana",
        2108: "BioOne (Zoological Society of Japan)",
        2109: "Ceramic Society of Japan",
        2110: "Society for Sedimentary Geology",
        2111: "Elsevier (Society for Range Management)",
        2112: "BioOne (Coastal Education and Research Foundation)",
        2113: "GeoScienceWorld",
        2114: "Japan Society of Physiological Anthropology",
        2115: "Society of Fiber Science & Technology Japan",
        2116: "Japan Society for Analytical Chemistry",
        2117: "Psychologia Society",
        2118: "Society of Petroleum Engineers",
        2119: "The Feinstein Institute for Medical Research"
              "(North Shore LIJ Research Institute)",
        2130: "Japanese Group Dynamics Association",
        2131: "Japanese Society of Toxicology",
        2132: "Japan Society of Personality Psychology",
        2133: "Japanese Society for the Study of Xenobiotics",
        2134: "American Society of Agronomy",
        2135: "Crop Science Society of America",
        2136: "Soil Science Society of America",
        2137: "MTT Agrifood Research Finland",
        2138: "Mineralogical Society of America",
        2139: "Social Science Electronic Publishing",
        2140: "Mathematical Sciences Publishers",
        2141: "Japan Poultry Science Association",
        2142: "Biophysical Society of Japan",
        2143: "Peeters Publishers",
        2144: "Future Science Ltd",
        2146: "American Society of Health System Pharmacists",
        2147: "Dove Medical Press",
        2148: "Alexandrine Press",
        2149: "Japanese Society of Tropical Medicine",
        2150: "Illuminating Engineering Society of Japan",
        2151: "Meteorological Society of Japan",
        2152: "University of Tokushima Faculty of Medicine",
        2153: "Historic Brass Society",
        2164: "American Society of Andrology",
        2165: "Springer - Adis",
        2166: "IWA Publishing",
        2167: "Informa UK (Multilingual Matters)",
        2168: "Logical Methods in Computer Science",
        2169: "Japanese Society of Internal Medicine",
        2170: "The Physiological Society of Japan",
        2171: "Japan Association on Odor Environment",
        2172: "US Department of Energy",
        2173: "Cornell Lab of Ornithology",
        2174: "Bentham Science",
        2175: "Water Environment Federation",
        2176: "Japan Neurosurgical Society",
        2177: "Japan Society for Clinical Immunology",
        2178: "Association for Symbolic Logic",
        2179: "BioOne (Castanea)",
        2180: "BioOne (Palynology)",
        2181: "BioOne (JANAS)",
        2182: "Canadian Association of Occupational Therapists",
        2183: "The Japan Academy",
        2184: "The Laser Society of Japan",
        2185: "The Japanese Association of Rural Medicine",
        2186: "Japan Prosthodontic Society",
        2187: "Japanese Society for Biological Sciences in Space",
        2188: "Japan Epidemiological Association",
        2189: "Johnson Graduate School of Management, Cornell University",
        2190: "Baywood Publishing",
        2191: "University of California Press",
        2192: "BioOne (Ursus)",
        2193: "BioOne (Wildlife Society Bulletin)",
        2194: "BioOne (Journal of Wildlife Management)",
        2195: "Fraunhofer-Institut fur Materialfluss und Logistik",
        2196: "JMIR Publications Inc.",
        2197: "Information Processing Society of Japan",
        2198: "Japanese Electrophoresis Society",
        2199: "Japan Society for Clinical Anesthesia",
        2200: "Morgan & Claypool",
        2201: "National Institute of Informatics",
        2202: "Berkeley Electronic Press",
        2203: "International Dose-Response Society",
        2204: "International Ocean Discovery Program - Management"
              "International",
        2205: "Geophysical Center of the Russian Academy of Sciences",
        2206: "Faculty of Mathematics, Kyushu University",
        2207: "Japan Welding Society",
        2208: "Japan Society of Civil Engineers",
        2209: "Tokyo Dental College",
        2210: "Worldwide Protein Data Bank",
        2211: "Atypon",
        2212: "Stewart Postharvest Solutions",
        2214: "American Roentgen Ray Society",
        2215: "American Society of Nephrology",
        2216: "International Phycological Society",
        2217: "Future Medicine",
        2218: "Edinburgh University Library",
        2219: "Railway Technical Research Institute",
        2220: "Biomedical Research Press",
        2221: "Cryogenic Association of Japan",
        2222: "The Japanese Association for Virology",
        2223: "Jornal de Pediatra/FSBP",
        2224: "Scientific Journal Publishers",
        2225: "Pontificia Universidad Catolica de Valparaiso",
        2298: "National Library of Serbia",
        2299: "Research Institute of Signal Processing",
        2300: "The Acarological Society of Japan",
        2301: "The Japan Academy of Neurosonology",
        2302: "The Keio Journal of Medicine",
        2303: "Japanese Society of Echocardiography",
        2304: "Symposium Journals",
        2305: "IUCN",
        2306: "Science Society of Thailand",
        2307: "JSTOR",
        2308: "American Accounting Association",
        2309: "Association for Vascular Access",
        2310: "Decker Medicine",
        2316: "ACTA Press",
        2317: "Kansas Entomological Society",
        2319: "The Angle Orthodontist",
        2320: "The Japan Institute of Metals",
        2321: "Intellectual Property Association of Japan",
        2322: "Japan Society for Aeronautical and Space Sciences",
        2323: "The Microbiology Research Foundation",
        2324: "The Society of Rubber Industry",
        2325: "Japanese Breast Cancer Society",
        2326: "BioOne (Ornithological Society of Japan)",
        2327: "Japanese Society of Veterinary Anesthesia and Surgery",
        2328: "Japan Society for Natural Disaster Science",
        2329: "The Japanese Society of Periodontology",
        2330: "Japanese Association for Oral Biology",
        2331: "The Japanese Society of Fisheries Science",
        2332: "Japanese Society of Allergology",
        2333: "The Behaviormetric Society of Japan",
        2334: "Nihon University School of Dentistry",
        2335: "The Japanese Society on Surgery for Cerebral Stroke",
        2336: "Western Japan Division of JDA",
        2337: "American Diabetes Association",
        2339: "Technical Education Faculty Gazi University",
        2340: "Acta Dermato-Venereologica",
        2341: "Operative Dentistry",
        2342: "Yamaguchi University Medical Association",
        2343: "The Geochemical Society of Japan",
        2344: "American Dental Society of Anethesiology",
        2345: "Association for the Advancement of Medical Instrumentation",
        2346: "The Tire Society",
        2349: "Department of Biomedical Imaging, University of Malaya",
        2350: "Society for Pediatric Pathology",
        2351: "Laser Institute of America",
        2352: "Society for Imaging Science & Technology",
        2353: "Elsevier - American Society for Investigative Pathology",
        2354: "Primate Society of Japan",
        2355: "The Iron and Steel Institute of Japan",
        2378: "Ernst Reinhardt, GmbH & Co. KG",
        2399: "Deomed Publishing",
        2422: "Scuola Normale Superiore - Edizioni della Normale",
        2457: "The Japan Section of the Regional Science Association"
              "International",
        2458: "University of Arizona",
        2459: "Ovid Technologies (Wolters Kluwer) - Italian"
              "Federation of Cardiology",
        2460: "American Veterinary Medical Association",
        2461: "SPVS - Sociedade Portuguesa de Vida Selvagem",
        2462: "EPP Publications",
        2463: "Japanese Society for Magnetic Resonance in Medicine",
        2464: "Japan Institute of Light Metals",
        2465: "Japan Association of Mineralogical Sciences",
        2466: "Ammons Scientific",
        2467: "Meteorological Research Institute, Japan Meteorological"
              "Agency",
        2468: "Japan Broncho-esophagological Society",
        2469: "CFA Institute",
        2470: "Research Foundation of CFA Institute",
        2471: "WHO Press",
        2472: "The Society of Materials Science Japan",
        2473: "The Mining and Materials Processing Institute of Japan",
        2474: "Japanese Society of Tribologists",
        2475: "American Journal of Science",
        2476: "Arachnological Society of Japan",
        2477: "Society of Computer Chemistry, Japan",
        2478: "De Gruyter Open Sp. z o.o.",
        2479: "Japanese Society of Sciences in Swimming and Water Exercise",
        2480: "The Society of Agricultural Meteorology of Japan",
        2481: "CODATA",
        2482: "The Japan Lung Cancer Society",
        2483: "University of California Press",
        2484: "Radiology Case Reports U of Washington",
        2485: "The Society for Hard Tissue Regenerative Biology",
        2486: "National Institute of Industrial Health",
        2487: "Muse - English Literature in Transition",
        2488: "The Japan Wood Research Society",
        2489: "Soil and Water Conservation Society",
        2490: "The Japanese Association of Rehabilitation Medicine",
        2491: "The Japanese Society on Thrombosis and Hemostasis",
        2492: "The Japanese Society of Inflammation and Regeneration",
        2493: "The Japan Society for Precision Engineering",
        2494: "The Technical Association of Photopolymers, Japan",
        2495: "WITPress",
        2496: "Japan Society for Higher Brain Dysfunction",
        2497: "Japan Society of Powder and Powder Metallurgy",
        2498: "University of Zagreb",
        2499: "International Food Policy Research Institute",
        2500: "Oceanside Publications",
        2501: "World Advertising Research Center",
        2502: "The Japanese Society for Animal Psychology",
        2503: "The Japanese Society for Horticultural Science",
        2504: "The Kyushu Dental Society",
        2505: "National Science Teachers Association",
        2506: "Indiana University Center for Genomics and Bioinformatics",
        2507: "Daaam International",
        2508: "Japanese Society of Animal Science",
        2509: "Pacific Northwest Fungi Project",
        2510: "The Commonwealth Fund",
        2511: "TASH",
        2512: "Japanese Society for Palliative Medicine",
        2513: "Informa UK (Lawrence Erlbaum Associates, Inc)",
        2514: "American Institute of Aeronautics and Astronautics",
        2515: "EDP Sciences",
        2516: "EDP Sciences",
        2517: "BioOne (Paleontological Society of Japan)",
        2518: "University of Cadiz Library",
        2519: "Journal of Orthopaedic & Sports Physical Therapy",
        2520: "Japanese Association of Mycotoxicology",
        2521: "Japanese Society of Water Treatment Biology",
        2522: "American Physical Therapy Association",
        2523: "Society of Petroleum Engineers International"
              "Petroleum Technology Conference",
        2524: "Japan Technical Association of the Pulp and Paper Industry",
        2525: "Japanese Society of Agricultural, Biological"
              "and Environmental Engineers and Scientists",
        2526: "Japan Society of Electrical Machining Engineers",
        2527: "American Society of Animal Science",
        2528: "The Electromagnetics Academy (EMW Publishing)",
        2529: "The Electromagnetics Academy (Piers Enterprises)",
        2530: "Japan Society for Laser Surgery and Medicine",
        2531: "Japanese Society of Spinal Surgery",
        2532: "The Electromagnetics Academy",
        2533: "Swiss Chemical Society",
        2534: "Japan Society of Naval Architects and Ocean Engineers",
        2535: "Department of Anatomy Keio University School of Medicine",
        2736: "Japanese Society of Veterinary Dermatology",
        2737: "USDA Forest Service, Research & Development",
        2738: "The Higher Education Electronic & Audio-Video Press",
        2739: "The Kurume Medical Journal",
        2740: "The Japanese Association for the Integrated"
              "Study of Dietary Habits",
        2741: "Frontiers in BioScience",
        2742: "California Institute for Human Science",
        2743: "The Japan Society of Veterinary Epidemiology",
        2744: "BioOne (Chelonian Conservation and Biology Journal)",
        2745: "Japan Society of Drug Delivery System",
        2746: "Wiley (Equine Veterinary Journal)",
        2747: "Bellwether Publishing",
        2748: "Mathematical Institute, Tohoku University",
        2749: "International Association for Bridge and Structural"
              "Engineering",
        2750: "The Association for Rural Planning",
        2751: "The Division of Chemical Information and Computer"
              "Sciences, Chemical Society of Japan",
        2752: "Bloomsbury Academic (Oxford International Publishers Ltd)",
        2753: "M.E. Sharpe",
        2754: "University of Veterinary and Pharmaceutical Sciences",
        2755: "Wiley Blackwell (Blackwell Publishing - Foundation"
              "for Cellular and Molecular Medicine)",
        2757: "Publications Office of the European Union",
        2760: "Publications Office of the European Union",
        2766: "Publications Office of the European Union",
        2778: "Publications Office of the European Union",
        2779: "Publications Office of the European Union",
        2807: "European Centre for Disease Control and Prevention (ECDC)",
        2811: "Publications Office of the European Union",
        2841: "Publications Office of the European Union",
        2870: "Publications Office of the European Union",
        2890: "Publications Office of the European Union",
        2901: "Publications Office of the European Union",
        2902: "Publications Office of the European Union",
        2903: "Publications Office of the European Union",
        2957: "The Japan Society of Hepatology",
        2958: "Japan Pancreas Society",
        2959: "Brill",
        2960: "Northwest Atlantic Fisheries Organization",
        2961: "Japan Laser Processing Society",
        2963: "Japan Society of Thermophysical Properties",
        2964: "Japan Society of Information and Knowledge",
        2965: "Japan Society on Water Environment",
        2966: "Wiley (AHRC Research Centre)",
        2967: "Society of Nuclear Medicine",
        2968: "Sage Publications - Bulletin of the Atomic Scientists",
        2969: "The Mathematical Society of Japan",
        2970: "Current Biodata",
        2971: "European Optical Society",
        2972: "The American School of Classical Studies at Athens",
        2973: "International Ocean Discovery Program Texas",
        2974: "Kitakanto Medical Society",
        2975: "Center for Psychiatric Rehabilitation",
        2976: "Informa UK (Human Frontier Science Program Publishing)",
        2977: "Research Institute for Mathematical Sciences,"
              "Kyoto University",
        2978: "Society of Advanced Science",
        2979: "Indiana University Press",
        2980: "BioOne (Ecoscience)",
        2981: "BioOne (Wildlife Biology)",
        2982: "BioOne (East African Natural History Society)",
        2983: "BioOne (National Shellfisheries Association)",
        2984: "BioOne (Pacific Science)",
        2985: "BioOne (Cactus and Succulent Society of America)",
        2986: "Theological Research Exchange Network",
        2987: "The American Mosquito Control Association",
        2988: "BioOne (The Biological Society of Washington)",
        2989: "National Inquiry Services Center (NISC)",
        2990: "BioOne (Politics and the Life Sciences)",
        2991: "Atlantis Press",
        2992: "BioOne (Carnegie Museum of Natural History)",
        2993: "BioOne (Society of Ethnobiology)",
        2994: "BioOne (Brazilian Herpetological Society)",
        2995: "Japanese Association for Chest Surgery",
        2996: "Tokyo Institute of Technology, Department of Mathematics",
        3057: "Publications Office of the European Union",
        3097: "Landscape Online",
        3098: "Agricultural History Society",
        3099: "BioOne (Museum of Comparative Zoology, Harvard University)",
        3100: "BioOne (Harvard University Botany Libraries)",
        3101: "BioOne (Kentucky Academy of Science)",
        3102: "American Educational Research Association",
        3103: "Allerton Press",
        3104: "The Down Syndrome Educational Trust",
        3105: "BioOne (Herpetological Society of Japan)",
        3106: "BioOne (Mammalogical Society of Japan)",
        3107: "Graduate School of Environmental Science, Okayama University",
        3108: "Centre for Bioscience",
        3109: "Informa UK (Informa Healthcare)",
        3110: "Informa UK (Regulatory Affairs)",
        3111: "Informa UK (PharmaProjects)",
        3112: "Erdkunde",
        3113: "Data Trace Publishing Company",
        3114: "Westerdijk Fungal Biodiversity Institute",
        3115: "Association for Computational Linguistics",
        3116: "Institute of Physical Optics",
        3117: "Japanese Society for Root Research",
        3118: "Japanese Society for Extremophiles",
        3119: "BioOne (New England Botanical Club)",
        3120: "BioOne (California Botanical Society)",
        3121: "Marshfield Clinic Research Foundation",
        3122: "American Board of Family Medicine",
        3123: "The Japanese Environmental Mutagen Society",
        3124: "The Society of Exploration Geophysicists of Japan",
        3125: "Vietnam Journals Online",
        3126: "Nepal Journals Online",
        3127: "Australian Journal of Information Systems",
        3128: "Scidea",
        3129: "Canadian Opthalmological Society",
        3130: "Architectural Institute of Japan",
        3131: "Vacuum Society of Japan",
        3132: "Sherborne Gibbs",
        3133: "US Geological Survey",
        3134: "Co-Action Publishing (Emerging Health Threats Forum CIC)",
        3135: "Marine Acoustics Society of Japan",
        3136: "Japanese Society for Food Science and Technology",
        3137: "Informa UK (Canadian Meteorological and Oceanographic"
              "Society)",
        3138: "University of Toronto Press Inc",
        3139: "Carl Hanser Verlag GMBH & Co KG",
        3140: "Czech Geological Survey",
        3141: "Transportation Research Board",
        3142: "Improbable Research, Inc.",
        3143: "The Japan Geriatrics Society",
        3144: "Department of Polymer EngineeringBudapest University"
              "of Technology & Economics, Scientific Society"
              "of Mechanical Engineering",
        3145: "Ediciones Profesionales de la Informacion SL",
        3146: "American Peanut Research & Education Society",
        3147: "The Japanese Society of Fish Pathology",
        3148: "Dietitians of Canada",
        3149: "Men's Studies Press, LLC",
        3150: "Bernoulli Society for Mathematical Statistics and Probability",
        3151: "Japan Concrete Institute",
        3152: "Informa UK (Beech Tree Publishing)",
        3153: "Scientific Web Journals (SWJ)",
        3154: "The Visualization Society of Japan",
        3155: "Informa UK (Air & Waste Management Association)",
        3156: "Japan Society for Fuzzy Theory and Intelligent Informatics",
        3157: "BioOne (American Entomological Society)",
        3158: "BioOne (Field Museum of Natural History)",
        3159: "BioOne (Torrey Botanical Society)",
        3160: "BioOne (Southern California Academy of Sciences)",
        3161: "Museum and Institute of Zoology at the Polish"
              "Academy of Sciences",
        3162: "Wiley (Comparative Legislative Research Center)",
        3163: "Medical Library Association",
        3164: "The Society for Free Radical Research Japan",
        3165: "The Japanese Society for Pediatric Nephrology",
        3166: "Lavoisier SAS",
        3167: "Berghahn Books",
        3168: "American Dairy Science Association",
        3169: "The Institute of Image Information and Television Engineers",
        3170: "Wiley (Opulus Press)",
        3171: "Journal of Neurosurgery Publishing Group",
        3172: "McFarland & Company, Inc.",
        3173: "The Japanese Society of Agricultural Informatics",
        3174: "American Society of Neuroradiology",
        3175: "Japan Society of Molecular Science",
        3176: "Estonian Academy Publishers",
        3177: "Center for Academic Publications Japan",
        3178: "Japan Society of Hydrology and Water Resources",
        3179: "The Japan Society of Ultrasonics in Medicine",
        3180: "Reactome",
        3182: "Elsevier - International Federation of Automatic"
              "Control (IFAC)",
        3183: "AB Svensk Papperstidning",
        3184: "Science Reviews 2000 LTD.",
        3186: "The Phytopathological Society of Japan",
        3187: "Informa UK(National Inquiry Services Center/African"
              "Journals Online)",
        3188: "Swiss Forestry Society",
        3189: "International Glaciological Society",
        3190: "Czech Geological Society",
        3191: "Japanese Society for Thermal Medicine",
        3192: "Japan Society for Immunology of Reproduction",
        3194: "AIS Educator Association",
        3195: "Phorte Editora",
        3196: "Vittorio Klostermann GMBH",
        3197: "White Horse Press",
        3198: "American Society of Agronomy",
        3199: "International Society of Chromosome Botany",
        3200: "Informa UK (Heldref Publications)",
        3201: "Centers For Disease Control and Prevention",
        3202: "College Art Association",
        3208: "The Japanese Geotechnical Society",
        3209: "The Society for Actinomycetes Japan",
        3210: "The Center for Fire Science and Technology,"
              "Research Institute for Science and Technology,"
              "Tokyo University of Science",
        3213: "Africa Magna Verlag",
        3217: "Verlag der Technischen Universitat Graz",
        3218: "vdf Hochschulverlag AG an der ETH Zurich",
        3221: "Gruppo Italiano Frattura",
        3224: "Barbara Budrich Publishers",
        3233: "IOS Press",
        3238: "Deutscher Arzte-Verlag GmbH",
        3256: "Dike Verlag AG",
        3280: "Franco Angeli",
        3285: "Copernicus GmbH",
        3289: "GEOMAR Helmholtz Centre for Ocean Research Kiel",
        3301: "Societa Geologica Italiana",
        3310: "National Institute for Health Research",
        3311: "Periodica Polytechnica, Budapest University"
              "of Technology and Economics",
        3312: "Yamashina Institute for Ornithology",
        3313: "Japan Landslide Society",
        3314: "The Japanese Society for Medical Mycology",
        3315: "Specjalisci Dermatolodzy",
        3316: "RMIT Publishing",
        3317: "Sage Publications (JRAAS Limited)",
        3318: "Royal Irish Academy",
        3319: "Chinese Geoscience Union",
        3320: "American Industrial Hygiene Association",
        3322: "Wiley (American Cancer Society)",
        3323: "Japan Society of Corrosion Engineering",
        3324: "Ferrata Storti Foundation",
        3325: "Croatian Medical Journal",
        3326: "Institute of Public Finance",
        3327: "Atomic Energy Society of Japan",
        3328: "J. Ross Publishing",
        3329: "Bangladesh Journals Online",
        3330: "The Japanese Society for Comparative Physiology"
              "and Biochemistry",
        3331: "Salem Press Inc.",
        3332: "Ecancer Global Foundation",
        3334: "ORNL Environmental Science Division",
        3335: "Korean Association of Medical Journal Editors (KAMJE)",
        3336: "University of Zagreb, Faculty of Science, Department"
              "of Mathematics",
        3337: "Anatomische Gesellschaft",
        3338: "Korean Society for Cytopathology (KAMJE)",
        3339: "Korean Society of Pediatric Nephrology",
        3340: "Korean Neurosurgical Society",
        3341: "Korean Opthalmological Society",
        3342: "Korean Society of Otorhinolaryngology - Head and Neck Surgery",
        3343: "Korean Society for Laboratory Medicine",
        3344: "Korean Pain Society",
        3345: "Korean Pediatric Society",
        3346: "Korean Academy of Medical Sciences",
        3347: "Korean Society for Parasitology",
        3348: "The Korean Society of Radiology",
        3349: "Yonsei University College of Medicine",
        3350: "The Korean Association for the Study of the Liver",
        3351: "CRESR, Sheffield Hallam University",
        3352: "Korea Health Personnel Licensing Examination Institute",
        3353: "The Japanese Society for Oral Pathology",
        3354: "Inter-Research Science Center",
        3356: "BioOne (The Raptor Research Foundation)",
        3357: "Aerospace Medical Association",
        3358: "Japanese Society for Food Hygiene and Safety",
        3359: "Institute of Information Science",
        3360: "ScienceWise Publishing",
        3361: "Armenian Association of Endoscopic Surgery",
        3362: "Practical Action Publishing",
        3363: "Japanese Association of Inorganic Phosphorus Chemistry",
        3365: "The Korean Institute of Metals and Materials",
        3366: "Edinburgh University Press",
        3367: "Uspekhi Fizicheskikh Nauk Journal",
        3368: "University of Wisconsin Press",
        3369: "Tethys, Journal of Weather & Climate of the"
              "Western Mediterranean",
        3370: "Institute of Life Cycle Assessment, Japan",
        3371: "Walsh Medical Media, LLC",
        3372: "Botanic Garden and Botanical Museum Berlin,"
              "Freie Universitaet Berlin",
        3373: "Japanese Society of Service Dog Research",
        3374: "BioOne (Bulletin of the Peabody Museum of Natural History)",
        3375: "BioOne (Natural Areas Journal)",
        3376: "BioOne (Journal of Vector Ecology)",
        3377: "BioOne (African Zoology)",
        3378: "BioOne (Human Biology, The International Journal"
              "of Population Biology and Genetics)",
        3379: "The Magnetics Society of Japan",
        3380: "Study Group of Microwave Surgery",
        3381: "Tobacco Science",
        3382: "Poultry Science Association",
        3383: "diaNet",
        3384: "Linkoping University Electronic Press",
        3386: "National Bureau of Economic Research",
        3387: "Japanese Society of Eucommia",
        3388: "Japanese Society of Pediatric Allergy and Clinical Immunology",
        3389: "Frontiers Media SA",
        3390: "MDPI AG",
        3391: "Regional Euro-Asian Biological Invasions Centre Oy - REABIC",
        3392: "Sociotechnology Research Network - Shakai Gijutsu Kenkyuukai",
        3393: "Korean Society of Coloproctology (KAMJE)",
        3394: "BioOne (Journal of Great Lakes Research)",
        3395: "Fundacao Oswaldo Cruz",
        3396: "International Federation of Infection Control",
        3397: "Institute of Noise Control Engineering",
        3398: "BioOne (Western North American Naturalist)",
        3399: "Royal College of General Practitioners",
        3400: "The Editorial Committee of Annals of Vascular Diseases",
        3401: "Wiley (Production and Operations Management)",
        3402: "Co-Action Publishing",
        3403: "BSI British Standards",
        3405: "Wiley (New York Academy of Sciences E-Briefings)",
        3406: "PERSEE Program",
        3407: "Nozomi Science",
        3408: "Japanese Association of Forensic Science and Technology",
        3409: "Institute of Systematics and Evolution of Animals,"
              "Polish Academy of Sciences",
        3410: "Faculty Opinions Ltd",
        3411: "The Nature Conservancy",
        3412: "Japanese Society for Bacteriology",
        3413: "Thieme (NuklearMedizin/NuclearMedicine)",
        3414: "Thieme (Methods of Information in Medicine)",
        3415: "Thieme (Veterinary and Comparative Orthopaedics"
              "and Traumatology)",
        3417: "Missouri Botanical Garden Press",
        3418: "Japan Academy of Midwifery",
        3719: "The Weed Science Society of Japan",
        3720: "Japanese Association for Petroleum Engineers",
        3721: "BioOne (Journal of the North Atlantic)",
        3722: "CAD Solutions, LLC.",
        3723: "Society for Underwater Technology",
        3724: "China Science Publishing & Media Ltd.",
        3725: "The Japanese Nematological Society",
        3726: "Peter Lang, International Academic Publishers",
        3727: "Cognizant, LLC",
        3728: "Vienna University of Technology, Institute of General Physics",
        3729: "International Forum for Biophilosophy",
        3730: "Portico",
        3731: "21st Century COE Program (Toplogical Science and Technology)",
        3732: "Botanical Society of America",
        3733: "University of California Agriculture and Natural"
              "Resources (UC ANR)",
        3735: "Schlager Group",
        3736: "Journal of Chinese Integrative Medicine Press",
        3737: "The e-Crystals Federation",
        3738: "Fundacao Educational de Ituverava",
        3739: "The Japanese Society of Limnology",
        3740: "The Materials Research Society of Korea",
        3741: "Korea Water Resources Association",
        3742: "Kyung Hee Oriental Medicine Research",
        3743: "Korean Society for Information Management",
        3744: "The Society of Naval Architects of Korea",
        3745: "Korea Information Processing Society",
        3746: "The Korean Society of Food Science and Nutrition",
        3747: "MultiMed Inc.",
        3748: "Baishideng Publishing Group Co",
        3749: "Mineralogical Association of Canada",
        3750: "Szczecinskie Towarzystwo Naukowe",
        3751: "The Middle East Journal",
        3752: "The Biological Survey of Canada",
        3753: "The Japanese Society of Lumbar Spine Disorders",
        3754: "Princeton Theological Seminary Library",
        3755: "Japanese Coral Reef Society",
        3756: "Society for Art and Science",
        3757: "Japan Society for Equilibrium Research",
        3758: "Springer - Psychonomic Society",
        3759: "The Japan Society of Tropical Ecology",
        3762: "Beilstein-Institut",
        3763: "Earthscan",
        3764: "Archaeological Institute of America",
        3765: "Semantics and Pragmatics",
        3766: "American Academy of Audiology",
        3767: "Naturalis Biodiversity Center",
        3768: "RTI Press",
        3769: "Japan Radioisotope Association",
        3774: "Biophilia Rehabilitation Acdemy",
        3775: "The Japan Institute of Energy",
        3776: "East Carolina University",
        3777: "The Japan Society of Acupuncture and Moxibustion",
        3788: "Shanghai Institute of Optics and Fine Mechanics",
        3789: "National Information Standards Organization",
        3790: "Duncker & Humblot GmbH",
        3791: "MyJoVE Corporation",
        3792: "Project Euclid",
        3793: "Japanese Society of Anti-Aging Medicine",
        3794: "Oxford Brookes University",
        3795: "The Korean Society of Mechanical Engineers",
        3796: "The Korean Society of Fisheries Technology",
        3797: "Osterreichische Pharmazeutische Gesellschaft",
        3798: "Theory in Action",
        3799: "Editorial Office of Earth Science",
        3800: "The Plankton Society of Japan (The Japanese"
              "Association of Benthology)",
        3801: "International Association for Fire Safety Science",
        3802: "Asian Society of Gynecologic Oncology; Korean"
              "Society of Gynecologic Oncology and Colposcopy",
        3803: "Korean Endocrine Society",
        3804: "Japan Association of Breast Cancer Screening",
        3805: "The Japan Epilepsy Society",
        3806: "Japan Kaatsu Training Society",
        3807: "Korean Journal of Optics and Photonics",
        3808: "International Society for Environmental Information Sciences",
        3809: "The Virtual Explorer Pty Ltd.",
        3810: "JTE Multimedia LLC",
        3811: "The Japanese Society for Multiphase Flow",
        3812: "Japan Organization of Clinical Dermatologists",
        3813: "S. Hirzel Verlag",
        3814: "Hindawi Publishing Corporation (Syrexe)",
        3815: "Cambridge University Press (Society for the"
              "Promotion of Roman Studies)",
        3816: "Elsevier - CIG Media Group LP",
        3817: "Telos Press",
        3818: "Justice Research & Statistics Association",
        3819: "Comparative Cognition Society",
        3820: "Japanese Society for Pharmacoepidemiology",
        3821: "Canadian Pharmacists Journal",
        3822: "Massage Therapy Foundation",
        3823: "International Medical Publisher (Fundacion de Neurociencias)",
        3824: "StemBook",
        3825: "Ecology and Civil Engineering Society",
        3826: "Informa UK (International Association of Hydraulic"
              "Engineering and Research (IAHR))",
        3827: "The Foot and Ankle Online Journal",
        3828: "Liverpool University Press",
        3829: "Laboratoire Central des Ponts et Chaussees",
        3830: "Asian Network for Scientific Information",
        3831: "Korean Pharmacopuncture Institute",
        3832: "SISEF",
        3833: "PharmaVentures",
        3834: "Digital Science Press",
        3835: "The Plant Genome",
        3836: "Tokyo Journal of Mathematics",
        3837: "Korean Society for Internet Information",
        3838: "JStage(Ornithological Society of Japan)",
        3839: "The Korean Society for Applied Biological Chemistry",
        3840: "Wolfram Research",
        3841: "International Wire & Cable Symposium",
        3842: "SIGMA",
        3843: "Sapiens Publishing, LTD",
        3844: "Science Publications",
        3845: "SEE",
        3846: "Vilnius Gediminas Technical University",
        3847: "American Astronomical Society",
        3848: "Iktisat Isletme ve Finans Dergisi",
        3849: "University of Defence",
        3850: "Research Publishing Services",
        3851: "International Medical Press",
        3852: "Mycological Society of America",
        3853: "Australian Museum",
        3854: "Chelonian Research Foundation",
        3855: "Journal of Infection in Developing Countries",
        3856: "Pontificia Universidad Catolica de Valparaiso",
        3857: "Korean Society for Therapeutic Radiology and Oncology",
        3858: "Springer Science and Business Media LLC",
        3860: "PhilJOL",
        3861: "The Japanese Society of Health and Human Ecology",
        3862: "The Japan Society of Coloproctology",
        3866: "Acta Physico-Chimica Sinica & University Chemistry"
              "Editorial Office, Peking University",
        3884: "Walter de Gruyter GmbH (European Journal of Nanomedicine)",
        3885: "Co-Action Publishing Medical Education Online",
        3886: "Inter-university Consortium for Political and Social Research",
        3887: "Sunmedia",
        3888: "Wolfram Research",
        3889: "ID Design 2012/DOOEL Skopje",
        3890: "International University Line",
        3891: "Danish Chemical Society",
        3892: "Spandidos Publications",
        3893: "Japanese Association for Acute Medicine",
        3894: "Journal of Advances in Modeling Earth Systems",
        3895: "Universidade Tecnologica Federal do Parana",
        3896: "International Bee Research Association",
        3897: "Pensoft Publishers",
        3898: "Lawrence and Wishart",
        3899: "The Journal of Rheumatology",
        3900: "Colegio Brasileiro de Atividade Fisica Saude e Esporte",
        3901: "Chinese Journal of Mechanical Engineering",
        3902: "Japanese Neural Network Society",
        3903: "Green Theory and Praxis: The Journal of Ecopedagogy",
        3904: "Korean Association of Internal Medicine",
        3905: "Pageant Media US",
        3906: "Scientific and Technological Research Council of Turkey",
        3908: "Albert Einstein College of Medicine",
        3916: "Grupo Comunicar",
        3917: "CAIRN",
        3918: "Japanese Society of Intensive Care Medicine",
        3919: "Japan Surgical Association",
        3920: "Wageningen Academic Publishers",
        3921: "Elsevier - Colegio Nacional de Opticos-Optometristas"
              "de Espana",
        3922: "Psychology and Neuroscience",
        3923: "Science Alert",
        3924: "Korean Medical Citation Index Project Board (KAMJE)",
        3925: "Japan Society of Transfusion Medicine and Cell Therapy",
        3926: "Omnia Publisher SL",
        3927: "John Rylands University Library",
        3928: "SLACK, Inc.",
        3934: "American Institute of Mathematical Sciences",
        3935: "Pravni Fakultet Sveucilista u Zagrebu - Law"
              "School of the University of Zagreb",
        3936: "hpsmedia",
        3937: "Japan Society for Oriental Medicine",
        3938: "Korean Physical Society",
        3939: "Informa UK (Georg D. W. Callwey GambH and Co., KG)",
        3940: "Royal Institution of Naval Architects",
        3941: "EduRad",
        3942: "Protein Structure Initiative Structural Genomic"
              "Knowledgebase Rutgers",
        3943: "De Montfort University",
        3944: "Turk Ortopedi ve Travmatoloji Dernegi",
        3945: "American Society for Nutrition",
        3946: "Korean Society of Medical Education",
        3947: "Korean Society of Infectious Diseases and Korean"
              "Society for Chemotherapy (KAMJE)",
        3949: "Cleveland Clinic Journal of Medicine",
        3950: "Oto-Rhino-Laryngological Society of Japan, Inc.",
        3951: "Society of Biomechanisms Japan",
        3952: "Lithuanian Academy of Sciences",
        3954: "BioOne (South Carolina Entomological Society)",
        3955: "BioOne (Northwest Scientific Association)",
        3956: "BioOne (Pacific Coast Entomological Society)",
        3957: "BioOne (Southern African Wildlife Management Association)",
        3958: "BioOne (Society of Southwestern Entomologists)",
        3959: "BioOne (Tree-Ring Society)",
        3960: "Japanese Society for Lymphoreticular Tissue Research",
        3961: "Korean Society for Preventive Medicine",
        3969: "Wanfang Data prefix (for IEEE acquisition)",
        3974: "Editorial Office of Journal of Global Change"
              "Data and Discovery",
        3982: "The Econometric Society",
        3983: "Transformative Works and Cultures",
        3984: "Instytut Bioinfobank Sp. z o.o.",
        3985: "Japan Society of Waste Management Experts",
        3986: "The Research Center of the Slovenian Academy"
              "of Sciences and Arts / Znanstvenoraziskovalni"
              "center Slovenske akademije znanosti in umetnosti (ZRC SAZU)",
        3987: "The Japan Institute of Heterocyclic Chemistry",
        3988: "Korean Neurological Association",
        3989: "Editorial CSIC",
        3990: "University Library/University of Twente",
        3991: "International Association of Online Engineering",
        3992: "College Publishing",
        3993: "Textile Bioengineering and Informatics Society",
        3994: "Center for East Asian Studies",
        3995: "Japan Stroke Society",
        3996: "U.S. Fish and Wildlife Service",
        3997: "EAGE Publications",
        3998: "University of Michigan Library",
        3999: "Japanese Society of Clinical Pharmacology and Therapeutics",
        4000: "OpenEdition",
        4001: "BioOne (Entomological Society of Southern Africa)",
        4002: "BioOne (Institute of Malacology)",
        4003: "BioOne (American Malacological Society)",
        4004: "Wiley (The Wildlife Society)",
        4005: "The Japanese Forest Society",
        4006: "Physics Essays Publication",
        4007: "Annals of Mathematics",
        4009: "Japanese Society for Dialysis Therapy",
        4010: "Japan Society of Lipid Nutrition",
        4011: "Japan Society of Colour Material",
        4012: "Japanese Society for Dental Materials and Devices",
        4013: "Universidade do Vale do Rio Dos Sinos - UNISINOS",
        4014: "Journal of Microbiology and Biotechnology",
        4015: "National Taiwan University",
        4016: "SciVee, Inc.",
        4017: "International Society for Gerontechnology",
        4018: "IGI Global",
        4019: "Japan Society of Coordination Chemistry",
        4020: "Japanese Heart Rhythm Society",
        4021: "Canada-China Clinical Medicine Study Association",
        4022: "Cardiofront, LLC",
        4023: "Clarivate Analytics (US)",
        4024: "Collegium Basilea",
        4025: "Universidade Estadual de Maringa",
        4026: "IS, GUC The Journal of Industrial Relations"
              "and Human Resoruces",
        4027: "Alaska Sea Grant",
        4028: "Trans Tech Publications, Ltd.",
        4029: "Society of Advocates for Sustainable Environment",
        4030: "Japanese College of Surgeons",
        4031: "Marine Technology Society",
        4032: "Pan Stanford Publishing",
        4033: "Ideas in Ecology and Evolution",
        4034: "Associacao de Apoio a Pesquisa em Saude Bucal",
        4035: "Japanese Society of Farm Work Research",
        4036: "Graduate School of Information Sciences, Tohoku University",
        4037: "AACN Publishing",
        4038: "Sri Lanka Journals Online",
        4039: "Cambridge University Press (Entomological Society of Canada)",
        4040: "Korean Society of Nursing Science",
        4041: "Korean Association of Orthodontists",
        4043: "Society of Petroleum Engineers Offshore Technology Conference",
        4044: "Okayama Medical Association",
        4045: "Norwegian Medical Association",
        4046: "The Korean Academy of Tuberculosis and Respiratory Diseases",
        4047: "The Korean Academy of Prosthodontics",
        4048: "Korean Breast Cancer Society",
        4049: "The American Association of Immunologists",
        4050: "American Helicopter Society",
        4051: "Korean Society for Bioinformatics and Systems Biology",
        4052: "Forum: Carbohydrates Coming of Age",
        4053: "National Nanomanufacturing Network",
        4054: "Max Planck Institute for Demographic Research",
        4055: "The Korean Orthopaedic Association",
        4056: "Genomic Standards Consortium",
        4057: "The Japan Sociological Society",
        4058: "Japanese Society of Environmental Infections",
        4059: "Aequitas Publishing",
        4060: "Food and Agriculture Organization of the United Nations",
        4061: "Hindawi Publishing Corporation (Sage-Hindawi"
              "Access to Research)",
        4062: "The Korean Society of Applied Pharmacology",
        4063: "Japanese Association of Electoral Studies",
        4064: "Institute of Mathematics, Polish Academy of Sciences",
        4065: "Elsevier - Mayo Clinic Proceedings",
        4066: "Allied Academies (OMICS)",
        4067: "SciELO Agencia Nacional de Investigación y Desarrollo (ANID)",
        4068: "Chonnam National University Medical School",
        4069: "Korean Academy of Women's Health Nursing",
        4070: "The Korean Society of Cardiology",
        4071: "IMAPS - International Microelectronics Assembly"
              "and Packaging Society",
        4072: "Sociedade Brasileira de Paleontologia",
        4073: "The Campbell Collaboration",
        4074: "Editions NecPlus",
        4076: "Springer (Cases Network, Ltd.)",
        4077: "Chinese Physiological Society",
        4078: "The Korean Rheumatism Association",
        4079: "George Washington University",
        4080: "The Peregrine Fund",
        4081: "PAGEPress Publications",
        4082: "The Korean Academy of Family Medicine",
        4084: "Institute of Hematology, Catholic University",
        4085: "Journal of Athletic Training/NATA",
        4086: "Theory of Computing Exchange",
        4087: "The Foundation Review",
        4088: "Physicians Postgraduate Press, Inc.",
        4089: "Biology Reports, Ltd.",
        4090: "Journal of Urban and Environmental Engineering",
        4091: "The Health Care Science Institute",
        4092: "Japan Society for Research on Emotions",
        4093: "Korean Diabetes Association",
        4094: "Korean Academy of Child Health Nursing",
        4095: "Natural Resources Canada",
        4096: "The Sedimentological Society of Japan",
        4097: "The Korean Society of Anesthesiologists",
        4098: "Mammal Research Institute",
        4099: "Mathematical Society of Japan",
        4100: "Journal of Human Sport and Exercise",
        4101: "Virtual Worlds Research Consortium",
        4102: "AOSIS",
        4103: "Medknow Publications",
        4104: "Primary Care Respiratory Society UK",
        4105: "Taiwanese Society of Molecular Medicine",
        4106: "Japan Society for Snow Engineering",
        4107: "Japan Society of Gymnastics for All",
        4108: "European Alliance for Innovation n.o.",
        4109: "Japan Society for Lactic Acid Bacteria",
        4110: "Korean Association of Immunobiologists",
        4111: "The Korean Urological Association",
        4112: "Austrian Veterinary Society",
        4113: "Informa UK (Journal of Maps)",
        4114: "IBERAMIA: Sociedad Iberoamericana de Inteligencia Artificial",
        4115: "Journal of Logic and Analysis",
        4116: "Japan Association for Quaternary Resarch",
        4127: "Department of Sport Management , University of Peloponnese",
        4128: "Business Expert Press",
        4129: "Accademia Italiana di Scienze Forestali",
        4130: "Asian Electric Vehicle Society",
        4131: "The Japan Society of High Pressure Science and Technology",
        4132: "The Korean Society of Pathologists and The"
              "Korean Society for Cytopathology",
        4133: "Environmental and Engineering Geophysical Society",
        4134: "The Korean Mathematical Society",
        4135: "Sage Publications - Reference E-Books",
        4136: "Instituto de Pesquisas Ambientais em Bacias Hidrograficas",
        4137: "Libertas Academica, Ltd.",
        4138: "Atlantic Geology",
        4139: "The Surface Finishing Society of Japan",
        4140: "American Society of Consultant Pharmacists",
        4141: "Agricultural Institute of Canada",
        4142: "The Korean Society of Veterinary Science",
        4143: "Korean Cancer Association",
        4144: "The Resources Processing Society of Japan",
        4145: "Japanese Association of Hydrological Sciences",
        4146: "Japanese Society of Sport Psychology",
        4147: "FBCommunication",
        4148: "New Prairie Press",
        4149: "AEPress, s.r.o.",
        4150: "The Korean Powder Metallurgy Institute",
        4151: "Pontifica Universidad Catolica de Valparaiso",
        4152: "Portuguese Society of Electrochemistry",
        4153: "Canadian Mathematical Society",
        4154: "Croatian Geological Survey",
        4155: "Newlands Press Ltd",
        4156: "AICIT",
        4157: "The Association of Japanese Geographers",
        4158: "American Association of Clinical Endocrinologists",
        4159: "Harvard University Press",
        4160: "International Potato Center",
        4161: "Landes Bioscience",
        4162: "The Korean Nutrition Society and The Korean"
              "Society of Community Nutrition",
        4163: "The Korean Nutrition Society",
        4164: "The Society of Powder Technology, Japan",
        4165: "The Kansai Plant Protection Society",
        4166: "The Korean Society of Gastroenterology",
        4167: "The Korean Society for Microbiology and The"
              "Korean Society of Virology (KAMJE)",
        4168: "The Korean Academy of Asthma, Allergy and Clinical"
              "Immunology and The Korean Academy of Pediatric"
              "Allergy and Respiratory Disease",
        4169: "Mathematical Association of America",
        4170: "AVES Publishing Co. Akademik Acil Tip Dergisi/Journal"
              "of Academic Emergency Medicine",
        4171: "EMS Press",
        4172: "OMICS Publishing Group",
        4173: "Norwegian Society of Automatic Control",
        4174: "The Korean Surgical Society",
        4175: "Frontiers Media SA",
        4176: "CoAction Publishing Libyan Journal of Medicine Society",
        4177: "Center for Applied Energy Research",
        4178: "Korean Society of Epidemiology (KAMJE)",
        4179: "New Labor Forum",
        4180: "Universite des Antilles et de la Guyane",
        4181: "Revista Neurociencias",
        4182: "Entomological Society of America Arthropod Management Tests",
        4183: "ACTA Endocrinologica Foundation",
        4184: "Korean Society of Spine Surgery",
        4185: "Revista Latina de Comunicacion Social",
        4186: "Faculty of Engineering, Chulalongkorn University",
        4187: "Daedalus Enterprises",
        4188: "The Textile Machinery Society of Japan",
        4189: "Society for Human Environmental Studies",
        4190: "Japanese Society of Veterinary Clinics",
        4191: "The Korean Ceramic Society",
        4192: "The International Journal of Digital Accounting Research",
        4193: "Rhinology",
        4194: "Central Fisheries Research Institute (SUMAE)",
        4195: "American Society of Agronomy",
        4196: "The Korean Physiological Society and The Korean"
              "Society of Pharmacology",
        4197: "King Abdulaziz University Scientific Publishing Centre",
        4199: "Morgan & Claypool",
        4200: "The Human Geographical Society of Japan",
        4201: "The London Student Journal of Medicine",
        4202: "Polska Akademia Nauk Instytut Paleobiologii"
              "Institute of Paleobiology, Polish Academy of Sciences)",
        4203: "Civil-Comp, Ltd.",
        4204: "Open Publishing Association",
        4205: "Royal Society of South Australia",
        4206: "Sistema de Bibliotecas UACH",
        4207: "Paleoantropology Society",
        4208: "Global Science Press",
        4209: "Taiwan Association for Aerosol Research",
        4210: "Smith School of Enterprise and the Environment",
        4211: "Consortium of Universities for the Advancement"
              "of Hydrologic Science, Inc.",
        4212: "Canadian Society of Hospital Pharmacists",
        4213: "Steklov Mathematical Institute",
        4214: "Rede Euroamericana de Motricidade Humana",
        4215: "Mercator - Revista de Geografia da UFC",
        4216: "The Philosophy of Science Society, Japan",
        4217: "Korea Ocean Research and Development Institute",
        4218: "Electronics and Telecommunications Research Institute",
        4219: "Sage Publications (Prufrock Press, Inc.)",
        4220: "Norwegian Nurses Organisation",
        4221: "The Norwegian Association of Masters of Science"
              "in Business/Sivilokonomene",
        4222: "The Institute for Leadership and Global Education",
        4223: "The Council on Employee Responsibilities and Rights",
        4234: "Japan Society of Family Sociology",
        4235: "The Korean Geriatrics Society",
        4236: "Scientific Research Publishing, Inc.",
        4237: "Programa de Pos Graduacao em Arquitetura e Urbanismo",
        4238: "Genetics and Molecular Research",
        4239: "Baishideng Publishing Group Co (World Journal of Diabetes)",
        4240: "Baishideng Publishing Group Co (World Journal"
              "of Gastrointestinal Surgery)",
        4241: "The Association for Plant Protection of Kyushu",
        4242: "Mulberry Technologies, Inc.",
        4243: "The Hebrew University Magnes Press",
        4244: "Europa Digital and Publishing",
        4245: "Spontaneous Generations",
        4246: "moved to DataCite",
        4247: "Society of Adaptive Science in Taiwan",
        4248: "International Journal of Oral Science, West"
              "China School of Stomatology",
        4249: "Scholarpedia",
        4250: "Korean Society of Echocardiography",
        4251: "Baishideng Publishing Group Co (World Journal"
              "of Gastrointestinal Oncology)",
        4252: "Baishideng Publishing Group Co (World Journal of Stem Cells)",
        4253: "Baishideng Publishing Group Co (World Journal"
              "of Gastrointestinal Endoscopy)",
        4254: "Baishideng Publishing Group Co (World Journal of Hepatology)",
        4255: "Lumitext Publishing",
        4256: "University of Plymouth",
        4257: "Oecologia Australis",
        4258: "The Korean Society of Medical Informatics",
        4259: "Japanese Association of Indian and Buddhist Studies",
        4260: "Institute of Food Technology",
        4261: "AVES - Turkish Society of Radiology",
        4262: "Electric Furnace Steel Forum",
        4263: "Japan Society of Engineering Geology",
        4264: "The Nihon University Medical Association",
        4265: "The Society for Antibacterial and Antifungal Agents, Japan",
        4266: "The Korean Society of Critical Care Medicine",
        4267: "INIST-CNRS",
        4268: "China Journal of Chinese Materia Medica",
        4269: "American Society of Tropical Medicine and Hygiene",
        4270: "Revista Universo Contabil",
        4271: "SAE International",
        4272: "Netbiblo",
        4273: "MAFTREE",
        4274: "Galenos Yayinevi",
        4275: "Korean Society for Library and Information Science",
        4276: "College of Occupational Therapists",
        4277: "NEN Netherlands Standardization Institute",
        4278: "American Journal of Health Promotion",
        4279: "Papers in Physics",
        4280: "Huelva University Press",
        4281: "EURP",
        4282: "The Sessile Organisms Society of Japan",
        4283: "The Korean Magnetics Society",
        4284: "Southern Economic Association",
        4285: "The Korean Society for Transplantation",
        4286: "Japanese Society of Medical Instrumentation",
        4287: "Japan Society of Photogrammetry and Remote Sensing",
        4288: "The Japan Association for Philosophy of Science",
        4289: "BioOne (Entomological Society of Washington)",
        4290: "BioOne (Freshwater Biological Association)",
        4291: "Baishideng Publishing Group Co",
        4292: "Baishideng Publishing Group Co",
        4293: "The Society of Laparoscopic and Robotic Surgeons",
        4294: "Seismological Society of Japan",
        4295: "Japan Audiological Society",
        4296: "Canadian Water Resources Association",
        4297: "North American Journal of Medical Sciences",
        4298: "ReVision Publishing",
        4299: "American Society for Aesthetics",
        4300: "Journal of Graduate Medical Education",
        4301: "TECSI",
        4302: "Photonics Society of Poland",
        4303: "Ashdin Publishing (OMICS)",
        4304: "Academy Publisher",
        4305: "Middle East Technical University, Faculty of Architecture",
        4306: "Korean Neuropsychiatric Association",
        4307: "Japanese Society for Engineering Education",
        4308: "Bogor Agricultural University",
        4309: "Journal of Gambling Issues",
        4310: "International Press of Boston, Inc.",
        4311: "National Speleological Society",
        4312: "Faculty of Arts, University of Ljubljana",
        4313: "The Korean Institute of Electrical and Electronic"
              "Material Engineers",
        4314: "African Journals Online",
        4315: "International Association for Food Protection",
        4316: "Universitatea Stefan cel Mare din Suceava",
        4317: "Medicina Oral, S.L.",
        4318: "Turkish Society of Gastroenterology",
        4319: "American Society of Limnology and Oceanography, Inc.",
        4320: "Dulwich Centre Publicaitons",
        4321: "SciELO Espana/Repisalud",
        4322: "Editora Cubo",
        4323: "Romanian Society of Legal Medicine",
        4324: "Informa UK (Routledge)",
        4325: "The Japan Society of Polymer Processing",
        4326: "The Japanese Society for Cardiovascular Surgery",
        4327: "Japanese Society of Nutrition and Food Science",
        4328: "Journal of Clinical and Analytical Medicine",
        4329: "Baishideng Publishing Group Co (World Journal of Radiology)",
        4330: "Baishideng Publishing Group Co (World Journal of Cardiology)",
        4331: "Baishideng Publishing Group Co (World Journal of Chemistry)",
        4332: "The Korean Society of Health Policy and Administration",
        4333: "Springer - The Korean Society of Pharmaceutical"
              "Sciences and Technology",
        4334: "Korea Concrete Institute",
        4335: "Institut Za Lokalno Samoupravo in Javna Narocila Maribor",
        4336: "Embrapa Florestas",
        4337: "Edward Elgar Publishing",
        4338: "Thieme (Applied Clinical Informatics)",
        4401: "Instituto Nazionale di Geofisica e Vulcanologia, INGV",
        4409: "Associazione Acque Sotterranee",
        4414: "EMH Swiss Medical Publishers, Ltd.",
        4415: "Istituto Superiore di Sanita",
        4467: "Uniwersytet Jagiellonski – Wydawnictwo Uniwersytetu"
              "Jagiellonskiego",
        4468: "Universita degli Studi di Milano-Bicocca Symphonya"
              "Emerging Issues in Management",
        4472: "DIAPHANES AG",
        4489: "The Korean Society of Mycology",
        4490: "The Korean Society of Phycology",
        4491: "Korean Society of Environmental Engineering",
        4992: "The Japanese Psychological Association",
        4993: "The Japanese Society of Strategies for Cancer"
              "Research and Therapy",
        4994: "Asociacion Japonesa de Hispanistas",
        4995: "Universitat Politecnica de Valencia",
        4996: "Association for Fire Ecology",
        4997: "Royal College of Physicians of Edinburgh",
        4999: "Akademi Doktolar Yayinevi",
        5000: "Earthquake Engineering Society of Korean",
        5001: "Oman Medical Journal",
        5002: "Dr. Akhilesh Pandey's Laboratory",
        5003: "Demos Medical Publishing",
        5004: "Desalination Publications",
        5005: "Jaypee Brothers Medical Publishing",
        5006: "NACE International",
        5007: "Universidade Federal de Santa Catarina",
        5008: "Associacao Educacional de Vitoria",
        5009: "The Editorial Office of Gut and Liver",
        5010: "The Korean Society for Plant Biotechnology",
        5011: "Korean Information and Communications Society",
        5012: "Korean Chemical Society",
        5013: "UK Simulation Society",
        5014: "AOTA Press",
        5015: "University of Toronto Medical Journal",
        5016: "UNESP - Universidade Estdual Paulista",
        5017: "International Ecostations",
        5018: "ZBW - German National Library of Economics",
        5019: "Soft Computing and Intelligent Information Systems",
        5020: "Fundacao Edson Queiroz",
        5021: "Korean Dermatological Association and The Korean"
              "Society for Investigative Dermatology",
        5022: "The English Linguistic Society of Japan",
        5023: "Japanese Society of Applied Statistics",
        5024: "Historical Society of English Studies in Japan",
        5025: "Japanese Leprosy Association",
        5026: "Tokyo Geographical Society",
        5027: "Pontificia Universidad Catolica de Valparaiso",
        5028: "Institute of Aeronautics and Space",
        5029: "Business History Society of Japan",
        5030: "Japan Chapter of International Geosynthetics Society",
        5031: "RIA",
        5032: "American Association for Agricultural Education",
        5033: "International Free and Open Source Software Law Review",
        5034: "Excellus Health Plan, Inc.",
        5035: "West-Japanese Society of Orthopedics & Traumatology",
        5036: "Department of Mathematics, Faculty of Science",
        5037: "Stilus Optimus",
        5038: "University of South Florida Libraries",
        5039: "Revista Brasileira de Ciencias Agrarias",
        5040: "Bloomsbury Academic",
        5041: "Rambam Health Corporation",
        5042: "Emerald - Pier Professional",
        5043: "E-Learning Network of Australia",
        5044: "Applied Science and Analysis",
        5045: "Korean Society of Hematology;"
              " Korean Society of Blood and Marrow Transplantation;"
              " Korean Society of Pediatric Hematology-Oncology;"
              " Korean Society on Thrombosis and Hemostasis",
        5046: "PMPH-USA, Ltd.",
        5047: "Terrapub",
        5048: "Biologic Institute",
        5049: "The Korean Society of Electrolyte Metabolism",
        5050: "The Korean Society for Noise and Vibration Engineering",
        5051: "Korean Academy of Periodontology",
        5052: "Springer - Real Academia de Ciencias Exactas,"
              "Fisicas y Naturales",
        5053: "Cevre Koruma ve Arastirma Vakfi",
        5054: "Wiley (Teachers of English to Speakers of Other"
              "Languages, Inc.)",
        5055: "Weston Medical Publishers",
        5056: "The Korean Society of Neurogastroenterology"
              "and Motility (KAMJE)",
        5057: "Japan  Society of Kansei Engineering",
        5058: "Public Health Research Foundation",
        5059: "The Society of Synthetic Organic Chemistry, Japan",
        5070: "eScholarship Publishing",
        5077: "Universite de Geneve",
        5080: "Turkish Association of Nervous and Mental Health",
        5081: "International Association of Chinese Professionals"
              "in Global Positioning Systems",
        5082: "XMLArchive",
        5083: "Healthcare Bulletin",
        5084: "Psychomusicology: Music, Mind and Brain",
        5085: "Journal of Forensic Economics",
        5086: "The Scientific and Technological Research Council of Turkey",
        5087: "Dialogue and Discourse",
        5088: "Instituto de Oceanografia - FURG",
        5089: "International Monetary Fund",
        5090: "Korean Society for Thoracic and Cardiovascular Surgery",
        5091: "Agentschap Plantentuin Meise",
        5092: "Historians of Netherlandish Art",
        5093: "Colegio Oficial de Psicologos de Madrid",
        5094: "Turkish National Committee for Air Pollution and Control",
        5095: "Coldev Editions",
        5096: "Institute of Human Cognition and Brain Science",
        5097: "Turkish Surgical Association",
        5098: "Global Digital Central",
        5099: "New World Publishing International, Inc.",
        5100: "Japan Ergonomics Society",
        5101: "OAhost",
        5102: "Centro de Ensino Unificado de Brasilia",
        5103: "Korean Society of Sport Biomechanics",
        5104: "Japan Institute of Electronics Packaging",
        5105: "The Japanes Society of Electrocardiology",
        5106: "Japan Society for Head and Neck Surgery",
        5107: "Society of Cosmetic Chemists of Japan",
        5108: "Japan Academic Collaborations Medical Society",
        5109: "Kyushu University",
        5110: "Japan Society of Engineering Geology",
        5111: "The Spectroscopical Society of Japan",
        5112: "The Japan Society of Logopedics and Phoniatrics",
        5113: "Instituto IMANI",
        5114: "Termedia Sp. z.o.o.",
        5115: "Korean Association of Anatomists",
        5116: "International Journal of Medical Education",
        5117: "Amsterdam University Press",
        5118: "Burke's Peerage and Gentry",
        5120: "Foundation of Computer Science",
        5121: "Academy and Industry Research Collaboration Center",
        5122: "Chinese Birds Press",
        5123: "Instituto Evandro Chagas",
        5124: "Korean Medical Association (KAMJE)",
        5125: "The Korean Association of Oral and Maxillofacial Surgeons",
        5126: "Excellus Health Plan, Inc.",
        5127: "Textrum, Ltd.",
        5128: "Estonian Association for Applied Linguistics",
        5129: "Comparative Politics",
        5130: "University of Technology, Sydney",
        5131: "American Journal of Comparative Law",
        5132: "Ecototoxicology and Environmental Contamination",
        5133: "Egyptian Journal of Surgery",
        5134: "Kyoto University Library",
        5135: "The European Union Studies Association - Japan",
        5136: "The Society of Life Support Technology",
        5137: "Turkish Neurosurgical Society",
        5138: "Advanced Research Journals",
        5139: "The Korean Society for Aeronautical & Space Sciences",
        5140: "The Korean Space Science Society",
        5141: "The Ecological Society of Korean",
        5142: "The Korean Society of Ginseng",
        5143: "The Ergonomics Society of Korea",
        5144: "King Faisal Specialist Hospital and Research Centre",
        5145: "The Korean Society of Clinical Microbiology",
        5146: "Federation of Turkish Pathology Societies",
        5147: "Atlas Publishing, LLC",
        5148: "Ottawa United Learning Academy",
        5149: "University of North Carolina Press",
        5150: "Centre Francais du Littoral",
        5151: "Editora Edgard Blucher, Ltda.",
        5152: "AVES Publishing Co.",
        5153: "Sociological Research Online",
        5154: "Universidad Autonoma Chapingo",
        5155: "European Journal of Chemistry",
        5162: "AMA Service GmbH",
        5166: "University of Bern",
        5171: "IBIMA Publishing",
        5172: "E-Content Management",
        5173: "Polish Urological Association",
        5174: "AVES Publishing Company - Trakya University"
              "Faculty of Medicine",
        5175: "Council on Social Work Education",
        5176: "Global Science and Technology Forum",
        5177: "Netherlands Tijschrift voor Tandheelkunde",
        5178: "Human Behavior and Evolution Society of Japan",
        5179: "The Japanese Association of Benthology",
        5180: "Japan Society of Gynecologic and Obstetric"
              "Endoscopy and Minimally Invasive Therapy",
        5181: "Japan Cetology Research Group",
        5182: "Japan Association of Ion Exchange",
        5183: "Japanese Society of Computational Statistics",
        5184: "Classical Association of the Middle West and South, Inc.",
        5185: "VBRI Sverige AB",
        5186: "Finnish Academy of Science and Letters",
        5187: "Korean Society of Animal Science and Technology",
        5188: "The Society of Materials Engineering for Resources of Japan",
        5189: "The Polargraphic Society of Japan",
        5190: "The Tohoku Geographical Association",
        5191: "Journal of International Agricultural and Extension Education",
        5192: "Northwestern University Press",
        5193: "Association for Experimental Education",
        5194: "Copernicus GmbH",
        5195: "University Library System, University of Pittsburgh",
        5196: "APGF-IFGW/UNICAMP",
        5197: "British Society for Plant Pathology",
        5198: "Center for Transportation Studies",
        5199: "Association of Sport Tourism and Recreation",
        5200: "Association of Lithuanian Serials",
        5201: "Image Processing On Line",
        5202: "Dipartimento di Economia, Universita di Perugia (IT)",
        5203: "University of Manitoba Libraries",
        5204: "Queensland University of Technology",
        5205: "Revista de Enfermagem, UFPE Online",
        5206: "University of Western Ontario, Western Libraries",
        5207: "The Korean Institute of Illuminating and Electrical"
              "Installation Engineers",
        5208: "ML Communications Co., Ltd",
        5209: "Universidad Complutense de Madrid",
        5210: "University of Illinois Libraries",
        5211: "Plaza y Valdes",
        5212: "Universidade Estadual de Ponta Grossa",
        5213: "Korean Continence Society",
        5214: "Indian Academy of Neurosciences",
        5215: "Historical Society of Pennsylvania",
        5216: "Universidade Federal de Goias",
        5217: "Korean Association for the Study of Intestinal"
              "Diseases (KAMJE)",
        5218: "Portularia, Revista de Trabajo Social",
        5219: "HACCP Consulting",
        5220: "Scitepress",
        5221: "Whiting & Birch, Ltd.",
        5222: "Logos Medical Publication",
        5223: "The Korean Society of Pediatric Gastroenterology,"
              "Hepatology and Nutrition",
        5224: "Massachusetts Historical Society",
        5225: "Japanese Association for Sport Management",
        5226: "Study Group of Aero Aqua Bio-mechanisms",
        5227: "The Japanese Skin Cancer Society",
        5228: "The Korean Society for Technology of Plasticity",
        5229: "The Korean Electrochemical Society",
        5230: "The Korean Gastric Cancer Association",
        5231: "Universidad de Malaga Faculty of Psychology",
        5232: "Revista Internacional de Ciencias del Deporte",
        5233: "SNSPMS Bucharest",
        5234: "The Japanese Society of Clinical Neuropsychopharmacology",
        5235: "Hart Publishing",
        5236: "mEDRA",
        5241: "The National Federation of the Blind",
        5242: "Computer Oriented Geology COGEO",
        5243: "Society for Social Work and Research, Inc.",
        5244: "British Machine Vision Association and Society"
              "for Pattern Recognition",
        5245: "Psion Journal",
        5246: "Journal of Chinese Pharmaceutical Sciences",
        5247: "The Grotowski Institute",
        5248: "Mycotaxon, Ltd.",
        5249: "Journal of Injury and Violence Research",
        5250: "University of Nebraska Press",
        5251: "Science Hub",
        5252: "BioOne (Museum National d'Histoire Naturelle, Paris, France)",
        5253: "BioOne (Netherlands Ornithologists' Union)",
        5254: "Rubber Division, ACS",
        5260: "The Charleston Co.",
        5261: "Spanish Journal of Rural Development",
        5262: "Turkish Nephrology Dialysis Transplantation",
        5263: "Australian Society for the Study of Labour History",
        5264: "The Japanese Society of Nutrition and Dietetics",
        5265: "The Japanese Society for Cognitive Psychology",
        5266: "The Institute of Positioning, Navigation and Timing of Japan",
        5267: "Growing Science",
        5268: "Freshwater Biological Association",
        5269: "Sociedade Paranaense de Matematica",
        5270: "European Space Agency",
        5271: "Scaninavian Journal of Work, Environment, and Health",
        5272: "Peytchinski Publishing Ltd.",
        5273: "National Osteoporosis Center",
        5274: "American Society of Sugarbeet Technologist",
        5275: "Remedica Medical Education and Publishing",
        5276: "Journal of Solid Waste Technology and Management",
        5292: "Korean Bone and Joint Tumor Society (KAMJE)",
        5293: "Korean Society for Fluid machinery",
        5294: "Universidad de la Sabana",
        5295: "Instituto de Economia Aplicada a la Empresa",
        5296: "Macrothink Institute, Inc.",
        5297: "Airiti Press, Inc.",
        5298: "Association for Applied Psychophysiology and Biofeedback",
        5299: "Surgisphere Corporation",
        5300: "CSTUG",
        5301: "Wichtig Publishing, SRL",
        5302: "Institute of Control, Robotics and Systems",
        5303: "The Korean Astronomical Society",
        5304: "Lyson Center for Civic Agriculture and Food Systems",
        5305: "American Society of International Law",
        5306: "Baishideng Publishing Group Co (World Journal"
              "of Clinical Oncology)",
        5307: "Korean Society for Agricultural Machinery",
        5308: "Vrlsoft, Inc.",
        5309: "The William and Mary Quarterly",
        5310: "Mayo Clinic Scientific Press",
        5311: "Journal of Spatial Information Science",
        5312: "Baishideng Publishing Group Co (World Journal of Orthopedics)",
        5313: "Baishideng Publishing Group Co (World Journal"
              "of Anesthesiology)",
        5314: "Baishideng Publishing Group Co (World Journal of Dermatology)",
        5315: "Baishideng Publishing Group Co (World Journal of Hematology)",
        5316: "Baishideng Publishing Group Co (World Journal of Neurology)",
        5317: "Baishideng Publishing Group Co (World Journal"
              "of Obstetrics and Gynecology)",
        5318: "Baishideng Publishing Group Co (World Journal"
              "of Opthalmology)",
        5319: "Baishideng Publishing Group Co (World Journal"
              "of Otorhinolaryngology)",
        5320: "Baishideng Publishing Group Co (World Journal of Respirology)",
        5321: "Baishideng Publishing Group Co (World Journal of Stomatology)",
        5322: "The Korean Environmental Science Society",
        5323: "Association for the Study of African American"
              "Life and History",
        5324: "Norwegian University of Science and Technology Library",
        5325: "The Pennsylvania State University Press",
        5326: "American Animal Hospital Association",
        5327: "Zeppelini Editorial e Comunicacao",
        5328: "Association for Career and Technical Education Research",
        5329: "FACECLA",
        5330: "American School Counselor Association",
        5331: "Japanese Society of Snow and Ice",
        5332: "Japan Society of Human Growth and Development",
        5333: "The Korean Society of Grassland and Forage Sciences",
        5334: "Ubiquity Press, Ltd.",
        5335: "UPF Editora",
        5336: "Turkiye Klinikleri",
        5337: "International Water Management Institute",
        5338: "The Korean Society of Environmental Agriculture",
        5339: "Hamad bin Khalifa University Press (HBKU Press)",
        5340: "Minnis Journals Pty Ltd.",
        5341: "CO2CRC",
        5342: "Michigan Historical Review",
        5343: "Bulletin of Marine Science",
        5344: "American Society for Enology and Viticulture",
        5345: "The Korean Institute of Building Construction",
        5346: "Japan Society of Spring Engineers",
        5347: "Instituto Tecnologico Autonomo de Mexico",
        5348: "EDORIUM Journals, Pvt., Ltd.",
        5349: "Center for Law and Policy",
        5350: "Yerkure Tanitim ve Yayincilik Hizmetleri",
        5351: "The Korean Statistical Society",
        5352: "Korean Society of Life Science",
        5353: "The University of Hong Kong Libraries",
        5354: "Universidad de Chile",
        5355: "Korea Basic Science Institute",
        5356: "The Society for Near Eastern Studies in Japan",
        5357: "The Stomatological Society, Japan",
        5358: "Herpetological Society of Japan",
        5359: "Japan Association for Wind Engineering",
        5360: "The Membrane Society of Japan",
        5361: "The Medical Society of Kansai Medical University",
        5362: "The Japanese Society for the History of Economic Thought",
        5363: "Japan Science Support Foundation",
        5364: "The Press at California State University, Fresno",
        5365: "World Health Organization, Western Pacific Regional Office",
        5366: "Malaysian Palm Oil Council",
        5367: "IP Publishing",
        5368: "Dentaquest",
        5369: "The Korean Sensors Society",
        5370: "The Korean Institute of Electrical Engineers",
        5371: "The Korean Hip Society",
        5372: "Faculty of Medicine, Chulalongkorn University,"
              "Asian Biomedicine",
        5373: "Institute of Advanced Scientific Research",
        5374: "Radiologia.hu Kft.",
        5376: "Sophia Publishing Group, Inc.",
        5377: "Latin America Journals Online",
        5378: "Indiana Magazine of History",
        5379: "Urban Planning Institute of the Republic of Slovenia",
        5380: "Universidade Federal do Parana",
        5381: "AITO - Association Internationale pour les"
              "Technologies Objets",
        5382: "Society of Economic Geologists",
        5383: "International Association for Sharing Knowledge"
              "and Sustainability",
        5384: "Norwegian Association of Optometry",
        5385: "The Korean Society of Neonatology",
        5386: "The Japanese Association of Indian and Buddhist Studies",
        5387: "The Fukushima Society of Medical Science",
        5388: "Korean Oncology Nursing Society",
        5389: "The Korean Society of Agricultural Engineers",
        5390: "Sustainable Building Research Center",
        5391: "Korean Institute of Intelligent Systems",
        5392: "The Korea Contents Association",
        5393: "Korean Association of Agriculture Medicine"
              "and Community Health",
        5394: "Korean Institute of Navigation and Port Research",
        5395: "The Korean Academy of Conservative Dentistry",
        5396: "Korean Society of Plastic and Reconstructive Surgeons",
        5397: "Korean Shoulder and Elbow Society",
        5398: "Media Peternakan",
        5399: "Oregon State University",
        5400: "Journal of Tropical Soils",
        5401: "Australian and New Zealand Society of the History of Medicine",
        5402: "Hindawi (International Scholarly Research Network)",
        5403: "Oregon Historical Society",
        5404: "Chinese Civilisation Centre",
        5405: "Taiwanese Society of Biomedical Engineering",
        5406: "University of Illinois Press",
        5407: "The Korean Society of Visualization",
        5408: "National Association of Geoscience Teachers",
        5409: "Baishideng Publishing Group Co (World Journal"
              "of Clinical Pediatrics)",
        5410: "Baishideng Publishing Group Co (World Journal"
              "of Clinical Urology)",
        5411: "Baishideng Publishing Group Co (World Journal of Immunology)",
        5412: "Baishideng Publishing Group Co (World Journal"
              "of Surgical Procedures)",
        5413: "MEDICRIT",
        5414: "Dustri-Verlag Dr. Karl Feistle",
        5415: "Asia Pacific Association of Allergy, Asthma,"
              "and Clinical Immunology",
        5416: "The Japan Institute of Power Electronics",
        5417: "Japan Association for Comparative Politics",
        5418: "ANPEGE - Revista",
        5419: "Brazilian Journal of Petroleum and Gas",
        5420: "University of Central Lancashire",
        5421: "Turkish Society of Public Health Specialists",
        5422: "Fordham University Press",
        5423: "Korean Society of Plant Pathology",
        5424: "Instituto Nacional de Investigacion y Tecnologia"
              "Agraria y Alimentaria",
        5425: "International Journal of Land Systems",
        5426: "The Japan Laryngological Association",
        5427: "Journal of Singularities",
        5428: "Publishing House of Pharmaceutical Care and Research",
        5429: "The International Association for the Study of Popular Music",
        5430: "Sciedu Press",
        5431: "Arachnologische Gesellschaft e.v.",
        5432: "Japan Society of Physical Education, Health"
              "and Sport Sciences",
        5433: "Universidade Estadual de Londrina",
        5434: "University of Debrecen/ Debreceni Egyetem",
        5435: "American Academy of Orthopaedic Surgeons",
        5436: "Dental Hypotheses",
        5437: "The Industrial Research Interchange",
        5450: "University of Bern",
        5453: "Japan Health Physics Society",
        5454: "Indonesian Society for Microbiology",
        5455: "ScopeMed",
        5456: "The Open University",
        5457: "University Clinical Centre Tuzla",
        5458: "The Japanese Society of Applied Glycoscience",
        5459: "New Zealand Society for Earthquake Engineering",
        5460: "Asociacion Mexicana de Comportamiento y Salud",
        5461: "Sociedad Mexicana de Medicina Conductual",
        5462: "The Japanese Psychological Association",
        5463: "VU E-Publishing",
        5465: "Academy of Management",
        5466: "Nihon University School of Dentistry at Matsudo"
              "Research Institute of Oral Science",
        5467: "The Korean Earth Science Society",
        5468: "Korean Society of Obstetrics and Gynecology",
        5469: "Korean Society of Interventional Neuroradiology",
        5470: "Pervasive Technology Institute",
        5471: "Institute for Information Policy",
        5472: "Marmara Medical Journal",
        5473: "VitJournals",
        5474: "Geological Survey of Slovenia",
        5476: "Faculty of Information Quarterly",
        5477: "Centro de Investigaciones Sociologicas",
        5478: "Korean Society Mass Spectrometry",
        5479: "Smithsonian Institution",
        5480: "National League for Nursing",
        5481: "Graduate School Khon Kaen University",
        5482: "Thieme (Hamostaseologie)",
        5483: "Korean Society for Biochemistry and Molecular"
              "Biology - BMB Reports",
        5484: "University of Debrecen/ Debreceni Egyetem,"
              "Debrecen University Press",
        5485: "University of Debrecen/ Debreceni Egyetem -"
              "Teaching Mathematics and Computer Science",
        5486: "University of Debrecen/ Debreceni Egyetem -"
              "Publicationes Mathematicae Debrecen",
        5487: "The Korean Society of Toxicology",
        5488: "Institute for Condensed Matter Physics",
        5489: "Canadian Urological Association Journal",
        5490: "GeroFam Journal",
        5491: "Occupational Safety and Health Research Institute, KOSHA",
        5492: "Baishideng Publishing Group Co (World Journal"
              "of Critical Care Medicine)",
        5493: "Baishideng Publishing Group Co (World Journal"
              "of Experimental Medicine)",
        5494: "Baishideng Publishing Group Co (World Journal"
              "of Hypertension)",
        5495: "Baishideng Publishing Group Co (World Journal"
              "of Clinical Infectious Diseases)",
        5496: "Baishideng Publishing Group Co (World Journal"
              "of Medical Genetics)",
        5497: "Baishideng Publishing Group Co (World Journal"
              "of Pharmacology)",
        5498: "Baishideng Publishing Group Co  (World Journal of Psychiatry)",
        5499: "Baishideng Publishing Group Co (World Journal"
              "of Rheumatology)",
        5500: "Baishideng Publishing Group Co (World Journal"
              "of Transplantation)",
        5501: "Baishideng Publishing Group Co (World Journal of Virology)",
        5502: "International Journal fo Wellbeing",
        5503: "St. Plum-Blossom Press Pty, Ltd.",
        5504: "Diagnosis Press, Ltd.",
        5505: "LookUs Bilisim A.S.",
        5506: "Jagiellonian University",
        5507: "Palacky University Olomouc",
        5508: "Journal of Hebrew Scriptures",
        5509: "Pacific Affairs",
        5510: "Oil Gas Scientific Research Project Institute",
        5511: "Japanese Society for Plant Cell and Molecular Biology",
        5512: "Japan Society for Southeast Asian Studies",
        5513: "Journal of Central European Agriculture",
        5514: "Sociedad Mexicana de Analisis de la Conducta",
        5515: "Korean Institute of Electromagnetic Engineering and Science",
        5516: "Korean Nuclear Society",
        5527: "Baishideng Publishing Group Co (World Journal of Nephrology)",
        5528: "Baishideng Publishing Group Co (World Journal"
              "of Translational Medicine)",
        5529: "The Chinese Society of Theoretical and Applied Mechanics",
        5530: "EManuscript Services",
        5531: "American Museum of Natural History",
        5532: "Korean Society of Agricultural and Forest Meteorology",
        5533: "Editora da Universidade Federal Fluminense",
        5534: "Korean Society for Sexual Medicine and Andrology",
        5535: "Korean Academy of Rehabilitation Medicine",
        5536: "The Korean Society of Poultry Science",
        5537: "BioOne (Edentata)",
        5538: "Editorial Glosa, S.L.",
        5539: "Canadian Center of Science and Education",
        5540: "Brazilian Society for Computational and Applied Mathematics",
        5541: "International Centre for Applied Thermodynamics",
        5542: "International Health Academy",
        5543: "Turkish Society of Cardiology",
        5545: "Faculty of Mechanical Engineering",
        5546: "Sociedad Argentina de Pediatria",
        5547: "International Association for Energy Economics",
        5548: "National Institute of Practical Analytical Meditation",
        5549: "International Journal of Students' Research",
        5550: "Faculty of Physical Education and Sport",
        5551: "Japan Atherosclerosis Society",
        5552: "Faculty of Forestry, University of Zagreb",
        5553: "Boom Uitgevers Den Haag",
        5554: "Sociedad Colombiana de Anestesiologia y Reanimacion",
        5556: "Tamkang Journal of Mathematics",
        5557: "Colegio Oficial de Bibliotecarios y Documentalistas"
              "de la Comunidad Valenciana",
        5558: "Canadian Institute of Forestry",
        5559: "Institute of Social Sciences Ivo Pilar",
        5560: "Verlag der Zeitschrift fuer Naturforsch",
        5561: "GEMS - Gender, Education, Music and Society",
        5562: "Croatian Chemical Society",
        5563: "Speleobiology Notes",
        5564: "Mongolia Journals Online",
        5565: "Universitat Autonoma de Barcelona",
        5566: "Slovenian Society for Stereology and Quantitative"
              "Image Analysis",
        5567: "Insight Knowledge",
        5568: "Centro Sperimentale Interuniversitario di Ricerca"
              "Stradale-Aeroportuale",
        5569: "SIIS - Centro de Documentacion y Estudios",
        5570: "European Defendology Center",
        5571: "The Editorial Committee of Synthesiology",
        5572: "Korean Society for Atmospheric Environment",
        5573: "The Institute of Electronics Engineers of Korea",
        5574: "Korean Society of Ocean Engineers",
        5575: "The Geological Society of Japan",
        5577: "International Dental Research",
        5578: "Bilimsel Tip Publishing House",
        5579: "Revista Neuropsicologia Latinoamericana",
        5580: "Internet Scientific Publishers, LLC",
        5581: "Revista Brasileira de Hematologia e Hemoterapia",
        5582: "International Research and Cooperation Association"
              "for Bio & Socio-Sciences Advancement",
        5584: "Proteomass",
        5585: "University Nove de Julho - Uninove",
        5586: "Polish Botanical Society",
        5587: "Machine Intelligence Research Labs",
        5588: "International Union Against Tuberculosis and Lung Disease",
        5589: "Canadian Aeronautics and Space Institute",
        5590: "Walden University",
        5592: "Faculty of Civil Engineering, University of Zagreb",
        5593: "STEF92 Technology",
        5594: "Society of Motion Picture and Television Engineers",
        5595: "Journal of Integrated Disaster Risk Management",
        5596: "Canadian Health Libraries Association",
        5597: "Sociedad Latinoamericana de Especialistas en"
              "Mamiferos Acuaticos",
        5598: "International Mycologica Association",
        5599: "International Association of Physical Chemists",
        5600: "Office of Research, Development and Information",
        5601: "Polish Society for Magnesium Research",
        5602: "Journal of Orofacial and Health Science",
        5603: "VM Media Sp zo.o. - VMGroup SK",
        5604: "Index Copernicus International",
        5605: "Insitututo Estudios Bursatiles",
        5606: "Baycinar Tibbi Yayincilik",
        5607: "The Korean Society for Brain and Neural Science",
        5608: "Journal of Athletic Training/NATA Athletic"
              "Training Education Journal",
        5609: "Japanese Society of Insurance Sciene",
        5610: "Japan Association for Earthquake Engineering",
        5611: "The Japanese Society for Neutron Science",
        5612: "Association for Slavic, East European, and Eurasian Studies",
        5613: "Croatian Sociological Association",
        5614: "The Institute for Research and Community Services ITB",
        5615: "The American Schools of Oriental Research",
        5616: "Asociacion Espanola de Fitosociologia",
        5617: "University of Oslo Library",
        5618: "Canadian Centre of Academic Art and Science",
        5619: "Algorithmic Finance, LLC",
        5620: "The Korean Society of Environmental Health and Toxicology",
        5621: "SF-TH, Inc.",
        5622: "University of Illinois Press",
        5623: "Canadian Institute of Geomatics",
        5624: "Korean Academy of Oral and Maxillofacial Radiology",
        5625: "Korean Association for Laboratory Animal Science (KAMJE)",
        5626: "Korean Institute of Information Scientists and Engineers",
        5627: "The Korean Academy of Clinical Electrophysiology",
        5628: "Faculdade de Desporto",
        5629: "Japanische Hegel-Gesellschaft",
        5630: "Japan Academy of Nursing Science",
        5631: "The Society of Practical Otolaryngology",
        5632: "Japanese Institute of Landscape Architecture",
        5633: "Medical Faculty, University in Nis",
        5634: "Journal of Internet Technology",
        5635: "The Korean Society of Systematic Zoology",
        5636: "Society of Geomagnetism and Earth, Planetary"
              "and Space Sciences",
        5637: "Japan Association for Urban Sociology",
        5638: "Geographic Information Systems Association of Japan",
        5639: "Pro Pharma Communications International",
        5640: "Insciences Organization",
        5641: "American Alliance for Health, Physical Education,"
              "Recreation and Dance",
        5642: "Claremont Colleges Library",
        5643: "Momentum Press",
        5644: "Academy of Sciences and Arts of Bosnia anad Herzegovina",
        5645: "Maralte, B.V.",
        5646: "The Korean Society of Hypertension (KAMJE)",
        5647: "The Japanese Society of Environmental Education",
        5648: "Japan Society of Immunology. and Allergology"
              "in Otolaryngology",
        5649: "Japanese Society of Pharmaceutical Health Care and Sciences",
        5650: "Japan Oil Chemists' Society",
        5651: "The Japanese Association of Administrative Science",
        5652: "The Japan Society of International Economics",
        5653: "The Korean Society for Reproductive Medicine",
        5654: "Revista ACTA Geografica",
        5655: "Southern Med Review",
        5656: "Korean Society of Applied Entomology",
        5657: "The Korean Society of Fisheries and Aquatic Science",
        5658: "The Korean Society of Wood Science Technology",
        5659: "Architectural Institute of Korea",
        5660: "The Korean Society of Weed Science",
        5661: "Nottingham University Press",
        5662: "Baishideng Publishing Group Co (World Journal of Methodology)",
        5663: "Aboriginal Policy Studies",
        5664: "American Academy of Sleep Medicine",
        5665: "Associated Professional Sleep Societies",
        5666: "Department of Mathematics, Kyungpook National University",
        5667: "Association of Genuine Traditional Korean Medicine",
        5668: "Korean Society of Environmental Health",
        5669: "Peeramed",
        5670: "The Oceanography Society",
        5671: "Croatian Anthropological Society",
        5672: "Institut d'Educacio Fisica de Catalunya",
        5673: "Institute for Social Research Zagreb",
        5674: "Japanese Society for Physiological Psychology"
              "and Psychophysiology",
        5682: "Editura Universitara",
        5685: "The Japanese Society of Plant Morphology",
        5686: "Japanese Society of Zoo and Wildlife Medicine",
        5687: "The Institute of Systems, Control and Information Engineers",
        5688: "American Journal of Pharmaceutical Education",
        5689: "Academic and Scientific Publishers",
        5690: "Kantoh Sociological Society",
        5691: "The Biometric Society of Japan",
        5692: "Societas Neurologica Japonica",
        5693: "Digital Journal of Opthamology",
        5694: "Australasian Medical Publishing Co. Pty. Ltd.",
        5695: "The Korean Institute of Surface Engineering",
        5696: "Blacksmith Institute",
        5697: "Institute of Oceanology, Polish Academy of Sciences",
        5698: "American Epilepsy Society",
        5699: "Modern Humanities Research Association",
        5700: "Business Department, School of Economics, Business"
              "& Accounting",
        5701: "Japanese Society of Pediatric Pulmonology",
        5702: "The Mass Spectrometry Society of Japan",
        5703: "Purdue University Press",
        5704: "Malaysian Orthopaedic Association",
        5705: "Statistica Sinica (Institute of Statistical Science)",
        5706: "Hungarian Association of Psychopharmacologists",
        5707: "Wilolud Journals",
        5708: "Semmelweis Egyetem (EKK Mentalhigiene Intezet)",
        5709: "University of Economics and Human Sciences in Warsaw",
        5710: "Asociacion Paleontologica Argentina",
        5711: "Military Operations Research Society",
        5712: "Sociedade Brasileira de Medicina de Familia e Comunidade",
        5713: "Asian Australasian Association of Animal Production Societies",
        5714: "Korean Carbon Society",
        5715: "Association for Natural Language Processing",
        5716: "World Agroforestry Centre",
        5717: "Korean Society for Exercise Nutrition",
        5718: "Korea Institute of Child Care and Education",
        5719: "Human Geographies",
        5720: "The Korean Society of Community Nutrition",
        5721: "Associazione Italiana di Telerilevamento",
        5723: "Korean Association of Child Studies",
        5724: "The GCSSEPM Foundation",
        5728: "Cornell SEAP",
        5729: "Information Engineering Research Institute",
        5730: "Association for Library Collections & Technical Services",
        5731: "Parenteral Drug Association, Inc.",
        5732: "Editorial Office of Chinese Journal of Cancer",
        5733: "BioOne (Natal Museum of South Africa)",
        5734: "Korean Society of Medical Genetics",
        5735: "BioOne (Finnish Zoological and Botanical Publishing Board)",
        5736: "Japan Association for Real Estate Sciences",
        5737: "Pappin Communications",
        5738: "Japan Association for Landscape Ecology",
        5739: "The Japan Fluid Power System Society",
        5740: "The Journal of AOAC International",
        5741: "Gemological Institute of America",
        5742: "Medi + World International",
        5743: "The American University in Cairo Press",
        5744: "University Press of Florida",
        5745: "Revista de Economia Aplicada",
        5746: "IDP Education Cambodia",
        5747: "Associacao Prudentina de Educacao e Cultura",
        5748: "TECSI",
        5749: "University of Minnesota Press",
        5750: "University of Buckingham Press",
        5751: "Resilience Alliance, Inc.",
        5752: "Pontificia Universidade Catolica de Minas Gerais",
        5753: "Sociedade Brasileira de Computacao - SB",
        5754: "Update Medical Publishing",
        5755: "Kaunas University of Technology (KTU)",
        5757: "The Korean Vacuum Society",
        5758: "The Korean Society for Vascular Surgery",
        5759: "The Japan Society of Computer Aided Sugery",
        5760: "Japan Association for Comparative Economic Studies",
        5761: "Editorial Committee of Annals of Thoracic and"
              "Cardiovascular Surgery",
        5762: "The Korea Academia-Industrial Cooperation Society",
        5763: "The Korean Society of Sports Medicine",
        5764: "The Korean Society of Dyers and Finishers",
        5765: "Korean Academy of Child and Adolescent Psychiatry",
        5766: "Social Entrepreneur Lauch Factor",
        5767: "Academy of Sciences and Arts of the Republic of Srpska",
        5768: "Journal of Applied Optics",
        5769: "ABEAT - Associacao Brasileira de Especialistas"
              "em Alta Tecnologia",
        5770: "Canadian Geriatrics Society",
        5771: "Nomos Verlag",
        5772: "IntechOpen",
        5773: "Faculdade de Economia, Administracao e Contabilidade",
        5774: "Stellenbosch University - SPIL",
        5775: "A. I. Rosu Cultural Scientific Foundation Fundatia"
              "cultural-stiintifica A. I. Rosu",
        5776: "Teachers' Training Faculty",
        5777: "Editora Unicentro",
        5778: "Faculty of Law, University of Tasmania",
        5779: "Hypothesis Journal",
        5780: "The Journal of Bioscience and Medicine",
        5781: "The Korean Welding and Joining Society",
        5782: "Khazar University Press",
        5783: "Revista Internacional de Relaciones Publicas",
        5784: "Stellenbosch University - OriON",
        5785: "Stellenbosch University - Per Linguam",
        5786: "Stellenbosch University - African Zoology",
        5787: "Stellenbosch University - Scientia Militaria",
        5788: "Stellenbosch University - Lexikos",
        5789: "Stellenbosch University - Global Media Journal"
              "African Edition",
        5790: "Hong Kong University Press",
        5791: "Business Valuation Review Journal",
        5792: "The Korean Knee Society",
        5793: "Japan Association on the Environmental Studies",
        5794: "Japanese Society of Oral and Maxillofacial Surgeons",
        5795: "The Japanese Society of Clinical Cytology",
        5796: "The Electrochemical Society of Japan",
        5797: "The Japanese Society for Neuroendovascular Therapy",
        5798: "Dicle Medical Journal/Dicle Tip Dergisi",
        5799: "Association of Health Investigations",
        5800: "Institute of Earth's Crust, Siberian Branch"
              "of Russian Academy of Sciences",
        5801: "Nucleo de Altos Estudos Amazonicos",
        5802: "Cellule MathDoc/CEDRAM",
        5803: "Japanese Society of Food Microbiology",
        5804: "Land and Housing Institute",
        5805: "The Korean Society for Clothing Industry",
        5806: "The Korean Society of Analytical Science",
        5807: "Korean Academic Society of Occupational Health Nursing",
        5808: "Korea Genome Organization",
        5810: "University Press of Kentucky",
        5811: "Western Journal of Emergency Medicine",
        5812: "Kowsar Medical Institute",
        5813: "Institute of Electronic and Information Technology",
        5814: "BioOne (Institute of Geographic Scienes and"
              "Natural Resources Research, Chinese Academy of Sciences)",
        5815: "MECS Publisher",
        5816: "Paradigm Publishers",
        5817: "Masaryk University Press",
        5818: "Association of Reptilian and Amphibian Veterinarians",
        5819: "Springer - (backfiles)",
        5820: "Centers for American Indian and Alaska Native Health",
        5821: "Iniciativa Digital Politecnica",
        5822: "Island Press",
        5823: "Japanese Association for Russian and East European Studies",
        5824: "AJIT-E Academic Journal of Information Technology",
        5825: "Technical Institute Bijeljina",
        5826: "Derm101.com",
        5827: "Centre for Scripture & Theology",
        5829: "International Digital Organization for Scientific Information",
        5830: "Clinics Cardive Publishing",
        5831: "The Homan Mathematical Society",
        5832: "Radiomatix",
        5833: "The Japanese Journal of Gastroenterological Surgery",
        5834: "Japanese Society for Dental Health",
        5835: "Journal of Experimental and Clinical Medicine",
        5836: "Institute of Agricultural Management",
        5837: "Medinews (Cardiology), Ltd.",
        5838: "Educational Research e-Journal",
        5839: "Revista Chilena de Neuropsicologia",
        5840: "Philosophy Documentation Center",
        5841: "Centro de Estudos Latino-Americanos sobre Cultura"
              "e Comunicacao",
        5842: "Stellenbosch University - SPILPlus",
        5843: "Japan Society for Oral Tumors",
        5844: "Japan Society of Marketing and Distribution",
        5845: "Research Society of Buddhism and Cultural Heritage",
        5846: "Acta Ecologica Sinica",
        5847: "World Journal of Emergency Medicine",
        5848: "GSE Research Limited",
        5849: "Society of American Foresters",
        5850: "The Korean Society of Clothing and Textiles",
        5851: "Korean Society for Food Science of Animal Resources",
        5852: "Museum National D'Histoire Naturelle",
        5853: "Korean Stroke Society",
        5854: "International Academy of Physical Therapy Research",
        5855: "The Korean Society for Energy",
        5856: "Korean Academy of Dental Sciences",
        5857: "Korean Society of Photoscience",
        5858: "Archives of Pathology and Laboratory Medicine",
        5859: "The Korean Association of Information Systems",
        5860: "American Library Association",
        5861: "Consortia Academia Publishing",
        5862: "St. Petersburg State Polytechnical University",
        5863: "Pediatric Pharmacy Advocay Group",
        5864: "Canadian Institute of Public Health Inspectors",
        5865: "Research Institute for Knowledge Content Development"
              "& Technology",
        5866: "National Academy of Dentistry",
        5867: "Medwave Estudios Limitada",
        5868: "Nara Institute of Science and Technology",
        5869: "International Association of Astacology",
        5871: "British Academy",
        5872: "Association for the Advancement of Psychological Science",
        5873: "The Japan Society of Christian Studies",
        5874: "The Food System Research Association of Japan",
        5875: "International Journal of Automation and Smart Technology",
        5876: "University Press of Colorado",
        5877: "Nepal Association of Medical Editors",
        5888: "Centers for Disease Control and Prevention",
        5890: "L&H Scientific Publishing LLC",
        5891: "Japan Association of Food Preservation Scientists",
        5892: "Universidade Vale do Rio Verde",
        5893: "Global War Studies",
        5894: "Associacao Portuguesa dos Recursos Hidricos",
        5895: "Institute of West Asian Studies",
        5896: "e-Bangi, Journal of Faculty of Social Sciences and Humanities",
        5897: "Academic Journals",
        5898: "Journal of Human-Robot Interaction",
        5899: "ISPACS GmbH",
        5900: "Sabanci University Information Center",
        5901: "Mediterranean Center of Social and Educational Research",
        5902: "Universidade Federal de Santa Maria",
        5903: "Universitaet Zuerich, Romanisches Seminar",
        5908: "Korean Institute of Fire Investigation",
        5909: "The Korean Institute of Broadcast and Media Engineers",
        5910: "Stellenbosch University - Zambian Journal of Physiotherapy",
        5911: "University of Debrecen/ Debreceni Egyetem - Pedagogia Online",
        5912: "thinkBiotech, LLC",
        5913: "Lockwood Press",
        5914: "Tropical Oceanography",
        5915: "Islamic Medical Association of North America",
        5916: "Journal of Advanced Marine Engineering and Technology",
        5917: "Japanese Association of Groundwater Hyrdology",
        5918: "Japan Agency for Marine-Earth Science and Technology",
        5919: "The Japanese Psychological Association",
        5920: "University of Huddersfield Press",
        5921: "The International Council for Traditional Music",
        5922: "Immanuel Kant Baltic Federal University",
        5923: "Scientific and Academic Publishing",
        5924: "Japanese Society of Animal Breeding and Genetics",
        5925: "The Vitamin Society of Japan",
        5926: "The Japanese Association of Educational Psychology",
        5927: "Japanese Society of Jaw Deformities",
        5928: "The Oceanographic Society of Japan",
        5929: "Administrative Issues Journal: Education, Practice"
              "and Research",
        5930: "Lesgaft National State University of Physical"
              "Education, Sport and Health, St. Petersburg",
        5931: "Dalhousie Journal of Interdisciplinary Management",
        5932: "Korean Society of Public Health Nursing",
        5933: "Korean Academy of Pediatric Dentistry",
        5934: "Korean Association of Human Ecology",
        5935: "GN1 Genesis Network",
        5936: "RNCSB Organization",
        5937: "Centre for Evaluation in Education and Science",
        5938: "The Japanese Society of Swine Science",
        5939: "The Society of Japanese Women Scientists",
        5940: "The Crystallographic Society of Japan",
        5941: "Mycobiology (KAMJE)",
        5942: "American Water Works Association",
        5943: "Mushroom Research Foundation",
        5944: "UNED - Universidad Nacional de Educacion a Distancia",
        5945: "Undergraduate Writing Center - UT Austin",
        5946: "The Korean Society of Gastrointestinal Endoscopy",
        5947: "EURICSE - European Research Institute on Cooperative"
              "and Social Enterprises",
        5948: "Cambridge University Press (Mathematical Association"
              "of America)",
        5949: "Cambridge University Press (CUP)",
        5950: "The MRE Foundation",
        5951: "National Council of Teachers of Mathematics",
        5952: "Stellenbosch University - Nederduitse Gereformeerde"
              "Teologiese Tydskrif",
        5953: "Korean Society of Muscle and Joint Health",
        5954: "ALife Robotics Corporation Ltd.",
        5955: "Japan Legal History Association",
        5956: "Society of Environmental Conservation Engineering",
        5957: "The Society of Naval Architects and Marine Engineers",
        5958: "Diva Enterprises Private Limited",
        5959: "KKMed Publications",
        5960: "Deutsche Zeitschrift Fur Sportmedizin/German"
              "Journal of Sports Medicine",
        5961: "Zonguldak Karaelmas University",
        5962: "Smithsonian Institution Biodiversity Heritage Library",
        5963: "The World Academic Publishing",
        5964: "PsychOpen GOLD Leibniz Institute for Psychology (ZPID)",
        5965: "Universidade do Estado de Santa Catarina",
        5966: "Alphamed Press",
        5977: "The Korean Academic Society of Nursing Education",
        5978: "International Phototherapy Association",
        5979: "Japanese Society of Tea Science and Technology",
        5980: "Japanese Urological Association",
        5981: "Japan Society for Head and Neck Cancer",
        5982: "The Academic Society of Japanese Garden",
        5983: "Japan Society for Comparative Endocrinology",
        5984: "Japanese Society for Laser Dentistry",
        5985: "Japan Society for Environmental Chemistry",
        5986: "Japanese Research Association for Food Irradiation",
        5987: "Japan Society of Sport Sociology",
        5988: "The Japan Institute of Marine Engineering",
        5989: "Japan Society for Graphic Science",
        5990: "Japan Wood Preserving Association",
        5991: "American Water Works Association - Opflow",
        5992: "American Academy of Cosmetic Surgery",
        5993: "PNG Publishers",
        5994: "Entomological Society of Indonesia/Perhimpunan"
              "Entomologi Indonesia",
        5995: "The Japan Society of Island Studies",
        5996: "The Geographic Education Society of Japan",
        5997: "Japan Society of Sports Industry",
        5998: "Japan Comparative Education Society",
        5999: "Korean Society of Plastic and Reconstructive Surgeons",
        6000: "Lifescience Global",
        6001: "Lithuanian Academy of Sciences",
        6002: "Baskent University Publishers",
        6003: "Society of Academy of Dermatology in Turkey",
        6004: "Harborside Press, LLC",
        6005: "Fundacao Educacional Aracatuba",
        6006: "Perspectivas Medicas - Faculty of Medicine",
        6007: "Human Resource Management Academic Research Society",
        6008: "Companhia Brasileira de Producao Cientifica",
        6009: "Japanese Society of Radiological Technology",
        6010: "Japan Society of Geoinformatics",
        6011: "Institute for Health Economics and Policy",
        6012: "Japanese Association of Water Resources and Environment",
        6013: "Brewing Society of Japan",
        6014: "Japanese Society of Oral Medicine",
        6015: "Association of Perioperative Registered Nurses",
        6016: "SZD - Slovenian Medical Association",
        6017: "Boston College University Libraries",
        6018: "Servicio de Publicaciones de la Universidad e Murcia",
        6019: "EMBL-EBI",
        6020: "Interscience Place",
        6021: "International Computer Science Company",
        6022: "Center for Music Learning",
        6023: "Shaghai Institute of Organic Chemistry",
        6024: "Marine Biological Association of India",
        6025: "Digital Information Research Foundation",
        6026: "Biomedical Informatics",
        6027: "Nordic Council of Ministers",
        6028: "National Institute of Standards and Technology",
        6029: "The Korea Academia - Industrial Cooperation Society",
        6030: "Association of Experimental Stroke and Translational Medicine",
        6031: "Asian Culture Society International",
        6032: "Japanese Society of Gout and Nucleic Acid Metabolism",
        6033: "The Japanese Association of Special Education",
        6034: "FACCAMP - Faculdade Campo Limpo Paulista",
        6035: "Universitat Jaume I",
        6036: "Publicaciones DYNA",
        6037: "Engineering Technology Press",
        6058: "Korean Association for the Study of Lung Cancer (KAMJE)",
        6059: "Aino University",
        6060: "Ivanovo State University of Chemistry and Technology",
        6061: "Fundacao Faculdade de Medicina",
        6062: "PACIS - Panamerican Association for Computational"
              "Interdisciplinary Sciences",
        6063: "Desafio Singular, Lda",
        6064: "Hindawi  (Scientifica)",
        6065: "Korean Society of Pediatric Endocrinology",
        6066: "Department of Food Science and Technology,"
              "Bogor Agricultural University",
        6087: "Korean Council of Science Editors",
        6088: "Integrated Publishing Association",
        6089: "The Japan Society for Composite Materials",
        6090: "Japan International Research Center for Agricultural Sciences",
        6106: "Korean Institute of Construction Engineering and Management",
        6107: "The Korean Housing Association",
        6108: "The Korean Society of Propulsion Engineers",
        6109: "The Korean Institute of Information and Communication"
              "Sciences",
        6110: "The Society of Air-Conditioning and Refrigeration"
              "Engineers of Korea",
        6111: "The Korea Association of Crystal Growth",
        6112: "Korea Society for Computational Fluids Engineering",
        6113: "The Korean Institute of Power Electrics",
        6114: "The Korean Oriental Medical Opthalmology and"
              "Otolaryngology and Dermatology Society",
        6115: "The Korean Home Economics Association",
        6116: "The Korea Association of Herbology",
        6117: "The Korean Microelectronics and Packaging Society",
        6118: "The Korean Society of Menopause",
        6219: "Medknow Publications",
        6339: "School of Statistics, Renmin University of China",
        6564: "Korean Magnetic Resonance Society",
        7120: "Universities Federation for Animal Welfare",
        7121: "Materials Science International Services GmbH",
        7122: "Society of Petroleum Engineers - Carbon Management"
              "Technology Conference",
        7123: "Springer Science and Business Media LLC",
        7124: "Institute of Molecular Biology and Genetics, NAS Ukraine",
        7125: "Proceedings of the Asia-Pacific Advanced Network",
        7126: "Cumhuriyet University Faculty of Dentistry",
        7127: "INOVAGRI",
        7128: "Hallym University Press",
        7129: "Hokkaido Sociological Association",
        7130: "The Japanese Society of Extra-Corporeal Techology in Medicine",
        7131: "The Chinese Linguistic Society of Japan",
        7132: "Journal of Rainwater Catchment Systems",
        7133: "Japan College of Angiology",
        7134: "Japanese Society of Phlebology",
        7135: "Cambridge University Press (Anthem Press)",
        7136: "Baltic Journal of Comparative and Clinical Systems Biology",
        7137: "KEDGE",
        7138: "AOTA Press (OT Practice Magazine)",
        7139: "AOTA Press",
        7140: "AOTA Press (AOTA Continuing Education)",
        7141: "Japanese Society for Theatre Research, Comparative"
              "Theatre Section",
        7142: "Japan Medical Library Association",
        7143: "Japan Society of Health Evaluation and Promiotion",
        7144: "Japanese Society of Stomatognathic Function",
        7145: "AOTA Press (AOTA Special Interest Section Quarterlies)",
        7146: "Aarhus University Library",
        7147: "Geografares",
        7148: "European Council for Modeling and Simulation",
        7149: "Sociedad Espanola de Optica",
        7150: "Ivyspring International Publisher",
        7151: "Faculty of Mathematics, Computer Science and"
              "Econometrics, University of Zielona Gora",
        7152: "University of Washington Libraries",
        7153: "Element d.o.o.",
        7154: "Geography Department, University of Sao Paulo",
        7155: "Journal of Algorithms and Applications",
        7156: "American Chinese Medical Exchange Society",
        7157: "Journal of Parkinsonism and RLS",
        7158: "Engineers Australia",
        7159: "American Water Works Association, IDA Journal",
        7160: "Czech University of Life Sciences Prague",
        7161: "Anadolu Journal of Agricultural Sciences",
        7162: "Fundacao Otorrinolaringologia",
        7163: "Institute of Geography and Spatial Organization,"
              "Polish Academy of Sciences",
        7164: "Japan Antibiotics Research Association",
        7165: "World Technopolis Association",
        7166: "Stellenbosch University - South African Journal"
              "of Industrial Engineering",
        7167: "Hindawi (Datasets International)",
        7168: "Faculty of Law, Banja Luka University",
        7169: "Adam Mickiewicz University",
        7170: "Senologic International Society",
        7171: "Association of Biomolecular Resource Facilities",
        7172: "University of Warsaw - Faculty of Management",
        7173: "Irish Journal of French Studies",
        7174: "Wyzsza Szkola Handlu I Uslug",
        7175: "Seed SRL",
        7176: "International Institute for Science, Technology and Education",
        7177: "LATEC",
        7178: "Hong Kong STM Publishing Co., Ltd.",
        7179: "Sociedad Iberoamericana de Pedagogia Social",
        7180: "Kosin University College of Medicine",
        7181: "Korean Cleft Palate-Craniofacial Association",
        7182: "North American Transplant Coordinators Organization",
        7183: "Society for American Archaeology",
        7184: "Asociacion Nacional de Historia y Anthropologia"
              "de los Cuidados",
        7185: "European Association of Geochemistry",
        7186: "Geological Society of Malaysia",
        7187: "Kolej Islam Darul Ridzuan",
        7188: "Hungarian Dermatological Society",
        7189: "International Global Health Society",
        7190: "Sheffield Hallam University",
        7191: "University of Massachusetts Medical School",
        7192: "Japan Society of Sport Anthropology",
        7193: "In Quarto - Editions EMS",
        7194: "Revista Especialidades",
        7195: "Asociacion Cientifica ICONO14",
        7196: "South African Medical Association NPC",
        7197: "Cumhuriyet Medical Journal",
        7198: "Associacao Academica de Propriedade Intelectual",
        7199: "Pediatric Oncall",
        7200: "ESIC Business and Marketing School",
        7201: "Asociacion Espanola de Economia Agraria",
        7202: "Consortium Erudit",
        7203: "Universitat de Valencia",
        7205: "AMSUS",
        7206: "Kozminski University",
        7207: "Digital Preservation Coalition",
        7208: "University of Chicago Press",
        7209: "The Carbon Society of Japan",
        7210: "The Robotics Society of Japan",
        7211: "The Japanese Society of Revegetation Technology",
        7212: "Zonguldak Karaelmas University - Karaelmas"
              "Science and Engineering Journal",
        7213: "Pontificia Universidade Catolica do Parana - PUCPR",
        7214: "The Brain Sciences Foundation",
        7215: "International Centre for Trade and Sustainable Development",
        7216: "UCTEA Chamber of Textile Engineers",
        7217: "Hindawi (Conference Papers in Science)",
        7218: "Japanese Political Science Association",
        7219: "Japanese Society of Sport Education",
        7220: "Vytautas Magnus University",
        7221: "National Institute of Japanese Literature",
        7222: "Japan Marketing Academy",
        7223: "Japan Society of Health Support Science",
        7224: "Consortium of Multiple Sclerosis Centers",
        7225: "University of Split, Faculty of Maritime Studies",
        7226: "Department of Forest Managment",
        7227: "Manchester University Press",
        7228: "Manchester University Press",
        7229: "National Science Museum of Korea",
        7230: "Korean Society of Cartoon and Animation Studies",
        7231: "The Korean Society of Oriental Neuropsychiatry",
        7232: "Korean Institute of Industrial Engineers",
        7233: "The Korea Society of Costume",
        7234: "The Korean Society for Composite Materials",
        7235: "Korean Society of Horticultural Science",
        7236: "The Institute of Webcasting, Internet and Telecommunication",
        7237: "Science Journal Publications",
        7238: "Fundacio per la Universitat Oberta de Catalunya",
        7240: "Marmara University Journal of Science",
        7241: "Our Dermatology Online",
        7242: "Institute of Continuous Media Mechanics",
        7243: "Herbert Publications",
        7244: "Cambridge Medicine Journal",
        7245: "Michigan Academy of Science, Arts and Letters",
        7246: "Journal of Peace, Conflict and Development",
        7247: "Journal of Turgut Ozal Medical Center",
        7248: "Japan Rhinologic Society",
        7249: "Rand Corporation",
        7250: "Riga Technical University",
        7251: "National and University Library of the Republic of Srpska",
        7255: "Australian Antarctic Division, Australia Department"
              "of Environment",
        7256: "NB-Media, Ltd.",
        7257: "Society of Urologic Nurses and Associates - SUNA",
        7262: "ESIC Business and Marketing School (REIMKE Journal)",
        7263: "ESIC Business and Marketing School, AdResearch Journal",
        7287: "PeerJ",
        7290: "University of Tennessee",
        7305: "KOREMA",
        7306: "Polish Geological Institute - National Research"
              "Institute (PGI-NRI)",
        7307: "Faculty of Transport and Traffic Sciences",
        7308: "Arquivos em Odontologia",
        7309: "Journal of Mobile Technology in Medicine",
        7310: "The Association for Regional Agricultural and"
              "Forestry Economics",
        7311: "University of Warsaw - Faculty of Modern Languages",
        7312: "Columbia University Press",
        7313: "Cambridge University Press (Nottingham University Press)",
        7314: "Asian Pacific Organization for Cancer Prevention",
        7315: "Society for Computational Design and Engineering",
        7316: "The Korean Hydrogen and New Energy Society",
        7317: "The Polymer Society of Korea",
        7318: "Korean Society of Food Culture",
        7319: "The Korean Society for Geospatial Informational Systems",
        7320: "Fondazione Pro Herbario Mediterraneo",
        7321: "International Journal of Soft Computing and"
              "Software Engineering",
        7322: "NEPAS",
        7323: "International Association of Engineering and Technology",
        7324: "Journal of Applied Pharmacuetical Science",
        7325: "Sociedad Espanola para la Conservacion y Estudio"
              "de los Mamiferos",
        7326: "American College of Physicians",
        7327: "Faculty of Economics, VSB Technical University of Ostrava",
        7328: "Recht Feur Deutschland GmbH",
        7329: "The Japan Institute of International Affairs",
        7330: "University Press of Colorado",
        7331: "Educational Travel Foundation",
        7332: "Immediate Science Ecology",
        7333: "International Association for Computerized Adaptive Testing",
        7336: "Academicus Journal",
        7341: "Fundacja Upowszechniajaca Wiedze i Nauke Cognitione",
        7342: "Associazione Acque Sotterranee",
        7343: "Associazione Acque Sotterranee",
        7358: "Led Edizioni Universitarie",
        7359: "Led Edizioni Universitarie",
        7424: "Elsevier - Central Mining Institute",
        7429: "Professioni Infermieristiche, CNAI",
        7433: "Fondazione CUEIM",
        7434: "Risklab at University of Toronto",
        7435: "Centro Universitario Italo Brasileiro",
        7436: "Omnipax Editora",
        7437: "Brazilian Center for Physican Research",
        7438: "Jurnalul de Chirurgie Journal of Surgery",
        7439: "Scholar Science Journals",
        7440: "Universidad de los Andes",
        7441: "Tomas Bata University in Zlin",
        7442: "International Institute of Design and Service",
        7443: "Problemata: International Journal of Philosophy",
        7444: "Fundacao Instituto de Administracao",
        7445: "Stellenbosch University - Akroterion",
        7446: "Associacao Aerospacial Brasileira",
        7447: "Faculdade de Tecnologia e Ciencias",
        7448: "International AIDS Society",
        7449: "The Minerals, Metals and Materials Society",
        7451: "Canadian Society for Bioengineering",
        7452: "Astro, Ltd.",
        7453: "Global Advances in Health and Medicine",
        7454: "Universitas Indonesia, Directorate of Research"
              "and Public Service",
        7455: "ISEKI Food Association",
        7456: "Turkish Online Journal of Design, Art and Communication",
        7457: "Cellular and Molecular Exercise Physiology",
        7458: "Centro de Investigacao e Estudos de Sociologica",
        7459: "James Nicholas Publishers",
        7460: "The Journal of Tourism and Research",
        7461: "Korean Society of Cerebrovascular Surgeons/Korean"
              "Society of Endovascular Surgery",
        7462: "Society of Exploration Geophysicists",
        7463: "Science and Education of the Bauman MSTU",
        7464: "The Korean Society of Clean Technology",
        7465: "Korean Data and Information Science Society",
        7466: "Korean Home Management Association",
        7467: "The Korean Society of Automotive Engineers",
        7468: "Korea Society of Mathematical Education",
        7469: "The Korean Society for Quality Management",
        7470: "Korean Society of Transportation",
        7471: "Institute of Korean Electrical and Electronics Engineers",
        7472: "Korean Society for Internet Information",
        7473: "The Rubber Society of Korea",
        7474: "Korean Society for Rock Mechanics",
        7475: "Korean Society of Adult Nursing",
        7476: "SciELO",
        7482: "Copernicus GmbH",
        7492: "International Association fo Sustainable Development"
              "and Management",
        7493: "Multiple Sclerosis Discovery Forum",
        7494: "AGH University of Science and Technology Press",
        7498: "Acta Physica Sinica, Chinese Physical Society"
              "and Institute of Physics, Chinese Academy of Sciences",
        7545: "Korea Technology Innovation Society",
        7546: "Prof. Marin Drinov Publishing House of BAS"
              "(Bulgarian Academy of Sciences)",
        7547: "American Podiatric Medical Association",
        7548: "Communications in Information Literacy",
        7549: "Faculty of Economics and Business, University of Maribor",
        7550: "Revista Mexicana de Biodiversidad",
        7551: "MIT Press",
        7552: "Stellenbosch University - African East Asian"
              "Affairs, The China Monitor",
        7553: "Stellenbosch University - South African Journal"
              "of Library and Information Science",
        7554: "eLife Sciences Publications, Ltd.",
        7555: "Journal of Biomedical Research",
        7556: "American Osteopathic Association",
        7557: "UiT The Arctic University of Norway",
        7558: "British Hydrological Society",
        7559: "Universidade Catolica Portuguesa",
        7560: "University of Texas Press",
        7561: "Scientific Annals of Computer Science",
        7562: "Faculty of Occupational Safety, University of Nis",
        7563: "Social Sciences Directory Limited",
        7564: "Baltzer Science Publishers",
        7565: "Language and Psychoanalysis",
        7566: "Physical Society of Japan",
        7567: "Japan Society of Applied Physics",
        7569: "Scrivener Publishing",
        7570: "Korean Society for the Study of Obesity",
        7571: "Japanese Educational Research Association",
        7572: "North Carolina Academy of Science, Inc.",
        7573: "Just Medical Media, Ltd.",
        7574: "Cambridge Journal of International and Comparative Law",
        7575: "Australian International Academic Centre",
        7577: "OsloMet - Oslo Metropolitan University",
        7578: "Fundacion Ciencia Joven",
        7579: "123Doc Education",
        7580: "The Korean Society for Legal Medicine (KAMJE)",
        7581: "Korean Academy of Pediatric Allergy and Respiratory"
              "Disease (KAMJE)",
        7582: "Korean Society of Earth and Exploration Geophysicists",
        7583: "Korea Game Society",
        7584: "Korea Technical Association of the Pulp and Paper Industry",
        7585: "The Korean Society of Pesticide Science",
        7586: "Korean Society of Biological Nursing Science",
        7587: "Korean Academic Society of Rehabilitation Nursing",
        7588: "World Literature Today",
        7589: "Wildlife Disease Association",
        7590: "Paris Legal Publishers",
        7591: "Cornell University Press",
        7592: "Estonian Literary Museum Scholary Press",
        7593: "Department of Biology",
        7594: "Revista da Biologia, Reitoria da Universidade de Sao Paulo",
        7595: "Faculty of Organizational Sciences",
        7596: "Karabuk University",
        7597: "ACOPIOS - Revista Iberica de Mineralogia",
        7598: "Chemical Science Transactions",
        7599: "Hanyang University College of Medicine (KAMJE)",
        7600: "The Japanese Society of Physical Fitness and Sports Medicine",
        7601: "Japan Society of Medical Entomology an Zoology",
        7602: "The Korean Society of Endoscopic and Laproscopic Surgeon",
        7603: "Springer - Global Science Journals",
        7704: "Korean College of Helicobacter and Upper Gastrointestinal"
              "Research",
        7705: "Instituto Nacional de Salud",
        7706: "International Journal of Eye Banking",
        7707: "Hamdan Medical Journal",
        7708: "City Net Scientific Research Center, Ltd. Belgrade",
        7709: "Journal of Negro Education",
        7710: "Pacific University Library",
        7711: "Fondazione Eni Enrico Mattei",
        7712: "ECCOMAS",
        7713: "Indian Journal of Medical Specialties",
        7714: "CNPS",
        7715: "Engineering Science and Technology Digital Library",
        7716: "Advanced Electromagnetics",
        7717: "PeerJ",
        7718: "International Association of Multidisciplinary Research",
        7719: "Philippine Association of Institutions for Research",
        7720: "Antibodypedia",
        7721: "Children, Youth and Environments Center",
        7722: "Boydell & Brewer Group, Ltd.",
        7723: "The Antioch Review",
        7724: "Nucelo de Pesquisa em Ecologia e Biodiversidade",
        7725: "Publications Office of the European Union",
        7726: "Columbia International Publishing",
        7727: "West Indian Medical Journal",
        7728: "Global Journal of Community Psychology Practice",
        7729: "North Atlantic Region Association for Counselor"
              "Education and Supervision",
        7730: "The Society  of Sasang Constitutional Medicine",
        7731: "Korea Institute of Fire Science and Engineering",
        7732: "The Plant  Resources Society of Korea",
        7733: "The Korean Radioactive Waste Society",
        7734: "Computational Structural Engineering Institute of Korea",
        7735: "The Korean Society of Manufacturing Technology Engineers",
        7736: "Korean Society of Precision Engineering",
        7737: "The Korean Operations and Management Science Society",
        7738: "Korean Association of Architectural History",
        7739: "Korean Academy of Fundamentals of Nursing",
        7740: "The Korean Society of Crop Science",
        7741: "The Korean Culture Association",
        7742: "The Korean Society of Radiology (KISTI)",
        7743: "Korean Institute of Educational Environment",
        7744: "CNU Journal of Agricultural Science",
        7745: "Korean Society of Soil Science and Fertilizer",
        7746: "The Korea Robotics Society",
        7747: "Institute of Forest Science, Kangwon National University",
        7748: "RCN Publishing",
        7749: "Dinamia'cet-IUL",
        7750: "Dundee Science Press",
        7751: "Royal Botanica Gardens and Domain Trust",
        7752: "Physical Education and Sport Faculty",
        7753: "Association of Technology and Science",
        7754: "Clinical Laboratory Publications",
        7755: "NMFS Scientific Publications Office",
        7756: "AMS Press, Inc.",
        7757: "Perspectives of New Music",
        7758: "Russell Sage Foundation",
        7759: "Cureus, Inc.",
        7760: "Instituto Juan de Herrera",
        7761: "Zeta Books",
        7762: "Korean Society of Clinical Nutrition",
        7763: "IACSIT Press",
        7764: "Pontificia Universidad Catolica de Chile",
        7765: "Manchester University Press",
        7766: "Martin Eve",
        7767: "Bohlau Verlag",
        7768: "Western Kentucky Libraries",
        7769: "Revista de Gestae e Secretariad",
        7770: "Universidad Catolica de Temuco",
        7771: "Purdue University (bepress)",
        7772: "Pacific University (bepress)",
        7773: "Instituto de Investigaciones Oceanologicas",
        7774: "Korean Vaccine Society",
        7775: "Revista Argentina de Cardiologia",
        7776: "The Acoustical Society of Korea",
        7777: "The Korean Foundrymens Society",
        7778: "The Association of Korean Oriental Pediatrics",
        7779: "The Korean Society for Nondestructive Testing",
        7780: "The Korean Society of Remote Sensing",
        7781: "Korean Society of Steel Construction",
        7782: "Korean Society for Railway",
        7783: "The Korean Society of Medicinal Crop Science",
        7784: "ANPTUR - Associacao Nacional de Pesquisa e"
              "Pos Graducao em Turismo",
        7785: "Adenine Press, Inc.",
        7788: "Bohlau Verlag",
        7790: "Swinburne University of Technology",
        7791: "Sumart Processing Society for Materials, Environment"
              "and Energy",
        7792: "Japanee Society of Limb Salvage and Podiatric Medicine",
        7793: "The Japanese Coronary Association",
        7809: "University of Hamburg, Biodiversitat, Evolution"
              "and Okologie der Pflanzen",
        7810: "Bridget Williams Books",
        7811: "Sociedad Argentina de Radiologia",
        7812: "The Permanente Press",
        7813: "Progress IPS",
        7814: "St. Louis University",
        7815: "White Globe Publications",
        7816: "Sada Institute",
        7817: "American Oriental Society",
        7818: "Asociacion Espanola de Ecologia Terrestre",
        7819: "FECAP Fundacao Escola de Comercio Alvares",
        7820: "Vocabulary Learning and Instruction",
        7821: "University of Alicante",
        7822: "Ondokuz Mayis Universitesi Egitim Fakultesi",
        7824: "Revista Brasileira de Herbicidas",
        7825: "Butler University, Irwin Library",
        7826: "Institute for Social Research",
        7827: "Mehmet Dursun Erdem",
        7828: "Liceo de Cagayan University",
        7829: "CEU Educational-Service Non-profit LLC",
        7830: "Scientific Knowledge Press",
        7831: "United Graduate School of Agricultural Science",
        7832: "Stellenbosch University - Missionalia",
        7833: "Stellenbosch University - Scriptura",
        7834: "Phoenix",
        7835: "Centrum Catolica",
        7836: "The Korean Solar Energy Society",
        7837: "The Korean Society of Marine Environment and Safety",
        7838: "Society for e-Business Studies",
        7839: "Korea Society of Fluid Power and Construction Equipment",
        7840: "Korea Information and Communications Society",
        7841: "Korean Society for Biotechnology and Bioengineering",
        7842: "The Korean Institute of Gas",
        7843: "Korean Geotechnical Society",
        7844: "The Korean Institute of Resources Recycling",
        7845: "The Microbiological Society of Korea",
        7846: "The Korean Society for Marine Environment and Energy",
        7847: "The Korean Society of Fish Pathology",
        7848: "Korean Society of Surveying, Geodesy, Photogrammetry"
              "and Cartography",
        7849: "The Korean Society for New and Renewable Energy",
        7850: "The Korean Society of Oceanography",
        7851: "Korean Society of Rural Planning",
        7852: "Korean Society of Sericultural Science",
        7853: "The Korean Society of Veterinary Service",
        7854: "Petrological Society of Korea",
        7855: "Korean Society of Road Engineers",
        7856: "The Korean Society of Community Living Science",
        7857: "Korean Society of Soil and Groundwater Environment",
        7858: "The Busan Gyeongnam Mathematical Society",
        7859: "Korean Institute of Educational Facilities",
        7860: "JCDR Research and Publications",
        7861: "Royal College of Physicians",
        7862: "Rzeszow University of Technology",
        7863: "American Institute of Ultrasound in Medicine",
        7864: "Brookings Institution Press",
        7865: "Journal of Agronomic Sciences",
        7866: "Instituto de Estudios Fiscales",
        7867: "Fundacao Universidade Regional de Blumenau",
        7868: "Akademizdatcenter Nauka",
        7869: "EMED Publishing",
        7870: "Canadian Periodical for Community Studies",
        7871: "Global Summitry Project, Munk School",
        7872: "BioOne (Association des Amis des Cryptogames)",
        7873: "EDAA",
        7874: "The Korean Audiological Society",
        7876: "Japanese Society for Medical Virtual Reality",
        7877: "Japanese Association of Sign Linguistics",
        7878: "Japan Society of Sarcoidosis and other Granulomatous"
              "Disorders",
        7879: "Society of Indoor Environment Japan",
        7880: "Global Business Research Center",
        7881: "Showa University Dental Society",
        7882: "Royal Zoological Society of New South Wales",
        7883: "Editorial Committee of Japanese Journal of"
              "Infectious Diseases, National Institute of"
              "Infectious Diseases",
        7884: "Turkce Edebiyat Kultur Egitim Dergisi",
        7885: "Kennesaw State University, Horace W. Sturgis Library",
        7886: "Hokkaido Geographical Society",
        7887: "The Japanese Congress of Neurological Surgeons",
        7888: "The University of Occupational and Environmental"
              "Health, Japan",
        7889: "The Japan Society for Hematopoietic Stem Cell Transplantation",
        7895: "International Journal of Alcohol and Drug Research",
        7896: "NAIK Research Institute of Agricultural Economics",
        7897: "Moksha Publishing House",
        7898: "MISTA Conference Series",
        7899: "Brighthall",
        7900: "Theta Foundation",
        7901: "International Oil Spill Conference",
        7902: "Araucaria Comunicacao Integrada",
        7903: "Academy of Taiwan Information Systems Research",
        7904: "Exact Sciences Department, Banat's Journal of Biotechnology",
        7905: "Bogdan Khmelnitsky Melitopol State Pedagogical University",
        7906: "Croatian Interdisciplinary Society",
        7916: "Columbia University Libraries/Information Services",
        7924: "Duke University Libraries",
        7926: "Middlebury College Library",
        7930: "U.S. Global Change Research Program",
        7934: "MorphoBank (Stony Brook University)",
        7939: "University of Alberta",
        9707: "Grand Valley State University Libraries",
        9708: "Korean Society of Computer Information",
        9709: "The Korea Society for Simulation",
        9710: "The Malacological Society of Korea",
        9711: "Korean Tunnelling Association",
        9712: "Korean Association for Spatial Structures",
        9713: "The Korean Institute of Chemical Engineers",
        9714: "The Korea Institute of Applied Superconductivity"
              "and Cryogenics",
        9715: "Korean Institute of Landscape Architecture",
        9716: "The Korea Society of Information Technology Services",
        9717: "Korea Multimedia Society",
        9718: "The Korea Society of Medical and Biological Engineering",
        9719: "The Korean Society of Economic and Environmental Geology",
        9720: "The Korea Society of Engineering Gelology",
        9721: "Korean Society of Food Science and Technology",
        9722: "The Korean Society for the Gifted",
        9723: "The Korea Society for Industrial Systems",
        9724: "Korean Society of Food and Cookery Science",
        9725: "The Korean Society of Tribologists and Lubrication Engineers",
        9726: "The Korea Society for Power System Engineering",
        9727: "The Mineralogical Society of Korea",
        9728: "Digital Contents Society",
        9729: "Korean Society of Electron Microscopy",
        9730: "Department of Nursing University of Michigan-Flint",
        9732: "Revista Brasileira de Estudos Politicos",
        9733: "UCTEA Chamber of Surveying and Cadastre Engineers",
        9734: "Sciencedomain International",
        9735: "Bioinfo Publications",
        9736: "Terapia Manual",
        9737: "History Studies International Journal of History",
        9738: "International College of Surgeons",
        9739: "Damar Cerrahi Dergisi - Turkish Journal of Vascular Surgery",
        9740: "College of Psychiatric and Neurologic Pharmacists",
        9741: "University Libraries, University of Nevada, Las Vegas",
        9742: "EDP Sciences",
        9743: "Grand Canyon University",
        9744: "Petra Christian University",
        9745: "Johns Hopkins Bloomberg School of Public Health,"
              "CEnter for Communication Programs",
        9746: "The Society of Instrument and Control Engineers",
        9747: "Japanese Association for Rural Studies",
        9748: "Institute of Human Culture Studies, Otsuma Womens University",
        9749: "Japan Institute of Navigation",
        9750: "Society of Antiquaries of Scotland",
        9751: "Centre for Theoretical Physics and Astrophysics",
        9752: "Transportation and Marketing Program, AMS, USDA",
        9753: "Coastal Engineering Reserach Council",
        9754: "Webmed Limited",
        9755: "Faculty of Food and Agriculture, United Arab"
              "Emerites University",
        9756: "Bonfring",
        9757: "Rudecolombia",
        9758: "Korean College of Neuropsychopharmacology",
        9759: "Japan Society of Health Promotion and Physical Therapy",
        9760: "Manchester University Press",
        9761: "Journal of Academic Social Science Studies",
        9762: "Digital Games Research Association Japan",
        9763: "Japanese Society for Sensory Evaluation",
        9764: "The Japan Society of Competitive Intelligence",
        9765: "Korean Society of Coastal and Ocean Engineers",
        9766: "The Korea Institute of Military Science and Technology",
        9767: "Bulletin of Chemical Reaction Engineering and Catalysis",
        9768: "IMDA Press",
        9769: "Energy Psychology Press",
        9770: "Entrepreneurship and Sustainability Center",
        9771: "Universidade Federal da Bahia",
        9772: "Japan Society for the Philosophy of Sport and"
              "Physical Education",
        9773: "The Japan Society for Technology and Plasticity",
        9774: "GSE Research Limited",
        9775: "Kafkas University",
        9776: "iSchools",
        9777: "The Company of Scientists",
        9778: "Joule Inc.",
        9779: "Pamukkale University Journal of Education",
        9780: "Laxmi Book Publication",
        9781: "Universidad Internacional de La Rioja",
        9782: "Division of International Special Education"
              "and Services  - DISES",
        9783: "University of Pennsylvania Press",
        9784: "Blay Publishers",
        9785: "Verlag Dr. Otto Schmidt KG",
        9786: "Korea Society of Waste Managment",
        9787: "Korean Society of Breeding Science",
        9788: "Associacao Brasileira de Psicologia",
        9789: "Universidade Federal do Estado do Rio de Janeiro UNIRIO",
        9790: "IOSR Journals",
        9791: "New Zealand Association of Psychotherapists",
        9792: "University of Worcester",
        9793: "The English Linguistic Society of Japan",
        9794: "The Japanese Society of Pediatric Cardiology"
              "and Cardiac Surgery",
        9795: "National Institute of Advanced Industrial Scicence",
        9796: "Turfgrass Society of Korea",
        9797: "The Society of Information Storage Systems",
        9798: "Korean Society of Hazard Mitigation",
        9799: "The Korean Society of Food and Nutrition",
        9876: "Inera test prefix",
        11001: "The Korean Society of Water and Wastewater",
        11002: "The Korean Society of Food Preservation",
        11003: "The Korean GNSS Society",
        11004: "Korean Society for Advanced Composite Structures",
        11005: "Korean Society for Bone and Mineral Research",
        11106: "Korean Thyroid Association",
        11107: "Pan Pacific Association of Input-Output Studies",
        11108: "The Korean Association of Geographic Information Studies",
        11109: "The Korean Society for Livestock Housing and Environment",
        11110: "The Korean Society of Plant Taxonomists",
        11111: "Korean Academy of Nursing Administration",
        11112: "Korea Institute for Structural Maintenance Inspection",
        11113: "Penerbit UTM Press",
        11114: "Redfame Publishing",
        11115: "Caspian Journal of Applied Sciences Research",
        11116: "Leuven University Press",
        11117: "Instituto Brasiliense de Direito Publico",
        11118: "Mendel University  Press",
        11119: "CADIME Centro Andaluz de Informacion de Medicamentos",
        11120: "The Higher Education Academy",
        11121: "International Journal of Optimization and"
               "Control: Theories and Applications",
        11123: "Revista de Contabilidade e Organizacoes",
        11124: "Joanna Briggs Institute",
        11125: "Compassos Coletivos",
        11126: "Stanford University Press",
        11127: "Gokaraju Rangaraju Educational Society",
        11128: "ARGESIM Arbeitsgemeinschaft Simulation News",
        11129: "DETAIL Business Information GmbH",
        11130: "Center for Economic Integration",
        11131: "OMICS - Agial Publishing House",
        11132: "INSPER - Instituto de Ensino e Pesquisa",
        11133: "The Psychological Record",
        11134: "National Center for Biotechnology",
        11135: "International Deep Drawing Research Group",
        11136: "Department of Qur'an and Hadith Studies",
        11137: "Instituto de Geociencias - UFRJ",
        11138: "CIC Edizioni Internazionali",
        11139: "Computer Assisted Language Instruction Consortium CALICO",
        11140: "Departmento Expresion Grafica y Cartografia",
        11141: "Council for British Archaeology",
        11142: "Journal of International Conference on Electrical"
               "Machines and Systems",
        11143: "Geographical Societ of Finland",
        11144: "Editorial Pontificia Universidad Javeriana",
        11145: "Biomath Forum",
        11146: "Stellenbosch University - SAJIE",
        11147: "Stellenbosch University - Acta Germanica",
        11148: "Stellenbosch University - Soil Survey Journal",
        11149: "he Korean Academy of Preventive Dentistry and Oral Health",
        11150: "The Japanese Association for Infectious Diseases",
        11151: "The Japan Society for Educational Sociology",
        11152: "SRUMB - Romanian Society for Ultrasonography"
               "in Medicine and Biology",
        11153: "Japan Academy of Critical Care Nursing",
        11154: "Japanese Association for the Study of Pain",
        11155: "Bangladesh Society for Veterinary Microbiology"
               "and Public Health",
        11156: "AIBR - Asociacion de Antropologos Iberoamericanos en Red",
        11157: "University of Otago Library",
        11158: "Systematic and Applied Acarology Society",
        11159: "Avestia Publishing",
        11160: "Asociacion Herpetologica Espanola",
        11161: "Korean Society of Odor Research and Engineering",
        11301: "Japan Society for Food Engineering",
        11560: "Japanese Association of Health Pyschology",
        11562: "Stellenbosch University - South African Economic"
               "History Annual",
        11563: "Stellenbosch University - Wound Healing Southern Africa",
        11564: "Stellenbosch University - African Population Studies",
        11565: "Chimera Innova Group",
        11566: "Chattagram Maa-O-Shishu Hospital Medical College",
        11567: "Institut Za Migracije I NarodnostiInstitute"
               "for Migration and Ethnic Studies",
        11568: "The Kangwon-Kyungki Mathematical Society",
        11569: "Baishideng Publishing Group Co (World Chinese"
               "Journal of Digestology)",
        11585: "University Library J. C. Senckenberg",
        11590: "Department of International Relations",
        11591: "Institute of Advanced Engineering and Science",
        11592: "Sia Tibit",
        11593: "Pediatric Endosurgery Group, Bangladesh",
        11594: "Galaxy Science",
        11595: "Japanese Society of Sheep Science",
        11596: "Japanese Association of Occupational Therapists",
        11597: "Korean Society for Indoor Environment",
        11598: "Seameo Biotrop",
        11599: "European HIV/AIDS and Infectious Diseases Academy",
        11600: "Fundacion Cinde",
        11601: "International Science and Engineering Society",
        11602: "Faculdade Independente do Nordeste",
        11603: "Ternopil State Medical University",
        11604: "Pan African Medical Journal",
        11605: "Nature Research Society",
        11606: "Universidade de Sao Paulo, Agencia USP de"
               "Gestao da Informacao Academica (AGUIA)",
        11607: "Quintessence Publishing",
        11608: "Croatian Nurses Association",
        11609: "Wildlife Information Liaison Development Society",
        11610: "Procon, Ltd.",
        11611: "Yonetim ve Ekonomi Arastirmalari Dergisi -"
               "Journal of Management and Economics Research",
        11612: "Res Philosophica",
        11613: "Croatian Society for Medical Biochemistry"
               "and Laboratory Medicine",
        11614: "Korean Society of Limnology",
        11615: "Christ University Bangalore - CU Journal of"
               "Non-linear Fluid Mechanics",
        11616: "Abant Izzet Baysal University Graduate School"
               "of Social Sciences",
        11617: "The Agricultural Economic Society of Tohoku",
        11618: "The Adhesion Society of Japan",
        11619: "Japan Association for African Studies",
        11620: "Korean Academy of Oral Biology",
        11621: "Russian Psychological Society",
        11622: "Singapore Medical Journal",
        11623: "Korean Society for Floricultural Science",
        11624: "Korea Society of Nature Conservation",
        11625: "Korean Association of Organic Agriculture",
        11626: "Korean Society of Environmental Biology",
        11627: "The Society of Korea Industrial and Systems Engineering",
        11628: "Korean Society for People, Plants and Environment",
        11629: "Korean Association for Particle and Aerosol Research",
        11630: "Institute for the Theology of Culture: New"
               "Wine, New Wineskins",
        11632: "Pietro Cipresso",
        11633: "European Paediatric Association",
        11634: "World Scholars",
        11635: "Om Vindhyavasini College of IT and Management",
        11636: "Geoscience Australia",
        11637: "Korean Association of Physical Anthropologists",
        11638: "Japanese Society for Surgical Metabolism and Nutrition",
        11639: "Society of Japan Science Teaching",
        11640: "Research Institute for Economics and Business"
               "Administration - Kobe University",
        11641: "Japan Gastrointestinal Endoscopy Society, Kanto Chapter",
        11642: "The Kyushu Association of Neuropsychiatry",
        11643: "Belgrade Centre for Security Policy",
        11644: "Korea Institute for International Economic Policy",
        11645: "CILIP Information Literacy Group",
        11646: "Magnolia Press",
        11647: "Open Book Publishers",
        11648: "Science Publishing Group",
        11649: "Institute of Slavic Studies of the Polish"
               "Academy of Sciences",
        11650: "The Mathematical Society of the Republic of China",
        11948: "Wilmington Scientific Publisher, LLC",
        12651: "The National Institute of Biological Resources",
        12652: "Korean Society of Civil Engineers",
        12653: "Korean Association of Agricultural Extension",
        12654: "The Korean Society of Conservation Science"
               "for Cultural Heritage",
        12655: "Korean Association for Disability and Health",
        12656: "The Korean Society for Heat Treatment",
        12657: "Bogucki Wydawnictwo Naukowe",
        12658: "Ilmi Etudler Dernegi - ILEM",
        12659: "International Scientific Information, Inc.",
        12660: "Fundacao Getulio Vargas",
        12661: "Instituto Agronomico de Pernambuco",
        12662: "Instituto para o Desenvolvimento da Educacao",
        12663: "SZVT Research and Development Center",
        12664: "Mycobiota",
        12665: "BioProcessing Journal",
        12666: "University of Debrecen/ Debreceni Egyetem - Novenytermeles",
        12667: "Japanese Society for Histocompatibility and Immunogenetics",
        12668: "Institute for Policy Studies and Social Innovation,"
               "University of Hyogo",
        12669: "Pakistan Journal of Medical Sciences",
        12670: "Technical Institute Bijeljina",
        12671: "The Korean Fracture Society",
        12672: "Korea Spatial Information Society",
        12673: "The Korea Navigation Institute",
        12674: "Korean Research Society of Physical Therapy",
        12675: "Emerging Theatre Research",
        12676: "China Communications Magazine, Co., Ltd.",
        12677: "Hans Publishers",
        12678: "J. Michael Ryan Publishing",
        12679: "Alliance for Children and Families",
        12680: "Romanian Association of Balneology",
        12681: "National Documentation Centre (EKT)",
        12684: "University of Bern",
        12685: "University of Basel",
        12687: "Thieme (Phlebologie)",
        12688: "F1000 Research Ltd",
        12689: "Reversal Theory Society",
        12690: "Defence Counsel Journal",
        12691: "Science and Education Publishing Co., Ltd.",
        12692: "International Journal of Biosciences",
        12693: "Institute of Physics, Polish Academy of Sciences",
        12694: "Scalable Computing: Practice and Experience",
        12695: "School of Business and Management, Institut"
               "Teknologui Bandung",
        12696: "EDUGAIT Press",
        12697: "University of Tartu",
        12698: "Consortium for Policy Research in Education",
        12699: "French-Vietnamese Association of Pulmonology",
        12700: "Obuda University",
        12701: "Yeungnam University College of Medicine",
        12702: "Byte Systems - Solucoes Digitais",
        12703: "Faculty Opinions Ltd",
        12704: "Institute of Systems Biology",
        12705: "International Association for Plant Taxonomy",
        12706: "Asociacion Interprofesional Desarrollo Agrario AIDA",
        12707: "Health Sciences Research Unit: Nursing",
        12708: "Delinat Instsitute",
        12709: "MESTE NGO",
        12710: "Russian New University",
        12711: "Iktisadi Girisim ve Is Ahlaki Dernegi (IGIAD)",
        12712: "Departamento de Empreendedorismo e Gestao da UFF",
        12713: "Two Golden Rules Publishing",
        12714: "Ege University Faculty of Fisheries",
        12715: "Healthy Aging Research (OMICS)",
        12716: "Faculty of Navigation",
        12717: "Korean Society of Developmental Biology",
        12718: "Institute of Animal Resources",
        12719: "Korean Society of International Agriculture",
        12720: "Engineering and Technology Publishing",
        12721: "Revista de Tecnologia da Informacao e Comunicacao",
        12722: "Revista Educacao Agricola Superior - ABEAS",
        12723: "Christ University Bangalore - Mapana Journal of Sciences",
        12724: "Christ University Bangalore - Artha Journal"
               "of Social Sciences",
        12725: "Christ University Bangalore - Ushus Journal"
               "of Business Management",
        12726: "Christ University Bangalore - Tattva Journal of Philosophy",
        12727: "Christ University Bangalore - Atna Journal"
               "of Tourism Studies",
        12728: "Christ University Bangalore - Christ University Law Journal",
        12729: "Research Institute of Veterinary Medicine",
        12730: "Bursa Ilahiyat Foundation",
        12731: "Science and Innovation Center",
        12732: "Academic Publications",
        12733: "Binary Information Press",
        12734: "International Engineering Journal",
        12735: "Science and Education Centre of North America",
        12736: "Energa SA",
        12737: "Infra-M Academic Publishing House",
        12738: "Egitim Danismanligi ve Arastirmalari Merkezi - EDAM",
        12739: "NWSA Academic Journals",
        12740: "Komitet Redakcyjno - Wydawniczy Polskiego"
               "Towarzystwa Psychiatrycznego",
        12741: "Entomologistas do Brasil",
        12742: "Samtid",
        12743: "Quanta",
        12744: "International Association of Applied Neuroscience",
        12745: "Early Theatre",
        12746: "The Southwest Respiratory and Critical Care"
               "Chronicles - SWRCCC",
        12747: "Journal Review Foundation",
        12748: "Usak University Journal of Material Sciences",
        12749: "Korean Society of Animal Reproduction",
        12750: "The Korean Society of Animal Reproduction"
               "and Biotechnology (KSARB)",
        12765: "German Federal Institute for Population Research",
        12771: "The Ewha Medical Journal, Ewha Womans University"
               "College of Medicine",
        12772: "The Korean Fiber Society",
        12773: "The Asian Association of World Historians",
        12774: "CABI Publishing - Evidence on Demand",
        12775: "Uniwersytet Mikolaja Kopernika/Nicolaus Copernicus"
               "University",
        12776: "Faculty of Materials, Metallurgy and Recycling",
        12777: "Faculty of Engineering Diponegoro University",
        12778: "The Political Science Association of Kasetsart University",
        12779: "Korean Dementia Association",
        12780: "Usak Universitesi Sosyal Bilimler Dergisi",
        12781: "AI Practitioner Journal",
        12782: "Japanese Society of Systematic Zoology",
        12783: "DEStech Publications",
        12784: "New Zealand College of Midwives",
        12785: "Deanship of Scientific Research",
        12786: "Korean Society for Neurorehabilitation",
        12787: "Karadeniz Arastirmalari Merkezi",
        12788: "Frontline Medical Communications, Inc.",
        12789: "Geological Association of Canada",
        12790: "Korean Society for Surgery of the Hand",
        12791: "The Korean Society for Bio-Environment Control",
        12792: "Institute of Industrial Applications Engineers",
        12793: "Korean Society for Clinical Pharmacology and Therapeutics",
        12794: "University of North Texas Libraries",
        12795: "Editorial Universidad de Sevilla",
        12796: "Centre for Advancement in Inclusive and Special Education",
        12797: "Ksiegarnia Akademicka Sp. z.o.o.",
        12798: "Revista Fluminense de Odontologia  - International"
               "Journal of Science Dentistry  - RFO",
        12799: "Korean Academy of Community Health Nursing",
        12800: "UCAM CCD",
        12801: "Dancecult: Journal of Electronic Dance Music Culture",
        12802: "Revista Latinoamericana e Investigacion en"
               "Matematica Educativa RELIME",
        12803: "The Scientific Society Akroasis",
        12804: "Colegio Mayor de Nuestra Senora del Rosario",
        12805: "Stanford Division of Pain Management",
        12806: "Journal of Leadership Education",
        12807: "University of Western Sydney SOHACA",
        12808: "Bilgi Terecem",
        12809: "Hong Kong Academy of Medicine Press",
        12811: "Korea Society of Health Service Management",
        12812: "Korea Safety Management and Science",
        12813: "Korea Institute of Ecological Architecture and Environment",
        12814: "Korean GeoSynthetics Society",
        12815: "The Korea Institute of Intelligent Transport Systems",
        12816: "Al Manhal FZ, LLC",
        12817: "Recent Science Publications",
        12818: "Revista da Faculdade de Direito da UFMG",
        12819: "Revista FSA",
        12820: "Brazilian Society of Physical Activity and Sport",
        12845: "CNBOP-PIB Centrum Naukowo-Badawcze Ochrony Przeciwpozarowej",
        12871: "Pisa University Press",
        12889: "SMC Media",
        12890: "SMC Media",
        12891: "IMR Press",
        12892: "IMR Press",
        12911: "Wydawnictwo Naukowe Gabriel Borowski (WNGB)",
        12912: "Wydawnictwo Naukowe Gabriel Borowski (WNGB)",
        12913: "Wydawnictwo Naukowe Gabriel Borowski (WNGB)",
        12921: "ICHB PAS Poznan Supercomputing and Networking Center",
        12922: "Human Performance Corsortium",
        12923: "Medical University of Lublin",
        12924: "Librello",
        12925: "Korean Oil Chemists' Society",
        12926: "American Society of Group Psychotherapy and"
               "Pyschodrama - ASGPP",
        12927: "Longwoods Publishing",
        12928: "Universitas Ahmad Dahlan",
        12929: "Journal of Literature and Science",
        12930: "National Academic Advising Association NACADA",
        12931: "Osterreichish-Rumanischer Akademischer Verein",
        12932: "Allergy, Asthma, and Immunology Association of Thailand",
        12933: "Asociacion Mexicana de Mastozoologia A.C.",
        12934: "The Korean Academy of Psychiatric and Mental Health Nursing",
        12935: "Japan Veterinary Medical Association",
        12936: "Tenri Foundation, Tenri Institute of Medical Research",
        12937: "Japanese Society for Information and Systems in Education",
        12938: "BMFH Press",
        12939: "The Korean Society of Fisheries Business Administration",
        12940: "The Korean Society of Fashion Business",
        12941: "The Korean Society for Industrial and Applied Mathematics",
        12942: "Living Reviews",
        12943: "Canadian Nuclear Laboratories Limited",
        12944: "Enviro Research Publishers",
        12945: "Science International Corporation",
        12946: "Max Planck Institute for Legal History and Legal Theory",
        12947: "EarthSpace",
        12948: "ECO-INFOSOC Research Center",
        12949: "Japan Association of Real Options and Strategy",
        12950: "Japanese Society of Radiation Safety Management",
        12951: "Japan Veterinary Cancer Society",
        12952: "BioOne (Elementa)",
        12953: "Revista de Ciencia da Madeira",
        12954: "Asian Pacific Prostate Society",
        12955: "CBU Research Institute s.r.o.",
        12956: "Turkish Journal of Pediatric Disease",
        12957: "Universidade do Estado do Rio de Janeiro UERJ",
        12958: "State Institution Luhansk Taras Shevchenko"
               "National University",
        12959: "Zalozba Vega, d.o.o.",
        12960: "Professional Association of Social Workers of Madrid",
        12961: "Associacio catalana de Salut Laboral",
        12962: "Lembaga Penelitian dan Pengabdian kepada Masyarakat ITS",
        12963: "Korean Academy of Speech-Language Pathology and Audiology",
        12964: "Korean Society of Sandplay Therapy",
        12965: "Korean Society of Exercise Rehabilitation",
        12966: "Sciknow Publications",
        12967: "International Society for Music Education",
        12968: "Mark Allen Group",
        12969: "Istituto Scientia",
        12970: "Synergy Publishers",
        12971: "Revista Agrotecnologia",
        12972: "Docuhut Publishing",
        12973: "Eurasian Society of Educational Research",
        12974: "Savvy Science Publisher",
        12975: "Rast Musicology Journal",
        12976: "Journal of Insect Biodiversity",
        12977: "BraDypUS",
        12978: "Journal of Analytic Theology",
        12979: "Revista de Contabilidade do Mestrado em Ciencias"
               "Contabeis da UERJ",
        12980: "Asian Pacific Journal of Tropical Medicine Press",
        12981: "Motif Akademi",
        12982: "Chiang Mai University",
        12983: "IJSR Publications",
        12984: "Ege Egitim Dergisi",
        12985: "The Korean Society for Aviation and Aeronautics",
        12986: "Infomesr",
        12987: "Yale University Press",
        12988: "Hikari, Ltd.",
        12989: "Techno-Press",
        12990: "European Journal of Research in Dentistry",
        12991: "Marmara Pharmaceutical Journal",
        12992: "International Language, Literature and Folklore"
               "Researchers Journal",
        12994: "Mediros Clinical Solutions",
        12995: "Bilig, Journal of Social Sciences of Turkish World",
        12996: "Gazi University",
        12997: "Korean Society of Lipidology and Atherosclerosis",
        12998: "Baishideng Publishing Group Co (World Journal"
               "of Clinical Cases)",
        12999: "American Water Works Association - Standards",
        13000: "The Korean Society for Fisheries and Marine"
               "Sciences Education",
        13001: "University of Wyoming Libraries",
        13002: "Journal of Agricultural Faculty of Gaziosmanpasa University",
        13003: "Crossref Publications",
        13004: "Korean Neurotraumatology Society",
        13005: "Oriental Scientific Publishing Company",
        13006: "Polish Hyperbaric Medicine and Technology Society",
        13007: "IEDP Ideas for Leaders",
        13008: "University of Iowa Libraries",
        13014: "University of Nebraska Consortium of Libraries - UNCL",
        13021: "Mason Publishing, George Mason University Libraries",
        13029: "Korean Association of Pediatric Surgeons",
        13030: "CLOCKSS Archive",
        13031: "American Society of Agricultural and Biological Engineers",
        13032: "Optimal Clinical Ltd.",
        13033: "Creative Decisions Foundation",
        13034: "Journal of Student Science and Technology",
        13035: "Instituto de Estudios Auriseculares",
        13036: "Society of Glass Technology",
        13037: "USCS Universidade Municipal de Sao Caetano do Sul",
        13038: "Edinburgh University Library",
        13039: "Crossref (Funder Registry)",
        13040: "International Journal of Pharmaceutical Sciences"
               "and Research",
        13041: "Korean Society of Preventive Veterinary Medicine",
        13042: "Sociedad Espanola de Pedagogia",
        13043: "Revista Desarrollo y Sociedad",
        13044: "SDEWES Centre",
        13045: "Korean Acupuncture and Moxibustion Medicine Society",
        13046: "Northeast Brazilian Health Journal",
        13047: "Korea Society of Environment and Ecology",
        13048: "The Society of Korean Medicine",
        13049: "The Korean Family Welfare Association",
        13050: "Korean Society for Food Engineering",
        13051: "Electronic Enlightenment Project",
        13052: "River Publishers",
        13053: "Instituto Politecnico Nacional/Centro de Investigacion"
               "en Computacion",
        13054: "Mevlana Internationl Journal of Education",
        13055: "Ross Science Publishers",
        13056: "Association for Child and Adolescent Mental Health (ACAMH)",
        13057: "UNS Solo",
        13058: "ANGRAD",
        13059: "FUNDACE",
        13060: "Institute of Sociology of the Academy of Sciences"
               "of the Czech Republic",
        13061: "ANPUR",
        13062: "Ege Universitesi Turk Dunyasi Incelemeleri Dergisi",
        13063: "Academy Health",
        13064: "The Korean Society of Speech Sciences",
        13065: "Korean Society of Dental Hygiene",
        13066: "The Korean Society of Physical Medicine",
        13067: "The Korean Institute of Electronic Communication Sciences",
        13068: "Arkiv Forlag & Tidskrift",
        13069: "Journal of Algebra Combinatorics Discrete"
               "Structures and Applications",
        13070: "Synatom Research, LLC",
        13071: "Revista Gestao & Conexoes - Management and"
               "Connections Journal",
        13072: "Whitaker Institute",
        13073: "Forest Products Society",
        13074: "Journal of Environmental Nanotechnology",
        13075: "Nofer Institute of Occupational Medicine",
        13076: "Turkish Journal of Orthodontics",
        13077: "The Electronic Materials Research Association",
        13078: "Korean Sleep Society",
        13079: "Klaipeda University",
        13080: "Lithuanian Research Centre for Agriculture and Forestry",
        13081: "The Korean Society for the History of Medicine",
        13082: "Crop Breeding and Applied Biotechnology",
        13083: "Revista Engenharia na Agricultura",
        13084: "Iberoamerican Journal of Industrial Engineering",
        13085: "Forschungsinstitut Freie Berufe",
        13086: "Revista Ecologia e Nutricao Florestal",
        13087: "The Korea Society of Environmental Restoration Technology",
        13088: "Korea Intelligent Information System Society",
        13089: "Korea Institute of Information Security and Cryptology",
        13090: "IBICT Test account",
        13092: "University of Bern",
        13101: "Japan Society of Erosion Control Engineering",
        13102: "Universidade Estadual de Feira de Santana",
        13103: "The Korean Society of Food Hygiene an Safety",
        13104: "Korean Society of Magnetic Resonance in Medicine",
        13105: "World Journal of Meta-Analysis",
        13106: "Korea Distribution Science Association",
        13107: "Indian Orthopaedic Research Group",
        13108: "Institute of Mathematics USC RAS",
        13109: "Vandenhoeck & Ruprecht GmbH & Co, KG",
        13110: "Wayne State University Press",
        13111: "INCAS - National Institute for Aerospace Research"
               "Elie Carafoli",
        13112: "Croatian Medical Association",
        13113: "Cedrus",
        13114: "Mediterranean Journal of Humanities",
        13115: "Revista Dialogos",
        13141: "Sachsische Landesbibliothek, Staats-"
               " und Universitatsbibliothek Dresden",
        13152: "Uni Bremen Campus GmbH",
        13156: "BioOne - Arachnology",
        13157: "BioOne - Ardeola",
        13158: "BioOne - Herzogia",
        13159: "Siirt Universitesi Sosyal Bilimler Enstitusu Dergisi",
        13160: "The Research Institute of Chosun Natural Science",
        13161: "Korean Institute of Building Information Modeling",
        13162: "Health Reform Observer - Observatoire des Reformes de Sante",
        13163: "Proceedings of Science Open Reviewed",
        13164: "Brno University of Technology",
        13165: "Mykolas Romeris University",
        13166: "Wyzsza Szkola Gospodarki Euroregionalnej im."
               "Alcide de Gasperi w Jozefowie",
        13167: "Faculty of Civil Engineering and Architecture Osijek",
        13168: "Institute of Rock Structure and Mechanics, AS CR",
        13169: "Pluto Journals",
        13170: "Institute of Postgraduate Studies, Syiah Kuala University",
        13171: "Mediterranean Journal of Chemistry",
        13172: "Open Access Publishing London",
        13173: "Harrassowitz Publishing House",
        13174: "Military Institute of Aviation Medicine",
        13175: "Southwest Journal of Pulmonary and Critical Care",
        13176: "Journal of Pattern Recognition Research",
        13177: "Institute of Public Administration and Politics"
               "- Icelandic Review of Politics and Administration",
        13178: "Journal of Nursing Practice Applications and"
               "Reviews of Research",
        13179: "Borderless Scicence Publishing",
        13180: "CLAWAR Association Limited",
        13181: "Faculty of Medicine, Universitas Indonesia",
        13182: "American Nuclear Society",
        13183: "Swedish Science Pioneers",
        13184: "Fundacion de Ciencias de al Salud",
        13185: "Ateneo de Manila University",
        13186: "Eastern Group Psychotherapy Society",
        13187: "Academic Publishing House Researcher",
        13188: "Avens Publishing Group",
        13189: "Horizon Research Publishing Co., Ltd.",
        14190: "Korean Recycled Construction Resource Institute",
        14191: "Korean Meteorological Society",
        14192: "Korean Society for Healthcare-associated Infection"
               "Control and Prevention",
        14193: "Korean Foot and Ankle Society",
        14194: "Michael Joanna Publications",
        14195: "Coimbra University Press",
        14196: "Sjournals",
        14197: "Alliance Against Traffic in Women Foundation",
        14198: "Publicaciones Universidad de Alicante",
        14199: "Institute of Plant Protection - National Research Institute",
        14200: "Association for the Advancement of Restorative Medicine",
        14201: "Ediciones Universidad de Salamanca",
        14202: "Veterinary World",
        14203: "Indonesian Institute of Sciences",
        14204: "Education & Psychology I+D+i",
        14205: "Pharma Publisher",
        14206: "Global Science Heritage",
        14207: "European Center of Sustainable Development",
        14208: "EBES Aras. Dan Egt. Yayin ve Org (Eurasia"
               "Business and Economics Society)",
        14209: "Sociedad Brasileira de Telecomunicacaoes",
        14210: "Editora UNIVALI",
        14211: "Revista de Empreendedorismo e Gestao de Pequenas"
               "Empresas - REGEPE",
        14212: "ISE Vera Cruz",
        14213: "The International Centre for Trade Union Rights ICTUR",
        14214: "Finnish Society of Forest Science",
        14216: "Korean Society of Clinical Oncology",
        14217: "Commonwealth Seretariat",
        14218: "Xia & He Publishing",
        14219: "American Dental Association",
        14220: "Vandenhoeck & Ruprecht GmbH & Co, KG - V&R Unipress",
        14221: "Edith Cowan University",
        14222: "Journal of Turkish Research Institute",
        14223: "Foundation for Sickle Cell Disease Research",
        14224: "Cold Spring Harbor Laboratory - CSHL Library & Archives",
        14225: "Journal of History School",
        14226: "Korean Society for Imaging Science and Technology",
        14227: "Dissolution Technologies",
        14228: "Jurnal Plastik Rekonstruksi",
        14229: "InPress Media Group, LLC",
        14230: "Karabuk University Journal of Institute of Social Sciences",
        14231: "Foundation Compositio Mathematica",
        14232: "University of Szeged",
        14233: "Asian Journal of Chemistry",
        14234: "University of Northampton",
        14235: "Bezmialem Vakif University",
        14236: "BCS Learning and Development",
        14237: "Society of Ethnobiology",
        14238: "Paediatrica Indonesiana  - Indonesian Pediatric Society",
        14239: "Centre for Postgraduate Studies",
        14240: "SAGE",
        14241: "Polskie Towarzystwo Geologiczne - Polish Geological Society",
        14242: "Associacao Medica de Brasilia",
        14243: "Japan Society of Applied Electromagnetics and Mechanics",
        14244: "Universidade Federal de Sao Carlos Programa"
               "de Pos-Graduacao em Educacao",
        14245: "The Korean Spinal Neurosurgery Society (KAMJE)",
        14246: "International Community of Spatial Planning"
               "and Sustainable Development",
        14247: "International Association of Lowland Technology",
        14248: "The Korean Society of Systems Engineering",
        14249: "Korean Society of Environmental Impact Assessment",
        14250: "Japan Cement Association",
        14251: "Crisis and Emergency Management:Theory and Praxis",
        14252: "Food Safety Commission, Cabinet Office",
        14253: "The Korean Society of Clinical Neurophysiology (KAMJE)",
        14254: "Centre of Sociological Research, NGO",
        14255: "Kherson State University",
        14256: "Croatian Association of Civil Engineers",
        14257: "Science and Engineering Research Support Society",
        14258: "Altai State University",
        14259: "VRI Press",
        14260: "Akshantala Enterprises Private Limited",
        14261: "Saxion",
        14262: "MERLOT Multimedia Educational Resource for"
               "Learning and Online Teaching",
        14263: "Association of Jewish Libraries",
        14264: "University of Queensland Library",
        14265: "International Association of School Librarianship",
        14266: "Petrodex Publication Group",
        14267: "Corvinus University of Budapest",
        14268: "ACTA Editora",
        14269: "Journal of Animal Behaviour and Biometeorology",
        14270: "Diagnosis Engineering Technology, LLP, Japan",
        14271: "Gesellschaft Anthroposophischer Arzte in Deutschland GAAD",
        14283: "SERDI",
        14292: "TOTBID Turk Ortopedi ve Travmatoloji Birligi Dernegi",
        14293: "ScienceOpen",
        14294: "Water Journal",
        14295: "Lepidus Tecnologia",
        14296: "School of Advanced Study",
        14297: "Edinburgh Napier University",
        14298: "The Movement Disorder Society",
        14299: "IJSER Publishing",
        14300: "Stavropol State Medical University",
        14301: "Society for Longitudinal and Life Course Studies",
        14302: "Open Access Pub",
        14303: "International Research Journals (OMICS)",
        14304: "Postdoc Journal",
        14305: "Syracuse University Libraries",
        14306: "Fundacion Espanola de Dietistas-Nutricionistas (FEDN)",
        14307: "American Association of Family and Consumer Sciences",
        14308: "Information Technologies in Education",
        14309: "American College of Gastroenterology",
        14310: "Hellenic Endocrine Society",
        14311: "Czech Technical University in Prague - Central Library",
        14312: "Noble Research Publishers",
        14313: "PIAP - Industrial Research Institute for Automation"
               "and Measurements",
        14314: "LUKASIEWICZ Research Network, Industrial Chemistry"
               "Research Institute",
        14315: "Penguin Random House Verlagsgruppe GmbH",
        14316: "Korean Society of Medcial Physics",
        14317: "Korean Society of Computational and Applied Mathematics",
        14318: "HAU, Journal of Ethnographic Theory",
        14319: "International Journal of Cancer Therapy and Oncology",
        14320: "Instytut Dermatologii Radoslaw Spiewak",
        14321: "Michigan State University Press",
        14322: "Publons",
        14323: "Marques Aviation Ltd.",
        14324: "UCL Press",
        14325: "University Press of Mississippi",
        14326: "Japanes Society for Medical and Biological Engineering",
        14327: "Japan Management Training Center",
        14328: "Korean Association for Multicultural Education",
        14329: "Korea Society of Management Information Systems",
        14330: "Yijun Institute of International Law",
        14331: "Fundamental Journals",
        14332: "Society of Vacuum Coaters",
        14333: "Center for Education Research",
        14334: "Indonesian Center for Animal Research and Devleopment, ICARD",
        14335: "Solutionwork",
        14336: "Aging and Disease",
        14337: "XML London",
        14338: "International Journal of Particle Therapy",
        14339: "NATO Science and Technology Organization",
        14340: "GH, Inc.",
        14341: "Endocrinology Research Centre",
        14342: "Society of Modern Grammar",
        14343: "Trends Ahead Scientific Journals",
        14344: "World Bird Names International Ornithologists Union",
        14345: "The Korean Society on Thrombosis and Hemostasis",
        14346: "The Korean Society of Safety",
        14347: "Korean Academy of Dental Technology",
        14348: "Korean Society for Molecular and Cellular Biology",
        14349: "Fundacion Universitaria Konrad Lorenz",
        14350: "Instituto de Geografia, UNAM",
        14351: "Society for the Preservation of Natural History"
               "Collections SPNHC",
        14352: "Korean Association of Information Education",
        14353: "The Sociolinguistic Society of Korea",
        14354: "The Yeats Society of Korea",
        14355: "Science and Engineering Publishing, Co.",
        14356: "Hosokawa Powder Technology Foundation",
        14357: "Federal Investigation Centre for Information and Control",
        14358: "American Society for Photogrammetry and Remote Sensing",
        14359: "American Concrete Institute",
        14360: "Medicine and Health Promotion Institute",
        14361: "Transcript Verlag",
        14362: "Korean Language Education Research Institute,"
               "Seoul National University",
        14363: "The Korean Association of Peace Studies",
        14364: "The T.S. Eliot Society of Korea",
        14365: "The Korean Academy of International Business",
        14366: "Korean Society of Ultrasound in Medicine",
        14367: "Korean Society for Health Education and Promotion",
        14368: "Korean Academy of Stomatognathic Function and Occlusion",
        14369: "The Society of Korean Medical Classics",
        14370: "East-West Nursing Research Institute",
        14371: "Korean Society for Quality in Health Care",
        14372: "Institute of Embedded Engineering of Korea",
        14373: "The Korea Dietetic Association",
        14374: "The Korean Academy of Oriental Medicne Prescription",
        14375: "Numerique Premium",
        14376: "The Korean Society for Literature and Religion",
        14377: "Korea Advertising Society",
        14378: "Korean Association for Chinese Studies",
        14379: "International Ocean Discovery Program - TAMU",
        14380: "Center for Art Studies",
        14381: "The Nammyonghak Studies Institute",
        14382: "SZTE Hungarian Scientific Society of the Silicate Industry",
        14383: "Konkuk University Climate Research Institute",
        14384: "The Korean Association of Language Sciences",
        14385: "Korean Society of Sport Psychology",
        14386: "The Korea Society for Innovation Management and Economics",
        14387: "The Korean Society for Practical Theology",
        14388: "The Korea Society for Dance Documentation",
        14389: "Japanese Society of Anxiety Disorders",
        14390: "Japan Society for the Study of Hypertension in Pregnancy",
        14391: "Asian Society of Human Services",
        14392: "Associacao Nacional de Programas de Pos-Graduacao"
               "em Ciencias Contabeis (ANPCONT)",
        14393: "EDUFU - Editora da Universidade Federal de Uberlandia",
        14394: "University of Warsaw - Faculty of Philosophy and Sociology",
        14395: "Journal of Divinity Faculty, Hitit University",
        14396: "Korean Academy of Human Resource Management",
        14397: "Institute of Agriculture and Life Science,"
               "Gyeongsang National University",
        14398: "The City Planning Institute of Japan",
        14399: "The Academy of Clinical Dentistry",
        14400: "The Korea Society of Digital Policy and Management",
        14401: "Korean Academy of Sleep Medicine",
        14402: "Korean Association of Maxillofacial Plastic"
               "and Reconstructive Surgeons",
        14403: "Chungcheong Mathematical Society",
        14404: "Korean Society of Archives and Records Managmenet",
        14405: "The Korean Society of Veterinary Science (KJVR)",
        14406: "Society for Meridian and Acupoint",
        14407: "Korean Association for Radiation Protection",
        14408: "The Korean Society of Emergency Medical Services",
        14409: "Universidad Nacional del Litoral",
        14410: "Revista Medica del Hospital Jose Carrasco Arteaga",
        14411: "Biology Centre AS CR",
        14412: "IMA Press, LLC",
        14413: "University of Debrecen/ Debreceni Egyetem - HERJ",
        14414: "STIE Perbanas Surabaya",
        14415: "Gradevinski Fakultet Subotica/Faculty of Civil"
               "Engineering Subotica",
        14416: "King Mongkut's University of Technology North Bangkok",
        14417: "ISPA - Instituto Universitario",
        14418: "Wesleyan University",
        14419: "Science Publishing Corporation",
        14420: "Foundation Consulting and Legal Protection of Citizens",
        14421: "Al-Jami'ah Research Centre",
        14422: "Universidad Pontificia Comillas",
        14423: "Southern Medical Association",
        14424: "International Journal of Construction Supply"
               "Chain Management",
        14425: "University of Malaya Press",
        14426: "University of the Western Cape Library Service",
        14427: "All-Russian National Academy of Mycology",
        14428: "Universite Catholique de Louvain",
        14429: "Defence Scientific Information and Documentation Centre",
        14430: "The Arctic Institute of North America",
        14431: "Research Institute of Asian Women",
        14432: "Faculty of Veterinary Medicine - Skopje",
        14433: "Association of American Geographers",
        14434: "IUScholarWorks",
        14435: "Science Practice",
        14436: "Dental Press International",
        14437: "Aperito Online Publishing Private Limited",
        14438: "Serbian Union of Surveyors",
        14439: "Royal Conservatoire of Scotland",
        14440: "Journal of Biological Methods",
        14441: "Japan Association for Evolutionary Economics",
        14442: "The Japan Primary Care Association",
        14443: "Korea Institute of Maritime Law",
        14444: "International Journal of Spine Surgery",
        14445: "Seventh Sense Research Group Journals",
        14446: "National Museum",
        14447: "Journal of New Materials for Electrochemical Systems",
        14448: "Wallace Center",
        14449: "Korean Breast Cancer Society",
        14450: "Conselho Federal de Farmacia",
        14451: "Economic Sciences",
        14452: "Monthly Review Foundation",
        14453: "University of Wollongog Library",
        14455: "ISEC Press",
        14464: "Technische Universitat Chemnitz",
        14474: "Physical Therapy Rehabilitation Science",
        14475: "Korean Society for Hospice and Palliative Care",
        14476: "Journal of Oral Medicine and Pain",
        14477: "The Korean Society for History of Mathematics",
        14478: "The Korean Society of Industrial and Engineering Chemistry",
        14479: "The Korean Opthalmic Optics Society",
        14480: "The Korean Society of Mushroom Science",
        14481: "Korean Geo-Environment Society",
        14482: "Universidad del Norte",
        14483: "Universidad Distrital Francisco Jose de Caldas",
        14484: "Journal of the Brazilian Air Transportation"
               "Research Society - JBATS",
        14485: "Paris Scholar Publishing",
        14486: "International Journal of Science Culture and Sport",
        14487: "Surname DNA Journal",
        14488: "Associacao Brasileira de Engenharia de Producao - ABEPRO",
        14489: "Izdatel'skii dom Spektr, LLC",
        14490: "The Japan Statistical Society",
        14491: "The Japanese Bird Banding Association",
        14492: "Department of Mathematics, Hokkaido University",
        14493: "The Korea Society of Mission Studies",
        14494: "The Japan Society of Nuclear and Radiochemical Sciences",
        14495: "The Japan Society for Industrial and Applied Mathematics",
        14496: "Diapedia",
        14497: "Rut Printer and Publisher",
        14498: "Samara State Technical University",
        14499: "Indonesian Journal of Pharmacy",
        14500: "Koya University",
        14501: "Ural Regional Nutrition Centre",
        14502: "Department of Textiles, Faculty of Natural"
               "Sciences and Engineering, University of Ljubljana",
        14503: "Texas Heart Institute Journal",
        14504: "American Association of Textile Chemists and"
               "Colorists - AATCC",
        14505: "SC ASERS SRL",
        14506: "American Anthropological Association",
        14507: "Education Policy Analysis Archives",
        14508: "Escuela de Ingenieria de Antioquia  - EIA",
        14509: "Alaska Division of Geological and Geophysical Surveys",
        14510: "American Romanian Academy of Arts and Sciences",
        14511: "American V-King Scientific Publishers",
        14512: "Oekom Publishers",
        14513: "Szechenyi Istvan University",
        14514: "Beykoz Akademi Dergisi",
        14515: "Russian Public Opinion Research Center",
        14516: "FahrenHouse",
        14517: "Korean Arthroscopy Society and Korean Orthopaedic"
               "Society for Sports Medicine",
        14518: "Current Research on Agriculture and Life Science",
        14519: "Korean Society of Occupational Therapy",
        14520: "Adiyaman University Journal of Social Sciences",
        14521: "Centro Universitario UNISOCIESC",
        14522: "Instituto de Botanica Darwinion (IBODA)",
        14523: "Ambra-Verlag",
        14524: "Association of Clinical Research Professionals",
        14525: "Jordan Journal of Civil Engineering",
        14526: "Kama Research Centre",
        14527: "Pegem Akademi Yayincilik Egitim Danismanlik Hizmetleri",
        14528: "Nurses and Midwives Association of Slovenia",
        14529: "FSAEIHE South Ural State University (National"
               "Research University)",
        14530: "Economic Research Institute, FEB RAS",
        14531: "ANO Clinic of NNIITO",
        14532: "Japanese Society for  Microcirculation",
        14533: "Japan Science Society of Biological Macromolecules",
        14534: "Pediatric Urology Case Reports",
        14568: "Associacao Profissional de Conservadores-Restauradores"
               "de Portugal",
        14569: "The Science and Information Organization",
        14570: "Content Press and Scientific Publishing",
        14571: "Brazilian Journal of Education, Technology"
               "and Society (BRAJETS)",
        14572: "Departamento de Educacao FCT/Unesp",
        14573: "ALTEX Edition",
        14574: "Online Journal of Rural Nursing and Health Care",
        14575: "Green Lines Institute of Sustainable Development",
        14576: "Proteopedia",
        14577: "Korean Institute of Rural Architecture",
        14578: "Korean Forest Society",
        14579: "The Membrane Society of Korea",
        14580: "Institute of APEC Collaborative Education",
        14581: "Korean Epilepsy Society",
        14582: "Dicle Universitesi Ziya Gokalp Egitim Fakultesi Dergisi",
        14583: "Revista Nativa",
        14591: "Associacao de Investigadores da Imagem em Movimento",
        14602: "ERFARET, Ldt.",
        14612: "Assonebb",
        14624: "Neurologie & Rehabilitation, Hippocampus Verlag",
        14635: "Col.legi Oficial de Psicologia de la Comunitat Valenciana",
        14639: "Pacini Editore",
        14657: "Scuola Normale Superiore - Edizioni della Normale",
        14664: "Societa Italiana di Vittimologia",
        14684: "Science and Education Research Council - COPEC",
        14685: "Federal University of Technology - Parana",
        14686: "Bartin Universitesi Egitim Fakultesi Dergisi",
        14687: "Uluslararasi Insan Bilimleri Dergisi",
        14688: "Global Science and Technology",
        14689: "Ani Publishing and Consulting Company",
        14690: "IESUR/FAAr",
        14691: "Stowarzyszenie Psychologia i Architektura"
               "(Society for Psychology and Architecture)",
        14692: "The Indonesian Phytopathological Society",
        14693: "Journal of Dentistry Indonesia",
        14694: "American Society of Clinical Oncology (Educational Books)",
        14695: "Korean Society for Emotion and Sensibility",
        14696: "Korea Regional Communication Research Association",
        14697: "The Korean Association for Science Education",
        14698: "The Korean Association of Child Care & Education",
        14699: "Korea Biblia Society for Library and Information Science",
        14700: "Korean Institute of Traditional Landscape Architecture",
        14701: "The Korean Association of Hepato-Biliary-Pancreatic Surgery",
        14702: "Korean Institute for Practical Engineering Education",
        14703: "Institute of International Cooperation on Agriculture",
        14704: "NeuroQuantology Journal",
        14705: "Research-Publishing.net",
        14706: "International Burch University",
        14707: "Asia Business Research Corporation",
        14708: "Polish Mathematical Society",
        14709: "SECEMU",
        14710: "Institute of Research and Community Services"
               "Diponegoro University (LPPM UNDIP)",
        14711: "The Hong Kong University of Science and Technology Library",
        14712: "Charles University in Prague, Karolinum Press",
        14713: "Rutgers University Libraries",
        14714: "North American Cartographic Information Society",
        14715: "CMB Association",
        14716: "International Journal of Technology",
        14717: "Journal of Surgical Arts",
        14718: "Editorial Universidad Catolica e Colombia",
        14719: "Horizon E-Publishing Group",
        14720: "Biotechnical Faculty, University of Ljubljana",
        14721: "TranScience Publishing",
        14722: "Internet Society",
        14723: "The Materials Research Society of Japan",
        14724: "Journal of Humanity",
        14725: "New Century Science Press. LLC",
        14726: "Peerage of Science",
        14727: "The Society of Korean Historical Linguistics",
        14728: "The Korean Society of Media and Arts",
        14729: "Korean Association of Converging Humanities",
        14730: "Korean Society for Aesthetic Plastic Surgery",
        14731: "Korean Association of International Studies",
        14732: "International Journal of Bioactive Products",
        14733: "CAD Solutions, LLC",
        14734: "Korean Society of Perinatology",
        14735: "Ambit Media, a.s.",
        14736: "Institute of Information Theory and Automation",
        14737: "ResearchersLinks Ltd",
        14738: "Scholar Publishing",
        14739: "Zaporozhye State Medical University",
        14740: "Elmer Press, Inc.",
        14741: "Inpressco",
        14742: "Australasian Society for Computers in Learning"
               "in Tertiary Education",
        14743: "Production Engineering Institute (PEI), Faculty"
               "of Mechanical Engineering",
        14744: "Kare Publishing",
        14745: "Infectious Disease and Control Branc (IDPCB)"
               "- Public Health Agency of Canada",
        14746: "Adam Mickiewicz University Poznan",
        14747: "McMaster Undergraduate Research Journal",
        14748: "Medical University Prof. Dr. Praskev Stoyanov - Varna",
        14763: "Internet Policy Review, Alexander von Humboldt"
               "Institute for Internet and Society",
        14769: "Korean Association of Art History Education",
        14770: "The Geological Society of Korea",
        14771: "Korean Society of Food and Agicultural Information Science",
        14772: "Center for the Study of Chinese Characters in Korea",
        14773: "The Corrosion Science Society of Korea",
        14774: "Korean Institute of Interior Design",
        14775: "The Korean Society of Manufacturing Process Engineers",
        14776: "The Korean Society of Pediatric Infectious Diseases",
        14777: "Korean Association of Urogentital Tract Infection"
               "and Imflammation",
        14778: "VLDB Endowment",
        14779: "E-Journal of Banking and Insurance Institute",
        14780: "M.U. Iktisadi ve Idari Bilimler Dergisi",
        14781: "Mamara Cografya Dergisi",
        14782: "Siyasal Bilimler Dergisi",
        14783: "Sosyal Bilimler Ensititusu Hakemli Dergisi Oneri",
        14784: "The Journal of Financial Researches and Studies",
        14785: "LymphoSign Journal Limited Partnership",
        14786: "EARLI",
        14787: "Society of Plant Reproductive Biologists",
        14788: "Editorial Board of International Journal of"
               "Fluid Machinery and Systems",
        14789: "The Jutendo Medical Journal",
        14790: "Japanese Association for Management Systems",
        14791: "The Korean Brain Tumor Society; The Korean"
               "Society for Neuro-Oncology (KAMJE)",
        14792: "Global e-Governance Research Institute",
        14793: "Fong Sum Wood Library",
        14794: "University of Debrecen/ Debreceni Egyetem- ICAI",
        14795: "Institute of Archaeology and Art History",
        14796: "Computational Hydraulics International",
        14797: "Methodist DeBakey Cardiovascular Journal",
        14798: "Croatian Journal of Fisheries",
        14799: "Environmental Biotechnology",
        14800: "Smart Science and Technology, LLC",
        14801: "Korean Institute of Information Technology",
        14802: "The Korean Movement Disorder Society",
        14803: "Stellenbosch University - African Public Procurement"
               "Law Journal",
        14804: "Stellenbosch University - Trends in Nursing",
        14805: "Vensel Publications",
        14806: "EMBnet Stichting",
        14807: "Independent Journal of Management and Production",
        14808: "Associacao Sergipana de Ciencia",
        14809: "University of Texas at San Antonio",
        14810: "Wireilla Scientific Publications",
        14811: "The Swedish Institute fo Children's Books",
        14812: "Cukurova University Faculty of Education",
        14813: "International Biophilia Rehabilitation Academy",
        14814: "Wiley (Physiological Reports)",
        14815: "Korean Society for Dental Materials",
        14816: "National Youth Policy Institute",
        14817: "The Japanese Language Association of Korea",
        14818: "the Korean Society for Augmentative and Alternative"
               "Communication",
        14819: "Korean Society for Correction Service",
        15170: "University Library of Pecs",
        15171: "International Society for Phytocosmetic Sciences",
        15172: "Griffith University",
        15173: "McMaster University Library Press",
        15174: "University of Guanajuato",
        15175: "Passagens",
        15176: "Institute of Ethnology and Folkkore Research",
        15177: "Hrvatski Sumarski Institut (Croatian Forest Research)",
        15178: "Forum Internacional de Comunicacion y Relaciones"
               "Publicas - FORUM XXI",
        15179: "The Institute of Economics, Zagreb",
        15180: "The Science Museum",
        15181: "Klaipeda University Press",
        15182: "Dicle Universitesi Sosyal Bilimler Enstitusu Dergisi",
        15183: "Spirituality, Leadership and Management, Inc.",
        15184: "Antiquity Publications",
        15185: "Forschungsinstitut zur Zukunft der Arbeit GmbH",
        15186: "Institute of Korean Cultural Studies Yeungnam University",
        15187: "Korea Society of Design Science",
        15188: "Korean Society of Oriental Pathology",
        15189: "Turan-Sam",
        15190: "Applied Systems, srl",
        15191: "National Weather Association",
        15192: "Progressive Science Publications",
        15193: "Polskie Towarzystwo Technologow Zywnosci Wydawnictwo"
               "Naukowe PTTZ",
        15194: "Instituto Brasileiro de Inovacao Financeira - IBRIF",
        15195: "Society for Sociological Science",
        15196: "Hungarian Central Statistical Office",
        15197: "Health Sciences Researches Society",
        15198: "Sociedad Espanola de Estudios de la Comunicacion"
               "Iberoamericana (SEECI)",
        15199: "Wydawnictwo SIGMA-NOT, sp. z.o.o.",
        15200: "The Winnower, LLC",
        15201: "Research Centre for Astronomy and Earth Sciences",
        15202: "Sociedade Unificada de Ensino Augusto Motta -UNISUAM",
        15203: "Universitaet Innsbruck - Innsbruck University Press",
        15204: "The Society of Korean Medicine Obstetrics and Gynecology",
        15205: "The Korean Society of Clinical Health Science",
        15206: "Asian Network for Public Opinion Research",
        15207: "Korea Convergence Society",
        15208: "Prague Development Center",
        15209: "Victoria University",
        15210: "Universidade Federal de Pelotas",
        15211: "Institute of Europe, Russian Academy of Sciences IERAS",
        15212: "Compuscript, Ltd.",
        15213: "University of Seville",
        15214: "The Japanese Society of Oral Diagnosis/Oral Medicine",
        15215: "Athabasca University Press",
        15216: "Knorus",
        15217: "Central Collector of Libraries BIBCOM",
        15218: "Hawler Medical University",
        15219: "Foundation for the Promotion and Accreditation"
               "of Economic Education",
        15220: "UNESCO Institute for Statistics",
        15221: "Hometrica Consulting",
        15222: "Private Enterprise, Politehperiodika",
        15223: "International STM Association",
        15224: "IRED PTE LTD.",
        15225: "Polskie Towarzystwo Pielegniarek Neurologicznych"
               "(Polish Assn. of Neurological Nurses)",
        15226: "Symbiosis Group",
        15227: "Organic Syntheses",
        15228: "Chempublishers",
        15229: "TAPAC: Theatre and Performance Across Cultures",
        15230: "Society of Cosmetic Scientists of Korea",
        15231: "The Korean Society of Combustion",
        15232: "American Registry of Professional Animal Scientists",
        15233: "University of Zagreb, Faculty of Science,"
               "Department of Geophysics",
        15234: "Pan-American Association of Ophthalmology",
        15235: "The East Asian Association of International Studies",
        15236: "Korean Academy of Preventive Dentistry",
        15237: "Association of Food Technology, Turkey",
        15238: "Turkish Journal of Islamic Economists",
        15239: "Cambria Press",
        15240: "Technical University of Liberec",
        15241: "The Professional Counselor at the National"
               "Board for Certified Counselors",
        15242: "International Institute of Engineers",
        15243: "International Research Centre for the Management"
               "of Degraded and Mining Lands",
        15244: "HARD Publishing Company",
        15245: "Japanese Association of Cardiovascular Intervention"
               "and Therapeutics",
        15246: "Hunkar Alevilik Bektasilik Akademik Arastimalar Dergisi",
        15247: "Divan Edebiyati Arastirmalari Dergisi",
        15248: "The Institute of Electronics, Information"
               "and Communication Engineers",
        15249: "Stellenbosch University - African Journal of Business Ethics",
        15250: "Journal of Odor and Indoor Environment",
        15251: "Virtual Company of Physics",
        15252: "EMBO",
        15253: "Rev Rene - Revista da Rede de Enfermagem do Nordeste",
        15254: "CD Remedies",
        15255: "Croatian Society of Chemical Engineers/HDKI",
        15256: "Swiss Medical Press, GmbH",
        15257: "International Social Sciences and Social Work"
               "Association AICTS",
        15258: "International Seed Testing Association",
        15259: "Polish Chitin Society",
        15260: "Associacao Brasileira de Criminalistica - ABC",
        15261: "Japan Association of Solvent Extraction",
        15262: "International Research and Cooperation Association"
               "for Bio & Socio-Sciences Advancment",
        15263: "Korean Association of Quality Assurance for"
               "Clinical Laboratory",
        15264: "The Korean Society of Pediatric Hematology-Oncology",
        15265: "Thieme GmbH (IMIA Yearbook of Medical Informatics)",
        15266: "Korean Resource Economics Association",
        15267: "The Korean Society of Elementary Science Education",
        15268: "Korean Society of Integrative Medicine",
        15269: "Korean Industrial Hygiene Association",
        15270: "Stellenbosch University - Social Work/Maatskaplike Werk",
        15271: "MAGIC- Making GRADE the Irresistible Choice",
        15272: "Literati Scientific and Publishers",
        15273: "Dalhousie Libraries",
        15274: "MAGGIOLI S.p.A.",
        15275: "LLC Science and Innovations",
        15276: "Odessa National Polytechnic University",
        15277: "ABCD Diabetes Care, LTD.",
        15278: "International Symposium on Molecular Spectroscopy",
        15279: "Korean Pancreatobiliary Association",
        15280: "Institute of Lifestyle Medicine",
        15281: "The Botanical Society of Japan",
        15282: "Universiti Malaysia Pahang Publishing",
        15283: "Korean Society for Stem Cell Research",
        15284: "The Korean Association of Human Development",
        15285: "Egitim Bilimleri Dergisi",
        15286: "The Polynesian Society",
        15287: "Marin Dracea National Research-Development"
               "Institute in Forestry",
        15288: "Alcohol Research Documentation, Inc.",
        15289: "Magyar Biomechanikai Tarsasag",
        15290: "University of Bialystok",
        15291: "University of Zadar",
        15292: "Faculty of Civil and Geodetic Engineering,"
               "University of Ljubljana",
        15293: "Novosibirsk State Pedagogical University",
        15294: "Universitas Negeri Semarang",
        15295: "Business and Management Studies: An International Journal",
        15296: "International Journal of Women's Health and"
               "Reproduction Studies",
        15297: "ScienceQ Publishing Group",
        15298: "KMK Scientific Press",
        15299: "Korean Studies Institute",
        15300: "Korea Association of Community Welfare Studies",
        15301: "Korea Environmental Policy and Administration Society",
        15302: "Engineering Sciences Press",
        15303: "Romanian Society of Experimental Applied Psychology (RSEAP)",
        15304: "University of Santiago de Compostela",
        15305: "ChanRe Healthcare and Research",
        15306: "Dr. Vasantrao Pawar Medical College",
        15307: "The Fibreculture Journal",
        15308: "Singidunum University",
        15309: "Sociedad Portuguesa de Psicologia da Saude",
        15310: "Journal of Evolution and Health",
        15311: "Selcuk Dental Journal",
        15312: "Eurasian Journal of Veterinary Sciences",
        15313: "Journal of Selcuk University Natural and Applied Science",
        15314: "Turkish Journal of Sport and Exercise",
        15315: "Turkiyat Arastirmalari Dergisi - Journal of"
               "Studies in Turkology",
        15316: "Selcuk Journal of Agriculture and Food Sciences",
        15317: "Selcuk University Journal of Engineering,"
               "Science, and Technology",
        15318: "The Journal of Fungus",
        15319: "International Journal of Mevlana Medical Sciences",
        15320: "Iconarp International Journal of Architecture and Planning",
        15321: "Genel Tip Dergisi",
        15322: "Selcuk Pediatri",
        15323: "The Graduate School of Advanced Imaging Science,"
               "Multimedia and Film at Chung-Ang University",
        15324: "The Korean Society for Clinical Laboratory Science",
        15325: "Institute of Electrical and Electronics Engineers"
               " - Bell Labs",
        15326: "COPD Foundation",
        15327: "IJOART Publication",
        15328: "Center of Physical Chemical Methods of Research and Analysis",
        15329: "Federacao Brasileira de Psicodrama",
        15330: "Vasyl Stefanyk Precarpathian National University",
        15331: "American Academy of Sleep Medicine",
        15332: "Universidad Santo Tomas",
        15333: "Korean Society of Arts and Cultural Management",
        15334: "The Korea Association of Foreign Languages Education",
        15335: "The Law Research Institute of Gachon University",
        15336: "The Institute of Environment Studies Daejeon University",
        15337: "Selcuk Universitesi Hukuk Fakultesi Dergisi",
        15338: "Marmara Sosyal Arastirmalar Dergisi",
        15339: "Marmara Universitesi Saglik Bilimleri Enstitusu Dergisi",
        15340: "Macroworld Egitim Hiz. Med. San. Tic. Ltd.",
        15341: "Academic Star Publishing Company",
        15342: "Mediterranean BioMedical Journals",
        15343: "Centro Universitario Sao Camilo - Sao Paulo",
        15344: "Graphy Publications",
        15345: "International Online Journal of Educational Sciences",
        15346: "Human Computation Institute",
        15347: "Wikiversity Journal of Medicine",
        15348: "Revista Psicologia: Teoria e Pratica",
        15350: "Colloquim",
        15351: "Center for the Blue Economy",
        15352: "Tusi Mathematical Research Group",
        15353: "University of Waterloo",
        15354: "Bonoi Science Advancement and Education, LLC",
        15355: "Economists for Peace and Security",
        15356: "Institute of Geography, Russian Academy of Sciences",
        15357: "Community of Azerbaijanis Living in Georgia",
        15358: "Verlag C.H.Beck oHG",
        15359: "Universidad Nacional de Costa Rica",
        15360: "FSBI SRIGR RAMS",
        15361: "FUNEP",
        15362: "The College of the Bahamas",
        15363: "Thinklab",
        15364: "University of Malaya, Academy of Islamic Studies",
        15365: "Loyola Marymount University",
        15366: "Servicio de Publicaciones de la Universidad"
               "Autonoma de Madrid",
        15367: "Temple University Press",
        15368: "Robert E. Kennedy Library, Cal Poly",
        15369: "The Showa University Society",
        15370: "Marmara Universitesi Ilahiyat Fakultesi Dergisi",
        15371: "Marmara Journal of European Studies",
        15372: "Publishing House SB RAS",
        15373: "The Global Journals",
        15374: "Evereth Publishing",
        15375: "RWS Verlag GmbH",
        15376: "BioResources",
        15377: "Avanti Publishers",
        15378: "Croatian Ethnological Society",
        15379: "Cosmos Scholars Publishing House",
        15380: "Sri Venkateswara Institute of Medical Sciences",
        15381: "Universidad Nacional Mayor de San Marcos,"
               "Vicerectorado de Investigacion",
        15382: "St. Tikhon's Orthodox University",
        15383: "Asia Press",
        15384: "Korean Society of Health Promotion and Disease Prevention",
        15385: "Centennial Library",
        15386: "Clujul Medical",
        15387: "Child Mind Institute",
        15388: "Vilnius University",
        15389: "Sel'skokhozyaistvennaya Biologiya Editorial Office (SBEO)",
        15390: "Turkish Education Association",
        15391: "Kharkiv State Academy of Physical Culture",
        15392: "Sociedade Brasileira de Protecao Radiologica - SBPR",
        15393: "Petrozavodsk State University",
        15394: "Embry-Riddle Aeronautical University/Hunt Library",
        15395: "Majalah Kedokteran Bandung (MKB)",
        15396: "Stichting European Real Estate Society",
        15397: "Society for Korean Bronze Culture",
        15398: "Institute of Computer Science, Polish Academy of Sciences",
        15399: "Chonnam National University of Fishing Technology Institute",
        15400: "Cultural Contents and Technology Research Institute",
        15401: "Indian Medico Journal",
        15402: "Engaged Scholar Journal",
        15403: "Romanian Society of Gastroenterology and Hepatology",
        15404: "Medical Student Research Journal",
        15405: "Cognitive-crcs",
        15406: "MedCrave Group",
        15407: "National Academy of Sciences of Ukraine (Co."
               "LTD Ukrinformnauka)",
        15408: "LP2M Universitas Islam Negeri (UIN) Syarif"
               "Hidayatullah Jakarta",
        15409: "Research Institute for Early Childhood Education",
        15410: "Indira Institute of Managment, Pune",
        15411: "FUTURE - Instituto Superior de Educacao e Saude Sinop EIRELI",
        15412: "Lexis Publisher (OMICS)",
        15413: "Academia Publishing",
        15414: "Slovak University of Agriculture in Nitra",
        15415: "Chitkara University Publications",
        15416: "Indonesian Journal of Clinical Pharmacy",
        15417: "Asociacion Argentina de Ortopedia y Traumatologia",
        15418: "Centro de Estudios Superiores de Baja California S.C.",
        15419: "Biomedical Research and Therapy",
        15420: "Radcliffe Group Ltd",
        15421: "Oles Honchar Dnipropetrovsk National University",
        15422: "American Medical Student Research Journal",
        15423: "Academy of Dental Education",
        15424: "The Feinstein Institute for Medical Research"
               "(North Shore LIJ Research Institute)",
        15425: "Facultad de Derecho",
        15426: "Russian Society of Regenerative Medicine and"
               "Medical Rehabilitation",
        15427: "Instituto Cervantes at Harvard University",
        15428: "American Association for Clinical Chemistry - CCTC",
        15429: "Society of Korean Medicine for Obesity Research",
        15430: "Korean Society of Cancer Prevention",
        15431: "International Journal of Tea Science",
        15432: "Korean Academy of Traditional Oncology",
        15433: "The Korean Society for Marine Biotechnology",
        15434: "The Korean Society of School Health",
        15435: "Institute for Liquid Atomization and Spray Systems - Korea",
        15436: "Ommega Online Publishers",
        15437: "Journal of Exercise Therapy and Rehabilitation",
        15438: "Joint Implant Surgery and Research Foundation",
        15439: "Polish Information Processing Society",
        15440: "Problemata: International Journal of Philosophy",
        15441: "Korean Society of Emergency Medicine",
        15442: "Wyzsza Szkola Spoleczno-Przyrodnicza im Wincentego"
               "Pola W Lublinie (Vincent Pol University)",
        15443: "Universidad de la Serena",
        15444: "Global Alliance of Marketing and Management Associations",
        15445: "University of Sheffield Conference Proceedings",
        15446: "Universidad Nacional de Colombia",
        15447: "San Francisco Estuary and Watershed Science",
        15448: "EDIPUCRS",
        15449: "WorldAcademics, BOSON Academic Publications",
        15450: "Servicios Ecologicos y Cientificos SA de CV",
        15451: "Universidade Estadual da Paraiba/Universidade"
               "Federal Rural de Pernambuco",
        15452: "University of Ostrava",
        15453: "WMU Department of Occupational Therapy",
        15460: "Staats-"
               " und Universitatsbibliothek Hamburg"
               "Carl von Ossietzky",
        15462: "University Library J. C. Senckenberg",
        15503: "Foundation Pro Scientia Publica",
        15504: "University of Natural Resources and Life Sciences"
               "BOKU Vienna",
        15505: "Integrated Interactive Information International, LLC",
        15506: "GEM-A",
        15507: "National Research Mordovia State University MRSU",
        15508: "Sphinx Knowledge House",
        15509: "United Publication for Health and Tech Research",
        15510: "Journals of Biomedical Presentations Smart"
               "Sci Solutions, Inc.",
        15511: "Turkish Family Physicians Association",
        15512: "Journal of Ecophysiology and Occupational Health",
        15513: "Association of Asian Art History",
        15514: "Institute for System Programming of the Russian"
               "Academy of Sciences",
        15515: "Society of Education",
        15516: "Faculty of Teacher Education, University of Zagreb",
        15517: "Universidad de Costa Rica",
        15518: "Scientfic Technical Centre TATA",
        15519: "The Center for Asia and Diaspora",
        15520: "Innovative Journal",
        15521: "Korean Society of Medical History",
        15522: "Korean Academy of Esthetic Dentistry",
        15523: "The Korean Society of Earth Science Education",
        15524: "Korean Society of Alcohol Science",
        15525: "Journal of Mediterranean Marine Sciences and Technology",
        15526: "European Journal of Research on Social Studies",
        15527: "European Journal of Research on Education",
        15528: "Revista Cientifica de Producao Animal",
        15529: "Todas as Letras: Revista de Lingua e Literatura",
        15530: "AAPG/Datapages",
        15531: "The Korean Society of Climate Change Research",
        15532: "Korea Association of Japanology",
        15533: "SDMIMD Journal of Management",
        15534: "Quotus Publishing",
        15535: "SRI Institute of Political and Mediametric Studies IPMS",
        15536: "Instituto Federal de Educacao Ciencia e Tecnologia"
               "Sul-Rio-Grandense",
        15537: "Saudi Medical Journal",
        15538: "South European Journal of Othodontics and"
               "Dentofacial Research",
        15539: "Institute of Legal Studies, Kyung Hee University",
        15540: "International Society for Neurofeedback Research (ISNR)",
        15541: "Shanghai Institute of Ceramics",
        15542: "Schleuen Verlag",
        15543: "Max Weber Studies",
        15544: "Vytautas Magnus University Agriculture Academy",
        15545: "Japanese Political Science Review",
        15546: "Technical University of Kosice, Faculty of"
               "Electrical Engineering and Informatics",
        15547: "Trakia University",
        15548: "Universitas Islam Negeri Imam Bonjol Padang",
        15549: "Journal of Eastern European and Central Asian Research",
        15550: "Scireps Corporation",
        15551: "Alexandru Ioan Cuza University",
        15552: "Revista de Ensino de Engenharia",
        15553: "BioOne (Society of Conservatoire at Jardin"
               "Botaniques de la Ville de Geneve)",
        15554: "Precast/Prestressed Concrete Institute",
        15555: "Intercom - Sociedade Brasileira de Estudos"
               "Interdisciplinares da Comunicacao",
        15556: "i-das",
        15557: "Medical Communications Sp. z.o.o.",
        15558: "Fundacion para el Analisis Estrategio y Desarrollo"
               "de la Pequena y Mediana Empresa (FAEDPYME)",
        15559: "VTeX",
        15560: "Check List Journal",
        15561: "Sergii Iermakov",
        15562: "DiscoverSys, Inc.",
        15563: "Society of Allied Health Services",
        15564: "Research Institute of Education Science, Jeju"
               "National University",
        15565: "The Society of Chung-Ang Language and Literature",
        15566: "Christian Journal for Global Health",
        15567: "Croatian Dairy Union",
        15568: "Real Academia de Medicina y Cirugia de Andalucia Oriental",
        15569: "The Open Atmospheric Society",
        15570: "Association of Slovenian Dermatovenerologists",
        15571: "Korean Society for Music Theory",
        15572: "Encounters/Encuentros/Rencontres on Education",
        15573: "Association of Test Publishers",
        15574: "Group of Companies, Med Expert, LLC",
        15575: "Sunan Gunung Djati State Islamic University of Bandung",
        15576: "The University of Agriculture in Krakow",
        15577: "Sedam Sunchang Foundation",
        15578: "Agency for Marine and Fisheries Research and Development",
        15579: "Science Gate Publishing PC",
        15580: "Greener Journals",
        15581: "Universidad de Navarra",
        15582: "Indian Journal of Peritoneal Dialysis",
        15583: "The Society for Chromatographic Sciences",
        15584: "University of Rzeszow",
        15585: "Centers for Disease Control MMWR Office",
        15586: "Codon Publications",
        15587: "Private Company Technology Center",
        15588: "Zaporizhhzhia National Technical University",
        15589: "Admiral Makarov National Unviersity of Shipbuilding",
        15590: "Asian Business Consortium",
        15591: "Medicinos Mintis",
        15592: "Fundacion Universidad de Oviedo",
        15593: "PNRPU Publishing Office",
        15594: "National Institute for Health Research",
        15595: "Global Journal of Enterprise Information System",
        15596: "The Korean Society for Microsurgery",
        15597: "The Korean Society of Art Theories",
        15598: "VSB Technical University of Ostrava, Faculty"
               "of Electrical Engineering and Computer Sciences",
        15599: "Instituto Metodista de Servicos Educacionais",
        15600: "Instituto Educacional Piracicabano da Igreja Metodista",
        15601: "Instituto Metodista Izabela Hendrix",
        15602: "Instituto Porto Alegre da Igreja Metodista",
        15603: "Instituto Metodista de Ensino Superior",
        15604: "Eurasian Publications",
        15605: "Journal of the ASEAN Federation of Endocrine"
               "Sciences (JAFES)",
        15606: "Fidelio Publishing",
        15607: "Robotics: Science and Systems Foundation",
        15608: "Faculty of Agriculture, Sebelas Maret University"
               "(Journal System)",
        15609: "GENES",
        15610: "Chumakov Institute of Poliomyelitis and Viral Encephalitides",
        15611: "Wroclaw University of Economics",
        15612: "University and Research Librarians' Association",
        15613: "Vellalar College for Women",
        15614: "Indian Association of Health, Research and Welfare (IAJRW)",
        15615: "International Society of University Colon"
               "and Rectal Surgeons",
        15616: "The Korean Society for Biomedical Laboratory Sciences",
        15617: "The Korean Association of Political Science"
               "and Communication",
        15618: "Research Group on Disalination and Industrial Outfall",
        15619: "Physiotherapy New Zealand",
        15620: "Centers for Disease Control and Prevention",
        15621: "International Journal of Physiotherapy",
        15622: "SPIIRAS",
        15623: "eSAT Publishing House",
        15624: "Society of Academy of Dermatology",
        15625: "Publishing House for Science and Technology"
               "Vietnam, Vietnam Academy of Science and Technology"
               "(Publications)",
        15626: "Linnaeus University",
        15627: "SolarLits",
        15628: "Instituto Federal de Educacao, Ciencia e Tecnologia"
               "do Rio Grande do Norte (IFRN)",
        15629: "Associated Students of the University of Nevada's"
               "Nevada State Undergraduate Research Journal",
        15630: "Japanese Association of Health Psychology",
        15631: "University of Bohol",
        15632: "Polish Society of Theoretical and Applied Mechanics",
        15633: "Pontifical Universtiy of John Paul II in Krakow",
        15634: "Medical Review Co., Ltd.",
        15635: "Wydawnictwo SIGMA-NOT, sp. z.o.o.",
        15636: "The Journal of Dental Panacea",
        15637: "Journal of Life Economics",
        15638: "School of Culture and Communication, The University"
               "of Melbourne",
        15639: "Teflin Journal",
        15640: "American Research Institute for Policy Development",
        15641: "University of Cape Town",
        15642: "State Islamic University (UIN) of Sunan, Ampel",
        15643: "Sotsial'no-Gumanitarnoe Znanie Publishing House",
        15644: "University of Zagreb School of Dental Medicine",
        15645: "University of Almeria Rd  de Universidades Lectoras",
        15646: "Wydawnictwo SIGMA-NOT, sp. z.o.o. (Gaz, Woda"
               "i Technika Sanitarna)",
        15647: "Wydawnictwo SIGMA-NOT, sp. z.o.o. (Cieplownictwo,"
               "Ogrzewnictwo, Wentylacja)",
        15648: "Universidad del Atlantico",
        15649: "Universidad de Santander - UDES",
        15650: "Hebrew Union College Press",
        15651: "Societa Editrice Esculapio",
        15652: "International Network for Korean Language and Culture",
        15653: "10.15653 (Tierarztl Prax Ausg G Grosstiere Nutztiere)",
        15654: "10.15654 (Tierarztl Prax Ausg K Kleintiere Heimtiere)",
        15655: "Media Watch",
        15656: "The Korean Academy of Clinical Geriatrics",
        15657: "Breakthrough Institute Rockefeller Philanthropy Advisors",
        15658: "Universidad Cesmag",
        15659: "Makale Sistemi/Geomes",
        15660: "Asociatia Pentru Inginerie Integrata Si Management"
               "Industrial",
        15661: "Verlag Holzhausen GmbH",
        15662: "Ess and Ess Research Publications",
        15663: "University of Waikato",
        15664: "University of St. Andrews Library",
        15665: "Universidad Autonoma del Caribe",
        15666: "ALOKI Ltd",
        15667: "Laboreal, FPCE, Universidade do Porto",
        15668: "Revista Artemis",
        15669: "The Atomic Energy Society of Japan",
        15670: "Han-Character and Classical Written Language Education",
        15671: "Hacettepe Journal of Biology and Chemistry",
        15672: "Hacettepe Journal of Mathematics and Statistics",
        15673: "Odessa National Academy of Food Technologies",
        15674: "Orthopaedics, Traumatology, and Prosthetics",
        15675: "A Fundacao para o Desenvolvimento de Bauru (FunDeB)",
        15676: "School of Electrical Engineering and Informatics (STEI) ITB",
        15677: "Africa Leather and Leather Products Institute (ALLPI)",
        15678: "Uniwersytet Ekonomiczny w Krakowie - Krakow"
               "University of Economics",
        15679: "Visio Mundi Science Limited",
        15680: "Ess and Ess Research Publications",
        15681: "Korean Society on Water Environment",
        15682: "Korea Institute of Healthcare Architecture",
        15683: "The Korean Society of Disaster Information",
        15684: "FORMATH Research Group",
        15685: "Sookmyung Institute of Multiculturalism Studies",
        15686: "Feminism and Korean Literature",
        15687: "Associacao de Estudos E Pesquisas Em Politicas"
               "E Practicas Curriculares",
        15688: "Volgograd State University",
        15689: "Instituto Brasileiro de Avaliacao Psicologica (IBAP)",
        15690: "Paediatrician Publishers LLC",
        15691: "Universidad Adolfo Ibanez",
        15692: "Korean Journal of Financial Law",
        15693: "International Journal of Advanced Information"
               "Science and Technology",
        15694: "Association for Medical Education in Europe (AMEE)",
        15695: "Vanderbilt University Library",
        15696: "Journal of Nursing and Socioenvironmental Health",
        15697: "CDL Crossref test prefix",
        15698: "Shared Science Publishers OG",
        15699: "Society of Biblical Literature/SBL Press",
        15700: "Education Association of South Africa",
        15701: "Korea Computer Graphics Society",
        15702: "Korean Association of Multimedia-Assisted Language Learning",
        15703: "Korean Counseling Association",
        15704: "Korean Society for History of Education",
        15705: "The Korean Poetics Studies Association",
        15706: "The Korean Service Management Society",
        15707: "Korea Employment Agency for the Disabled/Employ,emt"
               "Development Institute",
        15708: "The Korean Society for Curriculum Studies",
        15709: "Korea Institute for Health and Social Affairs",
        15710: "Korean Association for Government Accounting",
        15711: "Society of Korean Language and Literature",
        15712: "Association of East Asian Ancient Studies",
        15713: "Incessant Nature Science Publishers Pvt Ltd",
        15714: "Scandinavian Psychologist",
        15715: "Korean Society on Communication in Healthcare",
        15716: "The Association of Korean Drama",
        15717: "The Korean Society of Biology Education",
        15718: "The Discourse and Cognitive Linguistics Society of Korea",
        15719: "Global E-Business Association",
        15720: "Korea Association of Japanese Culture",
        15721: "Korea Society of Hermeneutics",
        15722: "Korea Distribution Science Association",
        15723: "Korea Consumer Agency",
        15724: "Korean Speech-Language and Hearing",
        15725: "Institute of Philosophy",
        15726: "The Society of Korean Dances Studies",
        15727: "The Russian Law Academy of the Ministry of Justice of the RF",
        15728: "Fucape Business School",
        15729: "Instituto Nanocell",
        15730: "Symposium Books, Ltd.",
        15731: "Classical Association of South Africa",
        15732: "Institute for English Cultural Studies",
        15733: "Institute of Japanese Studies",
        15734: "The Association of Korean Education",
        15735: "Korea Logistics Society",
        15736: "Institute of Humanities",
        15737: "The Korean Society for School Science",
        15738: "Korean Association for the Study of English"
               "Language and Linguistics, KASELL",
        15739: "Journal Issues Limited",
        15740: "Hind Agri Horticultural Society",
        15741: "Revista Bio Ciencias",
        15742: "Djokosoetono Research Center",
        15743: "Brazilian Journal of Internal Medicine",
        15744: "Annex Publishers, LLC",
        15745: "Religious Studies",
        15746: "Soonchunhyang Medical Research Institute",
        15747: "Korean Society for Parental and Enteral Nutrition",
        15748: "Japan Society for Simulation Technology",
        15749: "The Korean Association of Translation Studies",
        15750: "Institute for Philosophy",
        15751: "Contemporary Film Research Institute",
        15752: "Institute for the Translation of Korean Classics",
        15753: "Asian Journal of Education",
        15754: "The Korean Philosophy of Education Society",
        15755: "Foreign Studies Institute",
        15756: "Democratic Legal Studies Association",
        15757: "Center for Studies of Christian Thoughts and Culure",
        15758: "Korean Academy of Kinesiology",
        15759: "International Journal of Enterpreneurial Knowledge",
        15760: "Portland State University Library",
        15761: "Open Access Text",
        15762: "Towarzystwo Naukowe W Toruniu",
        15763: "University of Oklahoma Libraries",
        15764: "Scientific Online Publishing, Co., Ltd.",
        15765: "Institucion Universitaria Politecnico Grancolombiano",
        15766: "Association of American Medical Colleges",
        15767: "Feminist Studies",
        15768: "Revista Brasileir de Quiropraxia - Brazilian"
               "Journal of Chiropractic",
        15779: "UC Berkeley Law Library",
        15788: "Montana State University",
        15789: "SPb RAACI",
        15790: "Korean Society of Consumer Policy and Education",
        15791: "Japanese Association of Cardioangioscopy",
        15792: "Journal of Chinese Language and Literature",
        15793: "Korea Research Institute for Human Settlements",
        15794: "English Language and Literature Association of Korea",
        15795: "Yonsei Institute for North Korean Studies",
        15796: "The Korean Association for Feminist Studies"
               "in English Literature",
        15797: "The Hallym Academyn of Sciences",
        15798: "Korea Association for International Commerce and Information",
        15799: "Institute of Korean Independence Movement Studies",
        15800: "The Regional Association of Architectural Institute of Korea",
        15801: "The Korean Association of Ethics",
        15802: "Dnipropetrovsk National University of Railway Transport",
        15803: "IJNC Editorial Committee",
        15804: "Wydawnictwo Adam Marszalek",
        15805: "Turkiye Yesilay Cemiyeti",
        15806: "World Journal of Traditional Chinese Medicine",
        15807: "The Operations Research Society of Japan",
        15808: "Libronet Bilgi Hizmetleri ve Yazilim San. Tic. Ltd., Sti.",
        15809: "Brazilian Journal of Irrigation and Drainage - IRRIGA",
        15810: "The Korean Society for Investigative Cosmetology",
        15811: "The Journal of Korean Linguistics",
        15812: "Research Institute for Science Education",
        15813: "The Knowledge Management Society of Korea",
        15814: "Korean Academic Society for Public Relations",
        15815: "The Korean Society for Culture and Arts Education Studies",
        15816: "The Altaic Society of Korea",
        15818: "Institute for Humanities and Social Sciences",
        15819: "Korean Society of Art History",
        15820: "Research Institute of Social Science",
        15821: "Law Research Institute, University of Seoul",
        15822: "The Society for Korean Language and Literary Research",
        15823: "Vytautas Magnus University Education Academy",
        15824: "Acta Medica Anatolia",
        15825: "V.I. Shimakov Federal Research Center of Transplantology"
               "and Artificial Organs",
        15826: "Ural Federal University",
        15827: "Research Institute Centerprogamsystem, JSC",
        15828: "CNT Nanostroitelstvo",
        15829: "Silicea - Poligraf, LLC",
        15830: "Korean Marketing Association",
        15831: "Korean Society for the Studey of Physical Education",
        15832: "Ankara University Faculty of Agriculture",
        15833: "Korean Association for Educational Information and Media",
        15834: "Canadian Institute of Mining, Metallurgy, and Petroleum",
        15835: "University of Agricultural Sciences and Veterinary"
               "Medicine Cluj-Napoca",
        15836: "Medical Publishing, d.o.o.",
        15837: "Agora University of Oradea",
        15838: "Vologda Research Center of the RAS",
        15839: "English and American Cultural Studies",
        15840: "Society for Historical Studies of Ancient and Medieval China",
        15841: "Korea Society for Philosophy East-West",
        15842: "Korean Clinical Psychology Association",
        15843: "Korea Productivity Association",
        15844: "Pediatric Neurology Briefs Publishers",
        15845: "University of Bergen Library",
        15846: "KYEESHA, Ltd.",
        15847: "Instituto Universitario de Lisboa (ISCTE-IUL)",
        15848: "Sociedade Brasileria de Teoria e Historia de Historiografia",
        15849: "Alzaytoonah University of Jordan",
        15850: "International Journal of Integrated Health Sciences (IJIHS)",
        15851: "Jurnal Anestesi Perioperatif (JAP)",
        15852: "Online Journal of Analytic Combinatorics",
        15853: "Kamchatka Research Institute of Fisheries and Oceanography",
        15854: "Educational Research Institute, College of"
               "Education, Ewha Womans University",
        15855: "Korean Association of Social Welfare Policy",
        15856: "Association for Korea Public Administration History",
        15857: "Korean Society of Exercise Physiology",
        15858: "Korea Association of Teachers of English",
        15859: "Yeol-song Society of Classical Studies",
        15860: "Korean Generative Grammar Circle",
        15861: "Korean Society of Special Education",
        15862: "Publishing Company World of Science LLC",
        15863: "International Academy of Theoretical and Applied Sciences",
        15864: "Society for Makers, Artists, Researchers and Technologists",
        15865: "American Association of Chemistry Teachers",
        15866: "Praise Worthy Prize, s.r.l.",
        15867: "Gleeson Library Geschke Center",
        15868: "Foundation Center",
        15869: "Journal of the Human and Social Services Researchers",
        15870: "The Research Institute for Special Education"
               "and Rehabilitation Science",
        15871: "Revista Brasileira de Productos Agroindustriais",
        15872: "The Korean Workshop for the History of Culture",
        16873: "The Korean Literature Association",
        16874: "Korean Society of Study on Chinese Language and Culture",
        16875: "Society for Teaching English through Media",
        16876: "Korean Language Research Circle",
        16877: "Korean Society for Dance Studies",
        16878: "Galatasaray Universitesi Iletisim",
        16879: "The Korean Cartographic Association",
        16880: "The Study of the Eastern Classic",
        16881: "Institute of Social Sciences, Chungnam National University",
        16882: "Horticultural Studies (Hortis)",
        16883: "Istanbul University Faculty of Pharmacy",
        16884: "Korean Society for Rehabilitation of Persons"
               "with Disabilities",
        16885: "Society of Korean Traditional Costume",
        16886: "Agricultural University of Iceland",
        16887: "FIEP Bulletin Online",
        16888: "Centro Interdisciplinario de Investigaciones"
               "en Psicologia Matematic",
        16889: "Isomer Design",
        16890: "Tribunal Permanente de Revisao do MERCOSUL",
        16891: "Revista Interfaces: Saude, Humanas e Tecnologia",
        16892: "Library and Information Centre, Hungarian"
               "Academy of Sciences",
        16893: "International Association for Buddhist Thought and Culture",
        16894: "Journal of Western History",
        16895: "The Korean Association of Catholic Philosophers",
        16896: "Korean Association of Animal Assisted Psychotherapy",
        16897: "The Korean Isabu Society",
        16898: "Korean Association of Criminal Case Studies",
        16899: "Cagdas Tip Dergisi: Journal of Contemporary Medicine",
        16900: "Society of Onji Studies",
        16901: "Association of Western Art History",
        16905: "University of Bern",
        16910: "University of Bern",
        16912: "Korean Historical Association",
        16913: "Global Senior Health Promotion Institute",
        16914: "Korean Association for Advertising and Public Relations",
        16915: "Korean Physical Education Association for Women",
        16916: "Ana Dili Egitimi Dergisi ADED",
        16917: "Turkish Journal of Sociology",
        16918: "Balkan Libraries Union",
        16919: "Bozok Tip Dergisi",
        16920: "Rajarambapu Institute of Technology",
        16921: "CIESPAL",
        16922: "University of Wales Press/Gwasg Prifysgol Cymru",
        16923: "Sociedade Brasileira de Bioquimica e Biologia"
               "Molecular - SBBq",
        16924: "Revista de Ingenieria",
        16925: "Universidad Cooperativa de Colombia - UCC",
        16926: "Uniwersytet Humanistyczno-Przyrodniczy im."
               "Jana D?ugosza w Cz?stochowie",
        16927: "Global Institute of Forensic Research, LLC",
        16928: "PIDCC- Revista em Propriedade Intelectual"
               "- Direito Contemporaneo e Constituicao",
        16929: "Statistics and Probability African Society (SPAS)",
        16930: "Revista Catarinense da Ciencia Contabil",
        16931: "Annals of Surgical Hepatology",
        16932: "Japan Link Center",
        16933: "Institute for Foreign Language Education",
        16934: "International Studies Review",
        16935: "Ewha Journal of Social Sciences",
        16936: "Theology and Philosophy",
        16937: "Korea Culture and Tourism Institute",
        16938: "The Korea Academic Society of Trade Exhibition Studies",
        16939: "The Musicological Society of Korea",
        16940: "Yonsei Music Research Institute",
        16941: "Korean Society for Education Through Art",
        16942: "The Soong Sil Association for Historical Studies",
        16943: "Indian National Science Academy",
        16944: "Journal of Environment and Sociobiology",
        16945: "International Association for Humanistic Studies in Language",
        16946: "Korean Society for Schizophrenia Research",
        16947: "FSM Scholarly Journal of Humanities ands Social Sciences",
        16948: "Medical Bulletin of Zeynep Kamil",
        16949: "Turkish Journal of Computer and Mathematics Education",
        16950: "Inonu University Journal of Art and Design",
        16951: "Journal of Economics and Administrative Sciences",
        16952: "The Research Institute of Nursing Science,"
               "Seoul National University (KAMJE)",
        16953: "Dokuz Eylul University Graduate School of Socal Sciences",
        16954: "Baris Arastirmalari",
        16955: "Plant Protection Bulletin",
        16956: "Korean Association of Thyroid and Endocrine Surgeons",
        16957: "Institute for the Study of History",
        16958: "Dankook Center for Dispute Resolution",
        16959: "Institute for East Asian Cultures",
        16960: "The Law Research Institute of Hongik University",
        16961: "Institute of Police Science",
        16962: "ELK Education Consultants, Pvt., LTD.",
        16963: "Library of Medicine, LLC",
        16964: "Eye Reports",
        16965: "I MED Research Publications",
        16966: "Sci Forschen, Inc.",
        16967: "Fundacion Universitaria Ceipa",
        16968: "Korean Geomorphological Association",
        16969: "Turkish Bulleting of Entomology",
        16970: "Turkish Journal of Entomology",
        16971: "Vakiflar Dergisi",
        16972: "The Korean Society of Business Venturing",
        16973: "Journal of Governance Studies",
        16974: "Tax Law Association",
        16975: "The Korean Academy of Family Social Work",
        16976: "Studies on History Education",
        16977: "The Japanese Society of Cerebral Blood Flow and Metabolism",
        16978: "The Korean Society for Early Childhood Education and Care",
        16979: "The Japanese Modern Association of Korea",
        16980: "Korea International Trade Research Institute",
        16981: "Korean Library and Informatiomn Science Society",
        16982: "Koreanische Nietzsche Gesellschaft",
        16983: "Korean School Pyschology Association",
        16984: "Sakarya University Journal of Science",
        16985: "Marmara Turkiyat Arastirmalari Dergisi",
        16986: "Hacettepe University Journal of Education",
        16987: "Ankara Universitesi SBF Dergisi",
        16988: "Journal of the Faculty of Veterinary Medicine"
               "Istanbul University",
        16989: "The Journal of Turk-Islam World Social Studies",
        16990: "The Journal of Social Sciences",
        16991: "The Journal of International Education Sciences",
        16992: "The Journal of Academic Social Sciences",
        16993: "Stockholm University Press",
        16994: "Modern Academic Publishing Partners",
        16995: "Open Library of Humanities",
        16996: "Open Commications Press",
        16997: "University of Westminster Press",
        16998: "Korean Association for Arbitration Studies",
        16999: "Korea Association for Social Welfare Studies",
        17000: "Korean Association of Social Policy",
        17001: "The Society for Kugyol Studies",
        17002: "The Joongwon Linguistic Society of Korea",
        17003: "The Japanese Language and Literature Association of Korea",
        17004: "The Society for the Research of Chinese Novels",
        17005: "Korean Association of Labor Studies",
        17006: "Korean Society of Adapted Physical Activity and Exercise",
        17007: "Korean Lawyers Association",
        17008: "Institute of Legal Studies, Sungkyunkwan University",
        17009: "The Shakespeare Association of Korea",
        17010: "Associated Management Consultants, PVT, Ltd.",
        17011: "Jyvaskyla University Library",
        17012: "Marketing Aumentado",
        17013: "AISTI",
        17014: "Indonesian Journal of Geoscience",
        17015: "International Ataturk Alatoo University",
        17016: "Board of Governors of the Federal Reserve System",
        17017: "Journal of Fisheries",
        17018: "Croatian Conservation Institute",
        17047: "Hadtudomány",
        17048: "Eszterhazy Karoly Egyetem",
        17049: "Journal of Anatolia Nursing and Health Sciences",
        17050: "Kafkas University Faculty of Divinity Review",
        17051: "Ilkogretim Online",
        17052: "The Korean Society of Contemporary European Studies",
        17053: "Korean Consumption Culture Association",
        17054: "The Medieval and Early Modern English Studies"
               "Association of Korea",
        17055: "The Society of Korean Practical Arts Education",
        17056: "The Donam Language and Literature",
        17057: "Korea Association for History of Modern Art",
        17058: "APESC - Associacao Pro-Ensino em Santa Cruz do Sul",
        17059: "Institute of Economics of the Ural Branch of the RAS",
        17060: "Asociacion INFAD",
        17061: "The Sax Institute",
        17062: "Health Services Library, Creighton University",
        17063: "Brazilian Journal of Forensic Sciences, Medical"
               "Law and Bioethics",
        17064: "Istanbul Universitesi Iletisim Fakultesi Dergisi",
        17065: "Hacettepe University Journal of Economics"
               "and Administrative Sciences",
        17066: "Turkish Psychological Conseling and Guidance Journal",
        17067: "Asia Minor Studies",
        17068: "Korean Society for Local History and Culture",
        17069: "Korea Society of Lighting and Visual Environment",
        17070: "The Association of East Asian Ancient Studies",
        17071: "Scireps Corporation",
        17072: "Perm State University (PSU)",
        17073: "National University of Science and Technology MISiS",
        17074: "Faculdade de Letras, UFRJ",
        17075: "Consello da Cultura Galega",
        17076: "Karelian Research Centre of the Russian Academy of Sciences",
        17077: "University of Iowa",
        17078: "Reykjavik University School of Law",
        17079: "Korean Gerontological Nursing Society",
        17080: "Tropical Journal of Fisheries and Aquatic Sciences",
        17081: "Universidad Simon Bolivar",
        17082: "Queensland Museum",
        17083: "Serious Games Society",
        17084: "Komsomolsk-na-Amure State Technical Unversity",
        17085: "The Editorial Office of Anesthesia and Pain Medicine",
        17086: "The Tourism Sciences Society of Korea",
        17087: "Bombay Natural History Society",
        17088: "The Korea Society of Yang-Ming Studies",
        17089: "The Korean Association for Governance",
        17090: "Society of Korean Classical Woman Literature",
        17091: "Korean Society of World Music",
        17092: "The Korea Academy of International Business Education",
        17093: "Alphanumeric Journal",
        17094: "Ataturk Universitesi Veteriner Bilimleri Dergisi",
        17095: "Korea Reading Association",
        17096: "Journal of Istanbul University Faculty of Dentistry",
        17097: "Ataturk University Journal of the Agricultural Faculty",
        17098: "Ankara Medical Journal",
        17099: "Forestist",
        17100: "Nevsehir Bilim ve Teknoloji Dergisi",
        17101: "Bertalanffy Center for the Study of Systems Science",
        17102: "College of Natural Resources (Royal University of Bhutan)",
        17103: "Asociacion Espanola de Profesores de Derecho"
               "Internacional y Relaciones Internacional",
        17104: "Verlag C.H.Beck oHG",
        17105: "National Association of School Psychologists",
        17106: "Japanese Society of Biorheology",
        17107: "Kaleidoscope History",
        17108: "Acta Agronomica Ovariensis",
        17109: "Acta Zoologica Hungarica",
        17110: "Studia Botanica Hungarica",
        17111: "Fragmenta Palaeontologica Hungarica",
        17112: "Folia Entomologica Hungarica",
        17113: "Faculty of Food Technology and Biotechnology"
               "- University of Zagreb",
        17114: "University 1 December 1918 of Alba Iulia",
        17115: "LinkSciencePlace",
        17116: "Media Sphere Publishing Group",
        17117: "Consulting Company Ucom",
        17118: "Universite de Sherbrooke",
        17119: "The Japanese Society of Ryodoraku Medicine",
        17120: "Ondokuz Mayis University Review of the Faculty of Divinity",
        17121: "Route Educational and Social Science Journal",
        17122: "Ufa State Petroleum Technological University",
        17123: "Anatolia: Turizm Arastirmalri Dergisi",
        17124: "Istanbul Universitesi Siyasal Bilgiler Fakultesi Dergisi",
        17125: "Florida State University College of Medicine",
        17126: "Facultad de Odontologia, Universidad de Concepcion",
        17127: "CEGOT - Center of Studies on Geography and Spatial Planning",
        17128: "ESRI Press",
        17129: "Botanical Sciences, Sociedad Botanica de Mexico",
        17130: "International Journal of Management Economics and Business",
        17131: "Milel ve Nihal",
        17132: "Yale University Library",
        17133: "Journal of Turkology Research",
        17134: "The Journal of Defense Sciences",
        17135: "The Korean Society of Dental Hygiene Science",
        17136: "The Korean Society of Inherited Metabolic Disease",
        17137: "Korea Organic Resource Recycling Association",
        17138: "Tropical Grasslands - Forrajes Tropicales",
        17139: "Asociacion de Antropologia Biologia Argentina",
        17140: "Openventio Publishers",
        17141: "Facultad Latinoamericana de Ciencias Sociales (FLACSO)",
        17142: "Insan Akademika Publications",
        17143: "ABED - Associacao Brasileira de Educacao a Distancia",
        17144: "Bio 2.0",
        17145: "Betasciencepress Publishers",
        17146: "National Atomic Energy Agency of Indonesia (BATAN)",
        17147: "Alfons W. Gentner Verlag",
        17148: "Tejass Publisheers",
        17149: "Darswin Publishing House",
        17150: "Baikal State University",
        17151: "Universidad de Caldas",
        17152: "Gazi Egitim Fakultesi Dergisi",
        17153: "Eskisehir Osmangazi University Journal of"
               "Economics and Administrative Sciences",
        17154: "Applied Linguistics Association of Korea",
        17155: "Journal of Sports and Performance Researches",
        17156: "Korea Evangelical Theologial Society",
        17157: "Medicine Anthropology Theory",
        17158: "University of the Immaculate Conception",
        17159: "Academy of Science of South Africa",
        17160: "JOSHA, GmbH",
        17161: "The University of Kansas",
        17162: "Universidad Peruana Union",
        17163: "Salesian Polytechnic University of Ecuador",
        17164: "Erdeszettudomanyi Kozlemenyek",
        17165: "Kepzes es Gyakorlat",
        17184: "Editions des archives contemporaines",
        17206: "Pacific Early Childhood Education Research Association",
        17207: "Association of East Asian Social Thoughts",
        17208: "Korea Planners Association",
        17209: "Korean Society for Social Theory",
        17210: "The HCI Society of Korea",
        17211: "Turkish Geographical Review",
        17212: "Novosibirsk State Technical University",
        17213: "South Russian State Polytechnic University (NPI)",
        17214: "Acta Odontologica Turcica",
        17215: "Korean Association of Medical Law",
        17216: "Journal of Limnology and Freshwater Fisheries Research",
        17217: "Kamchatka State Technical University",
        17218: "Hitit Universitesi Sosyal Bilimler Enstitusu Dergisi",
        17219: "Wroclaw Medical University",
        17220: "Kiperonline Academic",
        17221: "Czech Academy of Agricultural Sciences",
        17222: "Institute of Metals and Technology",
        17223: "Tomsk State University",
        17224: "EIA Energy in Agriculture",
        17225: "Haemnet",
        17226: "The National Academies Press",
        17227: "Universidad Pedagogica Nacional",
        17228: "Avotaynu, Inc.",
        17229: "Open Medscience Limited",
        17230: "Universidad EAFIT",
        17231: "Centro de Estudos de Comunicacao e Sociedade,"
               "Universidade Minho",
        17232: "Korean Society for Educational Technology",
        17233: "Sosyoekonomi",
        17234: "Faculty of Humanities and Social Sciences,"
               "University of Zagreb",
        17235: "Sociedad Espanola de Patologia Digestiva (SEPD)",
        17236: "Gesellschaft Schweizer Tierarztinnen und Tierarzte",
        17238: "Kniga - Service Agency Ltd.",
        17239: "ARLE (International Associaton for Research in L1 Education)",
        17240: "Abant Izzet Baysal Universitesi Egitim Fakultesi Dergisi",
        17241: "Korean Society of Sleep Medicine",
        17242: "Magyar Vizivad Kozlemenyek",
        17243: "Magyar Aprovad Kozlemenyek",
        17244: "Egitimde Kuram ve Uygulama",
        17245: "The Korean Dental Society of Anesthesiology",
        17246: "Korea Design and Knowledge Society",
        17247: "The Society of Korean Language and Literature Education",
        17248: "Law Research Institute of KNU",
        17249: "Insitute of Cross-Cultural Studies",
        17250: "Institute for the Study of Language and Information,"
               "Kyung Hee University",
        17251: "Legal Research Institute Kookmin University",
        17252: "Institute for Legal Study",
        17253: "Sung Shin Women's University Educational Reseaerch Institute",
        17254: "Ewha Music Research Institute",
        17255: "E-Trade Review",
        17256: "Hanyang Economic Research Institute",
        17257: "HUFS Law Research Institute",
        17258: "The Institute of Korea Bibliography",
        17259: "The Classic and English Renaissance Literature"
               "Association of Korea",
        17260: "The Association of Korea Literature in Chinese",
        17261: "Pressacademia",
        17262: "Institute of Health and Environment",
        17263: "Journal of Language and Linguistic Studies",
        17264: "Marianna University Society of Medical Science",
        17265: "David Publishing Company",
        17266: "International Network for Social Network Analysts",
        17267: "Escola Bahiana de Medicine e Saude Publica",
        17268: "Universidad Nacional de Trujillo",
        17269: "Canadian Public Health Association",
        17270: "Wyzsza Szkola Logistyki (Poznan School of Logistics)",
        17271: "ANAP - Associacao Amigos de Natureza de Alta Paulista",
        17272: "Subsea Diving Contractor di Stefano Di Cagno Publisher",
        17273: "IndorSoft LLC",
        17274: "Infotech",
        17275: "Participatory Educational Research (Per)",
        17277: "Tambov State Technical University",
        17278: "International Journal of Educational Studies"
               "in Mathematics (IJESIM)",
        17279: "The Journal of the Korean Association of Geographic"
               "and Environmental Education",
        17280: "Korea Digital Design Council",
        17281: "The Hegel Society of Korea",
        17282: "Korean Elementary Moral Education Society",
        17283: "Korean Educational Idea Association",
        17284: "Fordham University Press - Undergradate Journals (FURJ)",
        17285: "CSRI Elektropribor",
        17286: "Korean Educational Psychology Association",
        17287: "Korean Academic Society of Business Administration",
        17288: "Chungwoon University Broadcasting and Arts"
               "Research Institute",
        17289: "Korea Society of Color Studies",
        17290: "The Linguistic Society of Korea",
        17291: "The Society of Korean Language and Literature",
        17292: "The Korean Association for Mongolian Studies",
        17293: "Dong-Bang Korean Chinese Language Society",
        17294: "Aurora Health Care, Inc.",
        17295: "Dogu Cografya Dergisi",
        17296: "The Korean Society of Bilingualism",
        17297: "Journal of the Research Society of Language and Literature",
        17298: "Korea International Economic Association",
        17299: "The Society of Eastern Philosophy",
        17300: "Association of Dongak Art History",
        17301: "Theological Forum",
        17302: "Arc Humanities Press",
        17303: "JScholar",
        17304: "Indonesian Journal of International Law",
        17305: "Association of Basic Medical Sciences of FBIH",
        17306: "Uniwersytet Przyrodniczy w Poznaniu (Poznan"
               "University of Life Sciences)",
        17307: "Wisconsin Space Grant Consortium",
        17308: "Voronezh State University",
        17309: "LLC OVS",
        17310: "National Tax Association",
        17311: "Science International",
        17312: "Harrington Park Press",
        17313: "Journal of Korean Language Education",
        17314: "Japanese Language and Culture Association of Korea",
        17315: "Korean Health Psychological Association",
        17316: "Korean Association of Feminist Philosophers",
        17317: "The Korea Society for Educational Law",
        17318: "The Korean Society for the Study of Anthropology"
               "of Education",
        17319: "The Society of Korean Classical Literature Education",
        17320: "Academy of Asian Studies, Dankook University",
        17321: "Institute for the Study of North Korean Life",
        17322: "Sage - Corwin",
        17323: "National Research University, Higher School"
               "of Economics (HSE)",
        17324: "International Fiscal Association of Korea",
        17325: "Sogang Journal of Philosophy",
        17326: "Journal of Humanities",
        17327: "Institute of Public Policy and Administration",
        17328: "Korea Assocation for Cultural Sociology",
        17329: "Associatio for Korea Culture Studies",
        17330: "Korea Environment Institute",
        17331: "Institute for Far Eastern Studies",
        17333: "Korean Society for Simulation in Nursing",
        17334: "Journal of Pediatric Sciences",
        17335: "Sakarya Universitesi Ilahiyat Fakultesi Dergisi",
        17336: "Istanbul Gelisim University Journal of Social Sciences",
        17337: "Korean Society of Vision Science",
        17338: "Active Aging Research Center",
        17339: "EJOVOC",
        17340: "Korean Neurological Association",
        17341: "Journal of the Faculty of Engineering and"
               "Architecture of Gazi University",
        17342: "Korean Insurance Academic Society",
        17343: "Medical Journal of Suleyman Demirel University",
        17344: "Slovenian Chemical Society",
        17345: "Universitat Rovira I Virgili",
        17346: "Escola Superior de Educacao de Paula Frassinetti",
        17347: "Innovative Pedagogical Technologies LLC",
        17348: "Botanical Research Institute of Texas",
        17349: "University of Lincoln, School of Film and"
               "Media and Changer Agency",
        17350: "Hittite Journal of Science and Engineering",
        17351: "Society for Social Studies of Science (4S)",
        17352: "Peertechz Publications Private Limited",
        17353: "All-Russia Petroleum Research Exploration Institute (VNIGRI)",
        17354: "IJSS Group of Journals",
        17355: "Regio",
        17356: "Intersections",
        17357: "Emergent Publications",
        17358: "Graduate Program of Management and Business,"
               "Bogor Agricultural University",
        17359: "GuvenGrup, Ltd. - International Journal of"
               "Family Child and Education",
        17360: "GuvenGrup, Ltd. - International Journal of"
               "Psychiatry and Psychological Researches",
        17361: "GuvenGrup., Ltd. - International Peer-Reviewed"
               "Journal of Communication and Humanities Researches",
        17362: "GuvenGrup, Ltd. - International Peer-Reviewed"
               "Journal of Nutrition Research",
        17363: "GuvenGrup, Ltd. - International Refereed Academic"
               "Journal of Sports",
        17364: "GuvenGrup Ltd. - International Referred Academic"
               "Social Sciences Journal",
        17365: "GuvenGrup, Ltd. - International Referred Journal"
               "of Design and Architecture",
        17366: "GuvenGrup, Ltd. - International Refereed Journal"
               "of Engineering and Sciences",
        17367: "GuvenGrup, Ltd. - International Refereed Journal"
               "of Gynaecological Diseases and Maternal and Child Health",
        17368: "GuvenGrup, Ltd. - International Refereed Journal"
               "of Humanities and Academic Sciences",
        17369: "GuvenGrup, Ltd. - International Referred Journal"
               "of Marketing and Market Researches",
        17370: "GuvenGrup, Ltd. - International Refereed Journal"
               "of Music Researches",
        17371: "GuvenGrup, Ltd. - International Referred Journal"
               "of Nursing Researches",
        17372: "GuvenGrup, Ltd. - International Refereed Journal"
               "of Orthopaedics Traumatology and Sports Medicine",
        17373: "GuvenGrup, Ltd. - International Refereed Journal"
               "of Research on Economics Management",
        17381: "CNBOP-PIB Centrum Naukowo-Badawcze Ochrony Przeciwpozarowej",
        17393: "Institute of Psychiatry and Neurology",
        17398: "Universidad de Extremadura - Servicio de Publicaciones",
        17409: "Kulttuuripoliittisen tutkimuksen vuosikirja",
        17428: "El Colegio de la Frontera Norte A.C.",
        17430: "Journal of Hearing Science",
        17431: "Nowa Audiofonologia",
        17450: "SoftPower",
        17454: "Lexis srl",
        17456: "Associazione Laureati/E In Lingue All",
        17473: "G. Giappichelli Editore srl",
        17474: "Artvin Coruh Universitesi Orman Fakultesi Dergisi",
        17475: "Kastamonu University Journal of Forestry Faculty",
        17476: "Korean Society for Metabolic and Bariatric Surgery",
        17477: "World Association for Triple Helix and Future"
               "Strategy Studies",
        17478: "Journal for Educating Gifted Young Scientists (JEGYS)",
        17479: "Korean Society of Acute Care Surgery",
        17480: "The Pharmaceutical Society of Korea",
        17481: "The Korean Ophthamological Society",
        17482: "Uludag University Journal of the Faculty of Engineering",
        17483: "Canadian Association of Schools of Nursing",
        17484: "YEDI: Journal of Art, Design and Science",
        17485: "Indian Society for Education and Environment",
        17486: "Studentska Sekcija za Neuroznanost",
        17487: "RFC Editor",
        17488: "Revista Mexicana de Ingenieria Biomedica",
        17489: "University of Debrecen/Biomechanica Hungarica",
        17490: "Sanat Tasarim Dergisi",
        17491: "Geological Society of India",
        17492: "Journal Press India",
        17493: "NITTE Management Review",
        17494: "Eskisehir Osmangazi University Journal of Social Sciences",
        17495: "The East Asian Society of Dietary Life",
        17496: "Yonsei University College of Medicine",
        17497: "Turkish History Education Journal",
        17498: "Dergi Karadeniz",
        17499: "Journal of Social Studies Education Research",
        17501: "The International Institute of Knowledge Management,"
               "Pvt, Ltd.",
        17502: "Methodos.revista de ciencias sociales",
        17503: "Agrivita, Journal of Agricultural Science (AJAS)",
        17504: "ZappyLab, Inc.",
        17505: "Journal for Person-Oriented Research, Lund"
               "University Library",
        17506: "Institute of Philosophy and Law, Ural Branch of the RAS",
        17507: "Academy Publication",
        17508: "Faculty of Food Technology Osijek",
        17509: "Universitas Pendidikan Indonesia",
        17510: "Wacana, Journal of the Humanities of Indonesia",
        17511: "Siddharth Health Research and Social Welfare Society",
        17512: "Czestochowa University of Technology",
        17513: "Eurasian Scientific and Industrial Chamber, Ltd.",
        17514: "International Association for Near-Death Studies",
        17515: "MIM Research Group",
        17516: "Siberian Federal University",
        17517: "Kahramanmaras Sutcu Imam Universitesi Tip Fakultesi Dergisi",
        17518: "Canakkale Arastirmalari Turk Yilligi Dergisi",
        17519: "The Apicultural Society of Korea",
        17520: "Biodiversity Science",
        17521: "Chinese Journal of Plant Ecology",
        17522: "Necatibey Faculty of Education Electronics"
               "Journal of Science and Mathematics Education",
        17523: "Instituto do Zootecnia",
        17524: "ABRACICON: Academia Brasileira de Ciencias Contabeis",
        17525: "Sociedade Brasileira de Virologia",
        17526: "Goverment Innovation and Productivity Institute",
        17527: "Korean Society of Aesthetics and Science of Art",
        17528: "Center for International Forestry Research (CIFOR)",
        17529: "Jurnal Rekayasa Elektrika (JRE)",
        17530: "The Journal of Engineering and Fundamentals",
        17531: "Polskie Naukowo-Techniczne Towarzystwo Eksploatacyjne",
        17532: "University of Sarajevo Faculty of Health Sciences",
        17533: "Universidad de Antioquia",
        17534: "Knowledge Ecology International",
        17535: "Croatian Operational Research Society",
        17536: "Asia Pacific Association for Gambling Studies",
        17537: "Institute of Mathematical Problems of Biology"
               "of RAS (IMPB RAS)",
        17538: "Physician Assistant Education Association",
        17539: "Amasya Universitesi Egitim Fakultesi Dergisi",
        17540: "Hikmet Yurdu Dusunce-Yorum Sosyal Bilimler Arastirma Dergisi",
        17541: "Optimum Journal of Economics and Management Sciences",
        17542: "Kitaibelia",
        17543: "Iskolakultura",
        17545: "Conselho Brasileiro do Oftalmologia",
        17546: "Medical Science and Discovery",
        17547: "Korean Society of Stress Medicine",
        17548: "Institute of Science Culture Exhibition and Design",
        17549: "People and Global Business Association",
        17550: "Journal of Academic Inquiries",
        17551: "Revista Geoamazonia",
        17552: "Boletim Amazonico de Geografia",
        17553: "Revista do Instituto Historico e Geografico do Para",
        17554: "ACT Publishing Group",
        17555: "The Korean Society of Veterinary Clinics",
        17556: "Erzincan University Journal of Education Faculty",
        17557: "Turkish Journal of Field Crops",
        17559: "Mechanical Engineering Faculty in Slavonski Brod",
        17560: "Vulkan-Verlag GmbH",
        17561: "Universidad de Jaen",
        17562: "Centro de Innovacion y Desarrollo Tecnologico en Computo",
        17563: "SBV Sociedade Brasileira de Vacuo",
        17564: "Universidade Tiradentes",
        17565: "Revista GESTA",
        17566: "Cadernos Ibero-Americanos de Direito Sanitario",
        17567: "The Journal of Faculty of Dentistry of Ataturk University",
        17568: "Ormancilik Arastirma Dergisi",
        17569: "Tukish Online Journal of Qualitative Inquiry",
        17570: "Evolve Publishing",
        17571: "Asian Scientific Publishers",
        17572: "Moment Journal",
        17573: "Faculty of Administration, University of Ljubljana",
        17574: "TESAM Ekonomik, Siyasal ve Stratejik Arastirmalar Merkezi",
        17575: "Associacao Portuguesa de Psicologia",
        17576: "Penerbit Universiti Kebangsaan Malaysia (UKM Press)",
        17577: "ESRSA Publications Pvt. Ltd.",
        17578: "Global Business Publications",
        17579: "Sociedad Espanola de Patologia Dual (SEPD)",
        17580: "Ore and Metals Publishing House",
        17581: "Botanical Garden - Institute of the Far Eastern"
               "Branch of the RAS",
        17582: "ResearchersLinks Ltd",
        17583: "Hipatia Press",
        17584: "Sociedad Colombiana de Ciencias Horticolas",
        17585: "Co-Action Publishing Cappelen Damms",
        17586: "ITMO University",
        17587: "New Technologies Publishing House",
        17588: "Ivanovo State Power University (ISPU)",
        17589: "Russian Law Journal",
        17602: "Duke University Libraries - Morphosource",
        17613: "Modern Language Association",
        17628: "European Chemical Bulletin",
        17640: "Korean Society of Water Science and Technology",
        17641: "The Korean Association for Play Therapy",
        17642: "Journal of Chitin and Chitosan",
        17643: "The Korea Society for Child Education (KSCE)",
        17644: "Hacettepe Journal of Sport Sciences",
        17645: "Cogitatio",
        17646: "KOME Journal",
        17647: "University of Seoul Institute of Seoul Studies",
        17648: "Galoa Events Proceedings",
        17649: "Ter es Tarsadalom Journal",
        17650: "Publishing House ABV Press",
        17651: "Institute of Polish Language",
        17652: "Sociedade Brasileira de Psicologia Organizacional"
               "e do Trabalho",
        17653: "ScienceScript, LLC",
        17654: "Pushpa Publishing House",
        17655: "Cerceau",
        17656: "Journal of Zankoy Sulaimani - Part A",
        17657: "Korea Distribution Association",
        17658: "Paul Mellon Centre for Studies In British Art",
        17659: "Journal of Case Reports",
        17660: "International Society for Horticultural Science (ISHS)",
        17661: "The Korean Institute of Information & Electronic"
               "Communication Technology",
        17662: "Korea Society of Digital Industry and Information Managment",
        17663: "Korean Wetlands Society",
        17664: "Korea Society of Geothermal Energy Engineers",
        17665: "EScola de Enfermagem Aurora de Afono Costa - UFF",
        17666: "ANPOCS",
        17667: "University of Debrecen/Recent Innovations in Mechatronics",
        17668: "Studia Slavica Savariensia",
        17670: "Magyar Pedagogia",
        17671: "International Journal of Informatics Technologies",
        17672: "Florence Nightingale Journal of Nursing",
        17673: "Samara State Technical University",
        17674: "Ufa State Academy of Arts",
        17675: "IFHAN",
        17676: "Hungarian Agricultural Engineering",
        17677: "P2M Invest",
        17678: "Bitlis Eren University Journal of Science and Technology",
        17679: "Inonu University Journal of the Faculty of Education",
        17680: "Erciyes Iletisim Dergisi",
        17681: "SAglik Bilimleri ve Meslekleri Dergisi",
        17682: "United States Sports Academy",
        17683: "INCDMTM",
        17684: "Diacronia",
        17685: "Croatian Society of Art Historians",
        17686: "Science and Education, Ltd.",
        17687: "Journal of Entrepreneurship and Business (JEB)",
        17688: "Catholic Theological Union",
        17689: "Penza Psychological Newsletter",
        17690: "The Finnish Society of Photogrammetry and Remote Sensing",
        17691: "Journal Sovremennye Tehnologii v Medicine",
        17692: "Laboratorio de Anatomia Comparada dos Vertebrados",
        17693: "Yunus Arastirma Bulteni",
        17694: "Balkan Journal of Electrical & Computer Engineering (BAJECE)",
        17695: "Faculdade de Enfermagem Nova Esperanca",
        17696: "Faculdade de Medicina de Sao Jose do Rio Preto - FAMERP",
        17697: "Institute of Business Management and Rural Development",
        17698: "Uni-FACEF Centro Universitario de Franca",
        17699: "Nucleo para o Desenvolvimento de Tecnologia"
               "e Ambientes Educacionais (NPT)",
        17700: "Journal of Agricultural Informatics",
        17701: "Karsztfejlodes",
        17702: "The Society of Adhesion and Interface, Korea",
        17703: "The International Promotion Agency of Culture Technology",
        17704: "History of the Earth Sciences Society HESS",
        17705: "Association for Information Systems",
        17706: "International Academy Publishing  (IAP)",
        17707: "Journal Agriculture and Forestry",
        17708: "The Slovenian Academy of Management",
        17709: "QUASAR, LLC",
        17710: "ASICE",
        17711: "Instituto Nacional de Psiquiatria Ramon de la Fuente Muniz",
        17712: "Neurosciences",
        17713: "Austrian Statistical Society",
        17714: "Gumushane University Journal of Science and"
               "Technology Institute",
        17715: "The Korean Society for the Study of Moral Education",
        17716: "Botanikai Kozlemenyek",
        17717: "Iskolakultura-konyvek",
        17718: "Turkish Online Journal of Distance Education",
        17719: "The Journal of International Social Research",
        17720: "The Classical Studies Foundation",
        17721: "Taras Shevchenko National University of Kyiv",
        17722: "Techmind Research Society",
        17723: "Society of American Archivists",
        17724: "The International Child Neurology Association (ICNA)",
        17725: "Editorial Board of Journal Radioelectronics,"
               "Nanosystems, Information Technology RENSIT",
        17726: "Pyatigorsk State University",
        17727: "KIMS Foundation and Research Center",
        17728: "Indonesian Food Technologists",
        17729: "Instytut Spawalnictwa",
        17730: "Society for Applied Anthropology",
        17731: "Entomological Society of Israel",
        17732: "The NCHERM Group, LLC",
        17735: "AEQUA & SEG",
        17736: "International Society of Offshore and Polar Engineers",
        17737: "Future Energy Service and Publishing",
        17738: "Austrian Geological Society (OGG)",
        17739: "EBA Conferences",
        17740: "Eurasian Academy of Sciences",
        17741: "The Geological Society of Finland",
        17742: "Immaginations: Journal of Cross-Cultural Image Studies",
        17743: "Audio Engineering Society",
        17744: "American Mental Health Counselors Association",
        17745: "Bumhan Philosophical Society",
        17746: "Institute of Archaeology and Ethnography SB RAS",
        17747: "Real Economy Publishing",
        17748: "Historical and Social Educational Ideas",
        17749: "IRBIS",
        17750: "Kazan Medical Journal",
        17751: "Daegu Historical Association",
        17752: "Journal of Security Strategies",
        17753: "Ekev Akademi Dergisi",
        17754: "Modern Studies in English",
        17755: "Electronic Journal of Social Sciences",
        17756: "United Scientific Group",
        17757: "The Family in America",
        17758: "Universal Researchers",
        17759: "Federal State-Financed Educational Institution"
               "of Higher Education Moscow State University"
               "of Psychology and Education",
        17760: "Northeastern University Library",
        17761: "International Association of Yoga Therapists",
        17762: "Auricle Technologies Pvt., Ltd.",
        17763: "Harvard Education Publishing Group",
        17764: "Institute of Environmental Sciencs and Technology (IEST)",
        17765: "Centro Universitario de Maringa",
        17766: "Raizes e Amidos Tropicais/Tropical Roots and Starches",
        17767: "Sociedade Brasileira de Dermatologia",
        17768: "Panorama of Brazilian Law",
        17770: "Rezekne Academy of Technologies",
        17771: "Faculdades Catolicas",
        17772: "Polskie Towarzystwo Ginekologiczne",
        17773: "Fondazione per la Flora italiana",
        17774: "Considering Disability Journal",
        17775: "Power System Technology Press",
        17776: "Cumhuriyet University Sciences Journal",
        17777: "Jangjeon Mathematical Society",
        17778: "Center for World Mission",
        17779: "Korean Academy of Oral and Maxillofacial Pathology",
        17780: "Kahramanmaras Sutcu Imam University Journal"
               "of Engineering Sciences",
        17781: "The Society of Digital Information and Wireless"
               "Communications (SDIWC)",
        17782: "Karam, Journal of Black Sea Studies",
        17783: "S/N Korean Humanities",
        17784: "Pro Ciencia Periodicos Cientificos",
        17785: "Korean Securities Law Association",
        17786: "Korean Society of Strategic Management",
        17787: "Institute of Social Science, Sogang University",
        17788: "Institute of Korean Studies",
        17789: "Korean Association of Socio-Historical Studies",
        17790: "Center for Korean Studies, Korea University",
        17791: "Institute for Cognitive Science",
        17792: "Research Institute of Korean Culture",
        17793: "Cesuca - Centro Ensino Superior de Cachoeirinha",
        17794: "Faculty of Mining, Geology and Petroleum Engineering",
        17795: "International Neuroscience Institute",
        17796: "The Journal of Clinical Pediatric Dentistry",
        17797: "CSurgeries",
        17798: "Bitlis Eren Universitesi Fen Bilimleri Dergisi",
        17799: "Termeszetfoldrajzi Kozlemenyek Journal",
        17800: "Amazonia, Organizacoes e Sustentabilidade",
        17801: "Revista Brasileira de Sementes",
        17802: "NII KPSSZ",
        17803: "Kutafin Moscow State Law University",
        17804: "Institute of Engineering Science, Ural Branch RAS",
        17805: "Moscow University for the Humanities",
        17806: "Sociedade Brasileira de Urologia",
        17807: "Instituto de Quimica - Univ. Federal do Mato Grosso do Sul",
        17808: "Programa de Pos Graduacao em Direito da PUC-Rio",
        17809: "Global Partnership on Development of Scientific"
               "Cooperation, LLC",
        17810: "Preschool Teacher Training College, Vrsac, Serbia",
        17811: "Universidad de Oviedo",
        17812: "Global Science Publishing Group",
        17813: "Mobilization Journal",
        17814: "SIMP Redakcja Mechanik - Agencja Wydawnicza SIMP",
        17815: "Forschungszentrum Julich, Zentralbibliotek",
        17816: "ECO-Vector",
        17817: "Korean Society for Neurotherapy",
        17818: "University of Dubrovnik",
        17819: "Felsefelogos",
        17820: "Korean Society of Ecology and Infrastructure Engineering",
        17821: "Sarada Ranganathan Endowment for Library Science",
        17822: "Osmanli Mirasi Arastirmalari Dergisi",
        17823: "Gumushane Universitesi Sosyal Bilimler Enstitusu"
               "Elektronik Dergisi",
        17824: "Yerbilimleri",
        17825: "Korea Logistics Research Association",
        17826: "Cukurova Medical Journal",
        17827: "Archives Medical Review Journal",
        17828: "Yalova Journal of Social Sciences",
        17829: "Marmara Iletisim Dergisi",
        17830: "European Agrophysical Institut",
        17831: "Enquiry:  The ARCC Journal of Arch Research",
        17832: "Orthopaedic Section, APTA, Inc.",
        17833: "Iowa Academy of Science",
        17834: "The Indian Society for Surface Science and Technology",
        17835: "Humanities Perspectives Limited",
        17836: "E-CONOM",
        17837: "Szent Istvan Egyetem",
        17838: "Korean Classical Literature Association",
        17839: "The Korean Society for the Gifted and Talented",
        17840: "Institute of Russian Studies",
        17841: "Korea Evangelical Counseling Society",
        17842: "Seokdang Academy",
        17843: "Instituto Nacional de Salud",
        17844: "Indonesian Society Fisheries Product Processing",
        17845: "Faculty of Medicine, Universiti Kebangsaan Malaysia",
        17846: "Constantine the Philosopher University in Nitra",
        17848: "W.E. Upjohn Institute for Employment Research",
        17849: "American Academy of Insurance Medicine",
        17850: "Geological Society of Norway",
        17851: "Faculdade de Letras da UFMG",
        17852: "Revista aSEPHallus de Orientacao Lacaniana",
        17853: "Russian State Vocational Pedagogical University",
        17854: "The Research Center for Korea Youth Culture",
        17855: "Institute of Latin American Studies",
        17856: "The Society of Asian Historical Studies",
        17857: "The Society for Hyo-won Historical Studies",
        17858: "Journal of Military and Information Scicence",
        17859: "Pamukkale University Journal of Divinity Faculty",
        17860: "Mersin University Journal of the Faculty of Education",
        17875: "Universitatsverlag Gottingen",
        17915: "Center for Nanotechnology in Society at UC Santa Barbara",
        17919: "Entrepreneur and Innovation Exchange (EIX.org)",
        17921: "Editora e Distribuidora Educacional",
        17922: "Russian State Social University",
        17923: "Language and Cognitive Science",
        17924: "Research Institute for Life and Culture Sogang University",
        17925: "Touch Medical Media LTD.",
        17926: "Korea Association of Law and Politics",
        17927: "Korean Association of Educational Methodology Studies",
        17928: "Japanese Association for Digital Humanities",
        17929: "The Japanese Society for Quality Control",
        17930: "University of Agronomic Sciences and Veterinary"
               "Medicine of Bucharest",
        17931: "Initia Via Editora Ltda",
        17932: "Istanbul Aydin University",
        17933: "Badan Litbang SDM Kementerian Komunikasi dan Informatika",
        17934: "Army Military History Institute",
        17935: "China Research Center",
        17936: "Pan-Korea English Teachers Association",
        17937: "21st Century Political Science Association",
        17938: "The Learned Society of Korean Drama and Theatre",
        17939: "Honam Institute of Humanities and Social Sciences",
        17940: "Endoscopy Gastrointestinal",
        17941: "Turkish Journal of Academic Gastroenterology",
        17942: "Jurnal of Continuing Medical Education",
        17943: "Education Technology Theory and Practice",
        17944: "The Medical Journal of Mustafa Kemal University",
        17945: "The Korean Society of Blood Transfusion",
        17946: "Korean Society of Radiological Science",
        17947: "Korean Film Association",
        17948: "Research Institute of Korean Studies",
        17949: "The Association of North-East Asian Cultures",
        17950: "Innovative Research Publications",
        17951: "Wydawnictwo Uniwersytetu Marii Curie-Skåodowskiej w Lublinie",
        17952: "Prompt Scientific Publishing",
        17953: "UCLA American Indian Studies Center",
        17954: "Akdeniz Medical Journal",
        17955: "Alexander Graham Bell Association for the"
               "Deaf and Hard of Hearing",
        17956: "MedStudy",
        17957: "Friends Science Publishers",
        17958: "Korean Society of Mechanical Technology",
        17959: "The Phonology-Morphology Circle of Korea",
        17960: "The English Linguistics Society of Korea",
        17961: "Korea Research Academy of Distribution and Management",
        17962: "Korea Association for Psychodrama and Sociodrama",
        17963: "Society for Korean Classical Chinese Education",
        17964: "Korean Association for Political Economy",
        17965: "Korean Society for Environmental Education",
        17966: "Korean Society for Medical Mycology",
        17968: "The Korean Society of Christian Religious Education",
        17969: "Agricultural Faculty",
        17970: "Indonesian Economist Association",
        17971: "Croatian Entomological Society",
        17972: "International Journal of Information, Communication"
               "Technology and Applications",
        17973: "MM Publishing, s.r.o.",
        17974: "Diagrama Editorial",
        17975: "STEM Fellowship",
        17976: "NPP Polis (Political Studies)",
        17977: "State University of Malang (UM)",
        17979: "Universidade da Coruna",
        17980: "Cancer Research Frontiers",
        17981: "Corporacion Universidad de la Costa - CUC",
        17982: "CBCD Colegio Brasileiro de Cirugia Digestiva",
        17983: "Kyoto University",
        17984: "Adiyaman Universitesi Egitim Bilimleri Dergisi",
        17985: "International Journal of Academic Research in Education",
        17986: "The Bulletin of Legal Medicine",
        17987: "International Cardiovascular Forum Journal",
        17988: "Council for Children with Behavioral Disorders",
        17989: "ENCODE Data Coordination Center",
        17990: "Aletheia - Associacao Cientifica e Cultural",
        17992: "Laeknabladid/The Icelandic Medical Journal",
        17993: "Area de Innovacion y Desarrollo, S.L. - 3 Ciencias",
        17994: "Academic and Educational Forum on International Relations",
        17995: "The Korean Society for German History",
        17996: "The Japanese Society of Nuclear Cardiology",
        17997: "The Center for Social Welfare Research",
        17998: "Korean Liver Cancer Association",
        17999: "The Society of History Education",
        18000: "Sathyabama University",
        18001: "Tobacco Regulatory Science Group",
        18002: "University of Leon",
        18003: "National University of Singapore",
        18004: "Instituto de Investigaciones en Ciencias de la Salud",
        18005: "Bowen Publishing Company",
        18006: "Journal of Experimental Biology and Agricultural Sciences",
        18007: "International Academy of Science and Higher Education",
        18008: "PE Polunina Elizareta Gennadievna",
        18009: "Journal of Computer and Education Research",
        18010: "Scientia Prima",
        18011: "Universidade Estadual Paulista - Campus de Tupa",
        18012: "Aufklarung",
        18013: "The Institute for Northeast Asia Research",
        18014: "Kyung Hee University Management Research Institute",
        18015: "The Korean Association for Saramdaum Education",
        18016: "KSU Journal of Agriculture and Nature",
        18017: "Journal of Istanbul Faculty of Medicine",
        18018: "The Institute for Legal Studies",
        18019: "Russian Ilizarov Scientific Centre Restorative"
               "Traumatology and Orthopaedics",
        18020: "Kesit Akademi",
        18021: "The Institute of the History of Christianity in Korea",
        18022: "Association Culturelle Franco-Coreenne",
        18023: "Korean Society for Early Childhood Education",
        18024: "Revista UNIANDRADE",
        18025: "Kafkas University Journal of Economics and"
               "Administrative Sciences",
        18026: "Celal Bayar University Journal of Social Sciences",
        18027: "Russian Society of Clinical Oncology",
        18028: "Revista de Gestao, Financas e Contabilidade",
        18029: "Publicacaoes Avulsas em Conservacao de Ecossistemas",
        18030: "Socio.hu",
        18031: "Ilmin International Relations Institute",
        18032: "Korean Academic Association of Business Administration",
        18033: "International Journal of Language Academy",
        18034: "ABC Journals",
        18035: "El Mednifico Journal",
        18036: "Anadolu Universitesi Bilim ve Teknoloji Dergisi"
               "C : Yasam Bilimleri ve Biyoteknoloji",
        18037: "Anadolu Universitesi Sosyal Bilimler Dergisi",
        18038: "Anadolu Universitesi Bilim ve Teknoloji Dergisi-A:"
               "Uygulamali Bilimler ve Muhendislik",
        18039: "Anadolu Universitesi Egitim Bilimleri Enstitusu Dergisi",
        18040: "Korean Ancient Historical Society",
        18041: "Universidad Libre",
        18042: "Centro de Estudios Politicos y Constitucionales",
        18043: "North Carolina Institute of Medicine",
        18044: "Medinform LTD",
        18045: "University of Rijeka, Faculty of Economics",
        18046: "Universidad Icesi",
        18047: "Faculty of Agrobiotechnical Sciences Osijek",
        18048: "Association for Promotion and Development"
               "of Maritime Industries",
        18049: "Journal of Contemporary Medicine and Dentistry",
        18050: "Universidad Cesar Vallejo",
        18051: "Universa Medicina",
        18052: "SciPress Ltd.",
        18053: "Journal of Clinical and Translational Research",
        18054: "Hrvatski Prirodoslovno Drustvo (Croatian Society"
               "for Natural Sciences)",
        18055: "Centro de Estudos Geograficos (IGOT) Universidade de Lisboa",
        18056: "South Egypt Cancer Institute",
        18057: "Hong Kong Institute of Steel Construction",
        18059: "Mercy College School of Business",
        18060: "IUPUI University Library",
        18061: "The Ohio State University Libraries",
        18062: "Case Western Reserve University",
        18063: "Whioce Publishing Pte Ltd.",
        18064: "Korean Academy of Sensory Integration",
        18065: "Instituto de Treinamento e Pesquisa em Gestalt Terapia",
        18066: "UNIVAP Universidade de Vale do Paraiba",
        18067: "Instituto Federal do Amapa",
        18068: "Associacao Convergir Sustentavel",
        18069: "Firat Universitesi Sosyal Bilimler Dergisi",
        18070: "Erciyes University Journal of Faculty of Economics"
               "and Administrative Sciences",
        18071: "Ideggyogyaszati Szemle Journal",
        18072: "Perspectivas Medicas da Faculdade de Medicine de Jundiai",
        18073: "Pan American Journal of Medical Thermology",
        18074: "Cankiri Karatekin Universitesi Iktisadi ve"
               "Idari Bilimler Fakultesi Dergisi",
        18075: "The Association of Japanology in East Asia",
        18076: "Korea Intelligent Information System Society",
        18077: "Institute of Chinese Studies, Hankuk University"
               "of Foreign Studies",
        18078: "American Studies Institute",
        18080: "Telecommunications Association Inc.",
        18081: "BM-Publisher",
        18082: "Turbomachinery Laboratory",
        18083: "Liquid Crystals and their Application",
        18084: "The Association of Baccalaureate Social Work"
               "Program Directors, Inc.",
        18085: "Journal of Latino/Latin American Studies",
        18086: "International Solar Energy Society (ISES)",
        18087: "APO Society of Specialists in Heart Failure",
        18088: "French Sciences Publishing Group",
        18089: "School of Management, Hospitality and Tourism,"
               "University of the Algarve",
        18090: "SMS Institute of Technology",
        18091: "Babasaheb Bhimrao Ambedkar University",
        18092: "International Journal of Economics and Administrative"
               "Studies",
        18093: "Scientific and Practical Reviewed Journal Pulmonology",
        18094: "Selcuk Iletisim",
        18095: "Modern English Society",
        18096: "Theory, Methodology, Practice",
        18097: "Institute of  Biomedical Chemistry",
        18099: "Forensicwayout",
        18100: "International Journal of Applied Mathematics,"
               "Electronics and Computers",
        18101: "Buryat State University",
        18102: "Society of Pansori",
        18103: "Knowledge Enterprises Journals",
        18104: "Korean Academy of Industrial Commerce",
        18105: "The Society for History of Education",
        18106: "The Society of Korean Photography",
        18107: "Institute of Global Affairs",
        18108: "Korean Society for Engineering Education",
        18109: "Institute for EU Studies",
        18110: "The Korean Linguistic Association",
        18111: "World Tourism Organization (UNWTO)",
        18117: "University of Pittsburgh Library",
        18122: "Boise State University",
        18125: "Forest Ecosystem Monitoring Cooperative",
        18130: "University of Virginia",
        18146: "Netherlands Institute for Sound and Vision",
        18162: "CRIFPE",
        18172: "Universidad de la Rioja",
        18173: "Journal of Science, Hanoi National University of Education",
        18174: "Wageningen University and Research",
        18175: "Universidad de los Andes - Centro de Investigacion"
               "y Formacion en Educacion (CIFE)",
        18176: "Esmon Publicidad, SA",
        18177: "The College Music Society",
        18178: "EJournal Publishing",
        18180: "Universidad ECCI",
        18181: "International Society for Technology in Arthroplasty",
        18182: "Turkish Journal of Forestry",
        18183: "Cadernos Cenpec",
        18184: "Publishing House Science",
        18185: "Erzincan Universitesi Fen Bilimleri Ensitusu Dergisi",
        18186: "Journal of Thermal Engineering",
        18187: "Pakistan Journal of Statistics and Operation"
               "Research (PJSOR)",
        18188: "Revista Scientia Agraria Paranaensis (SAP)",
        18189: "The Legal Studies Institute of Chosun University",
        18190: "Estudos Geologicos",
        18191: "International Journal of Philippine Science and Technology",
        18192: "University of Ottawa Library",
        18193: "SAHkartell",
        18194: "International Wader Study Group",
        18195: "Western Australian Museum",
        18196: "Universitas Muhammadiyah Yogyakarta",
        18197: "Surdurulebilir ve Engelsiz Bilim Egitimi",
        18198: "Institute of Low Temperature Energy Technology",
        18199: "St. Petersburg State University",
        18200: "Journal of Gifted Education and Creativity",
        18201: "International Journal of Intelligent Systems"
               "and Applications in Engineering",
        18202: "Faculty of Economics, University of Brawijaya",
        18203: "Medip Academy",
        18204: "International Society for Simulation Surgery",
        18205: "The Korean Society for Woman Psychology",
        18206: "Korean Association for Political and Diplomatic History",
        18207: "Korean Critical Sociological Association",
        18208: "The Korean Society of Design Culture",
        18209: "The International Association For Korean Language Education",
        18211: "The Korean Society for Human Resource Development",
        18212: "The Society for Chinese Cultural Research",
        18213: "The Classical Chinese Literature Association of Korea",
        18214: "Korean Association of EMG Electrodiagnostic Medicine",
        18215: "Kangwon National University Institute of Comparative"
               "Legal Studies",
        18216: "Confucianism Research Institute",
        18217: "Institute of Hispanic Studies",
        18218: "Institute for Mediterranean Studies",
        18219: "Daedong Institute for Korean Studies",
        18220: "Karadeniz Incelemeleri Dergisi",
        18221: "Beykent University Journal of Social Sciences",
        18222: "Fundacao Carlos Chagas",
        18223: "Revista Historia e Cultura",
        18224: "Pontifica Universidade Catolica de Goias -PUC Goias",
        18225: "IBICT",
        18226: "Universidade Caixias do Sul",
        18227: "Universidade Federal de Roraima",
        18228: "Izmir University Medical Journal",
        18229: "Kocatepe Medical Journal",
        18230: "The Korea Association of Yeolin Education",
        18231: "IP Innovative Publication Pvt Ltd",
        18232: "America Latina en la Historia Economics",
        18233: "Acta Pediatrica de Mexico",
        18234: "Instituto de Investigaciones Dr. Jose Maria Luis Mora",
        18235: "Felipe Herrera Library (Inter-American Development Bank)",
        18236: "E-Conservation Journal - University of Evora",
        18237: "Korean-German Academy of Economics and Management",
        18238: "Institute of Japanese Studies, Hallym University",
        18239: "Universidad Castilla la Mancha",
        18240: "Press of International Journal of Ophthalmology (IJO Press)",
        18241: "M&W Comunicacao Integrada Ltda - ME",
        18242: "Centro de Investigaciones Biologicas del Noroeste (CIBNOR)",
        18243: "International Society for Managing and Technical Editors",
        18244: "Journal of Modern Turkish History",
        18245: "International Journal of Automotive Engineering"
               "and Technologies",
        18247: "Revista EducaOnline",
        18248: "Revista de Educomunicacao Ambiental",
        18249: "Revista Hipertexto",
        18250: "Revista de Ensino de Ciencias e Engenharia",
        18251: "Eastern University",
        18252: "Aurora Publishing LLC (UK) Limited",
        18253: "The Institute of Korean Arts Therapy",
        18254: "LLC Integration Education and Science",
        18255: "P.G. Demidov Yaroslavl State University",
        18256: "Complexo de Ensino Superior Meridional S.A.",
        18257: "ACCEFYN - Academia Colombiana de Ciencias"
               "Exactas, Fisicas y Naturales",
        18258: "Experiment",
        18259: "Universidad Continental",
        18260: "American Society for Engineering Education",
        18261: "Scandinavian University Press / Universitetsforlaget AS",
        18262: "Graduate School of Science, Moscow University"
               "of Finance and Law",
        18263: "St. Mary's College of California",
        18264: "Fundacao CECIERJ",
        18265: "Instituto Federal de Educacao, Ciencia e Tecnologia"
               "da Paraiba",
        18267: "Prague University of Economics and Business",
        18268: "Sociedad Geologica Mexicana",
        18269: "Faculty of Mathematics and Science Education",
        18270: "Universidad el Bosque",
        18271: "Universidad Nacional Del Altiplano Puno Peru",
        18272: "Universidad San Francisco de Quito",
        18273: "Universidad Industrial de Santander",
        18274: "Borrowers and Lenders:  The Journal of Shakespeare"
               "and Appropriation",
        18275: "Lehigh University",
        18276: "University of Szczecin",
        18277: "Carleton College",
        18278: "Policy Studies Organization",
        18279: "Vilnius University Press - MIDAS Project",
        18280: "International Information and Engineering"
               "Technology Association",
        18281: "Research Publisher Inc.",
        18282: "PiscoMed Publishing Pte Ltd",
        18283: "European Association of Scientific Research",
        18284: "Research Institute for Social Science Keimyung University",
        18285: "GEONOMOS",
        18286: "HPE Ulyanovsk SAA Stolypin",
        18287: "Samara National Research University",
        18288: "Economic Policy",
        18289: "One Earth Future Foundation",
        18290: "Towarzystwo Naukowe KUL",
        18291: "Aarhus University Library",
        18292: "AILPcsh-Associacao Internacional de Ciencias"
               "Sociais e Humanas em Lingua Portuguesa",
        18293: "KSI Research, Inc.",
        18294: "Universidad Nacional de Lanus",
        18295: "Sultan Qaboos University Medical Journal",
        18296: "NZCER Press, New Zealand Council for Educational Research",
        18297: "University of Louisville",
        18298: "International Journal of Languages Education and Teaching",
        18299: "Cappadocia Journal of History and Social Sciences",
        18300: "International Journal of Religious Sciences",
        18301: "Researcher Social Science Studies",
        18302: "Journal of Educational Sciences",
        18303: "Trofimuk Institute of Petroleum Geology and"
               "Geophysics (SB RAS)",
        18304: "Revista Scripta Alumni",
        18305: "Revista Scripta Uniandrade",
        18306: "Progress in Geography",
        18307: "Journal of Lake Sciences",
        18308: "Sociedad Rio-Grandeense de Bioetic (SORBI)",
        18309: "ANPOLL",
        18310: "Associacao Brasileira da Rede Unida",
        18311: "Informatics Publishing Limited",
        18312: "UNIVAG Centro Universitario",
        18313: "Pacific Society for Reproductive Health Charitable Trust",
        18314: "Gratis Open Access Publishers LLC",
        18315: "Argumentum",
        18316: "Centro Universitario La Salle - UNILASALLE",
        18317: "Kelam Arastirmalari Dergisi (Journal of Kalam Studies)",
        18318: "Instytut Badan Literackich PAN/The Institute"
               "of Literary Research PAS",
        18319: "Developing Country Think Institute Ltd",
        18320: "VXL Publishers",
        18321: "Institute of Combustion Problems",
        18322: "Pozhnauka Publishing House",
        18323: "Togliatti State University",
        18324: "Bratsk State University",
        18325: "The Society of Korean Medicine Rehabilitation",
        18326: "IAIN Salatiga",
        18327: "Center for International Area Studies, HFUS",
        18328: "PLURA - Revista de Estudos de Religiao",
        18329: "The Energy and Resources Institute",
        18330: "Balai Penelitian dan Pengembangan Lingkungan"
               "Hidup dan Kehutanan Makassar",
        18331: "Greenwave Publishing of Canada",
        18332: "EU European Publishing",
        18333: "Korean Association for Public Administration",
        18334: "BIBLIO-GLOBUS Publishing House",
        18335: "European Regional Science Association",
        18338: "The Korean Society for Moral and Ethics Education",
        18339: "Korean Geomorphological Association",
        18341: "Korean Women's Institute, Ewha Womans University",
        18343: "JIPI, Lembaga Penelitian dan Pengabdian kepada Masyarakat",
        18344: "Institute for International Studies in Education",
        18345: "Turkiyat Mecmuasi",
        18346: "Korean Society for the Scientific Study of Subjectivity",
        18347: "Institute of History and Culture Hankuk University"
               "of Foreign Studies",
        18348: "Opuscula Zoologica",
        18349: "Magyar Nyelv",
        18350: "Institute for Poverty Alleviation and International"
               "Development",
        18351: "Revista Genero & Direito",
        18352: "Uopen Journals",
        18353: "Carcinological Society of Japan",
        18354: "Ege Strategic Research Journal",
        18355: "Slovenska Vzdelavacia a Obstaravacia s.r.o.",
        18356: "United Nations Publications",
        18357: "University of Victoria Libraries",
        18358: "Institute for Analytical Instrumentation of"
               "the Russian Academy of Sciences",
        18359: "Universidad Militar Nueva Granada",
        18360: "IFSP Itapetininga",
        18361: "Revista de Administracao e Negocios da Amazonia",
        18362: "Instituto Internacional de Analise Termica",
        18363: "Associacao Brasileira de Odontologia Rio de Janeiro (ABORJ)",
        18364: "Liceu Literario Portugues",
        18366: "Appris Editora e Livraria Eireli - ME",
        18367: "PHASELIS",
        18368: "Istanbul Journal of Sociological Studies",
        18369: "World Association of Chinese Character Studies",
        18370: "Publishing Office TRILIST",
        18371: "University of Banking of National Bank of Ukraine (Kyiv)",
        18372: "National Aviation University",
        18374: "International Academy of Business and Economics",
        18375: "Osaka University of Comprehensive Children"
               "Education and Osaka Jonan Women's Junior College",
        18376: "Excercise Fitness and Health Alliance",
        18377: "Informacao Arquivistica",
        18378: "Grupo Verde de Agroecologia e Abelhas",
        18379: "Universidade Veiga de Almeida",
        18380: "Columella",
        18381: "EconoQuantum",
        18382: "Jurnal Riset dan Aplikasi: Akuntansi dan Manajemen",
        18383: "Grapho, LLC",
        18384: "Moscow Regional State University",
        18387: "Escuela Nacional de Ciencias Biologicas",
        18388: "Polskie Towarzystwo Biochemiczne (Polish Biochemical"
               "Society)",
        18389: "Universidad de Los Andes - Facultad de Arquitectura y Diseno",
        18391: "Universidade Estadual da Paraiba (UEPB)",
        18392: "University of Debrecen/Metszetek",
        18393: "Eurasian Journal of Soil Science",
        18394: "Is ve Insan Dergisi",
        18395: "Publio Kiado",
        18396: "Korean Theatre Studies Association",
        18397: "The Korean Association for Comparative Government",
        18398: "The Korea Association for Local Government"
               "and Administration Studies",
        18399: "Academia Koreana",
        18400: "Teknik Dergi",
        18401: "Korean Counseling Association",
        18402: "Resources Science",
        18403: "E-Makalat Mezhep Arastirmalari Dergisi",
        18404: "International Journal of Education in Mathematics,"
               "Science and Technology",
        18405: "Revista Evidenciacao Contabil & Financas",
        18406: "IFSULDEMINAS (Instituto Federal do Sul de Minas)",
        18407: "Fundacao Centro Tecnologico de Juiz de Fora (FCT)",
        18408: "Australian Housing and Urban Research Institute (AHURI)",
        18409: "Paul V. Galvin Library/Illinois Institute of Technology",
        18410: "Level Up Business Center",
        18411: "Private Enterprise, Vladislav Ivanov",
        18412: "Kalvis",
        18413: "Belgorod National Research University",
        18414: "Kozgazdasagi Szemle",
        18415: "International Journal of Multicultural and"
               "Multireligious Understanding (IJMMU)",
        18421: "Association for Information Communication"
               "Technology Education and Science (UIKTEN)",
        18422: "Universitatsverlag Gottingen",
        18432: "ARI Journal",
        18433: "JPPS Journal",
        18435: "VAMP Journal",
        18438: "EBLIP Journal",
        18458: "University of Debrecen/Debreceni Egyetem",
        18466: "Celal Bayar University Journal of Science",
        18468: "Universidade Federal do Amapa",
        18469: "Povolzhskiy State University of Telecommunications"
               "and Informatics",
        18470: "Institute of Applied Ecology",
        18471: "Revista Baiana de Enfermagem",
        18472: "Editora de Livros IABS",
        18473: "BioOne (Lepidopterists' Society)",
        18474: "BioOne (Georgia Entomological Society)",
        18475: "BioOne (University of Puerto Rico at Mayaguez)",
        18476: "BioOne (Stuttgart State Museum of Natural History)",
        18477: "BioOne (Paleontological Institute at The University"
               "of Kansas)",
        18478: "IUFS Journal of Biology",
        18479: "Institute of Research and Journals",
        18480: "The Agricultural Economics Society of Japan",
        18481: "TIRAZH Publishing House",
        18482: "Journal of Adolescent Psychology and Psychiatry",
        18483: "Alkhaer Publications",
        18484: "Vitebsk State Medical University",
        18485: "Faculty of Philology, University of Belgrade",
        18486: "Machine Intelligence Research Labs, Gwalior",
        18487: "Fundacja Medigent",
        18488: "Conscientia Beam",
        18489: "South African Institute of Computer Scientists"
               "and Information Technologists",
        18490: "Sosyoloji Arastirmalari Dergisi",
        18491: "Beytulhikme An International Journal of Philosophy",
        18492: "Journal of Linguistics Research",
        18493: "Karamanoglu Mehmetbey Universitesi Sosyal"
               "ve Ekonomik Arastirmalar Dergisi",
        18494: "MYU K.K.",
        18495: "Faculty of Computer Science, Sriwijaya University",
        18496: "The Korea-Japan Historical Society",
        18497: "International Electronic Journal of Environmental Education",
        18498: "Review of the Faculty of Divinity of Amasya University",
        18499: "VSMU N.N. Burdenko",
        18500: "Saratov State University",
        18501: "Korea Heart Rhythm Society",
        18502: "Knowledge E",
        18503: "Nosov Magnitogorsk State Technical University",
        18504: "Facultad Latinoamericana de Ciencias Sociales,"
               "Mexico (FLACSO)",
        18505: "Cumhuriyet Universitesi Ilahiyat Fakultesi Dergisi",
        18506: "Anemon Mus Alparslan Universitesi Sosyal Bilimler Dergisi",
        18508: "LLC Relevant Medicine",
        18509: "Geobalcanica Society",
        18510: "GIAP Journals",
        18511: "Revista Brasileira de Ciencia e Mobimento",
        18512: "Revista Brasileira de Milho e Sorgo",
        18513: "Tarih Incelemeleri Dergisi",
        18514: "Mathematical Notes",
        18515: "Szent Istvan University Faculty of Economics"
               "and Social Sciences",
        18516: "Draw Science LLC",
        18517: "Insight Society",
        18519: "Society for Reproductive Biology and Comparative"
               "Endocrinology",
        18520: "Current Science Association",
        18521: "Konuralp Medical Journal",
        18522: "Southern Federal University",
        18523: "National University of Kyiv - Mohyla Academy",
        18524: "Odesa I.I. Mechnikov National University",
        18525: "The Korea Association of Clinical Ultrasound",
        18527: "Doctrina, Ltd.",
        18528: "Society of Gastrointestinal Intervention",
        18529: "The Korean Association of Psychoanalysis",
        18530: "Biztositas es Kockazat",
        18531: "Studia Mundi - Economica",
        18532: "Contemporary Analysis and Applied Mathematics",
        18533: "LAR Center Press",
        18534: "AR-Consult, LLC",
        18535: "Valley International",
        18536: "Tinkr",
        18537: "Universidad de Cuenca",
        18538: "Zayed University Press",
        18539: "Current Agriculture Science and Technology (CAST)",
        18540: "Universidade Federal de Vicosa",
        18541: "The Special Education Research Institute",
        18542: "Universidade Federal do Para",
        18543: "University of Deusto",
        18544: "PROQURAT Andrzej Prokurat",
        18546: "Institute of Education Press",
        18547: "Kernel Press UG",
        18548: "Asociacion Peruana de Reproduccion Animal",
        18549: "Centro de Investigaciones y Publicaciones"
               "Farmaceuticas (CIPF)",
        18550: "Rhazes, LLC",
        18551: "Russian Journal of Agricultural and Socio-Economic Sciences",
        18552: "Coventry University, Lanchester Library",
        18553: "Academy of Managed Care Pharmacy",
        18554: "Universidade Federal do Triangulo Minero",
        18555: "Korean Institute of Cultural Product Art and Design",
        18556: "Digital Surgery Limited",
        18558: "Fundacion Ignacio Larramendi",
        18559: "Poznan University of Economics",
        18560: "Cukurova Arastirmalari Dergisi",
        18561: "Revista Biota Amazonia",
        18562: "Universidad Metropolitana",
        18563: "Association Palaeovertebrata",
        18564: "Politeknik Negeri Jember",
        18565: "Bionika Media",
        18566: "Universidad Pontificia Bolivariana",
        18567: "Sociedad Espanola de Bioquimica y Biologia Molecular (SEBBM)",
        18568: "Escola Superior de Propaganda e Marketing (ESPM)",
        18569: "Nucleo de Estudos em Saude Publica",
        18570: "Associacao Brasileira de Etica e Odontologia Legal (ABOL)",
        18571: "Universidade Iguacu - Campus V",
        18572: "The Publishing Group Jurist",
        18573: "Cardiff University Press",
        18574: "NYU Press",
        18575: "Medical Society of the Republic of Srpska",
        18576: "Natural Sciences Publishing",
        18577: "FSUE Scientific Research Institute of Aviation Materials",
        18578: "Royal Pharmaceutical Society",
        18579: "Krupanidhi Educational Trust",
        18580: "Sociedade Brasileira de Engenharia de Televisao (SET)",
        18581: "Case Western Reserve University",
        18582: "Case Western Reserve University",
        18583: "Universidad Politecnica de Guanajuato",
        18584: "International Indigenous policy Journal",
        18585: "Secretariat of The Indonesian Biomedical Journal",
        18586: "Mus Alparslan Universitesi Fen Bilimleri Dergisi",
        18587: "Institute for Buddhist Culture",
        18588: "Asian Journal of Peacebuilding",
        18589: "Osmanli Arastirmalari",
        18590: "Marshall University Libraries",
        18591: "BJU International",
        18592: "IAIN Antasari",
        18593: "Universidade do Oeste de Santa Catarina",
        18594: "The Medical Journal of Haydarpasa Numune Training"
               "and Research Hospital",
        18595: "Civilacademy Journal of Social Sciences",
        18596: "Journal of the Turkish Chemical Society, Section"
               "A: Chemistry",
        18597: "Federacion Colombiana de Obstetricia y Ginecologia",
        18598: "Korean Society of Community-Based Occupational"
               "Therapy Association",
        18599: "Georesursy",
        18600: "Sulcus Publishing",
        18601: "Universidad Externado de Colombia",
        18602: "Faroe University Press",
        18603: "Sanat ve Tasarim Dergisi",
        18604: "Tourism Management Research Organization",
        18605: "Revista Cereus",
        18606: "Amazonia: Science and Health",
        18607: "Universidade Federal de Mato Grosso",
        18608: "Society for Learning Analytics Research",
        18609: "BioInsights Publishing, Ltd.",
        18610: "Giftet, Inc.",
        18611: "Comparative Politics",
        18612: "CNU Research Institute of Education",
        18613: "Maritime Faculty Journal - Denizcilik Fakultesi Dergisi",
        18614: "Dokuz Eylul Universitesi Tip Fakultesi Dergisi",
        18615: "Anadolu Ege Tarimsal Arastirmalar Enstitusu Dergisi",
        18616: "Fundacao Educacional de Criciuma- FUCRI",
        18617: "Liinc em Revista",
        18618: "Associacao Brasileira de Eletronica de Potencia SOBRAEP",
        18619: "FSBSI All-Russian Scientfic Research Institute"
               "of Vegetable Breeding and Seed Production",
        18620: "Foundation for the Development of Bone Marrow"
               "Transplantation",
        18621: "The European Research Journal",
        18622: "The Korean History Education Society",
        18623: "Editora Dom Helder",
        18624: "SENAI ISC",
        18625: "The Korean Association of Speech Communication",
        18627: "The Modern Linguistic Society of Korea",
        18628: "Urimalgeulhakhoe",
        18630: "Korean Association of Arabic Language and Literature",
        18631: "The Japanese Language and Literature Society of Korea",
        18632: "Impact Journals, LLC",
        18633: "Biotecnia",
        18634: "Universidad La Gran Colombia",
        18635: "Moscow Institute of Energy-Safety and Energy-Economy",
        18636: "Universidad Tecnologica del Choco Diego Luis Cordoba",
        18637: "Foundation for Open Access Statistic",
        18638: "Publishing Society",
        18639: "HATASO Enterprises LLC",
        18640: "International Journal of Information Security Engineering",
        18641: "Society of Biocontrol Advancement",
        18642: "Scientific Advances Publishers",
        18643: "Gineco, EU",
        18644: "SATA Publishing Corporation",
        18645: "Rishi Educational Society",
        18646: "New Millennium Discoveries",
        18647: "Journal of Jewish Studies",
        18648: "Jurisprudence Publishing House",
        18649: "The Korean English Education Society",
        18650: "Enliven Archive",
        18651: "Federal Reserve Bank of Kansas City",
        18652: "The Korean Society of Fashion Design",
        18653: "Association for Computational Linguistics (ACL)",
        18654: "Chinese Society for Mineralogy, Petrology, and Geochemistry",
        18655: "Pars Szoftverhaz Kft.",
        18656: "Journal of European Education",
        18657: "Yonetim ve Ekonomi",
        18658: "Korea Humanities Content Society",
        18659: "Korean Association of Corpus Linguistics",
        18660: "HunNet-Media Kft.",
        18661: "Revista Pauta Geral – Estudos em Jornalismo",
        18662: "Asociatia LUMEN",
        18663: "Turkish Journal of Clinics and Laboratory",
        18664: "Ukrainian State University of Railway Transport",
        18665: "Ithaka S&R",
        18666: "Sagamore Publishing, LLC",
        18667: "Escuela de Posgrados FAC",
        18668: "The Oil and Gas Institute - National Research Institute",
        18669: "Comhar Teoranta",
        18670: "American Society of Naval Engineers (ASNE)",
        18671: "Instituto de Pesquisa e Estudos Florestais (IPEF)",
        18672: "Revista Brasileira de Estudos Políticos da UFMG",
        18673: "Revista Gestao e Saude/NESPROM",
        18674: "Revista Exacta",
        18675: "Departamento de Educacao da Universidade Estadual"
               "Paulista — UNESP",
        18676: "Centro de Estudos e Pesquisas em Educacao,"
               "Cultura e Acao Comunitaria (CENPEC)",
        18677: "Centro Cientifico Conhecer",
        18678: "Duzce Medical Journal",
        18679: "Translational Neuroscience and Clinics",
        18680: "Hellenic Semiotic Society",
        18681: "Pakistan Society of Nematologists",
        18682: "Fundacion Universidad de Palermo",
        18683: "Asociatia Pentru Cresterea Vizibilitatii Cercetarii"
               "Stiintifice (ACVCS)",
        18684: "Biotecnologia en el sector Agropecuario y Agroindustrial",
        18685: "E3 Journals Ltd",
        18686: "Universe Scientific Publishing Pte. Ltd.",
        18687: "LACCEI (Latin American and Caribbean Consortium"
               "of Engineering Institutions)",
        18688: "Actual Problems of Theory and History of Art",
        18689: "Madridge Publishers, LLC",
        18690: "Universtiy of Maribor",
        18691: "Kultur ve Iletisim - Culture &Communication",
        18692: "LLC Polyforum Group",
        18693: "The Korean Society of Beauty and Art",
        18694: "Korea Philosophical Association",
        18695: "Case Western Reserve University",
        18696: "Reunir Revista de Administracao, Contabilidade"
               "e Sustentabilidade",
        18697: "African Journal of Food, Agriculture, Nutrition"
               "and Development",
        18698: "Bauman Moscow State Technical University",
        18699: "Institute of Cytology and Genetics, SB RAS",
        18700: "Korean Neurocritical Care Society",
        18701: "IMS Institute Noida",
        18702: "Asian Concrete Federation",
        18703: "Seoul International Law Academy",
        18704: "The Japan Language and Literature Association of Korea",
        18705: "Arterialnaya Gipertenziya",
        18706: "Gojoseon and Dangun Studies Association",
        18707: "East-West Art and Culture Studies Association",
        18708: "Korean Association of Christian Studies",
        18732: "HSSA Journal",
        18733: "CPI Journal",
        18737: "Southern Spaces",
        18740: "University of Alberta - Journal of Socialist Studies",
        18741: "University of Alberta - Journal of Professional"
               "Continuing and Online Education",
        18753: "Cantonal and University Library Fribourg",
        18754: "Seismo Verlag",
        18755: "University of Bern",
        18759: "Sociedade de Ensino Superior de Vitoria",
        18760: "INFOCOMP Journal of Computer Science",
        18761: "Associacao Paradigma - Centro de Ciencias"
               "e Tecnologia do Comportamento",
        18762: "Associacao Educacional de Araras",
        18763: "Revista do Tribunal de Contas do Estado de Minas Gerais",
        18764: "Universidade Federal do Maranhao",
        18765: "Journal of Extension Systems",
        18766: "InterEspaco: Revista de Geografia e Interdisciplinaridade",
        18767: "European Journal of Educational Studies",
        18768: "IJAEDU- International E-Journal of Advances in Education",
        18769: "IJASOS - International E-journal of Advances"
               "in Social Sciences",
        18770: "Korea Electric Power Corporation",
        18771: "Uyusmazlik Mahkemesi Dergisi",
        18772: "Wits University Press",
        18773: "NPS MedicineWise",
        18774: "Summa Psicologica UST",
        18775: "Inovatus Usluge d.o.o.",
        18776: "Texas Christian University",
        18777: "IEA Solar Heating and Cooling Programme",
        18778: "Uniwersytet Lodzki (University of Lodz)",
        18779: "Universidad Tecnica Estatal de Quevedo",
        18780: "Library of Technological Educational Institute of Athens",
        18781: "Revista Mexicana de Fitopatologia, Mexican"
               "journal of Phytopathology",
        18782: "Vital Biotech",
        18783: "Oscar M. Lopez for Climate Change Adaptation"
               "and Disaster Risk Management Foundation, Inc.",
        18784: "Office of Religious Research and Development",
        18785: "University of Southern Mississippi",
        18786: "Moscow Regional Research and Clinical Institute (MONIKI)",
        18787: "Korean Rhinologic Society",
        18788: "Revista Lugares de Educacao",
        18789: "The World Society of Taekwondo Culture",
        18790: "Sociedad Espanola de Paleontologia",
        18791: "Neurosurgical Atlas,Inc.",
        18792: "British Association for the Study of Religions (BASR)",
        18793: "Northern Institute",
        18794: "Medical University of Silesia",
        18795: "Mavi Atlas",
        18796: "Ugol Journal Edition, LLC",
        18797: "Canadian Arena of Applied Scientific Research",
        18798: "Peace, Conflict and Development:  An Interdisciplinary"
               "Journal",
        18799: "National Research Tomsk Polytechnic University",
        18800: "Sistema de Bibliotecas PUCP",
        18801: "Journal BiNET",
        18802: "DrJ Publishers, Inc.",
        18803: "Associacao Portuguesa de Sistemas de Informacao",
        18804: "Youngsan Theological Institute",
        18805: "Agricultural Research Communication Center",
        18806: "TESL Canada Federation",
        18807: "The Society of Service Science",
        18808: "Korea Parliamentary Research Institute",
        18809: "Neal A. Maxwell Institute for Religious Scholarship",
        18810: "Linked clinical trials regisdtry",
        18811: "International Society of Environmental Botanists",
        18812: "Societe d'Etudes Franco-Coreennes",
        18813: "Pierson Memorial Bible Institute",
        18814: "International Union of Geological Sciences",
        18815: "Faculdade Antonio Meneguetti",
        18816: "Revista Brasileira de Inovacao Tecnologica em Saude (R-BITS)",
        18817: "Universidade Estadual do Maranhao",
        18818: "Bloomberg LP",
        18819: "International Journal of Agricultural & Veterinary Sciences",
        18820: "AFRICAN SUN MeDIA",
        18821: "Izdatelstvo Meditsina",
        18822: "Yugra State University",
        18823: "The Journal of Asia TEFL",
        18824: "The French Language and Literature Association of Korea",
        18825: "International Review of Economics and Management",
        18826: "International Journal of Sport Education,"
               "Exercise and Training Science",
        18827: "The Journal of Medical Investigations",
        18828: "Revista Brasileira de Informacoes Cientificas",
        18829: "Revista de Pesquisa em Políticas Publicas",
        18830: "Programa de Pos-Graduacao em Arquitetura e"
               "Urbanismo - Univ. de Brasilia",
        18831: "DJ Publications",
        18832: "Research Institute of Brewing and Malting,"
               "Plc. (Vyzkumny Ustav Pivovarsky a Sladarsky)",
        18833: "Council on Undergraduate Research",
        18835: "Revista Juridica Uniandrade",
        18836: "Economic Analysis of Law Review",
        18837: "Revista Direito e Acao",
        18838: "Revista de Direito Internacional Economico e Tributario",
        18839: "Revista de Estudos e Pesquisas Avancadas do Terceiro Setor",
        18840: "Revista do Mestrado em Direito da Universidade"
               "Catolica de Brasilia: Escola de Direito",
        18841: "Institute of Japanese Studies Dankook University",
        18842: "The Korean Language and Culture Education Society",
        18843: "ERM Publications",
        18844: "Birlesik Dunya Yenilik Arastirma ve Yayincilik Merkezi",
        18845: "Instituto Tecnologico de Costa Rica",
        18846: "Centro de Investigaciones Biologicas del Noroeste (CIBNOR)",
        18847: "Grupo de Estudios en Seguridad Internacional (GESI)",
        18848: "Common Ground Research Networks",
        18849: "RCVS Knowledge",
        18850: "The Korean Ethics Education Association",
        18851: "Uluslararasi Turk Egitim Bilimleri Dergisi",
        18852: "Brand Design Association of Korea",
        18853: "The Jungang English Language and Literature"
               "Association of Korea",
        18854: "Korean Political Science Association",
        18855: "The Linguistic Society of Korea",
        18856: "The Association of Popular Narrative",
        18857: "The Korean Society of Physical Therapy",
        18858: "The Korean Society of Surgical Metabolism and Nutrition",
        18859: "The Social Science Research Institute",
        18860: "Maulana Malik Ibrahim State Islamic University",
        18861: "Universidad ORT Uruguay",
        18862: "Norske Dansekunstnere",
        18863: "Current Approaches in Psychiatry",
        18864: "Istanbul University Faculty of Fisheries",
        18865: "Ethnicity and Disease Inc",
        18866: "British Institute at Ankara",
        18867: "Sociedad Mexicana de Ingenieria Sismica A.C.",
        18868: "Saint Michael College of Caraga",
        18869: "CASRP; Center of Advanced Scientific Research"
               "and Publications",
        18870: "Laureate Education, Inc.",
        18871: "University of Guam Press",
        18872: "Dokuz Eylul Universitesi",
        18873: "Korean Society for Lacan and Contemporary Psychoanalysis",
        18874: "Nanzan University",
        18875: "Scholarena, LLC",
        18934: "Wiley",
        19026: "Maxwell Scientific Publication Corp.",
        19027: "Jurnal Akuakultur Indonesia",
        19028: "Department of Mechanical and Biosystem Engineering IPB",
        19029: "Applied mathematics & Computation Center",
        19030: "Clute Institute",
        19031: "Korean Home Economics Education Association",
        19032: "Koreanisch-Deutsche Gesellschaft fuer Sozlaiwssenschaften",
        19033: "The Society of Korean Semantics",
        19034: "Korean Association of Youth Welfare",
        19035: "The Institute for Northeast Asian Law Chonbuk"
               "National University",
        19036: "Academy for 21st Century Christian Society and Culture",
        19037: "Association of Global Studies Education",
        19038: "Space Applications Centre",
        19039: "Studies of the Ottoman Domain",
        19040: "Ecocycles",
        19041: "University of Debrecen/Applied Studies in"
               "Agribusiness and Commerce (APSTRACT)",
        19042: "The Korean Society for Thinking and Communication",
        19043: "Foundation of Nursing Studies",
        19044: "European Scientific Institute, ESI",
        19045: "Bolan Society for Pure and Applied Biology (BSPAB)",
        19046: "International Scientific Publishers",
        19047: "V.V. Dokuchaev Soil Science Institute",
        19048: "Association of Forensic Medical Experts",
        19049: "Daegu University Research Institute of the"
               "Korea Special Education",
        19050: "Modern Korean Literature Association",
        19051: "Korean Association of Sports and Entertainment Law",
        19052: "Universidad de La Salle",
        19053: "Universidad Pedagogica y Tecnologica de Colombia",
        19054: "Frederick Douglass Institute Collaborative",
        19055: "University of Debrecen/ Debreceni Egyetem"
               "- Acta Medicinae et Sociologica",
        19056: "International Journal of Medical and Dental Sciences",
        19057: "Journal of Special Education and Rehabilitation",
        19058: "Institute of Iberoamerican Studies",
        19059: "Mukaddime",
        19060: "Gazi Academic View",
        19061: "Institute of Chemical Research of Catalonia",
        19062: "Henri Coanda Air Force Academy",
        19063: "Instituto Nacional de Defensa de la Competencia"
               "y la Propiedad Intelectual (INDECOPI)",
        19064: "SciELO Agencia Nacional de Investigación y Desarrollo (ANID)",
        19065: "The Society for Asian Philosophy in Korea: SPK",
        19066: "The Korean Society for Cognitive Science",
        19067: "Institute of Governmental Studies",
        19068: "The Korean Society for Teaching English Literature",
        19069: "The Society of Korean Dialectology",
        19070: "SciDoc Publishers LLC",
        19071: "Scienceflora Publishers Pvt. Ltd.",
        19072: "International Journal of Engineering Technologies",
        19073: "Siberian Law Academy",
        19074: "Sibeocenter LLC",
        19075: "LLC Publishing House Internauka",
        19076: "Maden Tetkik ve Arama Dergisi",
        19077: "The Korean Society of Calligraphy",
        19078: "Korean Society for Science of Eastern Art",
        19079: "sdvig press",
        19080: "Juniper Publishers",
        19081: "Center for Environmental Research, Institute"
               "of Research and Community Services (IPB)",
        19082: "Mehr Publishing Group",
        19083: "Universidad Peruana de Ciencias Aplicadas",
        19084: "Sociedade de Ciencias Agrarias de Portugal",
        19085: "SCHOLEDGE Publishing",
        19086: "Alliance of Diamond Open Access Journals",
        19087: "Jurnal Veteriner",
        19088: "Institute of Development Studies",
        19089: "College of Humanities and Social Sciences, UAEU",
        19090: "Faculty of Philosophy, University of Novi Sad",
        19091: "Revista de Ciencias da Educacao",
        19092: "Instituto Rede de Pesquisa Empirica em Direito (REED)",
        19093: "Revista de Estudos Sociais",
        19094: "Revista Contextus Universidade Federal do Ceara",
        19095: "Revista e-ciencia",
        19096: "Review of International Law and Politics",
        19097: "The Korean Association of Space and Environment Research",
        19098: "Juniper Press",
        19099: "Vanguard University",
        19100: "Journal of General-Procedural Dermatology & Venereology",
        19101: "Association of Computer, Communication and"
               "Education for National Triumph Social and Welfare"
               "Society (ACCENTS)",
        19102: "MediaSphere Medical LLC",
        19103: "Burleigh Dodds Science Publishing Limited",
        19104: "Elyns Group LLC",
        19105: "Institut Agama Islam Negeri Madura",
        19106: "Journal of the Medical Sciences (Berkala Ilmu Kedokteran)",
        19107: "Romanian Association for Information Security"
               "Assurance (RAISA)",
        19108: "Koers Bulletin for Christian Scholarship",
        19109: "State Islamic University of Raden Fatah Palembang",
        19110: "IG Komi SC UB RAS",
        19111: "Bulletin of the Mineral Research and Exploration",
        19112: "RUSTEM (Russian Society for Trace Elements in Medicine)",
        19113: "SDU Journal of Natural and Applied Sciences",
        19114: "ACTS Theology Institute",
        19115: "The International Association of Comparative Korean Studies",
        19116: "The Criticism and Theory Society of Korea",
        19117: "Creativity Science Research Institute",
        19118: "The Korean Society for the Study of Educational Principles",
        19119: "Cineforum",
        19120: "The Pukyung Institute for Korean History",
        19121: "Wildlife Conservation Society",
        19122: "GEDIIB",
        19123: "Coordenacao de Publicacoes da Editora do Instituto"
               "Federal de Brasilia",
        19124: "Institute of Mathematics and Its Applications",
        19125: "Association for Research of MPS and Rare Diseases",
        19126: "Sakarya University Journal of Education",
        19127: "Middle Black Sea Journal of Health Science",
        19128: "Turkish Journal of Education",
        19129: "Gaziosmanpasa Universitesi Sosyal Bilimler"
               "Arastirmalari Dergisi",
        19130: "Instituto de Investigaciones Filologicas",
        19131: "Portuguese Journal of Mental Health Nursing",
        19132: "Faculdade de Biblioteconomia Comunicacao",
        19133: "Sociedade Latino-Americana de Estudos Imobiliarios",
        19134: "Eutomia - Revista de Literatura e Linguistica",
        19135: "CONSINTER",
        19136: "Universidad Juarez Autonoma de Tabasco",
        19137: "Universidad Nacional de La Pampa",
        19138: "Academic Journal, Al-Azhar University",
        19139: "International Academic Press",
        19140: "JMTR Publications, LLC",
        19141: "Instituto Adventista de Ensino",
        19142: "Revistas Processos Quimicos",
        19143: "Pontifica Universidade Catolica de Sao Paulo",
        19144: "Andrew John Publishing, Inc. (OMICS)",
        19145: "Gumushane University E-Journal of Faculty of Communication",
        19146: "PIBIC Unicamp",
        19147: "The World Bank",
        19148: "International Journal of Human and Behavioral Science",
        19149: "Water Resources and Irrigation Management",
        19150: "Society for Mining, Metallurgy & Exploration Inc.",
        19151: "Illinois Law Enforcement Training and Standards"
               "Board Executive Institute",
        19152: "Appellate Press B.V",
        19153: "Centro Latino Americano de Estudios en Informatica",
        19154: "Roskilde University Library",
        19155: "Geofisica Internacional",
        19156: "HPS - Health Publishing & Services SRL",
        19157: "National Association of Therapeutic Schools and Programs",
        19158: "The Korean Society for Women Psychology",
        19159: "Turkish Journal of Agricultural Research (TUTAD)",
        19160: "International Journal of Educational Researchers",
        19161: "Ege Journal of Medicine",
        19162: "The Association for the Historical Studies"
               "on Korean National Movement",
        19163: "Volgograd State Medical University",
        19164: "Northumbria University Library",
        19165: "The International Centre for Counter-Terrorism (ICCT)",
        19166: "Universitas Pelita Harapan",
        19167: "Weddell Publishing",
        19168: "Journal of Yasar University",
        19169: "The City History Compilation Committee of Busan",
        19170: "East European and Balkan Institute",
        19171: "Uludag Universitesi Egitim Fakultesi Dergisi",
        19172: "Korea Real Estate Analysts Association",
        19173: "Athabasca University Press",
        19174: "Revista Esferas Interprogramas de Pos-graduacao"
               "em Comunicacao do Centro-Oeste",
        19175: "RECOM (Revista de Enfermagem do Centro Oeste Mineiro)",
        19176: "Revista de Cultura Teologica",
        19177: "Universidade do Sul de Santa Catarina - UNISUL",
        19178: "Scientia Amazonia",
        19179: "Fundacao Municipal de Artes de Montenegro (FUNDARTE)",
        19180: "Essentia Editora",
        19181: "Federal Center of Theoretical and Applied"
               "Sociology of the Russian Academy of Sciences (FCTAS RAS)",
        19182: "CIRAD (Centre de Cooperation Internationale"
               "en Recherche Agronomique Pour le Developpement",
        19183: "Asociacion Colombiana de Profesores de Ingles - ASOCOPI",
        19184: "UPT Penerbitan Universitas Jember",
        19185: "Sciencematters",
        19195: "Wydawnictwo Uniwersytetu Wroclawskiego",
        19198: "Peperosso",
        19218: "LIBRUM Publishers and Editors LLC",
        19219: "AnEx Information Verlag",
        19245: "PuntoOrg",
        19249: "AIDAA Associazione Italiana di Aeronautica e Astronautica",
        19256: "Edra SpA",
        19261: "Institute of Chemistry of Academy of Science of Moldova",
        19263: "CREA-DC",
        19271: "Fundacja Rehasport Clinic",
        19277: "ALIES",
        20286: "Nova Explore Publications Inc.",
        20287: "Universidade da Beira Interior",
        20288: "Institute of Chinese Study Pusan National University",
        20289: "The Journal of Ege University Faculty of Agriculture",
        20290: "Anadolu University Journal of Science and"
               "Technology — B Theoretical Sciences",
        20291: "Ural State University of Railway Transport (USURT)",
        20292: "Central Institute of Cultural Heritage",
        20293: "Korean Philosophical Society",
        20294: "International Academy of Global Business and Trade",
        20295: "Petersburg State Transport University",
        20296: "Turkiye Sosyal Arastirmalar Dergisi",
        20297: "The Korean Academy of Scientific Criminal Investigation",
        20298: "Neva Electromagnetics, LLC",
        20299: "NAPICU (National Association of Psychiatric"
               "Intensive Care Units)",
        20302: "Croatian Natural History Museum",
        20303: "Journal and Publishing Unit Faculty of Law UGM",
        20304: "Humanitas Uluslararasi Sosyal Bilimler Dergisi",
        20305: "Korean Society of Interpretation and Translation Studies",
        20306: "Korean Comparative Education Society",
        20307: "The Korean Society of Pharmacognosy (KAMJE)",
        20308: "European Journal of Physics Education",
        20309: "Journal of Data and Information Science",
        20310: "Tambov State University - G.R. Derzhavin",
        20311: "Statisztikai Szemle",
        20312: "Dimenziok: matematikai kozlemenyek",
        20313: "Asociacion Paleontologica Argentina",
        20314: "Australian Literary Studies",
        20315: "Slovenian Forestry Institute",
        20316: "European Association of Science Editors",
        20318: "Universidad Carlos III de Madrid",
        20319: "Global Research & Development Services",
        20320: "Facultad de Ciencias de la Salud, UDES",
        20321: "Nile University of Nigeria",
        20322: "Littera Turca Journal of Turkish Language and Literature",
        20323: "Yaroslavl State Pedagogical University",
        20324: "Botanical Survey of India",
        20325: "The Korean Association of Coaching",
        20326: "Korea Evangelical Missiological Society",
        20328: "Vietnam Education Foundation Fellows and Scholars"
               "Association",
        20330: "Anthropologiai Kozlemenyek",
        20331: "Allattani Kozlemenyek",
        20332: "Termeszetvedelmi Kozlemenyek",
        20333: "Krasnoyarsk State Medical University",
        20334: "Vilnius Gediminas Technical University",
        20335: "Global World Academy Journal",
        20336: "Sociedade Brasileira de Sociologia",
        20337: "Programa de Pos-Graduacao em Ciencias da Linguagem"
               "- PPGCL/UNIVAS",
        20338: "Brazilian Journal for Motor Behavior",
        20339: "INOITs ALMAVEST Ltd.",
        20340: "Reaviz Medical University",
        20341: "Geologicva Belgica",
        20342: "Engineering Faculty of Hasanuddin University",
        20343: "International Society for the Scholarship"
               "of Teaching and Learning",
        20344: "Ordem dos Medicos",
        20353: "UC Observatories",
        20354: "Berkeley Linguistics Society",
        20355: "University of Alberta - Journal of Contemporary"
               "Issues in Education",
        20360: "University of Alberta- Language and Literacy",
        20361: "University of Alberta - The Deakin Review"
               "of Childrens Literature",
        20377: "Universitatsbibliothek Bamberg",
        20378: "Universitatsbibliothek Bamberg",
        20395: "Instituto Nacional de Educacao de Surdos",
        20396: "Universidade Estadual de Campinas",
        20397: "Fundacao Pedro Leopoldo - Revista Gestao and Tecnologia",
        20398: "Journal of Surgical and Clinical Research",
        20399: "Agora FIlosofica",
        20400: "Revista de Teologia e Ciencias da Religiao da UNICAP",
        20401: "Programa de Pos Graduacao em Administracao da UFF/VR",
        20402: "Korea Institute for Skin and Clinical Sciences",
        20403: "Siberian Research Institute of Geology, Geophysics"
               "and Mineral Resources",
        20404: "The Center for Japanese Studies",
        20405: "The Association for Korean Linguistics",
        20406: "Psychological Association of Culture and Social Issues",
        20407: "Fujita Medical Society",
        20408: "The Korean Society of Traumatology",
        20409: "Business and Economics Research Journal",
        20410: "Academy of Labor and Social Relations",
        20411: "Case Western Reserve University",
        20413: "International Study of Religion in Eastern"
               "and Central Europe Association",
        20414: "State Islamic University (UIN) Mataram",
        20415: "Electric",
        20416: "Societe de Philosophie des Sciences",
        20417: "New Zealand Ecological Society",
        20418: "SoftQ Sp. z o.o.",
        20419: "Drustvo psihologov Slovenije/Slovenian Psychologists'"
               "Association",
        20420: "Universidad de Las Palmas de Gran Canaria",
        20421: "Weil Cornell Medicine-Qatar",
        20422: "BPPKI Bandung - Kementerian Komunikasi Dan Informatika RI",
        20423: "Revista Internacional de Folkcomunicacao",
        20424: "Brazilian Journal of International Relations",
        20425: "International Society of Tea Science",
        20426: "Paralellus Revista de Estudos Religioso",
        20427: "Hacettepe University Journal of Turkish Studies",
        20428: "University of Science and Technology, Yemen",
        20429: "Georgia Southern University",
        20430: "Fondo de Cultura Economica",
        20431: "ARC Publications Pvt Ltd.",
        20432: "Editora Napoleao Ltda ME",
        20433: "The New Korean Philosophical Association",
        20434: "KNU Korean Institute of Management Consulting",
        20435: "Universidade Catolica Dom Bosco",
        20436: "Korean Ordo Economics Association",
        20437: "Korean Open Association for Early Childhood Education",
        20438: "Biblioteca do Centro de Educacao e Saude (CES)",
        20439: "Akademik Islam Arastirmalari Dergisi",
        20446: "Mattersburger Kreis",
        20447: "Shanxi Medical Periodical Press",
        20448: "Asian Educational Journal Publishing Group",
        20449: "Electrical Engineering Department, Andalas University",
        20450: "Society of Chemists and Technologists of Macedonia",
        20452: "Towarzystwo Internistow Polskich/Polish Society"
               "of Internal Medicine",
        20453: "Universidad Peruana Cayetano Heredia",
        20454: "Modern Science Publishers",
        20455: "American Innovative Medical Sciences and Technologies"
               "(AIMSCI) Inc.",
        20456: "Koreanische Gesellschaft fur Didaktik der"
               "Deutschen Sprache und Literatur",
        20457: "Soosun Historical Association",
        20458: "Din Bilimleri Dergisi",
        20459: "Chukyo University",
        20460: "Journal of Global Strategic Management",
        20461: "The Association of Korea Literary Theory and Criticism",
        20462: "Global e-Business Association",
        20463: "Korea Society for Exercise Nutrition",
        20464: "Korean Development Economics Association",
        20465: "Korea Internet of Things",
        20466: "Korean Society of Pressure Vessels and Piping",
        20467: "Journal for Human Caring",
        20468: "Arunai Publications Private Limited",
        20469: "KKG Publications",
        20470: "Czech Society for Systems Integration",
        20471: "Sestre Milosrdnice University Hospital Center"
               "(KBC Sestre milosrdnice)",
        20472: "International Institute of Social and Economic Sciences",
        20473: "Universitas Airlangga",
        20474: "TAF Publishing",
        20475: "Dentmedpub Research and Printing",
        20476: "Jurnal Ilmu Administrasi dan Organisasi: Bisnis & Birokrasi",
        20477: "Facultad de Agronomia/Instituto Nacional de"
               "Investigacion Agropecuaria",
        20478: "The Korean Society of Design Culture",
        20479: "Journal of Agricultural Faculty of Uludag University",
        20480: "LIBRI Kitap Tanitimi, Elestiri ve Ceviri Dergisi",
        20481: "Korea Society of Coastal Disaster Prevention",
        20482: "International Convergence Management Association",
        20483: "The Society of Korean Fiction",
        20484: "The Korean Association for Local Government Studies",
        20485: "Society of Automotive Engineers of Japan, Inc.",
        20486: "Islam Medeniyet Arastirmalari Dergisi (IMAD)",
        20487: "Korea Association of Secondary English Education",
        20488: "Anadolu University Journal of Art and Design",
        20489: "International Journal of Early Childhood Special"
               "Education (INT-JECSE)",
        20490: "Research Opinions in Animal and Veterinary Sciences",
        20491: "Journal of Business Research - Turk",
        20492: "Medical Journal of Ankara Training and Research Hospital",
        20493: "Birey ve Toplum Sosyal Bilimler Dergisi",
        20494: "University of Debrecen/Taplalkozasmarketing",
        20495: "Center for Southeast Asian Studies, Kyoto University",
        20496: "The Association of Korea Counseling Psychology"
               "Education Welfare",
        20497: "Korean Academy Welfare Counseling",
        20498: "East Asia Business Economics Association",
        20499: "Biblioteca da Presidencia da Republica",
        20500: "Revista Contemporanea de Educacao",
        20501: "Instituto Perspectivas e Novos Desafios de"
               "Humanizacao do Direito Civil-Constitucional",
        20502: "Revista Brasileira de Geomorfologia",
        20503: "Revista Carreiras e Pessoas (RECAPE)",
        20504: "OPUS",
        20505: "Contracampo UFF",
        20506: "O.I.E (World Organisation for Animal Health)",
        20507: "Nga Pae o te Maramatanga",
        20508: "International Journal of Renewable Energy Research",
        20509: "TEMPO",
        20510: "UK Journal of Pharmaceutical and Biosciences",
        20511: "Universidad San Ignacio de Loyola S.A.",
        20512: "Korean Association for Adult and Continuing Education",
        20513: "Revista de Medicina da UFC",
        20514: "Synapse, LLC",
        20515: "Osmangazi Journal of Medicine",
        20516: "The Research of the Korean Classic",
        20517: "OAE Publishing Inc.",
        20518: "Turkish Journal of Public Health",
        20519: "Divan: Disiplinlerarasi Calismalar Dergisi",
        20520: "JEL-KEP",
        20522: "Kyung Hee University Management Research Institute",
        20523: "Revista Eletronica Sapere Aude",
        20524: "Hellenic Society of Gastroenterology",
        20525: "Bussecon International",
        20526: "Physics Society of the Philippines (Samahang"
               "Pisika ng Pilipinas)",
        20527: "Center for Journal Management and Publication,"
               "Lambung Mangkurat University",
        20528: "Tulpar Academic Publishing",
        20529: "Forum for Medical Ethics Society",
        20530: "Whites Science Innovation Limited",
        20531: "The Forest Herbarium",
        20532: "Faculty of Electrical Engineering and Computing,"
               "Univ. of Zagreb",
        20533: "Infonomics Society",
        20534: "East West Association for Advanced Studies"
               "and Higher Education GmbH",
        20535: "Kyiv Politechnic Institute",
        20536: "Moscow Islamic Institute",
        20537: "Izhevsk Institute of Computer Science",
        20538: "Siberian State Medical University",
        20539: "Deadong Philosophical Association",
        20540: "International Academy of Physical Therapy Research",
        20541: "Center for Bioethics and Research, Nigeria",
        20542: "Primakov Institute of World Economy and International"
               "Relations",
        20543: "Balai Besar Kulit, Karet dan Plastik (Center"
               "for Leather, Rubber and Plastics)",
        20544: "University St. Kliment Ohridski - Bitola",
        20545: "PWD Group Publishing, Inc.",
        20546: "Excellent Publishers",
        20547: "Geist Science, Iqra University",
        20548: "Research Group Innoeduca, University of Malaga",
        20622: "Japan Language Testing Association",
        20659: "Japan Society of Forest Planning",
        20849: "July Press Pte. Ltd.",
        20850: "Glasstree",
        20851: "University of Adelaide Press",
        20852: "New Trends in Mathematical Science",
        20853: "Stellenbosch University - South African Journal"
               "of Higher Education",
        20854: "Beykent University Journal of Science and Engineering",
        20855: "International Institute of Acoustics and Vibration (IIAV)",
        20856: "NICEC, Ltd.",
        20857: "Research Institute of the Finnish Economy (ETLA)",
        20858: "Silesian University of Technology",
        20859: "Ar-Raniry, International Journal of Islamic Studies",
        20860: "Uluslararasi Sosyal ve Egitim Bilimleri Dergisi",
        20861: "OLIMP",
        20862: "Luchevaya Diagnostika",
        20863: "Natural Science and Discovery",
        20864: "The Studies of Korean Literature",
        20865: "Hankuk University of Foreign Studies, Language"
               "Research Institute",
        20867: "University of Rijeka, Faculty of Tourism and"
               "Hospitality Management",
        20868: "Universidad Politecnica de Madrid - University Library",
        20869: "Chamber of Financial Auditors of Romania",
        20870: "Universite de Bordeaux",
        20871: "Sekolah Tinggi Filsafat Islam (STFI) Sadra",
        20872: "Revista de Empreendedorismo Gestao de MPE",
        20873: "Universidade Federal do Tocantins",
        20874: "Tyumen Scientific Center of the SB RAS",
        20875: "Mehmet Akif Ersoy Universitesi Sosyal Bilimler"
               "Enstitusu Dergisi",
        20876: "International Journal of Biology Education",
        20878: "Culinary Society of Korea",
        20879: "Korean Society for Journalism and Communication Studies",
        20880: "Association of Korean Language Education Research",
        20881: "Institute of Korean Literature Studies",
        20882: "Socidrogalcohol",
        20883: "Poznan University of Medical Sciences",
        20884: "Universitas Jenderal Soedirman",
        20885: "Universitas Islam Indonesia (Islamic University"
               "of Indonesia)",
        20886: "Research, Development and Innovation Agency,"
               "Ministry of Environment and Forestry",
        20887: "Revista Datavenia - UEPB",
        20888: "RIDPHE-R - Revista Iberoamericana do Patrimonio"
               "Historico-Educativo",
        20889: "Instituto Brasileiro de Relacoes Internacionais",
        20890: "Revista Reflexus",
        20891: "Clio Arqueologica",
        20892: "Cancer Biology and Medicine",
        20894: "Integrated Intelligent Research (IIR)",
        20895: "LPPM Institut Teknologi Telkom Purwokerto",
        20896: "ACCB Publishing",
        20897: "Lectito BV",
        20898: "International Association for Shell and Spatial Structures",
        20899: "Journal of Public and Nonprofit Affairs",
        20900: "Hapres",
        20901: "Faculty of Political Science, University of Zagreb",
        20902: "Sai Scientific Communications",
        20903: "Macedonian Academy of Sciences and Arts",
        20904: "Institute of Theoretical and Applied Informatics,"
               "Polish Academy of Sciences",
        20906: "SWGE Sistemas",
        20907: "The Society of Korean Literary Therapy",
        20908: "IJARS Group",
        20909: "Korea Society of Packaging Science and Technology",
        20910: "The Society for Aerospace System Engineering",
        20911: "Faculdade Jesuita de Filosofia e Teologia — FAJE",
        20912: "Revista Direitos Culturais",
        20913: "State Public Scientific Technological Library SB RAS",
        20914: "FSBEI HE Voronezh State University of Engineering"
               "Technologies",
        20915: "Ural Research Institute of Metrology (UNIIM)",
        20916: "Russian Cognitive Linguistics Association",
        20919: "University of Sussex",
        20932: "Asociacion Castellano-Manchega de Sociologia (ACMS)",
        20933: "University of Dundee",
        20934: "Gandhara Publishers, LTD.",
        20935: "Academia.edu",
        20936: "Lawarence Press",
        20937: "Centro de Ciencias de la Atmosfera",
        20938: "BirdLife Australia, Ltd.",
        20939: "SOLAR, Revista de Filosofia Iberoamericana",
        20940: "Electoral Institute for Sustainable Democracy"
               "in Africa (EISA)",
        20941: "Synchro Publisher",
        20943: "SoftwareFirst, Ltd.",
        20944: "MDPI AG",
        20945: "Archives of Endocrinology and Metabolism",
        20946: "Revista Administracao em Dialogo",
        20947: "Associacao Brasileira de Estudos Populacionais",
        20948: "Keldysh Institute of Applied Mathematics",
        20949: "Revista Historia Hoje",
        20950: "Boletim do Instituto de Pesca",
        20951: "Faculdade Redentor",
        20952: "Revista Tempos e Espacos em Educacao",
        20953: "Dynasty Publishing House",
        20954: "Korean Association of Practical Arts Education",
        20955: "Federal Reserve Bank of St. Louis",
        20956: "Hasanuddin University, Faculty of Law",
        20957: "Strategic Studies and Agriculture Policy",
        20958: "Island Dynamics",
        20959: "World Journal of Pharmacy and Pharmaceutical Sciences",
        20960: "ARAN Ediciones",
        20961: "Universitas Sebelas Maret",
        20963: "The Chemical Society of Japan",
        20964: "ESG",
        20965: "Fuji Technology Press Ltd.",
        20966: "Polish Society of Child Neurologists",
        20967: "Science and Knowledge Research Society",
        20968: "New Delhi Institute of Management",
        20969: "Contemporary Clinical Medicine",
        20970: "Korean Academy of Social Welfare",
        20971: "Korean Council of Physical Multiple and Health Disabilities",
        20972: "Elementary Education Research Institute",
        20973: "Korea Institute for Future Politics",
        20974: "Heidegger Gesellschaft in Korea",
        20975: "The Korean Association for the Western Ancient"
               "History and Culture",
        20976: "Korean National Research Center for the Arts",
        20977: "Korean Society of Education Through Art",
        20978: "International Journal of Legal Progress",
        20979: "International Journal of Economics and Innovation",
        20980: "Acta Oncologica Turcica",
        20981: "Kaygi",
        20982: "The Quantitative Methods for Psychology",
        20983: "Noesis - Revista de Ciencias Sociales y Humanidades",
        20985: "Laikos Servicos Ltda",
        20986: "Inspira Network Group",
        20987: "Kenya Meteorological Society",
        20988: "Institute of Language and Information Studies",
        20989: "International Journal of Business, Economics,"
               "and Management Perspectives",
        20990: "Journal of Academic Researches and Studies",
        20991: "All Azimuth Journal of Foreign Policy and Peace",
        20992: "KeiMyung University Institute for Women's Studies",
        20993: "The Korean Society of School Social Work",
        20994: "Gwalior Management Academy",
        20995: "The Institute for Culture, Media and Entertainment Laws",
        20996: "Stolichnaya Izdatelskaya Kompaniyaizdat",
        20997: "The Korean Association for Survey Research",
        20998: "National Technical University Kharkiv Polytechnic Institute",
        20999: "UNAM (Norteamerica, Revista Academica del CISAN-UNAM)",
        21000: "American Society of Reclamation Sciences (ASRS)",
        21001: "Universitat de Lleida",
        21002: "Fakultas Ekonomi dan Bisnis Universitas Indonesia",
        21003: "Institute of Society Transformation",
        21004: "Yunnan Provincial Research Institute for Ethnic Arts",
        21005: "West Pomeranian University of Technology, Szczecin",
        21007: "University of Tennessee Health Science Center Library",
        21008: "Poznan University of Technology",
        21009: "Universitas Negeri Jakarta",
        21010: "African Traditional Herbal Medicine Supporters"
               "Initiative (ATHMSI)",
        21011: "Portuguese Association of Nutritionists",
        21012: "International Association for Fracture Mechanics"
               "of Concrete and Concrete Structures",
        21013: "Institute of Research Advances",
        21014: "IMEKO International Measurement Confederation",
        21015: "VFAST",
        21016: "Sons and Daughters Publishing House",
        21017: "Corporacion Universitaria Republicana",
        21018: "College of Communication and Public Relations",
        21019: "American Pharmacists Association",
        21020: "Hacettepe University Faculty of Health Sciences Journal",
        21021: "Osmanli Medeniyeti ve Arastirmalari Dergisi",
        21022: "Council on Tall Building and Urban Habitat Korea",
        21023: "Korea Money and Finance Association",
        21024: "Pusan National University Educational Development Institute",
        21026: "Korean Association for Local Government Studies",
        21027: "The Society of Korean Historical Manuscripts",
        21028: "Ilustre Colegio de Geologos (ICOG)",
        21029: "Agrar- es Kornyezetjog (Journal of Agricultural"
               "and Environmental Law)",
        21030: "Anyanyelv-pedagogia",
        21031: "Egitimde ve Psikolojide Olcme ve Degerlendirme Dergisi",
        21032: "The Korean Society of Health Informatics and Statistics",
        21033: "Federal Reserve Bank of Chicago",
        21034: "Federal Reserve Bank of Minneapolis",
        21035: "ARC Publishing",
        21036: "Latest Thinking, GmbH",
        21037: "AME Publishing Company",
        21038: "Kent State University",
        21039: "Winchester University Press",
        21040: "Endocrine Oncology and Metabolism",
        21041: "Alconpat Internacional",
        21042: "UP4 Institute of Sciences, S.L.",
        21043: "State Islamic College of Kudus",
        21044: "Research Foundation",
        21045: "Federal Research Institute for Health Organization"
               "and Informatics",
        21046: "Space Research Institute RAS",
        21047: "International Academy of Refrigeration (IAR)",
        21048: "Avinashilingam Institute for Home Science"
               "and Higher Education for Women",
        21049: "Center for Cross-Cultural Studies",
        21050: "The Society of Korean Christian Social Ethics",
        21051: "Korea University Peace and Democracy Institute",
        21052: "Korean Corporation Management Association",
        21053: "Korean Society of Otorhinolaryngology-Head and Neck Surgery",
        21054: "Dokuz Eylul University Review of the Faculty of Divinity",
        21055: "Russian Research Anti-Plague Institute Microbe",
        21056: "Revista de Direito Administrativo and Constitucional",
        21057: "Revista de Estudos e Pesquisas Sobre as Americas",
        21058: "Gyancity Research Consultancy",
        21059: "Buletin Peternakan",
        21060: "Communications in Inorganic Synthesis",
        21061: "Virginia Tech Libraries",
        21062: "Jan Evangelista Purkyne University in Usti nad Labem",
        21063: "ITP Press",
        21064: "Woman in Russian Society",
        21065: "Intellectual Consortium of Drug Discovery"
               "and Technology Development Incorporation",
        21066: "Croatian Association of Researchers in Children's Literature",
        21067: "University of Kanjuruhan Malang",
        21068: "Instituto de Investigacion de Recursos Biologicos"
               "Alexander von Humboldt (IAVH)",
        21069: "Acta Pediatrica Portuguesa",
        21070: "Universitas Muhammadiyah Sidoarjo",
        21071: "Cordoba University Press",
        21072: "IMBR RAS",
        21073: "Korea International Accounting Association",
        21074: "Korea Society for Consumer Advertising Psychology",
        21075: "The Korean Association for Children with Special Needs",
        21076: "Suleyman Demirel University Visionary Journal",
        21077: "Central Marine Fisheries Research Institute, Kochi",
        21078: "Journal of Systems Science and Information (JSSI)",
        21079: "U.S. Army Engineer Research and Development Center",
        21080: "American College of Cardiology",
        21081: "Alpha Chi National Honor Society",
        21082: "Indonesian Agency For Agricultural Research"
               "and Development (IAARD)",
        21083: "University of Guelph",
        21084: "The Association of Modern British and American"
               "Language and Literature",
        21085: "The Journal of Europe-Middle East Social Sciences Studies",
        21086: "The Korean Society of Living Environmental System",
        21087: "The New Association of English Language and Literature",
        21088: "Red Flower Publication Private, Ltd.",
        21089: "NIBD Journal of Health Sciences",
        21090: "International Journal of Advance Engineering"
               "and Research Development (IJAERD)",
        21091: "Science and Medicine, Inc.",
        21092: "HK Institute of Biologicals Standardization Ltd",
        21093: "IAIN Samarinda",
        21094: "Romanian Association of Geomorphologists",
        21095: "Adarsh Group of Institutions",
        21096: "Disegno",
        21097: "The Korea Society for Wellness",
        21098: "Bank Indonesia, Central Banking Research Department",
        21099: "Institute of Mathematics, University of Tsukuba",
        21100: "Educational Development Unit, University of Greenwich",
        21101: "National Institute of Public Health",
        21102: "Zant World Press",
        21103: "International Medical Research and Development Corporation",
        21104: "Institute of Ethnology, Czech Academy of Sciences",
        21105: "The Open Journal",
        21106: "Global Health and Education Projects, Inc. (GHEP)",
        21107: "University of Trunojoyo Madura",
        21108: "School of Computing, Telkom University",
        21109: "Kesmas: Jurnal Kesehatan Masyarakat Nasional",
        21110: "Instituto de Psicoterapia Relacional (IPR)",
        21111: "Universitas Darussalam Gontor",
        21112: "R.W. Steen Library, SFASU",
        21113: "ILIRIA International Review (IIR)",
        21114: "Asociacion Espanola de Economia del Trabajo",
        21115: "Jornal Brasileiro de Economia da Saude",
        21116: "Intellectus Revista Academica Digital",
        21117: "Revista Brasileira de Odontologia Legal",
        21118: "Administracao Publica e Gestao Social",
        21119: "Rede Brasileira Direito e Literatura (RDL)",
        21120: "University of Debrecen/Acta Geographica Debrecina"
               "Landscape and Environment",
        21121: "Ege Akademik Bakis (Ege Academic Review)",
        21122: "Belarusian National Technical University",
        21123: "College of Science for Women",
        21124: "The Asian Council of Science Editors",
        21125: "IATED Academy",
        21127: "Yaoyi International Publishing",
        21128: "Institute for Law and Public Policy",
        21129: "Korean Society of Peripheral Nervous System",
        21130: "Methodist Theological University",
        21131: "Korean Production And Operations Management Society",
        21132: "The Association of Theory and Practice of Private Law",
        21133: "Conversas Interdisciplinares",
        21134: "Universidad Miguel Hernandez de Elche",
        21135: "New York Botanical Garden",
        21136: "Institute of Mathematics, Czech Academy of Sciences",
        21137: "LPPM STKIP PGRI Pacitan",
        21138: "Asociacion Espanola de Geografia",
        21139: "Australian Water Association",
        21140: "Marine Corps University Press",
        21141: "Philippine Journal of Pathology",
        21142: "Universidad Cientifica del Sur",
        21143: "Jurnal Hukum dan Pembangunan",
        21144: "Federal Reserve Bank of Richmond",
        21145: "Remedium Privolzhie",
        21146: "Institute of Philosophy, Russian Academy of Sciences",
        21147: "Chinese Journal of Cancer Research",
        21148: "Korea Society of Child Welfare",
        21149: "Instituto Nacional de Salud Publica",
        21151: "E.T.S. de Ingenieria Agronomica",
        21152: "Multiphysics",
        21153: "Deakin University",
        21154: "STAIN Ponorogo",
        21156: "Faculty of Asian and Middle Eastern Studies",
        21157: "Faculty of Veterinary Medicine, Syiah Kuala"
               "University (IVMA)",
        21158: "Universidad EAN",
        21159: "The Japan Foundation, Sydney",
        21161: "Malaysian Journal of Microbiology",
        21162: "Pakistan Journal of Agricultural Sciences",
        21163: "Asociatia Geograhia Technica",
        21164: "Pomorski Uniwersytet Medyczny w Szczecinie",
        21165: "Grupo de Estudos Linguisticos do Estado de Sao Paulo",
        21166: "Instituto Federal Catarinense",
        21167: "C.Q.D.- Revista Eletronica Paulista de Matematica",
        21168: "Associacao Brasileira de Recursos Hidricos - ABRH",
        21169: "Temas em Saude",
        21170: "Revista Geonorte",
        21171: "Revista Gestao e Sociedade",
        21172: "S N Education Society",
        21173: "Kresge Library, Oakland University",
        21174: "Lazarus Scientific Publishing",
        21175: "RAD Centre",
        21176: "Orissa Journal of Otolaryngology and Head and Neck Surgery",
        21177: "FSBEIHE North Caucasian Institute of Mining"
               "and Metallurgy (STU)",
        21178: "FBU SPbFRI",
        21179: "Kemerovo Institute of Food Science and Technology",
        21180: "Kastamonu Universitesi Iktisadi ve Idari Bilimler"
               "Fakultesi Dergisi",
        21181: "Korean Association of Public Safety and Criminal Justice",
        21182: "Korean Minimally Invasive Spine Surgery Research Society",
        21183: "The Korea Society for Children's Media",
        21184: "Korea Entertainment Industry Association",
        21185: "The Institute of Humanities Research In Konkuk University",
        21186: "Industrial Promotion Institute",
        21187: "Korea Music Therapy Education Association",
        21188: "Korean Commercial Law Association",
        21189: "Journal of the Korean Urban Geographical Society",
        21191: "Korean Strategic Marketing Association",
        21192: "The Society for Chinese Language and Literature",
        21193: "Korean Social and Personality Psychological Association",
        21194: "Korean Journal of Gerontological Social Welfare",
        21195: "Design Institute, Inje University",
        21196: "Asia-Pacific Research Center",
        21197: "The Society for Cognitive Enhancement and Intervention",
        21199: "Western Field Ornithologists",
        21201: "Oxfam GB",
        21202: "Tatar Educational Centre Taglimat",
        21203: "Research Square",
        21204: "Revista Latino-americana de Jornalismo - ANCORA",
        21205: "Deu Muhendislik Fakultesi Fen ve Muhendislik",
        21206: "Revista Brasileira de Agropecuaria Sustentavel",
        21207: "Faculdade de Direita de Franca",
        21208: "Institute of Korean Literature and Arts",
        21209: "Transbaikal State University",
        21210: "Institute of Hispanic Studies",
        21211: "The Institute of the Humanities",
        21212: "The International Association of Area Studies (IAAS)",
        21213: "Korean Journal of Child Education and Care",
        21214: "Korean Journal of Early Childhood Special Education",
        21215: "The Korean Academy of Family Medicine",
        21216: "Hekima University College - African Peace Studies",
        21217: "Hekima University College - Hekima Review",
        21218: "Korea Photovoltaic Society",
        21219: "The Korea Database Society",
        21225: "Canadian Journal of University Continuing Education",
        21226: "East West Journal of Ukrainian Studies",
        21230: "University of Alberta - REEFMM",
        21236: "Defense Technical Information Center (DTIC)",
        21237: "Cliodynamics",
        21240: "Sektion Medienpadagogik der Deutschen Gesellschaft"
               "fur Erziehungswissenschaft - DGfE",
        21248: "University Library J. C. Senckenberg",
        21257: "Daniel Huerlimann",
        21260: "University of Bern",
        21266: "Cifra Ltd - Russian Agency for Digital Standardization"
               "(RADS)",
        21270: "Archives of Health Investigation",
        21271: "Salahaddin University - Erbil",
        21272: "Sumy State University",
        21273: "American Society for Horticultural Science",
        21274: "IAIN Tulungagung",
        21275: "International Journal of Science and Research",
        21276: "Marwah Infotech",
        21277: "Siauliai University",
        21278: "Faculty of Mechanical Engineering and Naval"
               "Architecture, Univ. of Zagreb",
        21279: "Mircea cel Batran Naval Academy",
        21281: "Associacao Portuguesa de Terapeutas da Fala"
               "(Portuguese Association of Speech Therapists)",
        21282: "Andrew John Publishing, Inc. (OMICS)",
        21283: "E-JournALL",
        21284: "Revista Elo - Dialogos em Extensao",
        21285: "Irkutsk National Research Technical University",
        21286: "The Korean Association for Public Society",
        21287: "Korea Women's Studies Institute",
        21288: "Rehabilitation Engineering And Assistive Technology"
               "Society of Korea",
        21289: "The Korean Society of Industrial Convergence",
        21290: "Korean Society of Community Based Occupational"
               "Therapy Association",
        21291: "Korean Association of Language Studies",
        21292: "New Terra",
        21293: "Tomsk State University of Control Systems"
               "and Radioelectronics (TUSUR)",
        21294: "Tomsk Cancer Research Institute",
        21295: "Belgorod University of Cooperation, Economics and Law",
        21296: "The Linguistic Science Society",
        21297: "The British and American Language and Literature"
               "Association of Korea",
        21298: "Korea Tourism Research Association",
        21299: "Korean Association for Visual Culture",
        21300: "Academy of Inventors",
        21301: "University of Belgrade - Faculty of Philosophy"
               "- Department of Ethnology and Anthropology",
        21303: "OU Scientific Route",
        21304: "IAP Intensive Care Chapter India",
        21305: "Oxford Pharamgenesis",
        21306: "Journal of International Dental Scienes (BEUN)",
        21307: "Exeley, Inc.",
        21308: "Asociacion Espanola de Ciencia Politica y"
               "de al Administracion",
        21310: "Connexions . International Professional Communication"
               "Journal",
        21311: "Facultad De Ingenieria Universidad Del Zulia",
        21312: "Riverside Architectural Press",
        21313: "University of Hawaii Press",
        21314: "Infopro Digital",
        21315: "Penerbit Universiti Sains Malaysia",
        21316: "The Journal of International Management Research",
        21317: "The Korean Society of Dance",
        21318: "The Society of Korean Folklore",
        21319: "K.R. Sapkal College of Management Studies",
        21320: "Practical Medicine Publishing House",
        21321: "Korean Association of Family Relations",
        21322: "Turquoise Sosyal Bilimler Dergisi",
        21323: "The Gorbatov's All-Russian Meat Research Institute",
        21324: "Dogal Afetler ve Cevre Dergisi",
        21325: "Journal of Tourism and Gastronomy Studies",
        21326: "Korea Society of Design Trend",
        21328: "SBS Foundation",
        21329: "Korean Human Resource Development Strategy Institute",
        21330: "Korean Music Therapy Association",
        21331: "The Korean Advertising and PR Practitioners Society",
        21333: "Korean Local Government Law Association",
        21406: "Acta Biologica Plantarum Agriensis (ABPA)",
        21417: "Adaptive Biotechnologies",
        21418: "UC Berkeley Geotechnical Engineering ResearchUC"
               "Berkeley Geotechnical",
        21425: "Frontiers of Biogeography",
        21426: "Biogeographia",
        21428: "PubPub",
        21432: "University of Alberta - Canadian Journal of"
               "Learning and Technology",
        21433: "International Conference on GIScience",
        21434: "Revista de Governanca Corporativa",
        21435: "Suomalaisen Kirjallisuuden Seura",
        21436: "Instituut voor Natuur- en Bosonderzoek (INBO)",
        21437: "International Speech Communication Association",
        21438: "Revista Brasileira de Gestao Ambiental e Sustentabilidade",
        21439: "IFCE",
        21440: "Ural State Mining University",
        21441: "Sosyal Guvence",
        21442: "Dongguk University Institute of Japan Studies",
        21443: "FSEI HPE Murmansk State Technical University",
        21445: "Journal of Hispanic Studies",
        21446: "Programa de Pos-graduacao em Ciencias Contabeis da UFRJ",
        21447: "The Korean Association of Urban Policies",
        21448: "International Journal of Secondary Metabolite",
        21449: "International Journal of Assessment Tools in Education",
        21450: "RAHIS - Revista de Administracao Hospitalar"
               "e Inovacao em Saude",
        21451: "Colegio Brasileiro de Reproducao Animal - CBRA",
        21452: "ABEC Publicacoes",
        21453: "The Association International Academy of Communicology",
        21454: "Romanian Society of Anaesthesia and Intensive Care",
        21455: "The Schmidt Institute of Physics of the Earth"
               "of the Russian Academy of Sciences (IPE RAS)",
        21456: "Diponegoro University",
        21457: "Korean Association for Religious Studies",
        21458: "Institute for Urban Humanities",
        21459: "The Korean Council for Children's Rights",
        21460: "Duta Wacana Christian University",
        21461: "Journal of Archaeology and Text",
        21462: "Yayasan Visi Intan Permata (Centrall)",
        21463: "Shima Publishing",
        21464: "Croatian Philosophical Society (Hrvatsko filozofsko drustvo)",
        21465: "Naklada Slap",
        21466: "Association of Anaethetists of Great Britain and Ireland",
        21467: "AIJR Publisher",
        21468: "Stichting SciPost",
        21469: "Computing Centre of the Russian Academy of Sciences",
        21470: "Sociedade Brasileira de Cirurgia Cardiovascular",
        21471: "American Portuguese Studies Association",
        21472: "Brazilian Journal of Biological Sciences",
        21473: "Notion Wave",
        21474: "International Journal Of Advanced Research",
        21475: "Southern Cross Publishing",
        21476: "Performance Philosophy",
        21477: "Sierra Journals",
        21478: "Korean Family Studies Association",
        21479: "Korean Association of Family Therapy",
        21480: "Korea Risk Management Society",
        21481: "The Japanese Culture Association of Korea",
        21482: "Korean Society for Buddhist Studies",
        21483: "Association for International Tea Culture",
        21484: "Korean Review of Organizational Studies",
        21485: "Center for Southeast Asian Studies, HUFS",
        21486: "Recreation",
        21487: "The Institute for Basic Social Science",
        21488: "Journal of Caucasian Studies",
        21489: "Korean Association of Health anad Medical Sociology",
        21490: "The Society for Studies of Korean History",
        21491: "Mnemosine Revista",
        21492: "Inonu Universitesi Hukuk Fakultesi Dergisi",
        21493: "Korean Society for Computer Game",
        21494: "ISTE OpenScience",
        21495: "Ustav Teoreticke a Aplikovane Mechaniky AC CR",
        21496: "Technical university of Kosice, Faculty of"
               "Mechanical Engineering",
        21497: "Selcuk Universitesi Edebiyat Fakultesi Dergisi",
        21498: "Ukrainian Institute for Plant Variety Examination",
        21499: "Moscow Region University of Technology (UNITECH)",
        21500: "Universidad de San Buenaventura",
        21501: "Universidad Catolica Luis Amigo",
        21502: "Libyan International Medical University",
        21503: "Universidad Alas Peruanas S.A.",
        21504: "Rhodes University",
        21505: "Australian Association for the Education of"
               "the Gifted and Talented, Ltd.",
        21506: "Ponte Academic Journal",
        21507: "Letra1 Servicos Editoriais",
        21508: "The National Academy of Pediatric Science and Innovation",
        21509: "Korea Youth Research Association",
        21510: "M. Akmullah BSPU",
        21511: "LLC CPC Business Perspectives",
        21512: "Universitas Bina Nusantara",
        21513: "FSBI IGKE",
        21514: "SPRI of Radiation Hygiene Prof. PV Ramzaev",
        21515: "Kuban State Agrarian University",
        21516: "Real Time, Ltd.",
        21517: "NRC Kurchatov Institute",
        21518: "Remedium, Ltd.",
        21519: "National Research Center Kurchatov Institute",
        21520: "International University of Struga",
        21521: "Medycyna Weterynaryjna - Redakcja",
        21522: "Texila International Marketing Management",
        21523: "Gatha Cognition",
        21524: "Acta Universitatis Gothoburgensis",
        21525: "Nordic Academic Press",
        21526: "Associacao Brasileira de Pavimentacao",
        21527: "Editora Unijui",
        21528: "Associacao Brasileira de Inteligencia Computacional"
               "- ABRICOM",
        21529: "IBEPES (Instituto Brasileiro de Estudos e Pesquisas Sociais)",
        21530: "Associacao Brasileira de Relacoes Internacionais - ABRI",
        21531: "Buletin Veteriner Udayana",
        21532: "ACFE Indonesia Chapter",
        21533: "International University of Sarajevo",
        21534: "Indonesian Aquaculture Society",
        21535: "UNSYS Digital",
        21536: "Deagu Gyeongbuk Society of Physical Education",
        21537: "Sport Science Institute of Keimyung University",
        21538: "Krasovskii Institute of Mathematics and Mechanics UB RAS",
        21539: "Korean Society of Dance Science",
        21540: "Korean Association of Lusophone Area Studies",
        21541: "Academic Platform Journal of Engineering and Science",
        21542: "Global Cardiology Science and Practice",
        21543: "Hungarian Demographic Research Institute",
        21544: "Escola de Guerra Naval",
        21545: "The Research Society of Language and Literature",
        21546: "Korean Language and Literature",
        21547: "Gaziantep University Journal of Social Sciences",
        21548: "Stellenbosch University - South African Journal"
               "of Enology and Viticulture",
        21549: "Nevelestudomany",
        21550: "Uludag Universitesi Fen Edebiyat Fakultesi"
               "Sosyal Bilimler Dergisi",
        21551: "Journal of History and Future",
        21552: "Lexxion Verlag",
        21553: "The Radio and Electronics Association of Vietnam (REV)",
        21554: "Institute for Human Rehabilitation",
        21555: "Universidad Panamericana",
        21556: "Edutec",
        21557: "East View Information Services, Inc.",
        21558: "Fullstopp GmbH",
        21559: "The Association of English Language and Literature in Korea",
        21560: "Sosyal Politika Calismalari Dergisi",
        21561: "The Korean Retina Society",
        21562: "Korean Sociological Association",
        21563: "Selcuk University Journal of Studies in Turcology",
        21564: "Yaroslav Mudryi National Law University",
        21565: "Ankara Universitesi Egitim Bilimleri Fakultesi"
               "Ozel Egitim Dergisi",
        21566: "Journal of Field Crops Central Research Institute",
        21567: "School of Management Sciences, Lucknow",
        21568: "The Choson Dynasty History Association",
        21569: "ANO NCR",
        21570: "International Association for Vegetation Science (IAVS)",
        21571: "International Center for Promotion of Enterprises",
        21572: "Signum Revista da ABREM",
        21573: "Revista Brasileira de Politica e Administracao"
               "da Educacao - RBPAE",
        21574: "Remipe - Revista da Micro e Pequenas Empresas"
               "e Empreendedorismo",
        21575: "Revista Mundi",
        21576: "Pensar Academico",
        21577: "Sociedade Brasileira de Quimica (SBQ)",
        21578: "UTFPR - Universidade Tecnologica Federal do"
               "Parana - Campus Guarapuava",
        21579: "IBGE",
        21580: "UIN Walisongo Semarang",
        21581: "Institute of Tourism Research, Hanyang University",
        21582: "Research Institute for Humanities",
        21583: "Revista Brasileria de Estudos Organizacionais - RBEO",
        21584: "Revista Geografia e Pesquisa",
        21585: "ICT in Practice",
        21586: "London Centre for Social Studies",
        21587: "nstitute of Indian Studies",
        21588: "Institute for Social Development and Policy Research",
        21589: "Legal Research Institute of Ajou University",
        21590: "MRI Books",
        21591: "The Korean Society for Western Medieval History",
        21592: "European Constitutional Law Association",
        21593: "Korean Journal of Head and Neck Oncology",
        21594: "Korea IT Marketing Association",
        21595: "JVE International Ltd.",
        21596: "Journal Academy",
        21597: "Journal of the Institute of Science and Technology",
        21598: "Korea Proprioceptive Neuromuscular Facilitation Association",
        21599: "Alternatives: Turkish Journal of International Relations",
        21600: "International Journal of Kurdish Studies",
        21601: "Veritas Publications LTD",
        21602: "Art-E Sanat Dergisi",
        21603: "Kemerovo State University",
        21604: "Revista Moringa - Artes do Espetaculo",
        21605: "Cukurova Universitesi Muhendislik-Mimarlik Fakultesi Dergisi",
        21606: "Design Research Society",
        21607: "Global Business Association",
        21608: "Egypts Presidential Specialized Council for"
               "Education and Scientific Research",
        21609: "Faculty of Computer Science, Universitas Indonesia",
        21610: "Editorum",
        21611: "QIRT Council",
        21612: "Yaratici Drama Dergisi",
        21613: "Mebas Medikal Basin Yayin",
        21614: "Celsius Publishing House",
        21615: "Universidad CES",
        21616: "Revotech Press (Research Evolution Technology Press)",
        21617: "Journal of Preventive and Rehabilitative Medicine",
        21618: "Faculty of Philology, University of Banja Luka",
        21619: "International Archives of Obstetrics and Gynecology",
        21620: "International Journal of Food and Allied Sciences",
        21621: "Office of Research, Innovation and Commercialization,"
               "Iqra University",
        21622: "Arab Academy for Science, Technology, and"
               "Maritime Transport (AASTMT)",
        21623: "Literacy in Composition Studies",
        21624: "Fontete Forskning",
        21625: "International Experts for Research Enrichment"
               "and Knowledge Exchange (IEREK)",
        21626: "LLC MedTestInfo",
        21627: "Stanford University Press",
        21628: "Revista FLAMMAE",
        21629: "Journal of Systems Engineering and Electronics",
        21630: "Sociedad de Ciencias Aranzadi",
        21631: "Sociedade Portuguesa de Psicologia",
        21632: "Ac  Prasetiya Mulya Publishing",
        21633: "Augusta University",
        21634: "Scinzer Scientific Publications",
        21635: "American Academy of Psychotherapists",
        21636: "Marshfield Clinic Research Institute",
        21637: "Gazdasag es tarsadalom",
        21638: "Saint Petersburg State University",
        21639: "Irkutsk Institute of the All-Russian State"
               "University of Justice",
        21640: "Universidad de LaSalle Bajio",
        21641: "Greater Poland Cancer Centre",
        21642: "Center for Global Trade Analysis",
        21643: "Igdir Universitesi Iktisadi ve Idari Bilimler Faku",
        21644: "Turkish World Journal of Educational Research",
        21645: "Intermedia International e-journal",
        21646: "Dusunce Platformu",
        21647: "Grenze Scientific Society",
        21648: "Gokhale Institute of Politics and Economics",
        21649: "Annals of King Edward Medical University",
        21650: "Korean Society of Systemic Theology",
        21651: "Association D'Etudes De La Culture Francaise"
               "Et Des Arts En France",
        21652: "Korean Association of Southeast Asian Studies",
        21653: "Turkish Journal of Physiotherapy and Rehabilitation",
        21654: "Dongnam Journal of Korean Language and Literature",
        21655: "International Journal of Software and Informatics (IJSI)",
        21656: "Applied Mathematics and Mechanics",
        21657: "Soil Water Journal",
        21658: "AABRM",
        21659: "Aesthetics Media Services",
        21660: "International Journal of Geomate",
        21661: "TSNS Interaktiv Plus, LLC",
        21662: "Mavlutov Institute of Mechanics",
        21663: "Association of Environmental and Engineering Geologists",
        21664: "Fronteiras: Journal of Social, Technological"
               "and Environmental Science",
        21665: "Revista Ambivalencias",
        21666: "Mugla Sitki Kocman University Journal of Education",
        21667: "Ryazan State Radio Engineering University RSREU",
        21668: "Federal Scientific Center for Medical and"
               "Preventive Health Risk Management Technologies",
        21669: "Revista TOMO",
        21670: "Instituto de Investigaciones Sociales",
        21671: "Revista Direito UFMS",
        21672: "FSBEI HE Astrakhan State University",
        21673: "Anadolu Klinigi Tip Bilimleri Dergisi",
        21674: "Revista Eletronica Cientifica da UERGS",
        21675: "Conselho Federal de Enfermagem - Cofen",
        21676: "Universidad del Magdalena",
        21677: "World Maritime University",
        21678: "Universidad del Pacifico",
        21679: "Universidad Maria Auxiliadora SC",
        21680: "Universidade Federal do Rio Grande do Norte - UFRN",
        21681: "A.L. Makarenko GI",
        21682: "OOO Grafika",
        21683: "Journal Dependability",
        21684: "Tyumen State University",
        21685: "Penza State University",
        21686: "Plekhanov Russian University of Economics (PRUE)",
        21687: "IBMP RAS",
        21688: "Institute of Circulation Pathology",
        21689: "Doctormedia",
        21690: "American Geographical Society",
        21691: "Global Science Journal Publishing",
        21692: "Human Anatomy and Physiology Society",
        21693: "Pulmonary Hypertension Association",
        21694: "American Research Journals",
        21695: "NOVA.ID.FCT",
        21696: "El Colegio de San Luis",
        21697: "Uniwersytet Kardynala Stefana Wyszynskiego",
        21698: "National Research and Development Institute"
               "for Industrial Ecology IND-ECOIND",
        21699: "El-Med-Pub Publisher",
        21700: "International Journal of Computing and Information Sciences",
        21701: "Instituto Geologico y Minero de Espana",
        21702: "Credo, LTD.",
        21703: "Universidad Catolica de la Santisima Concepcion",
        21704: "Universidad Nacional Agraria la Molina",
        21705: "Molecular and Cellular Biomedical Sciences (MCBS) Office",
        21706: "Klett-Cotta Journals",
        21707: "Gaia Scientia",
        21708: "Editora da Universidade Federal Rural do Semi-Arido"
               "- EdUFERSA",
        21709: "CASA: Cadernos de Semiotica Aplicada",
        21710: "Revista Cientifica Hermes",
        21711: "Sociedade Brasileira de Matematica",
        21712: "Latin American Journal of Energy Research",
        21713: "CAPES",
        21714: "ANPAD",
        21715: "Geochimica Brasiliensis",
        21716: "Institute of Language Research and Education",
        21717: "Institute for Legal Studies, Yonsei University",
        21718: "The Association of Eastern-Asia Buddhism and Culture",
        21719: "Korea Academic Society of Tourism Management",
        21720: "The Korea Comparative Literature Association",
        21721: "P2P & INOVACAO",
        21722: "Revista Brasileira de Pesquisa em Saude",
        21723: "Revista Ibero-Americana de Estudos em Educacao",
        21724: "REVER",
        21725: "Universidade Federal do Tocantins - UFT",
        21726: "Fundacao Educacional da Regiao de Joinville - Univille",
        21727: "Universidade Severino Sombra",
        21728: "Logeion Filosofia da Informacao",
        21729: "Korean Society of Disaster and Security",
        21730: "Instituto de Estudios Avanzados, Universidad"
               "de Santiago de Chile",
        21731: "The Society of Theology and Thought",
        21732: "Korean Publishing Science Society",
        21733: "Journal of International Scientific Researches",
        21734: "Turk Dunyasi Arastirmalari Dergisi",
        21735: "Turk Dunaysi Tarih Kultur Dergisi",
        21736: "International Journal of Commerce and Finance",
        21737: "Korean Academic Society of Accounting",
        21738: "Institute of Anthropology",
        21739: "Korean Academy of International Business Management",
        21740: "Korean Association of Southeast Asian Studies",
        21741: "Materials Research Forum, LLC",
        21742: "Global Vision Press",
        21743: "National Centre of Excellence in Analytical Chemistry",
        21744: "Suryasa and Sons",
        21745: "Revista Acta Científica",
        21746: "Adhya Biosciences Pvt. Ltd.",
        21747: "Universidade do Porto, Faculdade de Letras",
        21748: "American Oil Chemists Society",
        21749: "Trauma and Orthopaedic Surgery Journal",
        21750: "Reforesta",
        21751: "Croatian Medical Association Croatian Dermatovenereological"
               "Society",
        21752: "PPI-UKM Indonesian Student Association",
        21753: "Veterinaria Mexico OA",
        21754: "Universidad Nacional de Ingenieria",
        21755: "Magyar Traumatologia Ortopedia Kezsebeszet"
               "Plasztikai Sebeszet",
        21756: "Indian Lichenological Society",
        21757: "Agrotropica",
        21758: "Korean Society for Indian Studies",
        21759: "Chung-Ang Law Association",
        21760: "Charaka Publishers",
        21761: "SRMS Institute of Medical Sciences",
        21762: "FapUNIFESP (SciELO)",
        21763: "Turkish Journal of Family Medicine and Primary Care",
        21764: "Mehmet Akif Ersoy Universitesi Egitim Fakultesi Dergisi",
        21765: "Pediatric Practice and Research Journal",
        21766: "Servicio Sismologico Nacional",
        21767: "Scitechnol Biosoft Pvt. Ltd (OMICS)",
        21768: "Ejournal of Public Affairs",
        21769: "Bio-Protocol",
        21770: "Iore International",
        21771: "Center of Industrial Pollution Prevention Technology",
        21772: "Centro de Investigacion en Comportamiento"
               "Organizacional, Cincel SAS",
        21773: "Bogazici University",
        21774: "Universidad Santiago de Cali",
        21775: "Caister Academic Press",
        21776: "Brawijaya University",
        21777: "Moscow Witte University",
        21778: "CRI Electronics",
        21779: "Dagestan State University",
        21780: "JSC Vniizht",
        21781: "FSBSI Mental Health Research Cenetre",
        21782: "Academic Publishing House Geo, Ltd.",
        21783: "Revista Estudos Institucionais",
        21784: "PUZ Wloclawek",
        21785: "International Community for Auditory Display",
        21786: "Society for Science and Nature",
        21787: "Research and Development Agency, Ministry of Home Affairs",
        21788: "Portucalense University",
        21789: "Universidad de Bogota Jorge Tadeo Lozano",
        21790: "The Korean Balance Society",
        21791: "University of Debrecen/Muszaki es Menedzsment"
               "Tudomanyi Kozlemenyeknek",
        21792: "The Society of Japanese Language and Literature,"
               "Japanology Korea",
        21793: "Korean Language and Literature",
        21794: "Daedong Hanmun Association",
        21795: "The Korean Criminal Law Association",
        21796: "Science Education Research Institute",
        21797: "Korean Society For Measurement And Evaluation"
               "In Physical Education And Sports Science",
        21798: "Kadin ve Demokrasi Dernegi",
        21799: "Federal Reserve Bank of Philadelphia",
        21800: "Ciencia e Cultura",
        21801: "Principles and Practice of Clinical Research",
        21802: "Ivano-Frankivsk National Medical University",
        21803: "Corporacion Universitaria Americana",
        21804: "Stellenbosch University - African Journal of Nephrology",
        21805: "BioOne - The Bulletin of Zoological Nomenclature",
        21806: "Mission and Society Research Institute",
        21807: "The Korean Association of Northeast Asia Studies",
        21808: "The Korea Association of Japanese Education",
        21809: "Rilem Publications SARL",
        21810: "Simon Fraser University Library",
        21811: "Association of Hispanists of Korea",
        21812: "Korean Association of Sport Pedagogy",
        21813: "Coordinating Research Council, Inc. (CRC)",
        21814: "University of Minho",
        21815: "American Dental Education Association",
        21816: "Poly-Hybrid Forensic Consult",
        21817: "ENGG Journals Publications",
        21818: "Journal of Behavior and Applied Management",
        21819: "Simbiotics, LLC",
        21820: "Science Impact, Ltd.",
        21821: "Admiral Makarov State University of Maritime"
               "and Inland Shipping",
        21822: "FSB Educational Establishment of Higher Education"
               "Daghestan State Technical University",
        21823: "Vreden Russian Research Institute of Traumatology"
               "and Orthopedics",
        21824: "Instituto Centroamericano de Estudios Fiscales",
        21825: "Ghent University",
        21826: "Editora Letra1",
        21827: "University of Groningen Press",
        21828: "TI Journals",
        21829: "Instituto de Ecologia, A.C.",
        21830: "Escuela Militar de Cadetes Jose Maria Cordova",
        21831: "Universitas Negeri Yogyakarta",
        21832: "Channel View Publications, Ltd.",
        21833: "International Journal of Advanced and Applied Sciences",
        21834: "e-IPH Ltd.",
        21835: "Plekhanov University of Economics",
        21836: "Hippiatrika GmbH",
        21837: "Malaysian Institute of Planners",
        21838: "Iowa State University",
        21839: "Phoenix Research Publishers",
        21840: "Sociedad Iberoamericana de Informacion Cientifica (SIIC)",
        21841: "RCC Institute of Information Technology",
        21842: "PES University",
        21843: "Kalyani Government Engineering College",
        21844: "School of Management Sciences, Varanas",
        21845: "Competitio",
        21846: "Testneveles, Sport, Tudomany",
        21847: "Ukrainian Center for Cultural Studies",
        21848: "Korean Academy of Audiology",
        21849: "Korean Academy of Speech-Language Pathologies",
        21850: "Korea Grammar Education Circle",
        21851: "The Oral Biology Research Institute at Chosum University",
        21852: "Francis de Sales Scientific Society (Towarzystwo"
               "Naukowe Franciszka Salezego)",
        21853: "Journal of Health Design",
        21854: "Faculdad de Turismo y Gastronomia de UAEM",
        21855: "Universidad Tecnologico ECOTEC",
        21856: "V.Danilevsky Institute for Endocrine Pathology"
               "Problems of NAMSU",
        21857: "Croatian Academy of Sciences and Arts",
        21858: "Mazowieckie Biuro Planowania Regionalnego w Warszawie",
        21859: "Definitive International Institute of Life"
               "Science (DIILI Publication)",
        21860: "University of Rijeka, Faculty of Medicine",
        21861: "University of Zagreb, Department of Geography",
        21862: "Eotvos Lorand University (ELTE)",
        21863: "Publishing India Group",
        21864: "British Global Travel Health Association (BGTHA)",
        21865: "Associacao Iberoamericana de Diagnostico e"
               "Avaliacao Psicologica",
        21866: "Academy of East Asian Studies",
        21867: "University of Debrecen/Kozjavak",
        21868: "University of Debrecen/Public Goods and Governance",
        21869: "Southwest State University",
        21870: "National Medical Research Radiological Centre",
        21871: "Journal of Franchise Management",
        21872: "Institute of Industrial and Systems Engineers",
        21873: "Anticancer Research USA Inc.",
        21874: "Escola Nacional de Administracao Publica (ENAP)",
        21875: "Programa de Pos Graduacao em Direito da Universidade"
               "Federal do Rio de Janeiro",
        21876: "Revista Ciencias em Saude",
        21877: "Revista Brasileira de Analises Clinicas",
        21878: "Associacao Brasileira de Pesquisadores em"
               "Comunicacao e Politica (Compolitica)",
        21879: "Revista da FAEEBA",
        21880: "Revista IUS GENTIUM",
        21881: "Abril NEPA UFF",
        21882: "Revista Uninter de Comunicacao",
        21883: "Ioffe Institute Russian Academy of Sciences",
        21884: "International Journal of Modern Trends in"
               "Engineering and Research",
        21885: "West Virginia State Medical Association (WVSMA)",
        21886: "Rostov State Medical University",
        21887: "Society for Veterinary Science and Biotechnology",
        21888: "Korean Association of Government Studies",
        21889: "Tokyo University of Science",
        21890: "International Journal of Research in Education and Science",
        21891: "Journal of Education in Science, Environment and Health",
        21892: "Corporacion Universitaria del Caribe - CECAR",
        21893: "Scientific World Ltd.",
        21894: "Malaysian Palm Oil Board",
        21895: "Universidad Catolica Los Angeles de Chimbote",
        21896: "The Korean Society of Maternal and Child Health",
        21897: "Universidad de Cordoba",
        21898: "Dianoia Revista de Filosofia",
        21899: "First Edition Design Publishing",
        21900: "University of Illinois Main Library",
        21901: "Instituto Junguiano de Sao Paulo (IJUSP)",
        21902: "International Journal of Professional Business Review",
        21903: "Akademik Bakis Uluslararasi Hakemli Sosyal"
               "Bilimler E-Dergisi",
        21904: "We Ken",
        21905: "Bindura University of Science Education",
        21906: "Polskie Towarzystwo Teologiczne Sekcja Wydawnicza",
        21907: "Roman House Productions, Ltd.",
        21908: "SIRIM Industrial Research",
        21909: "Institute of Experimental Psychology",
        21910: "Revista Brasileira de Sociologia do Direito",
        21911: "Turkish National Society of Allergy and Clinical Immunology",
        21912: "Africa's Public Service Delivery and Performance Review",
        21913: "University of South Australia Library",
        21914: "Australian Mathematical Publishing Association, Inc.",
        21915: "Institute of Mathematics, Academia Sinica",
        21916: "Bureau of Labor Statistics",
        21917: "ICT Academy",
        21918: "Association for the Advancement of Artificial"
               "Intelligence (AAAI)",
        21919: "Instituto Mexicano de Ejecutivos de Finanzas",
        21920: "Revista Eletronica Cientifica Ensino Interdisciplinar",
        21921: "Society for Upliftment of Rural Economy (SURE)",
        21922: "Scholarly Research Journals",
        21923: "Muhendislik Bilimleri ve Tasarim Dergisi",
        21924: "Komunitas Ilmuwan dan Profesional Muslim Indonesia (KIPMI)",
        21925: "Minneapolis Heart Institute Foundation",
        21926: "LIDSEN Publishing Inc",
        21927: "Alma Ata University Press",
        21928: "University of Human Development",
        21929: "Abanico Veterinario",
        21930: "Corporacion Colombiana de Investigacion Agropecuaria"
               "(Corpoica)",
        21931: "Clinical Biotec",
        21932: "SACCoM - Sociedad Argentina para las Ciencias"
               "Cognitivas de la Musica",
        21933: "ENES Unidad Leon UNAM",
        21935: "Technische Universitat Chemnitz",
        21941: "Universitatsbibliothek Kiel",
        21971: "University of Alberta - Past Imperfect Journal",
        21977: "Journal for Learning through the Arts",
        21980: "University of California, Irvine",
        21991: "University of Alberta- Constitutional Forum Journal",
        21992: "TranscUlturAl Journal",
        22012: "University of Bern",
        22013: "University of Bern",
        22015: "University of Bern",
        22018: "University of Bern",
        22134: "Trace (Travaux et Recherches dans les Ameriques du Centre)",
        22135: "Graduate Program Sriwijaya University",
        22136: "El Colegio Mexiquense",
        22137: "Central Luzon State University International Affairs Office",
        22138: "Bulletin of the Ural Medical Academie Science",
        22139: "The Journal of Business Science",
        22140: "Rutgers University - Newark School of Public"
               "Affairs and Administration",
        22141: "Publishing House Zaslavsky",
        22142: "Institute of International Trade",
        22143: "Asia Culture Academy of Incorporated Association",
        22144: "Can Tho University",
        22145: "The Australian National University",
        22146: "Universitas Gadjah Mada",
        22147: "Ansari Education and Research Society",
        22148: "CA: Journal of Cultural Analytics",
        22149: "EEEIC International Publishing",
        22150: "SlaveFree Today",
        22151: "International Association for Political Science Students",
        22152: "Aspect-Oriented Software Association (AOSA)",
        22153: "Al-Khwarizmi Engineering Journal",
        22154: "Korean Society of Children's Literature and Education",
        22155: "Association for Future Early Childhood Education",
        22156: "Convergence Society for SMB",
        22157: "GDG",
        22158: "Scholink Co, Ltd.",
        22159: "Innovare Academic Sciences",
        22160: "African Studies Association of Australasia"
               "and the Pacific (AFSAAP)",
        22161: "AI Publications",
        22162: "Kalmyk Institute for Humanities of the RAS",
        22163: "Austrian Platform for Research and Technology"
               "Policy Evaluation (FTEVAL)",
        22164: "Indonesian Center for Social and Environmental"
               "Accounting Research and Development (ICSEARD)",
        22165: "Revista Enfermagem Atual In Derme",
        22166: "Revista Dissertatio de Filosofia",
        22167: "I-PECEGE",
        22168: "Programa de Extensao TEIA Entrepalavras",
        22169: "Revista Intersaberes",
        22170: "Ciencia e Cultura",
        22171: "Revista de Estudos Juridicos UNESP",
        22172: "Korean Society For Cognitive and Biological Psychology",
        22173: "Korean Society for the Sociology of Sport",
        22174: "Institute of Communication Research",
        22175: "Iowa State University",
        22176: "MayDay Group",
        22177: "Pavilion Publications",
        22178: "Publishing Center Dialog",
        22179: "Museo Argentino de Ciencia Naturales",
        22180: "Nano Advances Publishing Office",
        22181: "APESB - Associacao Portuguesa de Engenharia"
               "Sanitaria e Ambeintal",
        22182: "Institute for Political Studies",
        22183: "Research Nebula",
        22184: "Technosphera JSC",
        22185: "Papeles de Poblacion",
        22186: "Journal of Young Investigators",
        22187: "Revista de la Facultad de Derecho",
        22188: "Makadam",
        22189: "National Association of Directors of Nursing"
               "Administration in Long Term Care",
        22190: "University of Nis",
        22191: "Binghamton University",
        22192: "Darshan Publishers",
        22193: "Instituto Mexicano del Cemento y del Concreto, A.C.",
        22194: "Journal of Global Innovations in Agricultural"
               "and Social Sciences (JGIASS)",
        22195: "Revista Educacao em Foco (UFJF)",
        22196: "Revista Pedagogica",
        22197: "Instituto Brasileiro de Direito Processual Penal",
        22198: "Region y Sociedad, Revista de el Colegio de Sonora",
        22199: "Universidad Catolica del Norte - Chile",
        22200: "Pakistan Journal of Pharmaceutical Research",
        22201: "Universidad Nacional Autonoma de Mexico",
        22202: "STKIP PGRI Sumatera Barat",
        22203: "European Cells and Materials",
        22204: "Russian Foundation for Basic Research",
        22205: "South Indian Journal of Biological Sciences",
        22206: "Instituto Tecnologico de Santo Domingo (INTEC)",
        22207: "Journal of Pure and Applied Microbiology",
        22208: "Journal of Dentomaxillofacial Science",
        22209: "Corporacion Universitaria Remington",
        22210: "Hrvatsko filolosko drustvo (Croatian Philological Society)",
        22211: "Lukasiewicz Research Network - Institute of"
               "Industrial Organic Chemistry",
        22212: "Journal of Economic and Public Policy",
        22213: "Kalashnikov Izhevsk State Technical University",
        22214: "International Journal for Research in Applied"
               "Science and Engineering Technology (IJRASET)",
        22215: "Carleton University",
        22216: "Lembaga Layanan Pendidikan Tinggi - LLDIKTI -  Wilayah X",
        22217: "Urban Planning International",
        22218: "Cairo Center for Development Benchmarking",
        22219: "Universitas Muhammadiyah Malang",
        22220: "Physiology and Medicine Publishing, Pte. Ltd.",
        22221: "Climate Modelling Laboratory",
        22223: "Faculty of Mechanical Engineering, VSB Technical"
               "University of Ostrava",
        22224: "University Consortium for Geographic Information"
               "Science (UCGIS)",
        22225: "Universitas Warmadewa",
        22226: "Institute for Metals Superplasticity Problems of RAS",
        22227: "Moscow State University of Civil Engineering",
        22228: "Revista Territorios e Fronteiras",
        22229: "Ohio Counseling Association",
        22230: "CISP Journal Services",
        22231: "Agricultura, Sociedad y Desarrollo",
        22232: "Mizoram University",
        22233: "British Small Animal Veterinary Association",
        22234: "Culturales",
        22235: "Universidad Catolica de Uruguay",
        22236: "University of Muhammadiyah Prof. Hamka (UHAMKA)",
        22237: "Wayne State University Library System",
        22238: "Revista Cerrados",
        22239: "Vigilancia Sanitaria em Debate: Sociedade,"
               "Ciencia y Tecnologia",
        22240: "Co. Ltd. Ukrinformnauka",
        22241: "Revista Tocantinense de Geografia",
        22242: "The Korea Social Welfare Research Institute",
        22243: "Korean Academy of Leadership",
        22244: "Indian Association for Angiosperm Taxonomy (IAAT)",
        22245: "The Korean Academy of New Religions",
        22246: "The Society of Internal Korean Medicine",
        22247: "EverScience Publications",
        22248: "International University of Japan",
        22249: "Institute for the Future of Higher Education"
               "in Kyung Hee Cyber University",
        22250: "Amur State University",
        22251: "Korean Association For Learner-Centered Curriculum"
               "And Instruction",
        22252: "Inonu Universitesi Kultur ve Sanat Dergisi/Inonu"
               "University Journal of Culture and Art",
        22253: "The Korean Society for Seon Studies",
        22254: "Church History Society in Korea",
        22255: "Korean Association for Buddhist Studies",
        22256: "PUBVET",
        22257: "Korean Psychological Association",
        22258: "Asociacion Civil Universidad de Ciencias y Humanidades",
        22259: "Sryahwa Publications",
        22260: "International Association for Automation and"
               "Robotics in Construction (IAARC)",
        22261: "Veruscript",
        22262: "The Open University of Israel",
        22263: "Vitebsk State Medical University",
        22264: "CLIO: Revista de Pesquisa Historica",
        22265: "Asociacion Colombiana de Nefrologia",
        22266: "The Intelligent Networks and Systems Society",
        22267: "Universidad de Narino",
        22268: "Arab Society for Plant Protection",
        22269: "Places Journal",
        22270: "Society of Pharmaceutical Tecnocrats",
        22271: "AkiNik Publications",
        22272: "Korean Nietzsche Society",
        22273: "The Society of Modern Literary Theory",
        22274: "The Society of Korean Oral Literature",
        22275: "Global English Teachers Association",
        22276: "Ethnoscientia",
        22277: "Revista Gestao Organizacional",
        22278: "Secretaria da Saude do Estado da Bahia",
        22279: "Servico Nacional de Aprendizagem Comercial / SENAC SC",
        22280: "Revinter",
        22281: "Bryansk State University",
        22282: "The Online Journal of Recreation and Sport",
        22283: "Institute of Natural Science",
        22284: "The Academy of Qualitative Research",
        22285: "The Korean Society of Emergency Medical Services",
        22286: "Korean Association of Legal Philosophy",
        22287: "Revistas Integradas Campos Salles",
        22288: "Federacao Brasileira de Gastroenterologia",
        22289: "Psicologia e Saude em Debate",
        22290: "Jornal Brasileiro de Neurocirurgia",
        22291: "Associacao Brasileira de Cefaleia",
        22292: "Revista Meio Ambiente e Sustentabilidade",
        22293: "Faculdade Damas da Instrucao Crista",
        22294: "Departamento de Educacao UFV",
        22295: "Revista Grifos",
        22296: "Revista Brasileira de Estudos Urbanos e Regionais (RBEUR)",
        22297: "Dialogo das Letras",
        22298: "Revista Fisisenectus",
        22299: "Khon Kaen University",
        22300: "American Real Estate Society",
        22301: "IJHMCR ( International Journal Of Health Medicine"
               "and Current Research )",
        22302: "Riset Perkebunan Nusantara",
        22303: "University of Potensi Utama Medan",
        22304: "Padjadjaran Jurnal Ilmu Hukum",
        22305: "Universidad Nacional de la Patagonia Austral",
        22306: "4S go, s.r.o.",
        22307: "Edicions UIB",
        22308: "AWSELVA",
        22309: "Pedagogiatorteneti Szemle",
        22310: "Cadernos de Pos-Graduacao em Disturbios do Desenvolvimento",
        22311: "Medina Publishing Ltd.",
        22312: "SDU Journal of Health Sciences",
        22313: "Residential Environment Institute Of Korea",
        22314: "FSBI All Russian Research Institute for Mechanization"
               "in Agriculture (VIM)",
        22315: "University of Debrecen/Acta Classica Universitatis"
               "Scientiarum Debreceniensis",
        22316: "Philosophy of Coaching",
        22317: "Naba'a Al-Hayat Foundation for Medical Sciences"
               "and Health Care",
        22318: "International Society of the Learning Sciences",
        22319: "Revista Mexicana de Ciencias Pecuarias",
        22320: "Universidad del Bio Bio",
        22321: "Korea Distribution and Logistics Policy Association",
        22322: "Center for Handicraft and Batik",
        22323: "Sissa Medialab Srl",
        22324: "The International Journal of Human Resource"
               "Development: Practice, Policy and Research",
        22325: "Federacion Espanola de Sociologia",
        22326: "International Journal of Anatolia Sport Sciences",
        22327: "The Research Institute of Korean Education",
        22328: "Baltic Medical Education Center",
        22329: "University of Windsor Leddy Library",
        22330: "International Society for Human Ethology (ISHE)",
        22331: "Verein zur Forderung des Open Access Publizierens"
               "in den Quantenwissenschaften",
        22332: "Center for the Study of Material and Visual"
               "Cultures of Religion",
        22333: "Center for Civil Integration and Inter-Ethnic Relations",
        22334: "JBHOST Journal of Business on Hospitality and Tourism",
        22335: "Policia Nacional de Colombia",
        22336: "Fundatia Universitara Sanabuna",
        22337: "Publishing House ASV (Izdatelstvo ASV)",
        22338: "Andalas University",
        22339: "International Big History Association",
        22340: "Chungbuk National University Social Science Institute",
        22341: "JACSOnline Group Publisher",
        22342: "Indonesian Mathematical Society",
        22343: "Gorodets",
        22344: "Hankuk University of Foreign Studies Literature Studies",
        22345: "Korean Society for Urban History",
        22346: "The Korean Association for Religious Education",
        22347: "Revista Meta - Avaliacao",
        22348: "Revista Internacional de Educacao Superior",
        22349: "FSUE CRISM Prometey",
        22350: "Editora Fi",
        22351: "Faculdades EST",
        22352: "Universidad de Magallanes, Instituto de La Patagonia",
        22353: "National University of Mongolia",
        22354: "Asociacion Colombiana de Infectologia - ACIN",
        22355: "Associacao Portuguesa de Estudos sobre as Mulheres - APEM",
        22356: "Flinders University Library",
        22357: "Rochester Cathedral Research Guild",
        22358: "The Kielanowski Institute of Animal Physiology"
               "and Nutrition, PAS",
        22359: "Journal of Clinical Social Work and Health Intervention",
        22360: "Society for Modeling and Simulation International (SCS)",
        22361: "Journal of Facility Management Education and Research",
        22362: "IJCERT Publication House",
        22363: "Peoples' Friendship University of Russia",
        22364: "University of Latvia",
        22365: "Hellenic Psychiatric Association",
        22366: "Nigerian Young Academy",
        22367: "University of Economics in Katowice",
        22368: "The Korea Society of Energy and Climate Change Education",
        22369: "The Shodor Education Foundation, Inc.",
        22370: "Universidad de Valparaiso Chile",
        22371: "University of San Diego - Copley Library",
        22372: "Center for Korean History",
        22373: "Universitas Islam Negeri Ar-Raniry",
        22374: "Dougmar Publishing Group, Inc.",
        22375: "DBS Business Review",
        22376: "International Journal of Pharma and Bio Sciences",
        22377: "BRNSS Publication Hub",
        22378: "Sh.Marjani Institute of History of Tatarstan"
               "Academy of Sciences",
        22379: "Asociacion Colombiana de Neurologia",
        22380: "Instituto Colombiano de Antropologia e Historia",
        22381: "Addleton Academic Publishers",
        22382: "Society of Wood Science and Technology",
        22383: "Universidad Catolica de Manizales",
        22384: "Hyperion University",
        22385: "Journal of Communications Technology, Electronics"
               "and Computer Science (JCTECS)",
        22386: "Universidad Cientifica del Peru",
        22387: "Capacious: The Journal for Emerging Affect Inquiry",
        22388: "Departamento de Psicobiologia - UNIFESP",
        22389: "FSBI Center of Geodesy, Cartography, and SDI",
        22390: "Institute of Public Administration and Management,"
               "Russian Academy of National Economy and Public"
               "Administration (IPAM RANEPA)",
        22391: "Family Practice and Palliative Care",
        22392: "Suleyman Demirel Universitesi Egirdir Su Urunleri"
               "Fakultesi Dergisi",
        22393: "Barcaray Publishing Limited",
        22394: "The Russian Presidential Academy of National"
               "Economy and Public Administration",
        22395: "Universidad de Medellin",
        22396: "Marmara Universitesi Spor Bilimleri Dergisi",
        22397: "Wonkwang University Legal Research Institute",
        22398: "ESPM Rio de Janeiro",
        22399: "International Journal of Computational and"
               "Experimental Science and Engineering (IJCESEN)",
        22400: "Consejo de la Juventud de Extremadura",
        22401: "Al-Nahrain Journal of Science",
        22402: "UNAM Facultad de Estudios Superiores Iztacala",
        22403: "Universidad de Quintana Roo",
        22404: "ISTEM",
        22405: "FSBEIHE Tula State Lev Tolstoy Pedagogical University",
        22406: "Science and Business Publishing UK Ltd.",
        22407: "Instituto Federal de Educacao Ciencia e Tecnologia"
               "do Rio de Janeiro - IFRJ",
        22408: "Instituto de Cultura Tecnica Sociedade Civil Ltda",
        22409: "Pro Reitoria de Pesquisa, Pos Graduacao e Inovacao - UFF",
        22410: "Editora Univates",
        22411: "REDE — Revista Eletronica do PRODEMA",
        22412: "Russian State University of Tourism and Service",
        22413: "Vasundara Groups",
        22414: "Institute for Russian, East European, and Eurasian Studies",
        22415: "The Korean-Chinese Society of Law",
        22416: "Russian Gastroenterolgocial Society",
        22417: "The Institute for Social Science Research,"
               "Kyungpook National University",
        22418: "Institute of Social Science, Gangwon University",
        22419: "The Chinese Prose Association of Korea",
        22420: "Confederacao Nacional dos Trabalhadores em Educacao (CNTE)",
        22421: "Instituto de Estudos em Saude Coletiva - INESCO",
        22422: "Revista Temporalis",
        22423: "Konkuk Research Institute of Real Estate and Urban Studies",
        22424: "Korean Society of Milk Science and Biotechnology",
        22425: "Sejong University Language Research Institute",
        22426: "Australian Counselling Association, Inc.",
        22427: "NIEHS",
        22428: "MRVSA Pubishing House",
        22429: "Editorial de la Universidad de Cantabria",
        22430: "Instituto Tecnologico Metropolitano (ITM)",
        22431: "Escuela Superior de Administracion Publica",
        22432: "Fundacja Pedagogium",
        22433: "Canadian Journal of Nutrition",
        22434: "Wageningen Academic Publishers",
        22435: "Badan Penelitian dan Pengembangan Kesehatan",
        22436: "International Scientific Research Publications MY SDN. BHD.",
        22437: "Faculty of Education and Teacher Training, Jambi University",
        22438: "Triveni Enterprises",
        22439: "Copenhagen Business School",
        22440: "WERI Education Research Development Printing"
               "Publication, Ltd.",
        22441: "Universitas Mercu Buana",
        22442: "Journal of Liaquat University of Medical and"
               "Health Sciences (JLUMHS)",
        22443: "Royal Microscopical Society",
        22444: "Information Bulletin on Variable Stars (IBVS)",
        22445: "BSNV College Lucknow",
        22446: "Multiculture and Peace Institute",
        22447: "Asian Trade Association",
        22448: "Amur State Medical Academy",
        22449: "FSBSI MHI",
        22450: "Far Eastern State Agrarian University",
        22451: "Escuela de Postgrado Neumann",
        22452: "Univ. of Malaya",
        22453: "National Council for Scientific Research - Lebanon",
        22454: "Society of Teachers of Family Medicine",
        22455: "A. M. Gorky Institute of World Literature"
               "of the Russian Academy of Sciences",
        22456: "Universidade Federal do Rio Grande do Sul",
        22457: "House of Scientific Research",
        22458: "Universidad Estatal a Distancia",
        22459: "ANU Press",
        22460: "STKIP Siliwangi Bandung",
        22461: "The Institute of Clinical Bioethics at Saint"
               "Joseph's University",
        22462: "Undersea and Hyperbaric Medical Society (UHMS)",
        22463: "Universidad Francisco de Paula Santander",
        22464: "Diyalektolog - Ulusal Hakemlin Sosyal Arastirmalar Dergisi",
        22465: "Korean Urological Oncology Society",
        22466: "Artvin Coruh Universitesi Uluslararasi Sosyal"
               "Bilimler Dergisi",
        22467: "Korean Wound Management Society",
        22468: "Asian Society of Cardiovascular Imaging",
        22469: "Korean Society of Laryngology, Phoniatrics and Logopedics",
        22470: "Korean Society of Pediatric Emergency Medicine",
        22471: "J-Institute",
        22472: "Korean Physical Education Association for Girls and Women",
        22473: "Korean Association of Thai Studies",
        22474: "Korean Society of Eastern Art Studies",
        22475: "REBECA - Revista Brasileira de Estudos de"
               "Cinema e Audiovisual",
        22476: "Revista Critica",
        22477: "Revista Juventude e Politicas Publicas",
        22478: "Portal de Periodicos UFPB",
        22479: "Textura",
        22480: "Revista da Unifa",
        22481: "Universidade Estadual do Sudoeste da Bahia/Edicoes UESB",
        22482: "Editora da Universidade de Sorocaba - UNISO",
        22483: "Programa de Pos-Graduacao em Educacao — PPGE",
        22484: "Pos-Graduacao em Comunicacao e Cultura — PPGCC",
        22485: "The Association of Engineers, India",
        22486: "The Indian Institute of Welding",
        22487: "Fakultas MIPA Universitas Tadulako Palu",
        22488: "Oklahoma State University Library",
        22489: "Computing in Cardiology",
        22490: "Universidad Nacional Abierta y a Distancia",
        22491: "Tikinet Edicao Ltda. - EPP",
        22492: "The International Academic Forum (IAFOR)",
        22493: "University of Southampton",
        22494: "Institute of Cell Therapy",
        22495: "Virtus Interpress",
        22496: "Archyworld Publishing, Ltd.",
        22497: "Universidad Privada Antenor Orrego",
        22498: "Past Global Changes (PAGES)",
        22499: "Bureau of Meteorology, Australia",
        22500: "Sodality: Jurnal Sosiologi Pedesaan",
        22501: "Society for Artistic Research",
        22502: "Lumbini Medical College",
        22503: "Informacios Tarsadalom",
        22504: "Gwangju Catholic University Institute for Theology",
        22505: "The American Studies Association of Korea",
        22506: "Society of Toxicology, India",
        22507: "Corporacion Universitaria Lasallista",
        22508: "FSBEI HE Krasnoyarsk State Agrarian University"
               "Achinsk Branch",
        22509: "Korea Food Service Association",
        22510: "Korean Financial Management Association",
        22511: "Korean Association of Women's History",
        22512: "Istanbul Universitesi Kadin Arastirmalari Dergisi",
        22513: "Baldwin Wallace University",
        22514: "MRE Press",
        22515: "IAIN Surakarta",
        22516: "Asociacion Colombiana de Gastroenterologia",
        22517: "Universidad Tecnologica de Pereira - UTP",
        22518: "Universidad Sergio Arboleda",
        22519: "Corporacion Universitaria Rafael Nunez",
        22520: "Parantez Teknologji Ltd",
        22521: "Universitepark",
        22522: "Edward Bernays College of Communication Management",
        22523: "Universidad Libre - Bogota",
        22524: "Universidad Libre - Barranquilla",
        22525: "Universidad Libre - Cartagena",
        22526: "Universidad Libre - Cucuta",
        22527: "Universidad Libre - Pereira",
        22528: "Universidad Libre - Socorro",
        22529: "Universidad Catolica de Cordoba",
        22530: "Grupo de Investigacion Arte y Ciudad, UCM",
        22531: "Mugla Journal of Science and Technology",
        22532: "Journal of Transportation and Logistics",
        22533: "Atena Editora",
        22534: "Broadcasting and Arts Research Institute",
        22535: "Revista Cadernos de Pesquisa em Educacao",
        22536: "The Modern British and American Poetry Society of Korea",
        22537: "Korean Society of Clinical Toxicology",
        22538: "The Japan Center",
        22539: "Hanyang Peace Institute",
        22540: "Hylonome Publications",
        22541: "Authorea, Inc.",
        22542: "Nevada Water Resources Association",
        22543: "Valparaiso University",
        22544: "Colegio Profesional de Psicologos de Costa Rica",
        22545: "The Academy of Transdisciplinary Learning"
               "and Advanced Studies",
        22546: "Sociedad Gallega de Medicina Interna",
        22547: "Institute of Management Sciences",
        22548: "Lajnah Pentashihan Mushaf Al-Qur'an",
        22549: "Journal of Earth Energy Engineering",
        22550: "FUNCIVA Ediciones SL",
        22551: "Grigore T. Popa University of Medicine and Pharmacy",
        22552: "Jaipuria Institute of Management, India",
        22553: "The Korean Educational Administration Society",
        22554: "Irish Learning Technology Association (ILTA)",
        22555: "Institute of Business Management",
        22556: "The Convergence Tourism Contents Society",
        22557: "The Korean Language Society",
        22558: "The Korean Industrial Economic Association",
        22559: "Cyprus International University",
        22560: "Associacao Nacional de Educacao Catolica - ANEC",
        22561: "Contabilidade Vista e Revista",
        22562: "Cadernos do CEOM",
        22563: "Jornal de Assistencia Farmaceutica e Farmacoeconomia",
        22564: "Sociedade Brasileira de Geofisica",
        22565: "International Journal of Nutrology",
        22566: "Brazilian Journal of Information Science: Research Trends",
        22567: "Revista Eniac Pesquisa",
        22568: "Jornal de Estudos Espiritas",
        22569: "Conselho Nacional de Controle Interno - CONACI",
        22570: "Revista Ergodesign e HCI",
        22571: "Acta Brasiliensis",
        22572: "Media Research",
        22573: "Skyfox Publishing",
        22574: "The Society for the Promotion of Mechanism"
               "and Institution Design",
        22575: "The Japanese Society of Interventional Radiology",
        22576: "Society of Scientific Footprints",
        22577: "Colorado State University Libraries",
        22578: "College of Medicine, Alnahrain University",
        22579: "Universidad de los Llanos",
        22580: "Mondal I-Sciences",
        22581: "Mehran University of Engineering and Technology",
        22582: "NomadIT",
        22583: "Department of Arabic Studies and Islamic Civilization",
        22584: "Yukon University",
        22585: "Centro Internacional Virtual de Investigacion"
               "en Nutricion (CIVIN)",
        22586: "Croatian Institute of History (Hrvatski Institut"
               "za Povijest)",
        22587: "American-Eurasian Network for Scientific Information (AENSI)",
        22588: "The Korean Holistic Arts Therapy Association",
        22589: "Korea Academy of Care Management",
        22590: "The Korea Association for Early Childhood"
               "Education and Educare Welfare",
        22591: "Magyar Urologia",
        22592: "Revista Odontoestomatologia",
        22593: "Institute for Pure and Applied Mathematics (IPAM)",
        22594: "Defense Acquisition University Press",
        22595: "Atlanta University Center - Robert W. Woodruff Library",
        22596: "Faculty of Education, Harran University",
        22597: "Galaxy Tech Solutions",
        22598: "Faculty of Economics and Business, University of Zagreb",
        22599: "White Rose University Press",
        22600: "Investigacoes em Ensino de Ciencias (IENCI)",
        22601: "Tallinn University Press",
        22602: "Books on Demand",
        22603: "Japanese Society for Spine Surgery and Related Research",
        22604: "University of Colorado Denver",
        22605: "Rural and Remote Health",
        22606: "Isaac Scientific Publishing Co., Ltd.",
        22607: "KROS Publications",
        22608: "Asia Pacific Academy of Ophthalmology",
        22609: "Emertec Research and Development",
        22610: "AMH International Conferences and Seminars Organizing LLC",
        22611: "Jurnal Pendidikan Fisika",
        22612: "Iqra National University",
        22613: "Zeitschrift fur Praktische Philosophie",
        22614: "Revista Saber Cientifico",
        22615: "Journal of Geospatial Modelling",
        22616: "Latvia University of Life Sciences and Technologies",
        22617: "Asian Development Bank",
        22618: "SJT Trivent SRL",
        22619: "Centre for Multidisciplinary Research, Innovation"
               "and Collaboration (C-MRiC)",
        22620: "Agricultural University",
        22621: "Ottawa Field-Naturalists' Club",
        22622: "RossiSmith Academic Publishing, Ltd.",
        22623: "International Journal of Applied and Pure"
               "Science and Agriculture",
        22624: "Creative Research Publishers",
        22625: "SPRIDA",
        22626: "Korean Family Resource Management Association",
        22627: "Journal of Childrens Infections",
        22628: "Global Institute for Japanese Studies - Border Crossings",
        22629: "The Korean Academy of Business History",
        22630: "Warsaw University of Life Sciences — SGGW Press",
        22631: "eJournalPlus",
        22632: "Computer Science Laboratory Press",
        22633: "Revista Eletronica Politica e Gestao Educacional",
        22678: "Dae Han Society of Industrial Management",
        22682: "Korean Association for Business Communication",
        22789: "The Institute of Legal Studies Inha University",
        23713: "Haditechnika",
        23736: "Edizioni Minerva Medica",
        23740: "Territorial Identity and Development",
        23741: "Romanian Review of Geographical Education",
        23797: "Wachholtz Verlag GmbH",
        23799: "Schuren Verlag",
        23822: "Edra SpA",
        23823: "Royal Botanic Garden Edinburgh",
        23834: "The Journal of International Scientific Researches",
        23835: "Tasarim and Kuram",
        23836: "The Research of the Korean Classic Novel",
        23837: "Thiagarajar School of Management",
        23838: "Sungkyunkwan University School of Medicine",
        23839: "Korea Association of Business Education",
        23840: "The Korean Society for Agriculture Education"
               "and Human Resource Development",
        23841: "The Economic Geographical Society of Korea",
        23842: "Korea Insurance Research Institute",
        23843: "Korea Appraisal Society",
        23844: "Korea Counseling Psychological Association",
        23845: "Kalagatos",
        23846: "International Initiative for Impact Evaluation (3ie)",
        23847: "The Federation of Finnish Learned Societies",
        23848: "British Journal of Clinical Audit",
        23849: "North Pacific Anadromous Fish Commission",
        23850: "Servicio Nacional de Aprendizaje SENA - SENNOVA",
        23851: "Al-Mustansiriyah Journal of Science",
        23852: "Library of the Czech Academy of Sciences",
        23853: "Sociedad Espanola de Hidrologia Medica",
        23854: "Universidad Bernardo O'Higgins",
        23855: "Czech Botanical Society",
        23856: "Polonia University in Czestochowa",
        23857: "Polo de Capacitacion, Investigacion y Publicacion (POCAIP)",
        23858: "Institute of Archaeology and Ethnology, Polish"
               "Academy of Sciences",
        23859: "Cherepovets State University",
        23860: "University of Rhode Island",
        23861: "Albert Einstein College of Medicine",
        23862: "KIIT School of Management, KIIT University",
        23863: "Kalem Uluslararasi Egitim ve Insan Bilimleri Dergisi",
        23864: "Revista Com a Palavra o Professor",
        23865: "Cappelen Damm AS - Cappelen Damm Akademisk",
        23866: "Fundacio Barcelona Respiratory Network",
        23867: "The University of Texas at Austin, Bureau"
               "of Economic Geology",
        23868: "PJSC Human Stem Cells Institute",
        23869: "Journal of Biological Researches",
        23870: "Middle Atlantic Review of Latin American Studies",
        23871: "Revista Dimensoes",
        23872: "A.J. Institute of Management",
        23873: "IPO Association of Transplantologists",
        23874: "Acharya Bangalore B School",
        23875: "Korea Communication Association",
        23876: "The Korean Society of Nephrology",
        23877: "Triacon",
        23878: "Universidad Catolica de Santiago de Guayaquil",
        23879: "Eotvos Lorand University (ELTE) - Nevtani Ertesito",
        23880: "Medwin Publishers",
        23881: "Universidad Privada Boliviana",
        23882: "NMd - Nucleo Multidisciplinar",
        23883: "International Journal of Recent Trends in"
               "Engineering and Research",
        23884: "INESEG YAYINCILIK ARGE DANISMANLIK MUHENDISLIK"
               "SAN. VE TIC. LTD. STI.",
        23885: "Southern Scientific Center of the Russian"
               "Academy of Sciences",
        23886: "Faculty of Medicine Universitas Indonesia",
        23887: "Universitas Pendidikan Ganesha",
        23888: "Ryazan State Medical University",
        23889: "Swansea University",
        23890: "International Sustainable Aviation Research Society",
        23891: "Yuzunci Yil Universitesi Egitim Fakultesi Dergisi",
        23892: "Marmara Business Review",
        23893: "Acta Pharmaceutica Sciencia",
        23894: "Korean Journal of Comparative Criminal Law",
        23895: "Korea Development Institute",
        23896: "Korean Academy of Business History",
        23897: "Selcuk University Journal of Seljuk Studies",
        23898: "Litikum",
        23899: "Centro Latino-Americano de Estudos em Cultura - CLAEC",
        23900: "Revista Politicas Publicas e Cidades",
        23901: "Revista Ciencia em Extensao — RCE",
        23902: "Trakya University Journal of Natural Sciences",
        23903: "The Korean Association of Interpretation and"
               "Translation Education",
        23904: "The Society of the Eastern Classic",
        23905: "The Korean Society for Pastoral Care and Counseling",
        23906: "Instituto Portugues de Relacoes Internacionais,"
               "Universidade Nova de Lisboa",
        23907: "Academic Forensic Pathology International",
        23908: "The Society of Philosophical Studies",
        23909: "Korean Association of Christian Counseling and Psychology",
        23910: "Puspa Publishing House",
        23911: "Linha Unica - Edicao e Comunicacao em Saude e Desporto, Lda",
        23912: "Goodfellow Publishers, Ltd.",
        23913: "Centro de Estudios e Investigaciones para"
               "el Desarrollo Docente - CENID",
        23914: "JAS Arqueologia",
        23915: "Distill Working Group",
        23916: "Indonesian Counselor Association (IKI)",
        23917: "Universitas Muhammadiyah Surakarta",
        23918: "Ishik University",
        23919: "Institute of Electrical and Electronics Engineers",
        23920: "Fakultas Hukum UNPAD (FHUNPAD)",
        23921: "Quench Academy of Medical Education and Research"
               "(QAMER) Publishing House",
        23922: "The Japan Society of Coloproctology",
        23923: "Consejo General de Colegios Oficiales de Psicologos",
        23924: "Centro de Investigacion Social Avanzada (CISAV)",
        23925: "Pontifical Catholic University of Sao Paulo (PUC-SP)",
        23926: "Revista Pratica Docente",
        23927: "Insituto Historico e Geografico Brasileiro",
        23928: "Foldtani Kozlony",
        23929: "International Journal of Academic Value Studies"
               "(Javstudies JAVS)",
        23930: "Shri Ram Murti Smarak International Business School",
        23931: "Korean Association of Child Psychotherapy",
        23932: "Center for Crisis Society Studies",
        23933: "The Korean Society of Greco-Roman Studies",
        23934: "The Scientific and Practical Society of Emergency"
               "Medicine Physicians",
        23935: "Revista Invesigacion Cualitativa",
        23936: "Saludesa",
        23937: "ClinMed International Library",
        23938: "Gobierno de Navarra",
        23939: "Lviv Polytechnic National University",
        23940: "Totem Publisher, Inc.",
        23941: "Erasmus Journal for Philosophy and Economics",
        23942: "SidSam Profeza Technologies India Private Limited",
        23943: "Princeton University Press",
        23944: "Special Education and Rehabilitation Science Research Center",
        23945: "Korean Silhak Society",
        23946: "Kemerovo State Medical University",
        23947: "FSFEI HE Don State Technical University",
        23948: "The Korean Society of Health and Welfare",
        23949: "Korean Alliance for Health, Physical Education,"
               "Recreation, and Dance",
        23950: "JSC National Scientific Medical Research Center",
        23951: "Tomsk State Pedagogical University",
        23952: "Mathematical Research Press",
        23953: "Cloud Publications",
        23954: "Open Science Journal",
        23955: "Chemical Engineering Department, Syiah Kuala University",
        23956: "Advance Academic Publisher",
        23957: "BioExcel",
        23958: "International Scientific Invention Journals",
        23959: "Scientific Federation",
        23960: "Universitas Lampung",
        23961: "Federacion Latinoamericana de Sociedades Científicas"
               "de Estudiantes de Medicina (FELSOCEM)",
        23962: "Wits School of Literature, Language and Media (SLLM)",
        23963: "Colloquium - New Philologies",
        23964: "Research Consultants and Publisher",
        23965: "Early Childhood Australia",
        23966: "Uniwersytet Jagiellonski — Wydawnictwo Uniwersytetu"
               "Jagiellonskiego",
        23967: "Scipedia, S.L.",
        23968: "Saint Petersburg State University of Architecture"
               "and Civil Engineering (SPSUACE)",
        23969: "Universitas Pasundan",
        23970: "Agency for Healthcare Research and Quality",
        23971: "IAIN Palangka Raya",
        23972: "PGDesign / Universidade Federal do Rio Grande do Sul",
        23973: "Associacao Brasileira de Medicina Preventia"
               "e Administracao em Saude  - ABRAMPAS",
        23974: "International Journal of Librarianship",
        23975: "Basrah Journal of Veterinary Research",
        23976: "Middle East Institute Singapore",
        23977: "Clausius Scientific Press, Inc.",
        23978: "Informaatiotutkimus",
        23979: "Finnish Yearbook of Population Research",
        23980: "SQS-lehti",
        23981: "Fennia - International Journal of Geography",
        23982: "Virittaja",
        23983: "Media & viestinta",
        23984: "Trace - Finnish Journal for Human-Animal Studies",
        23985: "Etnomusikologian vuosikirja",
        23986: "Agricultural and Food Science",
        23987: "Science and Technology Studies",
        23988: "Ainedidaktiikka-lehti",
        23989: "Gerontologia",
        23990: "Sosiaalilaaketieteellisen aikakauslehti",
        23991: "Ethnologia Fennica",
        23992: "Kulttuuripoliittisen tutkimuksen vuosikirja",
        23993: "Studia Orientalia Electronica",
        23994: "Lahikuva",
        23995: "Tahiti",
        23996: "Finnish Journal of eHealth and eWelfare",
        23997: "Puhe ja kieli",
        23998: "Rakenteiden mekaniikka",
        23999: "OMF Publishing",
        24000: "STC Industry Safety CJSC",
        24001: "Infogain Publication",
        24002: "Universitas Atma Jaya Yogyakarta",
        24003: "EMITTER International Journal of Engineering Technology",
        24004: "The Korean Association For Patriots And Veterans"
               "Affairs Studies",
        24005: "Korean Special Education Association for Curriculum"
               "and Instruction",
        24006: "Jungseok Research Institute of International"
               "Logistics and Trade",
        24007: "Physical Education Lab of Chung-Ang University",
        24008: "Center for Korean Literature and Language Education",
        24009: "The Korean Society of Education for Hearing-Language"
               "Impairments",
        24010: "Seyahat ve Otel Isletmeciligi Dergisi",
        24011: "Bartin Orman Fakultesi Dergisi",
        24012: "Dicle Universitesi Muhendislik Fakultesi Muhendislik Dergisi",
        24013: "Journal of Management Economics Literature"
               "Islamic and Political Science",
        24014: "Universitas Islam Negeri Sultan Syarif Kasim Riau",
        24015: "Evidence Based Communications",
        24016: "Instituto Peruano de Orientacion Psicologica - IPOPS",
        24017: "Sulaimani Polytechnic University",
        24018: "European Open Access Publishing (Europa Publishing)",
        24019: "Fondazione Vasculab",
        24020: "The Korean Association for Local Public Enterprises",
        24021: "Revista Acta Ambiental Catarinense",
        24022: "A.N. Bakoulev National Scientific and Practical"
               "Center for Cardiovascular Surgery",
        24023: "Future Studies Research Journal: Trends and Strategies",
        24024: "Faculdade Frassinetti do Recife- FAFIRE",
        24025: "Cherkasy State Technological University",
        24026: "Lancet XXI STORIChChYa, LLC",
        24027: "National Scientific Centre Institute of Metrology",
        24028: "Institute of Geophysics of the National Academy"
               "of Sciences of Ukraine by S.I. Subbotin name",
        24029: "LEJANA - Revista Critica de Narrativa Breve",
        24030: "Essays on Conservatism",
        24031: "PH V. Ema",
        24032: "Empirical Research Press, Ltd.",
        24033: "Societe Mathematique de France",
        24034: "Sekolah Tinggi Ilmu Ekonomi Indonesia (STIESIA ) Surabaya",
        24035: "Department of Theology and Philosophy, UKM",
        24036: "Universitas Negeri Padang (UNP)",
        24037: "LLC Approbation",
        24038: "Ergon Creacion, S.A",
        24039: "Universidad Nacional Federico Villarreal",
        24040: "Matej Bel University in Banska Bystrica",
        24041: "ERA's Lucknow Medical College",
        24042: "Raden Intan State Islamic University of Lampung",
        24043: "Institute of Island Studies",
        24044: "Science Publishing Centre Sociosphere",
        24045: "Vedecko Vydavatelske Centrum Sociosfera-CZ, s.r.o.",
        24046: "Association for Research in Neuroeducation",
        24047: "International Plant Nutrition Institute",
        24048: "International Center for Business & Management"
               "Excellence (ICBME)",
        24049: "Florida Gulf Coast University",
        24050: "Universidad EIA",
        24051: "Mlodzi Naukowcy dr Jedrzej Nyckowiak",
        24052: "Centre for Business and Economic Research",
        24053: "UVK Verlagsgesellschaft mbH",
        24054: "Universidad de Pamplona",
        24055: "The Korean Association of Police Science",
        24056: "Korean Accounting Association (KAA)",
        24057: "Russian Geographical Society",
        24058: "Turk Kulturu lncelemeleri Dergisi",
        24059: "The Online Learning Consortium",
        24060: "Bashkir State Medical University",
        24061: "Higher State Educational Establishment of"
               "Ukraine Bukovinian State Medical University",
        24062: "The Korean Association of Practical Arts Education",
        24063: "HOSEO University Social Science Research Institute",
        24064: "International Workshop on Theory of Submanifolds",
        24065: "Universidade Federal do Oeste do Para",
        24066: "Ac Korean Association of Rusists",
        24067: "Educadora Sete de Setembro",
        24068: "Academia Brasileira de Direito Constitucional",
        24069: "Association of Science Editors and Publishers",
        24070: "Brazilian Journal of Veterinary Pathology",
        24071: "Sanata Dharma University",
        24072: "Peer Community In",
        24073: "St. John's University",
        24074: "Turkish Journal of Reproductive Medicine and Sugery",
        24075: "Pirogov Russian National Research Medical University",
        24076: "STMIK AMIKOM Yogyakarta",
        24077: "Editora e Livrarias Plena Ltd.",
        24078: "BPM Medica",
        24079: "Central Asian Journal of Medical Sciences",
        24080: "The Society of Hospital Pharmacists of Australia",
        24081: "NFC Institute of Engineering and Fertilizer Research",
        24082: "Alevi Bektasi Kultur Enstitusu",
        24083: "Australasian College of Health Service Management",
        24084: "AEDERMACP (European Association for the Development"
               "of Renewable Energies and Power Quality)",
        24085: "Stellenbosch University - Journal of Student"
               "Affairs in Africa",
        24086: "Cihan University of Erbil",
        24087: "The International Annals of Medicine",
        24088: "International Journal of Business and Economic Affairs",
        24089: "Jurusan Sistem Informasi",
        24090: "IAIN Purwokerto",
        24091: "Turkish Rhinologic Society",
        24092: "Current Research in Pharmaceutical Sciences",
        24093: "AWEJ Group",
        24094: "Union of Mechanical and Electrotechnical Engineers"
               "and Technicians of Serbia (SMEITS)",
        24095: "Health Promotion and Chronic Disease Prevention"
               "Branch (HPCDP) Public Health Agency of Canada",
        24096: "Versus Verlag AG",
        24097: "Wolfram Research",
        24098: "Pandit Ravishankar Shukla University",
        24099: "Faculty of Veterinary Medicine, University of Zagreb",
        24100: "Magyar Kemiai Folyoirat",
        24101: "Journal LOGOS",
        24102: "Science Target, Inc.",
        24103: "Global Century Press",
        24104: "Asian Academic Press Co., Limited",
        24105: "Global Institute for Research and Education (OMICS)",
        24106: "Kastamonu Egitim Dergisi",
        24107: "International Journal of Engineering and Applied Sciences",
        24108: "NPG Publishing",
        24109: "Instituto Nacional de Estudos e Pesquisas"
               "Educacionais Anisio Teixeira",
        24110: "Pediatria, Ltd.",
        24111: "Jurnal Riset Industri Hasil Hutan",
        24112: "Hong Kong Baptist University",
        24113: "Smart Moves",
        24114: "State University of Medan",
        24115: "Laplage em Revista",
        24116: "Educacao Matematica Debate",
        24117: "Transversal: International Journal for the"
               "Historiography of Science",
        24118: "Secretaria Municipal de Saude do Rio de Janeiro",
        24119: "Associacao de Docentes da Estacio de Sa (ADESA)",
        24120: "Embrapa Agroindustria Tropical",
        24121: "Diabetologia Hungarica",
        24122: "Institute of Business Research",
        24123: "University of Surabaya",
        24124: "University of Northern British Columbia",
        24125: "Association of Medical Doctors Sanamed",
        24126: "Biotechnology Research Center",
        24127: "Muhammadiyah Metro University",
        24128: "International Journal of Recent Advancement"
               "in Engineering and Research (IJRAER)",
        24129: "Imaidea Interactiva, S.L.",
        24130: "Association for the Development of Early Childhood"
               "Education in Turkey",
        24131: "Instytut Badan Edukacyjnych",
        24132: "University of West Bohemia",
        24133: "Universidad de las Fuerzas Armadas ESPE",
        24134: "Brief Encounters Postgraduate Journal",
        24135: "Auckland University of Technology (AUT) Library",
        24136: "Instytut Badan Gospodarczych / Institute of"
               "Economic Research",
        24137: "Asociacion Espanola de Investigacion de la Comunicacion",
        24138: "Croatian Communications and Information Society",
        24139: "Sumy State Pedagogical University named A.S.Makarenko",
        24140: "Universidade Lusofona de Humanidades e Tecnologias",
        24141: "Zdravstveno veleuciliste",
        24142: "Universidad Autonoma Latinoamericana",
        24143: "Astrakhan State Technical University",
        24144: "Uzhhorod National University",
        24145: "The Korean Institute of Public Affairs (Seoul"
               "National University)",
        24146: "Turk Kutuphaneciligi - Turkish Librarianship",
        24147: "Dostoevsky Omsk State University",
        24148: "Federal Reserve Bank of San Francisco",
        24149: "Federal Reserve Bank of Dallas",
        24150: "University of Missouri",
        24151: "National Research University of Electronic Technology (MIET)",
        24152: "The Korean Society of Nineteenth Century Literature"
               "in English",
        24153: "LLC Russian Science",
        24154: "Society for Promotion of Horticulture - Indian"
               "Institute of Horticultural Research",
        24155: "Turk Dunyasi Dergisi",
        24156: "Jurnal Ilmu Keluarga dan Konsumen",
        24157: "Arcadian Group S.A.",
        24158: "HORS Publishing House, LLC",
        24159: "Education Research Institute",
        24160: "Moscow Power Engineering Institute (MPEI)",
        24161: "The Society for Chinese Historical Researches",
        24162: "Estudios Irlandeses",
        24163: "International Journal of Applied Research and Technology",
        24164: "Balai Arkeologi Jawa Barat",
        24165: "Journal of Neurological Sciences (Turkish)",
        24166: "Internet Mathematics",
        24167: "Soegijapranata Catholic University",
        24168: "Revista Prumo",
        24169: "University of Debrecen/Debreceni Jogi Muhely",
        24170: "Stellenbosch University -SA Heart Journal",
        24171: "Korea Centers for Disease Control and Prevention",
        24172: "Korea Health Communication Association",
        24173: "Da Vinci Mirea Institute of General Education",
        24174: "Research Institute for Image and Cultural Contents",
        24175: "Scientific Big Data, S.L.",
        24176: "Universitas Muria Kudus",
        24177: "Hong Kong Institute for Research and Innovation"
               "in Technology",
        24178: "Research Plus Journals",
        24179: "Otorhinolaryngology and Head and Neck Surgery Society",
        24180: "Uluslararasi Tarim ve Yaban Hayati Bilimleri Dergisi",
        24181: "Tarim Ekonomisi Dergisi",
        24182: "Publishing Agency Science and Education",
        24183: "All Russian Public Organization Russian Scientific"
               "Society of Endovascular Diagnostic and Treatment",
        24184: "Korean Association for Research, Procedures"
               "and Education on Trauma",
        24185: "Sungshin Women's University Humanities Research Institute",
        24186: "Vakanuvis International Journal of Historical Researches",
        24187: "Institut National de la Statistique et des"
               "Etudes Economiques (INSEE)",
        24188: "Universidad de Sucre",
        24189: "Fund for Support and Development of Protected Areas",
        24190: "European Biotechnology Thematic Network Association",
        24191: "UiTM Press, Universiti Teknologi MARA",
        24192: "redfundamentos SL",
        24193: "Babes-Bolyai University",
        24194: "Publishing House Scientific Review",
        24195: "Ushynsky University",
        24196: "Public Library of Science (PLoS)",
        24197: "Universidad de Valladolid",
        24198: "Universitas Padjadjaran",
        24199: "Museums Victoria",
        24200: "SciTech Solutions",
        24201: "El Colegio de Mexico, A.C.",
        24202: "Hampshire Field Club and Archaeological Society",
        24203: "Asian Online Journals",
        24204: "Verein zur Forderung der Fachzeitschrift European"
               "Journal for Philosophy of Religion",
        24205: "Fundacion Aigle",
        24206: "LaborHistorico",
        24207: "Revista Latino-Americana de Marcapasso e Arritmia",
        24208: "Revista Brasileira de Educacao em Ciencia"
               "da Informacao (REBECIN)",
        24209: "Enriched Publications Pvt. Ltd.",
        24210: "Korean Association for Public Management",
        24211: "The Korean Society for the Study of Teacher Education",
        24212: "Nucleo de Estudos do Futuro NEF PUCSP",
        24213: "Stellenbosch University - South African Journal"
               "of Pre-hospital Emergency Care",
        24214: "SAE Publications",
        24215: "Universidad Nacional de La Plata",
        24216: "ibidem-Verlag",
        24217: "FUAM (Fundacioon Universidad Autonoma de Madrid)",
        24218: "Verizona Publisher Limited",
        24219: "Revista de Pesquisa Interdisciplinar",
        24220: "Cadernos de Fe e Cultura, Oculum Ensaios,"
               "Reflexao, Revista de Ciencias Medicas e Revista"
               "de Educacao da PUC-Campinas",
        24221: "Journal of Environmental Analysis and Progress - JEAP",
        24222: "Revista Stricto Sensu",
        24223: "NPO Energobezopasnost",
        24224: "OOO Centr naucnyh i obrazovatelnyh proektov",
        24225: "Korean Artificial Intelligence",
        24226: "Korean Society of Vocational Rehabilitation",
        24227: "The Institute of Korean Language and Literature"
               "in Chonnam National University",
        24228: "Kozlekedestudomanyi Szemle",
        24229: "Korean Evangelical Society of New Testament Studies",
        24230: "Korean Society for Industrial and Organizational Psychology",
        24231: "Research Institute of Curriculum and Instruction",
        24232: "Jeoloji Muhendisligi Dergisi",
        24233: "Biology Department, Faculty of Mathematics"
               "and Natural Sciences, Sriwijaya University",
        24234: "Armenian State Pedagogical University after Khacatur Abovyan",
        24235: "IAIN Syekh Nurjati Cirebon",
        24236: "SENA - Libraries",
        24237: "University of Diyala, College of Science",
        24238: "Publiverse Online S.R.L",
        24239: "IAIN Palu",
        24240: "Register of Qualified Genealogists",
        24241: "CIDOB (Barcelona Centre for International Affairs)",
        24242: "Library Juice Press",
        24243: "Rational Publication",
        24244: "Indonesia Society of Neuroanesthesia and Critical"
               "Care (INA-SNACC)",
        24245: "Nieto Editores",
        24246: "Universitas Kristen Satya Wacana",
        24247: "Transstellar Journal Publications and Research"
               "Consultancy Private Limited",
        24248: "East African Health Research Commission",
        24249: "Faculty of Philology, Lomonosov Moscow State University",
        24250: "Aurel Vlaicu University of Arad",
        24251: "HICSS Conference Office",
        24252: "Universitas Islam Negeri Alauddin Makassar",
        24253: "Colegio de Enfermeria de Caceres",
        24254: "Sociedad Mexicana de Ingenieria Biomedica, A.C.",
        24255: "Fundacja Heart Beat",
        24256: "Institut Agama Islam Negeri (IAIN) Palopo",
        24257: "Balai Bahasa Jawa Timur",
        24258: "Jurnal Borneo Administrator",
        24259: "Fakultas Kehutanan, Universitas Hasanuddin"
               "(Forestry Faculty, Hassannuddin Univ)",
        24260: "IAIN Pontianak",
        24261: "Associacao Internacional de Lusitanistas",
        24262: "E-Discovery Publication",
        24263: "National University of Food Technologies",
        24264: "Leather and Footwear Research Institute (INCDTP-ICPI)",
        24265: "Universidad de San Martin de Porres",
        24266: "Horticultural Research Institute",
        24267: "Universidad de Boyaca",
        24268: "Australian Catholic University",
        24269: "Universitas Muhammadiyah Ponorogo",
        24270: "The Educational Research Institute",
        24271: "University of Garmian",
        24272: "Zoological Research",
        24273: "Journal of Geoscience, Engineering, Environment"
               "and Technology (JGEET)",
        24274: "One Central Press",
        24275: "Universidad Autonoma Metropolitana",
        24276: "Revista Recien - Revista Cientifica de Enfermagem",
        24277: "Classica - Revista Brasileira de Estudos Classicos",
        24278: "Revista do Instituto Florestal",
        24279: "Journal of Management and Technology",
        24280: "Argumentos Pro-Educacao",
        24281: "Revista Remecs - Revista Multidisciplinar"
               "de Estudos Cientificos em Saude",
        24282: "Colab, Auckland University of Technology",
        24283: "Hellenic Regulatory Body of Nurses",
        24284: "The Korean Society for Urban Railway",
        24285: "Korean Association of Chinese Language Education",
        24286: "The Society Korean Children Literature",
        24287: "Fund Doctors, Innovations, Science for Children",
        24288: "Journal of Tourism Theory and Research",
        24289: "International Journal of Social Sciences and"
               "Education Research",
        24290: "Faculty of Sociology, Lomonosov Moscow State University",
        24291: "Uskonnontutkija — Religionsforskaren",
        24292: "Medical Education",
        24293: "PDS Patklin",
        24294: "EnPress Publisher",
        24295: "China Power Supply Society",
        24296: "JOMI, LLC",
        24297: "CIRWORLD",
        24298: "Disaster Nursing Global Leader Degree Program",
        24299: "Korea University Institute of Educational Research",
        24300: "The Korean Society of Child Welfare",
        24301: "Korea Academy of Mental Health Social Work",
        24302: "Universidade do Contestado - UnC",
        24303: "The Linguistic Association of Korea",
        24304: "The Korean College of Clinical Pharmacy",
        24305: "Caderno Eletronico de Ciencias Sociais",
        24306: "Association of European Schools of Planning (AESOP)",
        24307: "Polgari Szemle",
        24308: "International Association for Semiotic Studies",
        24309: "ASEPUMA",
        24310: "Malaga University",
        24311: "University of Economics Ho Chi Minh City",
        24312: "University of Central Punjab",
        24313: "International Society for Problem-Based Learning",
        24314: "Consejo Regional III Lima Metropolitana",
        24315: "Trakya Universitesi Egitim Fakultesi Dergisi",
        24316: "PROMETEICA - Revista de Filosofia y Ciencias",
        24317: "Revista Brasileira de Extensao Universitaria",
        24318: "Committee on Publication Ethics",
        24319: "Jurnal Teknologi Perikanan dan Ilmu Kelautan",
        24320: "Revista Electronica de Investigacin Educativa",
        24321: "Advanced Research Publications",
        24322: "Uitgeverij Verloren BV",
        24323: "Akademik Gida",
        24324: "Korean Association of International Association"
               "of Constitutional Law",
        24325: "South American Development Society Journal",
        24326: "Uniwersytet Przyrodniczy w Lublinie",
        24327: "Current Research",
        24328: "Letra e Imagem Editora e Producoes Ltda",
        24329: "Jurnal ASPIKOM",
        24330: "The International Electronic Journal of Algebra",
        24331: "International Journal of Educational Research Review",
        24332: "The Korean Society of Sports Physical Therapy",
        24333: "Korean Society of Old Testament Studies",
        24358: "Politechnika Lubelska",
        24361: "Performa",
        24411: "St. Petersburg Institute of History of the"
               "Russian Academy of Sciences",
        24415: "Leiden University Press",
        24434: "Seismo Verlag",
        24437: "University of Basel",
        24445: "Universitaet Zuerich, Philosophisches Seminar",
        24446: "Cantonal and University Library Fribourg",
        24450: "Universite de Geneve",
        24452: "University of Bern",
        24814: "Koreanische Gesellschaft Fuer Deutsche Sprachwissenschaft",
        24815: "LPPM Unsyiah",
        24816: "Content Ed Net Taiwan Limited",
        24817: "Center For Chemical & Packaging (Balai Besar"
               "Kimia & Kemasan)",
        24818: "Bucharest University of Economic Studies",
        24819: "NETSOL: New Trends in Social and Liberal Sciences",
        24820: "ARKAT USA, Inc.",
        24821: "Institut Seni Indonesia Yogyakarta",
        24822: "Revista Juridica Digital UANDES",
        24823: "Royal Botanic Garden Edinburgh",
        24824: "Editora CRV",
        24825: "Korean Association for Semiotic Studies",
        24826: "Korean Society for the Security Convergence Management",
        24827: "Systematic Theology Division Korea Evangelical"
               "Theological Society",
        24828: "The Korea Society of Buddhist Art History",
        24829: "Korean Buechner Society",
        24830: "Korean Society for German Language and Literature",
        24831: "Department of Agronomy and Horticulture",
        24832: "Badan Penelitian dan Pengembangan Kemdikbud",
        24833: "MGIMO University",
        24834: "Malmo University Library",
        24835: "Vidar, Ltd.",
        24836: "Estudios Sociales",
        24837: "Industrial and Organizational Psychology Association",
        24838: "Indian Phytopathological Society",
        24839: "Psi Chi, the International Honor Society in Psychology",
        24840: "University of Porto",
        24841: "Instituto de Investigaciones de la Amazonia Peruana - IIAP",
        24842: "International Journal of Computation and Applied Science",
        24843: "Universitas Udayana",
        24844: "Sociedad Mexicana de Investigacion y Divulgacion"
               "de la Educacion Matematica A. C.",
        24845: "Department of Chemistry Sriwijaya University",
        24846: "ICI Bucharest",
        24847: "Khayrallah Center for Lebanese Diaspora Studies",
        24848: "Mardjani Foundation for Support and Development"
               "of Research and Cultural Programs",
        24849: "Portuguese Geotechnical Society",
        24850: "Instituto Mexicano de Tecnologia del Agua",
        24851: "Texas Association for Bilingual Education (TABE)",
        24852: "A.Kh. Khalikov Archaeology Institute of the"
               "Academy of Sciences of Tatarstan",
        24853: "Universitas Muhammadiyah Jakarta",
        24854: "Konsorsium Psikologi Ilmiah Nusantara",
        24855: "Scientific Research Foundation XXI Century",
        24856: "UNTAG Semarang",
        24857: "RGSA- Revista de Gestao Social e Ambiental",
        24858: "Revista Dialogos Mediterranicos",
        24859: "Revista da Faculdade de Direito de Valenca",
        24860: "Comunicologia",
        24861: "Delictae: Revista de Estudos Interdisciplinares"
               "sobre o Delito",
        24862: "Fundacao Educacional de Formiga - FUOM",
        24863: "Ceuma Associacao de Ensino Superior",
        24864: "Centro interdisciplinar de Pesquisa em Educacao"
               "e Direito CIPED",
        24865: "IMLA (Arabic Teacher and Lecturer Association of Indonesia)",
        24866: "FSAEIHE Far Eastern Federal University (FEFU)",
        24867: "Faculty of Technical Sciences",
        24868: "Institute of Marine Engineering, Science and"
               "Technology (IMarEST)",
        24869: "Medicinska Naklada d.o.o.",
        24870: "Science Planet Inc.",
        24871: "The Indonesian Journal of Gastroenterology,"
               "Hepatology and Digestive Endoscopy",
        24872: "Rocky Mountain Geology, University of Wyoming",
        24873: "Sociedade Portuguesa de Estomatologia e Medicina"
               "Dentaria (SPEMD)",
        24874: "Faculty of Engineering, University of Kragujevac",
        24875: "Publicidad Permanyer, SLU",
        24876: "Senex:Journal of Aging Studies",
        24877: "Fincham Press",
        24878: "The Korea Entrepreneurship Society",
        24879: "Psicologia em Pesquisa",
        24880: "Mehmet Akif Ersoy Universitesi Veteriner Fakultesi Dergisi",
        24881: "KUKKIWON",
        24882: "Revista Educacao em Debate",
        24883: "Revista Inteligencia Competitiva",
        24884: "FSBEI HE I.P. Pavlov SPbSMU MOH Russia",
        24885: "Revista de Arqueologia da SAB",
        24886: "Korea Business Law Association",
        24887: "Oil Industry Corporation",
        24888: "Bunin Yelets State University",
        24889: "Dokuz Eylul Universitesi Isletme Fakultesi Dergisi",
        24890: "The Rome Foundation",
        24891: "Publishing House Finance and Credit",
        24892: "Russian Internet Publishing House",
        24893: "Andalas Journal of Public Health",
        24894: "Schwabe Verlagsgruppe AG",
        24895: "Geospatial Information Agency of The Republic of Indonesia",
        24896: "Journal of Microbiology and Biotechnology Research",
        24897: "American Research Foundation LLC",
        24898: "Turk Androloji Dernegi",
        24899: "DataOpen",
        24900: "International Journal of Safety Science",
        24901: "Relaciones Estudios de Historia y Sociedad",
        24902: "Universidade de Evora",
        24903: "Universitas Widya Gama Mahakam",
        24904: "Structurae",
        24905: "Universitas Pancasakti",
        24906: "Institute of Science, Education and Culture, Kolkata",
        24907: "Jeju National University Tourism, Business,"
               "and Economic Research Institute",
        24908: "Queen's University Library",
        24909: "ANSA Royals",
        24910: "SustaiN Society",
        24911: "Discover STM Publishing Ltd.",
        24912: "Universitas Tarumanagara",
        24913: "Romanian Journal of Applied Psychology, West"
               "University of Timisoara",
        24914: "Fakultas Ekonomika dan Bisnis",
        24915: "Associacao Portuguesa de Urologia",
        24916: "Archeologicke Centrum Olomouc",
        24917: "The Pedagogical University of Cracow/Uniwersytet"
               "Pedagogiczny w Krakowie",
        24918: "CourseSource",
        24919: "Drohobych Ivan Franko State Pedagogical University",
        24920: "Chinese Medical Sciences Journal",
        24921: "Carl Grossman Verlag",
        24922: "Doctoral Program in Tourism",
        24923: "LLC Publisher Stavrolit",
        24924: "FTMS College",
        24925: "Turkish Science and Technology Publishing (TURSTEP)",
        24926: "University of Minnesota",
        24927: "ICETA",
        24928: "IGLC",
        24929: "Universitas Wiraraja",
        24930: "IGG UB RAS",
        24931: "Russian Photodynamic Association",
        24932: "The Korean Society Of Educational Psychology",
        24933: "Casa de Nossa Senhora da Paz A.S.F",
        24934: "Educacao em Foco",
        24935: "Associacao Software Livre.Org",
        24936: "Competencia - Revista da Educacao Superior do Senac-RS",
        24937: "Krylov State Research Centre",
        24938: "Kirikkale Universitesi Tıp Fakultesi Dergisi",
        24939: "The Korean Association For Japanese History",
        24940: "International Journal of Innovative Research"
               "& Development (GlobeEdu)",
        24941: "International Journal of Current Research",
        24942: "Amity University, Gurgaon",
        24943: "Indian Institute for Human Settlements",
        24944: "Kresearch Corp",
        24945: "eRelation AG",
        24946: "cosmoanelixis",
        24947: "Bio Accent Group LLC",
        24948: "International Science Council",
        24949: "National University of Sciences and Technology",
        24950: "Sociedade Portuguesa de Medicina Interna",
        24951: "Sreyas Institute of Engineering and Technology",
        24952: "IAIN Padangsidimpuan",
        24953: "The Turkish Journal of Pediatrics",
        24954: "Marmara Iktisat Dergisi  (Marmara Journal of Economics)",
        24955: "Ankara Universitesi ilef Dergisi",
        24956: "Studies of Seon Culture",
        24957: "Korean Association for Housing Policy Studies",
        24958: "Institute for Russian and Altaic Studies Chungbuk University",
        24959: "National University of Pharmacy",
        24960: "Balai Riset dan Standardisasi Industri Padang",
        24961: "Asosiasi Agroindustri Indonesia (AGRIN)",
        24962: "The Publisher.org",
        24963: "International Joint Conferences on Artificial Intelligence",
        24964: "Program Studi Akuntansi Fakultas Ekonomi",
        24965: "Instituto Nacional de Administracion Publica",
        24966: "Herald Scholarly Open Access",
        24967: "Universitas Sang Bumi Ruwa Jurai",
        24968: "Swarthmore College",
        24969: "Center for Scientific Research and Development of Education",
        24970: "Jurnal Bina Hukum Lingkungan",
        24971: "Widya Teknika",
        24972: "International Journal of Transpersonal Studies",
        24973: "Open Therapeutics",
        24974: "Association of Mexican American Educators Journal",
        24975: "Centr Regionalnikh Issledovanij",
        24976: "Discovery Medicine",
        24977: "Revista Prima Facie",
        24978: "MODOS Revista de Historia da Arte",
        24979: "Universidade Estadual de Roraima",
        24980: "Fundacao Municipal de Educacao e Cultura de Santa Fe do Sul",
        24981: "International Society For Education Through Art- InSEA",
        24982: "Journal of Interaction Science",
        24983: "SciTeMed Publishing Group",
        24984: "University Service Publishing",
        24985: "Korea Institute of Sport Science",
        24986: "Korean Academy of Anxiety Disorders",
        24987: "Seoul National University Asia Center",
        24988: "Izmir Iktisat Dergisi",
        24989: "Facultas Verlags- und Buchhandels AG",
        24990: "The Association of Indonesian Nurse Education Center",
        24991: "Korean Association of Human Resource Development",
        24992: "Korean Hospitality and Tourism Academy",
        24993: "The Korea Association of Literature for Children"
               "and Young Adults",
        24994: "European Centre for Governance in Complexity",
        24995: "University of North Carolina at Greensboro",
        24996: "University of Baghdad College of Science",
        24997: "Korean Agricultural Economics Association",
        24998: "Mehmet Akif Ersoy Universitesi Saglik Bilimleri Dergisi",
        24999: "International Journal of Advanced Engineering"
               "and Management (IJOAEM)",
        25000: "Archives of Clinical and Experimental Medicine",
        25001: "Korean Business Education Association",
        25002: "Turk Immunoloji Dernegi",
        25003: "Review of Agrarian Studies",
        25004: "International Journal of Pharmaceutical Sciences"
               "and Drug Research",
        25005: "Avicenna Tajik State Medical University",
        25006: "Austrian Pharmacological Society  (APHAR)",
        25007: "Nawroz University",
        25008: "Jurnal Komuniksi ISKI",
        25009: "Universidad Veracruzana",
        25010: "Hypothes.is",
        25011: "Canadian Society for Clinical Investigation",
        25012: "Journals and Authors Publications",
        25013: "American Association for Thoracic Surgery",
        25014: "Facultad de Odontologia Universidad Nacional de Cordoba",
        25015: "Jurnal Penyuluhan",
        25016: "NRCERM EMERCOM of Russia",
        25017: "Khanty-Mansiysk State Medical Academy",
        25018: "LLC Gornaya Kniga",
        25019: "National University of Political Studies and"
               "Public Administration (SNSPA)",
        25020: "The Institute for Education and Research Gyeongin"
               "National University of Education",
        25021: "The Society for the Research of Chinese Language"
               "and Literature",
        25022: "The Research Society for the Korean Language Education",
        25023: "Korean Association for Public Security Administration",
        25024: "The Academy of Korean Studies",
        25025: "Facultad de Artes y Humanidades",
        25026: "Faculty of Pharmacy, Mulawarman University",
        25027: "Acta Graphica",
        25028: "Colegio Oficial de Ingenieros Agronomos de Centro y Canarias",
        25029: "Fundacio Universitaria Balmes",
        25030: "Nederlandse Commissie voor Stralingsdosimetrie (NCS)",
        25031: "Illiesia, International Journal for Stonefly Research",
        25032: "Claves Revista de Historia",
        25033: "Signum",
        25034: "Alanya Hamdullah Emin Pasa Universitesi",
        25035: "Bowling Green State University Libraries",
        25036: "Performance Health Academy",
        25037: "FKIP The University of Jember",
        25038: "AM Journal of Art and Media Studies",
        25039: "International Commission on Illumination",
        25040: "Danylo Halytskyi Lviv National Medical University",
        25041: "Fiat Justisia",
        25042: "Center of Technology (COT)",
        25043: "Corporacion de Ciencia y Tecnologia para el"
               "Desarrollo de la Industria Naval Maritima y"
               "Fluvial - COTECMAR",
        25044: "Fundacion Universitaria Tecnologico de Comfenalco - FUTC",
        25045: "Institute of Information Technology of Azerbaijan"
               "National Academy of Sciences",
        25046: "ASTES Journal",
        25047: "Politeknik Negeri Jember",
        25048: "Turkish Journal of Diabetes and Obesity",
        25049: "Transaction of Engineering and Applied Science",
        25050: "The Daesoon Academy of Sciences",
        25051: "The Korean Society for Narrative Education",
        25052: "Korean Society of Supply Chain Management",
        25053: "Educacao e Formacao",
        25054: "Universidad Surcolombiana",
        25055: "The Korean Society for Research on Nicotine and Tobacco",
        25056: "Korean College of Cosmetic Surgery",
        25057: "Institucion Universitaria de Envigado",
        25058: "Colegio Mayor de Cundinamarca",
        25059: "Retratos de Assentamentos",
        25060: "Residencia Pediatrica",
        25061: "Revista Brasileira Multidisciplinar - Rebram",
        25062: "Escuela Superior de Guerra",
        25063: "Asociacion Colombiana de Facultades de Odontologia",
        25064: "Mulkiye Dergisi",
        25065: "IPO Academy of Production Engineering Sciences and Practice",
        25066: "Revista Agropecuraria Tecnica - AGROTEC",
        25067: "Revista Sinais",
        25068: "DEDE KORKUT Uluslararasi Turk Dili ve Edebiyati"
               "Arastirmalari Dergisi",
        25069: "Strategic Public Management Journal",
        25070: "Revista de Economia e Agronegocio - REA",
        25071: "York University Libraries",
        25072: "Sekolah Tinggi Hukum Bandung",
        25073: "Vietnam National University Journal of Science",
        25074: "Universidad Academia de Humanismo Cristiano",
        25075: "Mining Institute Belgrade",
        25076: "The Business and Vocational Foreign Languages"
               "Teachers National Association",
        25077: "Universitas Andalas",
        25078: "Institut Hindu Dharma Negeri Denpasar",
        25079: "University of Kurdistan Hewler",
        25080: "SciPy",
        25081: "TathQeef Scientific Publishing",
        25082: "Syncsci Publishing Pte., Ltd.",
        25083: "Digital ProScholar Media",
        25084: "ECS Publications",
        25085: "Sociedad Entomologica Argentina",
        25086: "Institute for Eastern Classic Studies",
        25087: "Revista de Educacion Superior del Sur Global - RESUR",
        25088: "Complex Systems Publications, Inc.",
        25089: "Management Education and Research Institute (MERI)",
        25090: "Revista de Educacao Matematica",
        25091: "Novos Estudos - CEBRAP",
        25092: "Balikesir Universitesi Fen Bilimleri Enstitusu Dergisi",
        25093: "Institute of British and American Studies",
        25094: "Revista Texto Poetico",
        25095: "Muhasebe ve Finansman Dergisi",
        25096: "Anti Ageing Foundation",
        25097: "Facultad de Ciencias Economicas y Administrativas"
               "de la Universidad de Cuenca",
        25098: "Cihan University - Sulaimanyia",
        25099: "Journal of Nursing and Health",
        25100: "Universidad del Valle",
        25101: "Music and Entertainment Industry Educators"
               "Association (MEIEA)",
        25102: "International Association for Fuzzy Sets Management"
               "and Economy (SIGEF)",
        25103: "International Hellenic University",
        25104: "Badan Litbang Perhubungan",
        25105: "Universitas Trisakti",
        25106: "Scientific Society for History of Health Culture, Belgrade",
        25107: "Remedy Publications, LLC",
        25108: "International Organization for Legal Researches",
        25109: "Revista da Agu",
        25110: "Universidade Paranaense",
        25111: "CDAK Society of Communication Design",
        25112: "Associacao Pro-Ensino Superior em Novo Hamburgo",
        25113: "Farol -Revista de Estudos Organizacionais e Sociedade",
        25114: "Servico Nacional de Aprendizagem da Industria"
               "do Parana (Senai/PR)",
        25115: "Editorial Universidad de Almeria",
        25116: "Kozelitesek",
        25117: "Institute of Humanities and Social Sciences",
        25118: "Associacao Brasileira de Psiquiatria",
        25119: "Revista Praxis",
        25120: "James Cook University",
        25121: "Borgis, Ltd.",
        25122: "S.C.  JURNALUL PENTRU MEDICINA SI VIATA S.R.L",
        25123: "Veritas et Justitia",
        25124: "Telkom University",
        25125: "AD Publications",
        25126: "Fakultas Ilmu Komputer Universitas Brawijaya",
        25127: "Universidad Nacional Toribio Rodriguez de Mendoza",
        25128: "Ternopil Volodymyr Hnatiuk National Pedagogical University",
        25129: "Scire Science Pvt., Ltd.",
        25130: "Tikrit University",
        25131: "Geological Society of South Africa",
        25132: "Asociacion Argentina de Cirugia",
        25133: "Institute for Population and Social Research",
        25134: "University of Kuningan",
        25135: "ACG Publications",
        25136: "Aurora Group, s.r.o",
        25137: "United Association of Serbia for Quality UASQ",
        25138: "Kritike: An Online Journal of Philosophy",
        25139: "Dr. Soetomo University",
        25140: "Chernihiv Polytechnic National University",
        25141: "Biocore Group",
        25142: "Silesian University in Opava",
        25143: "Riga Stradins University",
        25144: "Institute of Acoustics Limited",
        25145: "University of La Laguna",
        25146: "KSPU named after V.P. Astafiev",
        25147: "STEP Academic Publisher",
        25148: "Florida International University",
        25149: "JCF Corporate Consulting Services",
        25150: "Dong-k Society of Language and Literature",
        25151: "The New Korean Association of English Language"
               "and Literature",
        25152: "Secondary Education Institute",
        25153: "Society of Preventative Korean Medicine",
        25154: "Latin America Research Commons",
        25155: "Australian Institute of Male Health and Studies (AIMHS)",
        25156: "Erbil Polytechnic University",
        25157: "Universitas Galuh Ciamis",
        25158: "Cultural Studies Association",
        25159: "UNISA Press",
        25160: "Brasiliana - Journal for Brazilian Studies",
        25161: "Sp. z o.o. Kozmenko Science Publishing",
        25162: "Franz Steiner Verlag GmbH",
        25163: "Eman Research, Ltd.",
        25164: "Chinese Nursing Research",
        25165: "International Journal of Agricultural and"
               "Biological Engineering (IJABE)",
        25166: "Human Journals",
        25167: "Uniwersytet Opolski",
        25168: "The Korean Association of Peace and Religious Studies",
        25169: "University of Sulaimani - College of Law and Politics",
        25170: "Atma Jaya Catholic University of Indonesia",
        25171: "Institute of Geophysics Polish Academy of Sciences",
        25172: "Southern Methodist University",
        25173: "Jawaharlal Nehru Tropical Botanic Garden and"
               "Research Institute",
        25174: "Society for Advancement of Wheat and Barley Research",
        25175: "National Institute of Rural Development and Panchayati Raj",
        25176: "Instituto de Investigacion en Ciencias Biomedicas",
        25177: "Sift Desk",
        25178: "The New Research of Tuva Editorial Board",
        25179: "Turkish Journal of Endocrinology and Metabolism",
        25180: "Axia Academic Publishers",
        25181: "Politeknik Negeri Lampung",
        25182: "Jurnal Gizi dan Pangan/Journal of Nutrition and Food",
        25183: "American OpenScience",
        25184: "Anales de la Facultad de Medicina",
        25185: "Humanidades: Revista de la Univeridad de Montevideo",
        25186: "Coffee Science",
        25187: "Codex - Revista de Estudos Classicos",
        25188: "Faculdade Luterana de Teologia",
        25189: "Associacao Brasileira de Linguistica",
        25190: "Expressao Catolica",
        25191: "Expressao Catolica Saude",
        25192: "Centro Universitario Autonomo do Brasil",
        25193: "Almanaque de Ciencia Politica",
        25194: "Seminario Adventista Latino-Americano de Teologia",
        25195: "University of Information Technology and Communications",
        25196: "Advances Publishing Co, Ltd.",
        25197: "Korea International Law Review",
        25198: "Orenburg State University",
        25199: "RPO Surgical Society - Wound and Wound Infections",
        25200: "Associacao Brasileira de Pesquisadores de Jornalismo",
        25201: "Hitelintezeti",
        25202: "The Association of Korean Geographers",
        25203: "Journal of Behavior at Work",
        25204: "Iktisadi Idari ve Siyasal Arastirmalar Dergisi",
        25205: "Novosibirsk State University (NSU)",
        25206: "FSB EIHE Omsk State Technical University",
        25207: "Kuban State Medical University",
        25208: "Rossijskoe Obschestvo Dermatovenerologov i Kosmetologov",
        25209: "Ailamazyan Program Systems Institute of Russian"
               "Academy of Sciences (PSI RAS)",
        25210: "Scientific and Technological Center of Unique"
               "Instrumentation of RAS",
        25211: "University of Engineering and Technology, Peshawar",
        25212: "Lebanese French University",
        25213: "Fundacion Tecnologica Antonio de Arevalo - TECNAR",
        25214: "Biteca",
        25215: "Redshine Publication",
        25216: "Pusat Penelitian dan Pengembangan Hukum dan"
               "Peradilan Mahkamah Agung RI",
        25217: "Institut Agama Islam Ma arif NU (IAIMNU) Metro Lampung",
        25218: "African Society of Food Science and Technology",
        25219: "Engineering Project Organization Society",
        25220: "Indonesian Nutrition Association",
        25221: "Federal Scientific Center of the East Asia"
               "Terrestrial Biodiversity, FEB RAS",
        25222: "Latin American Studies Association",
        25223: "BioOne (British Cactus and Succulent Society)",
        25224: "BioOne (The Plant Resources Center, The University"
               "of Texas at Austin)",
        25225: "BioOne (Institute of Vertebrate Biology, Academy"
               "of Sciences of the Czech Republic)",
        25226: "BioOne (British Ornithologists' Club)",
        25227: "BioOne (Dutch Bryological and Lichenological"
               "Society and Nordic Bryological Society)",
        25228: "Gumushane Universitesi Ilahiyat Fakultesi Dergisi (GIFAD)",
        25229: "Bulletin of Economic Theory and Analysis",
        25230: "All-Russia Research Institute of Oil Crops by Pustovoit V.S.",
        25231: "The Korea Association of Primary English Education",
        25232: "A Laboratory of Korean Studies",
        25233: "ICLEL Conferences",
        25234: "Faculty of Law Josip Juraj Strossmayer University of Osijek",
        25235: "Scientific Synergy LLC",
        25236: "Francis Academic Press Ltd.",
        25237: "Asociacion de Medicos Anestesiologos de Chile",
        25238: "The Academy of Medicine, Engineering, and"
               "Science of Texas (TAMEST)",
        25239: "Sudanese American Medical Association",
        25240: "Tropical Journal of Applied Natural Sciences (TJANS)",
        25241: "ROPOSTURO Romanian Association of Oral Rehabilitation"
               "and Posturotherapy",
        25242: "Institutos Superiores de Ensino do Censa",
        25243: "Revista Odontologia Clinico-Cientifica",
        25244: "Trilhas Filosoficas",
        25245: "Revista Direitos Sociais e Politicas Publicas UNIFAFIBE",
        25246: "Direito e Desenvolvimento",
        25247: "Universidade Catolica de Pernambuco",
        25248: "Revista Eletronica Acervo Saude",
        25249: "Sociedade Brasileira de Geologia",
        25250: "The Science Breaker",
        25251: "National Society for Cutaneous Medicine",
        25252: "Soil and Environment",
        25253: "Insight Turkey/SETA Vakfi",
        25254: "Auto/Fiction",
        25255: "COES and RJ, LLC",
        25256: "Pan-Pacific Association of Applied Linguistics",
        25257: "State Fire Academy of EMERCOM of Russia",
        25258: "Dr. Yashwant Research Labs Pvt. Ltd.",
        25259: "Scientific Scholar",
        25260: "Asociacion Argentina de Ecologia",
        25261: "Auraria Library, University of Colorado Denver",
        25262: "Niger Delta Medical Journal",
        25263: "BSD Research",
        25264: "The National University of Ostroh Academy",
        25265: "Representation of Azerbaijan International"
               "Diaspora Centre in Georgia (RAIDCG)",
        25266: "Journal of Advanced Research in Dynamical"
               "and Control Systems (JARDCS)",
        25267: "Servicio de Publicaciones de la Universidad de Cadiz",
        25268: "Marine and Coastal Research Institute INVEMAR",
        25269: "Center for Pulp and Paper",
        25270: "HMP Communications, LLC",
        25271: "Science Journal of University of Zakho",
        25272: "The Center for Political, Economic and Social"
               "Research (PESA)",
        25273: "Universitas PGRI Madiun",
        25274: "Bhatter College, Dantan",
        25275: "Asia Pacific Institute of Advanced Research (APIAR)",
        25276: "Publishing House Ophthalmology",
        25277: "Korean Association of Criminal Psychology",
        25278: "Sekolah Tinggi Theologia Jaffray Makassar",
        25279: "Saglik Akademisi Kastamonu",
        25280: "The Korean Gerontological Research Society",
        25281: "FSBI Russian State Library",
        25282: "Tip Egitimi Dunyasi",
        25283: "IBRAE RAN",
        25284: "Association of Anesthesiologists of Ukraine",
        25285: "The Centre for Independent Social Research (CISR)",
        25286: "Revista de Engenharia e Pesquisa Aplicada",
        25287: "Omer Halisdemir Universitesi Iktisadi ve Idari"
               "Bilimler Fakultesi Dergisi",
        25288: "Turkiye Jeoloji Bulteni",
        25289: "The Korean Society for Laser Medicine and Surgery",
        25290: "Friends of the Princeton University Library",
        25291: "Little William Bourke",
        25292: "Pusat Penelitian dan Pengabdian kepada Masyarakat (P3M)",
        25293: "Epiarte, s.l.",
        25294: "Akdeniz Universitesi Iktisadi ve Idari Bilimler Dergisi",
        25295: "Fiscaoeconomia",
        25296: "Geomarketing",
        25297: "Society for Art Education of Korea",
        25298: "Grodno State Medical University",
        25299: "UIR Press",
        25300: "MIS Quarterly",
        25301: "Journal of Pakistan Dental Association",
        25302: "Patient-Centered Outcomes Reserach Institute",
        25303: "World Researchers Associations",
        25304: "Association for Learning Technology",
        25305: "Ukranian Association of Neurosurgeons",
        25306: "Sosyal ve Kulturel Arastirmalar Dergisi",
        25307: "Spor Bilimleri Arastirmalari Dergisi",
        25308: "Adnan Menderes Universitesi Ziraat Fakultesi Dergisi",
        25309: "Korean Society of Evangelical Practical Theory",
        25310: "Australian Orthopaedic Association National"
               "Joint Replacement Registry (AOANJRR)",
        25311: "LPPM Hang Tuah Pekanbaru",
        25312: "Akademia Humanistyczno-Ekonomiczna w Lodzi",
        25313: "Internauka",
        25336: "University of Alberta - Canadian Studies in Population",
        25341: "Nanyang Technological University: NTU PRO-AM conference",
        25366: "Sachsische Landesbibliothek, Staats-"
               " und Universitatsbibliothek Dresden",
        25367: "Sachsische Landesbibliothek, Staats-"
               " und Universitatsbibliothek Dresden",
        25368: "Sachsische Landesbibliothek, Staats-"
               " und Universitatsbibliothek Dresden",
        25369: "Sachsische Landesbibliothek, Staats-"
               " und Universitatsbibliothek Dresden",
        25370: "Sachsische Landesbibliothek, Staats-"
               " und Universitatsbibliothek Dresden",
        25371: "Sachsische Landesbibliothek, Staats-"
               " und Universitatsbibliothek Dresden",
        25372: "Sachsische Landesbibliothek, Staats-"
               " und Universitatsbibliothek Dresden",
        25518: "University of Liege",
        25531: "Sachsische Landesbibliothek, Staats-"
               " und Universitatsbibliothek Dresden",
        25555: "Cifra Ltd - Russian Agency for Digital Standardization"
               "(RADS)",
        25557: "Russian Agency for Digital Standardization (RADS)",
        25630: "Cifra Ltd - Russian Agency for Digital Standardization"
               "(RADS)",
        25726: "Cifra Ltd - Russian Agency for Digital Standardization"
               "(RADS)",
        25796: "Registre de Dialyse Peritoneale de Langue Francaise (RDPLF)",
        26006: "Dynamic Processes in Geospheres, IDG RAS",
        26034: "Cantonal and University Library Fribourg",
        26043: "Universitaet Zuerich, Institut fuer Erziehungswissenschaft",
        26045: "Hochschule der Kunste Bern HKB",
        26119: "Special Astrophysical Observatory of the Russian"
               "Academy of Sciences",
        26128: "Sachsische Landesbibliothek, Staats-"
               " und Universitatsbibliothek Dresden",
        26140: "Association Professional Analytics of Autopoiesis Systems",
        26170: "Pedagogical Education in Russia, Ural State"
               "Pedagogical University Publisher",
        26226: "Morressier",
        26334: "Associacao Portuguesa de Linguistica",
        26351: "Ariel University Press",
        26353: "Societa di Ortoflorofrutticoltura Italiana (SOI)",
        26358: "Instituto Politecnico de Tomar",
        26359: "Instituto EPOMEX - Universidad Autonoma de Campeche",
        26363: "de Gruyter: Slovak Ethnology journal",
        26376: "TU1208 GPR Association",
        26389: "The Arab Journal of Sciences and Research Publishing",
        26407: "Stg ReAttach Therapy International Foundation",
        26414: "ILKE Ilim Kultur ve Egitim Dernegi",
        26415: "Knowledge Kingdom Publishing",
        26416: "MedicHub Media",
        26417: "EUSER",
        26418: "Tanjungpura University",
        26419: "AARP",
        26420: "Austin Publishing Group",
        26421: "Rinton Press",
        26422: "Asociacion Civil de Estudios Superiores (ACES)"
               "Universidad Austra",
        26423: "Universidad Estatal Peninsula de Santa Elena",
        26424: "Lucian Blaga Central University Library of Cluj",
        26425: "State University of Management",
        26426: "Institute For Kyeongki Cultural Studies",
        26427: "Korea Evangelical Historical Theological Society",
        26428: "FSBSI TINRO Center",
        26429: "Navandu Technologies",
        26430: "Cardiologia Hungarica",
        26431: "Association for Spanish and Portuguese Historical Studies",
        26432: "Fundacao Arnaldo Vieira de Carvalho",
        26433: "The Table and Food Coordinate Society of Korea",
        26434: "American Chemical Society",
        26435: "M.Gorky Donetsk National Medical University",
        26436: "Humanities Journal of University of Zakho",
        26437: "African Journal of Applied Research",
        26438: "ISROSET: International Scientific Research"
               "Organization for Science, Engineering and Technology",
        26439: "Universidad de Lima",
        26440: "International Healthcare Research Journal",
        26441: "Universidad de Piura",
        26442: "MediaMedica",
        26443: "McGill University Library and Archives",
        26444: "Institute of Rural Health",
        26445: "Sociedad Medicina Interna de Uruguay",
        26446: "Korea Society of Leisure, Recreation, and Park",
        26447: "Kozgazdasagi Szemle Alapitvany",
        26448: "Ariadna Ediciones",
        26449: "Social Sciences Studies Journal",
        26450: "International Journal of Social Humanities"
               "Sciences Research (JSHSR)",
        26451: "Animal Behavior and Cognition",
        26452: "GP Innovations Pvt. Ltd.",
        26453: "Online Turk Saglik Bilimleri Dergisi",
        26454: "Girisimcilik ve Kalkinma Dergisi",
        26455: "Matters of Behaviour",
        26456: "Tver State University",
        26457: "Universidad La Salle, Mexico",
        26458: "Fundatia Romania de Maine",
        26459: "Hue University",
        26460: "Program Studi Manajemen Fakultas Ekonomi",
        26461: "INNOTEC",
        26462: "Boletin de la Sociedad Zoologica del Uruguay",
        26463: "Rajiv Gandhi University of Health Sciences",
        26464: "Earth and Planetary Physics",
        26465: "Ottoman Journal of Tourism and Management Research",
        26466: "OPUS - Uluslararasi Toplum Arastirmalari Dergisi",
        26467: "Moscow State Institute of Civil Aviation",
        26468: "Trakya Universitesi Sosyal Bilimler Dergisi",
        26469: "Korean Society of Physiological and Anthropological Science",
        26470: "Korea CPTED Association",
        26471: "Asociatia Carpatica de Mediu si Stiintele Pamantului",
        26472: "KnowledgeCuddle Publication",
        26473: "ATLAANZ (Association of Tertiary Learning"
               "Advisors, Aotearoa/New Zealand)",
        26474: "Society of Automotive Engineers Australasia",
        26475: "Faculty of Islamic Studies, University Kebangsaan Malaysia",
        26476: "Inscribe Publications Pte. Ltd.",
        26477: "Journal of Baghdad College of Dentistry",
        26478: "The Learned Press",
        26479: "Life Science Informatics Publications",
        26480: "ZIbeline International Publishing",
        26481: "University of Maastricht",
        26482: "Annals of Liaquat University of Medical and Health Sciences",
        26483: "IJARCS International Journal of Advanced Research"
               "in Computer Science",
        26484: "Association of Teachers and Researchers in Asia (ATERA)",
        26485: "Lodzkie Towarzystwo Naukowe",
        26486: "Universitas Mercu Buana Yogyakarta",
        26487: "Fakultas Ekonomi dan Bisnis Unhas",
        26488: "Sabre Management Solutions Pvt. Ltd.",
        26489: "Revista de Ciencias Sociales",
        26490: "Universidad Nacional del Centro del Peru",
        26491: "Institute of Meteorology and Water Management"
               "- National Research Institute",
        26492: "National Parks Board",
        26493: "University of Primorska Press",
        26494: "Queen Mary University of London",
        26495: "Universidad Senor de Sipan SAC",
        26496: "Royal Belgian Institute of Natural Sciences (RBINS)",
        26497: "Sociedade Portuguesa de Endocrinologia Diabetes"
               "e Metabolismo",
        26498: "Savigny Verlagsgesellschaft mbH",
        26499: "Badan Pengembangan dan Pembinaan Bahasa",
        26500: "Journal of Advanced Research in Social Sciences"
               "and Humanities",
        26501: "Riphah International University",
        26502: "Fortune Journals",
        26503: "Digital Games Research Association",
        26504: "Economic and Social Research Institute",
        26505: "Diyala Journal of Medicine",
        26506: "American Institute of the History of Pharmacy",
        26507: "Asociacion Colombiana de Facultades de Ingenieriia (ACOFI)",
        26508: "Life Science Alliance",
        26509: "Federal Reserve Bank of Cleveland",
        26510: "iMedPharm",
        26511: "The Society for Environmental Technology in Korea",
        26512: "Biblioteca Central da UNB",
        26513: "Turkiye Ortadogu Calismalari Dergisi",
        26514: "Interfaces da Educacao",
        26515: "Zoological Survey of India",
        26516: "Irkutsk State University",
        26517: "Yeni Turk Edebiyati Arastirmalari",
        26518: "Siberian State Automobile and Highway University (SibADI)",
        26519: "Magma Research and Consultancy Pvt. Ltd",
        26520: "IFIASA (Ideas Forum International Academic"
               "and Scientific Association)",
        26521: "University of Debrecen - Pro Futuro",
        26522: "Brock University Library",
        26523: "Faculty of Teacher Training and Education,"
               "Mulawarman University",
        26524: "Eleyon Publishers",
        26525: "Forest Research Institute Malaysia",
        26526: "LLC Perspectiva",
        26527: "International Journal of Advance Research"
               "in Engineering, Science and Technology (IJAREST)",
        26528: "The Society of Korean and Chinese Humanities",
        26529: "Faculty of Education, University of Ljubljana",
        26530: "OAPEN Foundation",
        26531: "National Bank of Ukraine",
        26532: "Program Doktor Ilmu Hukum Unissula",
        26533: "STIE PGRI Dewantara Jombang",
        26534: "Philippine Federation of Chemistry Societies, Inc.",
        26535: "Institute of Global Perspectives on Medical Sciences",
        26536: "Family Science Association",
        26537: "Polytechnic of Porto",
        26538: "University of Benin",
        26539: "Pusat Kajian Bimbingan dan Konseling FIPPS Unindra",
        26540: "Environment and Social Welfare Society",
        26541: "International Journal of Advances in Applied"
               "Mathematics and Mechanics (IJAAMM)",
        26542: "Korean Society for Media Law, Ethics and Policy Research",
        26543: "International Research Journal of India",
        26544: "Revista Eletronica Evidencia e Enfermagem",
        26545: "Amazonian Journal of Plant Research",
        26546: "Magyar Nyelvtudomanyi Tarsasag",
        26547: "Direito Cultura e Cidadania",
        26548: "The Scandinavian Society of Korea",
        26549: "Synergy Publishing Pte. Ltd.",
        26550: "Australian College of Perioperative Nurses",
        26551: "STF Driyarkara - Diskursus",
        26552: "University of Zilina",
        26553: "Faculty of Public Health of Sriwijaya University",
        26554: "ARTS Publishing",
        26555: "Universitas Ahmad Dahlan, Kampus 3",
        26556: "Journal of Ethics and Social Philosophy",
        26557: "Litera",
        26558: "International Journal of Crop Science and Technology",
        26559: "Mersin Universitesi Saglik Bilimleri Dergisi",
        26560: "Istanbul Universitesi Iktisat Fakultesi Mecmuasi",
        26561: "Istanbul Universitesi Edebiyat Fakultesi Turk"
               "Dili ve Edebiyati Dergisi",
        26562: "AM Publications",
        26563: "Dobras",
        26564: "Fundacion Universidad Autonoma de Colombia",
        26565: "V. N. Karazin Kharkiv National University",
        26566: "Centro de Estudos em Geopolitica e Realcoes"
               "Internacionais (CENEGRI)",
        26567: "Marmara Saglik Hizmetleri Dergisi",
        26568: "Educa - Revista Multidisciplinar em Educacao",
        26569: "Shanti Braj Educational and Welfare Society",
        26570: "Islami Arastirmalar Dergisi",
        26571: "Revista REAMEC",
        26572: "Magnus Publishing and Distributors",
        26573: "AIMS International",
        26574: "TARUS Media",
        26575: "Dental Anthropology Association",
        26576: "STIKES PKU Muhammadiyah Surakarta",
        26577: "al-Farabi Kazakh National University",
        26578: "Balai Riset dan Standardisasi Industri Samarinda",
        26579: "Strategic Researches Academy",
        26580: "Psychtech and Health Journal",
        26581: "Sean Kingston Publishing",
        26582: "Faculty of Kinesiology, University of Zagreb",
        26583: "National Research Nuclear University MEPhI"
               "(Moscow Engineering Physics Institute)",
        26584: "Global Center for Arts and Culture Education",
        26585: "Chinese Studies Institute",
        26586: "Chinese Literary Society",
        26587: "Institute for Social Sciences, Gyeongsang"
               "National University",
        26588: "Korea Deposit Insurance Corporation",
        26589: "Cheongram Korean Language Education Association",
        26590: "Institute for the Study of Theology",
        26591: "The Japanese Education Association of Korea",
        26592: "The Korean Society of Inclusive Education",
        26593: "LPPM UNPAR",
        26594: "Universitas Pesantren Tinggi Darul Ulum (Unipdu)",
        26595: "Associacio Catalana de Comptabilitat i Direccio",
        26596: "World Public Health Nutrition Association",
        26597: "JHUP Journals Division",
        26598: "American University of Iraq, Sulaimani",
        26599: "Tsinghua University Press",
        26600: "Romanian Society of Biochemistry and Molecular Biology",
        26601: "Association of Pulmonologists from Republika Srpska",
        26602: "Singapore Management and Sports Science Institute",
        26603: "International Journal of Sports Physical Therapy",
        26604: "M2 Books",
        26605: "Medicina Veterinaria (UFRPE)",
        26606: "Korean Association of Addiction Crime",
        26607: "International Journal of Sustainable Lighting",
        26608: "American Numismatic Society",
        26609: "Naturepub Academics, Inc.",
        26610: "Balai Bahasa Jawa Barat",
        26611: "Statperson Publishing Coporation",
        26612: "Poder Judicial Cordoba",
        26613: "Academic Studies Press",
        26614: "Department of Wood Science and Technology,"
               "Biotechnical Faculty, University of Ljubljana",
        26615: "Assoc. for Computational Linguistics Bulgaria",
        26616: "National Institute for Occupational Safety"
               "and Health (NIOSH)/Education Information Division (EID)",
        26617: "Mental Health Research Institute",
        26618: "Publikasi Jurnal Ilmiah Akademik Universitas"
               "Muhammadiyah Makassar",
        26619: "Universidade Autonoma de Lisboa, EdiUal",
        26620: "Corporacion Minuto de Dios University - UNIMINUTO",
        26621: "Pontificia Universidad Catolica del Ecuador Sede Ibarra",
        26622: "Centre for Knowledge Futures",
        26623: "Universitas Semarang",
        26624: "Astronomical Society of the Pacific Conference Series",
        26625: "Wydawnictwo Czelej Sp. z o.o.",
        26626: "Editora Scienza",
        26627: "United Life Publications",
        26628: "Agenda Wydawnicza SIMP Przegla Spawalnictwa",
        26629: "International Journal of Maxillofacial Prosthetics",
        26630: "Politeknik Kesehatan Tanjungkarang",
        26631: "VostECO Ltd.",
        26632: "Seattle Science Foundation",
        26633: "Pan American Health Organization",
        26634: "i-manager Publications",
        26635: "Pasifika Medical Association",
        26636: "National Institute of Telecommunications",
        26637: "MKD Publishing House",
        26638: "STKIP Muhammadiyah Pringsewu Lampung",
        26639: "Masyarakat Sejarawan Indonesia",
        26640: "Direccion General Maritima - DIMAR",
        26641: "SE Dnipropetrovsk Medical Academy of Health"
               "Ministry of Ukraine",
        26642: "Zhytomyr State Technological University",
        26643: "Edupedia Publications",
        26644: "Sapientia Publishing Group",
        26645: "Revista de Direito da Cidade",
        26646: "Revista Quaestio Iuris",
        26647: "Petru Major University of Tirgu-Mures",
        26648: "JAMPAPER",
        26649: "University of Miskolc",
        26650: "Istanbul University",
        26651: "International Allelopathy Foundation",
        26652: "COMSATS Institute of Information Technology",
        26653: "Publishing House Economic Education",
        26654: "International Association for Global Culture",
        26655: "International Society of Communication and"
               "Development Between Universities (ISCDBU)",
        26656: "Rynnye Lyan Resources",
        26657: "Gulhane Medical Journal",
        26658: "Journal of Mosaic Research",
        26659: "Societatea Medicala Romana de Educatie Fizica"
               "si Sport/(SMREFS)",
        26660: "Asociatia Bibliotecarilor din Romania /ABR,"
               "Romanian Library Association",
        26661: "Zaporizhzhia National University",
        26662: "Novateur Publication LLP",
        26663: "Current Thoracic Surgery",
        26664: "Revista Brasileira da Historia da Midia",
        26665: "Hallym University Institute of Russian Studies",
        26666: "RMP Publications",
        26667: "CBE - Congresso Brasileiro de Educacao",
        26668: "Conselho Nacional de Pesquisa e Pos-Graduacao"
               "em Direito - CONPEDI",
        26669: "Revista de  Ciencia e Inovacao",
        26670: "IFTM Campus Uberlandia Centro",
        26671: "International Journal of Innovative Research and Growth",
        26672: "Anatolian Bryology",
        26673: "Revista Temas em Educacao e Saude",
        26674: "ImpressionsPress",
        26675: "Journal of Accounting and Business Education",
        26676: "Orebro Univeirsty Hospital, EVTM Program",
        26677: "Turk Turizm Arastirmalari Dergisi",
        26678: "Associacao Brasileira de Engenharia e Ciencias"
               "Mecanicas - ABCM",
        26679: "East Himalayan Society for Spermatophyte Taxonomy",
        26680: "Journal of Indian Ocean World Studies",
        26681: "Eastern Kentucky University",
        26682: "University of Duhok",
        26683: "NGO Allukrainian Association of Endovascular Neuroradiology",
        26684: "Researchtribe",
        26685: "Undergraduate Research in Natural and Clinical"
               "Science and Technology (URNCST) Journal",
        26686: "Victoria University of Wellington Library",
        26687: "IJAR APCBE Publishing",
        26688: "DeWitt Wallace Library, Macalester College",
        26689: "Bio-Byword Scientific Publishing, Pty. Ltd.",
        26690: "Self Regulation Institute",
        26691: "Center for Advanced Research in Integrated"
               "Future Society (CARIFS)",
        26692: "University of Sindh",
        26693: "Petro Mohyla Black Sea National University",
        26694: "Universidade Federal do Piaui",
        26695: "Marmara Universitesi Kadin ve Toplumsal Cinsiyet"
               "Arastirmalari Dergisi",
        26696: "Escuela de Postgrado - Universidad Catolica Santa Maria",
        26697: "Kharkiv Regional Public Organization - Culture of Health",
        26698: "Instituto de Astrofisica de Canarias",
        26699: "Journal of Ners and Midwifery",
        26700: "SH Urban Research Institute",
        26701: "European Mechanical Science",
        26702: "Revista Diorito",
        26703: "Journal of Commerce and Trade (JCT)",
        26704: "Revista Pensar Geografia",
        26705: "Payam Publishing, Pvt. Lt.",
        26706: "RAME Publishers",
        26707: "Revista Brasileira de Orientacao Profissional",
        26708: "SweDha Publications",
        26709: "Indian Potato Association",
        26710: "CSRC Publishing, Center for Sustainability"
               "Research and Consultancy",
        26711: "Journal of Mathematics Education",
        26712: "Journal of Holistic Performance",
        26713: "RGN Publications",
        26714: "LPPM Universitas Muhammadiyah Semarang",
        26715: "HealthMinds Consulting Pte Ltd",
        26716: "Armacost Library",
        26717: "Biomedical Research Network, LLC",
        26718: "Solar Energy Trade Shows",
        26719: "World Health Organization Regional Office"
               "for the Eastern Mediterranean (WHO/EMRO)",
        26720: "Moravske Zemsk Muzeum (Monrovian Museum)",
        26721: "Southeast Asian Ministers of Education Organization"
               "Regional Centre for Archaeology and Fine Arts",
        26722: "Seguro Social EsSalud",
        26723: "Asociacion Mexicana de Endoscopia Gastrointestinal",
        26724: "Scientific Society Anatomists, Histologists,"
               "and Embryologists, Ukraine",
        26725: "Extension Education Society",
        26726: "Association RPET",
        26727: "Community Medicine Faculties Association",
        26728: "International Journal of Disciplines In Economics"
               "and Administrative Sciences Studies (IDEAstudies)",
        26729: "Journal on Advances in Theoretical and Applied Informatics",
        26730: "T.F. Gorbachev Kuzbass State Technicak University",
        26731: "Irkutsk State Transport University",
        26732: "TP National Information Satellite System",
        26733: "Revista Em Tempo",
        26734: "The Korea Association of Future Education",
        26735: "Naif Arab University for Security Sciences",
        26736: "University of Richmond",
        26737: "STKIP Singkawang",
        26738: "New Health Concept sal",
        26739: "Tadqiqot",
        26740: "Universitas Negeri Surabaya",
        26741: "ABA Espana",
        26742: "Institut Seni Budaya Indonesia Bandung",
        26743: "Korean Association of Dance",
        26744: "*ASTTLR- Asociatia Specialistilor in Terapia"
               "Tulburarilor de Limbaj din Romania",
        26745: "Gazi Universitesi Iktisadi ve Idari Bilimler"
               "Fakultesi Dergisi",
        26746: "Aarya Publishing House",
        26747: "Anglo-American Publication",
        26748: "The Korean Society of Ocean Engineers",
        26749: "The Royal Society of Tasmania",
        26750: "University of Raparin",
        26751: "Universitas Muhammadiyah Kudus",
        26752: "Fundacion Universitaria Juan N. Corpas",
        26753: "STIKES Muhammadiyah Gombong",
        26754: "Universidad de Zaragoza",
        26755: "Institutul de Stiinte ale Educatiei (Institute"
               "of Educational Sciences)",
        26756: "Lebanese American University",
        26757: "Philippine Journal of Systematic Biology",
        26758: "Anthropological Researches and Studies",
        26759: "Vivekanada Institute of Professional Studies",
        26760: "Institut Teknologi Nasional, Bandung",
        26761: "International Journal of Research in Library Science",
        26762: "Zeetarz Publishing",
        26763: "Rewilding Europe",
        26764: "Korea Society of Acting Studies",
        26765: "Direct Research Journals",
        26766: "Private Publisher Chaban O.S.",
        26767: "Coloquio Revista do Desenvolvimento Regional",
        26768: "Scientific Blend Publications",
        26769: "International Journal of Innovative and Emerging"
               "Research in Engineering",
        26770: "Revista Phoinix",
        26771: "Laboratorio de Orientacao da Genese Organizacional"
               "- Universidade Federal de Santa Catarina",
        26772: "Caleb University",
        26773: "Montenegrin Sports Academy",
        26774: "Osrodek Pamiec i Przyszlosc",
        26775: "OpenPsych",
        26776: "Deer Hill Publications",
        26777: "SpringerNature",
        26778: "SpringerNature",
        26779: "Liga-Inform, Ltd.",
        26780: "Bates College",
        26781: "ClinTransMed, AB",
        26782: "Journal of Mechanics of Continua and Mathematical Sciences",
        26783: "iGenesis SciMed Pro, LLC",
        26784: "Asociacion Espanola de Contabilidad y Administracion"
               "de Empresas (AECA)",
        26785: "Direct Knowledge LLC",
        26786: "International Commission for Plant Pollinator Relations",
        26787: "Technomed Holdings LLC",
        26788: "Escuela de Posgrado de la Universidad Nacional del Altiplano",
        26789: "Urban Development Scientific Publishing Pte Ltd",
        26790: "Sociedade Brasileira de Oncologia Clinica",
        26791: "e-Sarkiyat Ilmi Arastirmalari Dergisi",
        26792: "Associacao Brasileira de Estudos de Defesa - ABED",
        26793: "Yadav Publications",
        26794: "Financial University under the Government"
               "of the Russian Federation",
        26795: "Minin University",
        26796: "University of Energy and Natural Resources",
        26797: "Academia Nacional de Estudios Politicos y"
               "Estrategios (ANEPE)",
        26798: "STMIK AKAKOM Yogyakarta",
        26799: "Setcor Media FZ, LLC",
        26800: "Croatian Medical Association",
        26801: "Neumann Janos Szamitogep-tudomanyi Tarsasag",
        26802: "Sycamore Global Publications",
        26803: "Society for Research and Knowledge Management",
        26804: "Ausasia Science and Technology Press Pty, Ltd.",
        26805: "Asosiasi Peneliti Manajemen Adat Indonesia",
        26806: "Philosophical Faculty of University of Hradec Kralove",
        26807: "Pontifica Universidad Catolica del Ecuador",
        26808: "RS Publication",
        26809: "Journal of Awareness",
        26810: "Perio J",
        26811: "SCAD Independent",
        26812: "Brandeis University",
        26813: "Sonoma State University",
        26814: "Communications in Plant Sciences",
        26815: "The Korean Child Neurology Society",
        26816: "Academy of Asian Business (AAB)",
        26817: "Institucion Universitaria Colombo Americana - UNICA",
        26818: "OSU Press",
        26819: "Sokoto Journal of Social Sciences",
        26820: "Saberes de Conocimiento",
        26821: "IJournals Publications",
        26822: "lnternational Electronic Journal of Elementary Education",
        26823: "Revista NUFEN: Phenomenology and Interdisciplinarity",
        26824: "Latin American Literary Review",
        26825: "University of Pitesti",
        26826: "La Trobe University",
        26827: "International Association of Applied Neuroscience",
        26828: "Research Society on Marijuana",
        26829: "Ghanshyam Binani Academy of Management Sciences",
        26830: "Symmetry: Culture and Science",
        26831: "Research Excellence Institute",
        26832: "Agriculture and Environmental Science Academy",
        26833: "International Journal of Engineering and Geoscience",
        26834: "Korean Society for Critical Inquiry of Childhood Education",
        26835: "Kocaeli Universitesi Mimarlik ve Yasam Dergisi",
        26836: "International Journal of Advanced Scientific"
               "Research and Development",
        26837: "TWMS Journal of Applied and Engineering Mathematics",
        26838: "Medico Edge Publication",
        26839: "Institute of Philosophical Studies",
        26840: "Korean Association of Island",
        26841: "African Virtual University",
        26842: "Iraq Natural History Research Center and Museum",
        26843: "Cruzeiro do Sul Educacional",
        26844: "The Korean Society of Elementary Physical Education",
        26845: "Korean Securities Association",
        26846: "Amity Directorate of Management and Allied Areas (ADMAA)",
        26847: "The Korean Association of Governance Studies",
        26848: "Revista Brasileira de Geografia Fisica",
        26849: "Boletim Tecnico do Senac/Senac Journal of Education and Work",
        26850: "Ecletica Quimica Journal",
        26851: "Revista Uruguaya de Ciencia Politica",
        26852: "Fundacion Universitaria Sanitas",
        26853: "Revista Fatec Zona Sul",
        26854: "Europe Culture Arts Association",
        26855: "Hill Publishing Group Inc.",
        26856: "Korean Academy of Management",
        26857: "Institute of Distance Education",
        26858: "Universitas Negeri Makassar",
        26859: "Barkat Ali Firaq Trust for Education and Research",
        26860: "National Society for Integration of Applied-Basic"
               "Medical Sciences",
        26861: "The Dance Society for Documentation and History",
        26862: "Korean Physical Therapy Science",
        26863: "The Korean Association of Regional Geographers",
        26864: "Psicologia, Conocimiento y Sociedad",
        26866: "Korean Institute of Electromagnetic Engineering and Science",
        26867: "Faculty of Economic Engineering",
        26868: "International Building Performance Simulation Association",
        26869: "Internet2",
        26870: "Kina Institute for International Knowledge Network,",
        26871: "Revistas Killkana",
        26872: "CEMADES",
        26873: "Veterinary Faculty, University of Ljubljana",
        26874: "Universitas Jenderal Achmad Yani",
        26875: "Association of Doctors for Holistic Care",
        26876: "Udako Euskal Unibertsitatea",
        26877: "Universitas PGRI Semarang",
        26878: "iGlobal Educational Services",
        26879: "Coquina Press",
        26880: "Ilmiah.ID",
        26881: "Uniwersytet Gdanski",
        26882: "Historia Agraria",
        26883: "Education and Technologies Association",
        26884: "Pidvodni Tehnologii",
        26885: "Universidad Iberoamericana",
        26886: "TK Meganom LLC",
        26887: "Institut Seni Indonesia Padang Panjang",
        26888: "GeoMetodika",
        26889: "Oxford Institute for Energy Studies",
        26890: "University of Dayton Libraries",
        26891: "Fakultas Kedokteran Universitas Riau",
        26892: "Noctua- Arqueologia e Patrimonio",
        26893: "Revista Mosaicum",
        26894: "Koreanisch-Deutsche Gesellschaft Fuer Erziehungswissenschaft",
        26895: "Revista Geosaberes",
        26896: "TEST-ZL Publishing",
        26897: "Russian State Agrarian University - Moscow"
               "Timiryazev Agricultural Academy",
        26898: "SFSCA RAS",
        26899: "The Journal of International Civilization Studies",
        26900: "Journal of Scientific Perspectives",
        26901: "Cafet Innova Technical Society",
        26902: "NIIC SB RAS",
        26903: "College of Medicine, Mustansiriya University",
        26904: "Research Publishing International",
        26905: "Universitas Merdeka Malang",
        26906: "Poltava National Technical Yuri Kondratyuk University",
        26907: "Kazan Federal University",
        26908: "Sociological Viewpoints",
        26909: "SE Engineering Center Drying Institute of"
               "Engineering Thermophysics",
        26910: "Universidad Tecnica de Babahoyo",
        26911: "Masters Program in Public Health, Sebelas Maret University",
        26912: "Scientific Dental Journal",
        26913: "The Centre for Philosophical Research",
        28914: "AEDEAN"
               " (Asociacion Espanola de Estudios Anglo-Norteamericanos)",
        28915: "Zinianz Publishing Pvt Ltd",
        28916: "Biome Scientia Sdn Bhd",
        28917: "Instituto Social del MERCOSUR",
        28918: "LP2M IAIN Pekalongan",
        28919: "SCIK Publishing Corporation",
        28920: "Diving and Hyperbaric Medicine Journal",
        28921: "Newredmars Education Pvt, Ltd.",
        28922: "Al-Qadisyiah Medical Journal",
        28923: "World Federation of Societies of Anaesthesiologists",
        28924: "SCIK Publishing Corporation",
        28925: "Borys Grinchenko Kyiv University",
        28926: "Universitas Nahdlatul Ulama Blitar",
        28927: "ABMS - Brazilian Association for Soil Mechanics"
               "and Geotechnical Engineering",
        28928: "Iztapalapa - Revista de Ciencias Sociales y Humanidades",
        28929: "The Duke Clinical Research Institute",
        28930: "Department of Marine Science and Technology",
        28931: "Centros de Integracion Juvenil, A.C.",
        28932: "Maranatha Christian University",
        28933: "eSciPub LLC",
        28934: "Institute of Economic Sciences",
        28935: "Dragonwell, LLC",
        28936: "Iraqi Journal of Market Research and Consumer Protection",
        28937: "Felix Meiner Verlag GmbH",
        28938: "Mattering Press",
        28939: "Institucio Alfons el Magnanim",
        28940: "Sociedad Mexicana de la Ciencia del Suelo A.C.",
        28941: "Weed Science Society of Pakistan",
        28942: "Uptodate In Medicine LLC",
        28943: "Lebanon Support",
        28944: "Institut Dirosat Islamiyah Al-Amien Prenduan",
        28945: "Informing Science Institute",
        28946: "Universitas Sriwijaya",
        28947: "Sociedad para el Estudio y Conservación de"
               "las Aves en México A.C.",
        28948: "Omer Halisdemir Universitesi Muhendislik Bilimleri Dergisi",
        28949: "Bilimname",
        28950: "Revista FOCO",
        28951: "Revista Brasileira de Biometria",
        28952: "Arquivos Brasileiros de Cirurgia Digestiva Express",
        28953: "Case Western Reserve University",
        28954: "Association Science2 (Science Squared)",
        28955: "Alinteri Zirai Bilimler Dergisi",
        28956: "Guvenlik Bilimleri Dergisi",
        28957: "Asociacion Colombiana de Medicina Fisica y Rehabilitacion",
        28959: "Balai Riset dan Standardisasi Industri Palembang",
        28960: "Universidad Autonoma de Sinaloa",
        28961: "Kursor",
        28962: "eccElearning",
        28963: "Everything Is Connected Press",
        28964: "Publishing Press",
        28965: "Cultura y Representaciones Sociales",
        28966: "Pacific States Marine Fisheries Commission",
        28967: "ZygoScient Research Insights Inc.",
        28968: "Catalyst Journal",
        28969: "Iraqi Journal of Embryos and Infertility Researches",
        28970: "Universitaria Agustiniana",
        28971: "James P. Adams Library, Rhode Island College",
        28972: "Korean Elementary Counselor Education Association",
        28973: "Paleomedicina Hungary",
        28974: "Idojaras",
        28975: "Jeonbuk Historical Association",
        28976: "Revista Brasileira de Pesquisa em Educacao em Ciencia",
        28977: "The Institute for Biblical Text Research of"
               "the Korean Bible Society",
        28978: "Natural and Engineering Sciences",
        28979: "Journal of Advanced Research in Natural and Applied Sciences",
        28980: "The Korean Journal of Educational Methodology Studies",
        28981: "Hikmet Akademik Edebiyat Dergisi",
        28982: "Journal of Surgery and Medicine",
        28983: "FSBEU HE Saratov SAU",
        28984: "Laurentian University Library",
        28985: "Journal of Science and Cycling",
        28986: "Audit Board of the Republic of Indonesia",
        28987: "International Academy for Research in Learning Disabilities",
        28988: "STAIN Parepare",
        28989: "SENATIK",
        28990: "Journal of Applied Instructional Design",
        28991: "Ital Publication",
        28992: "Universitas Pasundan Kampus 2",
        28993: "IMI LLC",
        28994: "Formifarma, LDA",
        28995: "Russian State University for the Humanities",
        28996: "Russian Dialysis Society",
        28997: "Revista Universitaria de la Educacion Fisica y el Deporte",
        28998: "Universidade Federal de Alogoas",
        28999: "The Pipeline Transport Institute",
        29000: "RumeliDE Dil ve Edebiyat Arastirmalari Dergisi",
        29001: "Cardiology Research Institute",
        29002: "Aksaray University Journal of Science and Engineering",
        29003: "LLC MAKS Press",
        29004: "The Association for Korean Modern and Contemporary History",
        29005: "Indian Society of Advancement of Canine Practice",
        29006: "P.P. Shirshov Institute of Oceanology, RAS",
        29007: "EasyChair",
        29008: "European Turbomachinery Society",
        29009: "International Foundation of the Future Horizons OU",
        29010: "Industrial Technologies, LTD",
        29011: "Gavin Publishers",
        29012: "Journal of Privacy and Confidentiality",
        29013: "Premier Publishing s.r.o.",
        29014: "Lithuanian Neurological Association",
        29015: "WSB University in Wroclaw Publication Department",
        29016: "Standard Analytics IO",
        29017: "LEMIGAS RD Center for Oil and Gas Technology",
        29018: "Centro de Investigacion y Desarrollo Profesional - CIDEPRO",
        29019: "Universidad Tecnologica Equinoccial",
        29020: "New York Business Global LLC",
        29021: "Portuguese Society of Dermatology and Venereology",
        29022: "Cadernos de Historia UFPE",
        29023: "Alanya Akademik Bakis",
        29024: "Levy Library Press",
        29025: "North-Ossetian State University named after"
               "Costa Levanovich Khetagurov",
        29026: "Opto-Electronic Advances",
        29027: "Mass Research Publication Pvt Ltd",
        29028: "National Gugak Center",
        29029: "Bingol Universitesi Sosyal Bilimler Enstitusu Dergisi",
        29030: "LC Publishing and Trading Dashkov Co.",
        29031: "Faculdade Aldete Maria Alves",
        29032: "Scholar Publications",
        29033: "Universidad Tecnica de Ambato",
        29034: "Dialectical Publishing",
        29035: "Universidad Catolica del Maule",
        29036: "University College of Business in Prague",
        29037: "UGM Press",
        29038: "Lesya Ukrainka Volyn National University",
        29039: "RIOR Publishing Center",
        29040: "STIE AAS Surakarta",
        29041: "Micropaleontology Press",
        29042: "BioAxis DNA Research Centre",
        29043: "Centro de Estudios Superiores de Mexico y"
               "Centroamerica (CESMECA)",
        29044: "International Centre for Transactional Analysis"
               "Qualifications",
        29045: "Class Publishing",
        29046: "Thomas Jefferson University",
        29047: "Instituto Colombiano del Petroleo",
        29048: "Mehmet Akif Ersoy Universitesi Fen Bilimleri"
               "Enstitusu Dergisi",
        29049: "The Research Journal of the Costume Culture",
        29050: "Harran Tarim ve Gida Dergisi",
        29051: "Revista EntreLinguas",
        29052: "Advance Educational Institute and Research Center",
        29053: "Pretoria University Law Press",
        29054: "Faisalabad Medical University",
        29055: "Journal of Chemistry and Chemical Sciences",
        29056: "Next-Generation Convergence Information Service"
               "Technology Society",
        29057: "Universidad Autonoma del Estado de Hidalgo",
        29058: "Medical Journal of Western Black Sea",
        29059: "Universidad Autonoma de Tamaulipas",
        29060: "Yong Loo Lin School of Medicine",
        29061: "JERAM : Journal of Education Research in Administration"
               "and Management",
        29062: "STIE Mahardika",
        29063: "African Journal of Reproductive Health (AJRH)",
        29064: "International Journal of Management and Administration",
        29065: "Usak Universitesi Egitim Arastirmalari Dergisi",
        29066: "RENOVA - Revista Brasileira de Ciencia, Conhecimento"
               "e Inovacao",
        29067: "Muhasebe ve Vergi Uygulamalari Dergisi",
        29068: "Honvedorvos",
        29069: "ForScience: Revista Cientifica do IFMG",
        29070: "Ignited Minds Pvt. Ltd.",
        29071: "Mathematics Education UHAMKA",
        29072: "Basrah Journal of Science",
        29073: "Ponteditora",
        29074: "American Society for Clinical Laboratory Science",
        29075: "Philadelphia Museum of Art",
        29076: "Universidad Estatal de Milagro (UNEMI)",
        29077: "Real Sociedad Espanola de Historia Natural",
        29078: "Procesos: Revista Ecuatoriana de Historia",
        29079: "College of Veterinary Medicine",
        29080: "Fakultas Sains dan Teknologi UINSA",
        29081: "Vasile Alecsandri University of Bacau",
        29082: "LPPM Stikes Bina Sehat PPNI Mojokerto",
        29083: "Hadronic Press, Inc.",
        29084: "Stikes Jenderal Achmad Yani Cimahi",
        29085: "Facet Publishing",
        29086: "University of KwaZulu-Natal",
        29087: "Colorado Library Consortium (CLiC)",
        29088: "International Center for Intelligent Research",
        29089: "Collegium Medicum, University of Warmia and Mazury",
        29090: "Faculty of Pharmacy, Mahidol University",
        29091: "Dr. Ludwig Reichert Verlag",
        29092: "Universidad Autonoma de la Ciudad de Mexico",
        29093: "International Journal of MC Square Scientific Research",
        29094: "Sociedad Iberica de Ictiologia",
        29095: "Korean Criminological Association",
        29096: "Korean Society for the Study of Elementary Education",
        29097: "Fundacion Universidad de America",
        29098: "Critical Romani Studies",
        29099: "STMIK Dharma Wacana",
        29100: "STKIP PGRI Tulungagung",
        29101: "Univ. Autonoma del Estado de Mexico- Convergencia"
               "Revista de Ciencias Sociales",
        29102: "Clinical Health Promotion Centre",
        29103: "LPPM Universitas Malikussaleh",
        29104: "UNESCO-IHP",
        29105: "Universidad Autonoma de Nuevo Leon",
        29106: "Finans Ekonomi ve Sosyal Arastirmalar Dergisi",
        29107: "Polish Rhetoric Society",
        29108: "Ege Universitesi Ziraat Fakultesi Dergisi",
        29109: "Gazi Universitesi Fen Bilimleri Dergisi Part"
               "C: Tasarim ve Teknoloji",
        29110: "SOYLEM Filoloji Dergisi",
        29111: "MNK Publications",
        29112: "Revista Uruguaya de Antropologia y Etnografia",
        29113: "The Society of Korean Practical Arts Education Research",
        29114: "Technical University of Varna",
        29115: "Survey Practice",
        29116: "Gerundium",
        29117: "Qatar University",
        29118: "Indonesian Petroleum Association",
        29119: "Politechnika Slaska - Silesian University of Technology",
        29120: "Indian Journal of Psychiatric Social Work",
        29121: "Granthaalayah Publications and Printers",
        29122: "Badan Pengkajian dan Penerapan Teknologi (BPPT)",
        29123: "Komisi Yudisial",
        29124: "Al Kut Journal of Economic and Administrative Sciences",
        29125: "International Journal of Electroactive Materials",
        29126: "ISAR - International Scientific and Research Journals",
        29127: "Facultad de Ciencias Sociales, UCC",
        29128: "Geomatik",
        29129: "Inonu Universitesi Egitim Bilimleri Enstitusu Dergisi",
        29130: "Duzce Universitesi Bilim ve Teknoloji Dergisi",
        29131: "Uluslararasi Iktisadi ve Idari Bilimler Dergisi",
        29132: "International Journal of Pure and Applied Sciences",
        29133: "Yuzuncu Yil Universitesi Tarim Bilimleri Dergisi",
        29134: "International Journal of Thermodynamics",
        29135: "Sanat Tarihi Dergisi",
        29136: "Mediterranean Agricultural Sciences",
        29137: "Uluslararasi Muhendislik Arastirma ve Gelistirme Dergisi",
        29138: "Narotama University",
        29139: "Asian International Journal of Social Sciences",
        29140: "Castledown Publishers",
        29141: "Ural State University of Economics",
        29142: "Editora TantaTinta",
        29143: "Institute for Future Growth",
        29144: "The Korean Society of Cognitive Therapeutic Exercise",
        29145: "University of Management and Technology",
        29146: "Revista ECO-Pos",
        29147: "Universidade Anhembi Morumbi",
        29148: "Revista Labor",
        29149: "Marketing and Tourism Review",
        29150: "Journal of Hyperspectral Remote Sensing",
        29151: "Fundacion Universitaria Monserrate - Unimonserrate",
        29152: "Korea Observer - Institute of Korean Studies",
        29153: "The Association of Korean Civil Judgement Enforcement Law",
        29154: "Seoul National University Institute for Japanese Studies",
        29155: "Veterinaria - Montevideo",
        29156: "InterCambios",
        29157: "Erzurum Teknik Universitesi Sosyal Bilimler"
               "Enstitusu Dergisi",
        29158: "American Academy of Psychiatry and the Law",
        29159: "The Korean Association of Area Studies",
        29160: "A D Publication",
        29161: "PharmaTutor Edu Labs",
        29162: "Faculty of Humanities and Social Sciences Osijek",
        29163: "Modern English Drama Association of Korea",
        29164: "Cambridge Encyclopedia of Anthropology",
        29165: "SAFE Network",
        29166: "Facultad de Odontologia",
        29167: "Natural Science Simulations and Engineering Laboratory",
        29168: "Project Management Institute",
        29169: "Set Publishers",
        29170: "Facultad de Ingenieria, UNAM",
        29171: "Afghanistan Centre at Kabul University",
        29172: "American Physical Society (APS)",
        29173: "University of Alberta Libraries",
        29174: "The Korean Association of Central Asian Studies",
        29175: "Korean Law-Related Education Association",
        29176: "Asociacion Colombiana de Dermatologia y Cirugia"
               "Dermatologica",
        29177: "Logisztika — Informatika — Menedzsment",
        29178: "Nevtani Ertesito",
        29179: "Egeszsegtudomany",
        29180: "Budapesti Gazdasagi Egyetem",
        29181: "Motricidades Revista da Sociedade de Pesquisa"
               "Qualitativa em Motricidade Humana",
        29182: "Historia Economica e Historia de Empresas",
        29183: "Mix Sustentavel",
        29184: "Revista Cientifica da Faculdade de Medicina de Campos",
        29185: "Hayvansal Uretim",
        29186: "The Korean Society of the History of Historiography",
        29187: "Journal of Sustainable Construction Materials"
               "and Technologies",
        29188: "UroMedia LLC",
        29189: "Korean Accounting Information Association",
        29190: "The Society of Ewha Korean Language and Literature",
        29191: "International Journal of Advances in Interdisciplinary"
               "Research",
        29192: "Cuadernos del CLAEH",
        29193: "Revista Medica del Uruguay",
        29194: "Al-Nahrain Journal for Engineering Sciences",
        29195: "Indian Academy of Sciences",
        29196: "University of Babylon",
        29197: "Pontifica Universidad Catolica Madre y Maestra",
        29198: "UK EPC Ltd",
        29199: "Northern California Open Access Publications",
        29200: "American Chemical Society (ACS)",
        29201: "Coordinacion de Publicaciones de la Escuela"
               "Superior de Economia",
        29202: "International Society of Philosophy and Cosmology",
        29203: "Finnish Mycological Society",
        29204: "Journal of Postharvest Technology",
        29205: "College of Medicine, University of Basrah",
        29206: "Jurnal Saintis",
        29207: "Ikatan Ahli Informatika Indonesia (IAII)",
        29208: "Ikatan Apoteker Indonesia - Sumatera Barat",
        29209: "Indonesia Prime",
        29210: "Indonesian Institute for Counseling, Education"
               "and Therapy (IICET)",
        29211: "Research Institute for Language and Information",
        29212: "Military History Institute, MND",
        29213: "Geumgang Center for Buddhist Studies",
        29214: "Daehan Academy of Management Information Systems",
        29215: "Pesquisa e Ensino em Ciecias Exatas e da Natureza",
        29216: "Uluslararasi Ekonomi Isletme ve Politika Dergisi",
        29217: "Usak University Journal of Social Sciences",
        29218: "Shri Ram Mutrti Smarak College of Engineering Technology",
        29219: "SNF Swedish Nutrition Foundation",
        29220: "Communications for Statistical Applications and Methods",
        29221: "The Journal of Curriculum and Evaluation",
        29222: "Oil and Gas Research Institute of the RAS",
        29223: "Associacao Instituto Tecnologico Vale",
        29224: "IInsan ve Insan Dergisi",
        29225: "The Korean Tea Society",
        29226: "Journal of Business Management and Economic Research",
        29227: "Polish Mineral Engineering Society",
        29228: "ASOS Yayinevi",
        29229: "V.I.Romanovskiy Institute of Mathematics",
        29230: "Program Studi Akuntansi FE-UST",
        29231: "Program Studi Manajemen FE-UST",
        29232: "Scientific Investigation in Dentistry",
        29233: "Suleyman Demirel Universitesi Fen Edebiyat"
               "Fakultesi Fen Dergisi",
        29234: "INNOPRO",
        29235: "Publishing House Belorusskaya Nauka",
        29236: "Asociacion Colombiana Ingenieros de Sistemas",
        29237: "Revista Educacao em Saude",
        29238: "Poltekkes Kemenkes Yogyakarta",
        29239: "Sekolah Tinggi Ilmu Pertanian Wuna",
        29240: "STAIN Curup",
        29241: "STIKES Yayasan RS.Dr. Soetomo Surabaya",
        29242: "Association of Research Libraries",
        29243: "Departement of Forest Resource Conservation and Ecotourism",
        29244: "Institut Pertanian Bogor",
        29245: "Sciaccess Publishers LLC",
        29246: "Revista Gestao, Inovacao e Negocios",
        29247: "Cientifica - Multidisciplinary Journal",
        29248: "Revista Juridica",
        29249: "Selcuk Universitesi Sosyal Bilimler Meslek"
               "Yuksekokulu Dergisi",
        29250: "Sinirsiz Egitim ve Arastirma Dergisi",
        29251: "International Journal of Perceptions in Public Health",
        29252: "Armenian Green Publishing Co.",
        29253: "JP Global Transtech",
        29254: "Ukrainian Medical Stomatological Academy",
        29255: "Balai Bahasa Bali",
        29256: "European Scientific Center Biological Marker s.r.o.",
        29257: "Fundacion San Pablo Andalucia CEU",
        29258: "Kazakh-German University",
        29259: "Faculty of Economics, Universitas Sriwijaya",
        29260: "Instituto Chileno Terapia Familiar",
        29261: "Pakistan Veterinary Journal",
        29262: "Colegio Mexicano de Inmunologia Clinica y Alergia, A. C.",
        29263: "Universidad de los Andes",
        29264: "Faculty of Economics and Business, Mulawarman University",
        29265: "CIDE, Revista Gestion y Politica Publica",
        29266: "Duke University Libraries",
        29267: "Mexican Journal of Biotechnology",
        29268: "Services Society",
        29269: "Inno Publisher",
        29270: "Greenwich University Pakistan",
        29271: "College of Physicians and Surgeons Pakistan",
        29272: "Techna Group srl",
        29273: "Journal of KEMA",
        29274: "Institute of East and West Studies",
        29275: "The Korean Society of Educational Studies in Mathematics",
        29276: "Revista EDAPECI",
        29277: "Revista Uruguaya de Cardiologia",
        29278: "Akademik Ziraat Dergisi",
        29279: "Korean Society of Technical Education and Training",
        29280: "Revista Amazonida",
        29281: "Revista Decifrar",
        29282: "Revista da Escola Superior de Direito Municipal",
        29283: "Materials for Future",
        29284: "XLE Science",
        29285: "Acta Pinteriana",
        29286: "Revista de Educacao Publica",
        29287: "Raizes no Direito",
        29288: "Ilahiyat Tetkikleri Dergisi",
        29289: "Mastology",
        29290: "Avid Science",
        29291: "The Korean Society of Law and Medicine",
        29292: "Journal of Integrated Circuits and Systems",
        29293: "Centro de Educacao Superior de Guanambi (CESG)",
        29294: "International Journal of Advanced Science"
               "and Engineering(IJASE)",
        29295: "Kharkiv National University of Civil Engineering"
               "and Architecture",
        29296: "Russkiy Vrach Publishing House",
        29297: "ORBIT",
        29298: "Revista Mexicana de Ciencias Forestales",
        29299: "Kirsehir Egitim Fakultesi Dergisi",
        29300: "Institut Agama Islam Negeri (IAIN) Bengkulu",
        29301: "Sokoto Journal of Management Studies",
        29302: "1 Decembrie 1918 University of Alba Iulia",
        29303: "Universitas Mataram",
        29304: "Journal of Al-Qadisiyah for Computer Science and Mathematics",
        29305: "Korean Legal Center",
        29306: "Korean Science Education Society for the Gifted",
        29307: "Center for Music Education at Sam Houston State University",
        29308: "Southern Spaces",
        29309: "Independent Medical Trust",
        29310: "Motu Economic and Public Policy Research",
        29311: "University of Leicester",
        29312: "Revista Mexicana de Ciencias Agricolas",
        29313: "Universitas Islam Bandung (Unisba)",
        29314: "MLS Educational Research",
        29315: "Academia CUF",
        29316: "Pope John Paul II State School of Higher Education"
               "in Biala Podlaska",
        29317: "Physico-Technical Society of Kazakhstan",
        29318: "The Anam Society for the Study of Education",
        29319: "Minimally Invasive Surgical Oncology",
        29320: "Sri Jai Narain P.G. College, Lucknow",
        29321: "Madras Agricultural Students Union",
        29322: "International Journal of Scientific and Research"
               "Publications (IJSRP)",
        29323: "The Korean Association For Studies Of Modern Chinese History",
        29324: "The Korean Society Of East-West Comparative Literature",
        29325: "NIST/OSAC",
        29326: "FGI Federal Centre for Animal Health (FGI ARRIA)",
        29327: "Even3",
        29328: "Heighten Science Publications Corporation",
        29329: "Pen Academic Publishing",
        29330: "Society of Contemporary Art Science",
        29331: "Korea Research Association of International Commerce",
        29332: "Suryasa and Sons",
        29333: "Modestum Limited",
        29334: "HanKuk Munhwa Sa Hak Hoe",
        29335: "Institute of Agriculture Science and Technology",
        29336: "Biodimension",
        29337: "IJS Press",
        29338: "Federal Reserve Bank of Atlanta",
        29339: "Korean Society for Public Health and Medicine",
        29340: "Centro de Investigaciones y Estudios Superiores"
               "en Antropologia Social",
        29341: "Centre for Advanced Medieval Studies",
        29342: "Callosum Neurology",
        29343: "The Kuwait Society for the Advancement of Arab Children",
        29344: "Universidad Catolica Silva Henriquez",
        29345: "Future Visions Journal",
        29346: "Journal of Cultural and Critical Studies",
        29347: "International Journal of Exlibris",
        29348: "Journal of Electronic Science and Technology",
        29349: "The Association of Korean Cultural and Historical"
               "Geographers",
        29350: "Al-Qadisiyah Journal of Pure Science",
        29351: "Sociedad Mexicana de Historia de la Educacion AC",
        29352: "Instituto Politecnico de Viseu",
        29353: "Fondazione Museo Delle Antichita Egizie",
        29354: "Polskie Towarzystwa Diagnostyki Technicznej",
        29355: "Igdir Universitesi Ilahiyat Fakultesi",
        29356: "Sociedad Quimica de Mexico, A.C.",
        29357: "Euro-Asian Accrediting Association",
        29358: "Studies and Scientific Researches. Economics Edition",
        29359: "Gdansk University of Physical Education and Sport",
        29360: "Majalah BIAM",
        29361: "International Association of GeoChemistry (IAGC)",
        29362: "Institut za Savremenu Istoriju",
        29363: "Fundacio Scito",
        29364: "Czech Moravian Psychological Society",
        29365: "Journal of Psychology and Behavioral Sciences of the UACJS",
        29366: "ICC - The International Language Association",
        29367: "Revista SODEBRAS",
        29368: "Fortune Institute of International Business",
        29369: "Vishwashanti Multipurpose Society India",
        29370: "Research Gateway Society",
        29371: "TAMAP",
        29372: "Revista Agronomia Brasileira",
        29373: "Revista Sem Aspas",
        29374: "Revista Brasileira de Medicina Veterinaria",
        29375: "Universidad Autonoma de Bucaramanga",
        29376: "Parbeszed",
        29377: "Revista Brasileira de Estudos em Seguranca Publica",
        29378: "Plurais - Revista Disciplinar",
        29379: "Journal of e-Democracy and Open Government",
        29380: "Anais Educon UFS",
        29381: "Revista da Sociedade de Cardiologia do Estado de Sao Paulo",
        29382: "Earthquake Science",
        29383: "Revista Desafio Online",
        29384: "Revista Brasileira de Fisica Medica",
        29385: "Apeejay Journal of Management  Technical Campus",
        29386: "Revista Eclesiastica Brasileira",
        29387: "Chandannagar Medical Research Association (CMRA)",
        29388: "Navegando Publicacoes",
        29389: "Royal Association for Science-led Socio-cultural"
               "Advancement (RASSA)",
        29390: "Canadian Society of Respiratory Therapists",
        29391: "American Welding Society",
        29392: "Inishmore Laser Scientific Publishing Ltd",
        29393: "Universidad de Concepcion",
        29394: "INDTEC, C.A.",
        29395: "Neadi TIS - Nucleo de Educacao a Distancia"
               "e Tecnologia da Informacao em Saude",
        29396: "CPRM",
        29397: "Instituto de Comunicacao e Informacao Cientifica"
               "e Tecnologica em Saude",
        29398: "Korean Association on Social Quality",
        29399: "Turk Noropsikiyatri Dernegi",
        29400: "Bayt Publications",
        29401: "The Society of Korean Language Education",
        29402: "Ewha Research Institute of Nursing Science (ERINS)",
        29403: "Korean Society for Language and Information",
        29404: "Revista Trabalho, Politica e Sociedade",
        29405: "Uhamka Press",
        29406: "Universitas Muhammadiyah Pontianak",
        29407: "Universitas Nusantara PGRI Kediri",
        29408: "Universitas Hamzanwadi",
        29409: "Iraqi Journal of Cancer and Medical Genetics",
        29410: "Qartuppi",
        29411: "Nineteenth-Century Art Worldwide",
        29412: "Federal Reserve Bank of Boston",
        29413: "FSPSI SCFHHRP",
        30435: "Edra SpA",
        30439: "Studia Bobolanum, Pontifical Faculty of Theology"
               "in Warsaw Collegium Bobolanum",
        30442: "Medical and Dental Consultants Association"
               "of Nigeria, OOUTH Chapter",
        30445: "Asociacion Anestesia Reannimacion Espana, AnestesiaR",
        30457: "KAIROS - Greek Theological Association for"
               "the Improvement of the Religious Education",
        30464: "Politechnika Koszalinska",
        30515: "Our Language",
        30516: "Bilge International Journal of Science and"
               "Technology Research",
        30517: "Cihannuma Tarih ve Cografya Arastirmalari Dergisi",
        30518: "Journal of Aviation",
        30519: "Advances in Hospitality and Tourism Research",
        30520: "The Journal of Social Science",
        30521: "Journal of Energy Systems",
        30522: "Osmanli Bilim Arastirmalari",
        30523: "Mutefekkir",
        30524: "University of Belgrade - Faculty of Architecture",
        30525: "Publishing House Baltija Publishing",
        30526: "College of Education for Pure Sciences Ibn Al-Haitham",
        30527: "Society for Korean Literature in Classical Chinese",
        30528: "Association of Life-span Studies",
        30529: "The Aviation Management Society of Korea",
        30530: "Korean Association of Slavic Languages",
        30531: "Human Development Science",
        30532: "Torch Trinity Center for Islamic Studies",
        30533: "MIIGAIK",
        30534: "The World Academy of Research in Science and Engineering",
        30535: "Society for Music Theory",
        30536: "Indonesian National Institute of Aeronautics"
               "and Space (LAPAN)",
        30537: "Sukkur IBA University",
        30538: "Ptolemy Scientific Research Press",
        30539: "Baghdad University College of Veterinary Medicine",
        30540: "Kielce University of Technology",
        30541: "Pakistan Institute of Developemnt Economics (PIDE)",
        30542: "The Journal of Collegiate Emergency Medical Services",
        30543: "Academy of IRMBR",
        30544: "Association of Metallurgical Engineers of Serbia",
        30545: "Universidad Americana",
        30546: "Trade Unions Republican Commiittee of Azerbaijan"
               "Water Economy Workers",
        30547: "Faculty of Journalism, Lomonosov Moscow State University",
        30548: "Japan Society for Vascular Failure",
        30549: "Editorial Committee of the Swedish Institutes"
               "at Athens and Rome (ECSI)",
        30550: "Fundacion Miguel Lillo",
        30551: "International Journal of Spermatology",
        30552: "Asociacion Universitaria de Educacion y Psicologia",
        30553: "Associacao Portuguesa de Sociologia",
        30554: "Universidad de Manizales",
        30555: "Social Justice",
        30556: "Puslitbang tekMIRA",
        30557: "Progedit - Progetti Editoriali srl",
        30558: "Sun Publications",
        30559: "Badan Penelitian dan Pengembangan Provinsi Sumatera Barat",
        30560: "Ideas Spread",
        30561: "Sinop Universitesi Sosyal Bilimler Dergisi",
        30562: "The Journal of Multidisciplinary Research in Sustainability",
        30563: "Uluslararasi Turk Lehce Arastirmalari Dergisi",
        30564: "Bilingual Publishing Co.",
        30565: "Acta Medica Alanya",
        30566: "International Journal of Business Society",
        30567: "Publicaciones Latinoamericanas SRL",
        30568: "The Journal of Turkic Language and Literature Surveys",
        30569: "Adiyaman Universitesi Saglik Bilimleri Dergisi",
        30570: "The Journal of Political Philosophy and Sociology"
               "of Politics",
        30571: "College Administration and Economics",
        30572: "University of Kufa",
        30573: "King Abdullah Petroleum Studies and Research"
               "Center (KAPSARC)",
        30574: "GSC Online Press",
        30575: "Faculty of Islamic Education and Teachers Training",
        30576: "Global Journal of Oral Science",
        30577: "Biomedical Research Foundation",
        30578: "Universidad Central - Colombia",
        30579: "The Institute of Medical Science, Chosun University",
        30580: "Korea Digital Publishing Society",
        30581: "Korean Society of Chuna Manual Medicine Spine and Nerves",
        30582: "Korea Copyright Commission",
        30583: "Katonai Logisztika",
        30584: "Revista Brasileira de Inovacao em Saude",
        30585: "University of Dubai",
        30586: "Politik Ekonomik Kuram",
        30587: "Universitas Muhammadiyah Gresik",
        30588: "Universitas Proklamasi 45 Yogyakarta",
        30589: "Indonesian Association for Public Administration IAPA",
        30590: "LPPM STIKES Guna Bangsa Yogyakarta",
        30591: "Politeknik Harapan Bersama Tegal",
        30592: "Korean Academy of Financial Consumer",
        30593: "Student Counseling Center at Sogang University",
        30594: "The Society of Korean Seowon",
        30595: "Lembaga Publikasi Ilmiah dan Penerbitan Universitas"
               "Muhammadiyah Purwokerto",
        30596: "Universitas Muhammadiyah Sumatera Utara",
        30597: "Fakultas Kesehatan Masyarakat Universitas Hasanuddin",
        30598: "Universitas Pattimura",
        30599: "STKIP Nurul Huda",
        30600: "STMIK IBBI",
        30601: "Universitas Abulyatama",
        30602: "Poltekkes Kemenkes Pontianak",
        30603: "Institut Agama Islam Negeri Sultan Amai Gorontalo",
        30604: "STIKES Aisyah Pringsewu Lampung",
        30605: "Universitas Cokroaminoto Palopo",
        30606: "Universitas Pasir Pengaraian",
        30607: "Kocatepe Veteriner Dergisi",
        30608: "Hungarian Journal of English and American Studies (HJEAS)",
        30609: "Journal of Experimental Techniques and Instrumentation",
        30610: "Journal18",
        30611: "Revista Dialectus",
        30612: "Universidade Federal de Grande Dourados",
        30613: "Current Research in Social Sciences",
        30614: "World Ocean Development Institute",
        30615: "Korean Association for Japanese Thought",
        30616: "Anatolian Journal of Botany",
        30617: "Dialogues",
        30618: "Caleidoscopio - Edicao e Artes",
        30619: "Ukumela: the South African Journal of Legal"
               "Reasoning, Writing and Education",
        30620: "Revista Pontos de Interrogacao",
        30621: "Dokuz Eyul Universitesi Saglik Bilimleri Enstitusu",
        30622: "Turkish Academic Research Review",
        30623: "Harran Universitesi Ilahiyat Fakultesi Dergisi",
        30624: "Ob-Ugric Institute of Applied Researches and Development",
        30625: "International Journal of Contemporary Tourism Research",
        30626: "TESAM Akademi",
        30627: "Cukurova Universitesi Ilahiyat Fakultesi Dergisi",
        30628: "GITR Film and Television School",
        30629: "Medical Informational Agency Publishers",
        30630: "Politeknik Negeri Padang",
        30631: "Universitas Islam Negeri (UIN) Sultan Thaha Saifuddin Jambi",
        30632: "Society of Petrophysicists and Well Log Analysts (SPWLA)",
        30633: "LPPM Syedza Saintika",
        30634: "International Journal of Green Technology",
        30635: "Journal of Agriculture Food and Development",
        30636: "Center for Experimental and Behavioral Public Administration",
        30637: "AMUB/Revue Medicale de Bruxelles",
        30638: "OAIMDD - EcoZone Publishing House",
        30639: "Korea Association of the Law of Civil Procedure",
        30640: "Universitas 45 Surabaya",
        30641: "Badan Penelitian dan Pengembangan Hukum dan HAM",
        30642: "Ikatan Konselor Indonesia (IKI)",
        30643: "STIKES Hang Tuah Surabaya",
        30644: "Sekolah Tinggi Ilmu Kesehatan Harapan Ibu Jambi",
        30645: "STIKOM Tunas Bangsa Pematangsiantar",
        30646: "STMIK Sinar Nusantara Surakarta",
        30647: "Sekolah Tinggi Pariwisata Trisakti",
        30648: "Sekolah Tinggi Teologi Intheos Surakarta",
        30649: "Universitas Hang Tuah",
        30650: "Universitas Muhammadiyah Kalimantan Timur",
        30651: "Universitas Muhammadiyah Surabaya",
        30652: "Universitas Riau",
        30653: "Universitas Mathla ul Anwar Banten",
        30654: "Mathews International Publishers",
        30655: "Journal of Physical Education and Sports Studies",
        30656: "Universitas Serang Raya",
        30657: "Stowarzyszenie Menedzerow Jakosci i Produkcji",
        30658: "Nicholson School of Communication, UCF",
        30659: "Universitas Islam Sultan Agung",
        30660: "AFinLA-e: Soveltavan kielitieteen tutkimuksia",
        30661: "AFinLAn vuosikirja",
        30662: "Ajatus",
        30663: "Alue ja Ymparisto",
        30664: "Approaching Religion",
        30665: "AVAIN - Kirjallisuudentutkimuksen aikakauslehti",
        30666: "Elore",
        30667: "Historiska och litteraturhistoriska studier",
        30668: "Janus Sosiaalipolitiikan ja sosiaalityön tutkimuksen"
               "aikakauslehti",
        30669: "Kosmopolis",
        30670: "Kulttuurintutkimus",
        30671: "Nordia Geographical Publications",
        30672: "Nordic Journal of Surveying and Real Estate Research",
        30673: "Sananjalka",
        30674: "Scripta Instituti Donneriani Aboensis",
        30675: "Sosiaalipedagoginen aikakauskirja",
        30676: "Suomen Antropologi: Journal of the Finnish"
               "Anthropological Society",
        30677: "Terra",
        30678: "Tribologia - Finnish Journal of Tribology",
        30679: "FSBSI No. Caucasian FSC of Horticulture, Viticulture,"
               "and Wine Making",
        30680: "ECO Journal",
        30681: "Universidade do Estado do Mato Grosso - UNEMAT",
        30682: "Bononia University Press",
        30683: "Neoplasia Research",
        30684: "University of Technology",
        30685: "Turkish Journal of Marketing",
        30686: "Mining Industry Journal (Gornay Promishlennost)",
        30687: "Edizioni Ca Foscari",
        30688: "Australian and New Zealand Student Services Association Inc.",
        30689: "Nordregio",
        30690: "Exeliq Publishing",
        30691: "Revista Eletronica Ludus Scientiae",
        30692: "Stratejik ve Sosyal Arastirmalar Dergisi",
        30693: "Korean Institute of Smart Media",
        30694: "Scientific and Technological Communications LLC",
        30695: "Russian Mineralogical Society",
        30696: "International Journal of Computer Engineering"
               "and Applications",
        30697: "Results in Fixed Point Theory and Applications",
        30698: "Revista Electronica de Conocimientos, Saberes y Practicas",
        30699: "Farname, Inc.",
        30700: "STMIK Pontianak",
        30701: "Indonesian Journal of Cardiology",
        30702: "Professional Edition Eastern Europe",
        30703: "Cumhuriyet International Journal of Education",
        30704: "Journal of Istanbul Veterinary Sciences",
        30705: "Educacao Quimica en Punto de Vista",
        30706: "Uygulamali Yerbilimleri Dergisi",
        30707: "Illinois State University",
        30708: "Mantar Dergisi",
        30709: "Max Planck Institute for the Study of Crime,"
               "Security and Law",
        30710: "American Association of Legal Nurse Consultants",
        30711: "Uluslararasi Ticaret ve Ekonomi Arastirmalari Dergisi",
        30712: "Revista Guara",
        30713: "JSC VNIIOENG",
        30714: "Experimental Biomedical Research",
        30715: "Doxa Revista Brasileira de Psicologia e Educacao",
        30716: "University of Debrecen/Regiokutatas Szemle",
        30717: "Finno Ugric Languages and Linguistics",
        30718: "Politikatudomanyi szemle",
        30719: "The Korean Association of Women's Studies",
        30720: "Ergoterapi ve Rehabilitasyon Dergisi",
        30721: "University of Food Technologies - Plovdiv",
        30722: "The University of Sydney Library",
        30723: "Iraqi Journal of Physics",
        30724: "Kazan State Power Engineering University",
        30725: "St. Petersburg State University of Culture",
        30726: "Innovative Research Developers and Publishers",
        30727: "Humanist Publishing House",
        30728: "Bor Dergisi",
        30729: "Business Style Publishers LLC",
        30730: "Institute of Marine Geology and Geophysics"
               "of the Far East Branch of the RAS",
        30731: "International Journal of Chemical and Physical"
               "Sciences (IJCPS)",
        30732: "Chhattisgarh Swami Vivekanand Technical University",
        30733: "Selcuk Tip Dergisi",
        30734: "IKIP PGRI Bojonegoro",
        30735: "Institut Dirosat Islamiyah Al-Amien Prenduan Sumenep",
        30736: "Universitas Islam Lamongan",
        30737: "Universitas Kediri",
        30738: "Universitas Sarjanawiyata Tamansiswa",
        30739: "IAI Darussalam Blokagung",
        30740: "Journal Kopertis 4",
        30741: "STIE Widya Gama Lumajang",
        30742: "Universitas Wijaya Kusuma Surabaya",
        30743: "Universitas Islam Sumatera Utara",
        30744: "Brazilian Journal of Analytical Chemistry",
        30746: "KTH Royal Institute of Technology",
        30747: "Fundacja General Elzbiety Zawackiej",
        30748: "Ivan Kozhedub Kharkiv National Air Force University KNAFU",
        30749: "Centro Cultural Justica Federal",
        30750: "IJPBR",
        30751: "Korea Fashion and Costume Design Association",
        30752: "Nordisk judaistik/Scandinavian Jewish Studies",
        30753: "Management Research Center",
        30754: "Croatian Library Association",
        30755: "Faculty of Sciences, University of Novi Sad",
        30756: "American Interventional Headache Society Annals"
               "of Headache Medicine",
        30757: "Institute for Applied and Pure Mathematics (IMPA)",
        30758: "FSBI Arctic and Antarctic Research Institute (FSBI AARI)",
        30759: "Institute of History and Archaeology of Ural"
               "Branch of Russian Academy of Science",
        30760: "INAKOS International Association of Korean"
               "Literacy and Cultural Studies",
        30761: "The Korean Society for Eco Early Childhood"
               "Education and Care",
        30762: "STAIN Kediri",
        30763: "Intervencion, Revista Internacional de Conservacion,"
               "Restauracion y Museologia",
        30764: "Russian Federal Centre of Forensic Science"
               "of the Ministry of Justice (RFCFS)",
        30765: "Faculty of Engineering, University of Rijeka",
        30766: "FARC of the North-East named N.V. Rudnitskogo",
        30767: "Dil ve Edebiyat Arastirmalari Dergisi",
        30768: "Korea Dance Movement Psychotherapy Association",
        30769: "Ulusal Spor Bilimleri Dergisi",
        30770: "Federation of State Medical Boards",
        30771: "CTOR Press",
        30772: "Al-Qadisiyah Journal for Engineering Sciences (QJES)",
        30773: "Korean Neuropsychiatric Association",
        30774: "The Association of Yeongju Language and Literature",
        30775: "Korean Music Education Society",
        30776: "The Korean Economic and Business Association",
        30777: "Secondary Education Institute, Kyungpook National University",
        30778: "Universidad de los Andes",
        30779: "CMM",
        30780: "Institute of Technical and Scientific Research",
        30781: "Revista Estudos e Pesquisas em Administracao (Repad)",
        30782: "Uludag Universitesi Veteriner Fakultesi Dergisi",
        30783: "Nevsehir Haci Bektas Veli Universitesi SBE Dergisi",
        30784: "Ekonomi Politika ve  Finanas Arastirmalari  Dergisi",
        30785: "Mimarlik Bilimleri ve Uygulamalari Dergisi",
        30786: "Journal of Education and Future",
        30787: "STIKES Aisyiyah Surakarta",
        30788: "Escuela Colombiana de Rehabilitacion",
        30789: "Asociacion Colombiana de Neumologia y Cirugia de Torax",
        30790: "Magyar Nyelvjarasok",
        30791: "Intercontact Science",
        30792: "Institute for Mongolian, Buddhist and Tibetan"
               "Studies of the SBRAS",
        30793: "Universitatea Petru Maior Targu Mures",
        30794: "Pamukkale Universitesi Sosyal Bilimler Enstitusu Dergisi",
        30795: "Associacao Brasileira de Medicina e Cirurgia"
               "do Tornozelo e Pe (ABTPe)",
        30796: "Ali Nihat Gokyigit Egitim, Saglik, Kultur,"
               "Sanat ve Dogal Varliklari Koruma Vakfi (ANG Vakfi)",
        30797: "Bilimsel Madencilik Dergisi",
        30798: "Mehmet Akif Ersoy Universitesi Iktisadi ve"
               "Idari Bilimler Fakultesi Dergisi",
        30799: "JACS Directory",
        30800: "Interdisciplinary Journal of Philosophy, Law and Economics",
        30801: "Acta INFOLOGICA",
        30802: "American Association for Laboratory Animal Science",
        30803: "Adnan Menderes Universitesi Sosyal Bilimler"
               "Enstitusu Dergisi",
        30804: "Cesm-i Cihan Tarih Kultur ve Sanat Arastirmalari E-Dergisi",
        30805: "Korean Journal of Agricultural Management and Policy",
        30806: "Korean Association for Christian Scholarship",
        30807: "The Korean School Mathematics Society",
        30808: "Interregional Public Organization Glaucoma Society",
        30809: "Moscow Pharmaceutical Society",
        30810: "Revista Ciencias da Sociedade",
        30811: "Politeknik Negeri Lhokseumawe",
        30812: "STMIK Bumigora Mataram",
        30813: "Universitas Bunda Mulia",
        30814: "Universitas Islam Riau",
        30815: "Universitas Nahdlatul Ulama Sumatera Utara",
        30816: "Fundatia Echo",
        30817: "African Performance Review APR",
        30818: "Badan Penelitian dan Pengembangan SDM Kominfo",
        30819: "Logos Verlag Berlin",
        30820: "Psychosozial Verlag",
        30821: "MIQOT Jurnal Ilmu ilmu Keislaman",
        30822: "Universitas Katolik Widya Mandira",
        30823: "Revista de Estudios del Campo de Montiel",
        30824: "The International Society of Hypertension",
        30825: "Uniwersytet Przyrodniczy we Wroclawiu",
        30826: "Torus Press",
        30827: "Editorial de la Universidad de Granada",
        30828: "EARDA",
        30829: "UIN Sumatera Utara Medan",
        30830: "TOBIDER",
        30831: "Kuramsal Egitimbilim Dergisi",
        30832: "The Korean Society of Music Education Technology",
        30833: "The Korea Society for Legal Theory and Practice",
        30834: "Branch of Indian Psychiatric Society (Kerala)",
        30835: "Plant Production Institute nd. a V.Ya. Yuriev"
               "of NAAS of Ukraine",
        30836: "Institute of Geological Sciences of the National"
               "Academy of Sciences of Ukraine",
        30837: "Kharkiv National University of Radioelectronics",
        30838: "SHEE Prydniprovska State Academy of Civil"
               "Engineering and Architecture",
        30839: "Zaporizhzhia State Engineering Academy",
        30840: "Research Institute of Ukrainian Studies",
        30841: "Professional Event, LLC",
        30842: "Institute for Linguistic Studies Russian Academy of Sciences",
        30843: "New Zealand Plant Protection Society",
        30844: "GITO mbH Verlag",
        30845: "Center for Promoting Ideas",
        30846: "Asociacion Sin Fines de Lucro Epistemonikos",
        30847: "Global Synergetic Foundation",
        30848: "Pakistan Journal of Botany",
        30849: "Sociedad Interamericana de Psicologia",
        30850: "Vestnik of the Russian Agricultural Science",
        30851: "Aatseel",
        30852: "Asia-Pacific Network for Global Change Research",
        30853: "Gramota Publishers",
        30854: "Universidad Autonoma de Manizales",
        30855: "Gazi Publishing",
        30856: "Thai Association for the Study of the Liver",
        30857: "Kyiv National University of Technologies and Design",
        30858: "Instytut Ekonomiki Rolnictwa i Gospodarki"
               "Zywnosciowej Panstwowy Instytut Badawczy",
        30859: "Asociacion Meteorologica Espanola",
        30860: "Fundacion Letamendi Forns",
        30861: "BAR Publishing",
        30862: "Universitas Papua",
        30863: "IAIN BONE",
        30864: "STIKOM Bali",
        30865: "STMIK Budi Darma",
        30866: "ELTE Media es Kommunikacio",
        30867: "Politeknik Kesehatan Kemenkes Aceh",
        30868: "Al Hidayah Press",
        30869: "Politeknik Gorontalo",
        30870: "FKIP Universitas Sultan Ageng Tirtayasa",
        30871: "Politeknik Negeri Batam",
        30872: "Universitas Mulawarman",
        30873: "Institut Informatika dan Bisnis (IIB) Darmajaya",
        30874: "PT. Kresna Acitya Nusantara Mediatama",
        30875: "World Trade Organization",
        30876: "Journal of Harmonized Research Publications",
        30877: "Desousa Foundation",
        30878: "Universidad Autonoma del Estado de Mexico",
        30879: "XML Link (KAMJE)",
        30880: "Penerbit UTHM",
        30881: "Ology Journals Kft",
        30882: "Universidad Nacional de Cordoba",
        30883: "Balai Arkeologi Yogyakarta",
        30884: "Uchitel Publishing House",
        30885: "Korea Institution and Economics Association",
        30886: "SOBEST Associacao Brasileira de Estomaterapia",
        30887: "Korean Association of Maritime Police Science",
        30888: "Kuprienko SV",
        30889: "Yolnat PE",
        30890: "Sergeieva I",
        30891: "Rhine Research Centre",
        30892: "Asociatia de Geografie",
        30893: "APEQ Associacao Portuguesa para o Estudo do Quaternario",
        30894: "Russian Space Systems",
        30895: "SCEEMP",
        30896: "ANO Veterinary Medicine Journal",
        30897: "International Journal of Environment and Geoinformatics",
        30898: "Kotelnikov Institute of Radioengineering and"
               "Electronics of Russian Academy of Sciences",
        30899: "Revista Brasileira de Direitos Fundamentais e Justica",
        30900: "e-Kafkas Egitim Arastirmalari Dergisi",
        30901: "FSBSI FRC N.I. Vavilov All-Russian Institute"
               "of Plant Genetic Resources",
        30902: "Korea Appraisal Board",
        30903: "Balkan Arastirma Enstitusu Dergisi",
        30904: "Pharma Research Library",
        30905: "Devir Educacao",
        30906: "Folium Publishing Company",
        30907: "College of Political Sciences",
        30908: "Trade Analysis and Development Agency",
        30909: "Volcanica",
        30910: "Turk Tarim ve Doga Bilimleri Dergisi",
        30911: "Yu. A. Kosygin Institute of Tectonics and"
               "Geophysics, Far Eastern Branch, RAS",
        30912: "Korean Social Work Counseling Academy",
        30913: "Alinteri Sosyal Bilimler Dergisi",
        30914: "FSBEI HE Mari State University",
        30915: "Ankara Barosu Dergisi",
        30916: "Korean Educational Research Association",
        30917: "Agentstvo Tvortheskih Tekhnologiy",
        30918: "Net Journals",
        30919: "Engineered Science Publisher",
        30920: "Universidad Nacional Mayor de San Marcos",
        30921: "Geodezia es Kartografia",
        30922: "Agile Libre",
        30923: "Red.es",
        30924: "Faculty of Economics, University of Split",
        30925: "University of Rijeka Faculty of Law",
        30926: "ECNU Review of Education",
        30927: "International Journal of Public Finance",
        30928: "Revista de Relatos de Casos do ABC",
        30929: "Kremenchuk Mykhailo Ostohradskyi National University",
        30930: "INNOTEC Gestion",
        30931: "Journal of Engineering Technology and Applied Science",
        30932: "FSBEO HPE Moscow State University of Railway"
               "Engineering (MIIT)",
        30933: "Korean Public Land Law Association",
        30934: "Kocaeli Universitesi Saglik Bilimleri Dergisi",
        30935: "Bastas Publications",
        30936: "Institute of Civil Society",
        30937: "Olimpianos - Journal of Olympic Studies",
        30938: "Boletim Cearense de Educacao e Historia da"
               "Matematica - BOCEHM",
        30939: "International Journal of Automotive Science and Technology",
        30940: "Korean Association for Qualitative Inquiry",
        30941: "Electrical Engineering Press Co. Ltd.",
        30942: "Master of Social Work Programme",
        30943: "APD SKEG",
        30944: "Asociacion Colombiana de Cirugia",
        30945: "EDIURCAMP",
        30946: "Russian Association of Pediatric Surgery",
        30947: "Koreanische Gesellschaft fuer Deutsche Sprache Und Literatur",
        30948: "The Natural Science Institute Wonkwang University",
        30949: "Dialogue",
        30950: "UACES",
        30951: "Onjourn Publishers Pvt. Ltd.",
        30952: "Societe Algerienne de Nutrition",
        30953: "Partners in Digital Health",
        30954: "New Delhi Publishers",
        30955: "University of the Aegean",
        30956: "Missouri Academy of Science",
        30957: "Center of Language and Culture Studies",
        30958: "Athens Institute for Education and Research ATINER",
        30959: "Balai Pelestarian Nilai Budaya Jawa Barat",
        30960: "Korean Association for Conservation of Nature",
        30961: "Language Education Institute, Seoul National University",
        30962: "E-compos",
        30963: "Arachnologische Mitteilungen",
        30964: "Ankara Universitesi Egitim Bilimleri Fakultesi Dergisi",
        30965: "Brill Deutschland GmbH",
        30966: "NATO Strategic Communications Centre of Excellence",
        30967: "International Journal of Current Research"
               "in Science, Engineering & Technology (IJCRSET)",
        30968: "Revista Brasileira de Farmacia Hospitalar"
               "e Servicos de Saude",
        30969: "Agropecuaria Cientifica no Semiarido",
        30970: "Ivan Franko National University of Lviv",
        30971: "Allama Iqbal Open University",
        30972: "Universidad Nacional del Nordeste",
        30973: "Universidad Autonoma del Estado de Morelos",
        30974: "Korean Association of International Culture Exchange",
        30975: "ARSRIPPI",
        30976: "Sosyal Ekonomik Arastirmalar Dergisi",
        30977: "Kharkiv National Automobile and Highway University",
        30978: "Publishing Company VIT-A-POL",
        30979: "Associacao Brasileira de Ensino Odontologico ABENO",
        30980: "The Ornithological Society of Korea",
        30981: "NGES STD",
        30982: "PIA HE MIA",
        30983: "IAIN Bukittinggi",
        30984: "IAIN Manado",
        30985: "Indonesian Indovisi Institute",
        30986: "Humanistica Lovaniensia",
        30987: "Bryansk State Technical University BSTU",
        30988: "Politeknik APP Jakarta",
        30989: "Universitas Jenderal Achmad Yani Yogyakarta",
        30990: "Pusat Studi Pendidikan Bahasa dan Sastra Indonesia"
               "Universitas Indraprasta PGRI",
        30991: "SR Informatics",
        30992: "The Korean Association of Party Studies",
        30993: "STEI Tazkia",
        30994: "Institut Ilmu Kesehatan STRADA Indonesia (STRADA"
               "Indonesia Health Sciences Institute)",
        30995: "Sekolah Tinggi Teologi Pelita Bangsa Jakarta",
        30996: "Universitas 17 Agustus 1945 Surabaya",
        30997: "Universitas Djuanda",
        30998: "Universitas Indraprasta PGRI",
        30999: "Universitas Islam Nusantara",
        31000: "Universitas Muhammadiyah Tangerang",
        31001: "Universitas Setia Budi",
        31002: "Universitas Tidar",
        31003: "U.S. Pharmacopeial Convention",
        31004: "Universitas Pahlawan Tuanku Tambusai",
        31005: "Sociedade Regional de Ensino e Saude LTDA",
        31006: "Girisimcilik Inovasyon ve Pazarlama Arastirmalari Dergisi",
        31007: "Institute of Humanities Research KBSC RAS",
        31008: "Institute of Korean Democracy (IKD)",
        31009: "Universitat Pompeu Fabra",
        31010: "Empress Publishing, LLC",
        31011: "Revista Enfermagem Atual",
        31012: "Brazil Publishing",
        31013: "Peak Health Center",
        31014: "Asian Institute of Research",
        31015: "International Journal of Agriculture Environment"
               "and Food Sciences",
        31016: "Institute of Fundamental and Applied Parasitology"
               "of Animals and Plants",
        31017: "Asociacion Civil Herpetologica Argentina",
        31018: "ANSF Publications",
        31019: "Turkiye Biyolojik Mucadele Dergisi",
        31020: "Mersin Universitesi Tip Fakultesi Lokman Hekim"
               "Tip Tarihi ve Folklorik Tip Dergisi",
        31021: "Boffin Access Limited",
        31022: "A-R Editions",
        31023: "Faculty of Economics and Business, The University of Lampung",
        31024: "N.S. Memorial Scientific Research and Education Society",
        31025: "Eurowaste SRL",
        31026: "Journal of Engineering",
        31027: "Ontario Drug Policy Research Network (ODPRN)",
        31028: "Balai Penelitian dan Pengembangan Irigasi",
        31029: "Daghestan Federal Research Center of Russian"
               "Academy of Science",
        31030: "Beuth Verlag GmbH",
        31031: "Crimson Publishers",
        31032: "BPAS Publishers LLP",
        31033: "Vandana Publications",
        31034: "Slavica",
        31035: "China Geology",
        31036: "Korean Society for Adolescents and Filial Piety Culture",
        31037: "The Society of Korean Thought and Culture",
        31038: "Research Desk Inc",
        31039: "Paragon International University",
        31040: "FSBSI Ufa Federal Research Centre of the RAS",
        31041: "The Convergence Institute for Asian Multicultural Studies",
        31042: "Anestesia, Analgesia, Reanimacion",
        31043: "Russian Research Institute of Farm Animal"
               "Genetics and Breeding",
        31044: "Science & Technology",
        31045: "The University of Maine",
        31046: "American Theological Library Association",
        31047: "Agriscientia",
        31048: "Revista del Museo de Antropologaa",
        31049: "Anuario de la Escuela de Historia Virtual",
        31050: "Estudios Digital",
        31051: "Revista Argentina de Anatomia Clinica",
        31052: "Revista de Salud Publica",
        31053: "Revista de la Facultad de Ciencias Medicas",
        31054: "Revista de la Facultad de Derecho",
        31055: "Boletín de la Sociedad Argentina de Botánica",
        31056: "Etica & Cine",
        31057: "Antiguos Jesuitas en Iberoamerica",
        31058: "International Technology and Science Publications",
        31059: "Crimean Astrophysical Observatory",
        31060: "Revista Brasileira de Seguranca Publica",
        31061: "Revista de Desenvolvimento e Politicas Publicas - REDEPP",
        31062: "AGROMETEOROS",
        31063: "Ural Branch of the RAS",
        31064: "Koreanische Gesellschaft fur Germanistik",
        31065: "Art History Association of Korea",
        31066: "Hwarang Dae Research Institute",
        31067: "Acibadem Universitesi Saglik Bilimleri Dergisi",
        31068: "Revista Teoria e Pesquisa",
        31069: "Journal of Applied Pharmaceutical Sciences and Research",
        31070: "Minerva Science Magazine",
        31071: "Kharkiv Medical Academy of Postgraduate Education KHMAPO",
        31072: "Revista FAEMA",
        31073: "Publishing House of National Academy Agrarian"
               "Sciences of Ukraine",
        31074: "Gyermeknevelss",
        31075: "The Road Association of Serbia Via-vita",
        31076: "Weight Stigma Conference",
        31077: "Centre for Ageing Better",
        31078: "Constitutional Court of the Republic of Indonesia",
        31079: "FSBEI HE Far Eastern State Transport University (FESTU)",
        31080: "Acta Scientific Publications Pvt. Ltd.",
        31081: "State Enterprise Ukrainian State Research"
               "Institute for Carbochemistry UKHIN",
        31082: "Healthcare of Kazakhstan",
        31083: "IMR Press",
        31084: "IG UFRC RAS",
        31085: "Izdatelstvo Prospekt LLC",
        31086: "The Turkish Journal of Geriatrics",
        31087: "LLC VNIGNI-2",
        31088: "FSBSI Research Institute of Human Morphology",
        31089: "FSBI Research Institute of Occupational Health RAMS",
        31090: "Fakultas Ilmu Komputer Universitas Narotama",
        31091: "Institut Seni Indonesia Denpasar",
        31092: "Politeknik Keuangan Negara STAN",
        31093: "Program Studi di Luar Kampus Utama (PSDKU)"
               "UNAIR di Banyuwangi",
        31094: "National Academy of Education",
        31095: "Universidad Espíritu Santo",
        31096: "Willamette University",
        31097: "The Kangwon Historical Society",
        31098: "Research Synergy Foundation",
        31099: "Sekolah Tinggi Psikologi Yogyakarta",
        31100: "STKIP Andi Matappa",
        31101: "Universitas Aisyiyah (UNISA) Yogyakarta",
        31102: "Universitas Islam Madura",
        31103: "Fakultas Ekonomi dan Bisnis, Universitas Narotama",
        31104: "Akademi Sekretari dan Manajemen Taruna Bakti",
        31105: "Balai Besar Penelitian dan Pengembangan Pelayanan Kessos",
        31106: "Fakultas Ekonomi dan Bisnis Universitas Islam Malang",
        31107: "FSBI Scientific and Research Financial Institute",
        31108: "Psychological Journal",
        31109: "ZAGENO",
        31110: "Makarenko Sumy State Pedagogical University"
               "Physical and Mathematical Education",
        31111: "Komarov Botanical Institute of the Russian"
               "Academy of Sciences",
        31112: "ISTEC - Instituto Superior de Tecnologias Avancadas",
        31113: "STIA Lan Bandung",
        31114: "FSFIS Institute for Design Problems in Microelectronics RAS",
        31115: "Korean Academy of Dysphagia Rehabilitation",
        31116: "Institute of Cytology of the Russian Academy of Sciences",
        31117: "Neurotak Publishing",
        31118: "California Institute of Technology Library",
        31119: "International Foundation for the Support of"
               "Social Research and Education Intersozic",
        31120: "Katre, Uluslararasi Insan Arastirmalari Dergisi",
        31121: "Tefsir Arastirmalari Dergisi",
        31122: "SineFilozofi",
        31123: "Akdeniz Universitesi Iletisim Fakultesi Dergisi",
        31124: "SAGE Publications",
        31125: "Hacettepe Universitesi Hemsirelik Fakultesi Dergisi",
        31126: "Akra Kultur Sanat ve Edebiyat Dergisi",
        31127: "Turkish Journal of Engineering",
        31128: "The Royal Australian College of General Practitioners",
        31129: "LUMA Centre Finland",
        31130: "The University of Danang",
        31131: "Journal of Athletic Performance and Nutrition",
        31132: "Institute for African Studies of the Russian"
               "Academy of Science",
        31133: "Koreanische Hesse-Gesellschaft",
        31134: "Archivos de Pediatria del Uruguay",
        31135: "Korean Society of Construction Legal Affairs",
        31136: "Korean Cleft Lip and Palate Association",
        31137: "The Korea Association of Education Consulting and Coaching",
        31138: "PCO Convin",
        31139: "Exemplum Tomasz Adamski",
        31140: "Institute of Biology of Komi Scientific Centre"
               "of the Ural Branch of the Russian Academy of Sciences",
        31141: "The Faculty of Law in Split",
        31142: "South Asia Management Association",
        31143: "Kabardino-Balkarian State University",
        31144: "A.P. Ershov Institute of Informatics Systems SB RAS",
        31145: "The Fund European Center for Quality",
        31146: "LLC Global Media Technology",
        31147: "International Association of Language and Literature",
        31148: "Military Health Policy Research Center",
        31149: "Research Parks Publishing",
        31150: "Global Research Network LLC",
        31151: "The Bombay Technologists",
        31152: "Korea Business Association",
        31153: "Center for Research and Development of Standardization",
        31154: "Universitas Klabat",
        31155: "National Excessive Force Institute",
        31156: "Lund University Library",
        31157: "Instituto Nacional de Neurologia y Neurocirugia"
               "Manuel Velasco Suarez",
        31158: "Korean Society for Educational Evaluation",
        31159: "Korean Society of MR Technology",
        31160: "Journal of Transcatheter Interventions",
        31161: "Dagestan State Pedagogical University",
        31162: "Russian Islamic Institute",
        31163: "Ufa Institute of Biology UFRC RAS",
        31164: "Revista Ecuatoriana de Investigaciones Agropecuaria",
        31165: "Networking Knowledge MeCCSA PGN Journal",
        31166: "OOO Zhurnal Voprosy Istorii",
        31167: "Sociedad Espanola de Ciencias Forestales",
        31168: "Institute of Slavic Studies of the Russian"
               "Academy of Sciences",
        31169: "Mesh Publisher",
        31170: "SEO/BirdLife",
        31171: "Vlast",
        31172: "Pusat Penelitian dan Pengembangan Badan Meteorologi"
               "Klimatologi dan Geofisika",
        31173: "Borno Medical Journal",
        31174: "Science and Education A New Dimension",
        31175: "Beyond the Horizon International Strategic Studies Group",
        31176: "Apertura",
        31177: "The Society of Study for Korean Music Education",
        31178: "Universitate din Bucuresti (University of Bucharest)",
        31179: "Facultad de Arquitectura y Diseño",
        31180: "Educational and Scientific Institute",
        31181: "Regional Association for Security and Crisis Management",
        31182: "Cubic Society",
        31183: "Investigational Medicinal Chemistry and Pharmacology",
        31184: "Pemberley Books (Publishing)",
        31185: "Wasit University",
        31186: "Badan Penerbitan Fakultas Pertanian (BPFP)"
               "Universitas Bengkulu",
        31187: "Maayan Laboratory",
        31188: "University of Ottawa Press",
        31189: "Clinical Exercise Physiology Association",
        31190: "Revista Mexicana de Neurociencia",
        31191: "African Journal of Current Medical Research",
        31192: "Krscanski Akademski Krug",
        31193: "Social Sciences Academic Press (China)",
        31194: "Biosciences and Biotechnology Research Center"
               "Institut Teknologi Bandung",
        31195: "Eurasian Journal of Forest Science",
        31196: "Harran Universitesi Veteriner Fakultesi Dergisi",
        31197: "Advances in the Theory of Nonlinear Analysis"
               "and its Application",
        31198: "IDEALKENT",
        31199: "Hak Is Uluslararasi Emek ve Toplum Dergisi",
        31200: "Mehmet Akif Ersoy Universitesi Uygulamali Bilimler Dergisi",
        31201: "International Journal of Health Management and Tourism",
        31202: "El-Cezeri: Journal of Science and Engineering",
        31203: "Asia-Europe Perspective Association",
        31204: "Oxford Academic Studies Press",
        31205: "Research Centre for East European Studies"
               "at the University of Bremen",
        31206: "Revista Mexicana de Orientacion Educativa",
        31207: "Universidad de Los Hemisferios",
        31208: "Volga Region Research Institute of Manufacture"
               "and Processing of Meat-and-Milk Production",
        31209: "TSI Press",
        31210: "Poltava State Agrarian Academy",
        31211: "Instituto Superior Miguel Torga",
        31212: "Institut Za Noviju Istoriju Srbije",
        31213: "Konversi",
        31214: "IJTHFAG Journal",
        31215: "Ukrainian Pulmonology Journal",
        31216: "Institute of Brain-Based Education, Korea"
               "National University of Education",
        31217: "University of Rijeka, Faculty of Maritime Studies",
        31218: "The Historical Society Of Korea",
        31219: "OSF Preprints",
        31220: "AgriXiv",
        31221: "Arabixiv",
        31222: "BITSS",
        31223: "EarthArXiv",
        31224: "engrXiv",
        31225: "FocUS Archive",
        31226: "Frenxiv",
        31227: "INA-Rxiv",
        31228: "LawArXiv",
        31229: "LIS Scholarship Archive",
        31230: "MarXiv",
        31231: "MindRxiv",
        31232: "NutriXiv",
        31233: "PaleorXiv",
        31234: "PsyArXiv",
        31235: "SocArXiv",
        31236: "SportRxiv",
        31237: "Thesis Commons",
        31238: "Associacao Paulista do Ministerio Publico",
        31239: "Revista Latino-Americana De Arqueologia Historica",
        31240: "Scientia Canonica",
        31241: "GI KSC RAS",
        31242: "Academy of Sciences of the Republic of Sakha (Yakutia)",
        31243: "Universidad Tecnica de Ambato - Carrera de Medicina",
        31244: "Waxmann",
        31245: "Council for British Research in the Levant",
        31246: "Nutrition Society of Malaysia",
        31247: "Arbeitsgemeinschaft fur Notfallmedizin",
        31248: "Integrity Mega Research Publishers",
        31249: "INION RAS",
        31250: "Peter the Great Museum of Anthropology and"
               "Ethnography of the Russian Academy of Sciences"
               "(The Kunstkamera)",
        31251: "Institute of Soil Science and Agrochemistry"
               "of the Siberian Branch",
        31252: "Revista Portuguesa de Saude Ocupacional",
        31253: "Universitas Buddhi Dharma",
        31254: "BioMed Research Publishers",
        31255: "Siberian Institute of Plant Physiology and"
               "Biochemistry of the Siberian Branch of RAS",
        31256: "UK-Robotics and Autonomous Systems (RAS) Network",
        31257: "University of Kufa - Faculty of Science",
        31258: "Lembaga Penelitian dan Pengabdian kepada Masyarakat"
               "Universitas Riau",
        31259: "Stony Brook University",
        31260: "Fundacion Universitaria de Ciencias de la Salud",
        31261: "University of Silesia in Katowice",
        31262: "Trnava University",
        31263: "Association of Austrian Librarians",
        31264: "International Journal of New Economics, Public"
               "Administration and Law",
        31265: "Stavanger University Library",
        31266: "Balai Besar Tekstil Kementerian Perindustrian",
        31267: "Schwabe Verlag GmbH",
        31268: "Kancelaria Sejmu/Chancellery of the Sejm",
        31269: "Information Society Research",
        31270: "Kiban Research Publications",
        31271: "Center for Future Studies",
        31272: "Mustansiriyah University College of Engineering",
        31273: "University of Warwick",
        31274: "Iowa State University",
        31275: "Society for Scientific Exploration",
        31276: "Vietnam Journal of Science Technology and Engineering",
        31278: "Foreign Policy Research Foundation",
        31279: "Stavropol State Agrarian University",
        31280: "Institute of Korean Theological Information Network Service",
        31281: "Gruenwald Laboratories",
        31282: "Indonesian Orthopaedic Association",
        31283: "Surgery Multimedia",
        31284: "Institut Teknologi Adhi Tama Surabaya",
        31285: "Agrociencia (Uruguay)",
        31286: "Towarzystwo Milosnikow Jezyka Polskiego",
        31287: "Fogyatekossag es Tarsadalom",
        31288: "Ukrainian Society of Ophthalmologists",
        31289: "Universitas Medan Area",
        31290: "Poltekkes Kemenkes Malang",
        31291: "Puslitbang Lektur, Khazanah Keagamaan dan"
               "Manajemen Organisasi",
        31292: "Sekolah Tinggi Pertanahan Nasional",
        31293: "Universitas 17 Agustus 1945 Samarinda",
        31294: "Universitas Bina Sarana Informatika",
        31295: "Suryasa and Sons",
        31296: "Knowledge Enterprise Journals - Copernicus",
        31297: "Institute of Public Administration Zagreb",
        31298: "Hrvatsko Sumarsko Drustvo",
        31299: "Faculty of Education and Rehabilitation Sciences,"
               "University of Zagreb",
        31300: "Research Trends, Ltd.",
        31301: "Institute of Biochemistry and Genetrics, UFRC RAS",
        31302: "Koreanische Gesellschaft fur Deutsch als Fremdsprache",
        31303: "Korea Real Estate Academy",
        31304: "Biotechnological Research",
        31305: "Research Review Publisher",
        31306: "Zavod za Istrazivanje i Razvoj Sigurnosti",
        31307: "Korean Association Of Public Policy",
        31308: "Korea Society for Sport Management",
        31309: "The Association for the Study of the Korean"
               "History of Thought",
        31310: "Research Institute of the Humanities",
        31311: "Universitas BSI",
        31312: "Institute for Peace and Conflict Research",
        31313: "The Society of Korean Literary Criticism",
        31314: "Universitas Muhammadiyah Gorontalo",
        31315: "Universitas Pembangunan Nasional Veteran Yogyakarta",
        31316: "Universitas PGRI Yogyakarta",
        31317: "Universitas Tamansiswa Padang",
        31318: "Ukrainian National Tchaikovsky Academy of Music",
        31319: "Dniprovsk State Technical University",
        31320: "Korean Society of Computed Tomographic Technology",
        31321: "The Korean Society of Medicine & Therapy Science.pdf",
        31322: "Sungshin University Educational Research Institute",
        31323: "Humanities Research Institute Incheon National University",
        31324: "Korean Association Of Regional Studies",
        31325: "Rhetoric Society of Korea",
        31326: "Universitas Trilogi",
        31327: "USN Scientific Journal, Universitas Sembilanbelas"
               "November Kolaka",
        31328: "Badan Penerbitan Universitas Widyagama Malang",
        31329: "Society For Ming-Qing Historical Studies",
        31330: "Balai Penelitian dab Pengembangan Agama Jakarta",
        31331: "IKIP Veteran Semarang",
        31332: "Institut Agama Islam Negeri Kendari",
        31333: "Korean Institute of Hazardous Materials",
        31334: "Institut Ilmu Sosial dan Manajemen STIAMI",
        31335: "The Historical Society of Yeoksa Silhak",
        31336: "The Korea Academic Society Of Tourism And Leisure",
        31337: "Institute of Philosophy and Theology of the Society of Jesus",
        31338: "University of Warsaw - Warsaw University Press",
        31339: "EPD Mukachevo State University",
        31340: "Institute of Biochemistry and Genetics - Subdivision"
               "of the Ufa Federal Research Centre of the Russian"
               "Academy of Sciences",
        31341: "Faculty of Organisation and Informatics",
        31342: "Korea Association for Drama/Theatre and Education",
        31343: "Vernadsky State Geological Museum RAS",
        31344: "International Journal of Human and Health Sciences IJHHS",
        31345: "Fire Investigation Society of Korea",
        31346: "Balai Besar Pengembangan SDM dan Penelitian"
               "Komunikasi dan Informatika Medan",
        31347: "The Institute of Dokdo in Yeungnam University",
        31348: "Czech Society of Ophthalmology of the CzMA",
        31349: "Sociedad Mexicana de Fisica A C",
        31350: "Mavipuro Polska Sp. z o.o.",
        31351: "College of Pharmacy University of Baghdad",
        31352: "The Education Research Institute",
        31353: "Landscape Architecture Foundation",
        31354: "Global Clinical Engineering Journal",
        31355: "International Network for Digital Innovation,"
               "Research & Education",
        31356: "University of North Dakota",
        31357: "University of Sri Jayewardenepura",
        31358: "Fakultas Teknik Elektronika dan Komputer Universitas"
               "Kristen Satya Wacana",
        31359: "Pravo Publishing House",
        31360: "FSBSI RRIFSC",
        31361: "IZMIRAN",
        31362: "Pamukkale Tip Dergisi",
        31363: "V.M. Bekhterev National Research Medical Center"
               "for Psychiatry and Neurology",
        31364: "Scientific Research Journal SCIRJ",
        31365: "Sociedade de Pediatria do Estado do Rio de Janeiro",
        31366: "Educational Research Institute of Kongju National University",
        31367: "FSBSI Agricultural Research Center Donskoy",
        31368: "Instituto Biologico",
        31369: "Korea Society of East Asian Comparative",
        31370: "Korean Society for the Biblical Theological Studies",
        31371: "Global Center for Japanese Studies",
        31372: "Asian American Pacific Islander Nurses Association (AAPINA)",
        31373: "Gdanski Uniwersytet Medyczny",
        31374: "Scandinavian Military Studies",
        31375: "Development of Management and Entrepreneurship"
               "Methods on Transport (ONMU)",
        31376: "Olexander Dovzhenko Hlukhiv National Pedagogical University",
        31377: "Avicenna Journals",
        31378: "Hanzehogeschool Groningen University of Applied Sciences",
        31379: "Clinical Anesthesiology and Intensive Care",
        31380: "Oral Roberts University",
        31381: "Universidad Ricardo Palma",
        31382: "Faculty of Sport and Physical Education, Novi Sad University",
        31383: "University of Sarajevo, Institute for Genetic"
               "Engineering and Biotechnology",
        31384: "Shaheed Zulfiqar Ali Bhutto Institute of Science"
               "and Technology",
        31385: "Jurnal Ledalero",
        31386: "University of Duhok College of Medicine",
        31387: "OSCM Forum",
        31388: "Tavria State Agrotechnological University",
        31389: "LSE Press",
        31390: "Louisiana State University Libraries",
        31391: "ITESO, A.C.",
        31392: "National Pedagogical Dragomanov University",
        31393: "Vinnytsia National Pyrogov Memorial Medical University",
        31394: "Ordfront Forlag",
        31395: "Uman National University of Horticulture",
        31396: "Edizioni Danaus sas",
        31397: "Community Development Academic",
        31398: "National Fisheries Research and Development Institute",
        31399: "ASM International",
        31400: "Digitalis Bolcseszet",
        31401: "Balkan, Black sea and Caspian sea Regional"
               "Network for Space Weather Studies",
        31402: "SDinformacion SAS",
        31403: "Sociedad Peruana de Obstetricia y Ginecologia",
        31405: "Ukrainian Association of Infection Control"
               "and Antimicrobial Resistance",
        31406: "Asocicacion Latinoamericana de Poblacion",
        31407: "IJEES",
        31408: "Tae Dong Institute of Classic Research",
        31409: "Korean Council on Social Welfare Education",
        31410: "Association of Economists and Managers of"
               "the Balkans - Udekom Balkan",
        31411: "Nottingham Contemporary",
        31412: "Academia Nacional de Policia",
        31413: "Nativa",
        31414: "Fundaco Educaional Inaciana Pe Saboia De Medeiros",
        31415: "Brazilian Journal of Natural Science",
        31416: "Revista Semiarido De Visu",
        31417: "Instituto Federal do Amazonas",
        31418: "Revista da ABPN",
        31419: "Boletim do Museu de Geociencias da Amazonia",
        31420: "Uakari",
        31421: "International Journal of Horticultural Science",
        31422: "Theoretical and Applied Engineering",
        31423: "Familia e Sociedade em Debate",
        31424: "International Computer Science and Engineering"
               "Society (ICSES)",
        31425: "Voprosy Literatury",
        31426: "International Journal of Advanced Multidisciplinary"
               "Scientific Research (IJAMSR)",
        31427: "Unijourn Publisher",
        31428: "Universidad Politecnica de Cartagena",
        31429: "Kuban State University",
        31430: "The Arab Center for Research and Policy Studies",
        31431: "IVS FEB RAS",
        31432: "International Centre for Scientific and Technical"
               "Information",
        31433: "ICARP FEBRAS",
        31434: "Revista Medica Sinergia",
        31435: "RS Global Sp. z O.O.",
        31436: "IIUM Press",
        31437: "Scientific Array",
        31438: "Transducer Research Foundation Inc. (TRF)",
        31439: "UNISCI",
        31440: "Don't forget the bubbles",
        31441: "International Quranic Studies Association",
        31442: "Economy of Agricultural and Processing Enterprises",
        31443: "East Siberian State Institute of Culture",
        31444: "JPM Press",
        31445: "Balai Pengembangan SDM dan Penelitian Kominfo Jakarta",
        31446: "Central States Communication Association",
        31447: "Instituto de Ciencias Sociais",
        31448: "KONSTRUKCE Media, s. r. o.",
        31449: "Slovenian Association Informatika",
        31450: "Institute of Nephrology of the National Academy"
               "of Medical Sciences",
        31451: "Avrasya Terim Dergisi",
        31452: "The Burlington Magazine",
        31453: "OOO KDU",
        31454: "Canakkale Onsekiz Mart Universitesi Uluslararasi"
               "Sosyal Bilimler Dergisi",
        31455: "Akademik Sosyal Arastirmalar Dergisi (Asya Studies)",
        31456: "Journal of Islamicjerusalem Studies",
        31457: "Dogu Anadolu Sosyal Bilimlerde Egilimler Dergisi",
        31458: "International e-Journal of Educational Studies",
        31459: "Turkish Journal of Kinesiology",
        31460: "Muhasebe Bilim Dunyası Dergisi",
        31461: "Yasam Becerileri Psikoloji Dergisi",
        31462: "Golden Light Publishing",
        31463: "Agri Ibrahim Cecen Universitesi Sosyal Bilimler"
               "Enstitusu Dergisi",
        31464: "Dil Egitimi ve Arastirmalari Dergisi",
        31465: "Edebi Elestiri Dergisi",
        31466: "Karadeniz Fen Bilimleri Dergisi",
        31467: "Uludag Aricilik Dergisi",
        31468: "Canadian Journal for Studies in Discourse"
               "Writing/Redactology",
        31469: "Fundacion Universitaria ESUMER",
        31470: "Pereiaslav-Khmelnytskyi Hryhorii Skovoroda"
               "State Pedagogical University",
        31471: "Ivano-Frankivsk National Technical University of Oil and Gas",
        31472: "Institute of Engineering Thermophysics of NAS of Ukraine",
        31473: "Leonid Pogorilyy Ukrainian Scientific Research Institute",
        31474: "Donetsk National Technical University",
        31475: "Khmelnytskyi Humanitarian-Pedagogical Academy",
        31476: "Thinkershub AB",
        31477: "The Central Bank of the Russian Federation",
        31478: "National Academy of Medicine",
        31479: "Universitas Jayabaya",
        31480: "Transpopmed",
        31481: "Odessa State Environmental University",
        31482: "Faculty of Military Health Sciences, University of Defence",
        31483: "Publishing House Sreda",
        31484: "Fondazione di ricerca Istituto Carlo Cattaneo",
        31485: "Abdul Latif Jameel Poverty Action Lab (J-PAL)",
        31486: "Ochsner Journal",
        31487: "Science Repository OU",
        31488: "MAK Periodical Library",
        31489: "Buketov Karaganda State University",
        31490: "VSB - Technical University of Ostrava",
        31491: "Ant Publishing",
        31492: "Associacao das Universidades de Lingua Portuguesa",
        31493: "Transfer of Innovative Technologies",
        31494: "Berdiansk State Pedagogical University",
        31495: "Korean Association of Manual Medicine",
        31496: "Revista Profissao Docente",
        31497: "Journal of Natural Resources",
        31498: "SHEI Pryazovskyi State Technical University",
        31499: "Pavlo Tychyna Uman State Pedagogical University",
        31500: "Modern Art Research Institute of the National"
               "Academy of Arts of Ukraine",
        31501: "Universidade Catolica de Brasilia",
        31502: "Social Science Research Institute Yonsei University",
        31503: "Balai Bahasa Riau",
        31504: "BPSDMP KOMINFO Surabaya",
        31505: "Revista Brasiliera de Terapia Comportamental"
               "e Cognitiva (RBTCC)",
        31506: "FISIP Universitas Sultan Ageng Tirtayasa",
        31507: "Management Accounting Association of Korea",
        31508: "Sociedade Brasileira de Enfermeiros Pediatras",
        31509: "Center for Ecological Problems and Productivity"
               "of Forests Russian Academy of Sciences",
        31510: "Interface Tecnologica",
        31511: "Elk Education Consultant Private Limited",
        31512: "Universidade Regional Integrada do Alto U\u200bruguai"
               "e das Missoes",
        31513: "Revista Linguistica",
        31514: "Revista Liberato",
        31515: "Autonomous Nonprofit Organization Publishing Dairy Industry",
        31516: "Kharkiv State Academy of Culture",
        31517: "Revista Saberes da Amazonia",
        31518: "Institute of History of the Russian Academy of Sciences",
        31519: "Academus Publishing",
        31520: "Institute of Market Problems and Economic"
               "and  Ecological Research of the NAS of Ukraine",
        31521: "Mykolayiv National Agrarian University",
        31522: "University of Zagreb, Faculty of Architecture",
        31523: "American Educational Association Multiple"
               "Linear Regression Special Interest Group",
        31524: "Bangkok Dusit Medical Service (BDMS)",
        31525: "Case Journals",
        31526: "Andromeda Publishing and Education Services",
        31527: "Asociacion Fisica Argentina",
        31528: "APUB Co. Ltd",
        31529: "Sarhad University of Science and IT Peshawar",
        31530: "Charmo University",
        31531: "Edwiser International",
        31532: "Rivera Publications Inc",
        31533: "Editora MV Valero",
        31534: "Faculty of Civil Engineering, Architecture"
               "and Geodesy, University of Split",
        31535: "Korean Association of Vietnamese Studies",
        31536: "Korean Society of Legislation Studies",
        31537: "IKIP PGRI Jember",
        31538: "Institut Pesantren Kh. Abdul Chalim Pact Mojokerto",
        31539: "IPM2KPE",
        31540: "LP4MK STKIP PGRI Lubuklinggau",
        31541: "Korean Forest Economics Society",
        31542: "MacEwan University Library",
        31543: "Program Studi Teknik Mesin Institut Teknologi Indonesia",
        31544: "Politeknik Sukabumi",
        31545: "Institute of Agrophysics Polish Academy of Sciences",
        31546: "Medical Press Open Access",
        31547: "Asia-Pacific Blood and Marrow Transplantation Group",
        31548: "National University of Life and Environmental"
               "Sciences of Ukraine",
        31549: "FSBEI HE NSMU MOH Russia",
        31550: "NP Rusmedical Group",
        31551: "ESKU named after S. Amanzholova",
        31552: "The Institute For History Studies",
        31553: "Korean Association For Policy Science",
        31554: "Buryat Scientific Center SB RAS",
        31555: "SPER Publications and Solutions Pvt. Ltd",
        31556: "Fund Gelios",
        31557: "EpiSmart Science Vector Ltd",
        31558: "Vasyl Stus Donetsk National University",
        31559: "Refaad for Studies and Research",
        31560: "Pimenta Cultural",
        31561: "TAYR Quarterly Journal",
        31562: "International Foundation for Korean Language"
               "& Culture Education",
        31563: "Bashkir State Agrarian University",
        31565: "Korean Writing Association",
        31566: "Journal of Arts",
        31567: "Social Science Development Journal",
        31568: "Atlas Journal",
        31569: "Journal Of Academic Social Resources",
        31570: "Prosperitas",
        31571: "LPPM IKIP PGRI Pontianak",
        31572: "Politeknik Aceh Selatan",
        31573: "Politeknik Negeri Pontianak",
        31574: "Sari Citra Surya",
        31575: "Sekolah Tinggi Ilmu ekonomi Keuangan Perbankan"
               "dan Pembangunan (STIE KBP)",
        31576: "Social Mentality And Researcher Thinkers Journal",
        31577: "Central Library of the Slovak Academy of Sciences",
        31578: "International Black Sea University",
        31579: "Auctores Publishing LLC",
        31580: "Readers Insight Publisher",
        31581: "Association for Bahai Studies",
        31582: "Rocky Mountain Association of Geologists",
        31583: "Iriss",
        31584: "Faculty of Medicine Prince of Songkla University",
        31585: "The British Blockchain Association",
        31586: "Trends Journal of Sciences Research",
        31587: "Research Institute for Early Childhood Education"
               "Korea National University of Education",
        31588: "FSBEI HE Kazan SAVM after N. E. Bauman",
        31589: "Journal Of Social Humanities and Administrative Sciences",
        31590: "European Journal of Science and Technology",
        31591: "Islam Sanat Tarih Edebiyat ve Musıkisi Dergisi ISTEM",
        31592: "Ahi Evran Universitesi Sosyal Bilimler Enstitusu Dergisi",
        31593: "International Journal of Energy Applications"
               "and Technologies",
        31594: "Commagene Journal of Biology",
        31595: "Sekolah Tinggi Kesejahteraan Sosial (STKS) Bandung",
        31596: "STIKES Cendekia Utama Kudus",
        31597: "STKIP PGRI Bangkalan",
        31598: "STMIK STIKOM Indonesia",
        31599: "Universitas Bhayangkara Jakarta Raya",
        31600: "Institute for the History of Material Culture"
               "Russian Academy of Science",
        31601: "Nova Univerza",
        31602: "Universitas Islam Kalimantan MAB Banjarmasin",
        31603: "Universitas Muhammadiyah Magelang",
        31604: "Universitas Muhamamadiyah Tapanuli Selatan",
        31605: "Universitas Sulawesi Barat",
        31606: "Korea Society of East Asian Comparative Literature",
        31607: "Revista Coletanea",
        31608: "Puncta",
        31609: "Fiziksel T?p ve Rehabilitasyon Bilimleri Dergisi",
        31610: "Zoological Institute of the Russian Academy of Sciences",
        31611: "Caribbean Herpetology",
        31612: "State Institution of Science Research and Practical Center",
        31613: "Ceramist",
        31614: "Tech Science Press",
        31615: "LLC CARTEC",
        31616: "Asian Spine Journal (ASJ)",
        31617: "Kyiv National University of Trade and Economics",
        31618: "Consulting Group Education and Science LLC",
        31619: "Consejo Nacional de Educacion",
        31620: "Amity University Madhya Pradesh Gwalior",
        31622: "Clinical Practice in Athletic Training",
        31623: "Journal Of Institute Of Economic Development"
               "And Social Researches",
        31624: "Turk Kulturu Ve Haci Bektas Veli Arastırma Dergisi",
        31625: "Institute of Social Sciences Soongsil University",
        31626: "Szemere Bertalan Magyar Rendvedelem-torteneti"
               "Tudomanyos Tarsasag",
        31627: "Rendvedelem-torteneti fuzetek",
        31628: "Rendvedelem-torteneti hirlevel",
        31629: "Universitas Maritim Raja Ali Haji",
        31630: "Institute of the Archaeology of the North",
        31631: "LLC Numicom",
        31632: "Dr Tarak Nath Podder Memorial Foundation",
        31633: "CBE - Congresso Brasileiro de Educaccao",
        31634: "Hanyang University Global Center for Japanese Studies",
        31635: "Chinese Chemical Society",
        31636: "Interdisciplinar Academy of Pain Medicine",
        31637: "Sinnergiak Social Innovation",
        31638: "International Journal for Pharmaceutical Research"
               "Scholars IJPRS",
        31639: "Revista Brasileira de Pesquisa sobre Formaccaao"
               "de Professores",
        31640: "LLC Information and Research Center Likarska Sprava",
        31641: "The City University of New York",
        31642: "University of Kufa - Faculty of Computer Science"
               "and Mathematics",
        31643: "Institute of Metallurgy and Ore Benefication",
        31644: "Universidad Autonoma de Chiapas",
        31645: "Journal of Independent Studies and Research",
        31646: "University of New South Wales",
        31647: "Koreanische Kafka Gesellschaft",
        31648: "Uniwersytet Warminsko-Mazurski",
        31649: "Vinnytsia National Technical University",
        31650: "Odessa State Academy of Civil Engineering and  Architecture",
        31651: "Bohdan Khmelnytsky National University Of Cherkasy",
        31652: "Vinnytsia Mykhailo Kotsiubynskyi State  Pedagogical"
               "University",
        31653: "National University Odessa Maritime Academy",
        31654: "Nizhyn Mykola Gogol State University",
        31655: "Asthma and Allergy",
        31656: "Corporacion Universitaria Sabaneta Unisabaneta",
        31657: "Pia Sociedade de Sao Paulo FAPCOM",
        31658: "Duksung Women's University The Institute for"
               "Humanties Research",
        31659: "Stroymaterialy",
        31660: "Industrial University of Tyumen",
        31661: "Salvia Medical Sciences Ltd",
        31662: "Japan Medical Association",
        31663: "Thi-Qar University Journal for Engineering Sciences",
        31664: "Institute of Art History",
        31665: "ISNFF and Associates Publishing Company",
        31666: "Institute of Chinese Characters and Korean"
               "Literature in Classical Chinese Korea University",
        31667: "The Academy of Korea Hospitality and Tourism",
        31668: "Universidade Estadual De Goias",
        31669: "Revista Herodoto",
        31670: "University of Miami Libraries",
        31671: "Dogus Universitesi Dergisi",
        31672: "Revista Derecho Public",
        31673: "State University of Telecommunications",
        31674: "Lincoln University College",
        31675: "Tomsk State University of Architecture and Building",
        31676: "Federal State Budgetary Scientific Institution"
               "All-Russian Horticultural Institute for Breeding"
               "Agrotechnology and Nursery",
        31677: "Federal State Educational Institution of Higher Education"
               "Novosibirsk State Agrarian University",
        31679: "Adam Akademi Sosyal Bilimler Dergisi",
        31680: "Gaziantep Universitesi Spor Bilimleri Dergisi",
        31681: "Journal of Educational Technology and Online Learning",
        31682: "Ayna Klinik Psikoloji Dergisi",
        31683: "Escola de Psicanalise dos Fóruns do Campo Lacaniano",
        31684: "Altai State Medical University",
        31685: "Ministry Of Finance - Fiscal Policy Agency",
        31686: "International Journal for Innovation Education and Research",
        31687: "Sociedad Argentina para el Estudio de los Mamiferos",
        31688: "Uniunea Medicala Balcanica",
        31689: "CMMB-Colegiul Medicilor Municipiul Bucuresti",
        31690: "Innovational Publishers",
        31691: "Korea Society Of Air Space Law And Policy",
        31692: "International Journal Education and Teaching",
        31693: "The Poputation Association of Korea",
        31694: "The Korean Society for the Philosophy of Sport"
               "Dance and Martial Arts",
        31695: "Sretechjournal Publication",
        31696: "Institute of Oriental Studies Russian Academy of Sciences",
        31697: "Korean Society of Petroleum and Sedimentary Geology",
        31698: "Sociedad Paraguaya de Pediatria",
        31699: "Iraqi Journal of Chemical and Petroleum Engineering",
        31700: "Applis Publishers",
        31701: "ENAC",
        31702: "Centro de Investigacion Cientifica y de Educacion"
               "Superior de Ensenada B.C.",
        31703: "Humanity Only - HO",
        31704: "Egitim Programlari ve Ogretim Dernegi",
        31705: "University of Moratuwa",
        31706: "International Energy and Environmental Research Institute",
        31707: "College of Optometrists in Vision Development",
        31708: "Institute of Biotechnology and Molecular Medicine",
        31709: "J. A. Alpha Business Research and Publishers Pvt Ltd",
        31710: "Psychological Association of the Philippines",
        31711: "Utah Geological Association",
        31712: "The Academy of Medical Sciences of Tajikistan",
        31713: "National University of Water and Environmental Engineering",
        31714: "Swedish Physical Society",
        31715: "Van Hall Larenstein University of Applied Science",
        31716: "Frontier Research Today S A",
        31717: "Institute for Safety Problems of Nuclear Power Plants",
        31718: "Ukrainian Medical  Stomatological Academy",
        31719: "A Journal of Academic Writing Assignments",
        31720: "Institute for Global and Area Studies",
        31721: "Kryvyi Rih National University",
        31722: "Ege Universitesi Devlet Turk Musikisi Konservatuvarı Dergisi",
        31723: "Odessa National A.V. Nezhdanova Academy of Music",
        31724: "Institute of Croatian Language and Linguistics",
        31725: "Central Astronomical Observatory at Pulkovo of RAS",
        31726: "Drzavni Arhiv U Pazinu",
        31727: "Zadruzna Stampa D.D.",
        31728: "Korean Society of Neurosonology",
        31729: "Journal of Nepal Medical Association (JNMA)",
        31730: "AfricaArXiv",
        31731: "Cherkasy Institute of Fire Safety named after"
               "Chornobyl Heroes of National University of"
               "Civil Defence of Ukraine",
        31732: "KROK University",
        31733: "The Dnipropetrovsk State University of Internal Affairs",
        31734: "Lviv National Agrarian University",
        31735: "The Chin-Tan Society",
        31736: "The Indian Society of Toxicology",
        31737: "Journal of New Economic Association",
        31739: "GGZ.NL",
        31740: "MSU Gensan Graduate School",
        31741: "International Journal of Health Research and"
               "Medico-Legal Practice",
        31742: "The Indian Society of Genetics and Plant Breeding",
        31743: "Katolicki Uniwersytet Lubelski Jana Pawla II",
        31744: "Sociedade Beneficente Israelita Brasileira"
               "Hospital Albert Einstein",
        31745: "Staroslavenski Institut",
        31746: "Linguagem e Educacao",
        31747: "Universidade Salvador - UNIFACS",
        31748: "Korean Society of Educational Gerontology",
        31749: "Andrzej Frycz Modrzewski Krakow University",
        31750: "Centro de Investigacao em Politica Economia"
               "e Sociedade (CIPES)",
        31751: "Gesellschaft fur Musiktheorie (GMTH)",
        31752: "The International Institute for Democracy"
               "and Electoral Assistance - International IDEA",
        31753: "Mississippi Academy of Sciences",
        31754: "Publishing House Nestor-Historia",
        31755: "African International Center for Scientific Research - AICSR",
        31756: "European Star Academic Publishing Ltd",
        31757: "European Star Academic Publishing Ltd",
        31758: "Physics Education Study Programme - UHAMKA",
        31759: "University of Babylon Environmental Research Center",
        31760: "Revista Mexicana de Ingenieria Quimica (RMIQ)",
        31761: "Center for a Public Anthropology",
        31762: "ASEAN Heart Journal",
        31763: "ASCEE Publications",
        31764: "Universitas Muhammadiyah Mataram",
        31765: "KAREN - Karadeniz Arastirmalari Enstitusu Dergisi",
        31766: "Asociacion Espanola de Psiquiatria del Nino"
               "y del Adolescente - Aepnya",
        31767: "National Academy of Statistics Accounting and Audit",
        31768: "RE Kavetsky Institute of Experimental Pathology"
               "Oncology and Radiobiology of National Academy"
               "of Sciences of Ukraine",
        31769: "Far East University Institute of Social Welfare",
        31770: "Quebec Nursing Intervention Research Network",
        31771: "Journal of Recreation and Tourism Research",
        31772: "Reshetnev Siberian State University of Science"
               "and Technology",
        31773: "Federal State-Funded Educational Institution"
               "of Higher Education Kemerovo State University of Cultures",
        31774: "Federal State Budget Scientific Establishment"
               "The Russian Scientific Research Institute of"
               "Land Improvement Problems",
        31775: "Southern Institute of Management",
        31776: "Russian State Scientific Center for Robotics"
               "and Technical Cybernetics",
        31777: "Systematic Theology Division of Korea Evangelical"
               "Theological Society",
        31778: "The Korean Society of Legal History",
        31779: "Korean Comparative Public Law Association",
        31780: "Ewha Institute of Unification Studies",
        31781: "Busan Toegye Studies Institute",
        31782: "Radiance Research Academy",
        31783: "Emergent Life Sciences Research",
        31784: "Polytechnic of Rijeka University",
        31785: "Algebra University College",
        31786: "MANU - International Council for Man and Nature",
        31787: "Stellenbosch University - Journal of Emerging"
               "African Scholarship",
        31788: "Rasayan Journal of Chemistry",
        31789: "IMSciD - International Multidisciplinary Scientific Database",
        31790: "INSPI",
        31791: "The Association For Korean Historical Studies",
        31792: "Universidade do Estado do Para",
        31793: "LLC Publishing house Medknyha",
        31794: "N. L. Dalmia Institute of Management Studies & Research",
        31795: "Balikesir Universitesi Sosyal Bilimler Enstitusu Dergisi",
        31796: "Eskisehir Osmangazi Universitesi Muhendislik"
               "ve Mimarlik Fakultesi Dergisi",
        31797: "Journal of Advances in VetBio Science and Techniques",
        31798: "Scientific Educational Studies",
        31799: "State University of Aerospace Instrumentation (SUAI)",
        31800: "Balai Pengembangan Media Televisi - BPMTV - Kemdikbud",
        31801: "Communications Faculty of Sciences University"
               "of Ankara Series A1 Mathematics and Statistics",
        31802: "Moscow Theological Academy",
        31803: "University North",
        31804: "Commonwealth of Englightenment of the Krasnoyarsk Territory",
        31805: "Academy Journal of Educational Sciences",
        31806: "Multifunctional center of applied studies, Krasnoyarsk",
        31807: "Turkish Journal of Water Science and Management",
        31808: "Menvia Solucoes em Informatica Ltda",
        31809: "Chunwon Research Society",
        31810: "Sociedad Espanola de Linguistica",
        31811: "Online Journal of Music Sciences",
        31812: "Kryvyi Rih State Pedagogical University",
        31813: "Kantor Bahasa Maluku Utara",
        31814: "National University of Civil Engineering (NUCE)",
        31815: "Pusat Penelitian dan Pengembangan Perumahan dan Permukiman",
        31816: "German Academic Exchange Service (DAAD)",
        31817: "Agricultural University Press",
        31818: "KNST (Korea society for Naval Science and Technology)",
        31819: "Vervuert Verlag oHG",
        31820: "Faculty of Humanities and Social Sciences"
               "University of Rijeka",
        31821: "Hanminjok Emunhakhoe",
        31822: "Disiplinlerarasi Akademik Turizm Dergisi",
        31823: "Catholic Faculty of Theology in Dakovo",
        31824: "North American Observatory on Health Systems and Policies",
        31825: "Vision Research",
        31826: "Gorgias Press LLC",
        31827: "The Society for the Education of Korean Language and Culture",
        31828: "Turk Psikologlar Dernegi",
        31829: "GUDAPURIS LLC",
        31830: "Gaurav Publications",
        31831: "Frelsi LLC",
        31832: "Sakarya Tip Dergisi",
        31833: "Order Sign of Honor the Institute of History,"
               "Language and Literature UFRC RAS",
        31834: "Sosyal Bilimler Dergisi",
        31835: "Roy Rosenzweig Center for History and New Media",
        31836: "Letras Historicas",
        31837: "Sociedad Cirugia del Uruguay",
        31838: "SynthesisHub  Advance Scientific Research",
        31839: "Institute for the Study of Law Dong-A University",
        31840: "Sociedad y Ambiente",
        31841: "Kardan University",
        31842: "Asosiasi Forum Inovasi Bisnis dan Manajemen (INOBIS)",
        31843: "Asosiasi Forum Manajemen Indonesia (FMI)",
        31844: "FKIP Universitas Wirara",
        31845: "Pusat Kajian dan Pelatihan dan Pendidikan"
               "Aparatur Lembaga Administrasi Negara (PKP2A I LAN)",
        31846: "Sekolah Tinggi Ilmu Ekonomi Pasaman",
        31847: "Unit Penerbitan dan Publikasi Ilmiah (UPPI) IAIN Kudus",
        31848: "Universitas Kebangsaan",
        31849: "Universitas Lancang Kuning",
        31850: "Universitas Muhammadiyah Parepare",
        31851: "Universitas PGRI Palembang",
        31852: "Egyetemi Meteorolgiai Fuzetek",
        31853: "LLC Integration Education and Science",
        31854: "Bonch-Bruevich State University of Telecommunications",
        31855: "Wolfram Research, Inc.",
        31856: "Redaction of Journal Public Policy",
        31857: "The Russian Academy of Sciences",
        31858: "University of KwaZulu-Natal Humanities Department",
        31859: "MIT Press - Blogs",
        31860: "Institute of Russian Literature Pushkinskij Dom RAN",
        31861: "Yuriy Fedkovych Chernivtsi National University",
        31862: "Federal State Budgetary Educational Institution"
               "of Higher Education Moscow Pedagogical State University",
        31863: "Special Education Research Institute Dankook University",
        31864: "Revista Conhecimento Contabil",
        31865: "SHEE Donbas State Pedagogical University",
        31866: "Kyiv National University of Culture and Arts",
        31867: "SE Institute of Grain Crops of National Academy"
               "of Agrarian Sciences of Ukraine",
        31868: "Institute of Automation and Electrometry",
        31869: "LPPM Universitas Muhammadiyah Sumatera Barat",
        31870: "The Society of Earth Scientists",
        31871: "Nextgen Research Publications",
        31872: "Kenkyu Publishing Group",
        31873: "Engineering Research Publication ERP",
        31874: "Institute of Higher Education NAES of Ukraine",
        31875: "Zeal Press",
        31876: "Espirales Revista Multidisciplinaria de Investigacion",
        31877: "AEOP - Associacao Enfermagem Oncologica Portuguesa",
        31878: "Sumathi Publications",
        31879: "Sociedad Chilena de Dermatologia",
        31880: "Glucksman Library University of Limerick",
        31881: "Seniko Studio Ltd",
        31882: "Laboratory of Intellect Ltd",
        31883: "Institute of Animal Reproduction and Food"
               "Research of the Polish Academy of Sciences",
        31884: "P3M Politeknik Negeri Indramayu",
        31885: "University of Helsinki Libraries",
        31886: "Journal of Recovery Science",
        31887: "Servier International",
        31888: "Korea Gerontological Society",
        31889: "The Society of Korean Language and Literature",
        31890: "Kharkiv State Zooveterynary Academy",
        31891: "Khmelnytskyi National University",
        31892: "Revista Brasileira de Pesquisa Auto Biografica",
        31893: "JABB - Journal of Animal Behaviour and Biometeorology",
        31894: "The Korea Rural Sociological Society",
        31895: "Croatian Society of Food Technologists Biotechnologists"
               "and Nutritionists",
        31896: "Croatian Society for Geometry and Graphics",
        31897: "Saint-Petersburg Mining University",
        31898: "Instituto de Pesca - APTA - SAA",
        31899: "Population Council",
        31900: "Institute of Laser for Postgraduate Studies",
        31901: "Kamla Raj Enterprises",
        31902: "Faculty of Philology - University of Montenegro",
        31903: "The Journal of Regional Economic Review of East Asia",
        31904: "Faculty of Islamic Studies - University of"
               "Muhammadiyah Jakarta",
        31905: "International Seismological Centre",
        31906: "Asociacion Universidad De La Salle",
        31907: "Green Publishers",
        31908: "Universidad Catolica de Pereira",
        31909: "Izmail State University of Humanities",
        31910: "Universidad de Ciencias Aplicadas Y Ambientales - UDCA",
        31911: "Archivo General de la Nacion",
        31912: "Vinogradov Russian Language Institute of the"
               "Russian Academy of Sciences",
        31913: "ScholarMate, Ltd.",
        31914: "Acta Agraria Kaposvariensis",
        31915: "HUNGARNET",
        31916: "ScholarGen Publishers",
        31917: "LLC Center TOMM",
        31918: "Soran University",
        31919: "Colegio de Profesionales del Trabajo Social de Puerto Rico",
        31920: "Adonis and Abbey Publishers",
        31921: "Fundacion Universitaria San Pablo CEU",
        31922: "Georgia State University - Library",
        31923: "PoolText",
        31924: "Natural Resources and Sustainable Development"
               "Journal - NRSDJ",
        31925: "Societatea de Stiinte Farmaceutice din Romania",
        31926: "Universitatea Transilvania Brasov",
        31927: "Academic Society of Event and Convention",
        31928: "Chetverta ?hvylia",
        31929: "Sunchon National University Namdo Cultural Research Center",
        31930: "Asiatic Research Institute Korea University",
        31931: "BioOne (Freshwater Mollusk Biology and Conservation)",
        31932: "STKIP Persada Khatulistiwa",
        31933: "LPPM Universitas EKASAKTI",
        31934: "Universitas Muhammadiyah Palu",
        31935: "Akbid Harapan Mulya Ponorogo",
        31936: "Universitas Muhammadiyah Sumatera Barat",
        31937: "Universitas Multimedia Nusantara",
        31938: "Universitas Nusa Bangsa",
        31939: "Organismo Autonomo de Museos y Centros del"
               "Cabildo de Tenerife",
        31940: "Politeknik Negeri Bali",
        31941: "Universitas Pekalongan",
        31942: "Universitas Wahid Hasyim Semerang",
        31943: "Universitas Wiralodra",
        31944: "UNITAS",
        31945: "Islamabad Policy Research Institute - IPRI",
        31946: "persist.ed",
        31947: "Fakultas Ilmu Sosial dan Ilmu Politik Universitas Hasanuddin",
        31948: "Universidad Mariana",
        31949: "Universitas Majalengka",
        31950: "DentaMed Publishing",
        31951: "Limnological Institute SB RAS",
        31952: "Croatia Scientific Society for the History of Health Culture",
        31953: "Senjsko muzejsko drustvo",
        31954: "Publicaciones Latinoamericanas SRL - Fronteras en Medicina",
        31955: "LPPM STIE Muhammadiyah Bandung",
        31956: "Ural State Forest Engineering University",
        31957: "Universitas Cendrawasih",
        31958: "Iain Batusangkar",
        31959: "P3M Politeknik Negeri Ambon",
        31960: "ILIN Institute",
        31961: "P3M Politeknik Negeri Banjarmasin",
        31962: "Politeknik Negeri Subang",
        31963: "Politeknik Negeri Ujung Pandang",
        31964: "Poltekkes Kemenkes Banjarmasin",
        31965: "Poltekkes Kemenkes Kupang",
        31966: "STIE Malang Kucecwara",
        31967: "P3M STIE Mandala Jember",
        31968: "Institute for History TextBooks Research",
        31969: "Balai Penelitian dan Pengembangan Agama Makassar",
        31970: "University of Alkhairaat Palu",
        31971: "University of Warsaw - Department of Political"
               "Science and International Studies",
        31972: "Salahaddin University- Erbil- Conferences",
        31973: "University of Baghdad - College of Arts",
        31974: "Association of Forensic Document Examination",
        31975: "The Society for Basic Medical Sciences",
        31976: "Sociedade Brasileira De Fitopatologia",
        31977: "Griot Revista de Filosofia",
        31978: "Agencia Estatal De Meteorologia",
        31979: "San Jose State University Library",
        31980: "Institut Pendidikan Indonesia",
        31981: "PreLights (The Company of Biologists)",
        31982: "The New Testament Society Of Korea",
        31983: "Poltekkes Kemenkes Semarang",
        31984: "Revista Odontologia Activa",
        31985: "The Society Chinese Literature",
        31986: "Rowan University Campbell Library",
        31987: "College of Information Engineering - Al-Nahrain University",
        31988: "Science Trends",
        31989: "Functional Food Center",
        31990: "Revista Agenda Politica",
        31991: "Revista de Odontologia Contemporanea - ROC",
        31992: "Moscow Polytechnic University",
        31993: "All-Russian Institute of Plant Protection",
        31994: "Instituto Vianna Junior Ltda",
        31995: "Anu Books",
        31996: "Ukrainian State Geological Research Institute",
        31997: "The Republican Research and Practical Center"
               "for Pediatric Surgery",
        31998: "The Korean Society Of Family Law",
        31999: "Unification and North Korean Law Studies",
        32000: "LLC Praktika",
        32001: "Sinema Arastirmalari Dergisi",
        32002: "Ombudsman Akademik",
        32003: "International Journal of Geography and Geography Education",
        32004: "Kajkavian Parliament",
        32005: "Sri Eshwar College of Engineering",
        32006: "National Society of Ecological Engineering"
               "and Environment Protection",
        32007: "Faculty of Medicine University of Baghdad",
        32008: "Saima Consult Ltd",
        32009: "UCAN Services",
        32010: "Azerbaijan State Oil and Industry University",
        32011: "Texas Academy of Science",
        32012: "Corporacion Universitaria Latinoamericana CUL",
        32013: "Society of Psychoceramics",
        32014: "National Academy of Sciences of the Republic of Kazakshtan",
        32015: "DOBA Business School",
        32028: "Archaeopress Publishing Ltd",
        32031: "Integral Transpersonal Institute",
        32038: "EUROKD Egitm Danismanlik Group",
        32044: "Positive Press",
        32054: "Uniciels",
        32070: "Consilium LLC",
        32074: "Siapec Servizi Srl",
        32078: "Rzecznik Finansowy (Insurance Ombudsman)",
        32098: "Edra SpA",
        32315: "Ikatan Peneliti Lingkungan Binaan Indonesia",
        32316: "Historical Studies in Education",
        32317: "Ekonomika APK",
        32318: "International Journal of Pharmacy and Technology",
        32319: "Society of Japanese Language Education Hong Kong",
        32320: "Educational Research Institute",
        32321: "Cukurova Universitesi Turkoloji Arastırmaları Dergisi",
        32322: "Journal of Health Sciences and Medicine",
        32323: "Universal Journal of Mathematics and Applications",
        32324: "Siberian Institute Of Management - The Russian"
               "Presidential Academy Of National Economy And"
               "Public Administration",
        32325: "Izmir Akdeniz Akademisi",
        32326: "National Research Lobachevsky State University"
               "of Nizhni Novgorod (UNN)",
        32327: "International Journal of Management Economics"
               "and Social Science",
        32328: "Turkish Journal of Forest Science",
        32329: "Universite Arastirmalari Dergisi",
        32330: "Nusha",
        32331: "Sosyal Guvenlik Dergisi",
        32332: "IAIN Metro Lampung",
        32333: "Korean Kant Society",
        32334: "O Que Nos Faz Pensar",
        32335: "O Debate Entre O Publico E O Privado",
        32336: "Instituto De Mantenca De Ensino Superior Ltda",
        32337: "Korean Academy Of Cardiorespiratory Physical Therapy",
        32338: "Non-Commercial Partnership Association Of"
               "Professionals For Regenerative Medicine (Diagnostics"
               "- Recovery - Rehabilitation)",
        32339: "JSC Chermetinformatsia",
        32340: "The Federal State Educational Institution"
               "Of Higher Education Altai State Institute Of Culture",
        32341: "The Korean Society for the Study of Career Education",
        32342: "Alfred Nobel University",
        32343: "The Institute of Education Development of Irkutsk Region",
        32344: "Legends Consulting Private Limited",
        32345: "Bogomolets National Medical University",
        32346: "University of Warsaw - Faculty of Education",
        32347: "Kyiv National University of Construction and Architecture",
        32348: "Universidad Nacional de Cordoba - Revista"
               "Argentina de Ciencias del Comportamiento",
        32349: "NURI00",
        32350: "University of Management and Technology",
        32351: "Fundacion MenteClara Internacional",
        32352: "The State Expert Center of the Ministry of Health of Ukraine",
        32353: "Hon. Prof. M.S. Bokarius Kharkiv Research"
               "Institute of Forensic Examinations",
        32354: "Korean Institute for Rhythmic Exercises",
        32355: "Institute for New Human Society",
        32356: "Revista Extensão em Ação",
        32357: "IJPRET",
        32358: "Revista Produção e Desenvolvimento",
        32359: "Revista Debates Insubmissos",
        32360: "Arquivos de Ciências do Mar",
        32361: "Revista de Direito",
        32362: "RTU MIREA",
        32363: "Print Publishing House",
        32364: "LLC Russian Medical Journal",
        32365: "The Korean Association for the Social History of Medicine",
        32366: "Donetsk Law Institute of MIA of Ukraine",
        32367: "Instituto Nacional de Investigacion e Innovacion Social",
        32368: "Forman Christian College - A Chartered University",
        32369: "Universidad Adventista de Bolivia",
        32370: "Intellectual Archive",
        32371: "Bendola S.p.r.L - BVBA",
        32372: "4ID",
        32373: "Association for Science and Engineering",
        32374: "Our Solar Siblings",
        32375: "Pacific Section of the American Association"
               "of Petroleum Geologists - PSAAPG",
        32376: "mediastudies.press",
        32377: "CVR College of Engineering",
        32378: "Irish College of Paramedics",
        32379: "Editora UFSM - FATEC",
        32380: "Escienta Ltd",
        32381: "Printspublications Private Limited",
        32382: "Poltekkes Kemenkes Makassar",
        32383: "Polish Pharmaceutical Society",
        32384: "European Association for Health Information"
               "and Libraries EAHIL",
        32385: "Associacao Portuguesa de Medicina Geral e Familiar",
        32386: "Bullaki Ltd.",
        32387: "Vereinigung zur Kritik der politischen Okonomie e.V.",
        32388: "Qeios Ltd",
        32389: "Environmental and Engineering Geophysical Society",
        32390: "The Korean Society of Mineral and Energy Resources Engineers",
        32391: "Albanian Society for Trauma and Emergency Surgery",
        32392: "Knowledge Enterprise Journals",
        32393: "Robertson Library, University of Prince Edward Island",
        32394: "National Institute of Public Health – National"
               "Institute of Hygiene",
        32395: "IJPRBS",
        32396: "University of Saskatchewan Library",
        32397: "Universidad Technologica de Bolivar",
        32398: "California State University Fullerton",
        32399: "Benemerita Universidad Autonoma de Puebla",
        32400: "Universitas Sam Ratulangi",
        32401: "Network Scientific-Research Periodical",
        32402: "Institute for Public Health of the National"
               "Academy of Medical Sciences of Ukraine",
        32403: "Ukrainian Academy of Printing",
        32404: "Revista de Agricultura Neotropical",
        32405: "The Institute of Pedagogy of the National"
               "Academy of Pedagogical Sciences of Ukraine",
        32406: "Revista Agraria Academica",
        32407: "Publicaciones Latinoamericanas SRL",
        32408: "Compositionality",
        32409: "STMIK Jakarta STI and K",
        32410: "University of Halabja",
        32411: "Revista Nos",
        32412: "Philippine Society of Otolaryngology-Head"
               "and Neck Surgery, Inc. (PSO-HNS)",
        32413: "Health Services Academy",
        32414: "N.N. Priorov National Medical Research Center"
               "Of Traumotology and Orthopedics",
        32415: "Scientia Publishing House LTD",
        32416: "Publishing House Logos Press",
        32417: "Urals State Agrarian University",
        32418: "Universidad Loyola Andalucia",
        32419: "Persatuan Perawat Nasional Indonesia",
        32420: "Ukrainian Association of Religion Researchers",
        32421: "Indonesian Urological Association",
        32422: "Institute of International Relations Prague",
        32423: "Balai Besar Logam dan Mesin",
        32424: "Pusat Pengelolaan Jurnal",
        32425: "Informatics Education",
        32426: "ENGRES - Engineering Research Technical Report",
        32427: "Korea Association of Local Administration",
        32428: "Korean Classical Poetry Association",
        32429: "The Korean Society of Health-System Pharmacists",
        32430: "Institute for Law of Science and Technology",
        32431: "The Korean Association Of Computer Education",
        32432: "Korean Association For Studies Of Philosophical Thought",
        32433: "Erciyes Journal of Education",
        32434: "SHEI Ukrainian State University of Chemical Technology",
        32435: "Environmental Smoke",
        32436: "College of the Holy Cross",
        32437: "NDSAN Network",
        32438: "World Energy and Environment Technology Ltd - WEENTECH",
        32439: "Plantae Scientia",
        32440: "Royal Spanish Academy of Medicine",
        32441: "International Scholars and Researchers Association",
        32442: "Asian Pacific Journal of Health Sciences",
        32443: "Latin American Journal of Telehealth",
        32444: "Tata Institute of Social Sciences TISS",
        32445: "Instituto Para O Desenvolvimento Democratico Ltda",
        32446: "FSUE VNIIMS All-Russian Research Institute"
               "of Metrological Service",
        32447: "Lviv State University of Life Safety",
        32448: "ENT Updates Journal",
        32449: "Turk Dunyası Incelemeleri Dergisi",
        32450: "Ankara Avrupa Calismalari Dergisi",
        32451: "The Korean Arts Psychotherapy Association",
        32452: "Academic Society of Science Engineering and Technology",
        32453: "Bogdan Khmelnytskyi National Academy of the"
               "State Border Guard Service of Ukraine",
        32454: "Sergo Ordshonikidze University",
        32455: "Thomson Journal",
        32456: "Econbiz Journal",
        32457: "Universidad Autonoma de Chile",
        32458: "Etnografski Muzej Zagreb",
        32459: "Centro Universitario Assuncao - Unifai",
        32460: "Central Scientific Library of the Urals Branch"
               "of the Russian Academy of Sciences",
        32461: "National Academy of Managerial Staff of Culture and Arts",
        32462: "LLC Publishing House Khleboproducty",
        32463: "OZZIE Publishers",
        32464: "Amp Complect Limited",
        32465: "The Korean Society for Sociology of Education",
        32466: "Editorial UFV",
        32467: "NESME - Nucleo de Estudos em Saude Mental",
        32468: "Banco de la República Colombia",
        32469: "University of Missouri Libraries",
        32470: "Cognitive Computational Neuroscience",
        32471: "LLC MORION",
        32472: "Meridians - The Journal of Acupuncture and Oriental Medicine",
        32473: "University of Florida George A Smathers Libraries",
        32474: "Lupine Publishers LLC",
        32475: "Societe entomologique de France",
        32476: "Academic Research Centre",
        32477: "STIE Widya Wiwaha",
        32478: "STAI Mahad Aly Al-Hikam Malang",
        32479: "EconJournals",
        32480: "Sociedad Cientifica del Paraguay",
        32481: "Delaware Academy of Medicine - Delaware Public"
               "Health Association",
        32482: "Heriot-Watt University",
        32483: "STIBANKS Al-Masoem",
        32484: "Salewangang Institute",
        32485: "Puslitbang KOPERTIP Indonesia",
        32486: "Politeknik Negeri Madiun",
        32487: "Politeknik Negeri Balikpapan",
        32488: "Puslitbang Bimas Agama dan Layanan Keagamaan"
               "Kementrian Agama",
        32489: "Sekolah Tinggi Agama Islam Negeri Sorong",
        32490: "Sekolah Tinggi Pendidikan Agama Kristen Malang",
        32491: "Masyarakat Iktiologi Indonesia",
        32492: "LPPM Universitas Darul Ulum",
        32493: "Universitas Pamulang",
        32494: "Askopis (Asosiasi Jurusan Komunikasi dan Penyiaran"
               "Islam-Indonesia)",
        32495: "Asosiasi Ilmu AlQuran dan Tafsir se Indonesia",
        32496: "Asian Journal Sections Indonesia",
        32497: "Politeknik Negeri Semarang",
        32498: "Revista (Con)textos Linguisticos",
        32500: "Fakultas Ekonomi Universitas Sains Al Qur_an"
               "(UNSIQ) Wonosobo",
        32501: "Fakultas Hukum Universitas Panca Bhakti",
        32502: "Universitas Muhammadiyah Palembang",
        32503: "Universitas Islam Kadiri",
        32504: "STIkes Surya Global Yogyakarta",
        32505: "IAIN Langsa",
        32506: "Islamicate Institute",
        32507: "Fakultas Agama Islam Universitas Ibn Khaldun Bogor",
        32508: "Viet Nam National University Ho Chi Minh City",
        32509: "Universitas Prof. Dr. Moestopo Beragama",
        32510: "KASB Institute of Technology",
        32511: "Politeknik Negeri Kupang",
        32512: "Junior Medical Research",
        32513: "Tbilisi Centre for Mathematical Sciences",
        32514: "Limited Liability Company PP ARIAL",
        32515: "Central Ukrainian National Technical University",
        32516: "Orenburg State Pedagogical University",
        32517: "Publishing House Education and Informatics",
        32518: "Lviv State University of Internal Affairs",
        32519: "Boletim Tecnico-Cientifico do CEPNOR",
        32520: "Universitas Islam Indragiri",
        32521: "Anuchin Research Institute and Museum of Anthropology,"
               "Lomonosov Moscow State University",
        32522: "Forestry Faculty, Mulawarman University",
        32523: "L. N. Gumilyov Eurasian National University",
        32524: "Universitas Katolik Musi Charitas",
        32525: "Guyra Paraguay",
        32526: "Faculty of Environment and Resource Studies"
               "- Mahidol University",
        32527: "KenzPub",
        32528: "Universitas Muhammadiyah Jember",
        32529: "Universitas Muhammadiyah Luwuk",
        32530: "Politeknik Pertanian Negeri Payakumbuh",
        32531: "Politeknik Katolik Saint Paul",
        32532: "Dokter Spesialis Onkologi Radiasi Indonesia",
        32533: "Yayasan Sukma",
        32534: "LPPM Universitas Muhammadiyah Cirebon",
        32535: "AIBPM Publisher",
        32536: "Asosiasi Institusi Pendidikan Kebidanan Muhammadiyah"
               "Aisyiyah",
        32537: "Balai Besar Keramik",
        32538: "ETN Centre Indonesia",
        32539: "Fakultas Kedokteran Universitas Sriwijaya",
        32540: "Prydniprovsk State Academyof Physical Culture and Sport",
        32541: "Instituto Superior de Formacion Docente Salome"
               "Urena (ISFODOSU)",
        32542: "The Korean Academy of Oral and Maxillofacial Implantology",
        32543: "National Civic League",
        32544: "Universidad Inca Garcilaso de la Vega",
        32545: "MDPI AG - Encyclopedia",
        32546: "STIE Ahmad Dahlan Jakarta",
        32547: "Ataturk Universitesi Guzel Sanatlar Enstitusu Dergisi",
        32548: "The American Society for Nondestructive Testing, Inc.",
        32549: "Ordine delle Professioni Infermieristiche di Napoli",
        32550: "Pustekkom",
        32551: "Medecine et Hygiene",
        32552: "Hacettepe University - Medical School",
        32553: "CAB PUBLISHER",
        32554: "Forum Dosen Akuntansi Sektor Publik",
        32555: "Robot Perception and Action Laboratory",
        32556: "Indonesian Society for Plant Taxonomy",
        32557: "USEFUL ACADEMY OF ENGINEERING AND COMPUTER SCIENCE INC.",
        32558: "LETRA Alapitvany",
        32559: "Europai tukor",
        32560: "Repulestudomanyi Kozlemenyek",
        32561: "Nemzetbiztonsagi Szemle",
        32562: "Muszaki Katonai Kozlony",
        32563: "Hadtudomanyi Szemle",
        32564: "Replika",
        32565: "Academic and Applied Research in Military"
               "and Public Management Science",
        32566: "Acta Humana",
        32567: "Hadmernok",
        32568: "Sakarya University Journal of Family Counseling"
               "and Education",
        32569: "Kocaeli Universitesi Mimarlik ve Tasarim Fakultesi"
               "- Resilience",
        32570: "Yildiz Teknik University - Uluslararasi Alan Egitimi Dergisi",
        32571: "International Journal of Chemistry and Technology",
        32572: "Ondokuz Mayis University - Guncel Turizm Arastirmalari"
               "Dergisi",
        32573: "Allamtudomanyi Muhelytanulmanyok",
        32574: "Tarsadalom es Honvedelem",
        32575: "Magyar Kozigazgatas",
        32576: "Nemzet es Biztonsag",
        32577: "Magyar Rendeszet",
        32578: "Bolyai Szemle",
        32579: "Adiyaman University - Mecuma",
        32580: "Korea Association of International Development"
               "and Cooperation",
        32581: "Korean Institute for Functional Medicine",
        32582: "Institute of Oceanography and Fisheries",
        32583: "Sekolah Tinggi Ilmu Kesehatan Kendal",
        32584: "Persatuan Perawat Nasional Indonesia Jawa Tengah",
        32585: "Universitas Veteran  Bangun Nusantara Sukoharjo",
        32586: "Tribunal de Contas do Estado do Ceara",
        32587: "Korean Neurointensive Care Society",
        32588: "Korean System Dynamics Society",
        32589: "Kyiv National Linguistic University",
        32590: "Folklor Akademii Dergiisi",
        32591: "Center for Open Access in Science",
        32592: "DoNotEdit",
        32593: "Shifa Tameer-e-Millat University",
        32594: "The International Association of Innovative"
               "Scientists (IAIS)",
        32595: "Melange Publications",
        32596: "Turkiye Kalp Ve Saglik Vakfi Iktisadi Isletmesi",
        32597: "James White Library",
        32598: "Negah Scientific Publisher",
        32599: "The Institute of Management and Economy Research",
        32600: "Hacettepe Universitesi Edebiyat Fakultesi Dergisi",
        32601: "Eurasian Journal of Applied Linguistics",
        32602: "Journal Of Accounting Finance And Auditing Studies (JAFAS)",
        32603: "St. Petersburg Electrotechnical University LETI",
        32604: "Computers, Materials and Continua (Tech Science Press)",
        32605: "Korean Enneagram Dynamic Psychological Academy",
        32606: "Indian Journal of Scientific Research",
        32607: "Acta Naturae Ltd",
        32608: "Institute of World History Russian Academy of Science",
        32609: "NP Voprosy Ekonomiki",
        32610: "Waveland Scientific Publications",
        32611: "Academic Association of Global Cultural Contents",
        32612: "University of Warsaw - Faculty of Applied Linguistics",
        32613: "The University of Notre Dame Australia",
        32614: "The R Foundation",
        32615: "Institute of Experimental Botany",
        32616: "Sekolah Tinggi Ilmu Tarbiyah Raden Wijaya Mojokerto",
        32617: "Entrepreneur and Innovation Exchange (EIX.org)",
        32618: "Polygraph",
        32619: "Delta State College of Education",
        32620: "National Aerospace University - Kharkiv Aviation Institute",
        32621: "Real Escuela Superior de Arte Dramatico",
        32622: "MG Aricent Private Limited",
        32623: "Nina B. Hollis Institute at Stetson University",
        32624: "Global Institute for Japanese Studies - Japanese"
               "Studies Journal",
        32625: "Korean Society for European Integration",
        32626: "Kamianets-Podilskyi National Ivan Ohiienko University",
        32627: "Universitas Masoem",
        32628: "Technoscience Academy",
        32629: "Frontier Scientific Publishing Pte Ltd",
        32630: "Badan Perencananaan Pembangunan Daerah Penelitian"
               "dan Pengembangan Kabupaten Sragen",
        32631: "Kharkiv National University of Internal Affairs",
        32632: "Ewha Legal Science Institute",
        32633: "Croatian Biological Society",
        32634: "Agrarian Science",
        32635: "Revista Brasileira De Cancerologia (RBC)",
        32636: "Institute of Agriculture of Carpathian Region of NAAS",
        32637: "Perhimpunan Dokter Spesialis Telinga Hidung"
               "Tenggorok Bedah Kepala Leher Indonesia - PERHATI-KL",
        32638: "AEE INTEC",
        32639: "P4M STIE Putra Bangsa",
        32640: "The Academic Society",
        32641: "Sociedad Chilena de Pediatria",
        32642: "University of Namibia",
        32643: "Hungarian Geographical Society (Magyar Foldrajzi Tarsasag)",
        32644: "BeyHealth UK Limited",
        32645: "Universidad Politecnica Estatal del Carchi - UPEC",
        32646: "Institute of Experimental Botany",
        32647: "Revista Brasileira de Mastozoologia",
        32648: "Proskolar Publishers",
        32649: "Anbar Journal of Agricultural Sciences",
        32650: "Asociacion Procompal",
        32651: "Economics of Agriculture of Russia",
        32652: "National University of Ukraine on Physical"
               "Education and Sport",
        32653: "Institute of History, Archeology and Ethnography",
        32654: "Universidad Nacional de Educacion Enrique Guzman y Valle",
        32655: "Research (NTU)",
        32656: "Research (Non NTU)",
        32657: "NYU Students Work",
        32658: "NTU Other",
        32659: "Stiepar Yapari Bandung",
        32660: "STIKES RS Baptis Kediri",
        32661: "Sekolah Tinggi Ilmu Hukum Muhammadiyah Takengon",
        32662: "Universitas Gorontalo",
        32663: "Universitas Prof. Dr. Hazairin SH",
        32664: "Sekolah Tinggi Informatika dan Komputer Indonesia",
        32665: "Universitas Nahdlathul Ulama Sunan Giri Bojonegoro",
        32666: "Sekolah Tinggi Ilmu Administrasi Mataram",
        32667: "Rumah Sakit Penyakit Infeksi Prof Dr. Sulianti Saroso",
        32668: "Poltekkes Kemenkes Jakarta",
        32669: "Village Institute",
        32670: "Institut Manajemen Koperasi Indonesia",
        32671: "Consejo Nacional de Acreditacion en Informatica"
               "y Computacion A.C.",
        32672: "Universitas Serambi Mekkah",
        32673: "Boletim da Sociedade Brasileira de Mastozoologia",
        32674: "STAR Scholars Network",
        32675: "Academy of Social Enterprise",
        32676: "Notitia D.O.O",
        32677: "Mansa STM Publishers",
        32678: "Lembaga Penelitian dan Pengabdian kepada Masyarakat"
               "(LP2M) Universitas Islam Negeri (UIN) Sultan"
               "Maulana Hasanuddin Banten",
        32679: "Puslitbang Sumber Daya Air Kementerian PUPR",
        32680: "Scientific Bulletin of the Odessa National"
               "Economic University",
        32681: "Institute for German Studies Seoul National University",
        32682: "STKIP PGRI Jombang",
        32683: "Novosibirsk State University of Architecture"
               "and Civil Engineering (Sibstrin)",
        32684: "Odessa State Academy of Technical Regulation and Quality",
        32685: "Servicio Geologico Colombiano",
        32686: "JSC FPH Business Express",
        32687: "Joint-Stock Company Chicot",
        32688: "Editora Bonecker",
        32689: "Interregional Academy of Personnel Management",
        32690: "University of Warsaw - Faculty of Oriental Studies",
        32691: "Institute of Philosophy and Law SB RAS",
        32692: "Darshan Institute of Engineering and Technology",
        32693: "Marine Geological Institute",
        32694: "Fakultas Syariah Institut Agama Islam Negeri Kerinci",
        32695: "Poltekkes Kemenkes Maluku",
        32696: "Universitas Muslim Nusantara Al Washliyah Medan",
        32697: "Komisi Pemberantasan Korupsi",
        32698: "RedWhite Press",
        32699: "LP3M Universitas Sains Al Quran",
        32700: "STIKes Maharani Malang",
        32701: "Faculty of Philosophy and Religious Studies",
        32702: "DKS Center",
        32703: "State University of Infrastructure and Technologies",
        32704: "Erdem",
        32705: "Yorum-Yonetim-Yontem Uluslararasi Yonetim-Ekonomi"
               "ve Felsefe Dergisi",
        32706: "Mersin Universitesi Beden Egitimi ve Spor Yuksekokulu",
        32707: "Erciyes Universitesi Veteriner Fakultesi Dergisi",
        32708: "Uludag Universitesi Tip Fakultesi Dergisi",
        32709: "Afyon Kocatepe Universitesi Sosyal Bilimler Dergisi",
        32710: "Tekstil Ve Konfeksiyon",
        32711: "Turkiye Ilahiyat Arastirmalari Dergisi",
        32712: "Fiocruz - Instituto de Tecnologia em Farmacos",
        32713: "Revista Direito and Paz",
        32714: "Research in Corpus Linguistics",
        32715: "Korean Society for Migration History",
        32716: "The SNU Society of Labor Law",
        32717: "Institute of Vegetable and Melon Growing of NAAS of Ukraine",
        32718: "Stepan Gzhytskyi National University of Veterinary"
               "Medicine and  Biotechnologies Lviv",
        32719: "Universidad Andina Simon Bolivar",
        32720: "International Digital Image Correlation Society",
        32721: "Canadian Tax Foundation",
        32722: "Politeknik Negeri Jakarta",
        32723: "Pipa Comunicacao Editorial",
        32724: "Phanes",
        32725: "University of South Bohemia in Ceske Budejovice",
        32726: "Foundation for Historical Outlook",
        32727: "Kennesaw State University, Horace W. Sturgis Library",
        32728: "Juraj Dobrila University of Pula",
        32729: "Puslitbang Pendidikan Agama dan Keagamaan"
               "Balitbang Kementerian Agama RI",
        32730: "Siec Badawcza Lukasiewicz - Instytut Metalurgii"
               "Zelaza im. St. Staszica",
        32731: "FiT Publishing",
        32732: "Tech Reviews Ltd",
        32733: "Integration ON LLC",
        32734: "Universitas Sumatera Utara",
        32735: "Universidad de Los Lagos, Chile",
        32736: "Sekolah Tinggi Manajemen Informatika & Komputer Atma Luhur",
        32737: "Institute of Catalysis and Inorganic Chemistry",
        32738: "Association of Engineering, Project, and Production"
               "Management",
        32739: "Uskudar University",
        32740: "Basrah Medical College - Department of Surgery",
        32741: "Eurofins NDSC Chimie Alimentaire",
        32742: "Max Weber Stiftung",
        32743: "Siberian Academic Book LLC",
        32744: "Perspectives of Science and Education",
        32745: "MWV Medizinisch Wissenschaftliche Verlagsgesellschaft"
               "mbH and Co. KG",
        32746: "Indian Journal of Mental Health and Neurosciences",
        32747: "US Department of Agriculture (USDA): National"
               "Agricultural Library (NAL)",
        32748: "Revista de Estudos de Cultura",
        32749: "Revista Cientifica Multidisciplinar Nucleo Do Conhecimento",
        32750: "European Scientific Journal Of Economic And"
               "Financial Innovation",
        32751: "Publishing House Lyudmila",
        32752: "National Academy of Management",
        32753: "Journal of Specialist",
        32754: "Militar Tehnical Academy Bucharest",
        32755: "Academy Of The State Penitentiary Servise",
        32756: "Pharmapress LLC",
        32757: "LabPro Media Limited Liability Company",
        32758: "The TEHINFORM Publishing Centre",
        32759: "Institute of Electronic Buddhist Texts & Culture Contents",
        32760: "REDD - Revista Espaco de Dialogo e Desconexao",
        32761: "Korea Society for Indian Philosophy",
        32762: "University of Rijeka, Faculty of Civil Engineering",
        32763: "Relawan Jurnal Indonesia",
        32764: "Relawan Jurnal Indonesia",
        32765: "Balai Besar Industri Agro",
        32766: "Real Academia Galega",
        32767: "Relawan Jurnal Indonesia",
        32768: "Archives of Breast Cancer",
        32769: "EnFoque Digital",
        32770: "Social Science and Business Research Network",
        32771: "Indonesian Society of Obstetrics and Gynecology",
        32772: "New World Publication Group",
        32773: "Medieval Institute Publications",
        32774: "Earth Science Dept, University of Florence",
        32775: "Laboratorio Nacional de Clima Espacial Instituto"
               "de Geofisica",
        32776: "Autonomous University of Yucatan",
        32777: "Researcher - European Journal of Humanities"
               "and Social Sciences",
        32778: "Scholar, LLC",
        32779: "Udruga Futura",
        32780: "National Research Foundation of Korea",
        32781: "Relawan Jurnal Indonesia",
        32782: "Publishing House Helvetica",
        32783: "Center for Sustainable Infrastructure Development",
        32784: "Universitatsverlag Gottingen",
        32785: "Mesford Publisher Inc",
        32786: "Volgograd State Agrarian University",
        32787: "Asosiasi Ilmu Hubungan Internasional Indonesia",
        32788: "Kazakh Ablai Khan University of International"
               "Relations and World Languages",
        32789: "iConferences",
        32790: "King Abdulaziz University -  Saudi Society"
               "of Internal Medicine",
        32791: "International Society for Orthomolecular Medicine",
        32792: "University of Thi Qar",
        32793: "Pengurus Besar Persatuan Dokter Gigi Indonesia",
        32794: "ST Bio-life",
        32795: "Universitas Hindu Indonesia",
        32796: "MINCOTUR",
        32797: "International Association for Time Use Research (IATUR)",
        32798: "University of Warsaw - Faculty of Polish Studies",
        32799: "Waakebiness-Bryce Institute for Indigenous Health",
        32800: "Consorci del Museu de Ciencies Naturals de Barcelona",
        32801: "Program Magister Kenotariatan Fakultas Hukum"
               "Universitas Lambung Mangkurat",
        32802: "Academy of Sciences Malaysia",
        32803: "European Federation of Associations and Centres"
               "of Irish Studies (EFACIS)",
        32804: "ISARA Solutions",
        32805: "Sekolah Tinggi Ilmu Kesehatan Jembrana",
        32806: "LP2M Institut Agama Islam Al-Khairat",
        32807: "Poltekkes Kemenkes Mataram",
        32808: "Faculdade Boas Novas de Ciencia Teologicas,"
               "Sociais e Biotecnologicas - FBNCTSB",
        32809: "Revista Brasileira de Nutricao",
        32810: "Revista Informacao em Pauta",
        32811: "Revista de Saude Publica do Parana",
        32812: "Sekolah Tinggi Ilmu Ekonomi Asia",
        32813: "Revista Ciencias Humanas",
        32814: "Sekolah Tinggi Ilmu Farmasi Makassar",
        32815: "STMIK Asia Malang",
        32816: "Fakultas Teknik Universitas Langlangbuana",
        32817: "Faculty of Science, University of Split",
        32818: "Sociedad Espanola de Medicina Interna - SEMI",
        32819: "Dnipro State Agrarian and Economic University",
        32820: "Ukrainian State Engineering Pedagogical Academy",
        32821: "Korean Association of Psychological Rehabilitation"
               "for Children",
        32822: "Scientific Research Institute of Oil and Gas"
               "Industry of National Joint-Stock Company Naftogaz of Ukraine",
        32823: "The Korean Society for Creative Information Culture",
        32824: "Facultad de Psicologia - Universidad de Buenos Aires",
        32825: "Publicaciones Latinoamericanas SRL",
        32826: "Cuadernos de Economia",
        32827: "Community Health Society Malaysia",
        32828: "Al-Mustansiriyah University - College of Dentistry",
        32829: "Centro de Investigacion y Capacitacion para"
               "el Desarrollo Regional (CINCADER)",
        32830: "Pusat Penelitian Dan Pemberdayaan Fiskal Dan Ekonomi Daerah",
        32831: "Akademi Keperawatan Dharma Husada Kediri",
        32832: "LPPM Universitas Ibn Khaldun Bogor",
        32833: "Fakultas Ekonomi dan Bisnis Islam Universitas"
               "Muhammadiyah Kendari",
        32834: "STIA LAN Jakarta",
        32835: "Institute of Vocational Education and Training,"
               "National Academy of Educational Sciences of Ukraine",
        32836: "University of Customs and Finance",
        32837: "National University - Odessa Law Academy",
        32838: "V.I. Vernadsky Taurida National University",
        32839: "Publishing House Young Scientist",
        32840: "Classic Private University",
        32841: "International Humanitarian University",
        32842: "Dniprovskii University of the Humanities",
        32843: "Black Sea Research Institute of Economics and Innovation",
        32844: "Private Institution Scientific Institute of Public Law",
        32845: "Sumy National Agrarian University",
        32846: "State Ecology Academy of Postgraduate Education"
               "and Management",
        32847: "International University of Business and Law",
        32848: "Institute of Irrigated Agriculture of the"
               "National Academy of Agrarian Sciences in Ukraine",
        32849: "Research Institute of Private Law and Entrepreneurship"
               "named after Academician F.H. Burchak of the"
               "National Academy of Legal Sciences of Ukraine",
        32850: "Odessa State University of Internal Affairs",
        32851: "Kherson State Agricultural University",
        32852: "University of Babylon - Department of Mechanical"
               "Engineering, Faculty of Engineering",
        32853: "Instituto Colombiano de Derecho Procesal",
        32854: "Colegio de Postgraduados",
        32855: "University of Texas at Arlington Libraries",
        32856: "Bangladesh Veterinary Association (BVA)",
        32857: "Paleontological Research Institution",
        32858: "Temblor",
        32859: "Ilia State University",
        32860: "Center for Glocal Culture and Social Empathy",
        32861: "Academic Research Publishing Group",
        32862: "Biblijski institut",
        32863: "Alta Astra LLC",
        32864: "The State Scientific Institution V.A. Belyi"
               "Metal-Polymer Research Institute of National"
               "Academy of Sciences",
        32865: "Texas Tech University",
        32866: "Network Design Lab - Transport Findings",
        32867: "Indonesian Society of Nephrology - InaSN",
        32868: "Saudi Food and Drug Authority",
        32869: "Edicions de la Universitat de Barcelona",
        32870: "Universidad de Guadalajara",
        32871: "University of San Jose-Recoletos",
        32872: "PsychOpen GOLD Leibniz Institute for Psychology (ZPID)",
        32873: "University of Nebraska Consortium of Libraries - UNCL",
        32874: "Zgodovinsko drustvo dr. Franca Kovacica v Mariboru",
        32875: "Association of Private International and Comparative"
               "Law Studies",
        32876: "Institute of Applied Astronomy of the Russian"
               "Academy of Sciences",
        32877: "Komunitas Dosen Indonesia",
        32878: "LLC M-Center",
        32879: "Islamia College Peshawar - Department of Psychology",
        32880: "Hypothekai",
        32881: "Aperio Press",
        32882: "Sprak- och litteraturcentrum, Lunds universitet",
        32883: "Sekolah Tinggi Ilmu Kesehatan Fort de Kock Bukittinggi",
        32884: "Ideas Publishing",
        32885: "The Institute of the Osteopathy and Holistic Medicine",
        32886: "Legislation Institute of The Verkhovna Rada of Ukraine",
        32887: "Revista do Departamento de Ciencias Sociais da UNIMONTES",
        32888: "Revista Cadernos de Gestao e Empreendedorismo",
        32889: "Indonesian Society for Biochemistry and Molecular Biology",
        32890: "UUM Press, Universiti Utara Malaysia",
        32891: "City Space Architecture",
        32892: "Medkrux",
        32893: "The International Journal of Business and"
               "Management Research",
        32894: "Kirkuk University",
        32895: "University of Medicine and Pharmacy at Ho Chi Minh City",
        32896: "Longe Medikal Sdn. Bhd",
        32897: "Universitas Sangga Buana YPKP",
        32898: "Naaj Consultancy",
        32899: "Tanta University",
        32900: "Institute of Animal Science of the National"
               "Academy of Agrarian Sciences of Ukraine",
        32901: "Melitopol Institute Of Public And Municipal"
               "Administration Of The Classic Private University",
        32902: "Communicable Diseases Intensive Care Association",
        32903: "Josip Juraj Strossmayer University of Osijek"
               "- Faculty of Education",
        32904: "ARRIV and W – Branch of the FSBSI FRARC",
        32905: "Revista Eletronica Simbio-Logias",
        32906: "Ministry of Education of Azerbaijan Republic",
        32907: "Research Outreach",
        32908: "Old City Publishing, Inc",
        32909: "Croatian Cartographic Society",
        32910: "Croatian Economic Association",
        32911: "Universidad Nacional Santiago Antunez de Mayolo",
        32912: "Tamaki Paenga Hira Auckland Museum",
        32913: "MIC Journal of Information and Communications Technology",
        32914: "Croatian Communication Association",
        32915: "Associacao Internacional Uniao das Americas",
        32916: "Inter-American Journal of Development and Research",
        32917: "Hiroshima University - Department of Mathematics",
        32918: "State Scientific and Technical Center for"
               "Nuclear and Radiation Safety",
        32919: "Department of Informatics and Cybernetics"
               "of Melitopol Bohdan Khmelnytsky State Pedagogical University",
        32920: "Ryerson University Library and Archives",
        32921: "Republican Center for Health Development",
        32922: "Poltekkes Kemenkes Pangkalpinang",
        32923: "IAIN Syaikh Abdurrahman Siddik Bangka Belitung",
        32924: "Sekolah Tinggi Manajemen IPMI",
        32925: "Turk Dili Arastırmalari Yillıgi - Belleten",
        32926: "Mutual Images",
        32927: "Stowarzyszenie Centrum Badan nad  Zaglada Zydow",
        32928: "The Korean Academy of Long-Term Care",
        32929: "Revista Cultura Agronomica",
        32930: "Nuances Estudos Sobre Educacao",
        32931: "A. N. Nesmeyanov Institute of Organoelement"
               "Compounds of the Russian Academy of Sciences - INEOS RAS",
        32932: "Publicacoes Ciencia e Vida, Lda",
        32933: "Research and Innovation Centre Pro-Akademia",
        32934: "Perkumpulan Dosen Pendidikan Guru Madrasah"
               "Ibtidaiyah se-Indonesia",
        32935: "TUMA GROUP",
        32936: "Prizren Social Science Journal",
        32937: "Meghana Publications",
        32938: "Universitas Timor",
        32939: "Institut Agama Islam Negeri Kerinci",
        32940: "Perkumpulan Sarjana Manajemen Pendidikan Islam Indonesia",
        32941: "Guncel Pediatri",
        32942: "EcoEvoRxiv",
        32943: "Banglarxiv",
        32944: "Persatuan Perawat Nasional Indonesia",
        32945: "Annals of Tropical Medicine, Visayas State University",
        32946: "Paleocultural Research Group",
        32947: "Al Mustansiriyah University - College of Pharmacy",
        32948: "Asian Medical Press Limited",
        32949: "Arkhaia Anatolika Anadolu Arkeolojisi Arastırmaları Dergisi",
        32950: "Recep Tayyip Erdogan Universitesi Ilahiyat Fakultesi Dergisi",
        32951: "Muhasebe ve Finans Incelemeleri Dergisi",
        32952: "Ataturk Iletisim Dergisi",
        32953: "Anadolu ve Balkan Arastirmalari Dergisi",
        32954: "Luqman Publishers",
        32955: "Near East University",
        32956: "Korean Association Of Computers and Accounting",
        32957: "Hacettepe Hukuk Fakultesi Dergisi",
        32958: "Gastroia Journal of Gastronomy and Travel Research",
        32959: "Turkish Online Journal of English Language Teaching",
        32960: "Ulusal Egitim Akademisi Dergisi",
        32961: "The Korean Society of Western Civilization",
        32962: "ANO Editorial Board Melioration and Water Management",
        32963: "Boletim do Curso de Medicina da UFSC",
        32964: "TAPPI",
        32965: "Viseu",
        32966: "Szellem es Tudomany",
        32967: "Egeszsegtudomanyi Kozlemenyek",
        32968: "Production Systems and Information Engineering",
        32969: "Vasbetonepites",
        32970: "Concrete Structures",
        32971: "Advanced Logistic Systems, Theory and Practice",
        32972: "Design of Machines and Structures",
        32973: "Journal of Computational and Applied Mechanics",
        32974: "Materials Science and Engineering",
        32975: "Gesta",
        32976: "Eszak-magyarorszagi Strategiai Fuzetek",
        32977: "Alkalmazott Nyelveszeti Kozlemenyek",
        32978: "Publicationes Universitatis Miskolciensis,"
               "Sectio Juridica et Politica",
        32979: "Publicationes Universitatis Miskolciensis,"
               "Sectio Philosophica",
        32980: "Miskolci Jogi Szemle",
        32981: "ATALAIA",
        32982: "Revista Brasileira de Saude e Seguranca no"
               "Trabalho - REBRAST",
        32983: "Research Centre of Industrial Problems of"
               "Development of NAS of Ukraine",
        32984: "Croatian Academy of Legal Sciences",
        32985: "Faculty of Electrical Engineering, Computer"
               "Science and Information Technology Osijek",
        32986: "Publishing House Triumph",
        32987: "State Scientific Institution - Institute of"
               "Educational Analytics",
        32988: "Publicacoes UNIFAL",
        32989: "Korea Association of Real Estate Law",
        32990: "Korean Society of Pediatric Critical Care Medicine",
        32991: "Historia Ambiental Latinoamericana y Caribena - HALAC",
        32992: "CEDLA - Centro de Estudios y Documentacion",
        32993: "The Public Association of Belorussian Stomatologists",
        32994: "Ragnar Nurkse Department of Innovation and Governance",
        32995: "Universidad Diego Portales",
        32996: "Al-Kindi Center for Research and Development",
        32997: "Universidad de Cartagena",
        32998: "AOTA Press",
        32999: "Kherson State University",
        33000: "MLS Psychology Research",
        33001: "Bait Al-Mashura Finance Consultations",
        33002: "The Grassroots Institute",
        33003: "Federal University Dutsin-Ma",
        33004: "Croatian Medical Association",
        33005: "University of Pembangunan Nasional Veteran Jawa Timur",
        33006: "LPPM STIKes Hafshawaty Pesantren Zainul Hasan Probolinggo",
        33007: "Puslitbangkesos Kementerian Sosial RI",
        33008: "Ubitech Solutions Pvt Ltd",
        33009: "Florida State University Libraries",
        33010: "Red de Investigadores Educativos Chihuahua AC",
        33011: "University of Colorado at Boulder",
        33012: "Institute of Navigation",
        33013: "Iraqi Scientific Business Management Society",
        33014: "The Blyth Institute",
        33015: "Dominican University of California",
        33016: "Nextjournal",
        33017: "Centro de Preparacion para la Ciencia y Tecnologia"
               "- Ceprecyt",
        33018: "Institute of Mechanics - National Academy"
               "of Sciences of Armenia",
        33019: "Universitas Bangka Belitung",
        33020: "Stmik Palangkaraya",
        33021: "Universitas Presiden",
        33022: "STMIK Indonesia Padang",
        33023: "LPPM STIKES Pemkab Jombang",
        33024: "Universitas Malahayati Bandar Lampung",
        33025: "Imperial Editora",
        33026: "Pegada Eletronica",
        33027: "Revista de Politicas Publicas do IPPMar",
        33028: "Ukrainian Scientific Research Institute of"
               "Ecological Problems",
        33029: "Geotar-Media Publishing Group",
        33030: "Geosciences and Engineering",
        33031: "Sarospataki Pedagogiai Fuzetek",
        33032: "Acta Carolus Robertus",
        33033: "Pro and Contra -  Central European Studies in Humanities",
        33034: "Paideia",
        33035: "Eger Journal of English Studies - EJES",
        33036: "Eger Journal of American Studies - EJAS",
        33037: "Konyv es Neveles",
        33038: "Journal of Central European Green Innovation - JCEGI",
        33039: "Annales Mathematicae et Informaticae - AMI",
        33040: "Acta Universitatis de Carolo Eszterhazy Nominatae"
               "Sectio Sport",
        33041: "Acta Universitatis de Carolo Eszterhazy Nominatae"
               "Sectio Biologiae",
        33042: "O.M.Beketov National University of Urban Economy in Kharkiv",
        33043: "Ball State University Library",
        33044: "Union Matematica Argentina",
        33045: "Research Institute for Fruit Growing",
        33046: "University of Debrecen/ Debreceni Egyetem",
        33047: "Perkumpulan Program Studi Pendidikan Guru"
               "Raudhatul Athfal Indonesia",
        33048: "Sobolev Institute of Mathematics",
        33049: "Scientific Adviser",
        33050: "iLearning Journal Center",
        33051: "Market Economy Institute of RAS - MEI RAS",
        33052: "Revista Interritorios",
        33053: "Fundacao Universidade de Cruz Alta",
        33054: "Revista da Abem",
        33055: "Editions Alphil",
        33056: "Editions Antipodes",
        33057: "Chronos Verlag",
        33058: "Seismo Verlag",
        33059: "Fakultas Hukum Universitas Samudra",
        33060: "STMIK Hang Tuah Pekanbaru",
        33061: "Universitas Slamet Riyadi",
        33062: "Universitas Binaniaga Indonesia",
        33063: "Uppsala University",
        33064: "Universidad Autonoma de Aguascalientes",
        33065: "Ulyanovsk State Pedagogical University",
        33066: "Akdema Bilisim Yayincilik ve Dan. Tic. Ltd. Sti",
        33067: "University of Warsaw - Centre for Europe",
        33068: "International Conference on Community Development - ICCD",
        33069: "Chronobiology in Medicine",
        33070: "The Gas Institute of the National Academy"
               "of Sciences of Ukraine",
        33071: "Social Science Research Institute - Locality"
               "and Globality - Korean Journal of Social Sciences",
        33072: "Philippine Society of Biochemistry and Molecular Biology",
        33073: "Polish Society of Microbiologists",
        33074: "Universitas Bosowa",
        33075: "Institute of Natural and Technical Systems",
        33076: "Turkish Dietetic Association Partner",
        33077: "University of Warsaw - Faculty of Journalism,"
               "Information and Book Studies",
        33078: "The Association Of Comparative Study Of World Literature",
        33079: "Journal of Microelectronic Manufacturing",
        33080: "STIA Al Gazali Barru",
        33081: "Revista Formacao",
        33082: "Odessa National Maritime University",
        33083: "Journal of Gastronomy Hospitality and Travel - JOGHAT",
        33084: "Universitas Muhammadiyah Palangkaraya",
        33085: "Institut Kesehatan Helvetia",
        33086: "Universitas Nahdlatul Ulama Surabaya",
        33087: "Universitas Batanghari Jambi",
        33088: "Poltekkes Kemenkes Bengkulu",
        33089: "Indonesian Society for Horticulture - Perhorti"
               "- Aceh Commissariat",
        33090: "Korea Safety Culture Institute",
        33091: "Al-Anbar Medical Journal",
        33092: "P.G. Sergiev Scientific and Educational Center"
               "of Parasitology",
        33093: "MUME00",
        33094: "International Academic Hub",
        33095: "University of Baghdad - College of Administration"
               "and Economics",
        33096: "Universitas Muslim Indonesia",
        33097: "International Next-Generation Convergence"
               "Technology Association",
        33098: "Private Higher Educational Institution King"
               "Danylo University",
        33099: "The National Defence University of Ukraine"
               "named after Ivan Cherniakhovskyi",
        33100: "University of Social Sciences and Humanities",
        33101: "Ukrmetallurginform STA Ltd",
        33102: "Universiti Sains Islam Malaysia",
        33103: "University of Technology, Baghdad",
        33104: "Balai Besar Industri Hasil Perkebunan",
        33105: "Direktorat Jenderal Perbendaharaan, Kementerian Keuangan",
        33106: "Baltic International Academy Ltd",
        33107: "University for Business and Technology",
        33108: "Ternopil Ivan Puluj National Technical University",
        33109: "Bangladesh Society for Veterinary Medicine",
        33110: "Instituto de Investigaciones Economicas y Empresariales",
        33111: "Kyiv National Economic University named after Vadym Hetman",
        33112: "The National and University Library of Iceland",
        33113: "Institute of Applied Mechanics of Russian"
               "Academy of Sciences",
        33114: "Academy for Design Innovation Management",
        33115: "Edicions A Peticio",
        33116: "Purnomo Yusgiantoro Center",
        33117: "Makerere University Business School",
        33118: "Open Access Publishers",
        33119: "Warsaw School of Economics",
        33120: "Institute for Social and Political Psychology,"
               "National Academy of Educational Science of Ukraine",
        33121: "Fakultas Hukum Universitas Narotama",
        33122: "SAINTIS Publishing",
        33123: "STIKes Kharisma Karawang",
        33124: "TKUMA Ukrainian Institute for Holocaust Studies",
        33125: "Soonchunhyang Exceptional Children Institute",
        33126: "Institute For Business & Economics Research",
        33127: "The Korean Association for Political and Diplomatic History",
        33128: "Hrvatsko Agronomsko Drustvo",
        33129: "METAS Adventist College",
        33130: "Asian Journal of Convergence in Technology",
        33131: "Institucion Universitaria Pascual Bravo",
        33132: "Fundacion Universitaria del Area Andina",
        33133: "Fundacion Universitaria Agraria de Colombia",
        33134: "Helsinki University Press",
        33135: "Society for Reformed Life Theology",
        33136: "Yuzhnoye State Design Office",
        33137: "University of Toronto Libraries - UOTL",
        33138: "University of Warsaw - Faculty of Economics Sciences",
        33139: "Rogers Publishing Corporation NFP",
        33140: "Opast Group LLC",
        33141: "Towarzystwo Naukowe Organizacji i Kierownictwa",
        33142: "Viser Technology Pte Ltd",
        33143: "Universitas Ubudiyah",
        33144: "ESRG Journals",
        33145: "National Research Center for Radiation Medicine"
               "of the NAMS of Ukraine",
        33146: "Institute of Accounting and Finance",
        33147: "Institute of Life Science and Natural Resources",
        33148: "Fundacao Joaquim Nabuco",
        33149: "Redbiz the Laboratory of Medical Business",
        33150: "Journal of ICT - Design - Engineering and"
               "Technological Science",
        33151: "Australasian College of Paramedicine",
        33152: "Journal of Management Practices - Humanities"
               "and Social Sciences",
        33153: "Institut Seni Indonesia Surakarta",
        33154: "Instituto Tecnologico de Sonora",
        33155: "Centro Andaluz de Medicina del Deporte",
        33156: "Institute of Philosophy in Chung-Ang University",
        33157: "Turkluk Arastırmaları Dergisi",
        33158: "Editora Mecenas Ltda",
        33159: "Revista Paulista de Enfermagem-REPEn",
        33160: "Tottori University Faculty of Medicine",
        33161: "Society of Korean Leprologists",
        33162: "The Korean Reliability Society",
        33163: "XMLink",
        33164: "BPSDMP Kominfo Yogyakarta",
        33165: "Faculty of Medicine Ramathibodi Hospital, Mahidol University",
        33166: "International Association for Educators and"
               "Researchers (IAER)",
        33167: "Instituto Superior de Ciencias Sociais e Politicas",
        33168: "Success Culture Press",
        33169: "S Open Access Open Journals Publishing SOAOJ",
        33170: "University of Babylon - Physical Education"
               "and Sports Sciences",
        33171: "Ankara University -  Faculty of Language and"
               "History-Geography",
        33172: "Indonesia Defense University",
        33173: "INFOTEKS",
        33174: "The Arctic Institute of North America (AINA)",
        33175: "Faculty of International Maritime Studies",
        33176: "Australasian Association for Clinical Biochemistry"
               "and Laboratory Medicine",
        33177: "Group of Studies of Asia and Latin America"
               "of the University of Buenos Aires",
        33178: "University College Cork",
        33179: "Ministry Of Defence - Slovenian Armed Forces",
        33180: "Drustvo MIDEM",
        33181: "The American College of Osteopathic Family"
               "Physicians - ACOFP",
        33182: "Transnational Press London",
        33183: "The West African Science Service Centre on"
               "Climate Change and Adapted Land Use",
        33184: "Bashkir State University",
        33185: "T-PRESS",
        33186: "State Public Scientific-Technical Library",
        33187: "Journal of Mathematical Sciences and Modelling",
        33188: "Veteriner Hekimler Dernegi Dergisi",
        33189: "Yarmachenko Institute of Special Education"
               "and Psychology of the NAES of Ukraine",
        33190: "Ukrainian Lingua-Information Fund Of The National"
               "Academy Of Sciences Of Ukraine",
        33191: "Revista Da Graduacao E Especializacao Em Direito",
        33192: "Faculty of Medicine Siriraj Hospital, Mahidol University",
        33193: "Emirates College for Education Sciences",
        33194: "Associacao Portuguesa dos Enfermeiros de Reabilitacao",
        33195: "University of Chitral",
        33196: "Verlag Osterreich GmbH",
        33197: "Universitas Widyatama",
        33198: "Universidad Privada Antonio Guillermo Urrelo",
        33199: "The Institute for Korean Christianity Culture",
        33200: "International Journal of Contemporary Educational Research",
        33201: "Iran Calismalari Dergisi",
        33202: "COMU Ziraat Fakultesi Dergisi",
        33203: "Maliye Finans Yazilari",
        33204: "Mucosa",
        33205: "Constructive Mathematical Analysis",
        33206: "MANAS Sosyal Arastirmalar Dergisi",
        33207: "Trakya Universitesi Edebiyat Fakultesi Dergisi",
        33208: "Revista Psicologia Clinica",
        33209: "Volyn Orthodox Theological Academy",
        33210: "Universidad Tecnologica Indoamerica",
        33211: "NHP Publications",
        33212: "Phoenix Publishing House Ltd",
        33213: "Yeongnam Toegye Studies Institute",
        33214: "Korean Ecological Engineering Society",
        33215: "SEISENSE Private, Ltd.",
        33216: "Volodymyr Dahl East Ukrainian National University",
        33217: "Massey University Library",
        33218: "Andover House Inc",
        33219: "Korea Credit Guarantee Fund",
        33220: "Ukrainian Research Institute of Forestry and"
               "Forest Melioration named after G. M. Vysotsky",
        33221: "Sekolah Tinggi Ilmu Kesehatan Indonesia Maju - STIKIM",
        33222: "STKIP Muhammadiyah Kuningan",
        33223: "Instytut Gospodarki Surowcami Mineralnymi"
               "i Energia Polskiej Akademii Nauk",
        33224: "Romanian Academy - Revue Roumaine De Chimie",
        33225: "Scientia Socialis Ltd",
        33226: "Polskie Wydawnictwo Ekonomiczne S.A.",
        33227: "Ankara Universitesi Ilahiyat Fakultesi Dergisi",
        33228: "SCRIBES - Brazilian Journal of Management"
               "and Secretarial Studies",
        33229: "Zwiazek Karaimow Polskich",
        33230: "Pusat Unggulan Riset-Penggembangan Lahan Suboptimal"
               "(PUR-PLSO)",
        33231: "Asociacion Espanola De Historia Economica",
        33232: "Maynooth University",
        33233: "Atlantica Editora",
        33234: "Playscript",
        33235: "Cambridge Media",
        33236: "North Caucasus University",
        33237: "Revista Geotemas",
        33238: "Revista Braseileira de Educacao em Ciencias"
               "e Educacao Matematica - ReBECEM",
        33239: "Revista Juridica Trabalho e Desenvolvimento Humano",
        33240: "Associacao Brasileira De Agroecologia",
        33241: "Cadernos do GPOSSHE On-line",
        33242: "Instituto Brasileiro de Direito Civil - IBDCivil",
        33243: "O.S. Popov Odesa National Academy of Telecommunications",
        33244: "University of the State Fiscal Service of Ukraine",
        33245: "The Bila Tserkva National Agrarian University",
        33246: "Shupyk National Medical Academy of Postgraduate Education",
        33247: "Actual Issues Of Clinicai And Preventive Medicine",
        33248: "Ukraine Business Publishing House",
        33249: "Zhytomyr National Agroecological University",
        33250: "Si Institute Of Pharmacology And Toxicology"
               "Of National Academy Of Medical Sciences Of Ukraine",
        33251: "Flight Academy of National Aviation University",
        33252: "Kangwon National University Humanities Institute",
        33253: "The Comments on the Korean Classical Texts",
        33254: "Institute of Archaeology",
        33255: "Universidad Nacional de Entre Rios",
        33256: "British Herpetological Society",
        33257: "Institute of Mechanics Lomonosov Moscow State University",
        33258: "Budapest International Research and Critics Institute",
        33259: "Journal of Livestock Science",
        33260: "ICT Association of Zambia",
        33261: "Society of Engineering Colleges and Institutes",
        33262: "Editorial Ciencia Digital",
        33263: "AMG Transcend Association",
        33264: "Universidad de las Artes y Ciencias de la"
               "Comunicacion - UNIACC",
        33265: "Norwegian Polar Institute",
        33266: "All-Russian Centre for Disaster Medicine “Zaschita”"
               "of Health Ministry of the RF",
        33267: "FGBNU - Rosinformagrotekh",
        33268: "Metszetek",
        33269: "The Ukrainian Civil Protection Research Institute",
        33270: "National Academy of Internal Affairs",
        33271: "Dnipro University of Technology",
        33272: "M. V. Ostrogradsky Poltava Regional Institute"
               "of In-Service Teacher Training",
        33273: "L. I. Medved Research Center of Preventive"
               "Toxicology, Food and Chemical Safety",
        33274: "Donetsk National University of Economics and"
               "Trade named after Mykhailo Tugan-Baranovsky",
        33275: "State Institution National Antarctic Scientific"
               "Center, Ministry of Education and Science of Ukraine",
        33276: "Central Economic Mathematical Institute of"
               "the Russian Academy of Sciences",
        33277: "CESaRE",
        33278: "Association System Economics",
        33279: "Journal of Khyber College of Dentistry",
        33280: "European University at Saint-Petersburg",
        33281: "Association for Tourism Development in Nigeria",
        33282: "University of Baghdad - College of Media",
        33283: "Springer Science and Business Media LLC",
        33284: "FSSI FRC BST RAS",
        33285: "Gubkin University",
        33286: "Omsk Scientific-Research Institute of Instrument"
               "Engineering - JSC ONIIP",
        33287: "Publishing house “Grani”",
        33288: "International Journal of Pharmaceutical, Chemical"
               "and Biological Sciences - IJPCBS",
        33289: "International Journal of Research in Pharmacy"
               "and Chemistry - IJRPC",
        33290: "Japan Transcatheter Valve Therapies",
        33291: "Museum of Anthropology and Ethnography, Kunstkamera, RAS",
        33292: "RESSI Journal",
        33293: "RPO for the Promotion of Institutes DE RAS",
        33294: "Institute of Religious Studies Lviv Museum"
               "of the History of Religion",
        33295: "Suchasna Stomatolohiya",
        33296: "School of Adaptive Management of Social And"
               "Pedagogical Systems",
        33297: "Badan Penelitian dan Pengembangan Daerah Kabupaten"
               "Luwu Utara",
        33298: "State University of Infrastructure and Technology",
        33299: "BPSDMP Kominfo Manado",
        33300: "Iqra University",
        33301: "National Economics University - Vietnam",
        33302: "Estonian National Museum",
        33303: "Performance Studies international",
        33304: "Universidad de Investigacion y Desarrollo",
        33305: "AIC - Economics, Management",
        33306: "Muallim Social Science and Humanities Society",
        33307: "Association for Advancement of Entomology",
        33308: "Istanbul Kultur University",
        33309: "Asclepius Open LLC",
        33310: "V ? Sukhomlynskyi Mykolaiv National University",
        33311: "Satu Mare College of Physicians",
        33312: "Indonesian Accountant Association - Compartment"
               "Educator Accountant",
        33313: "Association for Iron and Steel Technology",
        33314: "Nepal Health Research Council",
        33316: "The Chest and Heart Association of Bangladesh",
        33317: "Sir Syed University of Engineering and Technology",
        33318: "The University of the South Pacific",
        33319: "LPPM Universitas Merdeka Madiun",
        33320: "Macedonian Pharmaceutical Association",
        33321: "Australian Government Department of Health",
        33322: "Sekolah Tinggi Teknik-PLN",
        33323: "Universitas Kristen Artha Wacana",
        33324: "Universidad del Azuay",
        33325: "Healthcare Infection Society",
        33326: "Universidad Nacional Jorge Basadre Grohmann",
        33327: "East-European Law Research Center",
        33328: "The Egyptian Association of Medical Education",
        33329: "KY Publications",
        33330: "LPPM STMIK Royal",
        33331: "Badan Pembinaan Hukum Nasional Kementerian Hukum dan HAM RI",
        33332: "Pusat Survei Geologi",
        33333: "Escuela Politecnica Nacional",
        33334: "Sogang Institute for East Asian Studies (SIEAS)",
        33335: "The Society of Korean Language and Literature",
        33336: "Aikuiskasvatus",
        33337: "Arctos",
        33338: "Entomologica Fennica",
        33339: "Finnisch-Ugrische Forschungen",
        33340: "Suomalais-Ugrilaisen seuran Aikakauskirja",
        33341: "Suomalais-Ugrilainen Seura, Société Finno-Ougrienne",
        33342: "FMSERA Journal",
        33343: "Folk och musik",
        33344: "Helsinki Law Review",
        33345: "Idantutkimus",
        33346: "Joutsen/Svanen",
        33347: "Joutsen/Svanen. Erikoisjulkaisut.",
        33348: "Kasvatus",
        33349: "Instituto Andaluz de Patrimonio Historico - IAPH",
        33350: "Kasvatus ja Aika",
        33351: "Matkailututkimus",
        33352: "Prologi",
        33353: "Studia Celtica Fennica",
        33354: "Suomen Maataloustieteellisen Seuran Tiedote",
        33355: "Tekniikan waiheita",
        33356: "Temenos",
        33357: "Yhdyskuntasuunnittelu",
        33358: "Kansantaloudellinen Aikakauskirja",
        33359: "Finnish Economic Papers",
        33360: "Geonordeste",
        33361: "Revista Pesquisa Qualitativa - RPQ",
        33362: "Universidade Alto Vale Do Rio Do Peixe - Uniarp",
        33363: "Institut Agama Hindu Negeri Tampung Penyang Palangka Raya",
        33364: "Sekolah Tinggi Teknologi Garut",
        33365: "Universitas Teknokrat Indonesia",
        33366: "Unitri Press",
        33367: "Institut Agama Islam Tribakti Kediri",
        33368: "Fakultas Kesehatan Masyarakat Universitas Muslim Indonesia",
        33369: "UNIB Press",
        33370: "LP2M Sekolah Tinggi Ilmu Ekonomi IPWI Jakarta",
        33371: "National Cancer Center Dharmais Cancer Hospital",
        33372: "STMIK Amik Riau",
        33373: "Universitas Riau Kepulauan",
        33374: "Balai Konservasi Boorobudur",
        33375: "Visual Communication Design of Unikom",
        33376: "STIKOM INTERSTUDi",
        33377: "Sekolah Tinggi Ilmu Kesehatan RS Husada",
        33378: "Jurusan Penyuluhan Perikanan Sekolah Tinggi Perikanan",
        33379: "Universitas Islam Raden Rahmat Malang",
        33380: "Center of Pharmaceutical Analytics Ltd",
        33381: "Journal Of Medical And Surgical Intensive Care Medicine",
        33382: "Labour Law Association",
        33383: "Redakcia Zhurnala Svetotekhnika LLC",
        33384: "Siberian Research",
        33385: "e   National Collaborative for Improving the"
               "Clinical Learning Environment",
        33386: "CEIT S.A.",
        33387: "LPPM Universitas Khairun",
        33388: "The Korea Association For Industrial Security",
        33389: "Revista DESC - Direito Economia e Sociedade Contemporanea",
        33390: "HOMEROS",
        33391: "The Graduate School of the Humanities, Utrecht University",
        33392: "Diametros",
        33393: "Aboutscience Srl",
        33394: "LPPM IKIP Mataram",
        33395: "Politeknik Ganesha",
        33396: "Northern State Medical University",
        33397: "STATUT",
        33398: "Lviv National Music Academy named after Mykola Lysenko",
        33399: "Bingol Universitesi Iktisadi ve Idari Bilimler"
               "Fakultesi Dergisi",
        33400: "Kocaeli Universitesi Egitim Dergisi",
        33401: "Fundamental Journal of Mathematics and Applications",
        33402: "Ivan Krypyakevych Institute of Ukrainian Studies"
               "-  National Academy of Sciences of Ukraine",
        33403: "Review of International Geographical Education Online",
        33404: "ANASAY",
        33405: "National Academy of the National Guard of Ukraine",
        33406: "MOLESTO: Edebiyat Arastirmalari Dergisi",
        33407: "Institute of Information Technologies and"
               "Learning Tools of NAES of Ukraine",
        33408: "University of Civil Protection of the Ministry"
               "for Emergency Situations of the Republic of Belarus",
        33409: "Toprak Bilimi ve Bitki Besleme Dergisi",
        33410: "Hoseo Archaeological Society",
        33411: "50Sea",
        33412: "Universidad Tecnologica de Panama",
        33413: "National University Pedro Henriquez Urena",
        33414: "Universidad Tecnologica Nacional",
        33415: "Dinbilimleri Akademik Arastirma Dergisi",
        33416: "Isletme Ekonomi ve Yonetim Arastirmalari Dergisi",
        33417: "Toplum ve Sosyal Hizmet",
        33418: "Ataturk Universitesi Kazim Karabekir Egitim"
               "Fakultesi Dergisi",
        33419: "Ataturk Arastirma Merkezi Dergisi",
        33420: "Marife Dini Arastirmalar Dergisi",
        33421: "Instituto Nacional Materno Perinatal",
        33422: "Mokslines leidybos deimantas, MB",
        33423: "North American Business Press",
        33424: "Sci-Comm Consulting Ltd",
        33425: "SciVision Publishers LLC",
        33426: "Congreso de los Diputados",
        33427: "St. Elizabeth University of Health and Social"
               "Science - Bratislava, Slovakia",
        33428: "Villanova University",
        33429: "Central Bank of Nigeria",
        33430: "The Hong Kong Institution of Engineers",
        33431: "Belgi Dergisi",
        33432: "Yildirim Beyazit Hukuk Dergisi",
        33433: "Marmara Universitesi Hukuk Fakultesi Hukuk"
               "Arastirmalari Dergisi",
        33434: "Communications in Advanced Mathematical Sciences",
        33435: "Turkish Computational and Theoretical Chemistry",
        33436: "Romanian Journal of Information Technology"
               "and Automatic Control",
        33437: "Kahramanmaras Sutcu Imam Universitesi Sosyal"
               "Bilimler Dergisi",
        33438: "International Journal of Disabilities Sports"
               "and Health Sciences",
        33439: "Ergonomi",
        33440: "International Journal of Precision Agricultural Aviation",
        33441: "Josip Juraj Strossmayer University of Osijek,"
               "Faculty of Food Technology Osijek",
        33442: "Global Health Press Pte Ltd",
        33443: "Chungbuk National University National Technology"
               "Management Research Institute",
        33444: "Ukrainian Association of Neurology and Reflexology",
        33445: "Ukrainian Scientific Community",
        33446: "Korean Law and Society Association - KLSA",
        33447: "Paubrasilia",
        33448: "Research, Society and Development",
        33449: "Postgraduate of Universitas Bengkulu",
        33450: "Institute of Psychiatry and Neurology",
        33451: "SURYA",
        33452: "Asian Music Research Institute",
        33453: "Ukrainian Scientific Pharmacopoeial Center"
               "for Quality of Medicines",
        33454: "Postgraduate Institute for Public Health Worker"
               "under Health Ministry of Khabarovsk Krai",
        33455: "Indonesian Business and  Economics Research Network",
        33456: "Entrepreneurship and  Social Responsibility Research Network",
        33457: "International Journal of Health Services Research and Policy",
        33458: "Uluslararasi Iliskiler Dergisi",
        33459: "CBU Beden Egitimi ve Spor Bilimleri Dergisi",
        33460: "BEU Ilahiyat Fakultesi Dergisi",
        33461: "Uluslararasi Yonetim Bilisim Sistemleri ve"
               "Bilgisayar Bilimleri Dergisi",
        33462: "Tekirdag Ziraat Fakultesi Dergisi",
        33463: "Academy of the FPS of Russia",
        33464: "Dorduncu Kuvvet Uluslararasi Hakemli Dergi",
        33465: "Otraslevije Vedomosty Publishing House Ltd",
        33466: "Institute of Ethnocultural Education",
        33467: "Convergencias em Ciencia da Informacao",
        33468: "Umit Dogan Ustun",
        33469: "International Journal of Ancient History",
        33470: "Marshall University School of Medicine",
        33471: "Dong-Eui University Institute of Local Autonomy",
        33472: "African Journal of Biological Sciences",
        33473: "Riksawan Institute",
        33474: "Universitas Islam Malang",
        33475: "STIKES Widyagama Husada",
        33476: "Lembaga Penelitian Universitas YARSI",
        33477: "LP2M IAIN Ambon",
        33478: "Universitas Teknologi Surabaya",
        33479: "Universitas Ma Chung",
        33480: "PPPM STMIK Nusa Mandiri",
        33481: "STMIK Sumedang",
        33482: "STIKES Karsa Husada Garut",
        33483: "Ankara Universitesi Eczacilik Fakultesi Dergisi",
        33484: "Sinop Universitesi Fen Bilimleri Dergisi",
        33485: "Stikes Kapuas Raya",
        33486: "Akademi Kebidanan Panca Bhakti",
        33487: "STKIP Muhammadiyah Enrekang",
        33488: "Sekolah Tinggi Ilmu Komputer Yos Sudarso",
        33489: "Akademi Maritim Yogyakarta",
        33490: "Poltekkes Kemenkes Mamuju",
        33491: "Ille Mikhail Evgenievich",
        33492: "Australasian College of Road Safety",
        33493: "Mizo Academy of Sciences",
        33494: "Scion",
        33495: "Scienceweb Publishing",
        33496: "Gordon College of Education",
        33497: "Society for Philosophy of Emotion",
        33498: "LLC Publishing House In Jure",
        33499: "The Renaissance Group",
        33500: "BluePen Journals Publishing House",
        33501: "Revista Etos",
        33502: "The Korean Society for Multicultural Health",
        33503: "IKIP Budi Utomo Malang",
        33504: "Politeknik Manufaktur Negeri Bangka Belitung",
        33505: "Ikatan Sarjana Ilmu Perpustakaan dan Informasi"
               "Indonesia - ISIPII",
        33506: "Universitas Muhammadiyah Sorong",
        33507: "Pascasarjana IAINU Kebumen",
        33508: "Universitas Katolik Widya Mandala Surabaya",
        33509: "Sekolah Tinggi Ilmu Administrasi-Lembaga Adminstrasi"
               "Negara - STIA LAN - Makassar",
        33510: "Universitas Matana",
        33511: "Institut Ilmu Al Quran - IIQ - Jakarta",
        33512: "Fakultas Pertanian Universitas Sultan Ageng Tirtayasa",
        33513: "Ocimum Scientific Publishers Pty Ltd",
        33514: "Kyrgyz State University. I. Arabaev",
        33515: "C-International Archive",
        33516: "The Institute of Cost Accountants of India",
        33517: "Revista Uruguaya de Enfermeria",
        33518: "Institute of Humanities Research SCNU",
        33519: "Korea Wind Energy Association",
        33520: "AURAICA. Scripta a Societate Porthan edita",
        33521: "The Institute for Buddhist Study, Joong-Ang"
               "Sangha University",
        33522: "Korea National Maritime Museum",
        33523: "Korean Society of Intraoperative Neurophysiological"
               "monitoring",
        33524: "Nipissing University Library",
        33525: "Ordem dos Psicologos Portugueses",
        33526: "The British Association for Korean Studies",
        33527: "Chonnam Research Institute of Nursing Science",
        33528: "Korea Institute for Military Affairs",
        33529: "Russian Society of Angiologists and Vascular Surgeons",
        33530: "Innovative Defense Acquisition Society",
        33531: "Accent Graphics Communications and Publishing",
        33532: "Revemop",
        33533: "Fakultas Kedokteran UPN Veteran Jakarta",
        33534: "Centre for Security Governance",
        33535: "Dar Almandumah Inc",
        33536: "Indonesian Muslim University",
        33537: "Ankara University Graduate School of Social Sciences",
        33538: "University Library of Pecs",
        33539: "Universidad Femenina Del Sagrado Corazon",
        33540: "Utrecht University Library",
        33541: "UKI Press",
        33542: "Pavol Jozef Safarik University in Kosice",
        33543: "MAGNANIMITAS",
        33544: "Mapta Publishing Group",
        33545: "Comprehensive Publications",
        33546: "Belitung Nursing Journal",
        33547: "Wydawnictwo i Pracownia Archeologiczna PROFIL-ARCHEO",
        33548: "Science Diffusion Ltd",
        33549: "Institute of Physiology of the Czech Academy of Sciences",
        33550: "Reformed Center for Religion and Society",
        33551: "Universitas Negeri Semarang - Departmant of"
               "Guidance and Counseling",
        33552: "Iris Publishers LLC",
        33553: "Australian Tactical Medical Association",
        33554: "Universidad Nacional Hermilio Valdizan",
        33555: "Academic Research and Community Service Swiss"
               "German University",
        33556: "STIMART AMNI Semarang",
        33557: "Universitas Bina Darma",
        33558: "Universitas Islam 45",
        33559: "Lembaga Penelitian dan Penerbitan Hasil Penelitian"
               "Ensiklopedia",
        33560: "APTIRMIKI",
        33561: "Fakultas Hukum Universitas Halu Oleo",
        33562: "Korean Society for Naturopathy",
        33563: "Korea Social Safety Criminal Intelligence Association",
        33564: "IJEAST",
        33565: "Multidiszciplinaris kihivasok, sokszinu valaszok",
        33566: "Acta Scientiarum Socialum",
        33567: "Elelmiszer, taplalkozas es marketing",
        33568: "Regional and Business Studies",
        33569: "Anyanyelvi Kulturakozvetites",
        33570: "Central European journal of Gastroenterology and Hepatology",
        33571: "Politecnico Colombiano Jaime Isaza Cadavid",
        33572: "Jordan Engineers Association",
        33573: "Kundiiev Institute of Occupational Health"
               "of National Academy of Medical Sciences of Ukraine",
        33574: "Hellenic Society of Obstetrics and Gynecology",
        33575: "Korean Chinese Relations Institute",
        33576: "Korea Entertainment Society",
        33577: "Hetman Petro Sahaidachnyi National Army Academy",
        33578: "Program Studi PGSD FKIP Universitas Riau",
        33579: "STTNAS is Sekolah Tinggi Teknologi Nasional",
        33580: "Ovchinnikov Mikhail Arturovich",
        33581: "Belarusian State University",
        33582: "MedDocs Publishers LLC",
        33583: "Honor Frost Foundation",
        33584: "New Zealand Grassland Association",
        33585: "Czech Scientific Society for Mycology",
        33586: "CINTER",
        33587: "Front Science Inc.",
        33588: "Viguera Editores SLU",
        33589: "International Society of Hair Restoration Surgery",
        33590: "European Medical Group",
        33591: "College of Family Physicians Singapore",
        33592: "Lembaga Penelitian dan Pengabdian Masyarakat,"
               "Universitas Islam Syekh-Yusuf Tangerang",
        33593: "International Society for Concrete Pavements, Inc",
        33594: "Cell Physiol Biochem Press GmbH and Co KG",
        33595: "National University of the Altiplano - Professional"
               "School of Communication Sciences",
        33596: "University of Miami",
        33597: "Medtext Publishing Services Pvt Ltd",
        33598: "Albanian Sports Science Association",
        33599: "Society for the Advancement of Material and"
               "Process Engineering - SAMPE",
        33600: "Illinois Reading Council",
        33601: "Rukmini Devi Institute Of Advanced Studies",
        33602: "Croatian Society of Biologists in Health Care",
        33603: "Universitas swadaya Gunung Djati",
        33604: "Leksikografski zavod Miroslav Krleza",
        33605: "Academic Association of Management  and Administration"
               "- AVADA",
        33606: "Yonsei Law Association",
        33607: "Lithuanian Sports University",
        33608: "Shevchenko Institute of Literature of the"
               "National Academy of Sciences of Ukraine",
        33609: "V.I. Vernadsky Institute of General and Inorganic Chemistry",
        33610: "Instituto Nacional de Salud - Direccion de"
               "Vigilancia y Analisis del Riesgo en Salud Publica",
        33611: "AMED iD3 Catalyst Unit",
        33612: "University of Groningen",
        33613: "Antropoloji",
        33614: "Korean Correction Counseling Psychological Association",
        33615: "The Korean Society of Oral Health Science",
        33616: "Lege Artis Medicinae",
        33617: "UPO - Association Of Traditional And Non-"
               "Traditional Medicine Specialists Of Ukraine",
        33618: "Kibogora Polytechnic",
        33619: "Publishing Center Science and Practice",
        33620: "Edittec",
        33621: "DIGSUM",
        33622: "Izdatelstvo PGS",
        33623: "Faculty of Geology Lomonosov Moscow State University",
        33624: "Farwater LLC",
        33625: "Kharkiv State Academy of Design and Arts",
        33626: "Badan Penelitian dan Pengembangan Provinsi Sumatera Utara",
        33627: "Sekolah Tinggi Keguruan dan Ilmu Pendidikan (STKIP) Bim",
        33628: "Universitatea Alexandru Ioan Cuza Iasi",
        33629: "Ankara Universitesi Hukuk Fakultesi Dergisi",
        33630: "Ankara Universitesi Siyasal Bilgiler Fakultesi",
        33631: "Duzce Universitesi Saglik Bilimleri Enstitusu Dergisi",
        33632: "Federal State Budgetary Scientific Institution"
               "Federal Center for Toxicological, Radiation"
               "and Biological Safety",
        33633: "Universitas Dian Nuswantoro",
        33634: "Pesquisa Multidisciplinar em Ciencias Cardiovasculares",
        33635: "Faculdade de Tecnologia de Sertaozinho",
        33636: "Revista de Constitucionalizacao do Direito"
               "Brasileiro - RECONTO",
        33637: "Laborare",
        33638: "Institute for Human Studies of Kyungnam University",
        33639: "Research Institute for East-West Thought",
        33640: "University of Kerbala",
        33641: "The Korean Association For Lexicography - Korealex",
        33642: "The Center for Promoting Education and Research (CPER)",
        33643: "R. Glier Kyiv Institute of Music",
        33644: "State Enterprise - State Research Institute"
               "of Building Constructions",
        33645: "The Korean Society of Culture and Convergence",
        33646: "Universidade de Vigo - UVigo",
        33647: "Scientific Center for Biomedical Technologies"
               "of the Federal Medical-Biological Agency",
        33648: "Sekolah Tinggi Agama Islam - STAI - Al-Furqan Makassar",
        33649: "Politeknik Pariwisata Makassar",
        33650: "Universitas Nurul Jadid",
        33651: "Politeknik Medica Farma Husada Mataram",
        33652: "Balai Pelestarian Nilai Budaya Kalimantan Barat",
        33653: "STIKes Perintis Padang",
        33654: "STKIP PGRI Banjarmasin",
        33655: "Akademi Keperawatan Widya Husada Semarang",
        33656: "Masyarakat Pernaskahan Nusantara",
        33657: "HB Press STIKes Husada Borneo",
        33658: "Badan Perencanaan Pembangunan Daerah Kabupaten Pati",
        33659: "STKIP Paris Barantai",
        33660: "Akademi Fisioterapi Widya Husada Semarang",
        33661: "Fakultas Pertanian, Universitas Singaperbangsa Karawang",
        33662: "Cadernos do Tempo Presente",
        33663: "Koretsky Institute of State and Law of National"
               "Academy of Sciences of Ukraine",
        33664: "Japan Society for Studies in Journalism and"
               "Mass Communication",
        33665: "Electronic Scientists and Engineers Society",
        33666: "STIKES Widya Husada  Semarang",
        33667: "Alfmed LLC",
        33668: "Hypertonia es Nephrologia",
        33669: "Institut za jezik",
        33670: "Centro de Investigacion sobre Desarrollo Humano"
               "y Sociedad (Coideso)",
        33671: "LASALLE College of the Arts",
        33672: "Edicoes Esgotadas Publisher",
        33673: "Orkana Forlag AS",
        33674: "Wyzsza Szkola Biznesu i Przedsiebiorczosci"
               "w Ostrowcu Swietokrzyskim",
        33675: "Universitatverlag WINTER Heidelberg",
        33676: "EMUI EuroMed University",
        33677: "Azerbaijan National Academy of Sciences -"
               "Geology and Geophysics Institute",
        33678: "Czech Society of Cardiology",
        33679: "El Colegio de la Frontera Norte A.C.",
        33680: "Plant Science Archives",
        33681: "Paco Editorial",
        33682: "New York University",
        33683: "Scuola Universitaria Professionale Della Svizzera"
               "ItalianaD. Formazione E Apprendimento",
        33684: "Gambling Research Exchange Ontario",
        33685: "JiL Scientific Research Center",
        33686: "Central  Power Research Institute",
        33687: "EScience Press",
        33688: "Cografi Bilimler Dergisi",
        33689: "Spormetre",
        33690: "Dil Dergisi",
        33691: "Institute of Industrial Economics of the National"
               "Academy of Sciences of Ukraine",
        33692: "Avrasya Uluslararası Arastırmalar Dergisi",
        33693: "YUR-VAK",
        33694: "Askania Nova Institute Of Animal Breeding"
               "In The Steppe Regions Named After M.F. Ivanov",
        33695: "Asociatia Cercul de Stiinte Chirurgicale",
        33696: "Scientific Archives LLC",
        33697: "American Journal of Undergraduate Research",
        33698: "Nursing Research Forum",
        33699: "Chirurgie-Servis z.s",
        33700: "Alma Mater Europaea - ECM",
        33701: "Institut Pemerintahan Dalam Negeri",
        33702: "Raffles Connect Pte Ltd",
        33703: "Korean Association of Cognitive Behavioral Therapy",
        33704: "Al Bashaer Economic Journal",
        33705: "El Lougha Arabia Revue",
        33706: "Journal of Emergency Medicine Case Reports",
        33707: "Afyon Kocatepe Universitesi Iktisadi ve Idari"
               "Bilimler Fakultesi Dergisi",
        33708: "Akdeniz Kadin Calismalari ve Toplumsal Cinsiyet Dergisi",
        33709: "Ictimaiyat",
        33710: "SDU International Journal of Educational Studies",
        33711: "Yuzuncu Yil Universitesi Egitim Fakultesi Dergisi",
        33712: "Uluslararasi Yonetim Akademisi Dergisi",
        33713: "Ege Tip Bilimleri Dergisi",
        33714: "Marine Science and Technology Bulletin",
        33715: "Inonu Universitesi Saglik Hizmetleri Meslek"
               "Yuksek Okulu Dergisi",
        33716: "Balikesir Medical Journal",
        33717: "Dokuz Eylul Universitesi Hukuk Fakultesi Dergisi",
        33718: "Trabzon Ilahiyat Dergisi",
        33719: "Avrasya Uroonkoloji Dernegi",
        33720: "Karaelmas Is Sagligi ve Guvenligi Dergisi",
        33721: "Bilgi Yonetimi",
        33722: "Afro Eurasian Studies",
        33723: "R&S - Research Studies Anatolia Journal",
        33724: "Ziraat Muhendisligi",
        33725: "Mobilya ve Ahsap Malzeme Arastirmalari Dergisi",
        33726: "Revista AKEDIA - Versoes, Negligencias e Outros Mundos",
        33727: "Editura Thoth",
        33728: "Korea Institute for National Unification",
        33729: "The Korean Association For Persons With Autism",
        33730: "Institute of Agroecology and Environmental Management NAAN",
        33731: "Scientific Research Institute of Intellectual"
               "Property National Academy  of Law Sciences of Ukraine",
        33732: "Universidad Rey Juan Carlos",
        33733: "ACLC College of Butuan, Inc.",
        33734: "Fundacion Instituto Hipolito Unanue",
        33735: "MIND Group",
        33736: "UNIMAS Publisher",
        33737: "Global Power and Propulsion Society",
        33738: "Sociedad Peruana de Odontopediatria",
        33739: "International Scientific-Pedagogical Organization"
               "of Philologists - WEST-EAST",
        33740: "Science papers periodical of Donbas National"
               "Academy of Civil Engineering and Architecture",
        33741: "Institute of Haematology and Transfusiology"
               "of the National Academy of Medical Sciences of Ukraine",
        33742: "Polissia Institute of Agriculture of the National"
               "Academy of Agrarian Sciences",
        33743: "Institute of Dermatology and Vener?ology of"
               "National Academy of Medical Sciences of Ukraine",
        33744: "National Transport University",
        33745: "Saran Publications",
        33746: "Sekolah Tinggi Ilmu Kesehatan Faletehan",
        33747: "Sekolah Tinggi Ilmu Ekonomi Semarang",
        33748: "Perhimpunan Dokter Spesialis Radiologi Indonesia",
        33749: "Balai Riset dan Standardisasi Industri Manado",
        33750: "Center for Humanities and Innovation Studies",
        33751: "Universitas Pakuan",
        33752: "Universitas Hasyim Asy'ari Tebuireng Jombang",
        33753: "Lembaga Kajian Demokrasi dan Pemberdayaan Masyarakat",
        33754: "Institut Keislaman Abdullah Faqih - INKAFA - Gresik",
        33755: "STIKep PPNI Jawa Barat",
        33756: "Fakultas Hukum Universitas Negeri Gorontalo",
        33757: "STIKes Alifah Padang",
        33758: "LPSDI Bina Patria",
        33759: "APDFI - Asosiasi Pendidikan Diploma Farmasi Indonesia",
        33760: "LPPM STIH Putri Maharaja",
        33761: "Poltekkes Kemenkes Padang",
        33762: "University of Basrah, College of Medicine",
        33763: "State Educational-Scientific Establishment"
               "The Academy of Financial Management",
        33764: "Siberian State University of Geosystems and Technologies",
        33765: "Association for Promotion of Holistic Approach"
               "to Environment",
        33766: "Luhansk State University of Internal Affairs"
               "named after E.O. Didorenko",
        33767: "MediArXiv",
        33768: "Korea Society of Urban Environment",
        33769: "Communications Faculty of Sciences_University"
               "of Ankara_A2-A3",
        33770: "The Korean Society of Emotional and Behavioral Disorders",
        33771: "Indira Institute of Business Management",
        33772: "Universitas Halu Oleo",
        33773: "Journal of Universal Mathematics",
        33774: "Cambridge University Press pre-prints",
        33775: "Federal State Budgetary Scientific Institution"
               "All-Russian Rice Research Institute",
        33776: "Universidad de Huelva - UHU.pdf",
        33777: "Clinical Journal of Korean Association of Orthodontists",
        33778: "Korea Convergence Security Association",
        33779: "Scholarly-Methodical Center Innovative Art Studies",
        33780: "Revista Sinergia",
        33781: "Revista da Gama e Souza",
        33782: "Scientific Research Center Lukomorie",
        33783: "Berdyansk University Of Management and Business Ltd",
        33784: "Myronivka bulletin",
        33785: "Indian Dairy Association",
        33786: "e-Current Science",
        33787: "Centro de Estudos Geograficos (IGOT) Universidade de Lisboa",
        33788: "Expert Projects",
        33789: "Universidad Estatal de Bolivar",
        33790: "Gexin Publication",
        33791: "Academy of Medical Optics and Optometry",
        33792: "BlueOcean Perspectives Entwicklungsgesellschaft"
               "mbH and Co KG",
        33793: "Academic Perspective",
        33794: "Al-Qadisyiah Uni - Al-Qadisiyah Journal For"
               "Agriculture Sciences - QJAS",
        33795: "Politeknik Negeri Malang",
        33796: "West Africa Built Environment Research (WABER) Conference",
        33797: "The Centre for Conscious Design",
        33798: "Mecon Publications Association",
        33799: "Joklarannsoknafelag Islands",
        33800: "Museo Nacional de Historia Natural, Prof."
               "Eugenio de Jesus Marcano",
        33801: "Centro de Investigaciones Comerciales e Iniciativas"
               "Academicas",
        33802: "University of New Brunswick Libraries - UNB",
        33803: "Dedan Kimathi University of Technology",
        33804: "EScience Press Inc",
        33805: "Edelweiss Publications Inc",
        33806: "The Association of Professors of English and"
               "Translation at Arab Universities - APETAU",
        33807: "Institute for Scientific Research and Development",
        33808: "Clinical and Experimental Health Sciences",
        33809: "Revista Ciencia e Sustentabilidade",
        33810: "Seminar Nasional Struktural",
        33811: "Dar Al-utroha For Scientific Publishing",
        33812: "Egyptian Journal of Legal and Economic Studies",
        33813: "Private Enterprise PIIR",
        33814: "Federal Williams Research Center of Forage"
               "Production and Agroecology",
        33815: "Kherson State Maritime Academy",
        33816: "Grupo de Institutos Fundacoes e Empresas",
        33817: "Muhakeme Dergisi",
        33818: "International Econometric Review",
        33819: "Stockholmia - forskning och forlag",
        33820: "Perdoski",
        33821: "Sociedad de Lucha contra el Cancer del Ecuador"
               "- Matriz Guayaquil",
        33822: "FISIP UPN Veteran Jakarta",
        33823: "H-Net: Humanities & Social Sciences Online",
        33824: "National Institute of Psychology, Centre of"
               "Excellence, Quaid-i-Azam University",
        33825: "Academic Times",
        33826: "International Journal of Multidisciplinary"
               "Research And Studies",
        33827: "Directorate General of Budget - Finance Ministry"
               "of Indonesia",
        33828: "International Council of Associations for Science Education",
        33829: "Universidad Privada SISE",
        33830: "Universitas Terbuka",
        33831: "Eastern Mediterranean University - EMU",
        33832: "NADIA",
        33833: "Cifra Ltd - Russian Agency for Digital Standardization"
               "(RADS)",
        33834: "Brazilian Keynesian Review",
        33835: "The Cheonggye Historical Society",
        33836: "Interacao",
        33837: "Multi-Science Journal",
        33838: "National Academy of Fine Arts and Architecture",
        33839: "V. N. Bakul Institute for Superhard Materials,"
               "National Academy of Sciences of Ukraine",
        33840: "SI Institute of Nuclear Medicine and Diagnostic"
               "Radiology of NAMS of Ukraine",
        33841: "Modern Achievements of Geodesic Science and Industry",
        33842: "Bogdan Khmelnitsky Melitopol State Pedagogical University",
        33843: "Journal of Babylon Center for Humanities Studies",
        33844: "Canadian Institute for Knowledge Development",
        33845: "LLC AVIAN",
        33846: "Forum Ilmiah Kesehatan - Forikes",
        33847: "Institute of Certified Specialists",
        33848: "Pavlov First State Medical University",
        33849: "Joint Institute for High Temperatures of Russian"
               "Academy of Sciences",
        33850: "The Arab Institution for Education, Science and Arts",
        33851: "Korea Multimedia Society - English Version Journal",
        33852: "Lembaga Talif wa An-Nasyr - LTN - PBNU",
        33853: "STIT Islamic Village",
        33854: "Universitas Baiturrahmah",
        33855: "University of Babylon - The College Of Education"
               "For Humanitarian Studies",
        33856: "Sekolah Tinggi Teologi Injili Indonesia Surabaya",
        33857: "Universitas Patria Artha",
        33858: "The Universite Hassiba Benbouali de Chlef",
        33859: "LPPM Universitas Sari Mulia",
        33860: "Poltekkes Kemenkes Palu",
        33861: "Krasnodar Regional Public Veterinary Organization",
        33862: "STIKES Citra Delima Bangka Belitung",
        33863: "P3M - Politeknik Perkapalan Negeri Surabaya",
        33864: "METAFIZIKA International Journal of Philosophy"
               "and Interdisciplinary Studies",
        33865: "Scientific Platform",
        33866: "Pakistan Phytopathological Society",
        33867: "Sekolah Tinggi Ilmu Kesehatan Aisyiyah Bandung",
        33868: "The State Enterprise - The State Road Transport"
               "Research Institute - SE SRTRI",
        33869: "Proceedings Of The Ukrainian Mineralogical Society",
        33870: "Center of Scientific - Research and Statistical Innovation",
        33871: "Universidade Estadual do Parana - Unespar",
        33872: "Editora Universitaria EduFatecie",
        33873: "Russian Research Institute of Economics -"
               "Politics and Law in Science and Technology",
        33874: "The All Russian State University of Justice",
        33875: "Izdatelstvo - Soliton",
        33876: "Institute of Ethnology and Anthropology RAS",
        33877: "Association of Korean Genealogical Studies",
        33878: "Russian Association of Coloproctology",
        33879: "Full Universe Integrated Marketing Ltd",
        33880: "Eurasian Society of Family Medicine",
        33881: "Corporacion Universitaria Iberoamericana",
        33882: "Indian Dental Association",
        33883: "Sheri-i-Kashmir Institute of Medical Sciences",
        33884: "Universitas Putera Batam",
        33885: "Asociacion Mexicana para el Estudio de los Hongos AC",
        33886: "Kenyatta University",
        33887: "Research and Reviews",
        33888: "Multidisciplines",
        33889: "International Journal of Mathematical, Engineering"
               "and Management Sciences plus Mangey Ram",
        33890: "Universidad Internacional del Ecuador",
        33891: "Fogorvosi Szemle",
        33892: "Acta Pharmaceutica Hungarica",
        33893: "Financial and Economic Review",
        33894: "Muszaki Tudomanyos Kozlemenyek",
        33895: "Muszaki Tudomanyos Kozlemenyek",
        33896: "ELIPSA Dom Wydawniczy i Handlowy",
        33897: "Foundation University",
        33898: "Revista de Psiquiatria y Psicologia Humanista, S.L.",
        33899: "University of Mosul",
        33900: "The Korean Association for Policy Studies",
        33901: "Limited Liability Company AST 345",
        33902: "Journal of Pedagogical Research",
        33903: "Politeknik Meta Industri",
        33904: "European Journal of Forest Engineering",
        33905: "Bilecik Seyh Edebali Universitesi Sosyal Bilimler"
               "Enstitusu Dergisi",
        33906: "Musicologist",
        33907: "Turkish Journal of Educational Studies",
        33908: "Economy and Finance",
        33909: "Centre for Development of Advanced Computing",
        33910: "Herzens Russian Pedagogical University",
        33911: "Centro Universitario Luterano de Palmas",
        33912: "Revista PesquisAgro",
        33913: "Pratibha Spandan",
        33914: "Neelain University",
        33915: "West Virginia University Libraries",
        33916: "Al-Qadisiyah Journal of Administrative and Economics",
        33917: "Institute for Economic Strategies",
        33918: "Lietuvos Istorijos Institutas",
        33919: "New Bulgarian University",
        33920: "PANORAMA Publishing House",
        33921: "JIRIRI",
        33922: "Umudike Journal of Engineering and Technology",
        33923: "Acta Materialia Transylvanica",
        33924: "Acta Materialia Transylvanica",
        33925: "Periodontal Association - RPA",
        33926: "Gazdasag es Penzugy",
        33927: "University of Pannonia",
        33928: "Botanical Society of Britain and Ireland",
        33929: "The FaithX Project",
        33930: "Scientific Information Agency"
               "Science-technologies-information",
        33931: "Abant Izzet Baysal Universitesi Ilahiyat Fakultesi Dergisi",
        33932: "Institute of Industrial and Business Management"
               "of Kyungnam University",
        33933: "Russian State Hydrometeorological University",
        33934: "Initium",
        33935: "Sabinet Online Ltd",
        33936: "Empresa Publica de Servicios Generales e Ingenieria Utm",
        33937: "Universidad Autonoma de Yucatan Facultad de Economia",
        33938: "FSBSI FRC AESDRA VNIIESH",
        33939: "Postgraduate Institute of Management",
        33940: "Patient Safety Authority",
        33941: "Information Age",
        33942: "International University of Innovation Technologies",
        33943: "Journal of Dairy and Beef Cattle Farming",
        33944: "Yektaweb Publishing Company",
        33945: "Sami Publishing Company",
        33946: "Revista Atuacao - Ministerio Publico de Santa Catarina",
        33947: "Revistas Cientificas Eletronicas UNG",
        33948: "King Saud University",
        33949: "Korean Womens Development Institute",
        33950: "RSC Energia",
        33951: "The Arab Centre for Research and Studies in"
               "Library and Information Sciences",
        33952: "Federal State Budgetary Scientific Institution"
               "Scientific Research Institute of Agriculture of Crimea",
        33953: "Remah Research and Human Resources Development Center",
        33954: "Institute of International Studies - IIS",
        33955: "SE Ukrmetrteststandart",
        33956: "Tensoes Mundiais",
        33957: "Bangjunghwan Research Institute",
        33958: "Instituto Geologico",
        33959: "City University of Science and Information"
               "Technology, Peshawar",
        33960: "Fundacion Carolina",
        33961: "The Korean Electrochemical Society - English Journal",
        33962: "Romanian Society of Neurosurgery",
        33963: "Polskie Towarzystwo Kardiologiczne",
        33964: "Perum BULOG",
        33965: "IADIS - International Association for the"
               "Development of the Information Society",
        33966: "Health Education and Training Institute - HETI",
        33967: "Editorial Santa Paula SRL",
        33968: "HTWK Leipzig",
        33969: "Institute of Electronics and Computer",
        33970: "Instituto Tecnico Superior Espiritu Santo",
        33971: "University of Basrah - College of Engineering"
               "- Engineering Science Journal",
        33972: "Foley Center Library, Gonzaga University",
        33973: "International Society of Pleural Diseases",
        33974: "Rubatosis Publication",
        33975: "Universidad del Quindio",
        33976: "Islamic University of Gaza",
        33977: "Al-Quds Open University",
        33978: "Medforum",
        33979: "Orel State University",
        33980: "ERESEARCHCO",
        33981: "Trinity Trust",
        33982: "The Law Research Institute, Chungnam National University",
        33983: "Avtonomnaya nekommercheskaya organizatsiya"
               "Akademiya menedzhmenta i biznes-administrirovaniya",
        33984: "Scientific Journal of Physical Education,"
               "College of Physical Education, University of Babylon",
        33985: "Al Albayt University",
        33986: "Jinan University",
        33987: "Odessa National Economic University",
        33988: "Veterinary Journal of Ankara University",
        33989: "Poltava V.G.Korolenko National Pedagogical University",
        33990: "Lviv Regional Institute for Public Administration"
               "- LRIPA NAPA",
        33991: "Sekolah Tinggi Teologi Torsina",
        33992: "Jurusan Analis Kesehatan Poltekkes Kemenkes Denpasar",
        33993: "Academia Romana Filiala Cluj",
        33994: "Kyiv Scientific Research Institute of Forensic Expertise",
        33995: "The Polish Insurance Association",
        33996: "Centro de Investigacion y Desarrollo Ecuador",
        33997: "Asian Fisheries Society",
        33998: "LPPM STIKOM Dinamika Bangsa Jambi",
        33999: "The Courtauld Institute of Art",
        34000: "Association Monegasque pour la Recerche Academique"
               "sur le Yoga - AMARY",
        34001: "Centre for Research and Community Development"
               "- Islamic University of Nahdlatul Ulama Jepara",
        34002: "Mindanao State University - General Santos City",
        34003: "Sekolah Tinggi Ilmu Pertanian Muhammadiyah Sinjai",
        34004: "Yayasan SEKORA - Sekolah Olahraga",
        34005: "Universitas Islam As-Syafiiyah",
        34006: "Sekolah Tinggi Ilmu Ekonomi Indragiri - STIE-I - Rengat",
        34007: "Mahesa Research Institute",
        34008: "Akademi Keperawatan Kesdam I/Bukit Barisan Medan",
        34009: "Al-Anbar University journal of Economic and"
               "Administration Sciences, University of Anbar",
        34010: "Universitas Komputer Indonesia",
        34011: "Politeknik Kesehatan Kemenkes Bandung",
        34012: "Universitas Prima Indonesia",
        34013: "Sekolah Tinggi Pariwisata Bandung",
        34014: "Ulyanovsk State University",
        34015: "Penitentiary Association of Ukraine",
        34016: "Pakistan Journal of Biotechnology - PJBT",
        34017: "Personality Development Center HUMANUS",
        34018: "Associacao Brasileira de Cognicao e Artes Musicais",
        34019: "Universidade Federal de Juiz de Fora",
        34020: "Novosibirsk State University of Economics"
               "and Management - NSUEM",
        34021: "The London Academy of Science and Business Limited",
        34022: "All-Russian Research Institute of Work",
        34023: "Information and Publishing Centre Statistics of Russia",
        34024: "Universidade Federal de Sao Paulo",
        34025: "Chernivtsi Institute of Trade and Economics"
               "of Kyiv National University of Trade and Economics",
        34026: "Kyiv National Karpenko-Karyi University of"
               "Theater, Cinema and Television",
        34027: "University of Palestine",
        34028: "Zarqa University",
        34029: "Railway Transport of Ukraine",
        34030: "Limited Liability Company Publishing Service Internauka",
        34031: "BSTU named after V.G. Shukhov",
        34032: "Revista Ambientale",
        34033: "Revista IPT: Tecnologia e Informacao",
        34034: "Turkiye Odyologlar ve Konusma Bozukluklari Uzmanlari Dernegi",
        34035: "Kusuma Husada University",
        34036: "Organization for the Study of Communication"
               "Language and Gender",
        34037: "RISCOS - Associacao Portuguesa de Riscos Prevencao"
               "e Seguranca",
        34038: "Hansen Science Publishing Company Inc",
        34039: "Grupo de Investigacion - Hospital Cardiovascular"
               "del Nino de Cundinamarca",
        34040: "Bioclues Organization",
        34041: "Foreningen Lambda Nordica",
        34042: "College Language Association",
        34043: "North American Association of Christians in"
               "Social Work - NACSW",
        34044: "Kasetsart University and Development Institute",
        34045: "rubmedia AG",
        34046: "Admiral Ushakov State Maritime University",
        34047: "Imperial Publications",
        34048: "Advanced Computing and Communications Society - ACCS",
        34049: "Institute of Chemical Engineering, Bulgarian"
               "Academy of Sciences",
        34050: "Faculty of Cultural Sciences, Hasanuddin University",
        34051: "University of New Hampshire Libraries",
        34052: "Instituto de Salud Publica de Chile",
        34053: "University of Arkansas Press",
        34054: "Promote Ukraine",
        34055: "BodoArXiv",
        34056: "Anadolu Universitesi Egitim Fakultesi Dergisi",
        34057: "International Society of Pelviperineology - ISPP",
        34058: "Nigeria Physiotherapy Network",
        34059: "Revista Cientifica Jopef",
        34060: "Revista Esmat",
        34061: "Universidade Estadual de Alagoas",
        34062: "Advances in Forestry Science",
        34063: "Institut Ilmu Kesehatan Medika Persada Bali",
        34064: "Kharkiv I.P. Kotlyarevsky National University of Arts",
        34065: "Al-Aqsa University",
        34066: "Arab Planning Institute",
        34067: "American Society of Nephrology (ASN)",
        34068: "Clemson University",
        34069: "Amazonia Investiga",
        34070: "Instituto Superior Tecnologico de Formacion"
               "Profesional Administrativa y Comercial",
        34071: "Hue University of Medicine and Pharmacy",
        34072: "Ohio University Press",
        34073: "Universidad de Moron",
        34074: "Otago Polytechnic",
        34075: "Katolicki bogoslovni fakultet Sveucilista u Splitu",
        34076: "Ural State Law University",
        34077: "Rzhanov Institute Semiconductor Physics SB RAS",
        34078: "North-East Interdisciplinary Scientific Research"
               "Institute N.A. Shilo (NEISRI FEB RAS)",
        34079: "Mariupol State University",
        34080: "Ornis Svecica",
        34081: "Sekolah Tinggi Teologi Tawangmangu",
        34082: "Islami Ilimler Dergisi",
        34083: "Akademik Dil ve Edebiyat Dergisi",
        34084: "Journal of Biotechnology and Strategic Health Research",
        34085: "Bingol Universitesi Ilahiyat Fakultesi Dergisi",
        34086: "Recep Tayyip Erdogan Universitesi Sosyal Bilimler Dergisi",
        34087: "Celal Bayar Universitesi Saglık Bilimleri Enstitusu Dergisi",
        34088: "Kocaeli Journal of Science and Engineering",
        34089: "Korean Society of Nursing Research",
        34090: "Turist Rehberligi Dergisi",
        34091: "Abasyn University",
        34092: "Centro de Estudios Vallejianos",
        34093: "University of Baghdad - The Post-Graduate"
               "Institute for Accounting and Financial Studies - PGIAFS",
        34094: "LLC Medical Press",
        34095: "Turkish Spine Society",
        34096: "Editorial de la Facultad de Filosofia y Letras"
               "- Universidad de Buenos Aires",
        34097: "Communication Institute of Greece",
        34098: "Universidad Mayor de San Andres UMSA",
        34099: "European Star Academic Publishing Ltd",
        34100: "Tejgazdasag",
        34101: "Acta Agraria Debreceniensis",
        34102: "Italianistica Debreceniensis",
        34103: "Argumentum",
        34104: "Universe Publishing Group - UniversePG",
        34105: "Laboratory for Knowledge Management and E-Learning"
               "- The University of Hong Kong",
        34106: "The World of Science and Education",
        34107: "International Academic Express",
        34108: "Saglik Bilimleri Dergisi - Erciyes Universitesi",
        34109: "International Journal of Economics and Finance Studies",
        34110: "Turkish Journal of Forecasting",
        34111: "International Journal of eBusiness and eGovernment"
               "Studies - The Social Sciences Research Society",
        34112: "Revista Linha Mestra",
        34113: "Revista Leitura Teoria e Pratica",
        34114: "Fundacao de Apoio e Desenvolvimento ao Ensino,"
               "Pesquisa e Extensao",
        34115: "Brazilian Applied Science Review",
        34116: "Szent Istvan Egyetem Novenytermesztesi Intezet",
        34117: "Brazilian Journal of Development",
        34118: "Amar Telidji University of Laghouat",
        34119: "Brazilian Journal of Health Review",
        34120: "Kuwait University",
        34121: "Institute of Mathematical Machines and Systems"
               "Problems of the NAS of Ukraine",
        34122: "Korea Institute of Intellectual Property",
        34123: "Politeknik Statistika STIS",
        34124: "Universitas Victory Sorong",
        34125: "STKIP Pesisir Selatan",
        34126: "Pusat Air Tanah dan Geologi Tata Lingkungan, Badan Geologi",
        34127: "Politeknik LP3I Jakarta",
        34128: "Politeknik Negeri Tanah Laut",
        34129: "Federal State Budgetary Institution of Science"
               " - Kazan Scientific Center of Russian Academy of Science",
        34130: "Pitirim Sorokin Syktyvkar State University",
        34131: "FSBO - Yu. A. Gagarin Research and Test Cosmonaut"
               "Training Center",
        34132: "Institute of State Management of Petro Mohyla"
               "Black Sea National University",
        34133: "AAAS00",
        34134: "CETAPS - Centre for English Translation and"
               "Anglo-Portuguese Studies",
        34135: "Faculty of Social Sciences, University of"
               "Saints Cyril and Methodius",
        34136: "SEDERI",
        34137: "The Journal of International Lingual Social"
               "and Educational Sciences",
        34138: "Korean Technology Education Association",
        34139: "Society for Standards Certification and Safety",
        34140: "Brazilian Journal of Business",
        34141: "Grupo Anltyk S.A. de C.V.",
        34142: "H. S. Skovoroda Kharkiv National Pedagogical University",
        34143: "The Korea Transport Institute",
        34144: "Korean Society of Exhibition Design Studies",
        34145: "Politeknik Pembangunan Pertanian Malang",
        34146: "Yayasan Darul Hasanah Aceh",
        34147: "Badan Penelitian dan Pengembangan Daerah Provinsi Jawa Barat",
        34148: "Institut Informatika Indonesia Surabaya",
        34149: "PPM School of Management",
        34150: "STKIP Widya Yuwana",
        34151: "Institut Sains and Teknologi AKPRIND Yogyakarta",
        34152: "STIE Pelita Nusantara Semarang",
        34153: "Russian Academy of Sciences",
        34154: "European Academy of Applied and Social Sciences (EURAASS)",
        34155: "Sona College of Technology",
        34156: "Haufe-Lexware GmbH and Co. KG",
        34157: "Haufe-Lexware GmbH and Co. KG",
        34158: "International Council of Onomastic Sciences - ICOS",
        34159: "Institute of KoguryoBalhae",
        34160: "The Korean Dysphagia Society",
        34161: "Korean Association of Health Technology Assessment",
        34162: "Institute of Human Environment and Future",
        34163: "Korea Knowledge Information Technology Society",
        34164: "Inje Institute of Democracy and Autonomy",
        34165: "The Incheon Institute",
        34166: "Korean Association of Future Military Affairs",
        34167: "The Child Developmental Supporting Center",
        34168: "Institute for Francophone Studies",
        34169: "Central Scientific Research Institute of Armaments"
               "and Military Equipment of Armed Forces of Ukraine",
        34170: "Sophia Prima - Dialogue of Eternal Recurrence",
        34171: "IUMS Press",
        34172: "Maad Rayan Publishing Company",
        34173: "South-Russian Musical Anthology",
        34174: "Mentouri University of Constantine",
        34175: "Journal of Nutritional Oncology",
        34176: "Revista Icone",
        34177: "Editora da Universidade de Sao Paulo",
        34178: "Journal of Bioengineering and Technology Applied to Health",
        34179: "Revista Sergipana de Matematica e Educacao"
               "Matematica - ReviSeM",
        34180: "Revista Defesa e Seguranca - RDS",
        34181: "Servico Nacional de Aprendizagem Comercial"
               "- Santa Catarina - SENAC -SC",
        34182: "Revista Brasileira de Pesquisa em Administracao",
        34183: "The Academy of Customer Satisfaction Management",
        34184: "Korean Society for Structural Biology",
        34185: "National Metallurgical Academy of Ukraine",
        34186: "Kirklareli Universitesi Muhendislik ve Fen Bilimleri Dergisi",
        34187: "Croatian Leather and Footwear Society",
        34188: "BJAER - Brazilian Journal of Animal and Environmental"
               "Research",
        34189: "Ankara Haci Bayram Veli Universitesi",
        34190: "Academic Conferences International Ltd",
        34191: "Utah Geological Survey",
        34192: "Ucimed",
        34193: "Cactus Communications - Editage",
        34194: "Geological Survey of Denmark and Greenland",
        34195: "Etcetera Publications",
        34196: "International Microsimulation Association",
        34197: "ATS Scholar",
        34198: "Earthline Publishers",
        34199: "Yayasan Rumah Peneleh",
        34200: "Universitas Wisnuwardhana  - Arthavidya Jurnal"
               "Ilmiah Ekonomi",
        34201: "Universitas Islam Kalimantan  - AT Tadbir",
        34202: "IMANENSI - Forum Dosen Ekonomi dan Bisnis Islam",
        34203: "Universitas Pakuan Jimfe - Jurnal Ilmiah Manajemen"
               "Fakultas Ekonomi",
        34204: "Universitas Pakuan - JIAFC",
        34205: "Universitas Muhammadiyah Gresik  - JIATAX",
        34206: "Universitas Tadulako - JIMUT",
        34207: "Universitas Kristen Indonesia Paulus - Makassar"
               "- Jurnal Akuntansi Uki Paulus",
        34208: "STIE Trisaksti - Jurnal Bisnis Dan Akuntansi",
        34209: "Universitas Pembangunan Negara Veteran Jakarta"
               "- Jurnal Equity",
        34210: "Universitas Trilogi - Trilogi Business Research",
        34211: "Universitas Warmadewa - WMBJ - Warmadewa Management"
               "and Business Journal",
        34212: "Universitas Warmadewa - Warmadewa Economic"
               "Development Journal",
        34213: "Kharkiv Regional Institute of Public Administration",
        34214: "Academy for Quality",
        34215: "Pacific State Medical University",
        34216: "Kostroma State University",
        34217: "Knowledge Sector Initiative",
        34218: "IAEME Publication Chennai",
        34219: "Informatization and Communication Journal Editorial Board",
        34220: "Voronezh State University of Forestry and"
               "Technologies named after G.F. Morozov",
        34221: "Korean Political Science Society",
        34222: "The Korean Association of Criminal Procedure Law",
        34223: "International Cultural Institute",
        34224: "Korean Association for  Sexology",
        34225: "Korea International Cooperation Agency - KOICA",
        34226: "Research Institute for Gifted and Talented"
               "Education, Soongsil University",
        34227: "Research Institute of Creative Contents",
        34228: "The Korean Hyo Studies Association",
        34229: "V.M. Glushkov Institute of Cybernetics",
        34230: "Filistin Arastirmalari Dergisi",
        34231: "Journal of Internet Applications and Management",
        34232: "Pamukkale Journal of Eurasian Socioeconomic Studies",
        34233: "Tavukculuk Arastirma Dergisi",
        34234: "Degerler Egitimi Dergisi",
        34235: "FlashPub",
        34236: "Revista Peruana de Investigacion Educativa",
        34237: "American Shore and Beach Preservation Association",
        34238: "Thai Nguyen University - Journal of Science and Technology",
        34239: "Australian Society of Plastic Surgeons",
        34240: "The American Fiction Association of Korea",
        34241: "Korea Local Tax Association",
        34242: "ARIS",
        34243: "China Area Studies Association of Korea",
        34244: "Future Oriented Youth Society",
        34245: "Seowon University - Educational Research Institute",
        34246: "Ankara Haci Bayram Veli Universitesi Hukuk Fakultesi Dergisi",
        34247: "Artuklu Akademi",
        34248: "Black Sea Journal of Engineering and Science",
        34249: "National Institute of Special Education",
        34250: "Korean Society of Critical Care Nursing",
        34251: "Korea International Finance Association",
        34252: "Asian Cultural Studies Research Institute",
        34253: "Yayasan Bina Darma",
        34254: "Trinexcor, LLC",
        34255: "Study Group",
        34256: "IOR Press",
        34257: "Global Journals",
        34258: "Cuadernos de Literaturas de Irrealidad",
        34259: "Kambohwell Publisher Enterprises",
        34260: "Abdul Wali Khan University Mardan, Pakistan",
        34261: "The Korean Lesson Study Group for Social Studies",
        34262: "The Journal of Developmental Disabilities",
        34263: "Society of Occupational Therapy for the Aged and Dementia",
        34264: "Korean Aging-Frendly Industry Association",
        34265: "The Research Institute For The Mahan-Baekje Culture",
        34266: "Public Policy Research Institute - Korea University",
        34267: "Law Research Institute - Chungbuk National University",
        34268: "The Hoseo Archaeological Society",
        34269: "Moving Image Technology Association of Korea",
        34270: "Korean Society for Rhythmic Exercises",
        34271: "Korea Reformed Theology Society",
        34272: "Korean Institute of Forest Recreation and Welfare",
        34273: "The Korea Academy Of Business Ethics",
        34274: "The Korean Research Association For The Business Education",
        34275: "Korean Institute for Buddhist Studies",
        34276: "Strategy and Development Review",
        34277: "AL- HIKMA Scientific Research Center",
        34278: "Journal of Anbar University of Islamic Sciences."
               "College of Islamic Sciences",
        34279: "Iraqi Association of Information Technology",
        34280: "Revista ANNEP de Direito Processual",
        34281: "The Korean Association of Buddhist Professors",
        34282: "Korea Regional Economics Association",
        34283: "Korean Society of Golf Studies",
        34284: "Korean Society of Growth and Development",
        34285: "National Prosecution Academy of Ukraine",
        34286: "MEGAPOLIS",
        34287: "Zaporizhia Medical Academy of Post-Graduate Education",
        34288: "Kresnamedia Publisher",
        34289: "Universitas Siliwangi",
        34290: "FFPP - Psychedelic Science Review",
        34291: "Faculty of Theology, University of Ljubljana",
        34292: "Medlife Scientific Press Inc",
        34293: "Shanlax International Journals",
        34294: "Mindtech S.A.S",
        34295: "Cuadernos de Investigaciones de Ciencias de la Informacion",
        34296: "Association of Counseling Sexology and Sexual Wellness",
        34297: "BiomedGrid LLC",
        34298: "Institute of History",
        34299: "Department of Medical Research",
        34300: "Al-Marhaba Research and Publication Society",
        34301: "London College of United Knowledge",
        34302: "Technical University of Cluj Napoca",
        34303: "The Musicological Society Of Korea - NRFK",
        34304: "Sekolah Tinggi Ilmu Hukum - STIH - Muhammadiyah Bima",
        34305: "Sekolah Tinggi Ilmu Kesehatan Kuningan Garawangi",
        34306: "Pandawan",
        34307: "Sekolah Tinggi Agama Kristen Negeri Toraja",
        34308: "STIE DR KHEZ Muttaqien",
        34309: "Yayasan Jurnal Perempuan",
        34310: "karya",
        34311: "Indonesian Chemical Society",
        34312: "niversitas Negeri Gorontalo  - Fakultas Matematika dan IPA",
        34313: "TVZ Theologischer Verlag Zurich AG",
        34314: "University of Cincinnati Press",
        34315: "The Association for Institutional Research",
        34647: "Sachsische Landesbibliothek, Staats-"
               " und Universitatsbibliothek Dresden",
        34671: "Scientific Chronograph EOOD",
        34690: "Cifra Ltd - Russian Agency for Digital Standardization"
               "(RADS)",
        34739: "Siedlce University of Natural Sciences and Humanities",
        34766: "Quarterly Journal Fides et Ratio",
        34778: "Universitaet Zuerich, Institut fuer Kommunikationswissenschaft"
               "und Medienforschung",
        34836: "Centralne Laboratorium Kryminalistyczne Policji",
        34913: "Universite de Geneve",
        35116: "Technical University of Kosice - Faculty of Aeronautics",
        35117: "Stowarzyszenie Inzynierow i Technikow Komunikacji RP",
        35118: "Malaysian Society for Molecular Biology and Biotechnology",
        35119: "Kugler Publications",
        35120: "Institute of Knowledge Management",
        35121: "Indira Technologies",
        35122: "Hamilton Publishing Inc",
        35123: "Drustvo za geotehniku u Bosni i Hercegovini",
        35124: "Biochemical and Cellular Archives",
        35125: "Korea Association of Skills and Qualifications",
        35126: "Sekolah Tinggi Ilmu Manajemen Sukma",
        35127: "LPPM STAI Nazhatut Thullab Sampang",
        35128: "Research Institute for Language and Information"
               "Korea University",
        35129: "Universitas Atma Jaya Makassar",
        35130: "Sekolah Tinggi Ilmu Manajemen Indonesia  -"
               "STIMI -  Banjarmasin",
        35131: "The International Society of Health and Beauty",
        35132: "STIQ Wali Songo Situbondo",
        35133: "The Korean Society for School and Community Health Educatiom",
        35134: "Universitas Putra Indonesia YPTK Padang",
        35135: "Suwon Research Institute of Catholic Church History",
        35136: "Korea Real Estate Institute",
        35137: "Program Studi Magister Manajemen Universitas Krisnadwipayana",
        35138: "LPPM Universitas Winaya Mukti",
        35139: "Jurusan Teknik Sipil dan Perencanaan Universitas Sriwijaya",
        35140: "Korean Institute of Industrial Educators",
        35141: "Universitas Adiwangsa Jambi",
        35142: "The Korean Society of Property Law",
        35143: "Politeknik Caltex Riau",
        35144: "Research Institute of Nursing Science",
        35145: "Institut Bisnis dan Teknologi Pelita Indonesia",
        35146: "The Korean Society For The Study Of Environmental Philosophy",
        35147: "Police Science Institute in Korean National"
               "Police University",
        35148: "The Institute of Legal Studies Konkuk University",
        35149: "The Association of Korean Photo-Geographers",
        35150: "The Society of Korean Performance Art and Culture",
        35151: "Korea Youth Counseling and Welfare Institute",
        35152: "The Institute of Management Research, SNU",
        35153: "The Korean Association of Kubo Studies",
        35154: "Korea Association for Education and Rehabilitation"
               "of the Visually Impaired",
        35155: "Al Mustansiriya Journal for Arab and International Studies",
        35156: "Kasdi Merbah University - Ouargla",
        35157: "Dirassat Wa Abhath - The Arabic Journal Of"
               "Human And Social Sciences",
        35158: "Institute of International Maritime Affairs",
        35159: "The Korean Society of Sports Science",
        35160: "Society for the Study of Early Korean History",
        35161: "Association for Korean and Asian Political Thoughts",
        35162: "The Research Foundation of Korean Church History",
        35163: "Bagbahce Bilim Dergisi",
        35164: "Publishing House Plastic Ltd",
        35165: "Publishing Company Atmosphere",
        35166: "IIGF Institute",
        35167: "Al-Mustansiriyah University - College of Arts",
        35168: "Sociedade Civil Educacional Tuiuti Limitada",
        35169: "Focus Oral Research",
        35170: "Stricto Sensu Editora",
        35171: "UNIMAR- Universidade de Marilia",
        35172: "Revista Veterinaria e Zootecnia",
        35173: "Tourism Institute of Northeast Asia",
        35174: "Korean Society of Culture Industry",
        35175: "Korean Association of Regional Sociology",
        35176: "Korea MICE Tourism Society",
        35177: "Far Eastern State Medical University",
        35178: "Korean Association for Canadian Studies",
        35179: "Korean Society of Education for International Understanding",
        35180: "GeoScience Engineering",
        35181: "Transactions of the VSB - Technical University"
               "of Ostrava, Civil Engineering Series",
        35182: "Transactions of the VSB - Technical University"
               "of Ostrava, Safety Engineering Series",
        35183: "Research Institute of Industry and Economy,"
               "Chonbuk National University",
        35184: "Korean Society for Holistic Convergence Education",
        35185: "Korean Journal of Educational Therapist",
        35186: "Korea Neolithic Research Society",
        35187: "Hallym University Institute of Aging",
        35188: "United Nations University, World Institute"
               "for Development Economics Research",
        35189: "National University of Physical Education and Sport",
        35190: "Swiss Society of Paediatrics",
        35191: "Creative Connections s.r.o.",
        35192: "Applied Science Private University",
        35193: "Bilecik Seyh Edebali Universitesi Fen Bilimleri Dergisi",
        35194: "Universitas Suryakancana",
        35195: "Editorial  Group - Las Tunas University",
        35196: "Sociedad Mexicana de Fitogenetica A.C",
        35197: "Universidad Autonoma Indigena de Mexico",
        35198: "SCAN, z.s.",
        35199: "The Design Society",
        35200: "LP2M STMIK Mataram",
        35201: "Association of Arab Universities Journal for"
               "Education and Psychology",
        35202: "Health Policy Research Group, College of Medicine,"
               "University of Nigeria Enugu Campus",
        35203: "Institute of Oriental Studies",
        35204: "Newsletter Transcarpathian Academy of Arts,"
               "Bulletin of Scientific and Research Works",
        35205: "Institute of Horticulture of the National"
               "Academy of Agrarian Sciences of Ukraine",
        35206: "Journal of Apitherapy and Nature",
        35207: "Language Teaching and Educational Research",
        35208: "Environmental Research and Technology",
        35209: "Kahramanmaras Sutcu Imam Universitesi Ilahiyat"
               "Fakultesi Dergisi",
        35210: "Journal of Humanities and Social Sciences Invention",
        35211: "Volgograd State Technical University",
        35212: "Sekolah Tinggi Ilmu Ekonomi Kesatuan",
        35213: "National Association of Teachers of English",
        35214: "Financial Information Society of Korea",
        35215: "Constitutional Research Institute",
        35216: "Korean Institute of Spatial Design",
        35217: "Journal of Social Affairs",
        35218: "Universitatea Nationala de Arte George Enescu Iasi",
        35219: "Universitatea Dunarea de Jos din Galati",
        35220: "The Institute of Stomatology and Maxillo-Facial"
               "Surgery National Academy of Medical Science of Ukraine",
        35221: "Tamla Cultural Institute",
        35222: "The Institute of Humanities at Soonchunhyang University",
        35223: "Gyeongsang National University Law Research Institute",
        35224: "The Korean Association for Policy Development",
        35225: "The Korean Association of Ngo Studies - KANGOS",
        35226: "The Korean Society Of Knit Design",
        35227: "Han Yang Law Association",
        35228: "Korea Association Of Maritime Business",
        35229: "Journal of Anatolian Environmental and Animal Sciences",
        35230: "Catholic Institute of Bioethics",
        35231: "Pushkin Leningrad State University",
        35232: "Eskisehir Turk Dunyasi Uygulama ve Arastirma"
               "Merkezi Halk Sagligi Dergisi",
        35233: "Okuma Yazma Egitimi Arastirmalari",
        35234: "Firat Universitesi Muhendislik Bilimleri Dergisi",
        35235: "Uluslararasi Insan Calismalari Dergisi",
        35236: "Journal of Old Turkic Studies",
        35237: "Suleyman Demirel Universitesi Fen-Edebiyat"
               "Fakultesi Sosyal Bilimler Dergisi",
        35238: "Selcuk Universitesi Fen Fakultesi Fen Dergisi",
        35239: "Tarih Arastirmalari Dergisi",
        35240: "NYU - The Metropolitan Center for Research"
               "on Equity and the Transformation of Schools",
        35241: "Emerald - Emerald Open Research",
        35242: "Tribunal Supremo de Elecciones",
        35243: "Grand Mend Mexico, S.A. de C.V.",
        35244: "Lorch Potato Research Institute",
        35245: "Centro de Investigacion y de Estudios de America"
               "Latina, Mexico, A.C.",
        35246: "University of Baghdad - College of Law",
        35247: "Fundacion Mexicana de Estudios Políticos y"
               "Administrativos. A.C  - FUNDAMESPA",
        35248: "Longdom Group",
        35249: "Sociedad Chilena de Entomologia",
        35250: "Cappadocia University",
        35251: "Association of Nepalese Agricultural Professionals"
               "of Americas",
        35252: "Spirc - I and D Lda",
        35253: "Australian Early Medieval Association",
        35254: "Bishkek State University named after Academician"
               "Kusein Karasayev",
        35255: "John L Grove College of Business",
        35256: "Florida Geological Survey",
        35257: "Revista de Informatica Teorica e Aplicada",
        35258: "Revista Eletronica Nacional de Educacao Fisica",
        35259: "Bio-Manguinhos Fiocruz",
        35260: "Editora Sertaocult",
        35261: "Universitas Singaperbangsa Karawang",
        35262: "Journal of the Indian Academy of Geriatrics",
        35263: "Edizioni Casagrande",
        35264: "Scientific Research Institute-Federal Research"
               "Center for Project Evaluation and Consulting Services",
        35265: "Revista Cientifica Semana Academica",
        35266: "Surgut State University",
        35267: "Marine Mammal Council",
        35268: "Annales de Patrimoine",
        35269: "Almumarasat Allughawia - Language Practices",
        35270: "Magalatt Risalatt Al-Aalig Al-Arabi",
        35271: "Christian Theology Institute Center",
        35272: "The Research Institute of Advanced Engineering"
               "Technology - Chosun University",
        35273: "Korean Employment and Career Association",
        35274: "Korea Banking and Financial Law Association",
        35275: "Institute Of Law Studies Pusan National University",
        35276: "Institute for History of Medicine",
        35277: "Korean Alliance Of Martial Arts",
        35278: "Association of Obstetricians and Gynaecologists of Ukraine",
        35279: "European Wound Management Association",
        35280: "The Society of Korea Illusart",
        35281: "Contemporary Society and Multiculture",
        35282: "Rutgers University Center for Cognitive Sciences",
        35283: "The Institute for Educational Research",
        35284: "Al-Mustansiriyah University - College of Arts"
               "- Philosophy Journal",
        35285: "Westminster Theological Seminary - WTS",
        35286: "Universidad Catolica de Santa Maria",
        35287: "Theater der Zeit GmbH",
        35288: "BS Publishing Limited",
        35289: "Asociacion Educacion Editora",
        35290: "Universidad Tecnologica Israel",
        35291: "IJREAM Publishing House",
        35292: "Poder Judicial del Peru",
        35293: "University of Pretoria - Department of Philosophy",
        35294: "Von Karman Institute for Fluid Dynamics",
        35295: "Onati International Institute for the Sociology of Law",
        35296: "Center for Hellenic Studies",
        35297: "Mises Institute",
        35298: "Polar Knowledge Canada",
        35299: "The Korean Society of Pure Land Buddhism",
        35300: "The Institute of Paekche Culture, Kongju National University",
        35301: "The Korean Society For Medical Ethics",
        35302: "The Korea Society for The Study of Wooden Documents",
        35303: "The Society of Practice Folkloristics",
        35304: "The Korean Society for the Study of Curriculum Integration",
        35305: "Universidad Nacional de Rosario",
        35306: "Universidad Andina Nestor Caceres Velasquez",
        35307: "Association of Language Teachers in Southeast Asia - ALTSA",
        35308: "Universitas Teuku Umar",
        35309: "STAI Darul Hikmah Bangkalan",
        35310: "Sekolah Tinggi Ilmu Ekonomi Sutaatmadja",
        35311: "Program Studi Farmasi STIKES Mandala Waluya",
        35312: "Sekolah Tinggi Filsafat Teologi Widya Sasana",
        35313: "Politeknik Negeri Bandung",
        35314: "Politeknik Negeri Bengkalis",
        35315: "Universitas Stikubank",
        35316: "LP2M Universitas Ibrahimy",
        35317: "Hasanuddin Journal of Midwifery",
        35318: "Marine Corps University Press - Marine Corps History",
        35319: "Universidad Catolica Boliviana San Pablo",
        35320: "International Association of Sound and Audiovisual Archives",
        35321: "Lietuviu Kalbos Institutas",
        35322: "Faculty of Medicine Universiti Teknologi MARA",
        35323: "Associacao dos Docentes e Orientadores de"
               "Medicina Geral e Familiar",
        35324: "Pusat Kajian Kedokteran Dan Kesehatan Aceh",
        35325: "Akademi Kebidanan Budi Mulia Palembang",
        35326: "Universitas Muhammadiyah Buton",
        35327: "Universitas Mahasaraswati Mataram",
        35328: "STIKes Al Insyirah Pekanbaru",
        35329: "LPPM Universitas Al Asyariah Mandar",
        35330: "Balkar Scientific Center of Russian Academy",
        35331: "Caram Editores Eireli",
        35332: "Kyiv Orthodox Theological Academy",
        35333: "Marmara University",
        35334: "Universitas Borneo Tarakan",
        35335: "Institute of Computer Science",
        35336: "Institute of Cardio Technics - INCART",
        35337: "Mayas Publications - Emperor International"
               "Journal of Finance and Management Research",
        35338: "Mayas Publications - Emperor Journal of Economics"
               "and Social Science Research",
        35339: "Kharkiv National Medical University",
        35340: "Donetsk State University of Management",
        35341: "Afet ve Risk DergisI",
        35342: "Econder International Academic Journal",
        35343: "Kocaeli Universitesi Sosyal Bilimler Dergisi",
        35344: "Journal of Academic Perspective on Social Studies",
        35345: "Journal of Healthcare Management and Leadership",
        35346: "Anadolu Ogretmen Dergisi",
        35347: "Korea Cultural Heritage Association",
        35348: "Korea Credit Card Academic Society",
        35349: "Korean Association Of Tax and Accounting",
        35350: "Institute of Telecommunications and Global"
               "Information Space of the NAS of Ukraine",
        35351: "Geneesmiddelenbulletin Foundation",
        35352: "Accademia Gioenia di Scienze Naturali di Catania",
        35353: "Korean Neuro-Pain Society",
        35354: "Teknik Bilimler Dergisi",
        35355: "Fenix - Revista de Historia e Estudos Culturais",
        35356: "Universidade Estadual do Norte do Parana",
        35357: "Faculdade Sao Basilio Magno - FASBAM",
        35358: "Korean Society of Childrens Literature and Education",
        35359: "Korea Society of Ingrielogy",
        35360: "University of Gothenburg - Department of Languages"
               "and Literatures",
        35361: "The Korean Association on Intellectual Disabilities",
        35362: "Organizacion de Estados Iberoamericanos",
        35363: "Vidzeme University of Applied Sciences",
        35364: "Foundation University Islamabad",
        35365: "Cyprus Mental Health Organization",
        35366: "GRAPHIMEDIC SA DE CV",
        35367: "Islamic University of Applied Sciences Rotterdam",
        35368: "The Korean Association of Contemporary Japanese Studies",
        35369: "Institute for Peace and Unification Studies"
               "Seoul National University",
        35370: "University College of Technology Sarawak",
        35371: "Korean Society of Occupational and Environmental Medicine",
        35372: "Korea Maritime Institute",
        35373: "Korea Management Engineers Society",
        35374: "Dongyang Hanmoon Association",
        35375: "Sosyal Arastirmalar ve Yonetim Dergisi",
        35376: "Doctoral School of the University of Valladolid",
        35377: "Sakarya University Journal of Computer and"
               "Information Sciences",
        35378: "Gazi University Journal of Science",
        35379: "Cukurova Universitesi Sosyal Bilimler Enstitusu Dergisi",
        35380: "Pakistan Sugar Journal",
        35381: "Fundacion Koinonia",
        35382: "Tra Vinh University",
        35383: "Universidad Catolica Santo Toribio de Mogrovejo",
        35384: "STIE Indonesia Banking School",
        35385: "IG Publishing Pte Ltd",
        35386: "Usmanu Danfodiyo University Sokoto Dept. of"
               "Curriculum Studies",
        35387: "Ivan Ziaziun Institute of Pedagogical and"
               "Adult Education of the NAES of Ukraine",
        35388: "Buddhist Culture and Art Laboratory",
        35389: "Center for Interdisciplinary Research on China",
        35390: "The Sejong Institute",
        35391: "Economic Researcher Review",
        35392: "Journal of Financial, Accounting and Managerial Studies",
        35393: "Journal of Humanities and Social Sciences"
               "- Constantine 2 University",
        35394: "The Journal of Economics and Finance - University"
               "of Hassiba ben Bouali in Chlef",
        35395: "Journal of Human Sciences of Oum El Bouaghi University",
        35396: "Revista Presencia - Miradas desde y hacia la educacion",
        35397: "Institute of Aging Society, Hanyang University",
        35398: "Korean Association Of Bibliotherapy",
        35399: "The Ichthyological Society of Korea",
        35400: "AIA Standards and Quality",
        35401: "Scientific Research Institute - Ochapovsky"
               "Regional Clinical Hospital No 1",
        35402: "Kultura es Kozosseg",
        35403: "Elektronikai Technologia es Gyartasinformatika",
        35404: "Lelektan es hadviseles",
        35405: "OxIPO",
        35406: "Mesterseges intelligencia",
        35407: "Sociedad Argentina de Genetica",
        35408: "Yonetim Bilimleri Dergisi",
        35409: "Alpinus Publications",
        35410: "IJAEB Publication",
        35411: "National Organization of Disinfectionists",
        35412: "Korean Society For Cooperative Studies",
        35413: "Culture and Arts Contents Laboratory",
        35414: "Afyon Kocatepe Universitesi Fen Ve Muhendislik"
               "Bilimleri Dergisi",
        35415: "Sirnak Universitesi Ilahiyat Fakultesi Dergisi",
        35416: "Revista Geografia em Atos - Online",
        35417: "MARES EDITORES",
        35418: "Functional Plant Breeding Journal",
        35419: "The Society of Korean Literary Studies",
        35420: "The Busan, Ulsan, Gyeoungnam Branch of The"
               "Korean Society of Otolaryngology",
        35421: "The Society of Korean National Language and Literature",
        35422: "The Correction Welfare Society of Korea",
        35423: "Institute of Philosophy of H.S. Skovoroda"
               "of National Academy of Sciences of Ukraine",
        35424: "Instituto Panamericano de Geografia e Historia",
        35425: "Accreditation Council for Graduate Medical Education",
        35426: "Instituto Politecnico Nacional",
        35427: "Institute of State and Law Russian Academy"
               "of Sciences - ISL RAS",
        35428: "University of Avellaneda",
        35429: "ECORFAN",
        35430: "Episteme Health Inc",
        35431: "Korea Research Center of Ethnology",
        35432: "ORIPA NAPA under the President of Ukraine",
        35433: "Zhytomyr Ivan Franko State University",
        35434: "Association Medical Corps of the National"
               "Hospital Almanzor Aguinaga Asenjo",
        35435: "TMG Research gGmbH",
        35436: "Yulgok Society",
        35437: "Journal of Integrative Nursing",
        35438: "STKIP MBB PRESS",
        35439: "Universitas Tridharma",
        35440: "Harran Universitesi Tip Fakultesi Dergisi",
        35441: "The Society for Korean Music",
        35442: "Institute of Humanities and Art",
        35443: "Cooperative Management Institute",
        35444: "International Journal of Advanced Networking"
               "and Applications - IJANA",
        35445: "STAI Hubbulwathan Duri",
        35446: "Komunitas Manajemen Kompetitif",
        35447: "Universitas Harapan Medan",
        35448: "Faculty of Economics and Business - Universitas"
               "Sultan Ageng Tirtayasa",
        35449: "Universitas IBA Palembang",
        35450: "Balitbangda Provinsi Lampung",
        35451: "Institut Kesehatan Medistra Lubuk Pakam",
        35452: "International Journal of Current Approaches"
               "in Language, Education and Social Sciences",
        35453: "NED University of Engineering and Technology",
        35454: "Grupo Distribuna",
        35455: "BRK Global Healthcare Consulting Firm, LLC",
        35456: "USA Journal of Medical Research",
        35457: "Universitas Islam Balitar",
        35458: "Sains Global Institut",
        35459: "Biophysical Society",
        35460: "Faculty Of Medicine And Surgery - University"
               "Of Santo Tomas -  Manila",
        35461: "Universidad Salesiana, A. C.",
        35462: "Gobierno de Navarra. Departamento de Cultura y Deporte",
        35463: "Societatea Paleontologilor din Romania",
        35464: "Pedagogium Wyzsza Szkola Nauk Spolecznyc",
        35465: "Bulgarian Rheumatology Society",
        35466: "Universidad de Sevilla - Secretariado de Recursos"
               "Audiovisuales y Nuevas Tecnologias",
        35467: "War Studies University",
        35468: "Verlag Julius Klinkhardt KG",
        35469: "Science and Research Centre Koper",
        35470: "Institute for Problems in Mechanical Engineering"
               "of the Russian Academy of Sciences - IPME RAS",
        35471: "Moqarabat",
        35472: "Institut Teknologi Sumatra",
        35473: "Universitas Ngudi Waluyo",
        35474: "Sekolah Tinggi Ekonomi dan Bisnis Islam Lampung",
        35475: "Bappeda Kota Semarang",
        35476: "Korean Transactional Analysis Counseling Association",
        35477: "Ukrainian Research and Technology Center of"
               "Metallurgy Industry Energostal",
        35478: "The Asociatia Center for Socio-Economic Studies"
               "and Multiculturalism",
        35479: "Indonesian Vitreoretinal Society",
        35480: "University of Management and Administration in Zamosc",
        35481: "California Dental Association",
        35482: "Smith College",
        35483: "Association of Collegiate Schools of Architecture, Inc",
        35484: "Fatima Gohar Educational and Welfare Society",
        35485: "Instituto Centroamericano de Administracion Publica",
        35486: "Instituto Nacional de Innovacion y Transferencia"
               "en Tecnologia Agropecuaria - INTA",
        35487: "Instituto De Ciencias Juridicas De Puebla",
        35488: "Local Population Studies Society - LPSS",
        35489: "Blavatnik School of Government, University of Oxford",
        35490: "European Council for Computing in Construction",
        35491: "Centre for Criminal Justice and Development",
        35492: "Document Academy",
        35493: "The Meducator",
        35494: "Benemerita Universidad Autonoma de Puebla",
        35495: "Asian Journal of Agriculture and Biology",
        35496: "Woori Society of Korean Literature in Classical Chinese",
        35497: "Center for Indonesian Policy Studies",
        35498: "The Korea Tourism Enhancement Society",
        35499: "Revista Tabuleiro de Letras",
        35500: "Korean Society of Global Health",
        35501: "Revista DisSoL- Discurso, Sociedade e Linguagem",
        35502: "SG Publishing",
        35503: "International Journal of Medical Laboratory Research",
        35504: "The Society For Korean Philosophical History",
        35505: "The Institute for Legal Studies, Sogang University",
        35506: "The Institute of Japanese Studies, Kookmin University",
        35507: "Revista Engenharia de Interesse Social",
        35508: "Universitas Nusa Cendana",
        35509: "Instituto Nacional de Cancerologia",
        35510: "The Institute of Educational Research Chonnam"
               "National University",
        35511: "Soproni Egyetem Kiado",
        35512: "Revista de Agroecologia no Semiarido",
        35513: "Nature Research Centre -NRC",
        35514: "Maltepe Tıp Dergisi",
        35515: "Universitatsverlag Gottingen",
        35516: "The University of Jordan",
        35517: "Arab American University",
        35518: "Orientalism Studies",
        35519: "Hawlyat Al-Montada for Human Studies",
        35520: "Programa de Pos-Graduacao em Letras Vernaculas - PPGLEV",
        35521: "Unitas – Revista Eletronica de Teologia e"
               "Ciencias das Religioes",
        35522: "Estudos em Design",
        35523: "Ivanovo State Agricultural Academy named after D.K. Belyaev",
        35524: "Northern Trans-Ural State Agricultural University",
        35525: "National Assembly Budget Office",
        35526: "The Center for Regional Development",
        35527: "Korean Association of Financial Engineering",
        35528: "Helynevtorteneti Tanulmanyok",
        35529: "Sekolah Tinggi Bahasa Asing-Persahabatan Internasional Asia",
        35530: "The National Research and Development Institute"
               "for Textiles and Leather",
        35531: "The Scientific News Ltd",
        35532: "The Academy of Engineering and Education",
        35533: "Universidad Autonoma de Zacatecas - Francisco Garcia Salinas",
        35534: "Sciscan Publishing Limited",
        35535: "W. Szafer Institute of Botany Polish Academy of Sciences",
        35536: "Lahore School of Economics",
        35537: "Universidad Nacional de La Plata",
        35538: "University of Warsaw - Faculty of History",
        35539: "Institute of Literature, Folklore and Art"
               "of the University of Latvia",
        35540: "Geophysical Survey of the Russian Academy"
               "of Sciences - GS RAS",
        35541: "Hospital for Skin Diseases, Institute of Dermatology,"
               "Chinese Academy of Medical Sciences",
        35542: "EdArXiv",
        35543: "IndiaRxiv",
        35544: "Jumuga Journal of Education, Oral Studies,"
               "and Human Sciences",
        35546: "Kherson National Technical University",
        35547: "Magarach. Viticulture and Winemaking",
        35548: "Peoples Solidarity for Participatory Democracy - PSPD",
        35549: "Historical Reporter",
        35550: "Kharkiv National Agrarian University",
        35551: "Penzugyi Szemle",
        35552: "An-Najah National University",
        35553: "Elihyaa",
        35554: "Afak Ilmia Review, University Of Tamanrasset",
        35555: "Camelot Publishing",
        35556: "Stomatology for All - International Dental Review",
        35557: "Korea Association of Social Education",
        35558: "Associazione Coreana Di Letteratura E Linguistica Italian",
        35559: "Kyunghee University Institutes of Humanities",
        35560: "University Of Baghdad - College of Fine Arts",
        35561: "Society for Musicology in Ireland (SMI)",
        35562: "Universite Jean Moulin Lyon 3",
        35563: "Universidad Nacional San Luis Gonzaga de Ica",
        35564: "Academia Europea de Direccion y Economia de la Empresa",
        35565: "DentaQuest Partnership for Oral Health Advancement",
        35566: "International Society for Data Science and Analytics",
        35567: "Russian Research Institute for Integrated"
               "Water  Management and Protection - RosNIIVKh",
        35568: "LPPM Universitas Muhammadiyah Tasikmalaya",
        35569: "Universitas Subang",
        35570: "Center for Christian Culture",
        35571: "Moscow Medical - Social Institute named after"
               "Friedrich Haass",
        35572: "Editora da Universidade Federal de Campina Grande",
        35573: "Kukhak Institute",
        35574: "The Korean Society for Developmental Psychology",
        35575: "Fundacion Universitaria Catolica del Norte",
        35576: "FGBU IMCEUAOSMP ROSZDRAVNADZOR",
        35577: "Institute of Public Administration in the"
               "Sphere of Civil Protection",
        35578: "Eurasian Conference on Language and Social Sciences",
        35579: "Institute BelNIIS",
        35580: "Fakultas MIPA Universitas Negeri Makassar",
        35581: "Sekolah Tinggi Ilmu Ekonomi Pelita Indonesia",
        35582: "International Institute of Pediatrics and Child Health",
        35583: "Sekolah Tinggi Teknologi Pekanbaru",
        35584: "STIKes Hutama Abdi Husada Tulungagung",
        35585: "STMIK AKBA",
        35586: "Fakultas Hukum Universitas Pembangunan Nasional"
               "Veteran Jakarta",
        35587: "Brazilian Journal - Publicacoes e Editora",
        35588: "University of Santiago of Chile",
        35589: "Academy of Social Welfare and Law",
        35590: "Jurnal Ekonomi dan Bisnis",
        35591: "WAHANA - Jurnal Ekonomi, Manajemen dan Akuntansi",
        35592: "Jurnal Riset Bisnis - JRB",
        35593: "Accounting Profession Journal - APAJI",
        35594: "Korean Art Therapy Association",
        35595: "LLC Kartfond",
        35596: "Belarusian State University of Informatics"
               "and Radioelectronics",
        35597: "South Urals Federal Research Center for Mineralogy"
               "and Geoecology of the Urals Branch of the RAS",
        35598: "Theoretical and Applied Aspects of Device"
               "Development on Microcontrollers and FPGAs",
        35599: "Individual entrepreneur Turchanova Veronika Timurovna",
        35600: "Red CoPaLa - Constructores de Paz Latinoamericana",
        35601: "Alutroha House for Scientific Publishing",
        35602: "Research Centre Prague s.r.o",
        35603: "SGEM World Science",
        35604: "Ariadne Labs",
        35605: "Korean Association of Secretarial Studies",
        35606: "STIE Malangkucecwara Malang - Jurnal Akuntansi"
               "Bisnis Dan Manajemem",
        35607: "The Korean Academy Of Oral Anatomy",
        35608: "National Rural Education Association",
        35609: "Global Academy of Training and Research (GATR) Enterprise",
        35610: "HANKUK SIJOHAKHOE",
        35611: "Journal of Korea Trade",
        35612: "Asociacion Espanola de Optometristas Unidos",
        35613: "Center for Creative Leadership",
        35614: "Finnish Meteorological Institute",
        35615: "Korean Clinical Epilepsy Research Society",
        35616: "Korean Council of EMS Physicians",
        35617: "Ikatan Apoteker Indonesi",
        35618: "Hungarian Statistical Review",
        35619: "Rivne State University of Humanities",
        35620: "MERC Global",
        35621: "Revista Interdisciplinar em saude",
        35622: "Instituto Universitario de Investigacion Inudi Peru",
        35623: "The Ben-Zvi Institute",
        35624: "Institute of Mining Engineers of Chile",
        35625: "Narpavi Publishers",
        35626: "Universidad Catolica Sedes Sapientiae - Fondo Editorial",
        35627: "Federal Center for Hygiene and Epidemiology",
        35628: "Escuela Superior de Guerra Naval",
        35629: "Quest Journals",
        35630: "European Scientific Society",
        35631: "Global Academic Excellence (M) Sdn Bhd",
        35632: "International Institute of Islamic Thought",
        35633: "R and D National Institute for Agricultural"
               "and Food Industry Machinery - INMA Bucharest",
        35634: "Udmurt State University",
        35635: "Revista Cientifica FACIMED",
        35636: "Korea Tax Research Forum",
        35637: "Korea Lifelong Education and HRD Institute",
        35638: "National Folk Museum of Korea",
        35639: "The Korean Association of Child Welfare",
        35640: "Chem Publisher",
        35641: "Korean Practice Association for Early Childhood Education",
        35642: "Revista Macambira",
        35643: "Informatio",
        35644: "Dirasat",
        35645: "Gusur Al-Marrifat",
        35646: "KODISA International Conference on Business"
               "and Economics (KODISA.ICBE)",
        35647: "The Association Of Korean-Japanese National Studies",
        35648: "Institute of Development Studies - The Impact Initiative",
        35649: "Khulna University Business Administration Discipline",
        35650: "Social Science Research Council",
        35651: "Ewha Institute for the Humanities, Ewha Womans University",
        35652: "Indo Global Journal of Pharmaceutical Sciences",
        35653: "SRA Akademi Yayincilik",
        35654: "Alta Dharma Publisher",
        35655: "Ukraine Supreme Legal Council",
        35656: "Institute for Political Studies - Korea",
        35657: "Korea Elementary Art Education Association",
        35658: "Journal of Economics and Human Development",
        35659: "Latin American Semiotics Association deSigniS",
        35660: "Das Verlag",
        35661: "Bonfils Automation Technologies Private Ltd",
        35662: "University of Neuchatel",
        35663: "Colegio Medico del Peru",
        35664: "CESAMES Association",
        35665: "United Prime Publications",
        35666: "University of Sarajevo, Faculty of Science",
        35667: "Grupo Paradigma",
        35668: "State Scientific Institution - Ukrainian Institute"
               "of Scientific and Technical Expertise and Info",
        35669: "ForumXXI",
        35670: "Evaluar",
        35671: "STIMIK AMIKOM Purwokerto",
        35672: "Lakpesdam PCNU Kota Salatiga",
        35673: "Institut Agama Islam Negeri Bone",
        35674: "Kent Akademisi",
        35675: "Bayburt Egitim Fakultesi Dergisi",
        35676: "Serie Tecnica INIA",
        35677: "The Arab Journal of Science and Technology Education",
        35678: "International College Suan Sunandha Rajabhat University",
        35679: "CJSC ALKOR",
        35680: "The Beryl Institute",
        35681: "Institute for Information Recording of National"
               "Academy of Sciences of Ukraine",
        35682: "Mutah University",
        35683: "Journal of Contemporary Management",
        35684: "Sanglap",
        35685: "Revista Interacao Interdisciplinar",
        35686: "Institute of Archaeology of the CAS - Prague, v.v.i.",
        35687: "Sociedad de Cirujanos de Chile",
        35688: "Scientific Research Center Of Environmental"
               "Engineering And Construction",
        35689: "Advanced Multiple Inc",
        35690: "EDITIONS QUAE",
        35691: "Centre for Bioresource Research",
        35692: "Asociacion de Facultades de Administracion"
               "y Economia - Chile",
        35693: "FSBEI of Higher Education SamSMU of Ministry"
               "of Health of the Russian Federation",
        35694: "Yaroslavl State Agricultural Academy",
        35695: "Al Qalam University for Humanitarian and Applied Sciences",
        35696: "Thamar University - Faculty of Arts",
        35697: "Politeknik Negeri Bandung - Departmen Admininistrasi Niaga",
        35698: "Jerash University",
        35699: "Universidade Federal de Minas Gerais - Pro-Reitoria"
               "de Pesquisa",
        35700: "Instituto Federal De Educacao - Ciencia E"
               "Tecnologia De Santa Catarina-Ifsc",
        35701: "Revista da Casa da Geografia de Sobral",
        35702: "Magnus Med Club LLC - USA",
        35703: "The World Islamic Sciences and Education University",
        35704: "Chonnam National University  - Institute of Humanities",
        35705: "Bilge Strateji",
        35706: "LPPM Universitas Singaperbangsa Karawang -"
               "Research Department in Indonesia University",
        35707: "Universidad de Ibague",
        35708: "Institute for Semantic Computing Foundation",
        35709: "Association of Rice Research Workers",
        35710: "Al Mustansiriyah University - Centre for Arabic"
               "and International Studies",
        35711: "Artificial Intelligence in Medical Imaging",
        35712: "Artificial Intelligence in Gastroenterology",
        35713: "Artificial Intelligence in Cancer",
        35714: "Stratigraphy and Sedimentology of Oil-Gas Basins",
        35715: "Peter McMullin Centre on Statelessness - Melbourne"
               "University Law School",
        35716: "The Society of Economics and Development",
        35717: "ISCI Publishing",
        35718: "Institut Teknologi Kalimantan",
        35719: "IAIN Jember",
        35720: "LPPM An Nuur Purwodadi",
        35721: "Akademi Kebidanan Aisyiyah Pontianak",
        35722: "Sekolah Tinggi Ilmu Administrasi Tabalong",
        35723: "Al-Hayat Al-Istiqomah Foundation",
        35724: "Universitas Musamus Merauke",
        35725: "Fakultas Farmasi Universitas Mahasaraswati Denpasar",
        35726: "Politeknik Pertanian Negeri Kupang",
        35727: "Komisi Akreditasi Rumah Sakit - KARS",
        35728: "STIKes Wiyata Husada Samarinda",
        35729: "Jurusan Pariwisata Politeknik Negeri Manado",
        35730: "LPPM STIKes Prima Nusantara Bukittinggi",
        35731: "Korean Political Communication Association",
        35732: "Korean Society for Lactic Acid Bacteria and Probiotics",
        35733: "The Center for Global Teacher Education, Kongju"
               "National University",
        35734: "Korean Association For Rehabilitation Psychology",
        35735: "Miromanova Irina Vital evna",
        35736: "Korean Society of Consumer Studies",
        35737: "St Josephs College of Commerce",
        35738: "Institute of Textiles and Fashion Professionals,"
               "Ghana - ITFP-GH",
        35739: "Universidad Nacional del Centro de la Provincia"
               "de Buenos Aires",
        35740: "Ondefo-Verlag",
        35741: "Southwest Jiaotong University",
        35742: "Communication and Information Sciences",
        35743: "Faculty of Arts Law and Education",
        35744: "Budapesti Finnugor Fuzetek",
        35745: "International Institute of Knowledge Innovation"
               "and Invention Private Limited",
        35746: "Sekawan Institute",
        35747: "LP2M - UM Banjarmasin",
        35748: "Indonesia Center for Middle East Studies",
        35749: "Pesatuan Dokter Spesialis Mata Indonesia",
        35750: "Saint-Petersburg University of the Ministry"
               "of Internal Affairs of the Russian Federation",
        35751: "Korean Association Of African Studies",
        35752: "Federal Research and Production Center Joint-Stock"
               "Company - Mars - Research and Production Assoc",
        35753: "Revista Cientifica Hospital Santa Izabel",
        35754: "National Medical Research Center of Hematology"
               "of the Ministry of Health of the Russian Federation",
        35755: "Medical Association of Thailand",
        35756: "Universidad Marcelino Champagnat",
        35757: "Institute of Political Studies - Polish Academy of Sciences",
        35758: "The Electrochemical Society of India",
        35759: "Elewa Biosciences, F.a.C.T Ltd (K)",
        35760: "Gunadarma University",
        35761: "Spanish Association of Mental Health Nursing",
        35762: "Environmental Research Institute, Chulalongkorn University",
        35763: "Spanish Society of Research in Mathematics Education",
        35764: "The New York Academy of Medicine",
        35765: "Jesuit University Ignatianum - Ignatianum University Press",
        35766: "Centro Universitario CIFE",
        35767: "Canadian Society of Petroleum Geologists",
        35768: "Institute of Maha Bodhi Thought",
        35769: "Korean Environmental Law Association",
        35770: "Korean Competition Law Association",
        35771: "The 21st Century Association Of English Language"
               "And Literature",
        35772: "National Center for Global Health and Medicine (JST)",
        35773: "Institute of Global Politics",
        35774: "Ternopil National Economic University",
        35775: "LLC Publishing House - Science Today",
        35776: "MosvodokanalNIIproject",
        35777: "Magallat al-dirasat al-huquqiyat",
        35778: "Azzaytuna University",
        35779: "Centre Universitaire Relizane",
        35780: "Magallat al-qanun wa al-mugtama",
        35781: "Alandalus University for Science and Technology",
        35782: "ADPSE Catalactica Bucuresti",
        35783: "National Foundation Of Brief Therapy (JST)",
        35784: "Politechnika Lubelska",
        35785: "Smolensk State University",
        35786: "Curriculo sem Fronteiras",
        35787: "Islamabad Medical and Dental College",
        35788: "African Journal of Political Sciences Publications",
        35789: "The International Federation for Structural Concrete",
        35790: "Jurnal e-Biomedik",
        35791: "Eugenia",
        35792: "Zootec",
        35793: "Jurnal Teknik Informatika",
        35794: "Jurnal Pembangunan Ekonomi dan Keuangan Daerah",
        35795: "Lex Et Societatis",
        35796: "Kajian Linguistik",
        35797: "Jurnal Administrasi Bisnis",
        35798: "Kesmas",
        35799: "dCartesian",
        35800: "Platax",
        35801: "Vivabio",
        35802: "Wellcome",
        35803: "Kyrgyz State University of Construction, Transportation"
               "and Architecture n.a. N.Isanov",
        35804: "Royal Botanic Gardens - Kew",
        35805: "National Scientific Center of Surgery named"
               "after A.N. Syzganov",
        35806: "Sampoerna University - Faculty of Engineering and Technology",
        35807: "Legal and Management",
        35808: "ISMA SYC INT",
        35809: "All-Russian Research Institute of Irrigated Agriculture",
        35810: "The Egyptian Society of Cardiothoracic Surgery",
        35811: "Universidad de las Americas",
        35812: "Institutul de Chimie Macromoleculara Petru Poni",
        35813: "Hawliyat Al-Adab Wa Al-Lugat",
        35814: "Universitas Pancasila",
        35815: "European Radon Association",
        35816: "Akademi Keperawatan Sandi Karsa",
        35817: "Universitas Halu Oleo - Jurusan Ilmu Administrasi Publik",
        35818: "Editora IFMA",
        35819: "Instituto Federal de Educacao - Ciencia e"
               "Tecnologia do Rio Grande do Sul",
        35820: "The Korean Association for Contemporary Chinese Studies",
        35821: "The International Academic Forum of Korean"
               "Language and Literature",
        35822: "The Association of Chinese Language, Literature"
               "and Translation in Korea",
        35823: "Guhmok Publishing Acc",
        35824: "Faculties of Humanities and Theology at Lund"
               "University - Swedish Journal of Romanian Studies",
        35825: "27 Science Center",
        35826: "Ulusoy Egitim Danismanlik Yayincilik Ve Organizasyon",
        35827: "Korean Association of Pain Medicine",
        35828: "The English Teachers Association in Korea",
        35829: "STIE Bank BPD Jateng",
        35830: "Universidad Michoacana de San Nicolas de Hidalgo",
        35831: "Spotlight on Research",
        35832: "The Society Of Korean Modern Literrary Criticism",
        35833: "Journal of Modern Power Systems and Clean Energy",
        35834: "University of Central Missouri, Department"
               "of Mathematics and Computer Science",
        35835: "UNIRA",
        35836: "STEI SEBI - Jurnal Akuntansi dan Keuangan Islam - JAKIs",
        35837: "Politeknik Keuangan Negara STAN - Sumber Artikel"
               "Akuntansi Auditing dan Keuangan Vokasi",
        35838: "Jurnal Riset Akuntansi dan Perpajakan",
        35839: "REPIS",
        35840: "VIBGYOR ePress",
        35841: "Allied Academies",
        35842: "Pusat Penelitian dan Pengabdian Pada Masyarakat"
               "Universitas Respati Yogyakarta",
        35843: "Federation University Australia",
        35844: "University of Cincinnati -  Office of Innovation"
               "and Community Engagement",
        35845: "Khyber Medical University",
        35846: "Headway Global Research Consultancy Pte Ltd",
        35847: "ProLiteracy Worldwide",
        35848: "IOP Publishing",
        35849: "National Cereals Research Institute",
        35850: "Korean Academic Society of Taxation",
        35851: "Korean Society For Phenomenology",
        35852: "Russian Institute of Theatre Arts - GITIS",
        35853: "Liberal Arts University",
        35854: "Saint-Petersburg University of Management"
               "Technologies and Economics - UMTE",
        35855: "Korean Buddhist Counselling Society",
        35856: "Persatuan Dokter Gigi Indonesia Cabang Makassar",
        35857: "Ukrainian Research Institute of Refractories"
               "named after A.S. Berezhnoy",
        35858: "Theological Thought Institute",
        35859: "The Journal of Mesopotamian Studies",
        35860: "International Advanced Researches and Engineering Journal",
        35861: "Korean Association of Occupational Health Nurses",
        35862: "Scient Open Access",
        35863: "The Korean Medieval History Society",
        35864: "Etlik Veteriner Mikrobiyoloji Dergisi",
        35865: "Organization of Korean Historian",
        35866: "The Economic Research Institute",
        35867: "Soong Sil University Institute for the Legal Studies",
        35868: "Institute of Agrocultural Microbiology and"
               "Agro-industrial Manufacture of NAAS of Ukraine",
        35869: "University of Vigo",
        35870: "Lembaga KITA",
        35871: "Yayasan PKBM Lestari Jaya",
        35872: "Sekolah Tinggi Ilmu Kesehatan Estu Utomo",
        35873: "The Convergent Research Society Among Humanities,"
               "Sociology, Science, and Technology",
        35874: "STIKES Insan Cendekia Medika Jombang",
        35875: "Al-Ahliyya Amman University",
        35876: "Masyarakat Perkelapasawitan Indonesia",
        35877: "Yayasan Ahmar Cendekia Indonesia",
        35878: "Institut Pesantren Mathali ul Falah",
        35879: "Sekolah Tinggi Ilmu Kepolisian",
        35880: "Badan Pengembangan Sumber Daya Manusia Provinsi Jawa Barat",
        35881: "Korea Human Rights and Law-Related Education Association",
        35882: "Poltekkes Kemenkes Surabaya",
        35883: "The Korean Beauty Management Society",
        35885: "A.N.Severtsov Institute of Ecology and Evolution"
               "RAS  - IEE RAS",
        35886: "Sekolah Tinggi Desain Bali",
        35887: "All-Russian Research Institute for Use of"
               "Machinery and Petroleum Products in Agriculture",
        35888: "STAINU Madiun",
        35889: "STMIK Banjarbaru",
        35890: "Akademi Kebidanan Dharma Husada Kediri",
        35891: "Universitas Yudharta Pasuruan",
        35892: "STIKES Nani Hasanuddin Makassar",
        35893: "LPPM - Sekolah Tinggi Ilmu Kesehatan Siti Hajar Medan",
        35894: "Fakultas Teknologi Kedirgantaraan",
        35895: "Physical Society of Indonesia",
        35896: "IAI Al Hikmah Tuban",
        35897: "Institute Agama Islam Al-Qolam",
        35898: "Yayasan Aliansi Cendekiawan Indonesia Thailand",
        35899: "Faculty of Entrepreneurship Universitas Garut",
        35900: "Animal Husbandry, Agriculture Faculty, Gorontalo"
               "State University",
        35901: "Korean Constitutional Law Association",
        35902: "Urimal Society",
        35903: "TEANGA",
        35904: "STMA Trisakti, Jakarta",
        35905: "Institut Agama Islam Negeri Parepare",
        35906: "Universitas Muhammadiyah Palopo",
        35907: "Stikes Bina Generasi Polewali Mandar",
        35908: "Fakultas Ekonomi Universitas IGM",
        35909: "STT Stars Lub Luwuk Banggai",
        35910: "Poltekkes Kemenkes Jambi",
        35911: "Faculty of Marine Science and Fisheries, Hasanuddin"
               "University",
        35912: "Goodwood Publishing",
        35913: "STIKES Bethesda Yakkum",
        35914: "Universitas Andi Djemma",
        35915: "STIE Muhammadiyah Tanjung Redeb",
        35916: "Baiman Bauntung Batuah Center",
        35917: "Sekolah Tinggi Ilmu Manajemen YKPN Yogyakarta",
        35918: "Al-turat",
        35919: "Revista Brasileira de Sexualidade Humana",
        35920: "Analytica. Revista de Filosofia",
        35921: "Revista Jangada: Critica Literatura Artes",
        35922: "Program Studi Magister Teknik Perminyakan,"
               "Universitas Trisakti, Jakarta",
        35923: "Universitatea Vest din Timisoara",
        35924: "Forditastudomany",
        35925: "Multidiszciplinaris tudomanyok",
        35926: "Honvedsegi szemle",
        35927: "Winkler Gabor Szakkollegium",
        35928: "Capic Review",
        35929: "Museum d'Histoire naturelle of Geneva",
        35930: "Korean Juvenile Policy Association",
        35931: "Sekolah Tinggi Ilmu Qur an Amuntai",
        35932: "Universitas Bandar Lampung",
        35933: "University of Chemistry and Technology - Faculty"
               "of Environmental Technology",
        35934: "SEGi University SDN BHD",
        35935: "Seagull Publications",
        35936: "Universidad del Pacifico - Ecuador",
        35937: "Art and Possibility Consulting",
        35938: "Tashkent Institute of Irrigation and Agricultural"
               "Mechanization Engineers",
        35939: "Messenger Publications",
        35940: "Blue Eyes Intelligence Engineering and Sciences"
               "Engineering and Sciences Publication - BEIESP",
        35941: "Mulawarman University - Agricultural Faculty",
        35942: "IJCAB Publications",
        35943: "Mansoura University",
        35944: "ACRN Oxford Ltd",
        35945: "European University's Institute for the Research"
               "of Economic and Social Problems of Globalization",
        35946: "National Institute on Alcohol Abuse and Alcoholism",
        35947: "Progettinrete Srl",
        35948: "Project Department",
        35949: "St. John's University of Tanzaniza Press (SJUTP)",
        35950: "Al-Mustansiriya University - College of Basic Education",
        35951: "Communication and Methods",
        35952: "STIKes Muhammadiyah Pringsewu Lampung",
        35953: "Fundacao Oswaldo Cruz Brasilia - Fiocruz Brasilia",
        35954: "Ministerio de Defensa - Salud Militar",
        35955: "The Society for Chinese Humanities in Korea",
        35956: "Revista Latinoamericana de Estudos do Discurso",
        35957: "LPPM STMIK Global Informatika MDP",
        35958: "STIKes Prima Indonesia",
        35959: "STMIK DCC Kotabumi",
        35960: "Lembaga Penelitian dan Pengabdian Kepada Masyarakat"
               "Universitas Harapan Bangsa",
        35961: "Sekolah Tinggi Agama Islam Negeri Sultan Abdurrahman"
               "Kepulauan Riau",
        35962: "STBA LIA",
        35963: "Poltekkes Kemenkes Kalimantan Timur",
        35964: "LP3M Sekolah Tinggi Agama Islam Nahdlatul Wathan Samawa",
        35965: "Program Pascasarjana Universitas Bosowa",
        35966: "STIKES Karya Husada Kediri",
        35967: "Laboratorium Ilmu Pemerintahan FISIP Universitas Riau",
        35968: "Universitas Dirgantara Marsekal Suryadarma",
        35969: "STMIK Rosma",
        35970: "Politeknik Negeri Cilacap",
        35971: "Program Studi Kesehatan Masyarakat Universitas"
               "Negeri Gorontalo",
        35972: "Sekolah Tinggi Ilmu Ekonomi Pancasetia",
        35973: "PMIH Untag Semarang",
        35974: "LPPM Universitas Advent Indonesia",
        35975: "Aga Khan University Hospital",
        35976: "Pacific Oceanology Institute",
        35977: "10.35977",
        35978: "Korea Technology Innovation Society",
        35979: "Korea Administrative Law and Practice Association",
        35980: "The Korean Research Institute Of International"
               "Commerce and Law",
        35981: "Hanyang University The Third Sector Institute",
        35982: "The Korean Society for Chinese Studies",
        35983: "Korean Musicological Society",
        35984: "Centro Universitario da Fundacao Assis Gurgacz"
               "- Fag Journal Of Health",
        35985: "Santiago de Cali University",
        35986: "Korean Society for Affective Disorders",
        35987: "Laborjuris - Revista de Direito do Trabalho"
               "Processo do Trabalho e Direito da Seguridade Social",
        35988: "Sveikatos mokslai",
        35989: "Sekolah Tinggi Ilmu Pertanian Rejang Lebong",
        35990: "Fakultas Kedokteran Universitas Jenderal Ahmad Yani",
        35991: "Politeknik Perkapalan Negeri Surabaya",
        35992: "MLS Project, Design & Management",
        35993: "International Journal of Innovation in Teaching and Learning",
        35994: "Review of Human Rights (RHR)",
        35995: "Journal and Article Management System",
        35996: "Malaysian Association for Psychotherapy",
        35997: "Corporacion Universitaria Adventista",
        35998: "Berliner Wissenschafts-Verlag",
        35999: "Asociacion Espanola de Enfermeria Vascular y Heridas",
        36000: "Mediscience GmbH",
        36001: "PHM Society",
        36002: "LPPM Universitas Dhyana Pura",
        36003: "Universidad Privada Maria Serrana",
        36004: "National Institute for Economic Research",
        36005: "University for Development Studies",
        36006: "Editorial Octaedro",
        36007: "J. Selye University",
        36008: "Fundacion Iberoamericana De Las Industrias"
               "Culturales Y Creativas",
        36009: "Cato Institute",
        36010: "Faculdade De Tecnologia E Ciencias Da Bahia",
        36011: "Korean Society of Cardiovascular Disease Prevention",
        36012: "Encyclopaedic Publishing Pte Ltd",
        36013: "SciencePower Publishing House LLC",
        36014: "Phage Directory",
        36015: "Hospital de Especialidades Carlos Andrade Marin",
        36016: "Kharkiv Entomological Society",
        36017: "Ordine TSRM PSTRP di Napoli, Avellino, Benevento",
        36018: "Dev Sanskriti Vishwavidyalaya",
        36019: "Rutgers University Press",
        36020: "University of Maiduguri - College of Medical Sciences",
        36021: "University of North Carolina Wilmington -"
               "The Centers for Teaching Excellence and Faculty Leadership",
        36022: "STAI Al-Hamidiyah Jakarta",
        36023: "CASRE of the IGS of NASU",
        36024: "Journal of the Association of Arab Universities"
               "for Research in Higher Education",
        36025: "ARJ - Art Research Journal / Revista de Pesquisa em Artes",
        36026: "Revista Presenca Geografica",
        36027: "JSC Radio Engineering Corporation  - Vega",
        36028: "Volga Region Academy of Physical Culture, Sport and Tourism",
        36029: "Korean Academic Society of Financial Planning",
        36030: "National Academy for Public Administration"
               "under the President of Ukraine",
        36031: "The Korean Society of Health Education",
        36032: "Societe Coreenne de la Litterature Classique Francaise",
        36033: "Korean Social Science Research Council",
        36034: "Institute of Youngnam Culture",
        36035: "Assessment and Information, Centre State Enterprise",
        36036: "Mushroom Society of India",
        36037: "International Journal of Research in Engineering"
               "and Innovation",
        36038: "Russian Federal Research Institute of Fisheries"
               "and Oceanography",
        36039: "Coimbatore Institute of  Information Technology",
        36040: "LPPM ITN Malang",
        36041: "University of Technology, Baghdad",
        36042: "University of Technology, Baghdad",
        36043: "CHEAR Data Center at Icahn School of Medicine at Mount Sinai",
        36044: "Universidad de la Republica - ECOS",
        36045: "The Belgian Mathematical Society",
        36046: "Islamic University of Madinah",
        36047: "Journal of College of the Great Imam University",
        36048: "Balai Riset dan Standardisasi Industri Surabaya",
        36049: "Politeknik Kesehatan Kartini Bali",
        36050: "LPPM Sekolah Tinggi Teknologi Pagaralam",
        36051: "STIK Immanuel",
        36052: "Pusdiklat Tenaga Teknis Pendidikan dan Keagamaan",
        36053: "Sekolah Tinggi Ilmu Kesehatan Kepanjen",
        36054: "STMIK lKMl Cirebon",
        36055: "Universitas Sultan Ageng Tirtayasa - Fakultas Teknik",
        36056: "University of Aden -  Faculty of Economics"
               "and Administration",
        36057: "Universitas Ekasakti",
        36058: "Teachers Forum",
        36059: "Liha-Pres",
        36060: "Akademi Farmasi Sandi Karsa",
        36061: "Saint Claret College",
        36062: "West Bengal Veterinary Association",
        36063: "ASLE-KOREA",
        36064: "Korea Disabled Peoples Development Institute",
        36065: "Revista Odontologica do Brasil Central - ROBRAC",
        36066: "Composicao Revista de Ciencias Sociais da"
               "Universidade Federal de Mato Grosso do Sul",
        36067: "Program Studi Akuntansi Fakultas Ekonomi Universitas"
               "PGRI Yogyakarta",
        36068: "Romanian-American Organization",
        36069: "JoLS, Journal of Life Sciences",
        36070: "HexaTimes LLC USA",
        36071: "International CLIVAR Project Office",
        36072: "Global Alliance for Improved Nutrition - GAIN",
        36073: "Georgian Technical University, Techinformi",
        36074: "European Scientific Platform Publishing",
        36075: "European Association of Psychotherapy - EAP",
        36076: "American Society of Interventional Pain Physicians",
        36077: "Kufa Journal for Agricultural Sciences",
        36078: "Uzbek State World Languages University",
        36079: "Lamintang Education and Training Centre",
        36080: "Fikom Budi Luhur",
        36081: "P3M STIKes Bakti Tunas Husada",
        36082: "Politeknik Kesehatan Jakarta I",
        36083: "Pusat Penelitian dan Pengabdian Kepada MasyarakatP3M"
               "- STIE Sabang",
        36084: "Sekolah Tinggi Pertanian Kutai Timur",
        36085: "Universitas Muhammadiyah Bengkulu",
        36086: "Politeknik Kesehatan Kemenkes Palembang",
        36087: "Badan Penelitian dan Pengembangan Daerah Provinsi"
               "Kalimantan Timur",
        36088: "STIT Palapa Nusantara Lombok NTB",
        36089: "NHM Press",
        36090: "Universitas Indonesia Timur",
        36091: "Jerash Journal for Research and Studies",
        36092: "The Korean History of Science Society",
        36093: "The Korean Studies Institute",
        36094: "Indian Science News Association",
        36095: "Banco de Mexico",
        36096: "Bussecon International",
        36097: "Universidad San Gregorio de Portoviejo",
        36098: "The Stair Society",
        36099: "Addaiyan International Publishers",
        36100: "M.P. Shulgin State Road Research Institute State Enterprise",
        36101: "Sekolah Tinggi Ilmu Pelayaran - STIP - Jakarta",
        36102: "North Carolina Museum of Natural Sciences",
        36103: "University of Baghdad - College of Agriculture",
        36104: "Asociacion Colombiana de Medicina Interna",
        36105: "Universidad Anahuac Mexico",
        36106: "World Wide Journals",
        36107: "Moscow State University of Food Production",
        36108: "Lujosh Ventures Limited",
        36109: "Colegio de Medicos y Cirujanos de Guatemala",
        36110: "Pachhunga University College",
        36111: "Bangladesh Society for Veterinary Medicine",
        36112: "Revista ComCiencia",
        36113: "Universidade Estadual de Santa Cruz",
        36114: "The Korean Association of Literature and Film",
        36115: "Russian State Academy Of Arts For Special Needs Students",
        36118: "Edra SpA",
        36129: "Edra SpA",
        36130: "Edra SpA",
        36137: "Railway Research Institute",
        36138: "Edra SpA",
        36149: "Pacini Editore",
        36150: "Pacini Editore",
        36159: "ED MARKETING  AND COMMUNICATION di Edoardo Desiderio",
        36164: "Edra SpA",
        36171: "Societa Benefit AMD Srl SB",
        36175: "eDIzioNi AkME",
        36185: "Pacini Editore",
        36216: "Indonesian Society of Internal Medicine Bali",
        36217: "Ikatan Sarjana Ekonomi Indonesia Cabang Bandung",
        36218: "Medhawi Foundation",
        36219: "Institutul de Pshihiatrie Socola Iasi",
        36220: "Korean Association Of Victimology",
        36221: "Techmiracle Edusoft Llp",
        36222: "European Journal of Technique",
        36223: "Association for Future Music Education",
        36224: "SPJ Centre for Multidisciplinary Research",
        36225: "Universidad de la Republica - Centro Universitario Este",
        36226: "Universitas Islam Attahiriyah",
        36227: "Institute of Electrical and Electronics Engineers"
               "(IEEE) preprint",
        36228: "Wyzsza Szkola Ekonomiczno-Humanistyczna",
        36229: "Editora Poisson",
        36230: "Global Trading Consulting SH P.K",
        36231: "University of Baghdad - College of Education for Women",
        36232: "Universitas Pendidikan Muhammadiyah (UNIMUDA) Sorong",
        36233: "Central Research Institute for Epidemiology",
        36234: "Korea Association Of Cultural Economics",
        36235: "Knowledge Humanities Research Group",
        36236: "Joint Stock Company Compositebook",
        36237: "Korea Edutainment Society",
        36238: "Revista Academica Online",
        36239: "Revista de Divulgacao Cientifica Sena Aires",
        36240: "Erdelyi Tudomanyos Fuzetek",
        36241: "Tudomany - es Technikatorteneti Fuzetek",
        36242: "Muszaki Tudomanyos Fuzetek",
        36243: "Fiatal Muszakiak Tudomanyos Ulesszaka",
        36244: "Infocommunications Journal",
        36245: "Magyar Regeszet",
        36246: "Utugyi Lapok",
        36247: "Nizhny Novgorod Academy of the Ministry of"
               "Internal Affairs of Russia",
        36248: "Korean Insurance Law Association",
        36249: "4D",
        36250: "NORDEX Nonprofit Kft",
        36251: "School of Human Services and Social Work,"
               "Griffith University",
        36252: "Segedkonyvek a nyelvi kozvetitesrol",
        36253: "Firenze University Press",
        36254: "SWP Publishers BV",
        36255: "Exon Publications",
        36256: "Lasigo Publisher",
        36257: "Politeknik Negeri Sriwijaya",
        36258: "Arbeitstitel - Forum fur Leipziger Promovierende",
        36259: "EmeraldTree Publishing",
        36260: "Red Iberoamericana de Pedagogia",
        36261: "International Journal of Distance Education and E-Learning",
        36262: "Pembangunan Jaya University",
        36263: "University of Benin - Faculty of Environmental Sciences",
        36264: "Institute of Economics and Industrial Engineering,"
               "Siberian Branch",
        36265: "PublishingRealTime",
        36266: "Pubtexto",
        36268: "Ram Eesh Institute of Education, Greater NOIDA",
        36269: "Universitas Muhammadiyah Lampung",
        36270: "Sekolah Tinggi Alkitab Tiranus",
        36271: "Iberian American Journal of Podiatry",
        36272: "LPPM - STIE Sakti Alam Kerinci",
        36273: "Ivan Fedorov Book Chamber of Ukraine, State"
               "Scientific Institution",
        36274: "Yonsei Historical Association",
        36275: "Sekolah Tinggi Teknologi Industri (STTIND) Padang",
        36276: "Sekolah Tinggi Pariwisata AMPTA Yogyakarta",
        36277: "Universitas Balikpapan",
        36278: "Korea Society of Environmental Analysis",
        36279: "Asosiasi Program Studi Mandarin Indonesia (APSMI)",
        36280: "Stowarzyszenie Filozofii Prawa i Filozofii"
               "Spolecznej - Sekcja Polska IVR",
        36281: "Thammasat Eye Center",
        36282: "International Journal of Advanced Scientific"
               "Research and Management",
        36283: "Ziauddin University",
        36284: "Elon University Center for Engaged Learning",
        36285: "Transmathematica",
        36286: "Asociacion Peruana de Retorica",
        36287: "SETSCI",
        36288: "Open Source Robotics Foundation",
        36289: "Indonesian Mechanical Engineering Cooperation Agency",
        36290: "Solen s.r.o.",
        36291: "Non-Commercial Partnership Center for Biogenic"
               "Resources Humus Sapiens",
        36292: "Central Philippines State University",
        36293: "Ladoke Akintola University of Technology",
        36294: "Universitas Asahan",
        36295: "Africa Health Research Organization",
        36296: "Vidnovluvana energetika",
        36297: "Vallway Research Publication",
        36298: "Gerais: Revista Interinstitucional de Psicologia",
        36299: "The Korean Glaucoma Society",
        36300: "REMASP",
        36301: "Korean Society of Construction Health",
        36302: "Journal of Intelligence Researches and Mental Abilities",
        36303: "Medpharm Publications",
        36304: "Expeditions with MCUP",
        36305: "Nikitsky Botanical Gardens National Science Center",
        36306: "Konya Muhendislik Bilimleri Dergisi",
        36307: "Sekolah Tinggi Ilmu Kesehatan AKBIDYO",
        36308: "Sekolah Tinggi Ilmu Kesehatan Bhakti Mandala Husada Slawi",
        36309: "Jurnal Go Infotech STMIK AUB Surakarta",
        36310: "Jebi (Jurnal Ekonomi Bisnis Indonesia) STIE Bisnis Indonesia",
        36311: "Faculdade de Filosofia e Ciências",
        36312: "Lembaga Penelitian dan Pemberdayaan Masyarakat - LITPAM",
        36313: "ReDelve International Publications LLC",
        36314: "Centro Universitario de Oriente",
        36315: "World Institute for Advanced Research and Science",
        36316: "Global Clinical and Translational Research Institute",
        36317: "Kufa Journal of Arts - Faculty of Arts",
        36318: "Journal of Arabic Language and Literature - Faculty of Arts",
        36319: "Kufa Review - UNESCO Chair",
        36320: "Al-Kufa University Journal for Biology - Faculty of Science",
        36321: "Kufa Journal for Nursing Sciences - Faculty of Nursing",
        36322: "Journal of Kufa Studies Center - Kufa Studies Center",
        36323: "Kufa Journal of Legal and Political Sciences"
               "- Faculty of Law",
        36324: "Journal of the Faculty of Fiqh - Faculty of Jurisprudence",
        36325: "Al Gharee Journal for Economics and Administration"
               "Sciences - Faculty of Administration and Economic",
        36326: "Kufa Journal for Veterinary Medical Sciences"
               "- Faculty of Veterinary",
        36327: "Journal of the Faculty of Education for Girls"
               "of Humanities - Faculty of Education for Girls",
        36328: "Journal of Geographical Research - Faculty"
               "of Education for Girls",
        36329: "Kufa Journal for Chemistry -  Faculty of Education for Girls",
        36330: "Kufa Medical Journal - Faculty of Medicine",
        36331: "Agencia De Regulacion y Control Fito y Zoosanitario",
        36332: "Asociacion de Endodoncistas del Azuay",
        36333: "European Forest Institute",
        36334: "Modelling  and Simulation Society of Australia"
               "and New Zealand",
        36335: "Vietnam Meteorological and Hydrological Administration",
        36336: "Studio D - akustika s.r.o.",
        36337: "Erdelyi Muzeum-Egyesulet",
        36338: "Hungarian Archaeology",
        36339: "Forum Dosen Indonesia",
        36340: "Science and Art",
        36341: "LPPM Universitas Abdurrab",
        36342: "Fakultas Teknologi Informasi - Universitas Advent Indonesia",
        36343: "Southern Branch of the Heritage Institute",
        36344: "Cross Current Publisher",
        36345: "Korean Academy of Commodity Science & Technology",
        36346: "South Asian Research Publication",
        36347: "Scholars Academic and Scientific Publishers",
        36348: "Scholars Middle East Publishers",
        36349: "East African Scholars Publisher",
        36350: "Sekolah Tinggi Ilmu Komputer Binaniaga",
        36351: "Ophthalmological Society of Pakistan",
        36352: "LPPM STT Ibnu Sina",
        36353: "Journal of Historical and Social Studies",
        36354: "Korea Academic Society of Industrial Organization",
        36355: "Universitas Muara Bungo",
        36356: "Fakultas Hukum UNTAG Semarang",
        36357: "Research Institute of Human Ecology",
        36358: "Korean Society for Creativity Education",
        36359: "State Scientific Research Control Institute"
               "of Veterinary Medicinal Products and Feed Additives",
        36360: "Rouya turkiyyah",
        36361: "Tyumen State Medical University",
        36362: "Gumushane Universitesi Sosyal Bilimler Enstitusu"
               "Elektronik Dergisi",
        36363: "Revista de Extensao e Estudos Rurais - REVER",
        36364: "Western Music Research Institute",
        36365: "Korean Institute for Reformed Studies",
        36366: "The Forum on Education Abroad",
        36367: "Ludomedia Unipessoal Lda",
        36368: "Umea University Library",
        36369: "University of KwaZulu-Natal School of Management,"
               "IT and Governance",
        36370: "TTO Conference Ltd",
        36371: "Journal Port Science Research",
        36372: "The Jordan Journal of Libraries and Information Science",
        36373: "Erdelyi Muzeum-Egyesulet",
        36374: "Al-Nisour University College",
        36375: "CALNESTOR Knowledge Solutions Private Limited",
        36376: "STIKes Wira Medika Bali",
        36377: "Khabadze ZS",
        36378: "Universitas Islam Kuantan Singingi",
        36379: "STKIP PGRI Sumenep",
        36380: "Journal of World's Poultry Research",
        36381: "International Academy of Marine Science, Technology"
               "and Innovation",
        36382: "STMIK PPKIA Pradnya Paramita",
        36383: "Sekolah Tinggi Filsafat Driyarkara",
        36384: "EDIMECO",
        36385: "Universidad Nacional de Colombia - Faculty of Sciences",
        36386: "North-West University",
        36387: "Akademi Farmasi ISFI Banjarmasin",
        36388: "Fakultas Ilmu Administrasi Bisnis dan Ilmu"
               "Komunikasi Universitas Katolik Indonesia Atma Jaya",
        36389: "University of Warsaw - Faculty of Law and Administration",
        36390: "Telos: Revista de Estudios Interdisciplinarios"
               "en Ciencias Sociales, URBE",
        36391: "Museum of Russian Icons",
        36392: "Acta Neerlandica - Bijdragen tot de Neerlandistiek",
        36393: "Sociedad Peruana de Medicina Interna",
        36394: "University Of Sharjah - Scientific Publishing Unit",
        36395: "Internal Design",
        36396: "Media Services",
        36397: "Revista de Educacao Matematica e Tecnologica"
               "Iberoamericana, Universidade Federal de Pernambuco",
        36398: "Petroleo, Royalties e Regiao",
        36399: "University of Glasgow",
        36400: "FTD Resources Publisher",
        36401: "Innovative Healthcare Institute",
        36402: "Escuela Gallega de Administracion Publica - EGAP",
        36403: "Revista Espaco Aberto",
        36404: "Petroleum Exploration Society of Australia",
        36405: "Universita degli Studi di Firenze",
        36406: "Sekolah Tinggi llmu Ekonomi Indonesia",
        36407: "Imperium Institute",
        36408: "RSUP Dr. Kariadi Semarang",
        36409: "STIKES AR RUM Salatiga",
        36410: "Ceramic Research Institute",
        36411: "ANEP",
        36412: "Universitas Negeri Manado - Lembaga Penelitian"
               "dan Pengabdian Kepada Masyarakat",
        36413: "University of Santo Tomas - College of Rehabilitation"
               "Sciences",
        36414: "Fundacao Tiradentes",
        36415: "International Journal of Mountaineering and Climbing",
        36416: "Sociedade Brasileira de Pneumologia e Tisiologia",
        36417: "STAH Dharma Sentana Sulawesi Tengah",
        36418: "CV. Syntax Corporation Indonesia",
        36419: "STIKES Mamba'ul 'Ulum Surakarta",
        36420: "STAI Miftahul Ulum Pamekasan",
        36421: "Sekolah Tinggi Teologi Seminari Alkitab Asia Tenggara",
        36422: "National Association of Phthisiatricians",
        36423: "Universitas Perjuangan Tasikmalaya",
        36424: "Balai Pelestarian Nilai Budaya Sumatera Barat",
        36425: "Federal Research and Clinical Center for Resuscitation"
               "and Rehabilitation",
        36426: "GE Group",
        36427: "Central-European Journal of New Technologies"
               "in Research, Education and Practice",
        36428: "Revista da CGU",
        36429: "The Korea Institute for Religion and Culture",
        36430: "The Korean Society for Psychopathology and"
               "Psychiatric Classification",
        36431: "The Society for Open Parent Education",
        36432: "The Institute for Korean Historical Studies",
        36433: "Korea Anti-Corruption Law Association",
        36434: "Sekolah Tinggi Farmasi Indonesia Perintis",
        36435: "Jurnal Geofisika, Himpunan Ahli Geofisika Indonesia",
        36436: "Universidad de Cundinamarca",
        36437: "Academic and Scholar Publications",
        36438: "Journal of Orthopedic Surgery and Traumatology",
        36439: "Stadium - Hungarian Journal of Sport Sciences",
        36440: "Revista de Educacao Continuada em Medicina"
               "Veterinaria e Zootecnia do CRMV-SP",
        36441: "Universitas Sahid",
        36442: "Afyon Kocatepe Universitesi Akademik Muzik"
               "Arastirmalari Dergisi",
        36443: "Universidad de Burgos",
        36444: "Faculty of Medicine - Udayana University",
        36445: "Instituto de Estudios Internacionales",
        36446: "Sociedad Argentina de Analisis Filosofico",
        36447: "Polish Latin American Studies Association",
        36448: "Universitas Bandar Lampung Publication Center",
        36449: "Revista Tempos Historicos",
        36450: "Revista Perspectiva Geografica",
        36451: "Institut Ilmu Sosial dan Ilmu Politik (IISIP Jakarta)",
        36452: "Universitas Kristen Krida Wacana",
        36453: "Caderno de Educacao Fisica e Esporte",
        36454: "Ciencias Sociais Aplicadas em Revista",
        36455: "Jurnal Kesehatan Bakti Tunas Husada: Jurnal"
               "Ilmu-ilmu Keperawatan, Analis Kesehatan dan Farmasi",
        36456: "Universitas PGRI Adi Buana Surabaya",
        36457: "Persatuan Ahli Gizi Indonesia (Persagi)",
        36458: "Al-Ma mon College Journal",
        36459: "Korean Academy of Organization and Management",
        36460: "Mycological Society of India - MSI",
        36461: "Penza State Agrarian University",
        36462: "International Library of Science",
        36463: "International Display Workshops General Incorporated"
               "Association",
        36464: "Korea Rural Economic Institute Library",
        36465: "Pusat Penelitian dan Pengabdian Masyarakat"
               "STIKes BTH Tasikmalaya",
        36466: "The Korea Landscape Council",
        36467: "Jurnal Makro: Manajemen dan Kewirausahaan,"
               "Universitas Madura",
        36468: "Indian Pharmaceutical Association - IPA",
        36469: "The Journal of Health Economics and Outcomes Research",
        36470: "Faculdade Metropolitana Norte Riograndense - Famen",
        36471: "Journal Club for Condensed Matter Physics",
        36472: "Lycia Press London UK",
        36473: "University of Baghdad - College of Education"
               "- Ibn Rushd for Human Sciences",
        36474: "STIKES Bina Usada Bali",
        36475: "European University Scientific-Research Institute of Law",
        36476: "University of Haripur",
        36477: "Lviv University of Trade and Economics",
        36478: "Medwell Publications",
        36479: "Engineers Without Borders Australian",
        36480: "Wydawnictwo PRYMAT Mariusz Sliwowski",
        36481: "Daffodil International University",
        36482: "Revista Arquivos Brasileiros de Psicologia,"
               "Universidade Federal do Rio de Janeiro",
        36483: "Van Veterinary Journal",
        36484: "Liberal Dusunce Dergisi",
        36485: "Non-profit organization Nephrology",
        36486: "State Research Institute MIA Ukraine",
        36487: "Australian Centre for Geomechanics",
        36488: "Interregional Association for Clinical Microbiology"
               "and Antimicrobial Chemotherapy",
        36489: "MPM Comunicacao",
        36490: "Universitas Tjut Nyak Dhien",
        36491: "The Korean Association of Small Business Studies",
        36492: "The Chongram Historical Society",
        36493: "The Korean Society Of Chinese Studies",
        36494: "Cybercommunication Academic Society",
        36495: "Institute of Plant Protection of NAAS",
        36496: "Asociacion Psicoanalitica Uruguay",
        36497: "Perhimpunan Dokter Paru Indonesia",
        36498: "Korea Bigdata Society",
        36499: "Fakultas Teknik Universitas Wahid Hasyim",
        36500: "Instituto Superior Tecnologico del Azuay",
        36501: "Illinois Center for Transportation",
        36502: "Asploro Open Access Publications",
        36503: "Chettinad Academy of Research and Education",
        36504: "LLC Obyedinennaya Redaktsiya",
        36505: "International Society of Experimental Linguistics",
        36506: "Croatian State Archives",
        36507: "Amirit",
        36508: "Ryazan State Agrotechnological University"
               "named after P.A. Kostychev",
        36509: "Malardalen University",
        36510: "Leading English Education and Resource Network - LEARN",
        36511: "Nizhny Novgorod Academy of the Ministry of"
               "the Interior of Russia",
        36512: "Revista Cubana de Informacion en Ciencias de la Salud",
        36513: "SIGMA, Universitas Madura",
        36514: "Ministry of Justice",
        36515: "Catholic Theological Association of Korea",
        36516: "Cukurova Anestezi ve Cerrahi Bilimler Dergisi",
        36517: "Portal de Periodicos da UFC",
        36518: "HCA Healthcare",
        36519: "Doc Design and Informatics Co. Ltd.",
        36520: "STMIK Methodist Binjai",
        36521: "Center for Connected Health",
        36522: "Publishing & Printing House of Innovative Development",
        36523: "Han-Character Education Research Center",
        36524: "IFES – Instituto Federal do Espirito Santo",
        36525: "Poltekkes Kemenkes Jakarta II",
        36526: "Universitas PGRI Banyuwangi",
        36527: "Korea-China Social Science Association",
        36528: "AL- Mouhakiq Al-Hilly Journal for Legal and"
               "Political Science",
        36529: "Israa University Journal for Humanities",
        36530: "Revue Cahiers Economiques",
        36531: "Iraqi Journal of Desert Studies",
        36532: "Legal Research Institute of Korea University",
        36533: "Ziggurat Publishing",
        36534: "International Association for the Educational"
               "Role of Language",
        36535: "Russian Institute for Scientific and Technical"
               "Information - VINITI RAS",
        36536: "Delhi Ophthalmological Society",
        36537: "Shiv Shiksha Sadan Society",
        36538: "Engine Room Solutions Pty Ltd",
        36539: "Abad iqtisadiyat",
        36540: "Magallat al-igtihad li-l-dirasat al-qanuniyyat"
               "wa al-iqtisadiyyat - Imprime",
        36541: "Al Mansour Journal",
        36542: "Ermeni Arastirmalari",
        36543: "Kafkas Universitesi Iktisadi ve Idari Bilimler"
               "Fakultesi Dergisi",
        36544: "International Academy of Financial Consumers",
        36545: "Fundacion Vida Silvestre Mediterranea",
        36546: "Universitas Palembang",
        36547: "SciCell",
        36548: "Inventive Research Organization",
        36549: "Politeknik Sains dan Teknologi Wiratama Maluku Utara",
        36550: "Volodymyr Vynnychenko Central Ukrainian State"
               "Pedagogical University",
        36551: "University of Warsaw - American Studies Center",
        36552: "Pakistan Society of Neurosurgeons",
        36553: "Wiedza Medyczna Centre of Postgraduate Medical"
               "Education in Warsaw",
        36554: "Al-Istiqlal University Research Journal",
        36555: "Universitas Langlangnbuana",
        36556: "Revista Educacao Online",
        36557: "Brazilian Journal of Implantology and Health Sciences",
        36558: "Revista de Sistemas e Computacao - RSC",
        36559: "Revista Inteligencia Empresarial",
        36560: "Scientific Electronic Archives",
        36561: "Universidad de Montevideo",
        36562: "Revista Paraense de Contabilidade",
        36563: "Universitas Tulungagung",
        36564: "Computer Society of Nahdlatul Ulama",
        36565: "Sekolah Tinggi Ilmu Kesehatan Baiturrahim Jambi",
        36566: "STIKes Mandala Waluya Kendari",
        36567: "Balai Bahasa Jawa Tengah",
        36568: "Jurusan Kesehatan Lingkungan",
        36569: "LPPM STIKES Madani Yogyakarta",
        36570: "Dow University of Health Sciences",
        36571: "Arab Journal for Scientific Publishing - AJSP",
        36572: "Department of Museology - Lusophone University",
        36573: "Badr University in Cairo - BUC",
        36574: "The Indonesian Journal of Development Planning",
        36575: "Ateneum Szkola Wyzsza",
        36576: "Universidad Pontificia de Salamanca",
        36577: "Politeknik Kesehatan Karya Husada Yogyakarta",
        36578: "Fundacja Biografie Codziennosci",
        36579: "Asociacion Espanola de Perfusionistas",
        36580: "Instituto Nacional de Investigacion en Glaciares"
               "y Ecosistemas de Montana",
        36581: "Design Informational Technology Center",
        36582: "Al-Kunooze University College",
        36583: "Intersectional Knowledge Publishers",
        36584: "Infection Prevention and Control Canada",
        36585: "Pre-Construct Archaeology",
        36586: "University of Baghdad - College of Languages",
        36587: "STIE AUB Surakarta",
        36588: "STT BNKP Sundermann",
        36589: "Rawa Sains: Jurnal Sains STIPER Amuntai",
        36590: "LPPM Akper Yapenas 21 Maros",
        36591: "Council of Science Editors",
        36592: "Fundacao Fenix",
        36593: "Sociedad de Neurocirugia de Chile",
        36594: "STIE Pariwisata API Yogyakarta",
        36595: "STMIK Lombok",
        36596: "Universitas Boyolali",
        36597: "Direktorat Pascasarjana Pendidikan, Universitas"
               "Sarjanawiyata Tamansiswa",
        36598: "Dom Helder Revista de Direito",
        36599: "Editora Itacaiunas",
        36600: "Sekolah Tinggi Ilmu Ekonomi Atma Bhakti",
        36601: "Jurnal Migasian, Akademi Minyak dan Gas Balongan",
        36602: "Misurata University",
        36603: "Federal Research Center of Nutrition, Biotechnology"
               "and Food Safety",
        36604: "Far Eastern Scientific Center Of Physiology"
               "and Pathology of Respiration",
        36605: "The Japanese Standing Conference on Organizational Symbolism",
        36606: "Scientific Journal - WEST-EAST",
        36607: "Revista Eletronica da Faculdade Evangelica de Cere",
        36608: "Centennial Journal of Legal and Administrative Sciences",
        36609: "Bangladesh Public Administration Training Centre",
        36610: "Selva Andina Research Society",
        36611: "UP State Association of Otorhinolaryngologists of India",
        36612: "j2pubs",
        36613: "Vilans",
        36614: "Consortium for Computing Sciences in Colleges",
        36615: "University of Johannesburg",
        36616: "Ofort, LLC",
        36617: "NZ Penguin Initiative",
        36618: "Politeknik Pos Indonesia",
        36619: "Journal of Pharmacological Chemistry and Biological Sciences",
        36620: "Association of Buddhist Art",
        36621: "Maggallat Gami at al-Husayn bin Talal li-l-buhut - Print",
        36622: "Voronezh State Technical University",
        36623: "Korean Security Science Association",
        36624: "Fakultas Ilmu Sosial Dan Ilmu Politik Universitas Garut",
        36625: "Asosiasi Pusat Studi Wanita/Gender dan Anak Se-Indonesia",
        36626: "Politeknik Pembangunan Pertanian Yogyakarta"
               "Magelang Jurusan Peternakan",
        36627: "Grebennikov Publishing House",
        36628: "Korean Society of Heart Failure",
        36629: "Angarsk State Technical University",
        36630: "Pearl Mega Publishing Company",
        36631: "Universidad Rafael Landivar",
        36632: "Current Research Web",
        36633: "Utrecht University School of Law",
        36634: "Gauteng City - Region Observatory",
        36635: "Journal of Law and Mobility (JLM)",
        36636: "Fakultas Ekonomi dan Ilmu Sosial, Universitas"
               "Islam Raden Rahmat Malang",
        36637: "Korean Society of Robotic Gynecologic Surgery",
        36638: "Verinotio - Revista On-line de Filosofia e Ciencias Humanas",
        36639: "Michigan Business & Entrepreneurial Law Review (MBELR)",
        36640: "Michigan Journal of Environmental & Administrative"
               "Law (MJEAL)",
        36641: "Michigan Journal of Gender & Law (MJGL)",
        36642: "Michigan Journal of International Law (MJIL)/MJIL"
               "Opinio Juris (MJIL OJ)",
        36643: "Michigan Journal of Race & Law (MJRL)",
        36644: "Michigan Law Review (MLR)/Michigan Law Review"
               "Online (MLROnline)",
        36645: "Michigan Technology Law Review (MTLR)",
        36646: "University of Michigan Journal of Law Reform"
               "(JLR)/University of Michigan Journal of Law"
               "Reform Caveat (JLR Caveat)",
        36647: "Technoarete Research and Development Association",
        36648: "Insight Medical Publishing",
        36649: "Instituto Ecuatoriano de Arbitraje",
        36650: "UCLA Asian American Studies Center",
        36651: "Universidad Nacional Daniel Alcides Carrion",
        36652: "Innovative Mashinostroenie Publishing",
        36653: "Indonesian Conference Network",
        36654: "Kreasi Teknologi Informasi",
        36655: "Universitas HKBP Nommensen Medan",
        36656: "Inkes Deli Husada Deli Tua",
        36657: "Journal of Ibn Haldun Studies, Ibn Haldun University",
        36658: "Jurnal Al iidara Balad, Sekolah Tinggi Ilmu"
               "Administrasi Amunta",
        36659: "Revista DAE",
        36660: "Sociedade Brasileira de Cardiologia",
        36661: "Universidade Federal da Fronteira Sul",
        36662: "Conselho Nacional do Ministerio Publico",
        36663: "STKIP Agama Hindu Singaraja",
        36664: "Institute of Chemical  Technology (ICT)",
        36665: "Universitas Mahaputra Muhammad Yamin",
        36666: "Hong Kong Heptaviolet Cultural Publishing Co., Limited",
        36667: "IAID Ciamis, Jawa Barat",
        36668: "STAI Masjid Syuhada Yogyakarta",
        36669: "Korea Intellectual Property Society",
        36670: "STIT Al-Amin Kreo Tangerang",
        36671: "Pascasarjana Institut PTIQ Jakarta",
        36672: "International Journal of Ayurveda and Alternative Medicine",
        36673: "Uptodate Research Publication",
        36674: "Revista Mythos",
        36675: "Intisari Sains Medis",
        36676: "International Journal for Research Publication"
               "and Seminar, Shodh Sagar",
        36677: "Universidad Autonoma del Estado de Mexico",
        36678: "IJMAES Publisher",
        36679: "Fakultas Kedokteran Universitas Islam Al-Azhar",
        36680: "International Council for Research and Innovation"
               "in Building and Construction",
        36681: "Journal of Turkish Science Education",
        36682: "The College of European and Regional Studies",
        36683: "Orel State University of Economics and Trade",
        36684: "Chechen State University",
        36685: "Indonesian Public Health Association",
        36686: "Ariviyal Publishing",
        36687: "Institute for New Economic Thinking",
        36688: "European Wave and Tidal Energy Conference",
        36689: "University of Hradec Kralove",
        36690: "Scientific Center of Innovative Researches",
        36691: "Farmarus Print Media",
        36692: "Revista CPAQV",
        36693: "Wydawnictwo DiG I. Dacka-Gorzynska, S. Gorzynski Sp.J.",
        36694: "Jurnal Ilmu Manajemen dan Akuntansi Terapan"
               "(JIMAT) STIE Totalwin",
        36695: "Kyiv University of Law of NAS of Ukraine",
        36696: "Ocean Learning Center (OLC)",
        36697: "Korea Youth Activity Research Association",
        36698: "Nizhny Volga Ecocentre Ltd",
        36699: "Reading Acagora Association of Korea",
        36700: "Korean Urban Management Association",
        36701: "Sekolah Tinggi Ilmu Islam dan Bahasa Arab (STIBA) Makassar",
        36702: "Polish Librarians' Association",
        36703: "Instituto Joao Neorico",
        36704: "Editora UEMG - EdUEMG",
        36705: "Publishing House  KNIGA-MEMUAR",
        36706: "Universitas Sriwijaya - Pusat Inovasi Pembelajaran Unsri",
        36707: "Institucion Fernando el Catolico de la Excma."
               "Diputacion de Zaragoza",
        36708: "Vysoka skola technicka a ekonomicka v Ceskych Budejovicich",
        36709: "FKIP Universitas Halu Oleo",
        36710: "Institute of Oilseed Crops NAAS",
        36711: "Pediatric Rehabilitation",
        36712: "Studia Islamika, Center for the Study of Islam"
               "and Society (PPIM) Syarif Hidayatullah State"
               "Islamic University of Jakarta",
        36713: "EPRA JOURNALS",
        36714: "Korea Academic Society for Event Management",
        36715: "Institute of Public Administration",
        36716: "Universidad Tecnica Privada Cosmos",
        36717: "U.S. Association for Computational Mechanics",
        36718: "Krasnoyarsk State Agrarian University",
        36719: "International Journal of Academic Value Studies"
               "(Javstudies JAVS)",
        36720: "Akademi Keperawatan Kerta Cendekia",
        36721: "Pakistan Journal of Pharmaceutical Sciences",
        36722: "Universitas Al Azhar Indonesia",
        36723: "Jurnal Informatika STMIK Dumai",
        36724: "Media-Publisher",
        36725: "Agri-Environmental Sciences - AGRIES",
        36726: "The Korean Society of Art and Media",
        36727: "The Institute of Law & Policy, Jeju National University",
        36728: "Universitas Tunas Pembangunan",
        36729: "STIKES Aisyiyah Palembang",
        36730: "Leipzig University Repository",
        36731: "Cocuk ve Gelisim Dergisi",
        36732: "Revista Nova Paideia",
        36733: "Jurnal Santiaji Pendidikan of Mahasaraswati"
               "Denpasar University",
        36734: "Wade Group",
        36735: "Research Network Lukasiewicz - Institute of"
               "Organization and Management in Industry - ORGMASZ",
        36736: "Psiquiatria Psicodex S.L.",
        36737: "Instituto para la Investigacion Educativa"
               "y el Desarrollo Pedagogico",
        36738: "Malaysian Agricultural Research and Development"
               "Institute - MARDI",
        36739: "West East Institute LLC",
        36740: "ALUNA",
        36741: "Poltekkes Kemenkes Sorong",
        36742: "Scientific Club SOPHUS",
        36743: "Poltekkes Kemenkes Banten",
        36744: "Instytut Sztuki Polskiej Akademii Nauk",
        36745: "International Association for Court Administration",
        36746: "Sekolah Tinggi Ilmu Hukum (STIH) Amsir",
        36747: "Societe Coreenne d'Enseignement de Langue"
               "et Litterature Francaises",
        36748: "Japan Society for Surgical Wound Care",
        36749: "Asosiasi Institusi Pendidikan Kebidanan Indonesia (AIPKIND)",
        36750: "Building Production, State Enterprise 'V.Balytskyi"
               "Scientific Research Institute of Building Production"
               "(SE 'NDIBV')",
        36751: "Revista Direitos Humanos Fundamentais",
        36752: "The International Interdisciplinary Journal"
               "of Education - IIJE",
        36753: "Mathematical Sciences and Applications E-Notes",
        36754: "Sekolah Tinggi Meteorologi Klimatologi dan Geofisika",
        36755: "Department of Islamic Studies, Riphah International"
               "University",
        36756: "Institute of Construction Materials",
        36757: "Malkahtani",
        36758: "Research Centre for Management and Social Studies - RCMSS",
        36759: "Faculty of Veterinary Medicine, University of Ilorin",
        36760: "STIKES Al-Irsyad Al-Islamiyyah Cilacap",
        36761: "LPPM Universitas Teknologi Sumbawa",
        36762: "Badan Penelitian dan Pengembangan Provinsi Jawa Tengah",
        36763: "STIKes Payung Negeri Pekanbaru",
        36764: "Universitas Quality",
        36765: "Pusat Penelitian dan Pengembangan Rekarta Mataram",
        36766: "Indonesian Journal of Accounting and Governance"
               "based at Podomoro University, Jakarta-Indonesia",
        36767: "Universitas Tridinanti Palembang",
        36768: "Sekolah Tinggi Pendidikan Islam (STPI) Bina"
               "Insan Mulia Yogyakarta",
        36769: "STAI Asy-Syukriyyah",
        36770: "Ksiaznica Podlaska",
        36771: "United Arab Emirates University",
        36772: "Arabic Researcher ID",
        36773: "Brest State Technical University",
        36774: "STMIK Dipanegara Makassar",
        36775: "The Korean Association for the Study of Popular Music",
        36776: "Revista do Instituto Brasileiro de Seguranca Publica - RIBSP",
        36777: "Universiti Sains Malaysia Press",
        36778: "Sekolah Tinggi Ilmu Ekonomi Al-Washliyah",
        36779: "Escuela Superior Politecnica de Chimborazo",
        36780: "Ikatan Perawat Maternitas Indonesia Provinsi Jawa Barat",
        36781: "Sekolah Tinggi Agama Islam Al Fithrah",
        36782: "Universitas Bakrie",
        36783: "Revista Brasileira de Ciencia do Solo",
        36784: "Bandung Creative Movement",
        36785: "Balochistan University of Information Technology,"
               "Engineering and Management Sciences",
        36786: "Irish Hospice Foundation",
        36787: "Lembaga Layanan Pendidikan Wilayah IV",
        36788: "SahuarUS",
        36789: "SANUS",
        36790: "Epistemus",
        36791: "Trascender, Contabilidad y Gestion",
        36792: "Vertice Universitario",
        36793: "PSICUMEX",
        36794: "SAVIA",
        36795: "Reconstitucion de Instituciones",
        36796: "BIOLEX",
        36797: "Arte, entre parentesis",
        36798: "CONNOTAS",
        36799: "Estudios LAMBDA",
        36800: "MADGU",
        36801: "Universitatea Politechnica din Bucuresti",
        36802: "Lentera Dua Fondation",
        36803: "Indonesian Physical Medicine and Rehabilitation Association",
        36804: "National Scientific and Research Institute"
               "of Industrial Safety and Occupational Safety and Health",
        36805: "Universitas Buana Perjuangan Karawang",
        36806: "Fakultas Seni Rupa, Institut Kesenian Jakarta",
        36807: "St Petersburg State Institute of Technology"
               "(Technical University)",
        36808: "Forest Research Institute",
        36809: "Omsk State Pedagogical University",
        36810: "Revista de Desenvolvimento Economico",
        36811: "RAFT IT and Publications Pvt Limited",
        36812: "Revista Cientifica Pesquisa Agropecuaria Gaucha (PAG)",
        36813: "Badan Kerjasama Pusat Studi Lingkungan se-Indonesia",
        36814: "Plant Production Institute nd. a. V. Ya. Yuryev of NAAS",
        36815: "Universitas Islam Majapahit Mojokerto",
        36816: "Faculty of Archaeology - Fayoum University",
        36817: "Kulugyi szemle",
        36818: "State Institution - Institute of regional"
               "research named after M.I. Dolishniy of NAS of Ukraine",
        36819: "The Operational Research Society",
        36820: "MTA Konyvtar es Informacios Kozpont",
        36821: "CICS.NOVA",
        36822: "Nordic Press Stockholm KB",
        36823: "Tawasol for Studies and Cultural Awareness Foundation",
        36824: "Atelier Fluxus Virus",
        36825: "Revista de Investigacion en Tecnologias de la Informacion",
        36826: "Cornerstone Institute",
        36827: "Hogeschool Utrecht",
        36828: "University of West London",
        36829: "Universidad de San Carlos de Guatemala",
        36830: "Centre of Excellence in Accounting and Reporting"
               "for Cooperatives",
        36831: "Opscidia",
        36832: "Bangladesh English Language Teachers Association",
        36833: "Association of Speech-Language Pathologists"
               "of the Czech Republic",
        36834: "The Association of Faculties of Medicine of Canada",
        36835: "Koordinatorat Perguruan Tinggi Keagamaan Islam"
               "Swasta Wilayah IV Surabaya",
        36836: "Faculty of Physical Education and Sports,"
               "Alexandru Ioan Cuza - University of IASI, Romania",
        36837: "Chapman University",
        36838: "Terra Science and Education",
        36839: "Firat University",
        36840: "Institut Agama Islam Sunan Giri Bojonegoro",
        36841: "Universitas Abdurachman Saleh Situbondo",
        36842: "International Society of Ocean, Mechanical"
               "and Aerospace Scientists and Engineers (ISOMAse)",
        36843: "Jurnal Telaga Bahasa, Kantor Bahasa Gorontalo",
        36844: "Foundation of Computer Applications (FCA)",
        36845: "Joint Stock Company Scientific and Production"
               "Enterprise 'Pulsar'",
        36846: "Cukurova University, Agriculture Faculty",
        36847: "Korean National Security and Public Safety Association",
        36848: "Medworld Asia Publications",
        36849: "SanovaWorks",
        36850: "JOTE Publishers",
        36851: "California State University, Sacramento",
        36852: "Fundacion Accion Contra El Hambre",
        36853: "Nguyen Tat Thanh University",
        36854: "Widok. Fundacja Kultury Wizualnej",
        36855: "JG Research and Evaluation",
        36856: "Kenya Veterinary Association",
        36857: "Asociacion Nacional de Universidades e Instituciones"
               "de Educacion Superior A.C.",
        36858: "STIKES Dr. Soebandi Jember",
        36859: "Fakultas Ilmu Sosial Dan Ilmu Politik (FISIP)"
               "Universitas Jenderal Achmad Yani (UNJANI)",
        36860: "State University of Land Use Planning",
        36861: "Centro de Formacion en Educacion",
        36862: "Von Gunten and Co. AG - buch and netz",
        36863: "Management et Data Science",
        36864: "The Indonesian Society for Vascular and Endovascular"
               "Surgery - INASVS",
        36865: "Bulgarian Association for Promotion of Education"
               "and Science - BAPES",
        36866: "The Physiological Society",
        36867: "Institute of Economics Karelian Research Centre",
        36868: "Universitatea Gheorghe Asachi din Iasi",
        36869: "PangaderengJurnal hasil penelitian ilmu sosial"
               "dan humaniora,Balai pelestarian nilai budaya sulsel",
        36870: "Institut Teknologi Sains Bandung",
        36871: "The publishing house - SCIENTIFIC LIBRARY",
        36872: "Lepidoptera Research Foundation Inc",
        36873: "Universitas Palangka Raya",
        36874: "Instytut Europy Srodkowej",
        36875: "Narayana College of Nursing",
        36876: "JSMCentral LLC",
        36877: "HH Publisher",
        36878: "Centre for Defence and Security Studies, Massey University",
        36879: "Global Science Library",
        36880: "Eurasian Economists Association",
        36881: "Universidad Andina del Cusco",
        36882: "Revista Terceira Margem Amazonia",
        36883: "Journal of Fiscal And Regional Economy Studies,"
               "Papua University",
        36884: "IUT Press",
        36885: "State Museum of Natural History",
        36886: "School of Religion, Philosophy and Classics,"
               "Howard College Campus - UKZN",
        36887: "Drukarnia Madryd",
        36888: "Union de Universidades de America Latina y el Caribe",
        36889: "Korean Institute of Criminology",
        36890: "International Electronic Journal of Geometry,"
               "Person (Kazim ILARSLAN)",
        36891: "Anadolu, Ankara University",
        36892: "Tawasul International Centre for Publishing,"
               "Research and Dialogue",
        36893: "Scientific Journals International",
        36894: "Korea Commercial Cases Association",
        36895: "Indira School of Business Studies",
        36896: "College Academic Support Programs - CASP",
        36897: "Wroclaw Board of Scientific Technical Societies"
               "Federation NOT",
        36898: "Dixie State University",
        36899: "Pakistan Agricultural Scientists Forum",
        36900: "suburban e.V.",
        36901: "Universidad Catolica San Pablo",
        36902: "Sir Syed College of Education",
        36903: "Physiome Incorporated",
        36904: "Association scientifique et technique pour"
               "l'eau et l'environnement",
        36905: "Sport and Exercise Science New Zealand",
        36906: "Nizhnevartovsk State University",
        36907: "Korea Research Institute for Vocational Education & Training",
        36908: "Sekolah Tinggi Ekonomi dan Bisnis Syariah"
               "(STEBIS) Indo Global Mandiri",
        36909: "Journal of Engineering Research",
        36910: "Lutsknational Technical University",
        36911: "Poltekkes Kemenkes Medan",
        36912: "Misir Politik Arastirmalar Dernegi Iktisadi Isletmesi",
        36913: "ADHAPER: Jurnal Hukum Acara Perdata",
        36914: "STIKS Tarakanita",
        36915: "Institut Agama Islam (IAI DDI) Polewali Mandar",
        36916: "STIKes Panti Waluya Malang",
        36917: "STIA Lancang Kuning Dumai",
        36918: "Federal Center for Educational Legislation",
        36919: "Private Higher Educational Institution European University",
        36920: "Revista Estudos Sociedade e Agricultura",
        36921: "Polish Copernicus Society of Naturalist",
        36922: "Inno Science Press",
        36923: "International Collaboration of Research and Publications",
        36924: "Universiti Brunei Darussalam School of Business"
               "and Economics",
        36925: "Escola de Saude Publica Visconde Saboia",
        36926: "Editora Inovar",
        36927: "Association of Neurologists, Psychiatrists"
               "and Narcologists of Ukraine",
        36928: "Universitas Katolik Indonesia Santu Paulus",
        36929: "Poltekkes Kemenkes Riau",
        36930: "Ukrainian National Forestry University",
        36931: "Journal of Migration Affairs, Centre for Development"
               "Practice and Research, Tata Institute of Social Sciences",
        36932: "STIKES RS Anwar Medika",
        36933: "Hep Verlag AG",
        36934: "Williams College",
        36935: "National Museum of Eastern Carpathians",
        36936: "The Clay Minerals Society of India",
        36937: "ACA Publishing",
        36938: "Minnis Group Pty Ltd ATF the JJM Trust",
        36939: "University of Winnipeg",
        36940: "Merit Research Journals",
        36941: "Richtmann Publishing",
        36942: "Fundacao Universidade Federal do ABC - UFABC",
        36943: "Revista Comsertoes",
        36944: "Music Research Center, Hanyang University",
        36945: "Perm State Institute of Culture",
        36946: "Russian Register of Potentially Hazardous"
               "Chemical and Biological Substances",
        36947: "Geekerd",
        36948: "International Journal for Multidisciplinary Research (IJFMR)",
        36949: "Queens University Belfast",
        36950: "University of Bern",
        36951: "College of Nurses Aotearoa - NZ - Inc",
        36952: "Ukrainian Association of International Law",
        36953: "Action For Sustainable Efficacious Development and Awareness",
        36954: "Escuela Nacional Superior de Folklore Jose Maria Arguedas",
        36955: "Universidad Le Cordon Bleu",
        36956: "Nan Yang Academy of Sciences Pte Ltd",
        36957: "Journal of Aging and Innovation",
        36958: "Sistema de Estudios de Postgrado de la Universidad"
               "de San Carlos de Guatemala",
        36959: "sPage.direcT",
        36960: "University of Hawaii Press",
        36961: "Verlag Dr. Albert Bartens KG",
        36962: "Aspendos International Academy of Medical"
               "and Social Sciences. LTD",
        36963: "Nanjing University of Aeronautics and Astronautics",
        36964: "EUCLID",
        36965: "International Institute for Applied Knowledge"
               "Management - IIAKM",
        36966: "Interdisciplinary Research Journal and Archives",
        36967: "National Park Service",
        36968: "Peace and Development Foundation - PDF",
        36969: "Nordic Journal of European Law, Lund University",
        36970: "Editora da Universidade Estadual do Piaui",
        36971: "Akademi Keperawatan Pasar Rebo",
        36972: "Voice of Wesley: Jurnal Ilmiah Musik dan Agama,"
               "Sekolah Tinggi Teologia Wesley Methodist Indonesia",
        36973: "STIKES Indramayu",
        36974: "Jurnal Biopropal Industri, Balai Riset dan"
               "Standardisasi Industri Pontianak",
        36975: "Sekolah Tinggi Ilmu Ekonomi Riau",
        36976: "Journal of Applied Science, Institut Teknologi Sains Bandung",
        36977: "Revista Essentia",
        36978: "University of Rijeka",
        36979: "Kyrgyz-Russian Slavic University",
        36980: "Casa de Colon, Cabildo de Gran Canaria",
        36981: "The Palestinian Historical Journal",
        36982: "Universitas Indo Global Mandiri",
        36983: "Jurnal Akademi Pariwisata Medan, Politeknik Pariwisata Medan",
        36984: "STIKes MERCUBAKTIJAYA Padang",
        36985: "Universitas Simalungun",
        36986: "Perhimpunan Ahli Pertambangan Indonesia",
        36987: "Universitas Labuhan Batu",
        36988: "PERHUMAS",
        36989: "STKIP Subang",
        36990: "Poltekkes Kemenkes Kendari",
        36991: "Journal of Philia",
        36992: "Fininformservice NIKA",
        36993: "Research Journal Of English",
        36994: "Open International University of Human Developm?nt Ukraine",
        36995: "Universidad Nacional de Misiones",
        36996: "Instituto Nacional de Investigacion y Capacitacion"
               "Continua - Peru",
        36997: "Real Time EOOD",
        36998: "Sekolah Tinggi Ilmu Kesehatan Mutiara Mahakam Samarinda",
        36999: "Korean Association Of Criminology",
        37000: "Odessa State Agrarian University",
        37001: "Sociedade Brasileira de Educacao Matematica",
        37002: "Instituto Chico Mendes de Conservacao da Biodiversidade"
               "- ICBBio",
        37003: "REVISTA PhYLLOS",
        37004: "CREIN, S.L.",
        37005: "Dubna State University",
        37006: "Slavica Lundensia, Lund University",
        37007: "The Korean Association of Robotic Surgeons",
        37008: "Editora BAGAI",
        37009: "African Association for Sustainable Development",
        37010: "Neolectura",
        37011: "Widmaier Verlag",
        37012: "Lembaga Penelitian dan Pengabdian kepada Masyarakat"
               "Universitas Mohammad Husni Thamrin",
        37013: "STIKES Nasional",
        37014: "Perpustakaan Nasional RI",
        37015: "Pringma LLC",
        37016: "Shorenstein Center for Media, Politics, and Public Policy",
        37017: "Kenya Society of Environmental, Biological"
               "and Agricultural Engineers",
        37018: "Fatima Jinnah Medical University",
        37019: "IMAIOS",
        37020: "International Society for Tourism Research",
        37021: "Ram-Eesh Institute of Vocational and Technical Education",
        37022: "South Asian Academic Publications",
        37023: "Faculty of Geotechnical Engineering, University of Zagreb",
        37024: "Direktorat Sejarah, Direktorat Jenderal Kebudayaan,"
               "Kementerian Pendidikan dan Kebudayaan",
        37025: "State Scientific Research Forensic Centre"
               "of the Ministry of Internal Affairs of Ukraine",
        37026: "Rivne Oblast Institute of Postgraduate Pedagogical Education",
        37027: "Editions EHP-BEJUNE",
        37028: "Suryasa and Sons",
        37029: "Shaukat Khanum Memorial Cancer Hospital and Research Centre",
        37030: "AMIK Mahaputra Riau",
        37031: "Fakultas Teknik Universitas Negeri Gorontalo",
        37032: "Revista Mosaicos",
        37033: "Fakultas Matematika dan Ilmu Pengetahuan Alam,"
               "Universitas Negeri Manado",
        37034: "Rektorat Universitas Putra Indonesia YPTK Padang",
        37035: "Fakultas Syariah UIN Sultan Maulana Hasanuddin Banten",
        37036: "Sekolah Tinggi Ilmu Kesehatan Adi Husada",
        37037: "Fakultas Teknik Sipil dan Perencanaan, Universitas"
               "Bung Hatta",
        37038: "Revista Em Curso",
        37039: "Revista Brasileira de Oftalmologia",
        37040: "Czech Geographical Society",
        37041: "Salvador Dali Academy of Contemporary Arts",
        37042: "Asociacion para el Desarrollo del Practicum"
               "y de las Practicas Externas: Red de Practicum - REPPE",
        37043: "CICADIT, University of Bucharest",
        37044: "BioHackrXiv",
        37045: "Acta Silvatica et Lignaria Hungarica",
        37046: "Agribusiness Department Universitas Negeri Gorontalo",
        37047: "Journal of Oncological Sciences (JOS)",
        37048: "Sekolah Tinggi Ilmu Kesehatan Yatsi Tangerang",
        37049: "Servicio de Publicaciones de la Universidad de Cadiz-Book",
        37050: "ICI Berlin",
        37051: "Societe de Reanimation de Langue Francaise",
        37052: "Dewan Bahasa dan Pustaka",
        37053: "Meteo et Climat, Societe Francaise de la Meteorologie"
               "et du Climat",
        37054: "ProSystem",
        37055: "Wojskowa Akademia Techniczna w Warszawie",
        37056: "Muzeum Narodowe w Gdansku",
        37057: "Journal of Research Publication House LLC",
        37058: "LPPM-PMP Universitas Siliwangi Tasikmalaya",
        37059: "The Korean Association of Shipping and Logistics, Inc.",
        37060: "Korea Economic Education Association",
        37061: "Lembaga Penelitian Pengabdian Kepada Masyarakat"
               "Universitas Sriwijaya",
        37062: "Sociologisk Forskning, Swedish Sociological Association",
        37063: "STIKES ABDI Nusantara Jakarta",
        37064: "Fakultas Dakwah dan Komunikasi, Universitas"
               "Islam Negeri Sumatera Utara Medan",
        37065: "Hospital Regional Lambayeque",
        37066: "Asociacion Latinoamericana de la Papa - ALAP",
        37067: "Sociedade Brasileira de Psicopatologia Fenomeno-Estrutural",
        37068: "NASU Institute of Encyclopedic Research",
        37069: "Institute of Applied Mathematics and Mechanics"
               "of the National Academy of Sciences of Ukraine",
        37070: "Alkalmazott Matematikai Lapok",
        37071: "Tatar Research Institute of Agriculture",
        37072: "Societatea Romana de Sexologie Clinica si Procreere Umana",
        37073: "Universidad Nacional Autonoma de Huanta",
        37074: "Kaplan Higher Education Academy Pte Ltd",
        37075: "University of National and World Economy",
        37076: "Asociacion Colombiana de Otorrinolaringologia",
        37077: "Basrah Journal of Agricultural Sciences, College"
               "of Agriculture, University of Basrah",
        37078: "The Folger Shakespeare Library",
        37079: "UIN Sunan Kalijaga Yogyakarta",
        37080: "Nepal Police Hospital",
        37081: "Institut Pendidikan Tapanuli Selatan",
        37082: "International Journal of Innovative Research"
               "in Engineering and Multidisciplinary Physical"
               "Sciences (IJIRMPS)",
        37083: "National and University Library of Bosnia and Herzegovina",
        37084: "REMATEC",
        37085: "Jornal Memorial da Medicina",
        37086: "STIE YPUP Makassar",
        37087: "STMIK Komputama Majenang",
        37088: "Development and Research Center, China Geological Survey",
        37089: "Akademi Farmasi Indonesia Yogyakarta",
        37090: "Universitas Tulang Bawang Lampung",
        37091: "Ewha Historical Institutions",
        37092: "STAI Bumi Silampari Lubuklinggau",
        37093: "Bursa Uludag University",
        37094: "Adiyaman University",
        37095: "Akdeniz University",
        37096: "Ukrainian Association of Educational and Developmental"
               "Psychology",
        37097: "Ukrainian City and Regional Planning Research Center",
        37098: "Center of Paleoethnological Researches",
        37099: "Van Pelt and Opie Library, Michigan Technological University",
        37100: "Institute of Environmental Economics and Sustainable"
               "Development of the National Academy of Sciences of Ukraine",
        37101: "Institute for Physics of Mining Processes"
               "the National Academy Sciences of Ukraine",
        37102: "Far Eastern Branch of the Russian Academy of Sciences",
        37103: "The Korean Association of Public Policy",
        37104: "Indonesian Trust Health Journal, STIKes Murni Teguh",
        37105: "Centrum Rzeczoznawstwa Budowlanego Sp z. o.o.",
        37106: "Botswana Journal of Agriculture and Applied"
               "Sciences, Botswana University of Agriculture"
               "and Natural Resources",
        37107: "School of Health and Allied Sciences Faculty"
               "of Heath Sciences",
        37108: "Fakultas Adab dan Humaniora UIN Imam Bonjol Padang",
        37109: "The Wind Engineering Institute of Korea",
        37110: "Sekolah Tinggi Bahasa Asing IEC Jakarta",
        37111: "BRASPEN Journal",
        37112: "Bisman (Bisnis dan Manajemen): The Journal"
               "of Business and Management (Universitas Islam Majapahit)",
        37113: "The Interior Design - Interior Architecture"
               "Educators Association",
        37114: "Abanico Agroforestal",
        37115: "Revista Multidisciplinar do Sertao",
        37116: "Operador Nacional de Electricidad",
        37117: "Instituto Superior Tecnologico Portoviejo",
        37118: "International Journal of Development Research",
        37119: "University of Regina Library",
        37120: "The Institute of Education of Tsinghua University",
        37121: "Sciengtex Publishing",
        37122: "Korean Academy of Addiction Psychiatry",
        37123: "Sookmyung Research Institute of Humanities",
        37124: "Krasnodar regional public organization - Public"
               "Academy of innovative sustainable development",
        37125: "Universidad de la Republica",
        37126: "Artificial Intelligence in Gastrointestinal Endoscopy",
        37127: "Escuela Superior Tecnologica de Artes Debora Arango",
        37128: "Vinnytsia National Agrarian University",
        37129: "Odesa Military Academy",
        37130: "Universitatea Nationala de Arta Teatrala si"
               "Cinematografica Il. Caragiale",
        37131: "Lviv National Academy of Arts",
        37132: "Asociacion Espanola de Comprension Lectora",
        37133: "Istanbul Medipol Universitesi",
        37134: "Universiti Pendidikan Sultan Idris",
        37135: "Universidad Nacional de Chimborazo",
        37136: "Biskra University",
        37137: "Roa Iktissadia Review",
        37138: "Emir Abdelkader University of Islamic Sciences Review",
        37139: "The Challenge - University of Larbi Ben M'hidi"
               "Oum El Bouaghi",
        37140: "El Hiwar El Thakafi Review - University of Mostaganem",
        37141: "El-Khitab Journal",
        37142: "Donbass State Engineering Academy",
        37143: "Institute of Pig Breeding and AIP NAAS of Ukraine",
        37144: "Journal of Contemporary Islamic Thought",
        37145: "Pusat Pembinaan Analis Kebijkan, Lembaga Administrasi Negara",
        37146: "STAI Al-Azhar Gowa",
        37147: "ELTR Journal, Asosiasi Program Studi Pendidikan"
               "Bahasa Inggris",
        37148: "Puslitbang Sinergis Asa Professional",
        37149: "Jurusan Agribisnis Fakultas Pertanian Universitas Halu Oleo",
        37150: "Universitas Muhammadiyah Sukabumi",
        37151: "Sekolah Tinggi Ilmu Ekonomi STAN IM",
        37152: "The Korean Health & Fundamental Medical Science Society",
        37153: "Russian Association for Earthquake Engineering"
               "and Protection from Natural and Manmade Hazard - RAEE",
        37154: "IJOPEC Publication",
        37155: "Omniscient Pte Ltd",
        37156: "Revista Latino-Americana de Educacao em Astronomia - RELEA",
        37157: "REVISTA FIMCA",
        37158: "Edicoes UVA",
        37159: "Universitas Janabadra",
        37160: "Poltekkes Kemenkes Tasikmalaya",
        37161: "The Korean Research Association of Old Maps",
        37162: "Hydrometcenter of Russia",
        37163: "Shelves Review",
        37164: "Journal of Economic and Financial Research",
        37165: "Al-Riyada for Business Economics Journal",
        37166: "Knowledge Aggregats Review",
        37167: "Journal of Afak for Sciences",
        37168: "Tributaries Journal for Studies and Scientific"
               "Research in Social and Human Sciences",
        37169: "Journal of Research in Finance and Accounting",
        37170: "Finance and Business Economies Review",
        37171: "Al-Moacher Journal of Economic Studies",
        37172: "Contemporary Studies",
        37173: "Institute for Development and International Relations",
        37174: "Non-profit partnership Society of Interventional"
               "Oncoradiologists",
        37175: "Innovation Publishing House Pte. Ltd.",
        37176: "Intersecciones en Antropologia (IEA)",
        37177: "Espacios en Blanco (EB)",
        37178: "CA and C Press AB",
        37179: "MSD Publications",
        37180: "Cadernos PROARQ",
        37181: "Taegu Science University Defense Security Institute",
        37182: "Badan Penelitian dan Pengembangan Daerah Kota Kupang",
        37183: "King Fahad National Library",
        37184: "Liaquat National Hospital and Medical College",
        37185: "National University of Medical Sciences",
        37186: "Digital Cultures Research Centre Press, UWE Bristol",
        37187: "Buzdar Insights",
        37188: "Changchun Institute of Optics, Fine Mechanics"
               "and Physics, Chinese Academy of Sciences",
        37189: "University Library of Duisburg-Essen",
        37190: "Politechnika Wroclawska Oficyna Wydawnicza",
        37191: "Mapsci Digital Publisher OPC Pvt. Ltd.",
        37192: "The Korean Society Of Literary Education",
        37193: "Technical University of Cluj Napoca, North"
               "University Center of Baia Mare",
        37194: "The Indonesia Capital Market Institute",
        37195: "Universitas Ichsan Gorontalo",
        37196: "Institut Agama Kristen Negeri Ambon",
        37197: "Korean Finance Association",
        37198: "Stichting ArtEZ",
        37199: "POLIS University",
        37200: "Hampstead Psychological Associates",
        37201: "Sociedad Espanola de Quimioterapia",
        37202: "Korean Marketing Management Association",
        37203: "Kiev Institute of Business and Technology LLC",
        37204: "National Scientific Center - Institute of"
               "Agricultural Engineering and Electrification",
        37205: "Pedagoguskepzes",
        37206: "American Association of Physicists in Medicine (AAPM)",
        37207: "Climanosco",
        37208: "Glasgow Natural History Society",
        37209: "Jurnal Teknologi Bahan dan Barang Teknik,"
               "Balai Besar Bahan dan Barang Teknik",
        37210: "Korean Society for the Study of Vocational Education",
        37211: "Universidad de Ciencias y Artes de America Latina - UCAL",
        37212: "Journal of Cellular Neuroscience and Oxidative Stress",
        37213: "Canadian Journal of Applied Linguistics",
        37214: "American Finance Association",
        37215: "Antalya Il Milli Egitim Mudurlugu",
        37216: "Institut Agama Islam Hamzanwadi NW Pancor Lombok Timur",
        37217: "The Journal of Turkish Educational Sciences, Gazi University",
        37218: "Dafatir al-buhut al-ilmiyat - University Center"
               "of Tipaza Morsli Abdallah",
        37219: "Vistka LTD",
        37220: "Marine Intellectual Technologies",
        37221: "Asian Agricultural and Biological Engineering Association",
        37222: "Vasyl Stefanyk National Scientific Library"
               "of Ukraine in Lviv",
        37223: "MESOS Disiplinlerarasi Ortacag Calismalari Dergisi",
        37224: "Asociatia Jurnalul Pediatrului",
        37225: "FORMAEMPLEO",
        37226: "Fundacion para el Desarrollo de la Psicologia en el Caribe",
        37227: "Research Publishing Academy",
        37228: "Instituto de Altos Estudios Nacionales",
        37229: "Future Science Association",
        37230: "Ciudad y Territorio Estudios Territoriales",
        37231: "Universiti Sultan Zainal Abidin",
        37232: "Instytut Nauk Prawnych Polskiej Akademii Nauk",
        37233: "Ondokuz Mayis Universitesi Ozel Egitim Bolumu",
        37234: "The University of the West Indies (UWI Press)"
               "Centre - RHQ (Jamaica)",
        37235: "Revista Artes e Ensaios",
        37236: "The Electronic Journal of Combinatorics",
        37237: "Kanda University of International Studies",
        37238: "M. Utemissov West Kazakhstan State University",
        37239: "Zakon Publishing Group",
        37240: "Institute of Philosophy and Sociology, Polish"
               "Academy of Sciences",
        37241: "Journal of Applied and Theoretical Social Sciences",
        37242: "Premium e-Journal of Social Science",
        37243: "Korean Association of Health and Medical Sociology",
        37244: "The Korea English Language Testing Association",
        37245: "Korean Social Theory Association",
        37246: "GRID Architecture, Planning and Design Journal,"
               "Cankaya University",
        37247: "Vide Leaf",
        37248: "Korean Society for Political Thought",
        37249: "Asosiasi Dosen Perguruan Tinggi Islam",
        37250: "Badan Penelitian dan Pengembangan Daerah Provinsi Jambi",
        37251: "Cahaya Ilmu Cendekia",
        37252: "IIQ An Nur Yogyakarta",
        37253: "Universitas Internasional Batam",
        37254: "Korea Society of Industrial Design",
        37255: "Society for Manufacturing Engineers",
        37256: "Universal Wiser Publisher Pte. Ltd",
        37257: "Linguistic Research",
        37258: "Magallat al-qanun al-magribi - Dar Alsalam Publishing",
        37259: "al-madhab al-maliky - South Center for Cultural"
               "and Human Development",
        37260: "Universidad Jose Carlos Mariategui",
        37261: "University of the Basque Country",
        37262: "Fundacion Universitaria Europea de Relaciones"
               "Publicas (ESRP-FUERP)",
        37263: "Advance Business Research Network",
        37264: "Institute for Contemporary Study of Islam",
        37265: "Kashiwagi Solution Model Inc.",
        37266: "Society for Industrial and Systems Engineering",
        37267: "Association Of Chartered Physiotherapists In Neurology",
        37268: "Malaysian Public Health Physicians' Association",
        37269: "Perwakilan BKKBN Daerah Istimewa Yogyakarta",
        37270: "Korea Derivatives Association",
        37271: "NAIR Enterprise, Ivano-Frankivsk Oblast Organization"
               "of RSDP Institute of Social Policy",
        37272: "The Journal of Internet Electronic Commerce Research",
        37273: "Chemical Science Review and Letters",
        37274: "STIBA Arraayah Sukabumi",
        37275: "Scientia Psychiatrica,Perhimpunan Dokter Spesialis"
               "Kedokteran Jiwa Indonesia Region Sumatera Selatan",
        37276: "CV. Social Politic Genius (SIGn)",
        37277: "Institut Sains dan Teknologi Nasional",
        37278: "Universitas Informatika dan Bisnis Indonesia",
        37279: "V.I. Vernadsky Crimean Federal University",
        37280: "The Research Center for Silla Culture of Dongguk University",
        37281: "Genesis Sustainable Future Ltd.",
        37282: "Inference: International Review of Science",
        37283: "Vemu Institute of Technology",
        37284: "East African Nature and Science Organization",
        37285: "BSP Books Private Limited",
        37286: "STIT Muhammadiyah Paciran",
        37287: "Global Health Science Group",
        37288: "The Bukak History Academy",
        37289: "Universitas Megarezky",
        37290: "New Century Health Publishers LLC",
        37291: "Journal of Childhood, Education and Society",
        37292: "Universidad Privada de Pucallpa SAC",
        37293: "SAPIENTIAE, Universidade Oscar Ribas",
        37294: "Institut Teknologi dan Kesehatan Bali",
        37295: "STISIPOL Dharma Wacana",
        37296: "Asosiasi Idebahasa Kepri",
        37297: "Social Innovation Corporation Institute (SICI)",
        37298: "Ambitious Books Private Limited",
        37299: "Al Istihlal Journal",
        37300: "The Korean Society of Confucian Studies",
        37301: "Universitas Bung Hatta",
        37302: "Jurnal Bimas Islam, Kementerian Agama RI",
        37303: "LPPM Universitas Wisnuwardhana",
        37304: "Faculty of Teacher Training and Education,"
               "Universitas Palangka Raya",
        37305: "The Korean Bioethics Association",
        37306: "Badan Kependudukan Dan Keluarga Berencana Nasional",
        37307: "Erich Schmidt Verlag GmbH & Co. KG",
        37308: "Deep Foundations Institute",
        37309: "Victorian Health Promotion Foundation",
        37310: "Centro de Capacitacao Fisica do Exercito",
        37311: "Fakultas Olahraga dan Kesehatan Universitas Negeri Gorontalo",
        37312: "School of Design Universitas Pelita (SODUPH)",
        37313: "Samara Federal research center of the Russian"
               "Academy of Sciences",
        37314: "JNNCE Journal of Engineering and Management"
               "- JJEM, JNN College of Engineering",
        37315: "Sociedad de Traumatologia y Cirugia Ortopedica"
               "de la Comunidad Autonoma Valenciana",
        37316: "Excel Publications",
        37317: "Plant Breeding and Acclimatization Institute"
               "- National Research Institute",
        37318: "Sulime Diseno de Soluciones",
        37319: "Iraqi Association for Medical Research and Studies",
        37320: "International University of Finance",
        37321: "Publishing house Uston Ltd",
        37322: "Green Farming Editor and Publisher",
        37323: "Al-Lisan al-arabi Journal",
        37324: "Journal of Law and Business",
        37325: "Moroccan Journal of Criminal Law and Sciences",
        37326: "Journal of Research and Legal Studies",
        37327: "Epistemology (Journal of Islamic Social Sciences)",
        37328: "JAMSWAP Sekolah Tinggi Ilmu Ekonomi Gempol Pasuruan",
        37329: "Jayapangus Press",
        37330: "Jurnal Prima Ekonomika Sekolah Tinggi Ilmu"
               "Ekonomi YKP Yogyakarta",
        37331: "The Society for Ancient Korean History",
        37332: "Institute of Economics, Technologies and Entrepreneurship",
        37333: "International Association for Research on"
               "Service-Learning and Community Engagement",
        37334: "MUNDIS - Associacao Civica de Formacao e Cultura",
        37335: "International Journal of Entrepreneurial Knowledge",
        37336: "Institute of Nuclear Medicine and Diagnostic"
               "Radiology of the National Academy of Medical"
               "Sciences of Ukraine",
        37337: "Akbid Pelamonia Makassar",
        37338: "Politeknik Jambi",
        37339: "Politeknik Dharma Patria Kebumen",
        37340: "Musyawarah Guru Teknik Otomotif Jawa Timur",
        37341: "Poltekkes Kemenkes Surakarta",
        37342: "Social Welfare Policy and Practice",
        37343: "University of Warsaw - Polish Centre of Mediterranean"
               "Archaeology",
        37344: "TALIM: JOURNAL OF EDUCATION IN MUSLIM SOCIETIES"
               "AND COMMUNITIES, Onder Akademi",
        37345: "Facultad de Medicina de la Universidad Francisco Marroquin",
        37346: "Corte de Constitucionalidad de Guatemala",
        37347: "Tunisian Journal of Philosophical Studies",
        37348: "Sekolah Tinggi Agama Islam Hasan Jufri Bawean",
        37349: "Open Exploration Publishing",
        37350: "RIKEN",
        37351: "Asociacion Andaluza de Cirujanos (ASAC)",
        37352: "Sociedad Andaluza de Patologia Digestiva (SAPD)",
        37353: "Asia-Europe Institute, University Of Malaya",
        37354: "Universidad de Montemorelos",
        37355: "University of Finance and Administration",
        37356: "Colonel Malek Medical College Teachers' Association",
        37357: "Research and Education Promotion Association (REPA)",
        37358: "Revista de Chimie SRL",
        37359: "College of Physical Education and Sport Science",
        37360: "MS-Editions",
        37361: "Society for Academic Primary Care",
        37362: "Stikes Panrita Husada Bulukumba",
        37363: "Babali Health",
        37364: "Sekolah Tinggi Teologi Injili dan Kejuruan (STTIK) Kupang",
        37365: "Sekolah Tinggi Manajemen Informatika dan Komputer"
               "(STMIK) Widuri",
        37366: "LPPM Universitas Pelita Bangsa",
        37367: "Jurnal Perkeretaapian Indonesia, Politeknik"
               "Perkeretaapian Indonesia Madiun",
        37368: "Sekolah Tinggi Theologia Abdiel",
        37369: "Institut Teknologi Medan",
        37370: "Raizes Revista de Ciencias Sociais e Economicas",
        37371: "Kozbeszerzesi Ertesito Plusz",
        37372: "Magyar Rendeszettudomanyi Tarsasag",
        37373: "LPPMPK - Sekolah Tinggi Teknologi Muhammadiyah Cileungsi",
        37374: "Public Law",
        37375: "Sirte University",
        37376: "University of Benghazi",
        37377: "University of Ez-Zitouna",
        37378: "Jurnal Desentralisasi",
        37379: "The Korean Society Of Illustration Research",
        37380: "Adhou Communications AB",
        37381: "Tulane University",
        37382: "Centro Superior de Estudios Universitarios",
        37383: "Skyline University College",
        37384: "Liepaja University",
        37385: "Yayasan Riset dan Pengembangan Intelektual",
        37386: "Altai State Pedagogical University",
        37387: "Universidad Catolica Santa Maria La Antigua",
        37388: "Revista Construcao Psicopedagogica",
        37389: "ABEI Journal",
        37390: "AVANCA CINEMA , Cine-Clube de Avanca",
        37391: "FOREX Publication",
        37392: "Sievert Association",
        37393: "National Sports Academy - Vassil Levski",
        37394: "World Scientific and Engineering Academy and Society (WSEAS)",
        37395: "Sociedad Espanola de Enfermeria Oncologica",
        37396: "Asosiasi Prakarsa Indonesia Cerdas (APIC)",
        37397: "STDI Imam Syafi'i Jember",
        37398: "Banaras Hindu University",
        37399: "Russian State University of Justice",
        37400: "BHMS Editions",
        37401: "Masarat Journal",
        37402: "LPPM Akademi Kebidanan Harapan Ibu Pekalongan",
        37403: "Sekolah Tinggi Ilmu Ekonomi Sultan Agung",
        37404: "Muqabasat Journal",
        37405: "Academy of Economic Sciences of Ukraine",
        37406: "State Agrarian and Engineering University in Podilya",
        37407: "Korea Real Estate Society",
        37408: "Korean Society for Leisure Sciences",
        37409: "The Korean Association for the Social Studies",
        37410: "Comitetul National Roman al Consiliului Mondial al Energiei",
        37411: "PEDAGOGIKA Jurnal Ilmu Pendidikan Fakultas"
               "Ilmu Pendidikan Universitas Negeri Gorontalo",
        37412: "Institut Teknologi Yogyakarta",
        37413: "STIKes Insan Cendekia Husada Bojonegoro",
        37414: "Scientific Production Company Doza, Ltd",
        37415: "Studia Litteraria",
        37416: "University of Panama, Faculty of Medicine",
        37417: "Marcial Pons Ediciones Juridicas y Sociales, S.A.",
        37418: "Union of Researchers of Macedonia",
        37419: "Texas A and M University School of Law",
        37420: "Mason Publish Group",
        37421: "Hilaris SRL",
        37422: "Department of Animal Science and Technology,"
               "Faculty of Agriculture, Nnamdi Azikiwe University",
        37423: "Editora Conhecimento Livre",
        37424: "STMIK IM",
        37425: "East African Science and Technology Commission",
        37426: "WHO Regional Office for Europe",
        37427: "University of Zagreb, Faculty of Science,"
               "Department of Biology",
        37428: "Pasupati, Sekolah Tinggi Agama Hindu Dharma"
               "Nusantara Jakarta",
        37429: "Program Pascasarjana Universitas Kristen Indonesia Maluku",
        37430: "Journal Educational of Nursing based at Akademi"
               "Keperawatan RSPAD Gatot Soebroto Jakarta",
        37431: "Instituto Superior Tecnologico Ruminahui",
        37432: "African Field Epidemiology Network",
        37433: "Advancements in Agricultural Development, Inc.",
        37434: "International Association Welding",
        37435: "NUST Business School",
        37436: "Kharkiv Medical Society",
        37437: "The Institute for the Study of Science of"
               "the Russian Academy of Sciences - ISS RAS",
        37438: "Fakultas Teknologi Informasi - Universitas Merdeka Pasuruan",
        37439: "Dewan Riset DKI Jakarta",
        37440: "Center for European Social Law Implementation Issues",
        37441: "Central European Journal of Educational Research (CEJER)",
        37442: "All-Russian Dairy Research Institute",
        37443: "Federal State Autonomous Scientific Institution"
               "- Scientific Research Institute of the Bakery Industry",
        37444: "Revista Educacao, Psicologia e Interfaces",
        37445: "Institute of Archaeology, National Academy"
               "of Sciences of Ukraine",
        37446: "Journal of Innovative Agriculture",
        37447: "Budkavlen",
        37448: "Diakonian tutkimus",
        37449: "Ennen ja nyt",
        37450: "Hallinnon tutkimus",
        37451: "Kuntoutus",
        37452: "Politiikka",
        37453: "Trio",
        37454: "Tutkimus ja kritiikki",
        37455: "Tyoelaman tutkimus",
        37456: "Tyovaentutkimuksen Vuosikirja",
        37457: "Architectural Research in Finland",
        37458: "St. George Association",
        37459: "STIT Aqidah Usymuni Sumenep",
        37460: "Journal of Political Administrative and Local Studies",
        37461: "European Society of Cardiology",
        37462: "SU2 Foundation",
        37463: "Sociedad Peruana de Medicina Intensiva",
        37464: "Australian Nursing and Midwifery Federation",
        37465: "Sekolah Tinggi Teologi Kristus Alfa Omega",
        37466: "School of Dentistry, Faculty of Medicine, Udayana University",
        37467: "Global Knowledge Academics",
        37468: "Strategy of the Future",
        37469: "Autonomous non-profit scientific and medical"
               "organization - Questions of Oncology",
        37470: "STIE Widya Manggala",
        37471: "Jurnal Pendidikan Modern, STKIP Modern Ngawi",
        37472: "National Academy of Educational Sciences of Ukraine",
        37473: "Rescognito, Inc.",
        37474: "Azerbaijan Oil Industry",
        37475: "Revista Creatividad y Sociedad",
        37476: "Sekolah Tinggi Ilmu Ekonomi (STIE) Nobel Indonesia",
        37477: "Universitas Katolik Darma Cendika",
        37478: "Lembaga Penelitian dan Pengabdian Masyarakat"
               "Universitas Flores",
        37479: "Economic Faculty, Gorontalo State University",
        37480: "STIK Sint Carolus",
        37481: "Sekolah Menengah Kejuruan (SMK) Pustek",
        37482: "M.V. Lomonosov Northern (Arctic) Federal University",
        37483: "Journal of Comprehensive Pharmacy",
        37484: "Akademi Komunitas Manajemen Perhotelan Indonesia",
        37485: "Stroganov Moscow State Academy of Design and Applied Arts",
        37486: "Sociedade Entomologica do Brasil",
        37487: "BASSIRA Center for Research, Consulting and"
               "Learning Services",
        37488: "Journal of Economic and Financial Studies",
        37489: "Publishing House OKI",
        37490: "Pskov State University",
        37491: "Leonid Yuzkov Khmelnytskyi University of Management and Law",
        37492: "Ethnosphere",
        37493: "Bulletin of the North Caucasus Federal University",
        37494: "North Caucasus Federal University - Humanitarian"
               "and Legal Researches",
        37495: "North Caucasus Federal University - Science."
               "Innovations. Technologies",
        37496: "Revista Brasileira de Zootecnia",
        37497: "Centro Universitario Sao Francisco de Barreiras",
        37498: "Institute of Humanities, Keimyung University",
        37499: "JSC Vitae Litera",
        37500: "Chronopub Publications",
        37501: "Soil Science Society of Poland",
        37502: "Amanxo Publication",
        37503: "Institute of Biotechnology and Bioindustry,"
               "Korea National University of Transportation",
        37504: "Jurnal Manajerial Bisnis",
        37505: "Jurnal Aksa, Sekolah Tinggi Seni Rupa dan"
               "Desain (STSRD) VISI",
        37506: "Institute of Medico-legal Publications Private Limited",
        37507: "MAP (Jurnal Manajemen dan Administrasi Publik),"
               "Universitas Wijaya Putra",
        37508: "Revista Convergencia Lusiada",
        37509: "Revija za socijalna politika, Faculty of Philosophy,"
               "Ss. Cyril and Methodius Skopje",
        37510: "Godisen Zbornik, Faculty of Philosophy, Ss."
               "Cyril and Methodius Skopje",
        37511: "Fundacion Universitaria Los Libertadores",
        37512: "Kenya Institute of Food Science and Technology",
        37513: "Cornell University Library",
        37514: "The WAC Clearinghouse",
        37515: "Scientifica Press",
        37516: "Knowvel Journals",
        37517: "Australian Council for Educational Research",
        37518: "Universidad Nacional Autonoma de Chota",
        37519: "Armando Dado Editore SA",
        37520: "National Museum, Czech Republic",
        37521: "Pharmaceutical and Healthcare Sciences Society (PHSS)",
        37522: "Vilnius Academy of Arts",
        37523: "Samara Law Institute of the Federal Penitentiary"
               "Service of Russia",
        37524: "The Institute of Classical Studies of Korea",
        37525: "Pusat Pengembangan Sumber Daya Manusia Minyak dan Gas Bumi",
        37526: "American Nephrology Nurses Association",
        37527: "Sociedad Latinoamericana de Nutricion",
        37528: "University of Belgrade - Faculty of Transport"
               "and Traffic Engineering",
        37529: "Extempore",
        37530: "Pusat Studi Pendidikan dan Kreativitas Anak Purwokerto",
        37531: "Sekolah Tinggi Ilmu Ekonomi Amkop Makassar",
        37532: "Trade Science Inc",
        37533: "Centro Universitario de Sur-Oriente",
        37534: "B.P. Publishing",
        37535: "LSPR Communication and Business Institute",
        37536: "Universidad de Alcala",
        37537: "Facultad de Ingenieria del la Universidad de Buenos Aires",
        37538: "Joint Stock Company Research Center of Construction",
        37539: "HNRI National Development",
        37540: "Centre for International Peace and Stability (CIPS)",
        37541: "LLDIKTI Wilayah IX Sulawesi",
        37542: "Fakultas Tarbiyah, Institut PTIQ Jakarta",
        37543: "Centro Interdisciplinario de Ciencias Marinas",
        37544: "VDI Fachmedien GmbH and Co. KG",
        37545: "Haematology Society of Bangladesh",
        37546: "The Japan Association for Language Teaching (JALT)",
        37547: "The USA Journals",
        37548: "Silabas y Desafios",
        37549: "Anderson Publishing, Ltd.",
        37550: "Thu Dau Mot University",
        37551: "Sociedad Espanola de Enfermeria Nefrologica",
        37552: "Prodiko",
        37553: "Oxygenlink Publishing Company Limited",
        37554: "Archivos Espanoles de Urologia",
        37555: "National Dendrological Park - Sofiyivka",
        37556: "SZIC University of Peshawar",
        37557: "Global Journal of Public Health Medicine",
        37558: "Grupo Espanol del International Institute of Conservation",
        37559: "United Nations Institute for Disarmament Research",
        37560: "Matematichki Bilten, Union of Mathematicians of Macedonia",
        37561: "Korean Social Studies Association",
        37562: "Korean Society of Applied Science for Health"
               "and Beauty Culture",
        37563: "Institute of Cultural Heritage Conservation"
               "Sciences Kongju National University",
        37564: "ANO Editorial The Journal of Secondary Vocational",
        37565: "Korea Center for Public Choice Study",
        37566: "The National School of Judges of Ukraine",
        37567: "Institut Agama Islam Sultan Muhammad Syafiuddin Sambas",
        37568: "Universidad Nacional Autonoma de Mexico",
        37569: "Dalat University",
        37570: "Geological Society of Denmark",
        37571: "Universite du Quebec a Montreal - Departement de Didactique",
        37572: "Editora Artemis",
        37573: "American Society of Health-System Pharmacists",
        37574: "Leningrad Research Institute for Agricultural"
               "Science - BELOGORKA",
        37575: "Scientific Journal of King Faisal University",
        37576: "Three Quays Publishing",
        37577: "Universitas Insan Cendekia Mandiri (UICM)",
        37578: "Iowa State University Extension and Outreach",
        37579: "Ordem Iniciatica do Cruzeiro Divino",
        37580: "Society for Sugarcane Research and Development",
        37581: "The Korea Forest Bioenergy",
        37582: "Dankook University Institute for Convergence Society (DICS)",
        37583: "GERDER-Germanistler Dernegi",
        37584: "Antiquitas Byzantium Renascentia",
        37585: "Revista Higiene Alimentar",
        37586: "Autonomous non-profit organization-Society"
               "of Specialists in the Field Innovative Medical Technology",
        37587: "Revue d'Etudes Francaises",
        37588: "FILOGI",
        37589: "Samarkand State Institute of Foreign Languages",
        37590: "Association for Biology Laboratory Education",
        37591: "Consortium eLearning Network Pvt Ltd",
        37592: "Harvard Medical School Center for Bioethics",
        37593: "Pediatric Ethicscope: The Journal of Pediatric Bioethics",
        37594: "Universidad Metropolitana de Educacion, Ciencia y Tecnologia",
        37595: "Sekolah Tinggi Manajemen Informatika and Komputer"
               "(STMIK) LIKMI",
        37597: "African Journal for Physical Activity and"
               "Health Sciences, Tshwane University of Technology",
        37598: "Universitas Muhammadiyah Aceh",
        37599: "Ceviribilim ve Uygulamaları Dergisi, Hacettepe University",
        37600: "Jurnal Ekonomi Dan Bisnis dan Jurnal Teknik"
               "Informasi Dan Komputer, Politeknik Bisnis Indonesia",
        37601: "Politeknik Sawunggalih Aji",
        37602: "White Wing Publications",
        37603: "Comechingonia",
        37604: "Research Center for Biology-LIPI",
        37605: "Women University Mardan",
        37606: "Sekolah Tinggi Ilmu Administrasi Bina Taruna Gorontalo",
        37607: "The Gathering of Healthcare Simulation Technology"
               "Specialists Inc",
        37608: "Hospital General Docente Ambato",
        37609: "Academician Publishing",
        37610: "Asociacion para el Desarrollo de la Ingenieria"
               "de Organizacion",
        37611: "Instituto Superior Tecnologico Bolivariano de Tecnologia",
        37612: "Politeknik Bumi Akpelni",
        37613: "Journal of History of Sciences",
        37614: "Kola Science Centre",
        37615: "Sociedad Espanola de Imagen Cardiaca",
        37616: "Saudi Heart Association",
        37617: "Cherkassy Experimental Station of Bioresources of NAAS",
        37618: "PARADIGMA",
        37619: "Fatec de Sao Jose dos Campos Prof. Jessen Vidal",
        37620: "EthnoAnthropoZoom, Ss.Cyril and Methodius University, Skopje",
        37621: "National Academy of Medical Sciences of Ukraine",
        37622: "Research India Publications",
        37623: "Scientific Journal of Medical Research",
        37624: "International Research Publication House",
        37625: "University of New Haven - College of Business",
        37626: "WTM-Verlag",
        37627: "Institute for Cultural Research of the National"
               "Academy of Arts of Ukraine",
        37628: "Journals PUB",
        37629: "State Institution Scientific Hydrophysical"
               "Centre of the National Academy of Sciences of Ukraine",
        37630: "STKIP Taman Siswa Bima",
        37631: "Universitas Widya Mataram",
        37632: "Pushkin State Russian Language Institute",
        37633: "Periodico Tche Quimica",
        37634: "JSC Analityk",
        37635: "National Academy of Legal Sciences of Ukraine",
        37636: "Facultad de Ciencias de la Ingenieria y Tecnologia"
               "de la UABC",
        37637: "Universitas Panji Sakti",
        37638: "Perkumpulan Dosen Muda Bengkulu",
        37639: "STMIK Handayani",
        37640: "Jurnal Ilmu Pendidikan (JIP) STKIP Kusuma"
               "Negara, STKIP Kusuma Negara",
        37641: "Institut Bisnis dan Informatika Kesatuan",
        37642: "Korea Real Estate Management Academy",
        37643: "The International Association for Literature"
               "of Korean Residents Abroad",
        37644: "Administrative and Financial Sciences Review",
        37645: "Journal of Jordan Academy for Arabic",
        37646: "Universidad la Salle Pachuca",
        37647: "Institute of Traumatology and Orthopedics"
               "of the National Academy of Medical Sciences of Ukraine",
        37648: "IJRSSH Publication",
        37649: "Anbar Journal of Engineering Science",
        37650: "Iraqi Journal of Civil Engineering",
        37651: "Al-Anbar University Journal of Law and Political Sciences",
        37652: "Journal of University of Anbar for Pure Science",
        37653: "Journal of the University of Anbar for Humanities",
        37654: "Anbar University Journal of Language and Literature",
        37655: "University of Anbar Sport and Physical Education Sciences",
        37656: "Institute of Latin American Studies of the"
               "Russian Academy of Sciences ILA RAS",
        37657: "FGBU VNIIPO of EMERCOM of Russia",
        37658: "Macedonian Geographical Society",
        37659: "Ukrainian-American Concordia University",
        37660: "Acta Scientiarum Polonorum Agricultura, University"
               "of Science and Technology in Bydgoszcz",
        37661: "United Institute of Informatics Problems of"
               "the National Academy of Sciences of Belarus",
        37662: "Journal of Pharmaceutical Technology published"
               "by Mustafa Sinan Kaynak",
        37663: "FSBI TSUREN",
        37664: "Psikofarmakoloji Dernegi (Turkish Association"
               "for Psychpharmacology)",
        37665: "Surface Mount Technology Association",
        37666: "Fundacion Weber",
        37667: "Infomed-Verlags AG",
        37668: "Spanish Society for the Study of Popular Culture (SELICUP)",
        37669: "Milli Egitim Bakanligi Milli Egitim Dergisi",
        37670: "Orenburg State Agrarian University",
        37671: "Kantor Bahasa Kepulauan Bangka Belitung",
        37672: "UK Research Integrity Office",
        37673: "Jurnal Ekonomi Dan Bisnis Indonesia Fakultas"
               "Ekonomi Dan Bisnis Universitas Teknologi Sumbawa",
        37674: "Korea CEO Academy",
        37675: "Academic Society for Appropriate Technology",
        37676: "Unived Press",
        37677: "Siglo Diecinueve (Literatura hispanica), Universitas"
               "Castellae",
        37678: "Journal of Critical and Intensive Care, Turkish"
               "Society of Medical and Surgical Intensive Care",
        37679: "Turkish Radio and Television Corporation",
        37680: "Omah Jurnal Sunan Giri, INSURI Ponorogo",
        37681: "Research Institute of Semiconductor Physics"
               "and Microelectronics at National University"
               "of Uzbekistan named after M. Ulugbek",
        37682: "Xapori Socioambiental",
        37683: "Australian Society of Archivists",
        37684: "Centro de Investigacion y Proyectos en Ambiente"
               "y Desarrollo (CIPAD)",
        37685: "University of the Incarnate Word Libraries",
        37686: "Web of Open Science",
        37687: "Institute of Legislative Forecasts and Legal Expertise",
        37688: "Revista Cientifica da Faculdade Unimed",
        37689: "Acta Paulista de Enfermagem",
        37690: "Publisher House - Public Health Manager Ltd",
        37691: "Moscow Humanitarian Economic University",
        37692: "Bulletin of Academy of Advocacy of Ukraine,"
               "Academy of Advocacy of Ukraine",
        37693: "Public Journal of Semiotics, Lund University",
        37694: "Pusat Penelitian dan Pengabdian Pada Masyarakat"
               "(P3M) STIPER Dharma Wacana",
        37695: "Asosiasi Sinergi Pengabdi dan Pemberdaya Indonesia (ASPPI)",
        37696: "Namik Kemal Tip Dergisi, Tekirdag Namik Kemal University",
        37697: "Eskiyeni",
        37698: "UCYP Press, University College of Yayasan Pahang",
        37699: "Institute of General and Emergency Surgery"
               "Named after V.T. Zaitsev NAMS of Ukraine",
        37700: "Kharkiv Petro Vasylenko National Technical"
               "University of Agriculture",
        37701: "State Scientific Research Institute of Armament"
               "and Military Equipment Testing and Certification",
        37702: "Revista de Ensino em Engenharia",
        37703: "CEFPACS Consulting Limited",
        37704: "The Society of Korea and Japan Education",
        37705: "Cracow University of Technology",
        37706: "BAZNAS Center of Strategic Studies",
        37707: "Nuffield Department of Surgical Sciences",
        37708: "Southwest University Neofit Rilski",
        37709: "Koninklijk Nederlands Geologisch Mijnbouwkundig Genootschap",
        37710: "Ion Creanga State Pedagogical University",
        37711: "Universidad de Huanaco",
        37712: "Kamehameha Publishing",
        37713: "European Federation of Societies for Ultrasound"
               "in Medicine and Biology",
        37714: "Cortical Metrics LLC.",
        37715: "Universitas Ciputra Surabaya",
        37716: "International Society of Neuroscience Publishing"
               "Company Limited",
        37717: "James S. McDonnell Foundation",
        37718: "Svenska Arkeologiska Samfundet",
        37719: "San Beda University",
        37720: "Journal of Astrobiology and Space Science",
        37721: "Universitas Borobudur",
        37722: "Kosmos Publishers",
        37723: "University Medical and Dental College Faisalabad",
        37724: "Ryazan State University named for S. Yesenin",
        37725: "Association Internationale de management strategique (AIMS)",
        37726: "STIE Syariah Indonesia Purwakarta",
        37727: "The Korean Data Analysis Society",
        37728: "Jurnal Pendidikan Rokania, STKIP Rokania",
        37729: "Universitas Muhammadiyah Purworejo",
        37730: "BDK Semarang",
        37731: "Sekolah Tinggi Teologi Berea",
        37732: "The Korean Society of American History",
        37733: "Korean Society of Tax Law",
        37734: "Poltava University of Economics and Trade University",
        37735: "Journal of Camel Health",
        37736: "Korean Association for Literacy",
        37737: "Society for Clinical Epidemiology",
        37738: "All-Russia Research Institute of Hydraulic"
               "Engineering and Land Reclamation of A.N. Kostyakov"
               "(NIIG&M of A.N. Kostyakov)",
        37739: "St. Elizabeth University of Health and Social"
               "Science - Bratislava, Slovakia",
        37740: "Latin American Journal of Clinical Sciences"
               "and Medical Technology",
        37741: "Institute for Tourism",
        37742: "STKIP Pasundan",
        37743: "Korean Social History Association",
        37744: "Uganda Christian University",
        37745: "European Centre for Research Training and Development",
        37746: "National Autonomous University of Mexico",
        37747: "LLC Red Group",
        37748: "ANO -Perspective of Oncology",
        37749: "Yurincom Inter LLC",
        37750: "Scientific Research Institute of Informatics"
               "and Law of the NALS of Ukraine",
        37751: "Yayasan Eka Widya Nusantara",
        37752: "Union for Mental Health",
        37753: "Yayasan Tri Praja Karya Utama",
        37754: "Hasan Kalyoncu University",
        37755: "STKIP Budidaya",
        37756: "Biotechnology Kiosk",
        37757: "Medical Education Cooperation with Cuba",
        37758: "Sekolah Tinggi Agama Islam An-Nawawi Purworejo",
        37759: "Mkwawa University College of Education",
        37760: "Universitas Pandanaran",
        37761: "Sociedad Quimica del Peru",
        37762: "Gandhara University",
        37763: "Pulp and Paper Research Institute",
        37764: "The Nelson-Atkins Museum of Art",
        37765: "Managed Care and Healthcare Communications, LLC",
        37766: "International Platform of Registered Systematic"
               "Review and Meta-analysis Protocols (INPLASY)",
        37767: "Fundacion Universidad Pascal",
        37768: "Universidad Norbert Wiener",
        37769: "Vox. Philosophical Journal",
        37770: "Russian Association of Landscape Architects",
        37771: "Klabat Journal of Nursing, Universitas Klabat",
        37772: "Scientific and Research Institute of Providing"
               "Legal Framework for the Innovative Development",
        37773: "Indian Society for Ecological Economics (INSEE)",
        37774: "Pan American Health Organization",
        37775: "Mernoki es Informatikai Megoldasok - Engineering"
               "and IT Solutions",
        37776: "Universitas Batam",
        37777: "Disciplinarum Scientia: Ciencias da Saude",
        37778: "Disciplinarum Scientia: Sociais Aplicadas",
        37779: "Disciplinarum Scientia: Serie Ciencias Naturais"
               "e Tecnologicas",
        37780: "Disciplinarum Scientia: Serie Ciencias Humanas",
        37781: "Vidya",
        37782: "Thaumazein - Revista Online de Filosofia",
        37783: "Guncel Retina Dergisi",
        37784: "Revista Ipseitas",
        37785: "NAWI Arte Diseno y Comunicacion",
        37786: "American Journal of Oral Medicine",
        37787: "Universidad Nacional de Jaen",
        37788: "Innovative University of Eurasia",
        37789: "Matrix Rom",
        37790: "Acta Numismatica Hungarica",
        37791: "Moscow University for Industry and Finance - Synergy",
        37792: "Universitas Citra Bangsa",
        37793: "The Korea Association for Telecommunications Polices",
        37794: "International Journal of Scientific Research and Reviews",
        37795: "Association of Specialists in the Field of"
               "Pharmacology, Biology and Medicine",
        37796: "China Medical University",
        37797: "University Hospital for Infectious Diseases"
               "- Dr Fran Mihaljevic",
        37798: "International Council on Large Electric Systems"
               "- Cigre, Croatian National Committee",
        37799: "Malaysian Society of Soil Science",
        37800: "MedMedia Kazakhstan",
        37801: "Southeast Asian Regional Center for Graduate"
               "Study and Research in Agriculture (SEARCA)",
        37802: "Universitas Dinamika",
        37803: "Northern Arizona University College of Education",
        37804: "Turiba University",
        37805: "RESOLVE Network",
        37806: "Grupo Laceria",
        37807: "Gordon and Betty Moore Foundation",
        37808: "Southern Public Administration Education Foundation",
        37809: "Shah Abdul Latif University Khairpur",
        37810: "Faculdade FAEL",
        37811: "Asociacion Latinoamericana para el Avance de la Ciencia",
        37812: "Sekolah Tinggi Agama Islam (STAI) Al-Azhar Menganti Gresik",
        37813: "Techland Publishing Group Limited",
        37814: "Revista Cientifica da ABO",
        37815: "RTE Revista Tecnologica ESPOL",
        37816: "A. N. Kosygin Russian State University",
        37817: "Universitas Persada Indonesia Y.A.I",
        37818: "University of Sultan Ageng Tirtayasa-The Center"
               "of Excellence for Food Security",
        37819: "Eurasia Academic Publishing",
        37820: "STEI Hamfara Yogyakarta",
        37821: "Rajasthan University of Health Sciences",
        37822: "Research Institute of Psychiatry Ministry"
               "of Health of Ukraine",
        37823: "Institut Sains dan Teknologi Terpadu Surabaya",
        37824: "Universitas Qamarul Huda Badaruddin Bagu",
        37825: "Translational Medicine, University of Salerno",
        37826: "Lembaga Penelitian dan Pengabdian Kepada Masyarakat,"
               "Sekolah Tinggi Ilmu Komunikasi - AWS",
        37827: "Shevchenko Scientific Society",
        37828: "Institute for Biodiversity and Ecology",
        37829: "Health Foundation",
        37830: "INE-Instituto Nacional de Estadistica de Espana",
        37831: "Akademi Keperawatan Panti Kosala Surakarta",
        37832: "Universitas Katolik Widya Karya Malang",
        37833: "International Coconut Community",
        37834: "Journal of Contemporary Philology, Ss Cyril"
               "and Methodius University, B Koneski Faculty of Philology",
        37835: "Taras Shevchenko Regional Humanitarian-Pedagogical"
               "Academy of Kremenets",
        37836: "Center for Assistance to Social Development"
               "named after Mykola Pyrohov",
        37837: "Directorate-General for Rendering Services"
               "to Diplomatic Missions",
        37838: "Estudios Socioterritoriales. Revista de Geografia",
        37839: "Asia Institute, University of Melbourne",
        37840: "Archivo General de la Nacion - Peru",
        37841: "Indian Dental Association Madras Branch",
        37842: "STKIP Pangeran Dharma Kusuma",
        37843: "Grupo Docentes 20 C.A.",
        37844: "Glokom Katarakt",
        37845: "Gazi Goz Vakfi",
        37846: "The Convergence of Humanities, Social Science"
               "an Art's Academy",
        37847: "Uluslararasi Turk Dunyasi Turizm Arastirmalari Dergisi",
        37848: "AKBID Bhakti Husada Cikarang",
        37849: "Sekolah Tinggi Ilmu Administrasi Pembangunan Jember",
        37850: "STIT Al-Fatah Siman Lamongan",
        37851: "Korean Academy of Neural Rehabilitation",
        37852: "Open Books at Lund University",
        37853: "Pesquisa e Ensino",
        37854: "Alford Council of International English and"
               "Literature Journal",
        37855: "Society for the Advancement of Horticulture",
        37856: "Brazilian Journal of Agriculture",
        37857: "terAs Law Review : Jurnal Hukum Humaniter"
               "dan HAM, terAs, FH Universitas Trisakti",
        37858: "Lembaga Penelitian and Pengabdian Masyarakat"
               "(LPPM) STISIPOL Candradimuka Palembang",
        37859: "LPPM Universitas Muhammadiyah Riau",
        37860: "The Research Center of Religions, Wonkwang University",
        37861: "Meat Industry Journal LLC",
        37862: "A&AePortal",
        37863: "Institute of Mathematics National Academy"
               "of Sciences of Ukraine",
        37864: "Luhansk State Medical University",
        37865: "SixScientia Resources",
        37866: "Metis Presses SARL",
        37867: "Gujarat University",
        37868: "Research and Development Academy",
        37869: "Institute for Research on Innovation and Industrial"
               "System (IRIS)",
        37870: "Association Marocaine pour l'Amelioration"
               "de la Qualite de l'Enseignement",
        37871: "SciRes Literature LLC",
        37872: "Malawi University of Science and Technology",
        37873: "The Ordo Iuris Institute for Legal Culture",
        37874: "Sekolah Tinggi Farmasi Muhammadiyah Cirebon",
        37875: "Sekolah Tinggi Teknologi Angkatan Laut (STTAL)",
        37876: "Asosiasi Dosen Hukum Keluarga Islam (ADHKI) Indonesia",
        37877: "IT and Law Research Institute, Kyungpook National University",
        37878: "Oil Research and Engineering Center of the"
               "National Engineering Academy of the Republic of Kazakhstan",
        37879: "Turk Tarih Kurumu",
        37880: "Journal of Economics and Administrative Sciences"
               "based at Sivas Cumhuriyet University",
        37881: "NeuroPharmac Journal",
        37882: "Scientific Technologies LLC",
        37883: "Saber CDCH-UCV",
        37884: "Kazakh National Agrarian University",
        37885: "Editora Cientifica Digital",
        37886: "Faculty of Organisation Studies Novo Mesto",
        37887: "Fakultas Kesehatan Masyarakat Universitas Halu Oleo",
        37888: "STIEM Bongaya",
        37889: "Perhimpunan Polimer Indonesia (HPI)",
        37890: "Volga State University of Water Transport",
        37891: "Jurusan Pendidikan Fisika FKIP Unipa",
        37892: "Federal State Institution of Science Institute"
               "of Linguistics of the Russian Academy of Sciences",
        37893: "Fakultas Hukum Universitas Krisnadwipayana",
        37894: "Anuario IEHS",
        37895: "Skobelkin State Scientific Center of Laser Medicine",
        37896: "Engineering Skill Development",
        37897: "AMALTEA Medical Publishing House",
        37898: "Association for Spiritual Psychology and Counseling",
        37899: "Newinera Publisher",
        37900: "The Renal Association",
        37901: "California Pharmacists Association",
        37902: "The Korea Association of English Reading Education",
        37903: "Smolensk State Medical University",
        37904: "TANGER Ltd",
        37905: "Universitas Negeri Gorontalo",
        37906: "iSTEAM Co.",
        37907: "University of the Philippines School of Economics",
        37908: "Mustafa Kemal University Journal of Agricultural"
               "Sciences, Hatay Mustafa Kemal University",
        37909: "Kryachkov Novosibirsk State University of"
               "Architecture, Design and Arts",
        37910: "Revista Digital de Postgrado",
        37911: "SFI Press",
        37912: "Wagga Journal of Medicine",
        37913: "RJS Medical Press LLC",
        37914: "Escola Superior de Saude Norte da Cruz Vermelha Portuguesa",
        37915: "Kherson Academy of Continuing Education",
        37916: "Programa de Pos-Graduacao Stricto Sensu em"
               "Arquitetura e Urbanismo da Universidade Sao Judas Tadeu",
        37917: "Iraqi Journal for Electrical and Electronic"
               "Engineering (IJEEE)",
        37918: "The Korea Association of Child Care and Education",
        37919: "Institute of the Ukrainian Language of National"
               "Academy of Sciences of Ukraine",
        37920: "South African Statistical Association (SASA)",
        37921: "Chan Zuckerberg Initiative (CZI)",
        37922: "Prestige Institute of Management",
        37923: "The Korean Society for Phlebology",
        37924: "Honam Historical Association",
        37925: "LLC PH Pig Breeding",
        37926: "Korea Labor and Employment Relations Association",
        37927: "Instituto de Estudios Albacetenses Don Juan Manuel",
        37928: "Nubebe Mibyeong Research Institute",
        37929: "Natural Volatiles and Essential Oils",
        37930: "Institute of New Industrial Development n.a. S.Y. Vitte",
        37931: "Informativo Equestre",
        37932: "STIE Muhammadiyah Jakarta",
        37933: "NIPES Publications",
        37934: "Akademia Baru Publishing",
        37935: "ARTIS-IHA - School of Arts and Humanities"
               "of the University of Lisbon, Portugal",
        37936: "ECTI",
        37937: "Elix Techno Media Pvt. Ltd.",
        37938: "Enso Consulting Pty Ltd",
        37939: "Rawalpindi Medical University",
        37940: "University of Fallujah",
        37941: "United Nations University International Institute"
               "for Global Health",
        37942: "NWV Verlag GmbH",
        37943: "Astana IT University",
        37944: "Institute of Defense Acquisition Program",
        37945: "The Body of Chartered Certified Accountants"
               "and Licensed Accountants of Romania",
        37946: "British Society for Haematology",
        37947: "InnoRenew CoE",
        37948: "Dr. Meghnad Saha College",
        37949: "STISIP Widyapuri Mandiri",
        37950: "HK-Publishing",
        37951: "Associacao Educativa Evangelica",
        37952: "Innovation Publishing House Butlerov Heritage Ltd",
        37953: "The Russian Academy of Art",
        37954: "Revista Electronica Sinergias Educativas",
        37955: "Centro Sur",
        37956: "Journal of Business and Entrepreneurial Studies",
        37957: "Revista Tecnologica Ciencia y Educacion Edwards Deming",
        37958: "Journal of American Health",
        37959: "Centrosur",
        37960: "Revista Venezolana de Gerencia",
        37961: "Revista de Ciencias Sociales",
        37962: "RADS Journal of Pharmacy and Pharmaceutical Sciences",
        37963: "IBERC - Responsabilidade Civil",
        37964: "Canadian Society of Physician Leaders",
        37965: "Intelligence Science and Technology Press Inc.",
        37966: "KIU, Sri Lanka",
        37967: "Hankukeomunhakhoe",
        37968: "Sekolah Tinggi Agama Islam Al-Musaddadiyah Garut",
        37969: "Associacao Brasileira de Psicologia Social",
        37970: "Australian Population Studies",
        37971: "RADIAL: Jurnal Peradaban Sains, Rekayasa dan"
               "Teknologi STITEK Bina Taruna Gorontalo",
        37972: "Yakovlev Chuvash State Pedagogical University",
        37973: "Kazan Law Institute of MIA Russia",
        37974: "Vrije Universiteit Amsterdam Faculty of Law",
        37975: "Faculty of Arts and Humanities, UEA",
        37976: "Instituto Nacional de Neurologia y Neurocirugia",
        37977: "Society of Christian Philosophers",
        37978: "Frontier Science Associates",
        37979: "Escuela de Formacion de Infanteria de Marina",
        37980: "Infomedic International",
        37981: "Humanities Research Institute, Sogang University",
        37982: "Croatian Meteorological Society",
        37983: "International Journal of Dental Materials",
        37984: "Russian University of Cooperation",
        37985: "Perkumpulan Pengelola Jurnal PAUD Indonesia",
        37986: "Pusat Bahasa IAIN Sultan Amai Gorontalo",
        37987: "Medicines of Ukraine",
        37988: "Clinical Dentistry",
        37989: "Gumushane University",
        37990: "Medical Records - International Medical Journal",
        37991: "Sosyolojik Dusun",
        37992: "EJPB, ISPB, CPBG - Tamil Nadu Agricultural University",
        37993: "Krasnoyarsk Regional Youth Public Organization"
               "'Union of Young Scientists of Krasnoyarsk'",
        37994: "Journal of Entrepreneurship and Innovation Management",
        37995: "Journal of the Turkish Chemical Society, Section"
               "C: Chemical Education",
        37996: "Institute of Honam Studies",
        37997: "The Korean Society of Beauty Cultural Arts",
        37998: "Institute for Liberal Arts and Sciences Education",
        37999: "Uluslararasi Dil, Edebiyat ve Kultur Arastirmalari Dergisi",
        38000: "Journal of Universal History Studies",
        38001: "International Journal of Life Sciences and"
               "Biotechnology, International Society of Academicians",
        38002: "Trafik ve Ulasim Arastirmalari Dergisi",
        38003: "Faculty of Humanities and Social Sciences,"
               "University of Split",
        38004: "Sosyal Bilimler Akademi Dergisi",
        38005: "The European Journal for Sport Sciences Technology",
        38006: "Publishing House Znanie-M",
        38007: "Scholar Publishing Group Ltd",
        38008: "Hellenic Aviation Society",
        38009: "Ekonomi Isletme Maliye Arastirmalari Dergisi",
        38010: "Perkumpulan Dosen Program Studi Desain Komunikasi"
               "Visual Indonesia",
        38011: "Indonesian Center for Environmental Law (ICEL)",
        38012: "Journal Badati, Fakultas Ilmu Sosial dan Ilmu Politik, UKIM",
        38013: "Almaz-Antei Air and Space Defence Corporation",
        38014: "East European Institute of Psychology (Ukraine - France)",
        38015: "Uluslararasi Sosyal Bilgilerde Yeni Yaklasimlar"
               "Dergisi (IJONASS)",
        38016: "Journal of Intelligent Systems: Theory and"
               "Applications, Harun TASKIN",
        38017: "Fundacion Universitaria Juan de Castellanos",
        38018: "The Irish Longitudinal Study on Ageing",
        38019: "The Korea Academy of Case Management",
        38020: "Global Bioethics Enquiry Journal",
        38021: "Akdeniz Spor Bilimleri Dergisi (Mediterranean"
               "Journal of Sport Science)",
        38022: "Asian Journal of Hospital Pharmacy",
        38023: "Weblaw",
        38024: "American Review of Political Economy",
        38025: "National Medical Research Center For Rehabilitation"
               "And Balneology",
        38026: "Indonesia Social Justice Network",
        38027: "Alanya Hamdullah Emin Pasa Universitesi",
        38028: "Melentiev Energy Systems Institute Siberian"
               "Branch of the Russian Academy of Sciences",
        38029: "Babcock Medical Society",
        38030: "Index Journal",
        38031: "Drustvo za pomoc otrokom s presnovnimi motnjami",
        38032: "SciEnPG",
        38033: "RILALE UAC",
        38034: "Instituto Noos",
        38035: "Yayasan Dharma Indonesia Tercinta (Dinasti)",
        38036: "Gas Turbine Society of Japan",
        38037: "STIKes Dharma Husada Bandung",
        38038: "Akademi Komunitas Negeri Aceh Barat",
        38039: "Holy Spirit University of Kaslik",
        38040: "Universitas Muhammadiyah Lamongan",
        38041: "Sekolah Tinggi Ilmu Komunikasi Indonesia Maju",
        38042: "Tarla Bitkileri Merkez Arastirma Enstitusu",
        38043: "Universitas Pendidikan Nasional",
        38044: "Limited Liability Company Commonwealth",
        38045: "Asociatia de Biosiguranta si Biosecuritate"
               "din Republica Moldova",
        38046: "Ewha Institute for Biomedical Law and Ethics",
        38047: "Canoa do Tempo",
        38048: "Pusat Penelitian dan Pengabdian Kepada Masyarakat"
               "STKIP Citra Bakti",
        38049: "Caminhos da Historia",
        38050: "Faculty of Economics, Lomonosov Moscow State University",
        38051: "University of Rochester Press",
        38052: "STT Gamaliel",
        38053: "Anatolian Current Medical Journal",
        38054: "International Association of Experts on Earthquake"
               "Engineering",
        38055: "Ryerson University",
        38056: "Consejo Nacional para la Ensenanza y la Investigacion"
               "de las Ciencias de la Comunicacion",
        38057: "Bucak Isletme Fakultesi Dergisi, Burdur Mehmet"
               "Akif Ersoy Universitesi",
        38058: "International Journal of Science Letters",
        38059: "Turkish Journal of Biodiversity, Artvin Coruh University",
        38060: "KARE (International Comparative Journal of"
               "Literature, History and Philosophy), Erciyes University",
        38061: "Natural and Applied Sciences Journal, Izmir"
               "Democracy University",
        38062: "PT Media Krafon Sejahtera",
        38063: "EJONS International Journal of Mathematic,"
               "Engineering and Natural Sciences",
        38064: "Euroasia Journal of Social Sciences and Humanities",
        38065: "Euroasia Journal of Mathematics, Engineering,"
               "Natural and Medical Sciences",
        38066: "EJONS International Journal on Mathematics,"
               "Engineering and Natural Sciences",
        38067: "Maheswari Publishers",
        38068: "Korea Association of Chinese Linguistics",
        38069: "European Distance and E-Learning Network (EDEN)",
        38070: "Hamburger Edition HIS Verlagsges. mbH",
        38071: "Universitatsbibliothek Kiel",
        38072: "Universitatsbibliothek Kiel",
        38073: "Institut Agama Islam Darullughah Waddawah Bangil Pasuruan",
        38074: "CV Oceania Press",
        38075: "Tatar Pasundan, Balai Diklat Keagamaan Bandung",
        38076: "Insan Doktor Ekonomi Indonesia (IDEI)",
        38077: "The Korean Association of Russian Language and Literature",
        38078: "Asian Comparative Folklore Society",
        38079: "Istanbul Gelisim University Journal of Health"
               "Sciences, Istanbul Gelisim University",
        38080: "Yuksabipyungsa",
        38081: "Korean Society For History Of Physical Education,"
               "Sport And Dance",
        38082: "The Institute for Social Work Research, College"
               "of Social, Ewha Womans University",
        38083: "The Institute of Nursing Science, Kyungpook"
               "National University",
        38084: "Korean Police Studies Association",
        38085: "Institute of Socio-Political Research",
        38086: "Main Information and Analytical Center of"
               "the Ministry of Education of the Republic of Belarus",
        38087: "Logos University International",
        38088: "Journal of Innovative Science and Engineering,"
               "Bursa Technical University",
        38089: "Mugla Sitki Kocman Universitesi Egitim Fakultesi"
               "Temel Egitim Bolumu",
        38090: "Revista Educacao e Ciencias Sociais",
        38091: "Sekolah Tinggi Teologi Arastamar Bengkulu",
        38092: "The Journal of Pediatric Academy, Erciyes University",
        38093: "Current Perspectives on Medicinal and Aromatic"
               "Plants (CUPMAP)",
        38094: "Interdisciplinary Publishing Academia",
        38095: "Doctor's Herald",
        38096: "Problems of Biology and Medicine",
        38097: "OOO DSM-Press",
        38098: "Institute of Psychology Russian Academy of Sciences",
        38099: "Police Administration, Dae-Kyeung University",
        38100: "Korea Housing and Urban Guarantee Corporation",
        38101: "STMIK Bina Sarana Global",
        38102: "Jurnal Surya Muda, STIKES Muhammadiyah Kendal",
        38103: "Hue Central Hospital",
        38104: "Lithuania Business University of Applied Sciences",
        38105: "MIT Science Policy Review",
        38106: "Liaquat Medical Research Journal",
        38107: "sui generis Verlag",
        38108: "Ordu University Journal of Nursing Studies, Ordu University",
        38109: "Intermedservice Ltd",
        38110: "International Journal of Oral Health and Medical Research",
        38111: "MYBO Group",
        38112: "Research and Education Development Society (REDS)",
        38113: "The Korean Academy of Taoism and Culture",
        38114: "STKIP PAMANE TALINO",
        38115: "The Academic Society of Global Business Administration",
        38116: "Instituto de Pesquisa Economica Aplicada - IPEA",
        38117: "Formacao em Movimento",
        38118: "University of Oklahoma Press",
        38119: "Center for Applied Cultural Sciences",
        38120: "Bandirma Onyedi Eylul Universitesi",
        38121: "The Korea Port Economic Association",
        38122: "Aksaray University Journal of Institute of Social Sciences",
        38123: "Universidad Finis Terrae",
        38124: "International Journal of Innovative Science"
               "and Research Technology",
        38125: "Journal of Biomedical Science Network LLC",
        38126: "Journal of Science Policy and Governance, Inc.",
        38127: "The University of Queensland Law Journal (UQLJ)",
        38128: "Ciencia y Filosofia",
        38129: "The International Centre for Civil Political Science",
        38130: "Constanta Maritime University",
        38131: "Korea Private International Law Association",
        38132: "Academy for Environment and Life Sciences",
        38133: "Institute for Legal Studies Chonnam National University",
        38134: "National Association of Korean Local Government Studies",
        38135: "Center for Public Interest and Human Rights"
               "Law Chonnam National University",
        38136: "Ministry of Health, Ankara City Hospital",
        38137: "Veterinary Pharmacology and Toxicology Association",
        38138: "Journal of Multidisciplinary Dental Research,"
               "International Dental Educationists Association",
        38139: "Stichting Tijdschrift Sociologie",
        38140: "University of the Free State",
        38141: "Federacion Nacional de Cafeteros de Colombia (CENICAFE)",
        38142: "PT Keberlanjutan Strategis Indonesia",
        38143: "Magyar Nyelvor",
        38144: "Tavol-keleti Tanulmanyok",
        38145: "The Hungarian Historical Review",
        38146: "Belugyi szemle",
        38147: "Facultad de Ciencias Economicas y Empresariales - USFX",
        38148: "Hanoi University of Public Health",
        38149: "Shanghai Huihua Medical Science and Technology Co., Ltd",
        38150: "Horizon Publisher India",
        38151: "Ahmet Kelesoglu Egitim Fakultesi Dergisi",
        38152: "Brazilian Journal of Technology",
        38153: "STIT Al Marhalah Al'Ulya Bekasi",
        38154: "Cappadocia Journal of Area Studies (CJAS),"
               "Cappadocia University",
        38155: "Karadeniz Sosyal Bilimler Dergisi (The Black"
               "Sea Journal of Social Sciences), Giresun University",
        38156: "Universitas Wijaya Putra",
        38157: "Research Innovation Initiative",
        38158: "The European Union Studies Association of Korea",
        38159: "Noyam Publishers",
        38160: "West African Association of Physician Assistants",
        38161: "Federal State Budgetary Educational Institution"
               "of Higher Education Khabarovsk State University"
               "of Economics and Law",
        38162: "Korean Studies Information (KSI)",
        38163: "Science and Society, ANO EE Siberian Institute"
               "of Practical Psychology, Pedagogy and Social Work",
        38164: "Asian Journal of Pharmaceutical Education and Research",
        38165: "Jurnal Kesehatan, Sekolah Tinggi Ilmu Kesehatan Cirebon",
        38166: "mstayyar@mandumah.com",
        38167: "dirasat qanuniyat",
        38168: "Mawarid (Susat)",
        38169: "Al-lisaniyyat (Imprime)",
        38170: "Journal of Advanced Economic Research",
        38171: "Magallat dirasat fi ilm nafs al-sḥat",
        38172: "Amarat fi al-lugat wa al-adab wa al-naqd",
        38173: "Marcau Flavius Cristian",
        38174: "International Journal of Research in Medicine",
        38175: "Journal of Medical, published by Ibrahim Ikizceli",
        38176: "Korean Public Law Association",
        38177: "Nemeth Publishers",
        38178: "Centro de Estudios Publicos",
        38179: "International Journal of Clinical Research",
        38180: "Sociedad Peruana de Derecho Internacional",
        38181: "Center of Endourology Endocenter",
        38182: "Self-employment Fursov A.L.",
        38183: "G.V. Davydova GI",
        38184: "Studies of Koreans Abroad",
        38185: "Korean Association of Cultural Studies",
        38186: "Revista Latinoamericana de Difusion Cientifica,"
               "Editorial Difusion Cientifica",
        38187: "Revista Ensino de Geografia (Recife)",
        38188: "VUZF Reviev, University of Finance, Business"
               "and Entrepreneurship",
        38189: "Sekolah Tinggi Teologi Berita Hidup",
        38190: "Europska poslovna skola Zagreb / European"
               "Business School Zagreb",
        38191: "Asociacion Espanola de Ciencia Regional",
        38192: "British Association of Physicians of Indian Origin",
        38193: "International Journal of Research In Commerce"
               "and Management Studies",
        38194: "Aspikom Wilayah Riau",
        38195: "Urban Design Institute of Korea",
        38196: "Korean Womens Association for Communication Studies",
        38197: "Free Economic Society of Russia",
        38198: "Journal of Management Scholarship",
        38199: "The Korean Society of Ethics",
        38200: "Institute for Insurance and Knowledge Myong Ji University",
        38201: "South African National Biodiversity Institute",
        38202: "High Rate Consulting Publications",
        38203: "Foreign Trade University",
        38204: "Politeknik LP3I",
        38205: "Swami Paramhansa Swastha Siksha Evam Paryavaran"
               "Sanrakshan Samiti",
        38206: "The Medical Journal of Southern California Clinicians",
        38207: "Acquaint Publications",
        38208: "Alwaha Scientific Publishing Services SARL",
        38209: "Taiwan Association of Orthodontists",
        38210: "RussianStudiesHu",
        38211: "Journal of Applied Research in Plant Sciences (JOARPS)",
        38212: "The Journal of Food and Drug Analysis (JFDA),"
               "Food and Drug Administration, Taiwan (TFDA)",
        38213: "OHS Academy",
        38214: "STID Muhammad Natsir",
        38215: "Jurnal Teras Kesehatan, Politeknik Al Islam Bandung",
        46216: "Bouna Research & Studies",
        46217: "Maaallat Ansinat li-l-buhut wa al-dirasat",
        46218: "Ministry of Health",
        46219: "Sociedad Chilena de Educacion Matematica",
        46220: "BioPhorum Operations Group",
        46221: "Nurture Science Publishing Group",
        46222: "AFRICAJOURNALS",
        46223: "Ho Chi Minh City Open University",
        46224: "Economic Consultant",
        46225: "Criminal Investigation Institute at Korean"
               "National Police University",
        46226: "Korea Research Institute for Strategy",
        46227: "Institute of Personality Education",
        46228: "Economic Institute, Yonsei University",
        46229: "Institut Shanti Bhuana",
        46230: "Revista Linguagem em Foco",
        46231: "Akademiar - The Journal of Academic Islamic Studies",
        46232: "Hrvatsko botanicko drustvo",
        46233: "LPPM STIKes Karya Kesehatan",
        46234: "Chinese Center for Disease Control and Prevention",
        46235: "Russian Society of Immunology",
        46236: "Isparta University of Applied Sciences",
        46237: "Adnan Menderes Universitesi Saglik Bilimleri"
               "Fakultesi Dergisi",
        46238: "Journal of Business in The Digital Age (JOBDA)",
        46239: "Karamanoglu Mehmetbey University",
        46240: "Lawyers' Training Centre under the Ministry of Justice",
        46241: "Lown Institute",
        46242: "Industrial University of Ho Chi Minh City (IUH)",
        46243: "Longman Publishers",
        46244: "STKIP Bina Bangsa Getsempena",
        46245: "Indonesia Approach Education",
        46246: "Aerospace Medical Association of Korea",
        46247: "The Society of Namdo Folklore",
        46248: "Korea Institute of Design Research Society",
        46249: "Korean Association of Christian Economists",
        46250: "kulturarastirmalari",
        46251: "Korea Society Of Innovation",
        46252: "Fakultas Perikanan dan Ilmu Kelautan Universitas Papua",
        46253: "Resbee Publisher",
        46254: "IEOM Society International",
        46255: "Asociacion de Amigos del Museo Nacional de Artes Decorativas",
        46256: "RMC Journal of Social Sciences and Humanities,"
               "Research and Management Consultants (Pvt) Ltd",
        46257: "Universitas Islam Jakarta",
        46258: "The James Joyce Society of Korea",
        46259: "The Korean Society for Western History",
        46260: "Korean Society for Learning and Performance",
        46261: "The Journal of Tasan Studies",
        46262: "Korean Theatre Education Association",
        46263: "Center for Religious Studies",
        46264: "The Toegye Studies Institute",
        46265: "Alliance of Bioversity International and CIAT",
        46266: "Nippon Publications",
        46267: "Journal of Tropical Meteorology",
        46268: "The Korean Surgical Ultrasound Society",
        46269: "Askesis",
        46270: "The Institute of Humanities",
        46271: "Korean Society of International Economic Law",
        46272: "Moscow State Institute of International Relations",
        46273: "Journal of Business and Entrepreneurship,"
               "Sampoerna University",
        46274: "Instituto de Ciencias Penais",
        46275: "American Association of State Climatologists",
        46276: "Universidad Nacional Hermilio Valdizan - Facultad"
               "de Ciencias de la Educacion",
        46277: "Editions de Linguistique et de Philologie",
        46278: "Neuropsychologie Clinique et Appliquee",
        46279: "ANO Pravovoye Prosveshcheniye",
        46280: "Magyar Konyvtarosok Egyesulete",
        46281: "Centre for Research on Islamic Banking and"
               "Finance and Business",
        46282: "Bratislava Law Review",
        46283: "Musaica Archaeologica",
        46284: "Muzeologia a kulturne dedicstvo",
        46285: "Acta Physica Universitatis Comenianae",
        46286: "Marketing Science and Inspiration",
        46287: "Journal of Human Resource Management",
        46288: "Testimonia theologica",
        46289: "University of Gloucestershire",
        46290: "ICGE Ltd.",
        46291: "Iktisadi Kalkinma ve Sosyal Arastirmalar Dernegi",
        46292: "American Spinal Injury Association",
        46293: "Digital Forensics (4N6)",
        46294: "Universidade Lusofona do Porto (COFAC)",
        46295: "Sociedad Espanola de Beneficencia",
        46296: "IBKNORGCORP S.A.",
        46297: "National Institute for Materials Science",
        46298: "Centre pour la Communication Scientifique Directe",
        46299: "International Science Group",
        46300: "North Atlantic University Union (NAUN)",
        46301: "Colmena Verlag",
        46302: "ELC Publishing",
        46303: "OpenEd Network",
        46304: "Reschrone Medico Publisher",
        46305: "Sekolah Tinggi Teologi Sumatera Utara",
        46306: "Universitas Bina Bangsa",
        46307: "Universitas Kristen Indonesia",
        46308: "Keimyung University School of Medicine",
        46309: "Biological Diversity and Conservation",
        46310: "Turkish Journal of Internal Medicine",
        46311: "Editora UNINGA",
        46312: "Perspectivas da Educacao Matematica",
        46313: "Al-mi yar (Qasantinat)",
        46314: "Al-Mutargim",
        46315: "Dirasat Insaniyat wa Igtima Iyat",
        46316: "Al-muhtarif (Al-galfat)",
        46317: "Dirasat iqtisadiyat (um al-bawaqi)",
        46318: "Scientia Publishing",
        46319: "Sri Siddhartha Medical College and Research Centre",
        46320: "Eurasian Law Journal",
        46321: "BUGU Dil ve Egitim Dergisi",
        46322: "Korea Research Institute for National Strategy",
        46323: "Asia-Pacific Chiropractic Journal",
        46324: "South-West University",
        46325: "University of Tlemcen",
        46326: "Hanoi University of Mining and Geology",
        46327: "Medical and Surgical Research Journals Group (MSRJGroup)",
        46328: "ISTES Organization",
        46329: "Society of Labor Law Theory and Profession",
        46330: "The Kyong-In Society For Public Administration",
        46331: "Incheon Studies Institute, Incheon National University",
        46332: "Ahi Evran Medical Journal, Kirsehir Ahi Evran University",
        46333: "Society for Global Studies and Research",
        46334: "Institute of Theological Studies in Hanshin University",
        46335: "Innovative Scientific Publication",
        46336: "Research Collaboration Community (RCC)",
        46337: "METEOCEANICS UNIP. LDA.",
        46338: "IJETAE Publication House",
        46339: "Institut Agama Islam Negeri Ternate - IAIN - Ternate",
        46340: "Berostav Druzstvo",
        46341: "M.M. Gryshko National Botanical Garden of the NAS of Ukraine",
        46342: "Revista Brasileira de Cunicultura",
        46343: "Internacional Multidisciplinary Journal of the Brazil",
        46344: "Innovative Association",
        46345: "The Korean Society for Eighteenth-Century English Literature",
        46346: "Humanities Research Institute Chungnam National University",
        46347: "Adichunchanagiri Institute of Medical Sciences",
        46348: "Sekolah Tinggi Teologi Injil Bhakti Caraka Batam",
        46349: "The Association of Humanities and the Social Sciences",
        46350: "The Korean Association for Terrorism Studies",
        46351: "Anatomical Society of UP Chapter",
        46352: "Faculty of Philosophy, University of Sarajevo",
        46353: "Kilis 7 Aralik Universitesi Ilahiyat Fakultesi Dergisi",
        46354: "Cal-Tek Srl",
        46355: "Indonesian Hip and Knee Society",
        46356: "Warta Farmasi, Politeknik Bina Husada Kendari",
        46357: "Boletim do Museu Paraense Emilio Goeldi",
        46358: "Revista Parajas",
        46359: "Ethnobiology Society of Indonesia",
        46360: "Globus Journals",
        46361: "National Institute of Economic Research",
        46362: "Sekolah Tinggi Teologi Widya Agape and Indonesia"
               "Christian Religion Theologian Association",
        46363: "Asociacion Civil Universidad Catolica de Trujillo"
               "UCT Benedicto XVI",
        46364: "University of Agder",
        46365: "Agrisains, Politeknik Hasnur",
        46366: "Ikatan Ahli Kesehatan Masyarakat Indonesia (IAKMI)",
        46367: "Sekolah Tinggi Ilmu Ekonomi (STIE) Syariah Bengkalis",
        46368: "Sekolah Tinggi Keguruan dan Ilmu Pendidikan (STKIP) Melawi",
        46369: "Jurnal Logistik Bisnis Politeknik Pos Indonesia",
        46370: "Society for World Environment, Food and Technology",
        46371: "STIMIK ESQ",
        46372: "Mardin Artuklu University",
        46373: "Halic University",
        46374: "Society for Education in Anesthesia",
        46375: "Revista Encantar",
        46376: "Global Foundation for Nursing Profession and Social Services",
        46377: "Asesorias y tutorias para la Investigacion"
               "Cientifica en la Educacion Puig-Salabarria S.C.",
        46378: "International Research Journal of Science and Technology",
        46379: "Journal of Science, Computing, and Engineering Research",
        46380: "Red Iberoamericana de Medio Ambiente (REIMA, A.C)",
        46381: "Sociedad Espanola de Investigacion Criminologica",
        46382: "Mediterranean Journal of Basic and Applied Sciences",
        46383: "Editora De Castro",
        46384: "Canakkale Onsekiz Mart University Journal"
               "of Marine Sciences and Fisheries, Canakkale"
               "Onsekiz Mart University",
        46385: "Turk Spor Bilimleri Dergisi",
        46386: "Concept Tech Publishing",
        46387: "Mühendislik Bilimleri ve Araştırmaları Dergisi",
        46388: "The Sexology Institute of Romania",
        46389: "Sanatatea Press Group SRL",
        46390: "National Research And Development Institute"
               "For Cryogenic And Isotopic Technologies",
        46391: "Pontificia Universidade Catolica do Rio de Janeiro",
        46392: "The Korean Association of General Education",
        46393: "Pharmacology and Pharmacotherapy",
        46394: "The Research Institute of Physical Education"
               "and Sports Science",
        46395: "International Journal of Research in Applied"
               "Engineering, Science and Technology",
        46396: "The Korean-Japanese Journal of Economics and"
               "Management Studies",
        46397: "Humanities Research Institute (HRI)",
        46398: "Cuestiones Politicas, Universidad del Zulia",
        46399: "Makina Muhendisleri Odasi (Chamber of Mechanical"
               "Engineering)",
        46400: "Uluslararasi Uygur Arastirmalari Dergisi, Adem Oger",
        46401: "Albuquerque: revista de historia",
        46402: "Samvakti Journals",
        46403: "Eszterhazy Karoly Egyetem",
        46404: "Jimma University",
        46405: "Nirvana Psychosocial Care Center and Research Institute",
        46406: "The Korean Society of International Law",
        46407: "Korean Association For Communication and Information Studies",
        46408: "Journal Vox Dei, Sekolah Tinggi Teologi Ekumene Jakarta",
        46409: "University of St. Augustine for Health Sciences Library",
        46410: "Holy Knight Publications",
        46411: "SOAPHYS (Societe Ouest Africaine de Physique)",
        46412: "Austrian Academy of Psychology (AAP)",
        46413: "Bandirma Onyedi Eylul Universitesi Saglik"
               "Bilimleri ve Arastirmalari Dergisi",
        46414: "Elderly Issues Research Journal",
        46415: "Institute of Social Sciences at Dongguk University",
        46416: "The Korean Cadastre Information Association",
        46417: "Journal of Animal Physiology and Nutrition",
        46418: "Saint Petersburg State University of Industrial"
               "Technologies and Design",
        46419: "Croatian Veterinary Institute, Zagreb",
        46420: "Pantanal Editora",
        46421: "Associacao Nacional de Tecnologia do Ambiente"
               "Construido (ANTAC)",
        46422: "The Journal of History and Social Sciences,"
               "University of Karachi",
        46423: "Istanbul Sabahattin Zaim University",
        46424: "PHARMANEST : An International Journal of Advances"
               "in Pharmaceutical Sciences",
        46425: "The University of the West Indies (UWI Press) Mona (Jamaica)",
        46426: "BPSDMP Kominfo Banjarmasin",
        46427: "Geochemical Society",
        46428: "Earl K. Long Library, University of New Orleans",
        46429: "Agricultural Experiment Station, University"
               "of Puerto Rico-Mayaguez",
        46430: "Programming Historian",
        46431: "Middle East Journal of Applied Science & Technology",
        46432: "A Helyi Ertek",
        46433: "Historia Nostra",
        46434: "Acta Universitatis de Carolo Eszterhazy Nominate:"
               "Germanistische Studien",
        46435: "Acta Universitatis de Carolo Eszterhazy Nominatae."
               "Sectio philosophica",
        46436: "Acta Universitatis de Carolo Eszterhazy Nominatae."
               "Sectio paedagogica",
        46437: "Acta Universitatis de Carolo Eszterhay Nominatae."
               "Sectio linguistica Hungarica",
        46438: "Acta Universitatis de Carolo Eszterhazy Nominatae."
               "Sectio historiae",
        46439: "ProBiologists LLC",
        46440: "Acta Universitatis de Carolo Eszterhazy Nominatae."
               "Sectio litterarum",
        46441: "Erdorezervatum Fuzetek",
        46442: "Uluslararasi Kulturel ve Sosyal Arastirmalar Dergisi",
        46443: "Academia de Ciencias Administrativas A.C.",
        46444: "LPPM IKIP Saraswati",
        46445: "Sekolah Tinggi Teologi Simpson",
        46446: "Swiss National Science Foundation",
        46447: "Politeknik Keselamatan Transportasi Jalan",
        46448: "Afkar wa afak",
        46449: "The Mirae English Language and Literature Assoication",
        46450: "Universidad Fidelitas",
        46451: "Clifford Publishing Limited",
        46452: "International Journal of Western Black Sea"
               "Social and Humanities Sciences",
        46453: "Turkish Society for Geomorphology",
        46454: "SUMEDHA JOURNAL OF MANAGEMENT, CMR College"
               "of Engineering & Technology",
        46455: "Helbing and Lichtenhahn Verlag AG (CH) and Co.",
        46456: "Pusbindiklatren Bappenas",
        46457: "Theban Publishing LLC",
        46458: "Faculty of Economics University of Mostar",
        46459: "International Society of Personalized Medicine",
        46460: "International Journal of Innovative Engineering Applications",
        46461: "Smart Freight Centre",
        46462: "National Association for Professional Development Schools",
        46463: "International Journal of Recreation and Sport Science",
        46464: "Turkish Journal of Earthquake Research, Disaster"
               "and Emergency Management Authority",
        46465: "Journal of Industrial Engineering",
        46466: "Iraqi Dental Journal",
        46467: "Elisava Barcelona School of Design and Engineering",
        46468: "Sociedad Argentina de Analisis Politico",
        46469: "Ulster Institute for Social Research",
        46470: "Researcherstore",
        46471: "GigaScience Press",
        46472: "Sophia Centre Press",
        46473: "World Complexity Science Academy",
        46474: "Journals of Design Studio",
        46475: "Foundation for Orphan and Rare Lung Disease (FORLD)",
        46476: "Instituto de Estudios Peruanos",
        46477: "ICOM International Committee for Museums and"
               "Collections of Instruments and Music",
        46478: "EScholar Publication Services",
        46479: "Aisthesis Verlag",
        46480: "GDEON",
        46481: "Nigerian Society of Physical Sciences",
        46482: "Erzincan Binali Yildirim Universitesi Iktisadi"
               "ve Idari Bilimler Fakultesi Dergisi",
        46483: "E-Journal of Dokuz Eylul University Nursing Faculty",
        46484: "Politeknik Ilmu Pelayaran (PIP) Semaran",
        46485: "Journal of the Faculty of Arts and Media",
        46486: "Mezhdunarodnaya Media Gruppa",
        46487: "The Korean Society Of Modern Chinese Literature",
        46488: "Technoscience Publications",
        46489: "OKTAN PRINT s.r.o.",
        46490: "Baltic Forestry",
        46491: "Politeknik Penerbangan Surabaya",
        46492: "Agricultural Technology Development Society",
        46493: "EDITORIAL BOARD FOREIGN AFFAIRS (FOREIGN AFFAIRS-UA)",
        46494: "Sekolah Tinggi Teologi Baptis Indonesia",
        46495: "Sekolah Tinggi Theologia Nazarene Indonesia",
        46496: "Jurnal Medula, Fakultas Kedokteran Universitas Halu Oleo",
        46497: "The Archives of Rheumatology",
        46498: "Coordinacion de Investigacion e Innovacion UPEL-IPB",
        46499: "Wochenschau Verlag Dr. Kurt Debus GmbH",
        46500: "Wallstein Verlag GmbH",
        46501: "International Journal for Modern Trends in"
               "Science and Technology (IJMTST)",
        46502: "Editorial Primmate",
        46503: "Asociacion para la Formacion y la Investigacion"
               "en Ciencias Economicas y Sociales",
        46504: "Park University",
        46505: "International Journal of Biological Innovations",
        46506: "International Society of Culture and Arts",
        46507: "Pusat Penelitian Ilmu Sosial dan Humaniora Kontemporer",
        46508: "STIKes Wijaya Husada",
        46509: "Airman: Jurnal Teknik dan Keselamatan Transportasi,"
               "Politeknik Penerbangan Makassar",
        46510: "Akademi Komunitas Negeri Putra Sang Fajar Blitar",
        46511: "Journal of Science and Humanities",
        46512: "Al-Bahtiyat",
        46513: "Muqarabat falsafiyat",
        46514: "10.46514",
        46515: "Jordanian Education Journal",
        46516: "Bau, Centre Universitari de Disseny",
        46517: "SEAMEO Regional Centre for QITEP in Mathematics",
        46518: "Korean Society of Neuromuscular Disorders",
        46519: "International Journal of 3D Printing Technologies"
               "and Digital Industry",
        46520: "Journal of BRSA Banking and Financial Markets,"
               "Banking Regulation and Supervision Agency of Turkey",
        46521: "Pakistan Association for Women's Studies",
        46522: "Universitatea de Arta din Targu Mures",
        46523: "Journal of Advance Research in Science and"
               "Social Science, IMHRC",
        46524: "Journal of Business and Social Science Research,"
               "Ace Institute of Management",
        46525: "REVISTA ENCONTROS TEOLOGICOS",
        46526: "Associacao Brasileira de Cultura de Tecidos de Plantas",
        46527: "YuMed Text Publications",
        46528: "JUNI KHYAT",
        46529: "Proceedings of SOCIOINT International Conference"
               "on Education and Education of Social Sciences",
        46530: "Instituto Tecnologico y de Estudios Superiores de Monterrey",
        46531: "Sociedade Portuguesa de Neurologia",
        46532: "IJAICT India Publications",
        46533: "Scimpact, LLC",
        46534: "Green Roofs for Healthy Cities",
        46535: "Muzeul National de Istorie a Romaniei",
        46536: "Peoples University of Medical and Health Science"
               "for Women Nawabshah, Shaheed Benazirabad",
        46537: "Faculty of Science, Universiti Brunei Darussalam",
        46538: "National Heritage Language Resource Center",
        46539: "Limited Liability Company Scientific Industrial"
               "Enterprise - Genesis. Frontier. Science",
        46540: "Danmarks Frie Forskningsfond",
        46541: "University of Novi Sad, Faculty of Economics in Subotica",
        46542: "International Pharmaceutical Federation (FIP)",
        46543: "Facultad de Teologia San Isidoro de Sevilla",
        46544: "Technical University of Kosice",
        46545: "American Center of Science and Education",
        46546: "Editorial FLACSO Ecuador",
        46547: "Istanbul Medipol Universitesi Hukuk Fakultesi Dergisi",
        46548: "XXI Century: Resumes of the Past and Challenges"
               "of the Present plus, PenzSTU",
        46549: "Fakultas Peternakan Universitas Papua",
        46550: "Editora Ilustracao",
        46551: "Universidade Estadual de Montes Claros (UNIIMONTES)",
        46552: "Sekolah Tinggi Agama Islam Nahdlatul Ulama Kotabumi Lampung",
        46553: "Fundacion Universidad Catolica Argentina",
        46554: "Samara State University of Economics",
        46555: "Vyzkumny ustav vodohospodarsky T. G. Masaryka, v. v. i.",
        46556: "Global Alliance for Incinerator Alternatives (GAIA)",
        46557: "Asia-Pacific Applied Economics Association",
        46558: "Sekolah Tinggi Teologi Injili Setia Siau",
        46559: "Balai Riset dan Standardisasi Industri Bandar Lampung",
        46560: "Meritum - Revista de Direito da Universidade FUMEC",
        46561: "The Institute for Peace Affairs",
        46562: "Scholars for English Studies in Korea",
        46563: "National Medical Research Center for Childrens Health",
        46564: "Institute for Research and Development India",
        46565: "Yeshwantrao Chavan College of Engineering",
        46566: "Institute of Management and Socio-economic Development, LLC",
        46567: "Asosiasi Teolog Indonesia",
        46568: "University of Karachi",
        46569: "San Francisco State University",
        46570: "University of Toledo",
        46571: "Corporacion Universitaria Autonoma del Cauca",
        46572: "NATURENGS MTU Journal of Engineering and Natural"
               "Sciences, Malatya Turgut Ozal University",
        46573: "Tver State Technical University",
        46574: "Minangkabau Scholar Institute",
        46575: "AgriHumanis: Journal of Agriculture and Human"
               "Resource Development Studies, Balai Pelatihan Pertanian Jambi",
        46576: "Universitas Dharmawangsa",
        46577: "Korea National University of Arts",
        46578: "Harran University Journal of Engineering, Harran University",
        46579: "Scitech Central Inc., USA",
        46580: "University of Melbourne",
        46581: "International Academy of Microwave and Optical"
               "Technology (IAMOT)",
        46582: "G N Corporation Co Ltd",
        46583: "Universidad Catolica de Valencia San Vicente Martir",
        46584: "Faculty of Montenegrin Language and Literature",
        46585: "Univerzita Pardubice",
        46586: "Universitatsbibliothek der Ruhr-Universitat Bochum",
        46587: "Institute of Sustainable Development, Environmental"
               "and Scientific Research (ISDESR)",
        46588: "INVURNUS",
        46589: "Revista de Investigacion Academica sin Frontera",
        46590: "Research Journal of Textile and Leather",
        46591: "P.P. Revista JURNALUL UMANITAR MODERN S.R.L.",
        46592: "Turkish Journal of Agricultural Engineering Research",
        46593: "Bold Rapid Net. Pvt. Ltd.",
        46594: "ROSOMED",
        46595: "Ozcan Gungor",
        46596: "Kebikec, Insan Bilimleri icin Kaynak Arastirmalari Dergisi",
        46597: "Instituto de Psicologia - USP",
        46598: "Impact Journal of Science and Technology,"
               "Impact Care Foundation",
        46599: "Gyor-Moson-Sopron Megyei Epitesz Kamara",
        46600: "Pakistan Society of Religions (PSR)",
        46601: "Jurnal JURIDICA Universitas Gunung Rinjani",
        46602: "Chemical Society of Nigeria",
        46603: "East Delta University",
        46604: "Taiwan Association of Engineering and Technology Innovation",
        46605: "Sociedad Iberoamericana de Estudios Heideggerianos",
        46606: "Gitoya Centre for Academic Research and Dissemination",
        46607: "International Ayurvedic Medical Journal",
        46608: "University Bordeaux Montaigne",
        46609: "MIJP Publication",
        46610: "MAT Journals (A unit of ARV Infomedia Pvt. Ltd.)",
        46611: "Journal of Medical Sciences, Aksaray University",
        46612: "Korea Association of Chinese Language and Literature",
        46613: "Sociedad Chilena de Gastroenterologia",
        46614: "International Journal of Old Uyghur Studies",
        46615: "Izdatelstvo Prospekt",
        46616: "Revista Cientifica Educacao Online",
        46617: "SRO media",
        46618: "Red de Centros de Investigacion en Matematica Educativa",
        46619: "Tridha Scholars Publishing Pvt Ltd",
        46620: "International Union of Radio Science (URSI)",
        46621: "Society of Family Planning",
        46622: "The Southern African Geography Teachers' Association (SAGTA)",
        46623: "National Translation Mission, Central Institute"
               "of Indian Languages",
        46624: "Epixpub Publication",
        46625: "International Journal of Futures Research and Development",
        46626: "Asian Forum For Corrections",
        46627: "SI - Scientific Online Journal",
        46628: "Artuklu Human and Social Science Journal,"
               "Mardin Artuklu University",
        46629: "Istanbul Yeni Yuzyil Universitesi, Yeni Yuzyil"
               "Journal of Medical Sciences",
        46630: "University of Nis - Faculty of Philosophy",
        46631: "Universidad ESAN",
        46632: "REST Publisher",
        46633: "Scholars Publishing, LLC",
        46634: "Fundacion Universitaria Maria Cano",
        46635: "REVISE - Revista Integrativa de Inovacao Tecnologica"
               "nas Ciencias da Saude",
        46636: "Instituto Federal do Norte de Minas Gerais",
        46637: "Southeast Asia Mental Health and Counseling Association",
        46638: "Perhimpunan Mikologi Indonesia (Mikoina)",
        46639: "The Donghak Society",
        46640: "Center for Philosophy of Media and Media Research",
        46641: "Medeniyet Sanat Dergisi, Istanbul Medeniyet University",
        46642: "Lecturas: Educacion Fisica y Deportes",
        46643: "Perkumpulan Doktor Indonesia Maju",
        46644: "DNIPROPETROVSK RESEARCH INSTITUTE OF FORENSIC EXAMINATIONS",
        46645: "Scholars for English Studies in Korea",
        46646: "International Sakharov Environmental Institute"
               "of Belarusian State University",
        46647: "International Journal of Engineering Technology"
               "and Management Sciences",
        46648: "Gevher Nesibe Journal, IESDR",
        46649: "Al-Furat Journal of Innovations in Electronics"
               "and Computer Engineering, ATU",
        46650: "Universitas Dwijendra",
        46651: "The Archaeological Society Malta",
        46652: "Centro de Investigaciones en Ciencias Sociales"
               "y Humanidades desde America Latina",
        46653: "Pakistan Society for Horticultural Science",
        46654: "IJAAR Publishing",
        46655: "Fe Journal",
        46656: "ACCESS",
        46657: "Laboratory of Sustainable Development and"
               "Computer Science (LDDI)",
        46658: "JBI",
        46659: "Escuela Colombiana de Ingenieria Julio Garavito",
        46660: "International Journal of Economic and Environmental"
               "Geology (IJEEG), published by SEGMITE",
        46661: "Universidad Pablo de Olavide",
        46662: "Lahore College for Women University, Lahore",
        46663: "Jinnah Sindh Medical University",
        46664: "Sociedade Paulista de Radiologia e Diagnostico por Imagem",
        46665: "The Association of Korean Economic Studies",
        46666: "Kazakh Research Institute of Economy of Agro-Industrial"
               "Complex and Rural Development",
        46667: "Revista de Ensino de Biologia",
        46668: "STIKes Panti Rapih Yogyakarta",
        46669: "Korea Sport Society",
        46670: "The Korean Sensors Society",
        46671: "UFM Madrid",
        46672: "Libertas medunarodno sveuciliste",
        46673: "Adab",
        46674: "Bulletin of Sudanese Studies",
        46675: "Revista Brasileira de Ciencias Biomedicas",
        46676: "International Journal on Food, Agriculture,"
               "and Natural Resources (IJFANRES), Jember University",
        46677: "Compendium Cuadernos de Economia y Administracion",
        46678: "Plant Biology iPosters",
        46679: "CSMFL Publications",
        46680: "Himpunan Pengembang Kurikulum Indonesia Wilayah Jawa Tengah",
        46681: "Consejo de Formacion en Educacion",
        46682: "IJHBS Publication",
        46683: "Journal of Molecular Virology and Immunology",
        46684: "Educational and Instructional Center for Railway"
               "Transportation",
        46685: "DAADStudien",
        46686: "HSIMonitor",
        46687: "Department of English Studies, Faculty of"
               "Humanities, Konstantin Preslavsky University of Shumen",
        46688: "Archivos de historia del movimiento obrero y la izquierda",
        46689: "Joint-Stock Company - Yuzhuralzoloto Group Of Companies",
        46690: "Yandy Scientific Press",
        46691: "Guarda Polytechnic Institute",
        46692: "Bristol University Press",
        46693: "National Remote Sensing Centre",
        46694: "Korean Association of Slavic-Eurasian Studies",
        46695: "Asian Society for Sport Convergence Sciences",
        46696: "REVISTA AFRICA E AFRICANIDADES",
        46697: "Academy of International Business",
        46698: "Vladikavkaz Scientific Centre of the Russian"
               "Academy of Sciences",
        46699: "Fundacao Universitaria do Desenvolvimento do Oeste Fundeste",
        46700: "Asiatic Society for Social Science Research",
        46701: "Blood and Genomics",
        46702: "Jurnal Ilmu Kepolisian",
        46703: "Fakultas Kehutanan Universitas Papua",
        46704: "Krishma P.G. College of Education",
        46705: "Federation of Sport Pedagogues of Republic of Macedonia",
        46706: "Sociedad de Psiquiatria del Uruguay",
        46707: "University of Birmingham Library Services",
        46708: "Lembaga Kajian Islam dan Hukum Islam, Faculty"
               "of Law, Universitas Indonesia",
        46709: "Ohio Public Health Association",
        46710: "Centre d Estudis Demografics (CED)",
        46711: "Association science et bien commun",
        46712: "Carolina Small Business Development Fund",
        46713: "Bulgarian Defence Institute - BDI",
        46714: "Simons Foundation",
        46715: "SkepticMed Publishers",
        46716: "International Society for the Study of Trauma"
               "and Dissociation",
        46717: "Union of Iraqi Geologists (UIG)",
        46718: "Open Access Journal of Biogeneric Science and Research",
        46719: "Dynamic Publishers",
        46720: "FISITA",
        46721: "Fundacion para la Educacion y Calidad en Investigacion"
               "Medica Ecuatoriana (FECIM-ECUADOR)",
        46722: "Al-Hikmah Research and Publication Center",
        46723: "Endless Science Ltd",
        46724: "Noospheric Studies",
        46725: "Intelligentsia and the World",
        46726: "Ivanovo State University. Series: The Humanities",
        46727: "Revista de Studii Socio-Umane",
        46728: "Revista Psihologie. Pedagogie speciala. Asistenta sociala",
        46729: "Inara",
        46730: "Jurnal Administrasi Politik dan Sosial, Fakultas"
               "Ilmu Sosial dan Ilmu Politik Universitas Riau",
        46731: "Relicario: Revista do MAS",
        46732: "Acta Fisheries and Aquatic Resources (ActaPesca News)",
        46733: "Ss. Cyril and Methodius University, Faculty"
               "of Physical Education, Sport and Health",
        46734: "Centro de Investigacion de la Sociedad del Conocimiento",
        46735: "Instituto Andaluz de Administracion Publica",
        46736: "Eurasian Scientific and Medical Journal Sino",
        46737: "Ekonomi Maliye Isletme Dergisi",
        46738: "The Korean Association for Laser, Dermatology and Trichology",
        46739: "Debreceni Reformatus Hittudomanyi Egyetem",
        46740: "ALKU Fen Bilimleri Dergisi, Alanya Alaaddin"
               "Keykubat University",
        46741: "Vologda Institute of Law and Economics of"
               "the Federal Penal Service of Russia",
        46742: "Institute of Developmental Physiology RAE",
        46743: "Nova Southeastern University",
        46744: "Academia Peruana de la Lengua",
        46745: "ILMA University",
        46746: "Brooke: Action for Working Horses and Donkeys",
        46747: "The College of Family Physicians of Canada",
        46748: "Iskusstvo Evrazii",
        46749: "AKPER Fatmawati",
        46750: "Jurnal Economica, ISEI Cabang Pekanbaru",
        46751: "National Public Law Review",
        46752: "Associacao Nacional de Pesquisadores e Professores"
               "de Historia das Americas (ANPHLAC)",
        46753: "Poincare Publishers",
        46754: "Penerbit UMT, Universiti Malaysia Terengganu",
        46755: "Vietnam Association of Obstetrics and Gynecology - VAGO",
        46756: "Food Standards Agency",
        46757: "Sekolah Tinggi Ilmu Ekonomi PGRI Sukabumi",
        46758: "Korean Law and Economics Association",
        46759: "Irish Interdisciplinary Journal of Science and Research",
        46760: "The Jungbu Archaeological Society",
        46761: "Revista Cientifica da Faculdade de Balsas",
        46762: "Maarif Mektepleri-International Journal of"
               "Educational Sciences",
        46763: "Goce Delchev University - Shtip",
        46764: "STMIK Syaikh Zainuddin NW Anjani-Lombok Timur",
        46765: "Journal of Bone Marrow Transplantation and Cellular Therapy",
        46766: "Inovacus Publication",
        46767: "Kenan Fellows Program for Teacher Leadership",
        46768: "Sociedad Argentina de Coloproctologia",
        46769: "Centre for Democracy, Research and Development (CEDRED)",
        46770: "Atomic Spectroscopy Press Limited",
        46771: "Buske Verlag",
        46772: "Universitas Muhadi Setiabudi",
        46773: "STAI Muhammadiyah Probolinggo",
        46774: "Journal of Publikasi Penelitian Terapan dan"
               "Kebijakan, Balitbangda Sumatera Selatan",
        46775: "Institute of Global Business Research",
        46776: "Journal of Arts Entrepreneurship Education,"
               "University of Memphis",
        46777: "Prodi Pendidikan Agama Kristen, FKIP, Universitas"
               "Kristen Indonesia",
        46778: "International Journal of Turkish Education"
               "Sciences, Gaziosmanpasa University",
        46779: "University of Bihac, Faculty of Economics",
        46780: "Centro Universitario de Zacapa",
        46781: "STAI Diniyah Pekanbaru",
        46782: "Belarus State Economic University",
        46783: "Electronic Scientific Journal Intellectualization"
               "of Logistics and Supply Chain Management",
        46784: "Scientific Veterinary Institute 'Novi Sad'",
        46785: "CIID Journal",
        46786: "Philosophy of Education Society of Australasia",
        46787: "California State University, Office of the Chancellor",
        46788: "Medical Practice Scientific Society",
        46789: "Revista Brasileira de Educacao em Geografia",
        46790: "Erzincan Universitesi Sosyal Bilimler Enstitusu"
               "Dergisi, Erzincan Binali Yuldurum Universitesi",
        46791: "Unique Journals Communication",
        46792: "Faculty of Engineering, Federal University Oye-Ekiti",
        46793: "University Library in Kragujevac",
        46794: "Universidad Nacional Hermilio Valdizan - Escuela de Posgrado",
        46795: "International Journal of Health care and Biological"
               "Sciences (IJHCBS)",
        46796: "International Journal of Pharmacognosy and Chemistry (IJPC)",
        46797: "International Journal of Alternative and Complementary"
               "Medicine (IJACM)",
        46798: "Indian Association of Applied Microbiologists (IAAM)",
        46799: "Ridwan Institute",
        46800: "Jurnal Ilmu Bedah Indonesia, Universitas Indonesia",
        46801: "Instituto Brasileiro de Direito Tributario",
        46802: "Revista Movimentos Sociais e Dinamicas Espaciais",
        46803: "Ibtida Journal of Elementary School, Sekolah"
               "Tinggi Agama Islam Nahdlatul Ulama Kotabumi Lampung",
        46804: "Maine Medical Center",
        46805: "Dogma, Philosophy and Human Sciences Review",
        46806: "Institut Bisnis dan Informatika Kwik Kian Gie",
        46807: "Pusat Penelitian Badan Keahlian Sekretariat Jenderal DPR RI",
        46808: "Politeknik Indonusa Surakarta",
        46809: "Global Talent Academy",
        46810: "Turk Doga ve Fen Dergisi, Bingol University",
        46811: "Academy of Health Professionals",
        46812: "Revista Agora",
        46813: "Problems of Atomic Science and Technology",
        46814: "Latin American Journal of Development",
        46815: "Akademi Keperawatan Ngesti Waluyo Temanggung",
        46816: "Asosiasi Pengajar Hukum Adat Indonesia",
        46817: "Sekolah Tinggi Teologi Kalimantan",
        46818: "Centro de Estudos Juridicos da Procuradoria"
               "Geral do Estado do Rio de Janeiro",
        46819: "Tanc es Neveles",
        46820: "Journal of Science Engineering Technology and Management",
        46821: "Fakultas Ekonomi dan Bisnis Universitas Bhayangkara Surabaya",
        46822: "STKIP Muhammadiyah Sampit",
        46823: "Institute for Historical Studies at Chung-Ang University",
        46824: "GAW Bukit Kototabang - Badan Meteorologi Klimatologi"
               "dan Geofisika",
        46825: "Holy Synod of Bishops of the Serbian Orthodox Church",
        46826: "University of Agriculture and Forestry, Hue University",
        46827: "Open Access Publishing Group",
        46828: "Ocean Publication",
        46829: "University of Ghana",
        46830: "World Resources Institute",
        46831: "The Association of Pediatric Surgeons of Pakistan",
        46832: "Journal of Mines Metals and Fuels",
        46833: "Revista Paulista de Reumatologia",
        46834: "Association of Korean Modern and Contemporary Art History",
        46835: "Seoul Institute of Economic Research",
        46836: "Arsip Nasional Republik Indonesia",
        46837: "Politeknik Internasional Bali",
        46838: "MAN Insan Cendekia Jambi",
        46839: "Sekolah Tinggi Ilmu Hukum Sumpah Pemuda Palembang",
        46840: "Centro de diseno, cine y television",
        46841: "Institutul de Cercetare a Calitatii Vieti",
        46842: "Cientifica, Revista Mexicana de Ingenieria"
               "Electromecanica y de Sistemas, IPN",
        46843: "Yayasan Keluarga Guru Mandiri",
        46844: "BLITHE Research Initiative",
        46845: "Kaliningrad State Technical University",
        46846: "Politeknik Meta Industri Cikarang",
        46847: "Ukrainian Military Medical Academy",
        46848: "Revista Teorico-Pratico",
        46849: "Journal of Economics and Administrative Sciences,"
               "Giresun University",
        46850: "Society for Institutional Analysis",
        46851: "INDEF - Institute for Development of Economics and Finance",
        46852: "Agricultural Economics and Social Science"
               "Research Association (AESSRA)",
        46853: "The European Association of Counselling Psychology",
        46854: "Journal of Philosophy",
        46855: "Applied Energy Innovation Institute (AEii)",
        46856: "Pan American League of Associations of Rheumatology (PANLAR)",
        46857: "University of Toamasina",
        46858: "Dr. Vithalrao Vikhe Patil Foundation, Ahmednagar",
        46859: "Faculdades Catolicas - Departamento de Teologia",
        46860: "Chandigarh Group of Colleges, Mohali",
        46861: "Bulletin Mineralogie Petrologie",
        46862: "Fakultas Kedokteran Gigi Universitas Mahasaraswati Denpasar",
        46863: "Ecocene: Cappadocia Journal of Environmental"
               "Humanities, Cappadocia University",
        46864: "Joint Institute of Mechanical Engineering"
               "of the National Academy of Sciences of Belarus",
        46865: "ISOASPSH of N.D. Kondratieff",
        46866: "Journal of Social Sciences, Humanities and"
               "Research in Education (JOSSHE)",
        46867: "International Society of Comparative Psychology",
        46868: "Akademik Tarih ve Dusunce Dergisi",
        46869: "The State Institution - Institute of World"
               "History of the National Academy of Sciences of Ukraine",
        46870: "STAIN Majene",
        46871: "Gaziantep Islam Science and Technology University",
        46872: "IEDSR Association",
        46873: "Głowny Instytut Gornictwa (Central Mining Institute)",
        46874: "Komisi Pemilihan Umum Republik Indonesia",
        46875: "Faculdade do Centro Oeste Paulista",
        46876: "Journal of Agriculture",
        46877: "Mecila Working Papers Series",
        46878: "Praxia - Revista on line de Educacao Fisica da UEG",
        46879: "Institute for Medical Radiology and Oncology"
               "of NAMS of Ukraine",
        46880: "METHOMIKA:Jurnala Manajemen Informatika dan"
               "Komputerisasi Akuntansi, Universitas Methodist Indonesia",
        46881: "Olabisi Onabanjo University",
        46882: "International Scholars Journals",
        46883: "MJH Life Sciences",
        46884: "Aeres Hogeschool",
        46885: "Nucleo de Desenvolvimento dos Tecnicos de Radiologia",
        46886: "Knowledge Words Publications",
        46887: "Northwest Territories Geological Survey",
        46888: "Badan Perencanaan Pembangunan Penelitian Pembangunan"
               "Daerah Provinsi Nusa Tenggara Timur",
        46889: "Athenaeum Scientific Publishers",
        46890: "Science Letters",
        46891: "Badan Penelitian dan Pengembangan Daerah Kota Baubau",
        46892: "Achieve Behavioral Health",
        46893: "UYEP Merkezi",
        46894: "The Korean Association for Immigration Policy"
               "and Administration",
        46895: "Mita Society for Library and Information Science",
        46896: "Journal of Islamic Civilization and Culture",
        46897: "Uluslararasi hayvancilik arastirma ve egitim"
               "merkezi mudurlugu Livestock Studies dergisi",
        46898: "Rfb Editora",
        46899: "Jurnal Ekonomi Dan Perbankan Syariah, STEI SEBI",
        46900: "Archives of Pediatric Neurosurgery",
        46901: "Defensoria Publica da Uniao",
        46902: "Folha de Rosto: Revista de Biblioteconomia"
               "e Ciencia da Informacao",
        46903: "Gomal Journal of Medical Sciences",
        46904: "Editura Electra",
        46905: "Aligarh Muslim University",
        46906: "CAOS Revista Eletronica de Ciencias Sociais",
        46907: "Social Integration Research Center, Kangwon"
               "National University",
        46908: "Universidad Nacional Autonoma de Tayacaja"
               "Daniel Hernandez Morillo",
        46909: "University of Agricultural Sciences and Veterinary"
               "Medicine in Iasi",
        46910: "American Bilingual University",
        46911: "Victorian Popular Fiction Association",
        46912: "Cprint Publishers (CPP)",
        46913: "NNC Institute of Beekeeping named after P.I. Prokopovich",
        46914: "Turan University",
        46915: "Zero Prod. Ltd. Sti.",
        46916: "ICSP - New Science",
        46917: "VERN University",
        46918: "Universitas Muslim Maros",
        46919: "Archives of Health",
        46920: "Russian Energy Agency",
        46921: "Instituto Federal de Educacao, Ciencia e Tecnologia"
               "do Triangulo Mineiro",
        46922: "Personnel Training Institute of the State"
               "Employment Service of Ukraine",
        46923: "Universitas Teknologi Yogyakarta",
        46924: "Jurnal Ilmu Hukum Kyadiren, Sekolah Tinggi"
               "Ilmu Hukum Biak, Papua",
        46925: "Revista de la Universidad del Zulia, Universidad del Zulia",
        46926: "MusMat",
        46927: "Gheorghe-Gavrila Hognogi",
        46928: "Sosyal Bilimler Dergisi, Istanbul Ticaret Universitesi",
        46929: "Sekolah Tinggi Teologi Baptis Jakarta",
        46930: "Universitas Darma Agung",
        46931: "Archivum Anatolicum, Ankara University",
        46932: "South Florida Journal of Development",
        46933: "Diegesis: Jurnal Teologi, STT Bethel Indonesia",
        46934: "Ikatan Prostodonsia Indonesia",
        46935: "University of Georgia Press",
        46936: "US Department of Energy",
        46937: "Jurnal Komunikasi Pembangunan",
        46938: "Theory of Science",
        46939: "Valahia University of Targoviste - Journal"
               "of Science and Arts",
        46940: "SunKrist Publishing",
        46941: "European Integration Studies",
        46942: "Studia Iurisprudentiae Doctorandorum Miskolciensium",
        46943: "Realize Eventos Cientificos e Editora LTDA",
        46944: "ASP Editions",
        46945: "Bangladesh Physiotherapy Association (BPA)",
        46946: "Centro Espanol de Investigaciones Coreanas (CEIC)",
        46947: "Jamal Mohamed College",
        46948: "Global Science Research Journals",
        46949: "Heyttu",
        46950: "The Representative Office of the Institute"
               "for War and Peace Reporting in Kyrgyzstan",
        46951: "Pennsylvania Association for Colleges and Teacher Educators",
        46952: "Revista Brasileira de Ensino de Jornalismo (Rebej)",
        46953: "Yonetim Hizmetleri Genel Mudurlugu",
        46954: "Facultad de Humanidades de la Universidad"
               "de San Carlos de Guatemala",
        46955: "Ataturk Yolu Dergisi, Ankara Universitesi"
               "Turk Inkilap Tarihi Enstitusu",
        46956: "International Journal of Indigenous Herbs and Drugs",
        46957: "Contradictions: A Journal for Critical Thought",
        46958: "Revista Clinica Veterinaria",
        46959: "Journal of Empirical Economics and Social"
               "Sciences, Bandirma Onyedi Eylul University",
        46960: "Nizhny Novgorod State Technical University"
               "n.a. R.E. Alekseev",
        46961: "Politeknik Negeri Media Kreatif",
        46962: "Journal FORTEI-JEERI, Forum Pendidikan Tinggi"
               "Teknik Elektro Indonesia",
        46963: "STAI Auliaurrasyidin Tembilahan",
        46964: "Center for Research and Community Service",
        46965: "Jurnal Christian Humaniora, Institut Agama"
               "Kristen Negeri Tarutung",
        46966: "Research and Education Development (READ) Institute",
        46967: "The Korean Society for the Economics and Finance"
               "of Education",
        46968: "Al-Darah",
        46969: "Turkish Journal of Womens Health and Neonatology,"
               "University of Health Sciences",
        46970: "International Forum of Management Scholars",
        46971: "Ankara Saglik Bilimleri Dergisi",
        46972: "Korolov Zhytomyr Military Institute",
        46973: "Vestnik RGUPS, Rostov State Transport University",
        46974: "Sekolah Tinggi Teologi Sriwijaya",
        46975: "Sekolah Tinggi Manajemen Immi",
        46976: "Pusat Penelitian dan Pengembangan Polri",
        46977: "Lincoln Research and Publications Limited",
        46978: "Larix International",
        46979: "Revista Brasileira de Neurologia",
        46980: "Nordmedizdat",
        46981: "South Florida Journal of Health",
        46982: "Global Journal of Medical Therapeutics",
        46983: "Korean Association of Probation and Parole Services",
        46984: "STMIK Widya Cipta Dharma",
        46985: "Kathmandu School of Law Review, Kathmandu"
               "School of Law, Nepal",
        46986: "ASA Publications",
        46987: "Association Research Center Peresvet",
        46988: "Gulf University",
        46989: "AquacultureHub",
        46990: "Invepy y Asociados S.C.",
        46991: "Yerevan State University",
        46992: "Philippine National Philosophical Research Society",
        46993: "Mohammed Bin Rashid School of Government",
        46994: "Instituto Aragones de Ciencias de la Salud (IACS)",
        46995: "GuiaSalud",
        46996: "International Institute of Administrative Sciences",
        46997: "Sociedad Medica Ecuatoriana de Neurologia",
        46998: "International Journal of Clinical Studies"
               "and Medical Case Reports",
        46999: "Duolingo, Inc.",
        47000: "Turkish Journal of Mathematics and Computer"
               "Science, Association of Mathematicians",
        47001: "International Research Journal of Innovations"
               "in Engineering and Technology",
        47002: "Universitas Mulia",
        47003: "Asociacion Uruguaya Historia Economica",
        47004: "cemmap",
        47005: "Associacao Brasileira de Medicina Legal e Pericias Medicas",
        47006: "Jurnal Edu Riligia, Pascasarjana UIN Sumatera Utara Medan",
        47007: "Universitas Esa Unggul",
        47008: "College of Medicine and Health Science, DireDawa University",
        47009: "Society for Healthcare & Research Development",
        47010: "The Russian Journal of Cognitive Science",
        47011: "Jordan Journal of Physics",
        47012: "Jordan Journal of Modern Languages & Literature",
        47013: "Jordan Journal of Mathematics and Statistics",
        47014: "Jordan Journal of Chemistry",
        47015: "Jordan Journal of Educational Sciences",
        47016: "Jordan Journal of the Arts",
        47017: "Abhath Al-Yarmouk",
        47018: "Asia Counselling and Coaching Society",
        47019: "Centre on African Philanthropy and Social Investment (WBS)",
        47020: "Locality and Communication",
        47021: "Papanin Institute for Biology of Inland Waters"
               "Russian Academy of Sciences",
        47022: "Backup Books Editora",
        47023: "Tree of Evolution",
        47024: "Sekolah Tinggi Manajemen Informatika dan Komputer GICI",
        47025: "Seminari Tinggi St. Fransiskus Xaverius Ambon",
        47026: "I.N. Ulianov Chuvash State University",
        47027: "Dicle University Journal of Faculty of Veterinary Medicine",
        47028: "Kelompok Peneliti Muda Universitas Negeri Jakarta",
        47029: "Agrociencia",
        47030: "STISIP Muhammadiyah Sinjai",
        47031: "Alma Jordan Library",
        47032: "Mona Library",
        47033: "Informatic and Data Science Journal STMIK"
               "Muhammadiyah Banten",
        47034: "Faculty of Public Health, University Indonesia",
        47035: "Facultad de Teologia de Granada",
        47036: "Scholar Chatter LLC",
        47037: "Applied Computational Electromagnetics Society",
        47038: "Truman Pierce Institute",
        47039: "Badan Penelitian dan Pengembangan Daerah Provinsi"
               "Papua Barat",
        47040: "University of Bucharest, Doctoral School of"
               "Faculty of Psychology and Educational Science",
        47041: "Orpheus Instituut",
        47042: "Critical Social Welfare Academy",
        47043: "Asosiasi Filsafat Teologi Indonesia dan Asosiasi"
               "Ahli Filsafat Keilahian Indonesia",
        47044: "Revista Epicas",
        47045: "Corvus Publisher",
        47046: "Acta Comenia",
        47047: "Cahaya Surya",
        47048: "Korea Association of World History and Culture",
        47049: "Marine Engineering Bureau",
        47050: "Fundacja Rozwoju Systemu Edukacji",
        47051: "Tropical Herping",
        47052: "Publicar Editorial Medica",
        47053: "Joint Data Center on Forced Displacement",
        47054: "Faculty of Philosophy, Skopje",
        47055: "Ural State University of Architecture and Art",
        47056: "Publishing House RAMS",
        47057: "The Society of Korean Literary Creative Writing",
        47058: "Journal of the Academy, Asociacion de Universidades del Peru",
        47059: "Centivens Institute of Innovative Research",
        47060: "Austrian Association for American Studies",
        47061: "Presencing Institute, Inc.",
        47062: "Save the Environment",
        47063: "Ss. Cyril and Methodius University in Skopje,"
               "Faculty of Economics-Skopje",
        47064: "Science Wiki, Inc.",
        47065: "Forum Kerjasama Pendidikan Tinggi (FKPT)",
        47066: "Associacao Brasileira de Fisioterapia Cardiorrespiratoria"
               "e Fisioterapia em Terapia Intensiva",
        47067: "South Punjab Center for Research and Development (SPCRD)",
        47068: "Asociatia Romana de Pteridologie",
        47069: "Revista de Estudios Ambientales",
        47070: "Mahadev Publications",
        47071: "Shri Prasanna Vitthala Education and Charitable Trust (Reg)",
        47072: "Demiryolu Muhendisligi Dergisi, Demiryolu"
               "Muhendisleri Dernegi",
        47073: "OpenCiencia Preprints",
        47074: "Historical Studies on Central Europe",
        47075: "Central European Cultures",
        47076: "Perhimpunan Intelektual Muslim Indonesia",
        47077: "STAI An-Najah Indonesia Mandiri Sidoarjo",
        47078: "Central European Journal of Comparative Law",
        47079: "Madl Ferenc Osszehasonlito Jogi Intezet",
        47080: "Universitas Banten Jaya",
        47081: "Neuroscience Society of Nigeria",
        47082: "European University Foundation",
        47083: "Cadernos de Astronomia",
        47084: "Middle East Studies Center (MESC)",
        47085: "Korean Food Marketing Association",
        47086: "Proceedings of International Mathematical Sciences",
        47087: "Maltepe Journal of Mathematics",
        47088: "Uluslararasi Tarih Arastirmalari Dergisi",
        47089: "idil Ural Arastirmalari Dergisi",
        47090: "Ibnu Jaya Konsultan",
        47091: "Turkish Studies-Comparative Religious Studies Journal",
        47092: "Cronica Tributaria",
        47093: "Sechenov University",
        47094: "Editora Omnis Scientia",
        47095: "Revista Odontologica do Hospital de Aeronautica de Canoas",
        47096: "Revista de Direito Publico Contemporaneo e"
               "Revista de Direito da Adm. Publica",
        47097: "Pamukkale University Journal of Business Research",
        47098: "Bayburt Universitesi Ilahiyat Fakultesi",
        47099: "Conference Proceedings of Science and Technology",
        47100: "Research Support Center Publications",
        47101: "Sociedad Espanola de Historia de las Ciencias y las Tecnicas",
        47102: "Academy of Medicine, Singapore",
        47103: "Bilturk",
        47104: "EE Publisher",
        47105: "Nitel Sosyal Bilimler",
        47106: "Root Work Journal",
        47107: "INIF E-DERGI, Inonu University",
        47108: "Journal of Ideas in Health",
        47109: "Sciami",
        47110: "Igdir University",
        47111: "Jurusan Teknik Informatika, Universitas Palangka Raya",
        47112: "Journal of Science and Engineering, Necmettin"
               "Erbakan University",
        47113: "NHS Future-Focused Finance",
        47114: "Indian Society of Soil Survey and Land Use Planning",
        47115: "Ondokuz Mayis University",
        47116: "Future Convergence Technology Research Society",
        47117: "The Centre of Expertise on Child Sexual Abuse",
        47118: "Manisa Celal Bayar University Soma Vocational"
               "School Technical Sciences Journal",
        47119: "International Journal of Research Publications",
        47120: "National Physical Laboratory",
        47121: "Acarological Studies, DergiPark",
        47122: "Vietnam Association of Diabetes and Endocrinology",
        47123: "Bibliotheque de l'Universite Laval",
        47124: "ViraVerita E-Journal: Interdisciplinary Encounter",
        47125: "School of Environmental Science and Management,"
               "University of the Philippines Los Banos",
        47126: "North South University",
        47127: "Koforidua Technical University",
        47128: "International Journal of Research and Development"
               "in Pharmacy and Life Sciences (IJRDPL)",
        47129: "Bartin University Journal of Faculty of Economics"
               "and Administration Sciences, Bartin University",
        47130: "Bitlis Eren University",
        47131: "Sekolah Tinggi Agama Kristen Teruna Bhakti",
        47132: "Christian Reading, Saint Petersburg Theological Academy",
        47133: "Kuaarara",
        47134: "Indonesian Journal Publisher",
        47135: "Scriptura Indonesia",
        47136: "ASBU Law Journal",
        47137: "Usak University Journal of Engineering Sciences",
        47138: "Sirnak Universitesi Iktisadi ve Idari Yaklasimlar Dergisi",
        47139: "Balkanlarda Turk Dili ve Edebiyati Arastirmalari Dergisi",
        47140: "Kirklareli University Journal of Social Sciences,"
               "Kirklareli University",
        47141: "Geriatik Bilimler Dergisi (Journal of Geriatric"
               "Science), Geriatrik Bilimler Dernegi (Associaton ofg",
        47142: "General Education and Citizen",
        47143: "Institute of Marine Biology of the NAS of Ukraine",
        47144: "Pakistan Cardiac Society",
        47145: "Din ve Bilim-Mus Alparslan Universitesi Islami"
               "Ilimler Fakultesi Dergisi, Mus Alparslan Universitesi",
        47146: "Revista Brasileira de Musica",
        47147: "Kahramanmaras Sutcu Imam Universitesi",
        47148: "Federal State Budgetary Institution - All-Russian"
               "Research Geological Oil Institute",
        47149: "Revista Praticas Educativas, Memorias e Oralidades",
        47150: "Logical Creations (Private) Limited",
        47151: "Sekolah Tinggi Ilmu Ekonomi Wira Bhakti Makassar",
        47152: "Institute of Criminological and Sociological Research",
        47153: "Yayasan Profesional Muda Cendekia",
        47154: "SRIPTA : Jurnal Teologi dan Pelayanan Kontekstual,"
               "STT Ebenhaezer Tanjung Enim",
        47155: "Maarif Mektepleri-International Journal of"
               "Social and Humanistic Sciences",
        47156: "Journal of Individual Differences in Education",
        47157: "Journal of Interdisciplinary Education: Theory and Practice",
        47158: "The University of the West Indies (UWI Press)"
               "Cavehill (Barbados)",
        47159: "The University of the West Indies (UWI Press)"
               "Open Campus (Virtual - Online)",
        47160: "The University of the West Indies (UWI Press)"
               "Five Island (Antigua & Barbuda)",
        47161: "The University of the West Indies (UWI Press)"
               "St Augustine (Trinidad)",
        47162: "Societatea Romana de Morfologie",
        47163: "Agrociencia (Mexico)",
        47164: "Perpetual Innovation Media Pvt. Ltd.",
        47165: "Yayasan Pendidikan Intan Cendekia",
        47166: "Sekolah Tinggi Teologia Paulus Medan",
        47167: "Sekolah Tinggi Alkitab Jember",
        47168: "Revista Brasileira de Energia",
        47169: "Akademik Siyer Dergisi, Kahramanmaras Sutcu"
               "Imam Universitesi",
        47170: "Occam Academic Press Ltd",
        47171: "AVI Publishers",
        47172: "South Florida Publishing",
        47173: "Journal of Al-Mayadine Al-Iktissadia",
        47174: "Latin American Publicacoes",
        47175: "Randwick International Research and Analysis Institute",
        47176: "Academic World Research",
        47177: "Institute for Global Business Research",
        47178: "Universitas Kristen Indonesia Toraja",
        47179: "Associacao Brasileira de Custos",
        47180: "Editora Mentes Abertas",
        47181: "University of Zurich",
        47182: "Editorial Verbo Divino",
        47183: "Federal Medical Biological Agency",
        47184: "Trillium GmbH Medizinischer Fachverlag",
        47185: "Institucion Universitaria Digital de Antioquia",
        47186: "Universidad Privada de Huancayo Franklin Roosevelt",
        47187: "Escuela Superior Politecnica de Chimborazo",
        47188: "Galvanotekh",
        47189: "Universidad Tecnologica Empresarial de Guayaquil",
        47190: "Universidad Nacional de Juliaca",
        47191: "Everant Journals",
        47192: "Instituto de Innovacion y Formacion Cientifica",
        47193: "Instituto Nacional de Investigacion y Desarrollo"
               "Pesquero (INIDEP)",
        47194: "Indonesian Operations Research Association",
        47195: "Centro de Documentacion e Investigacion de"
               "la Culturas de Izquierdas (CeDInCI)",
        47196: "Editorial Biotecnologica S.R.L.",
        47197: "Federacion Espanola de Asociaciones de Docentes"
               "de Educacion Fisica (FEADEF)",
        47198: "Kementerian Ketenagakerjaan Republik Indonesia",
        47199: "Jurnal STIP-AP, STIPER Agrobisnis Perkebunan, Medan",
        47200: "Universitas Cokroaminoto Yogyakarta",
        47201: "STIE Kertanegara Malang",
        47202: "Media Akuntansi, STIE St. Pignatelli Surakarta",
        47203: "Indian Association of Preventive and Social Medicine",
        47204: "Knowledge Empowerment Foundation",
        47205: "Pakistan Social Sciences Research Institute (PSSRI)",
        47206: "International Universities Strength and Conditioning"
               "Association",
        47207: "Revista Baiana de Educacao Matematica",
        47208: "Sociedade em Debate",
        47209: "Fundacao Universidade Federal de Rondonia",
        47210: "The Association of Otolaryngologists of India, West Bengal",
        47211: "Ana Publishing Private Limited",
        47212: "Alianza de Investigadores Internacionales SAS",
        47213: "Kayon Media",
        47214: "Research and Experience Journal",
        47215: "Asian Journal of Instruction",
        47216: "The Literacy Trek",
        47217: "UNITED LIGHTENING VISION",
        47218: "Jurnal Kesehatan Panca Bhakti Lampung (JKPBL),"
               "STIKes Panca Bhakti Lampung",
        47219: "Sekolah Tinggi Teknologi Industri dan Farmasi Bogor",
        47220: "Korean Management Practice Association",
        47221: "Tangible Journal at STIE Tri Dharma Nusantara",
        47222: "Revista Interinstitucional Brasileira de Terapia Ocupacional",
        47223: "International Research Journal of Ayurveda & Yoga",
        47224: "Instituto de Administracao e Gestao Educacional Ltda",
        47225: "Magyar Gerontologia",
        47226: "Tennessee Academy of Science",
        47227: "Department of Clinical Pharmacy and Pharmacy"
               "Practice, University of Benin",
        47228: "Casemate Publishers and Book Distributors",
        47229: "Historic England",
        47230: "Universidad Estatal del Sur de Manabi (UNESUM)",
        47231: "Covenant University, Ota, Nigeria",
        47232: "Sekolah Tinggi Ilmu Ekonomi Trisna Negara",
        47233: "Universitas Dharma Andalas",
        47234: "Sociedad Matematica Mexicana A.C.",
        47235: "Revista de Morfologia Urbana",
        47236: "Instituto Federal de Educacao Ciencia e Tecnologia"
               "do Tocantins",
        47237: "Editora Senso",
        47238: "The Scientific Club of Al Manahil Association",
        47239: "Direktorat GTK Dikmen Diksus, Kemendikbud",
        47240: "Revista da Escola Superior de Guerra",
        47241: "LAS Journal",
        47242: "Editora In Vivo",
        47243: "Journal of Original Studies",
        47244: "Escuela Superior Politecnica de Chimborazo",
        47245: "UMR7044 ARCHIMEDE",
        47246: "Societatea de Geografie din Romania - Filiala Prahova",
        47247: "V & V Editora",
        47248: "Pivot Science Publication Corp.",
        47249: "Revista Brasileira de Alfabetizacao",
        47250: "Interdisciplinar - Revista de Estudos em Lingua e Literatura",
        47251: "Sociedad Espanola de Documentacion e Informacion Científica",
        47252: "International Journal of Creative Future and"
               "Heritage (TENIAT)",
        47253: "Journal of Tropical Resources and Sustainable"
               "Science (JTRSS)",
        47254: "International Online Journal of Language,"
               "Communication, and Humanities (INSANIAH)",
        47255: "Jurnal Ilmiah Spectral, Sekolah Tinggi Bahasa"
               "Asing Pontianak",
        47256: "Sekolah Tinggi Pariwisata Ambarrukmo (STIPRAM) Yogyakarta",
        47257: "Bilge Uluslararasi Sosyal Arastirmalar Dergisi",
        47258: "Kolpa Editores",
        47259: "Global Scientific Publications",
        47260: "Scientific Press International Limited",
        47261: "GNT-Verlag GmbH",
        47262: "The Science Publishers",
        47263: "Sarawak Research Society",
        47264: "LASSIJ IDEA",
        47265: "CJIM Publishing Group",
        47266: "Ministry of National Development Planning/Bappenas RI",
        47267: "National Library Republic of Uzbekistan",
        47268: "Fakultas Hukum Universitas Pattimura",
        47269: "Kantor Bahasa Kepulauan Riau",
        47270: "Revista Alembra",
        47271: "Jogi- es tarsadalomtudomanyi kozlemenyek a"
               "kozossegek szolgalataban",
        47272: "Kozigazgatasi es Infokommunikacios Jogi PhD Tanulmanyok",
        47273: "Acta Periodica",
        47274: "Revista de Derecho, Universidad de Montevideo",
        47275: "Scholit Publishing LLC",
        47276: "Lepra",
        47277: "Dorma Trading Est",
        47278: "Unique Scientific Publishers",
        47279: "Zakarpatian State Agricultural Experimental"
               "Station of NAAS of Ukraine",
        47280: "Revista de la Facultad de Agronomia, Universidad del Zulia",
        47281: "Fastabiq : Jurnal Studi Islam Fakultas Ilmu"
               "Keislaman Muhamadiyah Bandung",
        47282: "Economica",
        47283: "Revista Tecnologica da Fatec Americana",
        47284: "Cadernos de Campo: Revista de Ciencias Sociais",
        47285: "Scientific Research Initiative",
        47286: "Universidad Catolica de Oriente",
        47287: "American Chemical Society (C&EN)",
        47288: "Tarek Waly Center Architecture & Heritage",
        47289: "The AI Robotics Ethics Society",
        47290: "FLE Learning Ltd",
        47291: "Institute for Economic and Social Research (LPEM FEB UI)",
        47292: "STMIK Amikbandung",
        47293: "Institute for Humanities Jeju National University",
        47294: "KSBDA (Korea Society of Basic Design & Art)",
        47295: "Universidade Regional do Cariri",
        47296: "Centro Universitario Sagrado Coracao - Unisagrado",
        47297: "World Scientific Publishing House Ltd",
        47298: "The CALA Unit, SOAS University of London",
        47299: "Black Sea Journal of Management and Marketing",
        47300: "Universidad Internacional de Ciencia y Tecnologia",
        47301: "Find Publishing and Research",
        47302: "Institute of Statistical Research and Training"
               "(ISRT), University of Dhaka",
        47303: "Akademi Kebidanan Palu",
        47304: "Sekolah Tinggi Teologi Pelita Dunia",
        47305: "Journal of Liberty and International Affairs,"
               "Institute for Research and European Studies - Bitola",
        47306: "Associacao Caruaruense de Ensino Superior"
               "e Tecnico - ASCES-UNITA",
        47307: "Gaceta Medica de Caracas",
        47308: "Revista de Derecho Procesal Laboral",
        47309: "Institute of Strategic Research of the Republic"
               "of Bashkortostan",
        47310: "International Academic and Research Consortium",
        47311: "International Journal Of English and Studies",
        47312: "Asosiasi Fakultas Ekonomi & Bisnis Indonesia",
        47313: "Universitas Nasional",
        47314: "Universidad de Concepcion - CABID",
        47315: "State Archival Service of Ukraine",
        47316: "South Kazakhstan Medical Academy",
        47317: "STIKES Wira Husada Yogyakarta",
        47318: "Ludika Centro de acompanamiento y desarrollo infantil",
        47319: "Revista de Direitos Fundamentais e Tributacao",
        47320: "Revista da Associacao Brasileira de Nutricao - RASBRAN",
        47321: "Revista Pesquisas e Praticas Educativas",
        47322: "Uniqbu Journal of Exact Sciences, Universitas Iqra Buru",
        47323: "Uniqbu Journal of Social Sciences, Universitas Iqra Buru",
        47324: "Akademi Ilmu Komputer Ternate",
        47325: "Ogranak Matice hrvatske u Daruvaru",
        47326: "Ontario COVID-19 Science Advisory Table (OCSAT)",
        47327: "UNIKA Journal of Health Sciences",
        47328: "Pro-Reitoria de Pesquisa e Pos-Graduacai - UFV",
        47329: "Faculty of Economic and Business, Ngurah Rai University",
        47330: "Design Computation",
        47331: "Association of Oncologists of Uzbekistan",
        47332: "Publishing house URSS",
        47333: "Modernism and Postmodernism Studies Network",
        47334: "The Korean Academic Group of Comparative Study on East Asia",
        47335: "Jurnal EMA (Ekonomi Manajemen Akuntansi),"
               "Universitas Merdeka Pasuruan",
        47336: "LLC NIVA",
        47337: "Magallat al-ibda al-riyadi",
        47338: "Journal of Neonatal Surgery",
        47339: "British Columbia Institute of Technology (BCIT)",
        47340: "Millennium Journals",
        47341: "Business & Economics Society International (B&ESI)",
        47342: "Stiftung Munchner Sicherheitskonferenz gGmbH",
        47343: "Department of Graduate Medical Education, HCA Healthcare",
        47344: "Suleyman Demirel University, Kaskelen, Kazakhstan",
        47345: "Instituto de Ciencias Juridico-Politicas",
        47346: "London Tech Press",
        47347: "Instituto Geologico Minero y Metalurgico",
        47348: "Juta and Company (Pty) Ltd",
        47349: "Perhimpunan Biologi Indonesia",
        47350: "Matrosov Institute for System Dynamics and"
               "Control Theory of Siberian Branch of RAS",
        47351: "Centre for Problems of Control and Security",
        47352: "Pandawa Institute",
        47353: "Universitas 45 Mataram",
        47354: "Universitas Fajar",
        47355: "Java Scientific Academy",
        47356: "Turkish Studies - Social Sciences",
        47357: "UFAM Business Review - UFAMBR",
        47358: "ETU Sentez Iktisadi ve Idari Bilimler Dergisi,"
               "Erzurum Teknik Universitesi",
        47359: "Aegitas Publishing House",
        47360: "Mediar Press",
        47361: "The Centre for Education and Methodics of the FAS Russia",
        47362: "Electronic Journal of Social and Strategic Studies",
        47363: "Scientific Research and Community Ltd",
        47364: "WWF Russia",
        47365: "Litinfinite Journal",
        47366: "Universidad del Pacifico - Colombia",
        47367: "Ivanovo State Polytechnic University",
        47368: "University of Zurich, IKMZ – Department of"
               "Communication and Media Research",
        47369: "Revista Eletronica de Estudos Integrados em"
               "Discurso e Argumentacao",
        47370: "Maikop State Technological University",
        47371: "The Mycological Society of Japan",
        47372: "University of Aden",
        47373: "Revista de la Facultad de Ciencias de la Salud"
               "Universidad del Cauca",
        47374: "Novedades colombianas",
        47375: "Justicia y derecho",
        47376: "Filosofia Publishing House",
        47377: "Editora Nomada",
        47378: "ASBM University",
        47379: "Wiener Wissenschafts-, Forschungs- und Technologiefonds"
               "(WWTF)",
        47380: "Universidad Pedagogica Nacional, Mexico",
        47381: "The Society for the Provision of Education"
               "in Rural Australia (SPERA)",
        47382: "Archeologicky ustav AV CR, Brno, v. v. i.",
        47383: "Asociatia Romana pentru Promovarea Artelor Spectacolului",
        47384: "Sekolah Tinggi Perikanan dan Kelautan Palu",
        47385: "Fundacao Oswaldo Aranha - FOA",
        47386: "Revista Cientifica de Estudios Urbano Regionales"
               "Hatso-Hnini, UNINSUS",
        47387: "Ahlimedia Press",
        47388: "Nizhny Novgorod Linguistics University",
        47389: "Australian Institute for Disaster Resilience",
        47390: "LLC SCIENCEPROBLEMS TEAM",
        47391: "Pakistan Medical Association",
        47392: "RSP Science Hub",
        47393: "The University of Danang - University of Science"
               "and Education, Vietnam",
        47394: "Publishing House Polyterra",
        47395: "SeekingPro",
        47396: "DECENT Database",
        47397: "TeX Users Group",
        47398: "ILTEK : Jurnal Teknologi, Universitas Islam Makassar",
        47399: "Asosiasi Psikologi Islam",
        47400: "IAI Al Khoziny Buduran Sidoarjo",
        47401: "Revista Sergipana de Educacao Ambiental",
        47402: "Editora e-Publicar",
        47403: "Institute of Research Publishing Project",
        47404: "Bulletin of the KSMTU, FSBSI AzNIIRKH",
        47405: "Secholian Publication",
        47406: "Optimization in Engineering Research",
        47407: "MediaMedica - MMA Limited Liability Company",
        47408: "Association for Learning Development in Higher Education",
        47409: "The Institute of Psychotherapy, Psychological"
               "Counseling and Clinical Supervision (IPCS)",
        47410: "BabuHawaii Foundation",
        47411: "Badan Wakaf Indonesia",
        47412: "Faculty of Engineering, The University of The West Indies",
        47413: "Vidya - A Journal, Gujarat University",
        47414: "Institute of Bioenergy Crops and Sugar Beet"
               "National Academy of Agrarian Sciences of Ukraine",
        47415: "Global Research Institute for Arts and Culture Education",
        47416: "Asian Pacific Journal of Dentistry",
        47417: "The Yeongnam Archaeological Society",
        47418: "UAQUIRI - Revista do Programa de Pos-Graduacao em Geografia",
        47419: "Department of Chemistry and Biochemistry,"
               "Al-Nahrain University",
        47420: "Deutscher Arzteverlag GmbH",
        47421: "Librairie Droz S.A.",
        47422: "Professionals On Line SAC",
        47423: "Turkish Studies - Educational Sciences",
        47424: "Tekirdag Namik Kemal Universitesi Ilahiyat"
               "Fakultesi Tasavvur / Tekirdag Ilahiyat Dergisi",
        47425: "Siirt Universitesi Ilahiyat Fakultesi Dergisi",
        47426: "Facultad de Ciencias Economicas y Administracion",
        47427: "Korean Association of Military Counseling",
        47428: "Facultad de Ciencias Sociales, UDELAR",
        47429: "Korean Society for Laboratory Medicine",
        47430: "Umaru Musa YarAdua University Katsina NG",
        47431: "Program Studi Ilmu Pemerintahan STPMD APMD",
        47432: "Sindh Agriculture University",
        47433: "Editura Doxologia, Mitropolia Moldovei si Bucovinei",
        47434: "Journal of Educational Research in Developing Areas",
        47435: "Institut Agama Islam Muhammadiyah Sinjai",
        47436: "Faculty of Fine Arts - Cairo",
        47437: "Eskisehir Osmangazi Universitesi Tarih Bolumu Dergisi",
        47438: "Izvestia Voronezh State Pedagogical University,"
               "Voronezh State Pedagogical University",
        47439: "The Korean Archaeological Society",
        47440: "Journal of Agriculture, Food and Environment",
        47441: "Badan Penelitian dan Pengembangan Provinsi"
               "Kalimantan Selatan",
        47442: "Behavior & Law Journal, Fundacion Universitaria"
               "Behavior & Law",
        47443: "Shahin Digital Publisher",
        47444: "AMOGJ - Angolan Mineral, Oil & Gas Journal, SA",
        47445: "UK Wealth Tax Commission",
        47446: "Observatoire des Abeilles",
        47447: "Turkish Journal of Sports Medicine, Turkish"
               "Sports Medicine Association",
        47448: "Karakalpak State University",
        47449: "CientMed",
        47450: "The Kostanai Academy of the Ministry of Internal"
               "Affairs of the Republic of Kazakhstan named"
               "after Shyrakbek Kabylbayev",
        47451: "European Scientific e-Journal",
        47452: "Bokkos Journal of Applied Scientific Reports",
        47453: "Institut Agama Islam Bunga Bangsa Cirebon",
        47454: "STAI Al-Anwar Sarang Rembang",
        47455: "Inovacao & Tecnologia Social",
        47456: "Universidade Federal do Espirito Santo",
        47457: "Sekolah Tinggi Teologi Injili Arastamar (SETIA) Jakarta",
        47458: "Center of Excellence of Healthcare and Pharmacy Specialists",
        47459: "General Jonas Zemaitis Military Academy of Lithuania",
        47460: "AutanaBooks",
        47461: "International Society for Occupational Ergonomics and Safety",
        47462: "Berkshire Publishing Group",
        47463: "Universidad de Otavalo",
        47464: "Hospital Metropolitano",
        47465: "PCP Publishing",
        47466: "Sekolah Tinggi Agama Islam (STAI) ALHIKMAH",
        47467: "Institut Agama Islam Nasional Laa Roiba Bogor",
        47468: "Charak Samhita Research, Training and Skill"
               "Development Centre",
        47469: "International Society for Energy and Environmental"
               "Sustainability",
        47470: "Federal Scientific Center for Hygiene F.F.Erisman",
        47471: "PIBOC DVO RAN",
        47472: "International Society of City and Regional Planners",
        47473: "Italian Association of Financial Industry"
               "Risk Managers (AIFIRM)",
        47474: "National Institute of Public Health",
        47475: "Chelyabinsk State University",
        47476: "European Association for Studies in Screen Translation",
        47477: "International Journal of Science and Education",
        47478: "Istanbul Umut Hareketi Dernegi",
        47479: "Aksaray University",
        47480: "Turk Isi Bilimi Ve Teknigi Dernegi",
        47481: "Yildiz Technical University",
        47482: "Alban Tanitim Ltd. STI",
        47483: "Revista Brasileira de Estetica Cientifica",
        47484: "Telegeco",
        47485: "Uniscience Publishers LLC",
        47486: "Australian Research Data Commons (ARDC)",
        47487: "Instituto Nacional Cardiovascular - Carlos"
               "Alberto Peschiera Carrillo",
        47488: "DHR Health Institute for Research and Development",
        47489: "Shaikh Zayed Medical Complex Lahore",
        47490: "Center of Study Radicalism and Terrorism Indonesian"
               "Police Science College",
        47491: "Land Journal, D4 Akuntansi Keuangan Politeknik Pos Indonesia",
        47492: "Sekolah Tinggi Pariwisata Mataram",
        47493: "Bolu Abant Izzet Baysal Universitesi, Tip"
               "Fakultesi, Abant Tip Dergisi",
        47494: "Academic Journal Development Service",
        47495: "Osmaniye Korkut Ata University J Natural Applied Sciences",
        47496: "NatureLibrary",
        47497: "Editorial Universidad del Cauca",
        47498: "Sekolah Tinggi Agama Islam Negeri Teungku Dirundeng Meulaboh",
        47499: "Asociacion Colombiana de Ciencias Biologicas",
        47500: "National Museum of the Republic of Kazakhstan",
        47501: "LLC Institute Information Technologies",
        47502: "Izmir Katip Celebi University, Mizanu l-Haqq:"
               "Journal of Islamic Studies",
        47503: "Journal of Innovative Research in Social Studies",
        47504: "International Journal of Agriculture, Biology"
               "& Environmental Sciences",
        47505: "International Journal of Research in Social"
               "Science & Humanities",
        47506: "STIKES Mataram",
        47507: "Indonesian Society of Obstetric Anesthesia and Critical Care",
        47508: "BRDU International Journal of Multidisciplinary Research",
        47509: "ARF (Academic Research Foundations) India",
        47510: "The East Asian Trade Association",
        47511: "SAPS Academy",
        47512: "Mersin University Journal of Maritime Faculty,"
               "Mersin University",
        47513: "International Association for Music and Medicine",
        47514: "Department of Physics, Kaduna State University",
        47515: "SAGE Publications",
        47516: "Ethnographica et Folkloristica Carpathica",
        47517: "Arts and Humanities",
        47518: "Revista Femass",
        47519: "Editora Ibero-Americana de Educacao",
        47520: "Society for Jeju Studies",
        47521: "Badan Perencanaan Pembangunan Daerah Kabupaten"
               "Indragiri Hilir",
        47522: "Jurnal Mitra Kesehatan,STIKes Mitra Keluarga",
        47523: "Canary Intelligence And Analytics Pte. Ltd.",
        47524: "Credence Press Limited",
        47525: "Uluslararasi Anadolu Sosyal Bilimler Dergisi",
        47526: "Khoja Ahmed Yasawi International Kazakh-Turkish University",
        47527: "The Northeast Asia Association of Ancient History",
        47528: "Plant Protection and Quarantine Journal",
        47529: "National Alliance of Medicine and Sports -"
               "Healthy Generation",
        47530: "EDULEAD: Journal of Christian Education and Leadership",
        47531: "Mantech Publications",
        47532: "Universitas Mahendradatta",
        47533: "National Engineering Academy of the Republic of Kazakhstan",
        47534: "International Anesthesia Research Society",
        47535: "Faculty of Economic Sciences, University of Oradea",
        47536: "International Whaling Commission",
        47537: "Circulo de Estudios de la Literatura Picarescca"
               "y Celestinesc (CELPYC)",
        47538: "Amplamente Cursos e Formacao Continuada",
        47539: "Poltekkes Kemenkes Jayapura",
        47540: "Literacy Institute",
        47541: "Jurnal Kadera Bahasa, Balai Bahasa Provinsi Sulawesi Utara",
        47542: "Sakarya Universitesi Isletme Enstitusu Dergisi,"
               "Sakarya University",
        47543: "Sekolah Tinggi Teologi Iman",
        47544: "International Organization for Health, Sports,"
               "and Kinesiology",
        47545: "Faculty of Engineering, Lagos State University, Nigeria",
        47546: "The Research Gate Corporation",
        47547: "Cytokines and Inflammation Company Limited",
        47548: "Interdisciplinary Research and International"
               "Strategy (IRIS) Institute",
        47549: "Prodi Teknik Mesin, Intitut Teknologi Nasional Malang",
        47550: "Revista Cuestiones Economicas del Banco Central del Ecuador",
        47551: "Revista Mbote",
        47552: "Ayurveda Sahiti Prabha",
        47553: "University of Murcia",
        47554: "Instituto Dominicano de Evaluacion e Investigacion"
               "de la Calidad Educativa IDEICE",
        47555: "Science Diplomacy Center",
        47556: "World Association for Sustainable Development (WASD)",
        47557: "Academia Mexicana de Investigacion Turistica, A.C.",
        47558: "SBH Health System",
        47559: "Fundacja Rehasport Clinic",
        47560: "STIKes William Booth Surabaya",
        47561: "STMIK Subang",
        47562: "Sekolah Tinggi Teologi Bethel Indonesia",
        47563: "The Military & Culture Association of Korea-Japan",
        47564: "Viipurin Suomalaisen Kirjallisuusseuran Toimitteita",
        47565: "Turk Nefroloji, Diyaliz ve Transplantasyon"
               "Hemsireleri Dernegi",
        47566: "Sociedad Mexicana de Ciencia y Tecnologia"
               "de Superficies y Materiales A.C.",
        47567: "NGO Freedom and Democracy Foundation",
        47568: "User-Centred Energy Systems TCP",
        47569: "Linkin Science",
        47570: "Institutul National de Cercetare Dezvoltare Victor Babes",
        47571: "Ataturk Universitesi Guzel Sanatlar Fakultesi: Sanat Dergisi",
        47572: "Medical Journal of Mugla Sitki Kocman University,"
               "Mugla Sitki Kocman University",
        47573: "AYA Editora",
        47574: "Lembaga Studi Agama dan Masyarakat Aceh",
        47575: "STIKES Bhakti Husada Mulia Madiun",
        47576: "Publishing house Pegas",
        47577: "Technium Science",
        47578: "National Association of Advisors for the Health Professions",
        47579: "Fundacion Espanola para la formacion e investigacion"
               "en ciencias biomedicas y patologia oncologica",
        47580: "European Star Academic Publishing Ltd",
        47581: "JSC University Book",
        47582: "Journal of Medicine and Palliative Care",
        47583: "Global Research Online",
        47584: "Foodservice Management Society of Korea",
        47585: "Editora Vecher",
        47586: "Dr. JK Research Foundation",
        47587: "Science Archives",
        47588: "Central University of Technology",
        47589: "Adalya, Koc University",
        47590: "Medical University of Silesia",
        47591: "Revista sobre Excelencia em Gestao e Qualidade",
        47592: "Mundo Economico",
        47593: "Revista ABC Imagem Cardiovascular",
        47594: "Adam Matthew Digital",
        47595: "Revista de Ciencias Juridicas e Sociais - IURJ",
        47596: "Sekolah Tinggi Teologi Aletheia",
        47597: "Seok Juseon Memorial Museum, Dankook University",
        47598: "Bashkir Institute of Social Technologies",
        47599: "Pusat Sumber Daya Mineral Batubara dan Panas Bumi",
        47600: "Jurnal Teknik Sipil Terapan, Politeknik Negeri Manado",
        47601: "Lembaga Kajian dan Pengembangan Profesi Pendidik Indonesia",
        47602: "Journal of Positive School Psychology",
        47603: "Sociedad Colombiana de Mastozoologia",
        47604: "IPR Journals and Books (International Peer"
               "Reviewed Journals and Books)",
        47605: "Research Center for Climate Change, University of Indonesia",
        47606: "The Venezuelan Civil Association for Social"
               "Research for Humanity (ACVENISPROH)",
        47607: "RESAIM Publishing",
        47608: "Ministry of Tourism and Creative Economy",
        47609: "The Academy of Research",
        47610: "American Journal of Environmental Biology",
        47611: "rScroll",
        47612: "National Academy of Sciences of Belarus",
        47613: "Istanbul Bilgi University",
        47614: "Arete Political Philosophy Journal",
        47615: "International Social Sciences Education Journal",
        47616: "Asian Multicultural Research Study",
        47617: "Vestnik MSUT STANKIN, MSUT STANKIN",
        47618: "Association/Society of Orthopaedics Surgical"
               "and Dental Sciences",
        47619: "Research Institute for Healthcare Organization"
               "and Medical Management",
        47620: "Middle East Forum",
        47621: "Multi-Science Research (MSR)",
        47622: "African Minds",
        47623: "IVAP (Herri Arduralaritzaren Euskal Erakundea)",
        47624: "KTV Kongress & Zeitschrift GmbH",
        47625: "STIT Sunan Giri Bima",
        47626: "EDITORA SCIENTIFIC",
        47627: "Gradus - Revista Brasileira de Fonologia de Laboratorio",
        47628: "Sekolah Tinggi Theologi Jaffray Jakarta",
        47629: "Bulletin of the Academy of Law and Management"
               "(Institute of Business Career)",
        47630: "Kopint Konjuktura Kutatasi Alapitvany",
        47631: "Baynoon Centre for Studies and Development",
        47632: "Ukrainian Catholic University",
        47633: "Universidad Tecnica Nacional",
        47634: "The Canadian Counselling and Psychotherapy Association",
        47635: "The Korea Learning Disabilities Association",
        47636: "Global Knowledge and Convergence Association",
        47637: "Universitas Muhammadiyah Kotabumi",
        47638: "Assosiasi Dosen Muhammadiyah Magister Administrasi"
               "Rumah Sakit",
        47639: "School Press Publisher",
        47640: "Russian Entomological Society",
        47641: "Palestine Technical College - Deir El-Balah",
        47642: "Protestant Theological University",
        47643: "Limited Liability Company - Publishing Law and State Press",
        47644: "Turkish Studies - Economics, Finance, Politics",
        47645: "Saint Petersburg Institute (Branch) of the"
               "All-Russian State University of Justice",
        47646: "Cocuk Vakfi",
        47647: "Universitas Jabal Ghafur",
        47648: "Holy Family Red Crescent Medical College",
        47649: "Kh.Dosmukhamedov Atyrau University",
        47650: "Universitas Pancasakti Makassar",
        47651: "MAARIF Institute",
        47652: "STEI Tholabul Ilmi",
        47653: "Sekolah Tinggi Farmasi Muhammadiyah Tangerang",
        47654: "Asia University",
        47655: "Badan Litbang dan Diklat Kementerian Agama RI",
        47656: "University of Puerto Rico Medical Sciences Campus",
        47657: "Department of Information Management, University"
               "of the Punjab, Lahore",
        47658: "Mondru Editora LTDA",
        47659: "Membrana Institute",
        47660: "Sociedade Brasileira de Reumatologia",
        47661: "Anais de Filosofia Classica",
        47662: "Universitas Alwashliyah Medan",
        47663: "Jurnal Manajemen Eka Prasetya, Sekolah Tinggi"
               "Ilmu Ekonomi Eka Prasetya",
        47664: "IZDATELSTVO KIR",
        47665: "Malaysian Society of Parasitology and Tropical Medicine",
        47666: "Corporacion Universitaria Autonoma de Narino",
        47667: "Information-integrated Global Society Studies",
        47668: "STKIP PGRI Situbondo",
        47669: "Eurasia Institutes",
        47670: "Westcliff University",
        47671: "Universa BV",
        47672: "AJPO JOURNALS",
        47673: "MacroMolecule Wiki",
        47674: "Bristol University Press",
        47675: "Ukrainian Scientific Railway Car Building Research Institute",
        47676: "The Korean Association for Early Childhood"
               "Education and Care",
        47677: "Revista Glauks - Departamento de Letras",
        47678: "The Canadian Journal of Higher Education/la"
               "Revue canadienne d'enseignement superieur",
        47679: "Health, Utan Kayu Publishing",
        47680: "Universidade Federal de Minas Gerais - Pro-Reitoria"
               "de Pesquisa",
        47681: "Revista Conhecimento em Acao",
        47682: "RINTERPAP - Revista Interdisciplinar de Pesquisas Aplicadas",
        47683: "KKI Elemzések",
        47684: "Korea Coaching Development Center",
        47685: "Universitas 17 Agustus 1945 Cirebon",
        47686: "STIE YPPI Rembang",
        47687: "Politeknik Pembangunan Pertanian Manokwari",
        47688: "Reserve Bank of Australia",
        47689: "inScience LLC",
        47690: "Science World Publishing",
        47691: "Journal of Illusion, Ritsumeikan University",
        47692: "Cadernos de historia da ciencia",
        47693: "Advances in Nutritional Sciences",
        47694: "Escritas do Tempo - Revista do Programa de"
               "Pos-Graduacao em Historia da Unifesspa",
        47695: "Centro Universitário UNIEURO",
        47696: "Proceedings of ADVED International Conference"
               "on Advances in Education",
        47697: "African Leadership Centre",
        47698: "Consejo para Prevenir y Eliminar la Discriminacion"
               "de la Ciudad de Mexico",
        47699: "Cathays High School",
        47700: "Journal of Islamic Economics and Finance Studies",
        47701: "Universitas Duta Bangsa Surakarta",
        47702: "Selcuklu Medeniyeti Arastirmalari Dergisi (SEMA)",
        47703: "University of International Business",
        47704: "Stencila",
        47705: "Faculty of Dentistry Khalij-Libya Dental College",
        47706: "Foreign Policy Review",
        47707: "Külügyi Szemle",
        47708: "KKI Tanulmányok",
        47709: "Information Technology and Science (ITScience)",
        47710: "STIKES Nahdlatul Ulama Tuban",
        47711: "The Institute of Economic Forecasting of the"
               "Russian Academy of Sciences",
        47712: "Facultad de Ciencias Juridicas y Politicas"
               "de la Universidad Nacional del Altiplano de Puno",
        47713: "Society of Basic and Applied Physiology",
        47714: "Uluslararasi Egitim Bilim ve Teknoloji Dergisi",
        47715: "Jupiter Publications Consortium",
        47716: "Magestic Technology Solutions Private Limited",
        47717: "Turkish Journal of Surgery",
        47718: "Poltekkes Kemenkes Manado",
        47719: "Iraqi Academics Syndicate",
        47720: "Habibia Research Academy",
        47721: "Skies Educational",
        47722: "International Center for Research & Development",
        47723: "University of Baghdad - Al-Kindy College of Medicine",
        47724: "Drustvo radioloskih inzenirjev Slovenije - DRI",
        47725: "Subdireccion de Investigacion, Servicio Nacional"
               "del Patrimonio Cultural",
        47726: "Somato Publications",
        47727: "Morarji Desai National Institute of Yoga",
        47728: "Sekolah Tinggi Ilmu Pertanian Farming Semarang",
        47729: "Ikatan Sarjana Ekonomi Indonesia Cabang Semarang",
        47730: "Lampuhyang, STKIP Agama Hindu Amlapura",
        47731: "Journal of Education, Miftahul Ulum Bangkalan College",
        47732: "STAI AL-Falah Banjarbaru",
        47733: "Global Journal of Cancer Case Reports",
        47734: "Portal de Periodicos do Instituto Federal"
               "de Educacao, Ciencia e Tecnologia de Sao Paulo",
        47735: "The Pusan Archaeological Society",
        47736: "Majelis Pendidikan Dasar dan Menengah Pimpinan"
               "Wilayah Muhammadiyah Daerah Istimewa Yogyakarta",
        47737: "Perm Agrarian Journal, FSBEI HE Perm SATU",
        47738: "Bright Publisher",
        47739: "J-SciMed Central, LLC",
        47740: "University for Development Studies Library",
        47741: "Direccion de Investigacion Criminal e Interpol - DIJIN",
        47742: "The Institute of Academic Research and Publication",
        47743: "Editura Universitatii Alexandru Ioan Cuza din Iasi",
        47744: "Barnard Press Inc",
        47745: "Universitatea Sapientia din municipiul Cluj-Napoca",
        47746: "Frontiers in Medical Case Reports",
        47747: "Training & Research Institute Jeramba Ilmu Sukses",
        47748: "Turkish Journal of Veterinary Research",
        47749: "Universidade Estadual de Campinas - Repositorio"
               "Institucional",
        47750: "Siree Journals",
        47751: "South Kazakhstan State Pedagogical University",
        47752: "Sumerianz Publication",
        47753: "Faculty of Social and Political Sciences,"
               "Universitas Sriwijaya",
        47754: "Sekolah Tinggi Teologi Amanat Agung",
        47755: "Pragma Journals LLP",
        47756: "Asociacion Mexicana de Interaccion humano-Computadora"
               "(AMexIHC)",
        47757: "Practal",
        47758: "SRP Center",
        47759: "Institut Agama Islam Uluwiyah Mojokerto",
        47760: "Zain Publications",
        47761: "River Campus Libraries, University of Rochester",
        47762: "University for Development Studies Library",
        47763: "Macedonain Heraldic Society",
        47764: "Revista Internacional Educon",
        47765: "Central research institute of geological prospecting"
               "for base and precious metals",
        47766: "Program Studi Manajemen Pendidikan Islam - IAIN Lhokseumawe",
        47767: "Politeknik Negeri Sambas",
        47768: "Sekolah Tinggi Ilmu Ekonomi Gentiaras",
        47769: "Istanbul Sabahattin Zaim University Journal"
               "of Institute of Science and Technology",
        47770: "Faculty of Education, Karamanoglu Mehmetbey University",
        47771: "Journal of Communication Aksaray, Aksaray University",
        47772: "International Journal of Research and Innovation"
               "in Social Science",
        47773: "Science and Humanities Journal",
        47774: "Institute of Ionosphere NAS and MES of Ukraine",
        47775: "STMIK Widya Pratama",
        47776: "Universitas Nahdlatul Ulama Indonesia",
        47777: "Journal of Humanities and Social Sciences,"
               "Cankaya University",
        47778: "Avrasya Spor Bilimleri ve Egitim Dergisi",
        47779: "American Journal of Health Studies",
        47780: "Journal of Inland Fisheries Society of India",
        47781: "ASIA-PACIFIC",
        47782: "Osterreichische Bankwissenschaftliche Gesellschaft",
        47783: "Literasi Kita Indonesia",
        47784: "Sociedad Materno Fetal",
        47785: "Urbana-Urban Affairs & Public Policy",
        47786: "Balboa Park Online Collaborative",
        47787: "Pan Africa Science Journal (PASJ)",
        47788: "University of Exeter Press",
        47789: "Bicol University",
        47790: "Instituto de Estudios Turolenses",
        47791: "Ribeirao Grafica e Editora",
        47792: "Instituto Espanol de Oceanografia",
        47793: "Humanistic Perspective",
        47794: "ITSK RS dr. Soepraoen",
        47795: "Whitehouse Publishing",
        47796: "Universidad Privada de Tacna",
        47797: "Instituto de Investigacion Universidad Nacional"
               "Autonoma de Tayacaja Daniel Hernandez Morillo",
        47798: "Alwasl University",
        47799: "Prathima Institute of Medical Sciences",
        47800: "Petrovietnam Journal",
        47801: "Korean Association for Infant Mental Health",
        47802: "Muzeul Etnografic al Transilvaniei",
        47803: "Societatea Romana de Cardiologie",
        47804: "All-Russian Research and Technological Institute"
               "of Biological Industry",
        47805: "Publisher Otchiy Dom",
        47806: "International Journal of Educational Spectrum",
        47807: "Universidad de Sonora books",
        47808: "Editorial Agricola Tuxpan",
        47809: "Academia Nationala de Muzica Gheorghe Dima",
        47810: "Institute for Bulgarian Language Prof. Lyubomir"
               "Andreychin at the Bulgarian Academy of Sciences",
        47811: "Khesar Gyalpo University of Medical Sciences of Bhutan",
        47812: "Digital Intelligentsia Consultancy Services",
        47813: "Krasnoyarsk Science and Technology City Hall",
        47814: "International Journal of Social Science Research and Review",
        47815: "Gramya Krishi Vikas Samiti",
        47816: "Academic Support, International Institute of Islamic Thought",
        47818: "Journal of Design to Resilience in Architecture & Planning",
        47819: "Humanidades e Tecnologia (FINOM)",
        47820: "RECIMA21 - Revista Cientifica Multidisciplinar",
        47821: "Revista Linguasagem",
        47822: "Bionorte",
        47823: "Revista Politica Hoje",
        47824: "Sekolah Tinggi Ilmu Ekonomi Yapman Majene",
        47825: "Korean Association for Geriatric Psychiatry",
        47826: "Pascasarjana Universitas Riau",
        47827: "Indonesian Institute for Corporate Learning"
               "and Studies (IICLS)",
        47828: "Jurnal Ilmuwan Administrasi Negara AsIAN (Asosiasi"
               "Ilmuwan Administrasi Negara)",
        47829: "Annals of Clinical and Medical Case Reports",
        47830: "Ikatan Dokter Indonesia",
        47831: "College of Education, Mustansiriyah University",
        47832: "Rimar Academy Egitim Arastirma ve Danismanlik"
               "Limited Sirketi",
        47833: "Neumann Janos Egyetem",
        47834: "Uluslararasi Turkce Ogretimi Arastirmalari Dergisi",
        47835: "Monre Akademi",
        47836: "Universiti Putra Malaysia",
        47837: "Caliphate Medical Journal",
        47838: "Annals of Clinical and Experimental Medicine",
        47839: "Research Institute for Intelligent Computer Systems",
        47840: "Ingenieria Agronomica, Facultad Ciencias Agrarias,"
               "Universidad Nacional Hermilio Valdizan",
        47841: "Asosiasi Dosen PkM Indonesia (ADPI)",
        47842: "Journal of Urban Technology and Sustainability",
        47843: "PREVENTIVE AND CLINICAL MEDICINE, North-Western"
               "State Medical University named after I.I. Mechnikov",
        47844: "Turkish Studies - Information Technologies"
               "and Applied Sciences",
        47845: "Turkish Studies - Language and Literature",
        47846: "Turkish Studies - Historical Analysis",
        47847: "Universidad de la Amazonia",
        47848: "United Research Publication",
        47849: "IJEPS Publications",
        47850: "Institute of Philosophy and Law of the Siberian"
               "Branch of the Russian Academy of Sciences",
        47851: "Department of Mass Communication Nnamdi Azikiwe University",
        47852: "BON VIEW PUBLISHING PTE",
        47853: "The Korean Society of Fisheries and Aquatic Science",
        47854: "Bibliotheque de l'Universite Laval",
        47855: "State Institution - D. F. Chebotarev Institute"
               "of Gerontology NAMS of Ukraine",
        47856: "Elixir Publishers",
        47857: "Iquz Galaxy Publisher",
        47858: "Korean Association of Music Therapy and Psychological"
               "Rehabilitation",
        47859: "Stikes Kesdam IX/Udayana",
        47860: "Sekolah Tinggi Ilmu Ekonomi Dewantara",
        47861: "Sekolah Tinggi Agama Buddha Nalanda",
        47862: "Apples - Journal of Applied Language Studies",
        47863: "Sociedade Cientifica de Pedagogia do Desporto",
        47864: "Sociedad Colombiana de la Ciencia del Suelo",
        47865: "Revista IGOBERNANZA, Instituto Internacional de Gobierno",
        47866: "Vietnamese Journal of Food Control, National"
               "Institute for Food Control",
        47867: "Hipertextus - Revista Digital",
        47868: "Scandia: Tidskrift for historisk forskning, Lund University",
        47869: "University of Transport and Communications",
        47870: "Atha Mais Editora Ltda",
        47871: "Instituto de Pesquisas Jardim Botanico do Rio de Janeiro",
        47872: "Centro de Investigacion y Docencia Economicas (CIDE)",
        47873: "National Center of Public Health and Analyses (NCPHA)",
        47874: "Al-Quds University",
        47875: "WellBeing International",
        47876: "Institutul de Lingvistica Iorgu Iordan - Al."
               "Rosetti al Academiei Romane",
        47877: "Journal of Addiction Problems, V. Serbsky NMRCPN",
        47878: "Instituto Lauro de Souza Lima",
        47879: "Epitaya Propriedade Intelectual e Editora LTDA",
        47880: "International Information Institute",
        47881: "Ghana Journal of Science, Technology and Development - GJSTD",
        47882: "GenBio",
        47883: "Sheik Zayed Medical College",
        47884: "Journal of Water Engineering and Management",
        47885: "Fakultas Ekonomi Universitas Islam Al-Azhar",
        47886: "American Fisheries Society",
        47887: "Sekolah Tinggi Agama Islam (STAI) Jamiatut"
               "Tarbiyah Lhoksukon Aceh Utara",
        47888: "University of Warsaw - Faculty of Archeology",
        47889: "University of Warsaw - Faculty of Culture and Arts",
        47890: "HELICS Online Publishing Private Limited",
        47891: "Sabratha University",
        47892: "Sociedad de Gastroenterologia del Peru",
        47893: "Interscience Research Network",
        47894: "Micropaleontological Foundation MicroPress Europe",
        47895: "University of the Philippines Manila",
        47896: "Pusat Penelitian dan Pengabdian Kepada Masyarakat"
               "(P3M) STIE H. Agus Salim Bukittinggi",
        47897: "International Scientific and Vocational Studies Journal",
        47898: "International Journal of Eastern Anatolia"
               "Science Engineering and Design",
        47899: "Izmir Sosyal Bilimler Dergisi",
        47900: "Sekolah Tinggi Teologi Cipanas",
        47901: "Pusat Studi dan Pengembangan Pelayanan Kaum Muda",
        47902: "Institut Agama Islam Agus Salim Metro Lampung",
        47903: "Sekolah Tinggi Ekonomi Syariah Islamic Village",
        47904: "SKIT Research Journal, Swami Keshvanand Institute"
               "of Technology, Management & Gramothan",
        47905: "Interregional Association of State and Law Theorists",
        47906: "Instituto da Defesa Nacional",
        47907: "Instituto Juridico da Universidade de Coimbra",
        47908: "Candlin & Mynard ePublishing Limited",
        47909: "ColNes",
        47910: "Institute for Applied Mathematics, Far Eastern"
               "Branch, Russian Academy of Sciences",
        47911: "National Institute for Legislative and Democratic"
               "Studies (NILDS), National Assembly, Nigeria",
        47912: "Society for Clinical Management",
        47913: "Sosiora Journal",
        47914: "Liberal Press",
        47915: "CSIR-National Environmental Engineering Research Institute",
        47916: "Journal of Knowledge Engineering and Management",
        47917: "WW Media GmbH",
        47918: "Universitas Bina Mandiri Gorontalo",
        47919: "Fundacion de estudios parlamentarios y del"
               "Estado autonomico Manuel Gimenez Abad",
        47920: "Universitas Bali Internasional",
        47921: "Aquatic Bioresources & Environment, FSBSI"
               "VNIRO, Azov-Black Sea Branch of the FSBSI VNIRO (AzNIIRKH)",
        47922: "Centre for Samoan Studies - National University of Samoa",
        47923: "Union of Concerned Scientists",
        47924: "Fundacion Cenit",
        47925: "Philosophy of Education Society",
        47926: "STTIKOM Insan Unggul Cilegon",
        47927: "STMIK Dharmapala Riau",
        47928: "Adyghe (Circassian) International Academy of Sciences",
        47929: "Institute for Research on Social Phenomena",
        47930: "SITRE. Simposio Internacional, trabalho, relacaes"
               "de trabalho, educacao e identidade",
        47931: "Plato Editorial",
        47932: "Ortadogu Etutleri, Center for Middle Eastern Studies",
        47933: "International Journal of Engineering and Innovative Research",
        47934: "Iktisadi ve Idari Bilimler Fakultesi Dergisi,"
               "Trakya University",
        47935: "Cocuk, Edebiyat ve Dil Egitimi Dergisi",
        47936: "Korean Encephalitis and Neuroinflammation Society",
        47937: "South Asian Journal Of Cranio-Maxillofacial & Dental Surgery",
        47938: "Smithsonian Institution - National Museum"
               "of the American Indian",
        47939: "Chongqing VIP Information Co.,Ltd.",
        47940: "T. K. Zhurgenov Kazakh National Academy of Arts",
        47941: "CARI Journals Limited",
        47942: "Universitas Sahid Surakarta",
        47943: "Artes Liberales sp. z o.o.",
        47944: "Indian Science and Technology Foundation",
        47945: "IAIN Sorong",
        47946: "Revista NERA",
        47947: "International Journal of Nature and Life Sciences",
        47948: "Karamanoglu Mehmetbey Universitesi Edebiyat"
               "Fakultesi Dergisi",
        47949: "The Womens Studies Center",
        47950: "Editura ARA - Arhitectura Restaurare Arheologie",
        47951: "Journal of Media and Religion Studies, Erciyes University",
        47952: "University Goettingen",
        47953: "SAE International",
        47954: "The Entrepreneurship Club",
        47955: "Journal of Korean Burn Society",
        47956: "Balkan Music And Art Journal",
        47957: "International Journal of Pharmaceutics and Drug Analysis",
        47958: "JSC Scientific Research Institute of Economics"
               "and Organization of Maintenance Supply",
        47959: "African Journal of Engineering and Technology",
        47960: "Sveuciliste u Mostaru",
        47961: "Institucion Universitaria Comando de Educacion y Doctrina",
        47962: "Institute of Electrical and Electronics Engineers"
               "(IEEE) StandardDocs",
        47963: "University of Cape Coast",
        47964: "European Association for Structural Dynamics EASD",
        47965: "Facultad de Humanidades y Ciencias de la Educacion",
        47966: "Universidad de la Empresa",
        47967: "South West Doctoral Training Partnership",
        47968: "Grand Academic Portal",
        47969: "International Peer-Reviewed Media and Communication"
               "Research Journal,Aydin Adnan Menderes University",
        47970: "Tanri Abeng University",
        47971: "Institut Agama Islam Latifah Mubarokiyah",
        47972: "The Vietnam Journal of Cardiovascular and Thoracic Surgery",
        47973: "Journal of Pediatric Research and Practice,"
               "Vietnam National Childrens Hospital",
        47974: "Taru Publications",
        47975: "International Journal of Digital Law",
        47976: "Revista Brasileira de Historia da Matematica - RBHM",
        47977: "Anais do Uso Publico em Unidades de Conservacao",
        47978: "Technical University of Sofia",
        47979: "Oriental Institute, Czech Academy of Sciences",
        47980: "International Islamic Academy of Uzbekistan",
        47981: "Military Institute of Science and Technology (MIST)",
        47982: "Stichting OpenAccess Foundation",
        47983: "Advertising and Information Agency in the"
               "Caucasian Mineral Waters",
        47984: "Parlamento Vasco / Eusko Legebiltzarra",
        47985: "Stichting Liliane Fonds",
        47986: "Hackers to Hackers Conference",
        47987: "All Nations University",
        47988: "American Nurses Association - New York",
        47989: "University of Boras, Faculty of Librarianship,"
               "Information, Education and IT",
        47990: "Asociacion Latinoamericana de Odontopediatria",
        47991: "CMJ Publishing Solutions",
        47992: "Srinivas University",
        47993: "Universidad Mayor de San Simon",
        47994: "Uluslararasi Sosyal Bilimler Akademi Dergisi (USBAD)",
        47995: "STISIPOL Pahlawan 12 Sungailiat",
        47996: "STAIINDO Jakarta",
        47997: "Shree Dadu Educational Society",
        47998: "Iletisim Kuram ve Arastirma Dergisi",
        47999: "Bosnjacka zajednica kulture Preporod - Institut"
               "za bosnjacke studije",
        48000: "Journal of Korean Association for Qualitative Research",
        48001: "QTanalytics India",
        48002: "Vidya Kutir Publications",
        48003: "Koninklijke Nederlandse Oudheidkundige Bond",
        48004: "Revista Artigos Estrategicos",
        48005: "Revista de Tecnologia Aplicada - RTA",
        48006: "Associacao Brasileira de Antropologia",
        48007: "Esély - Társadalom- és szociálpolitikai folyóirat",
        48008: "National Geographical Society of India (NGSI)",
        48009: "International Association for Computer Information",
        48010: "Institute for Philosophy, Political Science"
               "& Religion Studies of Committee Science of MES of the RK",
        48011: "Sociedade Brasileira de Automatica",
        48012: "The Bulletin of Izhevsk State Agricultural Academy",
        48013: "Inlight Publisher",
        48014: "Science Footprint Press Co., Limited",
        48015: "Lomonosov Moscow State University, School of World Politics",
        48016: "EDUNEAL - Editora da Universidade Estadual de Alagoas",
        48017: "Diversitas Journal - Periodico Científico Multidisciplinar",
        48018: "Hospital Vozandes Quito - HVQ SA",
        48019: "Office of the California Surgeon General",
        48020: "Kiev Institute of Modern Psychology and Psychotherapy",
        48021: "Editora Dialetica",
        48022: "Korean Society for Microbiology and Biotechnology",
        48023: "Ideas and Innovations Journal, Moscow",
        48024: "Sekolah Tinggi Ilmu Ekonomi Pemuda",
        48025: "Revista de Geomorfologia",
        48026: "Association of Medical Radiology Engineers"
               "in Federation of Bosnia and Herzegovina",
        48027: "National Museum of Natural History, BAS",
        48028: "First Assured Brilliant International Ltd",
        48029: "The Trained Nurses Association of India",
        48030: "Conferencia Franciscana de Espana y Portugal (CONFRES)",
        48031: "Mindanao State University at Naawan",
        48032: "NLA Hogskolen",
        48033: "The K Association of Education Research",
        48034: "Apabistia Press",
        48035: "Universidad Publica de Navarra. Biblioteca",
        48036: "Shaheed Zulfiqar Ali Bhutto Medical University, Pakistan",
        48037: "University Hospitals of Morecambe Bay NHS Foundation Trust",
        48038: "Red Latinoamericana Trabajo y Trabajadores",
        48039: "Technical University of Mombasa",
        48040: "Szaknyelvoktatok es - Kutatok Orszagos Egyesulete",
        48041: "Acadents",
        48042: "STIE Surya Nusantara",
        48043: "Universidad Tecnologica de la Mixteca",
        48044: "International Society of Arboriculture",
        48045: "European Association of Fish Pathologists (EAFP)",
        48046: "Baghdad Medical Journal of Students",
        48047: "Institute for Advanced Studies",
        48048: "College of Graduate Studies, Walailak University",
        48049: "Food Safety Authority of Ireland",
        48050: "Melanoma Research Alliance",
        48051: "Revista Cientifica de Estetica e Cosmetologia",
        48052: "University of Sarajevo",
        48053: "Turkish Journal of Geoscience",
        48054: "The Korea Society of Homiletics",
        48055: "PT. Pilar Pengembangan Profesi",
        48056: "Fakultas Teknik UNDAR",
        48057: "Sociedad Argentina de Hematologia",
        48058: "Kasa Rolniczego Ubezpieczenia Spolecznego",
        48059: "Orebro University",
        48060: "The Global Health Network, University of Oxford",
        48061: "Sociedad Argentina de Nutricion",
        48062: "Association for Computer, Electronics and Education (ACEE)",
        48063: "Instituto de Perfeccionamiento y Estudios Superiores",
        48064: "Siirt University",
        48065: "Southern State Agricultural Research Station"
               "of the Institute of Water Problems and Land Reclamation",
        48066: "Social Sciences: Theory and Practice",
        48067: "International Journal of Active Learning",
        48068: "Turkiye Rusya Arastirmalari Dergisi (TRAD)",
        48069: "Akdeniz Sanat, Akdeniz University",
        48070: "Journal of Social Sciences Institute, Erciyes University",
        48071: "University of Health Sciences Turkey",
        48072: "Instituto Brasileiro de Petroleo e Gas",
        48073: "Morula Editorial",
        48074: "Aceno - Revista de Antropologia do Centro-Oeste",
        48075: "Universidade Estadual do Oeste do Parana - UNIOESTE",
        48076: "Ukrainian Research and Design Institute of"
               "Building Materials and Products",
        48077: "Scientific Horizons",
        48078: "RJI Publisher",
        48079: "Akper Hermina Manggala Husada",
        48080: "PalArch Foundation",
        48081: "Toraighyrov University",
        48082: "Editorial Espacios",
        48083: "Societe Internationale dUrologie",
        48084: "Engineering, Technology & Applied Science Research",
        48085: "Brain Health Alliance",
        48086: "Future Biochemistry and Bioscience",
        48087: "Association pour la Recherche Pharmaceutique"
               "et l Enrichissement des Connaissances (ARPEC)",
        48088: "European Association of Geographers, EUROGEO ivzw",
        48089: "Foundations of Ophthalmology",
        48090: "Fundacao Stemmer para Pesquisa, Desenvolvimento e Inovacao",
        48091: "Georgetown Scientific Research Journal",
        48092: "Akademi Kebidanan Ummi Khasanah",
        48093: "Program Pascasarjana Universitas Sjakhyakirti",
        48094: "STIT Raudhatul Ulum",
        48095: "Care Comm",
        48096: "Institute of Bast Crops of the National Academy"
               "of Agrarian Sciences of Ukraine",
        48097: "Faculdade Metropolitana da Grande Recife",
        48098: "Revista LibertAcao",
        48099: "Revista da Micro e Pequena Empresa - RMPE",
        48100: "Management & Economics Research Journal, Ziane"
               "Achour University of Djelfa",
        48101: "Uppsala Medical Society",
        48102: "Universidad Iberoamericana, A. C.",
        48103: "Al-Balqa Applied University",
        48104: "Revista Venezolana de Cirugía",
        48105: "Children's Tumor Foundation",
        48106: "Verein philosophie.ch",
        48107: "Trinidad and Tobago Medical Association",
        48108: "Jurnal BPPK, Badan Pendidikan dan Pelatihan"
               "Keuangan Kementerian Keuangan RI",
        48109: "SDN 195/V Pinang Merah",
        48110: "Impact Research",
        48111: "Pakistan Endocrine and Thyroid Surgeons Association (PETSA)",
        48112: "International Research and Publishing Academy",
        48113: "FSBSI - All-Russian Scientific Research Institute"
               "of Tobacco, Makhorka and Tobacco Products",
        48114: "Theory and Methods of Physical Culture, Kazakh"
               "Academy of Sports and Tourism",
        48115: "Humanities Institute, Pusan National University",
        48116: "Orijentalni institut Univerziteta u Sarajevu"
               "(University of Sarajevo - Oriental Institute)",
        48117: "Hikmet Dernegi",
        48118: "Kent ve Cevre Arastirmalari Dergisi",
        48119: "Journal of Tourism, Leisure and Hospitality,"
               "Anadolu University",
        48120: "Ortacag Arastirmalari Dergisi",
        48121: "Journal of International Health Sciences and Management",
        48122: "Amisos",
        48123: "Turkish Journal of Remote Sensing and GIS",
        48124: "Halic University Journal of Health Sciences",
        48125: "Jurnal Teologi Biblika, Sekolah Tinggi Teologi"
               "Biblika Jakarta",
        48126: "Coal Energy Technology Institute of the National"
               "Academy of Sciences of Ukraine",
        48127: "Scientific Methodical Centre Scientia Educologica",
        48128: "Faculty of Geography and Regional Studies"
               "(Wydzial Geografii i Studiow Regionalnych)",
        48129: "Kuwait Journal of Science",
        48130: "Maximum Academic Press",
        48131: "Cagatay Sarp",
        48132: "Harvard Deusto Business Research, EAE Business School",
        48133: "Igdir Universitesi Spor Bilimleri Dergisi",
        48134: "JURKESSUTRA (jurnal kesehatan surya nusantara)"
               "Akper Surya Nusantara",
        48135: "Indonesian Mathematics Educators Society (I-MES)",
        48136: "Omsk State Agrarian University named after P.A. Stolypin",
        48137: "Geoeconomics of Energetics, Institute of Diaspora"
               "and Integration",
        48138: "Caucasian Journal of Science",
        48139: "Kulliye, Ankara Yildirim Beyazit University",
        48140: "Digital Editora",
        48141: "Southern Brazilian Journal of Chemistry",
        48142: "Revista Organizacoes Rurais & Agroindustriais (OR&A)",
        48143: "Revista RDAI",
        48144: "Universitas Muhammadiyah Pekajangan Pekalongan",
        48145: "Tokat Gaziosmanpasa Universitesi Sosyal Bilimler"
               "Arastirmalari Dergisi",
        48146: "ODU Sosyal Bilimler Arastirmalari Dergisi (ODUSOBIAD)",
        48147: "Uluslararasi Disiplinler Arasi Dil Arastirmalari Dergisi",
        48148: "Lund Journal of English Studies, Lund University",
        48149: "Association on Communications, Information,"
               "Electronic and Energy Systems",
        48150: "American Research and Publication Center",
        48151: "Fitzpatrick Clan Society",
        48152: "Regents of the University of California, Berkeley",
        48153: "Association of Radiographers of Nigeria",
        48154: "University of Bucharest, Faculty of Sociology"
               "and Social Work",
        48155: "Australian and New Zealand Law and History Society",
        48156: "Worldwide Dragonfly Association",
        48157: "iScholar",
        48158: "Grupo Pacifico",
        48159: "Revista do Instituto de Direito Constitucional e Cidadania",
        48160: "Universidad Nacional de Quilmes",
        48161: "Qubahan Organization for Development",
        48162: "Universidad Nacional de Cuyo",
        48163: "Facultad de Ciencias de la Educacion - Universidad"
               "de la Empresa",
        48164: "Federal State Budgetary Educational Institution"
               "of Higher Education - Samara State Institute of Culture",
        48165: "South Asian Journal of Social Sciences and Humanities",
        48166: "Tugba Yanpar Yelken",
        48167: "NHL Stenden University of Applied Sciences",
        48168: "Universidad Nacional de San Agustin",
        48169: "Ecuatesis",
        48170: "Research Centre for the Study of Music, Media,"
               "and Place (MMaP)",
        48171: "Jurnal WASAKA Critical Law Review, Lembaga"
               "Pendidikan Dan Publikasi Riset Ilmu Hukum",
        48172: "University of South Carolina Press",
        48173: "Hoedspruit Development Publication",
        48174: "BAYTEREK Journal of International Academic Research",
        48175: "Naksh Solutions",
        48176: "Eskisehir Medical Journal, Eskisehir City Hospital",
        48177: "Ikatan Senat Mahasiswa Farmasi Seluruh Indonesia",
        48178: "Intersection Magazine of the State University of Alagoas",
        48179: "Revext - Extension Journal of the State University"
               "of Alagoas",
        48180: "Revista Ambientale of the State University of Alagoas",
        48181: "Pascasarjana Universitas Sultan Ageng Tirtayasa",
        48182: "Politeknik Manufaktur Bandung",
        48183: "Revista Cientifica Multidisciplinar Brilliant Mind",
        48184: "Almaty Technological University JSC",
        48185: "SABA Publishing",
        48186: "STIKes Respati",
        48187: "Nanotechnology Research Group",
        48188: "Sveuciliste u Splitu",
        48189: "New Literaria",
        48190: "Universidad Tecnica de Machala",
        48191: "AKAFARMA Sunan Giri Ponorogo",
        48192: "Politeknik Ilmu Pelayaran Makassar",
        48193: "Revista Mexicana de Urologia based at Sociedad"
               "Mexicana de Urologia. Colegio de Profesionales AC.",
        48194: "Progressive Publications",
        48195: "SEPE - Simposio de Ensino, Pesquisa e Extensao",
        48196: "Center for Agri-Fisheries and Biosystems Mechanization"
               "(BIOMECH)",
        48197: "The Augmented Text Company LTD",
        48198: "Nigerian Journal of Pure and Applied Sciences",
        48199: "Journal of Urban Planning and Architecture of Afghanistan",
        48200: "Bulletin of the Russian-Armenian University",
        48201: "Zhiganov Kazan State Conservatoire",
        48202: "Lychnos, Lund University",
        48203: "Turkish Journal of Shiite Studies",
        48204: "Universidad de Panama",
        48205: "Gurukula Kangri (Deemed to be University)",
        48206: "The Korean Career, Entrepreneurship & Business Association",
        48207: "E-papers Servicos Editoriais Ltda",
        48208: "Headache Medicine",
        48209: "Arco Editores",
        48210: "Revista Temporis[acao]",
        48211: "Web Revista SOCIODIALETO",
        48212: "Revista Culturas & Fronteiras",
        48213: "Centro de Estudos Migratorios",
        48214: "United Nations University Institute for the"
               "Advanced Study of Sustainability (UNU-IAS)",
        48215: "Zhende Business Consulting Co. (Private) Limited",
        48216: "SCIO Publishing",
        48279: "Pacini Editore",
        48286: "Edra SpA",
        48294: "Scalpendi Editore",
        48295: "Giordano Editore",
        48297: "Asociacion Espanola de Asesores Fiscales (AEDAF)",
        50505: "Crossref Test",
        51117: "Medeniyet ve Toplum Dergisi, Necmettin Erbakan University",
        51118: "Konya Sanat, Necmettin Erbakan University",
        51119: "Necmettin Erbakan Universitesi Eregli Egitim"
               "Fakultesi, Necmettin Erbakan University",
        51120: "Necmettin Erbakan Universitesi Hukuk Fakultesi"
               "Dergisi, Necmettin Erbakan University",
        51121: "Necmettin Erbakan Universitesi Ilahiyat Fakultesi"
               "Dergisi, Necmettin Erbakan University",
        51122: "Necmettin Erbakan Universitesi Dis Hekimligi"
               "Dergisi (NEU Dent J), Necmettin Erbakan University",
        51123: "Journal of General Health Sciences (JGEHES),"
               "Necmettin Erbakan University",
        51124: "Necmettin Erbakan Universitesi Siyasal Bilgiler"
               "Fakultesi Dergisi, Necmettin Erbakan University",
        51125: "Jurusan Ekonomi Pembangunan Fakultas Ekonomi"
               "dan Bisnis Universitas Pattimura",
        51126: "Rede Academica das Ciencias da Saude da Lusofonia",
        51127: "Akhtar Saeed Medical & Dental College, Lahore, Pakistan",
        51128: "Vetinfo",
        51129: "Phytochemistry and Ayurveda Society",
        51130: "MONOMAX Limited Liability Company",
        51131: "Basaveshwara Medical College & Hospital, Chitradurga,"
               "Karnataka",
        51132: "Jurnal Teknologika, Sekolah Tinggi Teknologi Wastukancana",
        51133: "Studia Mediaevalia Bohemica",
        51134: "Institute of Contemporary History, Czech Academy of Sciences",
        51135: "Lembaga Layanan Perguruan Tinggi (LLDikti) Wilayah XII",
        51136: "Swedish Journal of BioScience Research",
        51137: "Wohllebe & Ross Publishing",
        51138: "Forsyningstilsynet",
        51139: "Modern Nyelvoktatas",
        51140: "The Institute Theology Science Florida USA (ITS-USA)",
        51141: "International Research Establishment for Energy"
               "and Environment",
        51142: "Western Sydney University",
        51143: "STIKES Suaka Insan",
        51144: "Necmettin Erbakan University Press",
        51145: "Journal of Life Science and Biomedicine",
        51146: "Connection Scientific Journal - CSJ",
        51147: "Revista Criminalistica e Medicina Legal",
        51148: "Journal of Art and Architecture Studies",
        51149: "Polish European Communities Studies Association (PESCSA)",
        51150: "Gaples Institute for Integrative Cardiology",
        51151: "Institute of Social Sciences and Humanities, Skopje",
        51152: "Journal of Basic and Applied Research in Biomedicine",
        51153: "Karachi Institute of Economics and Technology",
        51154: "Association of History Students - Toma Arhidakon",
        51155: "Korean Academy of Multicultural Welfare",
        51156: "The Northeast Asian Business and Economics Association",
        51157: "Morphology",
        51158: "Universitas Kahuripan Kediri",
        51159: "MediSys EduTech Pvt. Ltd.",
        51160: "Publishing House Sphere",
        51161: "Revista Multidisciplinar em Saude",
        51162: "Brazilian Congress",
        51163: "Creative Generation",
        51164: "World Alliance for Arts Education",
        51165: "International Teaching Artists Collaborative",
        51166: "Scottish Educational Review",
        51167: "Israel Chemical Society (ICS)",
        51168: "Students Journal of Health Research Africa",
        51169: "Dinas DIKPORA Daerah Istimewa Yogyakarta",
        51170: "Universitas Pradita",
        51171: "STIK Makassar",
        51172: "Badan Riset dan Inovasi Daerah Provinsi Bali",
        51173: "Middle Technical University",
        51174: "Centre for Defence Research",
        51175: "Univerzitet u Beogradu  -  Arhitektonski fakultet",
        51176: "The economy: strategy and practice, Institute"
               "of Economics Science of the Republic of Kazakhstan",
        51177: "Sosyal Bilimler Dergisi - Kayseri University"
               "Journal of Social Sciences",
        51178: "CV. Pusdikra Mitra Jaya",
        51179: "Universitas Almuslim",
        51180: "Laboratory of Humanitarian Projects",
        51181: "Fakultas Kesehatan Masyarakat Universitas Indonesia",
        51182: "Fakultas Ekonomi & Bisnis, Universitas Khairun",
        51183: "Dunamelleki Reformatus Egyhazkerulet Raday Gyujtemeny",
        51184: "Dunamelleki Reformatus Egyhazkerulet Raday Gyujtemeny",
        51185: "Universite de Geneve",
        51186: "Universite de Geneve",
        51187: "FSUE Central Scientific Research Automobile"
               "and Automotive Engines Institute (FSUE NAMI)",
        51188: "Universidad Central de Chile",
        51189: "Revista Multidisciplinar em Educacao e Meio Ambiente",
        51190: "Sekolah Tinggi Ilmu Adab dan Budaya Islam Riyadlul Ulum",
        51191: "INSAM Institute for Contemporary Art, Music and Technology",
        51192: "Almubin Islamic Scientific Journal, Institut"
               "Ummul Quro Al-Islami Bogor",
        51193: "Malwa International Journal's Publication",
        51194: "Institute of Astronomy of the Russian Academy of Sciences",
        51195: "Jurnal Ekonomi Integra, STIE Indonesia Pontianak",
        51196: "Faculty of Sociology (Wydzial Socjologii)",
        51197: "Llapanchikpaq: Justice",
        51198: "Ukrainian Association of Sinologists",
        51199: "The Vietnam Association of Endolaparoscopic Surgeons",
        51200: "Universiti Malaysia Sabah (UMS)",
        51201: "ADD Technologies",
        51202: "VDI Verlag GmbH",
        51203: "Presses Universitaires de Caen",
        51204: "Univerzitet u Beogradu - Pravni fakultet",
        51205: "Lestu Editora, Grafica e Consultoria Ltda.",
        51206: "Camara dos Deputados",
        51207: "Revista Psicopedagogia",
        51208: "Revista Saude e Inovacao",
        51209: "Municipal Academy of Circus and Variety Arts",
        51210: "Combinatorial Theory",
        51211: "Universitas Bina Insani",
        51212: "Jurnal DInamika Pendidikan, FKIP Universitas"
               "Kristen Indonesia",
        51213: "LPPM Universitas Merdeka Pasuruan",
        51214: "Kuras Institute",
        51215: "Irkutsk State Agrarian University named after A.A. Ezhevsky",
        51216: "Theological Collection of Tambov Theological"
               "Seminary, Tambov Theological Seminary",
        51217: "Cognition and Experience, Moscow Institute of Psychoanalysis",
        51218: "Joint Super Computer Center of the Russian"
               "Academy of Sciences",
        51219: "United Research Forum",
        51220: "Society for Himalayan Action Research and"
               "Development (SHARAD)",
        51221: "University of South Carolina Libraries",
        51222: "International Journal of Martial Arts",
        51223: "Korean Society of Martial Arts",
        51224: "Society for Transparency, Openness, and Replication"
               "in Kinesiology",
        51225: "Akademi Farmasi Dwi Farma",
        51226: "Jurnal As-Salam, STAI Darussalam Lampung",
        51227: "Online Journal of Animal and Feed Research",
        51228: "Journal of Civil Engineering and Urbanism",
        51229: "Turkish Journal of Public Administration",
        51230: "Rehabilitacion Integral",
        51231: "Journal of Biomedical and Medical Sciences",
        51232: "Moscow Center of Advanced Sport Technologies",
        51233: "Ltd Radomir-M+",
        51234: "Associacao Brasileira de Enfermagem",
        51235: "Faculty of Criminalistics, Criminology and"
               "Security Studies, University of Sarajevo",
        51236: "Academy of Watercolor and Fine Arts of Sergey Andriaka",
        51237: "University of Sarajevo Institute for History",
        51238: "Institute of Jesuit Sources",
        51239: "Newport Institute of Communications and Economics, Karachi",
        51240: "Seoul School of Integrated Sciences and Technologies",
        51241: "Turkish Journal of Natural Sciences",
        51242: "Turkish Journal of Educational Research",
        51243: "Turkish Journal of Management and Economics",
        51244: "International Journal of Research and Scientific Innovation",
        51245: "Ramanujan College",
        51246: "Faculty of Tourism and Hospitality, Ohrid",
        51247: "Instituto Tecnologico Superior Jubones",
        51248: "Indian Association of Biomedical Scientists",
        51249: "Periodicojs",
        51250: "Prof Pubs LLC",
        51251: "Revista Cientifica Internacional I+D+I",
        51252: "Universidad Nacional de San Martin",
        51253: "Army Medical College",
        51254: "Mari Scientific Research Institute of Language,"
               "Literature and History named after V.M. Vasilyev",
        51255: "St. Petersburg Institute of History of the"
               "Russian Academy of Sciences",
        51256: "HealthCom Media",
        51257: "Editions Techniques de l Ingenieur",
        51258: "Academia de Stiinte Agricole si Silvice 'Gheorghe"
               "Ionescu-Sisesti'",
        51259: "AFKAR",
        51260: "Escuela Superior Politecnica Agropecuaria"
               "de Manabi Manuel Felix Lopez",
        51261: "Yuksek Ihtisas Universitesi Saglik Bilimleri"
               "Dergisi, Yuksek Ihtisas Universitesi",
        51262: "Eurasian Journal of Toxicology, Emergency"
               "Physicians Association of Turkey",
        51263: "STIM Lasharan Jaya Makassar",
        51264: "Masyarakat Limnologi Indonesia",
        51265: "Global Finance Society",
        51266: "Jurnal Borneo Akcaya, Badan Penelitian dan"
               "Pengembangan Provinsi Kalimantan Barat",
        51267: "National University of Physical Education and Sport",
        51268: "Prime Scholars Library",
        51269: "Appalachian State University",
        51270: "Canadian Archaeological Association",
        51271: "MediHealth Academy",
        51272: "The North American Chapter of the International"
               "Group for the Psychology of Mathematics Education",
        51273: "Services Institute of Medical Sciences, Lahore",
        51274: "Universidad Autonoma de Santo Domingo (UASD)",
        51275: "Alim - Journal of Islamic Education, Kopertais"
               "Wilayah - DKI Jakarta and Banten",
        51276: "Natural Aceh",
        51277: "Sekolah Tinggi Ilmu Ekonomi SBI Yogyakarta",
        51278: "CV. Creative Tugu Pena",
        51279: "STIA Nusantara Sakti Sungai Penuh",
        51280: "Yildiz Journal of Educational Research, Yildiz"
               "Technical University",
        51281: "Revista de Instrumentos, Modelos e Politicas"
               "em Avaliacao Educacional",
        51282: "Edicoes Verona Comercio de Livros e Revistas Ltda - ME",
        51283: "Revista Corpoconsciencia",
        51284: "Revista Brasileira de Pesquisas Juridicas",
        51285: "ITD Perspektiva",
        51286: "Research Institute of Science and Technology (RISAT)",
        51287: "Current Trends in Translation Teaching and Learning E",
        51288: "Sociedad de Obstetricia y Ginecologia de Venezuela",
        51289: "Jurnal Penelitian Teori Dan Terapan Akuntansi,"
               "STIE Kesuma Negara Blitar",
        51290: "Dumlupinar University Journal of Social Sciences",
        51291: "WellBeing International Publications",
        51292: "CV. Caraka Khatulistiwa",
        51293: "Socrates Journal of Interdisciplinary Social Researches",
        51294: "GOBEKLITEPE Saglik Bilimleri Dergisi",
        51295: "Digital Communication Journal",
        51296: "New Era Journal of Interdisciplinary Social Studies",
        51297: "Rumeli Tarih Arastirmalari Dergisi",
        51298: "Vietnam Medical Journal, Vietnam Medical Association",
        51299: "Buddhist Institute of Sungchol Thought",
        51300: "Luminous Insights LLC",
        51301: "Satbayev University",
        51302: "CEF. Centro de Estudios Financieros",
        51303: "INTBAU Spain",
        51304: "BioVat",
        51305: "Ustav pro ceskou literaturu AV CR, v. v. i.",
        51306: "Instituto Otavaleno de Antropologia",
        51307: "Yerevan Brusov State University of Languages"
               "and Social Sciences",
        51308: "Revista Continentes",
        51309: "Pan-Ukrainian Association of Traumatology and Osteosynthesis",
        51310: "Fakultas Teknologi Pertanian Universitas Papua",
        51311: "Institut Agama Islam Yasni Bungo",
        51312: "Krasnodar University of the Ministry of the"
               "Interior of Russia",
        51313: "Freeside Europe Online Academic Journal",
        51314: "Faculty of Educational Studies of Lomonosov"
               "Moscow State University",
        51315: "Kerns Verlag",
        51316: "Hanoi University of Science and Technology",
        51317: "Editon Consortium Publishing",
        51318: "Cifra Ltd - Russian Agency for Digital Standardization"
               "(RADS)",
        51319: "International Journal of Advance Scientific"
               "Research and Engineering Trends",
        51320: "Conselho Regional de Contabilidade de Minas Gerais",
        51321: "Journal of Sports Korean Medicine & Clinical"
               "Pharmacopuncture",
        51322: "Tambovskiy poligraficheskiy soyuz",
        51323: "AKS University, Satna",
        51324: "Uniedusul Editora",
        51325: "EuroMid Academy of Business and Technology",
        51326: "Colegio Oficial de Enfermeria de Malaga",
        51327: "International Association of Management Spirituality"
               "& Religion",
        51328: "Association Medecine et Philosophie",
        51329: "International Virtual Ophthalmic Research Center",
        51330: "Global Affairs Review",
        51331: "Association of Education and Thought Fettah Efendi",
        51332: "Indian Psychiatric Society, West Bengal Branch",
        51333: "Indian Society of Professional Social Work (ISPSW)",
        51334: "Politeknik Bisnis dan Pasar Modal",
        51335: "The Organoid Society",
        51336: "STKIP Pembangunan Indonesia",
        51337: "Alumni scientiae Bohemicae, z.s.",
        51338: "Sociedade Portuguesa de Psiquiatria e Saude Mental",
        51339: "Institut Agama Islam Sunan Kalijogo Malang",
        51340: "Grafica e Editora Gaucha Ltda",
        51341: "Revista Contabilidade Gestao e Governanca",
        51342: "Paulus Law Journal, Universitas Kristen Indonesia Paulus",
        51343: "Universidad Nacional de San Antonio Abad del Cusco",
        51344: "Birne Publishing House",
        51345: "AlMaarif University College (AUC)",
        51346: "Tashkent State Technical University",
        51347: "International Seminar on Urban Form",
        51348: "Uzbekistan Research Online",
        51349: "Faculty of Geography and History, University"
               "of Las Palmas de Gran Canaria",
        51350: "Healthcare of Kyrgyzstan",
        51351: "Jurnal TIMES, STMIK TIME",
        51352: "Jurnal Ilmiah Manuntung, Sekolah Tinggi Ilmu"
               "Kesehatan Samarinda",
        51353: "Universitas Paramadina",
        51354: "Kyrgyz-Turkish Manas University",
        51355: "Journal of Research in STEM Education, Hacettepe"
               "Science Center",
        51356: "Sociedade Portuguesa de Psicanalise",
        51357: "University of Ontario Institute of Technology Library",
        51358: "Sociedade Brasileira de Design da Informacao",
        51359: "Universidade Federal de Pernambuco",
        51360: "Editora ZH4",
        51361: "Instituto Federal do Piaui",
        51362: "Loma Linda Publishing Company",
        51363: "Cantonal and University Library Fribourg",
        51364: "Humanities Research Centre",
        51365: "The Society of Life Sciences, Satna",
        51366: "Procuradoria Geral do Trabalho",
        51367: "Editora Educitec",
        51368: "Joint-Stock Company Scientific and Production"
               "Association - ORION",
        51369: "National Academy of the Security Service of Ukraine",
        51370: "Masyarakat Hukum Pidana dan Kriminologi Indonesia",
        51371: "Physical Education Pedagogues Association, Ljubuski",
        51372: "Bioagro, Universidad Centroccidental Lisandro Alvarado",
        51373: "Reciti Kiado",
        51374: "West African Postgraduate College of Medical"
               "Laboratory Science",
        51375: "G-CSC, University of Frankfurt",
        51376: "Aksara Rentaka Siar",
        51377: "Pusat Pungutan Zakat (PPZ-MAIWP)",
        51378: "Universidad Centroamericana Jose Siemon Canas"
               "UCA El Salvador-Biblioteca P. Florentino Idoate, S.J.",
        51379: "Institute of Pedagogy, Psychology and Social Problems",
        51380: "Gomal University Journal of Research",
        51381: "University of Applied Sciences Velika Gorica",
        51382: "Technische Universitat Chemnitz",
        51383: "Mevlut AYDOGMUS",
        51384: "Certamen",
        51385: "Intersecciones en Comunicacion",
        51386: "Science Research Publication",
        51387: "New England Statistical Society",
        51388: "Digital Promise Global",
        51389: "Debates & Equacoes, Unipessoal Lda",
        51390: "Eastern Mennonite University",
        51391: "ASTRA National Museum Complex",
        51392: "Enterprise Sadeg EIRL",
        51393: "Huatuo Culture Media Co., Limited",
        51394: "National Crisisonomy Institute, Chungbuk National University",
        51395: "The Korean Society of Philosophical Practice",
        51396: "Journal of Natural Resource Conservation and Management",
        51397: "Open Access International Journal of Science and Engineering",
        51398: "INCT-INEU (Instituto Nacional de Ciencia e"
               "Tecnologia para Estudos sobre os Estados Unidos)",
        51399: "Reunina - A Revista Cientifica da Faculdade Unina",
        51400: "National Taiwan Ocean University",
        51401: "Jurnal Informatika, Teknologi dan Sains Universitas"
               "Teknologi Sumbawa",
        51402: "Pimpinan Wilayah Muhammadiyah Jawa Tengah",
        51403: "Tap chi Y hoc du phong (Vietnam Journal of"
               "Preventive Medicine-VJPM)",
        51404: "MetaMind: Journal of Artificial Intelligence"
               "and Philosophy of Mind",
        51405: "Association of Arab Universities Journal for Arts",
        51406: "Federal University of Agriculture, Abeokuta (FUNAAB)",
        51407: "Malaysian Paediatric Association",
        51408: "Institute for Informatics and Automation Problems - NAS RA",
        51409: "Academys Herald, Private Institution of Higher"
               "Education Moscow Academy of Entrepreneurship",
        51410: "CIRU - Centar za istrazivanje i razvoj upravljanja d.o.o.",
        51411: "Rahat ul Quloob Research Academy, Quetta",
        51412: "The Pharmaceutical Society of Nigeria",
        51413: "Hubungan Internasional Universitas Pertamina",
        51414: "Stockholm Environment Institute",
        51415: "Durban University of Technology",
        51416: "Association for Innovation in Building and Construction",
        51417: "Institut Catala d Arqueologia Classica, ICAC",
        51418: "Rave Reviewer Publishing",
        51419: "Balance Profi LLC",
        51420: "The Learning Partnership",
        51421: "Nordic Digital Publishing",
        51422: "Revista de Enfermeria Neurologica INNN",
        51423: "Non-governmental Organization - Research and"
               "Education Centre - SUCCESSFUL",
        51424: "Al-araqah Foundation",
        51425: "International Islamic Schools Alliance",
        51426: "Science Foundation Ireland",
        51427: "Faculdade de Letras, Universidade de Lisboa",
        51428: "Sociological Review Foundation",
        51429: "Egyptian Society for Medical Microbiology",
        51430: "Southern Academic Publisher",
        51431: "Universidad Nacional Jose Faustino Sanchez Carrion",
        51432: "Proud Pen Limited",
        51433: "Biblioteca Nacional del Peru",
        51434: "University of Economics and Innovation in Lublin",
        51435: "Turkish Journal of Analytical Chemistry.",
        51436: "Sociedad(e) Iberica de Micologia",
        51437: "Meditech Foundation",
        51438: "HERMES DATA SA DE CV",
        51439: "Korea TESOL",
        51440: "Universidad Nacional San Cristobal de Huamanga",
        51441: "University of Health Sciences Lahore",
        51442: "Armenian Genocide Museum-Institute Foundation",
        51443: "Centro Iberico de Restauracion Fluvial",
        51444: "Art Architecture Design Research",
        51445: "Asociacion Espanola de Audiologia",
        51446: "International Journal of Research in Ayurveda"
               "and Medical Sciences",
        51447: "Uludag University Faculty of Theology",
        51448: "Insan Medeniyet Hareketi",
        51449: "Advance Journal of Pharmaceutical Sciences",
        51450: "Tokat Gaziosmanpasa Universitesi Islami Ilimler"
               "Fakultesi, Tokat Ilmiyat Dergisi",
        51451: "Neumologia Pediatrica, Sociedad Chilena de"
               "Neumologia Pediatrica",
        51452: "S.Seifullin Kazakh Agro Technical University",
        51453: "Tap chi Khoa hoc, Dai hoc Tan Trao",
        51454: "Universitas Muhammadiyah Kendari",
        51455: "Polymatheia",
        51456: "International Journal of Engineering and Innovative"
               "Technology (IJEIT)",
        51457: "Korean Association of Coaching Psychology",
        51458: "Bee Studies- Apiculture Research Institute",
        51459: "Federal University of Technology Akure",
        51460: "Dokuz Eylul University",
        51461: "VostokSibAcademCenter",
        51462: "The Korean Society for Nursing Ethics",
        51463: "Contemporary Pediatric Dentistry",
        51464: "Nunavut Services O.U.",
        51465: "Sekolah Tinggi Teologi Tenggarong",
        51466: "Journal of Electrical Engineering and Information"
               "Technologies",
        51467: "The Animation Society of Korea",
        51468: "Sekolah Tinggi Ilmu Tarbiyah Madani Yogyakarta",
        51469: "Gunes Vakfi",
        51470: "Plant Archives",
        51471: "International Journal of Entrepreneurship Case Study (IJECS)",
        51472: "SIELAE - Seminario Interdisciplinar",
        51473: "RCMOS - Revista Científica Multidisciplinar O Saber",
        51474: "Kathmandu University School of Education",
        51475: "Journal of Educational and Management Studies",
        51476: "Sekolah Tinggi Agama Islam (STAI) Binamadani",
        51477: "Middle East Journal of Science",
        51478: "Atlas of Genetics and Cytogenetics in Oncology"
               "and Hematology",
        51479: "Czech Dental Chamber",
        51480: "Polish Communication Association",
        51481: "Colegio de Medicos y Cirujanos de Costa Rica",
        51482: "STAI Sebelas April Sumedang",
        51483: "SvedbergOpen",
        51484: "Osrodek Rozwoju Spoleczenstwa Obywatelskiego",
        51485: "Signals and Systems Laboratory, University"
               "Mhamed Bougara of Boumerdes",
        51486: "CIC Lembaga Riset dan Konsultan Sosial",
        51487: "Jurnal Kesehatan dan Sains (JKS) STIKES Griya Husada Sumbawa",
        51488: "LLP International Educational Corporation",
        51489: "Turkish Journal of Remote Sensing",
        51490: "Oksident",
        51491: "Magyar Hadtudomanyi Tarsasag",
        51492: "California Fish and Wildlife Journal, California"
               "Department of Fish and Wildlife",
        51493: "Journal of Archaeology, Ege University",
        51494: "Jurnal Pendidikan Dasar Flobamorata (JPDF)"
               "Program Studi Pendidikan Guru Sekolah Dasar (PGSD)",
        51495: "Direktorat Pendidikan Masyarakat dan Pendidikan"
               "Khusus, Ditjen PAUD Dikdas Dikmen, Kemdikbud",
        51496: "Journal of Global Medicine, Ibadan Medical Specialist Group",
        51497: "Reflexao Academica",
        51498: "Journal of Mahad Aly Al Fithrah: PUTIH (Jurnal"
               "Pengetahuan tentang Ilmu dan Hikmah)",
        51499: "Sekolah Tinggi Keguruan dan Ilmu Pendidikan Syekh Manshur",
        51500: "Online-Publishing Lie Gudzyk",
        51501: "National Academy of Forensic Engineers",
        51502: "Dunya Veteriner Dergisi",
        51503: "Gelisim ve Psikoloji Dergisi (GPD)",
        51504: "Shandong University of Science and Technology",
        51505: "International Journal of Medical Science and Health Research",
        51506: "Institute of Islamic Studies, Punjab University, Lahore",
        51507: "Medical Association of Pharmacopuncture Institute",
        51508: "Proceedings of INTCESS International Conference"
               "on Education and Education of Social Sciences",
        51509: "Acta Universitatis Upsaliensis",
        51510: "Politeknik Negeri Medan",
        51511: "Indonesian Young Scientist Group",
        51512: "STIE Muara Teweh",
        51513: "Journal of Intelligent Transportation Systems"
               "and Applications",
        51514: "Journal of Science and Technological Researches",
        51515: "Academy of Music, University of Sarajevo",
        51516: "Israa University Journal for Scientific Conferences",
        51517: "Universitas Tamansiswa Palembang",
        51518: "Jurnal Lentera:Jurnal studi Pendidikan, Yayasan"
               "Pembina Lembaga Pendidikan (YPLP) PGRI Sumbawa Barat",
        51519: "Asosiasi Perguruan Tinggi Informatika dan"
               "Komputer (APTIKOM) Sumsel",
        51520: "RAS Publishers LLC",
        51521: "Boston Science Publishing",
        51522: "United Edition of the Federal Penitentiary Service of Russia",
        51523: "Gomel State Medical University",
        51524: "Uluslararasi Hukuk ve Sosyal Bilim Arastirmalari Dergisi",
        51525: "Journal of Hospitality and Tourism Issues",
        51526: "Knowledge-based Engineering and Sciences",
        51527: "Sudanese Online Research Journal",
        51528: "Revista Publicando Publishing House",
        51529: "Perkumpulan Program Studi Pendidikan Islam"
               "Anak Usia Dini Indonesia",
        51530: "Sekolah Tinggi Manajemen Informatika dan Komputer DCI",
        51531: "Korkut Ata Turkiyat Arastirmalari Dergisi",
        51532: "Fruit Research Institute Egirdir Isparta Turkey",
        51533: "Insan Bilimleri Dergisi, Ondokuz Mayis University",
        51534: "Turkiye Insansiz Hava Araclari Dergisi",
        51535: "Journal of Teacher Education and Lifelong Learning",
        51536: "Ankara Yildirim Beyazit Universitesi Turkiye"
               "Saglik Bilimleri ve Arastirmalari Dergisi",
        51537: "Akif Akgul",
        51538: "Deniz Bedir",
        51539: "Avrasya Arastirma Gelistirme Bilim ve Teknoloji"
               "Merkezi Limited Sirketi",
        51540: "International Journal of Filologia",
        51541: "Nicel Bilimler Dergisi based at ESOGU Istatistik"
               "Danismanlik Uygulama ve Arastirma Merkezi",
        51542: "International Journal of Scientific Advances",
        51543: "Jurnal Pembelajaran Sastra, HISKI Komisariat Malang",
        51544: "Universitas Sari Mutiara Indonesia",
        51545: "Katanov Khakass State University Bulletin,"
               "Katanov Khakass State University",
        51546: "The Journal of School and University Medicine",
        51547: "Foundation of Public Initiatives",
        51548: "Journal of Communication Technology",
        51549: "Journal of Recreation and Landscape, Jeonbuk"
               "National University",
        51550: "Noble Academic Publisher",
        51551: "Stratejik Arastirmalar ve Verimlilik Genel"
               "Mudurlugu Verimlilik Dergisi",
        51552: "Turkish Journal of Landscape Research",
        51553: "Journal of Theology, Bozok University",
        51554: "The Institute of Lithuanian Literature and Folklore",
        51555: "International Conference on Aesthetics and"
               "the Sciences of Art",
        51556: "Perhimpunan Pakar Gizi Dan Pangan (Pergizi Pangan) DPD NTT",
        51557: "Fakultas Teknik, Universitas Andi Djemma",
        51558: "University of Tuzla",
        51559: "Sanglah Hospital - Udayana University",
        51560: "OpenCFD Ltd",
        51561: "Institute of Psychology, Czech Academy of Sciences",
        51562: "Faculty of Law, Tekirdag Namik Kemal University",
        51563: "CIDER",
        51564: "EGOB",
        51565: "Facultad de Administración",
        51566: "Facultad de Artes y Humanidades",
        51567: "Facultad de Ciencias",
        51568: "Facultad de Ciencias Sociales",
        51569: "Facultad de Economía",
        51570: "Facultad de Educación",
        51571: "Facultad de Ingeniería",
        51572: "Facultad de Medicina",
        51573: "Ediciones Uniandes",
        51574: "Education and Talent Development Center Indonesia"
               "(ETDC Indonesia)",
        51575: "Ankara Sosyal Bilimler Universitesi ATEBE Dergisi",
        51576: "Yegah Musiki Dergisi",
        51577: "Perkumpulan Intelektual Madani Indonesia",
        51578: "Jurnal Sains Teknologi Transportasi Maritim,"
               "Akademi Maritim Suaka Bahari Cirebon",
        51579: "Almaty Academy of Economy and Statistic",
        51580: "A.B. Bekturov Institute of Chemical Sciences",
        51581: "Facultad de Medicina - Dr. Aurelio Melean",
        51582: "Scientific Publishing Center InterConf",
        51583: "International Journal of Latest Technology"
               "in Engineering, Management & Applied Science",
        51584: "International Journal of Research and Innovation"
               "in Applied Science",
        51585: "German Multidisciplinary Publishing Center",
        51586: "Agenda Publishing House Limited",
        51587: "DP Express-obyava",
        51588: "European Association for Architectural Education",
        51589: "Jurnal AgroSainTa: Widyaiswara Membangun Bangsa,"
               "Pusat Pelatihan Pertanian, Kementerian Pertanian",
        51590: "Waraqat: Jurnal Ilmu-Ilmu Keislaman, STAI As-Sunnah",
        51591: "STTII Yogyakarta",
        51592: "KULLIYAT Osmanli Arastirmalari Dergisi",
        51593: "Center for Security and Emerging Technology (CSET)",
        51594: "Fair East Publishers",
        51595: "Resilience Research, Training and Consulting",
        51596: "SPACE Studies of Planning and Architecture Ltd",
        51597: "Clinica Guayaquil",
        51598: "Asociacion Archiveros de Castilla y Leon",
        51599: "Institute of Eastern European Research and Consulting",
        51600: "Asia Social Science Academy",
        51601: "International Journal of Educational Research"
               "& Social Sciences (IJERSC)",
        51602: "LENTERA KAJI",
        51603: "Socio-Cultural Awareness Human & Resources"
               "Advancement Association (SAHARAA), Pakistan",
        51604: "International Journal of Pharmaceutics & Drug Research",
        51605: "Mesned Ilahiyat Arastirmalari Dergisi",
        51606: "Bioenergy Studies, Black Sea Agricultural Research Institute",
        51607: "Veterinarski fakultet, Univerzitet u Sarajevu",
        51608: "Noncommercial Organization - Institution of"
               "Forensic Construction and Technological Expertise",
        51609: "Mordovian State Pedagogical University named"
               "after M. E. Evseviev",
        51610: "Al-Razi University",
        51611: "International Association of Research Scholars",
        51612: "Jurnal Ilmiah Teunuleh",
        51613: "Korean Youth Counseling Association",
        51614: "STIT Darul Fattah Bandar Lampung",
        51615: "Sekolah Tinggi Teologi Yerusalem Baru Manado",
        51616: "Konsorsium Cendekiawan Indonesia",
        51617: "The Korean Association of Rehabilitation and Bankruptcy Law",
        51618: "LLC AGROMASHPRO",
        51619: "Svensk Teologisk Kvartalskrift",
        51620: "EKOlab",
        51621: "Journal of Womens Research-ATAKAD, Ataturk University",
        51622: "Universitas HKBP Nommensen Medan",
        51623: "Nic Art",
        51624: "Szociologiai szemle",
        51625: "Deutscher Psychologen Verlag GmbH",
        51626: "Skeena Publishers",
        51627: "Glohea Inc.",
        51628: "Neurons Behavior Data Analysis and Behavior",
        51629: "HAR PRESS INDONESIA",
        51630: "Asosiasi Peneliti Sriwijaya",
        51631: "International Centre Art and Education",
        51632: "OpenScience, Research Group Omnibus",
        51633: "Center for Support of Horticultural, Horticultural"
               "Societies and Garage Cooperatives",
        51634: "Law and State Journal, M. Narikbayev KAZGUU University",
        51635: "Agency of Advanced Scientific Research",
        51636: "The Korean Association For Thinking Development",
        51637: "Journal of International Museum Education",
        51638: "The Korean Society of Geriatric Neurosurgery",
        51639: "Journal of the NB BSTU named after Shukhov:"
               "MechMath, BSTU n. a. V G Shukhov (Novorossiysk Branch)",
        51640: "MCPHS University Library",
        51641: "Instituto de Investigaciones Ambientales del"
               "Pacifico - John Von Neumann",
        51642: "Postgraduate Medical Institute, Lahore General Hospital",
        51643: "Asociacion Colombiana de Hematologia y Oncologia",
        51644: "Wilfrid Laurier University",
        51645: "Kosuyolu Heart Journal, Health Sciences University",
        51646: "Center for Solar Energy Research and Studies",
        51647: "Fundacja Instytut Spraw Administracji Publicznej",
        51648: "Ayurveda Campus IOM",
        51649: "Ayurveda Medical Development and Research Foundation (AMDRF)",
        51650: "Veleuciliste u Sibeniku",
        51651: "Perkumpulan Kualitama Edukatika Indonesia",
        51652: "Korea Speech, Media & Communication Association",
        51653: "Journal of Teaching History in School",
        51654: "Slavic Greek Latin Academy",
        51655: "E-Journal of Law",
        51656: "Psycause",
        51657: "Revue internationale du CRIRES : innover dans"
               "la tradition de Vygotsky",
        51658: "Annals of Basic and Medical Sciences",
        51659: "Management & Business Academy",
        51660: "Editic",
        51661: "British Association for Chinese Studies",
        51662: "Asosiasi Staf Akademik Perguruan Tinggi Seluruh"
               "Indonesia (ASASI)",
        51663: "Institut za novejso zgodovino",
        51664: "Artium, Hasan Kalyoncu University",
        51665: "Al-Ramzan Trust",
        51666: "Korean Society of Gastrointestinal Surgery",
        51667: "Institut Agama Kristen Negeri Manado",
        51668: "Fundacion VECA",
        51669: "OM Consulting Limited",
        51670: "Associacao Brasileira de Harmonizacao OroFacial",
        51671: "PAIFORI - The Society of Indonesian Exercise Physiology",
        51672: "STAI Serdang Lubuk Pakam",
        51673: "Universitas Nahdlatul Wathan Mataram",
        51674: "Sekolah Tinggi Ilmu Kesehatan Tamalatea Makassar",
        51675: "Institut Agama Islam Nahdlatul Ulama (IAINU) Tuban",
        51676: "Poltorak",
        51677: "Interregional social organization National Tourism Academy",
        51678: "Federal State-Financed Scientific Institution"
               "State Institute for Art Studies",
        51679: "Ophiussa, Revista do Centro de Arqueologia"
               "da Universidade de Lisboa",
        51680: "Ekonomski fakultet u Osijeku",
        51681: "Center for Positive Sexuality",
        51682: "Technova Publications",
        51683: "Deep QuanTech, LLC",
        51684: "First10EM",
        51685: "Journal of Quantitative Description: Digital Media",
        51686: "Hermeneutische Blätter",
        51687: "Hajnal István Kör - Társadalomtörténeti Egyesület",
        51688: "Sekolah Tinggi Teologi Reformed Injili Internasional",
        51689: "Fakultas Ilmu Pendidikan Kristen",
        51690: "Jurnal Ayurveda Medistra, Sekolah Tinggi Ilmu"
               "Kesehatan Medistra Indonesia",
        51691: "Mukhin M.N. Publisher",
        51692: "Association for Business Development and Non-Commercial"
               "Projects Holding-Spectrum (St. Petersburg)",
        51693: "Base de Dados Cientificos da UFPR",
        51694: "Advances in Weed Science",
        51695: "Sanare - revista de Politicas Publicas",
        51696: "Revista Semestral de Direito Economico",
        51697: "Graduate School of Economics and Management",
        51698: "Blanquerna - Universitat Ramon Llull",
        51699: "Academic Journal Publishing",
        51700: "STAI Darul Kamal NW Kembang Kerang",
        51701: "Universidad Peruana Los Andes",
        51702: "Eskisehir Osmangazi Universitesi Ilahiyat"
               "Fakultesi Dergisi at Eskisehir Osmangazi University",
        51703: "Dili Institute of Technology",
        51704: "Czech Journal of Civil Engineering",
        51705: "Annamalai University",
        51706: "Khortytsia National Educational Rehabilitation Academy",
        51707: "Junior Academy of Sciences of Ukraine",
        51708: "Suryasa and Sons",
        51709: "Shaheed Benazir Bhutto Women University Peshawar, Pakistan",
        51710: "Indian Association of Sedimentologists",
        51711: "Research Institute for Subtropical Agriculture"
               "and Biotechnology, Jeju National University",
        51712: "STIKES Bhakti Husada Bengkulu",
        51713: "Triatma Mulya University",
        51714: "Universitas Panca Sakti Bekasi",
        51715: "ID Solutions Publisher",
        51716: "Tashwirul Afkar, Lembaga Kajian dan Pengembangan"
               "Sumberdaya Manusia, Pengurus Besar Nahdlatul Ulama",
        51717: "STIMIK Bina Bangsa",
        51718: "International Anesthesia Research Society",
        51719: "Gazi Husrev-begova biblioteka Sarajevo",
        51720: "Management Control Review",
        51721: "Revista Cientifica Interdisciplinar Multiplos Acessos",
        51722: "Revista INGI - Indicacao Geografica e Inovacao",
        51723: "Fundacao de Ensino e Pesquisa em Ciencias da Saude",
        51724: "International Society of Educational Research",
        51725: "Journal of Research in Education and Society",
        51726: "Journal of Language Research",
        51727: "Turkish Business Journal",
        51728: "Islamski pedagoski fakultet Univerziteta u Zenici",
        51729: "STAI Pelabuhanratu",
        51730: "Sekolah Tinggi Teologi Excelsius",
        51731: "Canadian Journal of Health Technologies, CADTH",
        51732: "NUST Journal of Social Sciences & Humanities",
        51733: "Instytut Rozwoju Miast i Regionow",
        51734: "University of Texas Rio Grande Valley, University Library",
        51735: "International Journal of Intelligent Communication,"
               "Computing and Networks",
        51736: "Editorial Tecnocientifica Americana And Sinergia Academica",
        51737: "SUNTEXT REVIEWS LLC",
        51738: "Kultura - Polis",
        51739: "Organisation of Academic Scholars, Institutions,"
               "and Societies (OASIS)",
        51740: "Lietuvos nacionaline Martyno Mazvydo biblioteka",
        51741: "Faculty of Social Work, University of Ljubljana",
        51742: "Universitas Karimun",
        51743: "Fundacion Universitaria Espanola",
        51744: "Centre of Excellence for Development Impact"
               "and Learning (CEDIL)",
        51745: "The North African Journal of Food and Nutrition"
               "Research (NAJFNR)",
        51746: "University of Ljubljana, Faculty of Arts",
        51747: "Universitas Panca Marga",
        51748: "The Global Academia Publisher",
        51749: "Scholar Center",
        51750: "Early Modern Low Countries",
        51751: "Dutch Journal of Applied Linguistics (DuJAL)",
        51752: "Journal of Tourism Economics and Business Studies",
        51753: "Frontiers in Life Sciences and Related Technologies",
        51754: "Journal of Cumhuriyet University Health Sciences"
               "Institute, Sivas Cumhuriyet University",
        51755: "Turkish Veterinary Journal, Sivas Cumhuriyet University",
        51756: "Marine and Life Sciences",
        51757: "International Journal of Epidemiology and Health Sciences",
        51758: "Arabian Gulf University",
        51759: "Federal State Budgetary Scientific Institution"
               "- Federal Beekeeping Research Centre (FSBSI - FBRC)",
        51760: "Vestnik IPB (Vestnik Professionalnyh Buhgalterov),"
               "Institute of professional accountants of Russia (IPAR)",
        51761: "Tuberous Sclerosis Alliance",
        51762: "Ural State Pedagogical University",
        51763: "Department of International Relations, Bandirma"
               "Onyedi Eylul Universitesi",
        51764: "Kirsehir Ahi Evran University",
        51765: "Turkish Journal of Land Management, Mersin University",
        51766: "Moroccan Association for Research and Ethics",
        51767: "The Bhopal School of Social Sciences",
        51768: "Society for Human Transformation and Research",
        51769: "BMGN - Low Countries Historical Review, KNHG",
        51770: "Sekolah Tinggi Teologi Kibaid",
        51771: "Universitas Haji Sumatera Utara",
        51772: "Ngabari: Jurnal Studi Islam dan Sosial, Institut"
               "Agama Islam Riyadlatul Mujahidin (IAIRM) Ngabar",
        51773: "Association of Muslim Community Development (AMCA)",
        51774: "Sekolah Tinggi Ilmu Ekonomi Yapan",
        51775: "Bulletin of Almaty university of Power Engineering"
               "and Telecommunications",
        51776: "Skvortsovia. International Journal of Salicology"
               "and Plant Biology. Institute Botanic Garden UB RAS",
        51777: "RELIEF - revue electronique de litterature"
               "francaise, Radboud University",
        51778: "Revista Mais Educacao",
        51779: "Editora Blimunda",
        51780: "ALFAHELIX",
        51781: "Editora Colab",
        51782: "University Oran 1, Ahmed BENBELLA, Algeria",
        51783: "Tarsadalomtudomanyi kutatokozpont",
        51784: "Igazsagugyi Miniszterium",
        51785: "Journal of Aviation Research, Maltepe University",
        51786: "Korean Society for French History",
        51787: "International Science and Technology Publishing Pte.Ltd.",
        51788: "Tashkent State University of Law",
        51789: "Korean Society of CardioMetabolic Syndrome",
        51790: "Scientific Research Institute of System Analysis",
        51791: "Nigerian Society for Animal Production (NSAP)",
        51792: "Universitas Selamat Sri",
        51793: "Open Systems Publications",
        51794: "Institute of Animal Husbandry Mechanization"
               "- Branch of FS Agroengineering Center VIM",
        51795: "Pedro & Joao Editores LTDA M.E.",
        51796: "Escola de Enfermagem de Ribeirão Preto",
        51797: "Interativa Design",
        51798: "Sapienza: International Journal of Interdisciplinary Studies",
        51799: "EMAIS Editora",
        51800: "Ege Universitesi Ege Cografya Dergisi",
        51801: "Turkish Journal of Range and Forage Science,"
               "Society of Range and Forage Science",
        51802: "Istanbul Journal of Arabic Studies (ISTANBULJAS)",
        51803: "Yildiz Social Science Review, Yildiz Technical University",
        51804: "Universitas Maarif Hasyim Latif Sidoarjo",
        51805: "Sekolah Tinggi Ekonomi Bisnis Syariah Bina Mandiri",
        51806: "Institut Agama Islam Hamzanwadi NW Lombok Timur",
        51807: "Maaseudun uusi aika ry",
        51808: "Suomen Muinaismuistoyhdistys",
        51809: "Tutkijaliitto ry",
        51810: "Poliittisen talouden tutkimuksen seura",
        51811: "Suomen kasvatuksen ja koulutuksen historian seura",
        51812: "BirdLife Finland",
        51813: "Siirtolaisuusinstituutti",
        51814: "Uusfilologinen yhdistys",
        51815: "Society of Social and Economic Research in"
               "the Universities of Turku",
        51816: "Suomen musiikkitieteellinen seura",
        51817: "Perkumpulan Pengelola Jurnal Bahasa dan Sastra"
               "Indonesia serta Pengajarannya (PPJBSIP)",
        51818: "Salesian Journal of Humanities and Social"
               "Sciences, Salesian College",
        51819: "National Association of Social and Applied"
               "Gerontology (NASAG)",
        51820: "Departamentul de Educatie din cadrul Bisericii"
               "Baptiste Maranata din Timisoara",
        51821: "Acta Gastro Enterologica Belgica",
        51822: "Izdatelstvo Avers",
        51823: "Expert Center for Arctic Development PORA",
        51825: "Law Faculty, Universitas Sultan Ageng Tirtayasa",
        51826: "Universitas Kapuas Sintang",
        51827: "Akademi Pariwisata dan Perhotelan Darma Agung",
        51828: "Sekolah Tinggi Teologi Studi Alkitab untuk"
               "Pengembangan Pedesaan Indonesia",
        51829: "Drassana, Museu Maritim de Barcelona",
        51830: "Journal of Urban Learning, Teaching, and Research",
        51831: "Northern Institute (branch) All-Russian State"
               "University of Justice",
        51832: "LLC Scientific and Consulting Center",
        51833: "Publishing House Practica Inc.",
        51834: "Granat Publisher",
        51835: "Ikatan Ahli Geologi Indonesia",
        51836: "STKIP PGRI Sidoarjo",
        51837: "The Islamic University College Journal",
        51838: "Journal El-Bahith in Human and Social Sciences",
        51839: "Journal of Land Law and the Environment",
        51840: "Algerian Journal of Law and Political Sciences",
        51841: "Tobna Journal for Academic Scientific Studies",
        51842: "Journal of Business Administration and Economic Studies",
        51843: "NCSL International",
        51844: "Scientific notes, Khujand State University"
               "named after academician B. Gafurov",
        51845: "National Association of Scholars",
        51846: "The University of Lahore",
        51847: "Polaris",
        51848: "Thematics Journals Publication PVT. LTD.",
        51849: "Jurnal Penelitian Pendidikan, Psikologi Dan"
               "Kesehatan (J-P3K)",
        51850: "Masyarakat Peneliti Kayu Indonesia",
        51851: "STIKes Sapta Bakti",
        51852: "Politeknik Pembangunan Pertanian Bogor",
        51853: "Journal of Research in Data Science",
        51854: "The Ben Gurion University of the Negev",
        51855: "Korean Oral History Association",
        51856: "Federal state budget institution - Oksky state"
               "natural biosphere zapovednik",
        51857: "Medical Dental Institute",
        51858: "K·Business Management Research Institute (KBMRI)",
        51859: "Editora Amplla",
        51860: "Revista Conatus - Filosofia de Spinoza",
        51861: "Desenvolvimento em Debate",
        51862: "Fundacion Observatorio Multidisciplinario"
               "para la Construccion del Conocimiento",
        51863: "Universitat de Valencia, Proyecto Parnaseo",
        51864: "Editora dos Editores Eireli",
        51865: "Universitatea Petrol-Gaze din Ploiesti",
        51866: "Academy of Family Physicians of Malaysia",
        51867: "AJER Publishing",
        51868: "Stanford Computational Antitrust",
        51869: "Association for Educational Communication and Technology",
        51870: "Metropolitan University Prague",
        51871: "North-Caucasian Federal Scientific And Clinical"
               "Center Of The Federal Medical And Biological Agency",
        51872: "Progressive Research Journal of Arts & Humanities,"
               "Sarangaa Literary & Cultural Society (SLCS)",
        51873: "Journal of Holistic and Health Science, STIKes HOLISTIK",
        51874: "PGRI Kota Semarang",
        51875: "Institut Daarul Quran",
        51876: "STMIK Catur Sakti Kendari",
        51877: "Sekolah Tinggi Ilmu Ekonomi Rahmaniyah",
        51878: "Pusat Pengembangan Pendidikan dan Penelitian Indonesia",
        51879: "Research Walkers",
        51880: "Historia Oral",
        51881: "STIE Indonesia Malang",
        51882: "Jurnal Akuntansi Manajemen Madani at STIE Madani Balikpapan",
        51883: "Academy of Pedagogical Sciences",
        51884: "Bulletin of Tajik National University",
        51885: "D. Serikbayev East Kazakhstan Technical University",
        51886: "Kazakh Research Institute of Soil Science"
               "and Agrochemistry named after UU. Uspanov",
        51887: "Jurnal Penelitian Farmasi Indonesia, Sekolah"
               "Tinggi Ilmu Farmasi Riau",
        51888: "Fakultas Kesehatan Masyarakat Univeristas Tompotika Luwuk",
        51889: "Abai Kazakh National Pedagogical University",
        51890: "Gazprom Neft",
        51891: "Revista Ibero-Americana de Humanidades, Ciencias e Educacao",
        51892: "Youth STEM 2030",
        51893: "College Of Intensive Care Medicine Of Australia"
               "And New Zealand",
        51894: "Michigan State University College of Osteopathic"
               "Medicine Statewide Campus System",
        51895: "Georgian National University SEU",
        51896: "Servicios Academicos Intercontinentales S.L.",
        51897: "InterAlia: A Journal of Queer Studies",
        51898: "STIKES Bahrul Ulum Jombang",
        51899: "The Indonesian Association of Women, Gender & Child Studies",
        51900: "Ushuluddin dan Studi Islam, UIN Sumatera Utara",
        51901: "Sekolah Tinggi Teknologi Ronggolawe Cepu",
        51902: "Jurnal Providensi, STT Anugerah Sinagoge Medan",
        51903: "Universitas Sains dan Teknologi Komputer",
        51904: "Innovation and Experiment in Education LLC",
        51905: "Non-Profit Enterprise - Editorial Board of"
               "the Journal Mir Peremen",
        51906: "Publishing House Primary School and Education",
        51907: "National Institute of Professional Accountants,"
               "Financial Managers and Economists",
        51908: "Publishing house Saratovsky source",
        51909: "Revista de Ensino Ciencia e Inovacao em Saude (RECIS)",
        51910: "International Journal of High Dilution Research - IJHDR",
        51911: "Editora OESTE",
        51912: "O Alferes",
        51913: "Revista de Educacao Linguagem e Literatura - REVELLI",
        51914: "Revista Brasileira de Musicoterapia",
        51915: "Revista Espanola de la Transparencia, ACREDITRA",
        51916: "Aksh – The Advance Journal",
        51917: "Dialogo Publishing House",
        51918: "Central-European Horizons",
        51919: "Programa de Historia das Ciencias e das Tecnicas"
               "e Epistemologia",
        51920: "tanzilmultazam@umsida.ac.id",
        51921: "Majalah Ilmiah Cakrawala Hukum Fakultas Hukum"
               "Universitas Wijayakusuma Purwokerto",
        51922: "Belarusian State Medical University",
        51923: "REPAE - Revista de Ensino e Pesquisa em Administracao"
               "e Engenharia",
        51924: "Revista Thesis",
        51925: "Yayasan Islam Nusantara Center",
        51926: "ISTE Group",
        51927: "Universidad Autonoma Espana de Durango",
        51928: "Academic Publishing",
        51929: "The Brazilian Society of Anatomy (SBA)",
        51930: "University of Baghdad - College of Islamic Sciences",
        51931: "Open Access Journal of Clinical Surgery",
        51932: "CAD/EDA, MODELING AND SIMULATION IN MODERN"
               "ELECTRONICS, Bryansk State Technical University",
        51933: "Universitas Aufa Royhan",
        51934: "Journal of Medical Innovation and Technology,"
               "Eskisehir Osmangazi University",
        51935: "Faculty of Electrical Engineering, University of Ljubljana",
        51936: "Faculty of Social Sciences, University of Ljubljana",
        51937: "Academy for Theatre, Radio, Film and Television,"
               "University of Ljubljana",
        51938: "Academy of Fine Arts and Design, University of Ljubljana",
        51939: "Faculty of Computer and Information Science,"
               "University of Ljubljana",
        51940: "Faculty of Law, University of Ljubljana",
        51941: "Archives of Medical and Clinical Research",
        51942: "Tula State Museum of Weapons",
        51943: "Ch. Valikhanov Institute of History and Ethnology",
        51944: "Moscow Psychological and Social University",
        51945: "Journal of Cukurova University Faculty of"
               "Economics and Administrative Sciences, Cukurova University",
        51946: "Turkiye lidar dergisi (Mersin University)",
        51947: "Uluslararasi Akademik Yonetim Bilimleri Dergisi",
        51948: "Acikogretim Uygulamalari ve Arastirmalari"
               "Dergisi, Anadolu University",
        51949: "Revista Querubim - Letras - Ciencias Humanas"
               "- Ciencias Sociais",
        51950: "CCS Grafica Editora Comercio o Servico Ltda",
        51951: "Travessias Interativas",
        51952: "Bristol University Press",
        51953: "Revista de Direito Notarial",
        51954: "Bureau of Historical and Cultural Expertise",
        51955: "Moscow State Technical University of Civil Aviation",
        51956: "Emres Publishers Private Limited",
        51957: "Indian Association of Preventive and Social"
               "Medicine - Gujarat Chapter",
        51958: "Al Ain University",
        51959: "Current Bioscience",
        51960: "Journal of Instructional Technologies and"
               "Teacher Education, Karadeniz Technical University",
        51961: "Upper-Volga Federal Agricultural Research Center",
        51962: "Basis, Volga Region Cooperative Institute"
               "(Branch) of Russian University of Cooperation",
        51963: "Gazi Entomolojik Arastirmalar Dernegi",
        51964: "International Institute of Social History",
        51965: "Vestnik of Volzhsky University named after V. N. Tatishchev",
        51966: "Kerala Veterinary and Animal Sciences University",
        51967: "Politeknik Pertanian Negeri Samarinda",
        51968: "Navida Editora",
        51969: "Kirklareli Universitesi Sosyal Bilimler Meslek"
               "Yuksekokulu Dergisi",
        51970: "Turkish Federation of Animal Science (Zootekni Federasyonu)",
        51971: "Kemala Indonesia Publisher",
        51972: "Turk Fen Ve Saglik Dergisi",
        51973: "Journal of Halal and Ethical Research, Istanbul"
               "Sabahattin Zaim University",
        51974: "Instytut Polnocny im. W. Ketrzynskiego",
        51975: "The Nigerian Society of Chemical Engineers",
        51976: "Institute of Business Management, GLA University",
        51977: "Universitas Adhirajasa Reswara Sanjaya",
        51978: "Politeknik Pertanian Negeri Pangkajene Kepulauan",
        51979: "Korean Society of Sport and Leisure Studies",
        51980: "Siberian Law Institute of the MIA of Russia",
        51981: "Polar Geophysical Institute",
        51982: "Ozel Bagimlilik Ltd.",
        51983: "The Research Publication",
        51984: "Sebha University",
        51985: "Bahria University Medical and Dental College",
        51986: "Education Research and Rural Community Development Forum",
        51987: "Instituto Universitario Hospital Italiano de Buenos Aires",
        51988: "Jurnal Teknik Sipil Cendekia, Universitas Winaya Mukti",
        51989: "Kyiv Regional Center of the National Academy"
               "of Legal Sciences of Ukraine",
        51990: "Journal of Al-Farahidi Arts",
        51991: "Management and Development Research and Studies Review",
        51992: "Journal of Research and Economic Studies",
        51993: "Es-saoura journal for human and social studies",
        51994: "Mediterranean International University Journal",
        51995: "Revista Intercontinental de Gestao Desportiva",
        51996: "Imprensa Universitaria da Universidade Federal do Ceara",
        51997: "Sekolah Tinggi Ilmu Kesehatan (STIKes) YPIB Majalengka",
        51998: "STMIK Antar Bangsa",
        51999: "Sekolah Tinggi Teknologi Mutu Muhammadiyah",
        52000: "Inovbook Publications",
        52001: "Tekirdag Viticulture Research Institute",
        52002: "FSBI SRC Planeta",
        52003: "Erzurum Technical University Faculty of Letters"
               "History Journal",
        52004: "Journal of the European Mosquito Control Association",
        88888: "Crossref Test",
    },
}

def org(s: DOI | Sequence[int],
        d: dict = ORGANIZATION) -> str | dict | None:
    k = s.prefix() if isinstance(s, DOI) else s
    while k:
        if k[0] not in d:
            return None
        d = d[k[0]]
        k = k[1:]
    return d
